<h3>{{ project.model.name }}</h3>
<p>Connected to <code>{{ project.backend }}</code></p>
{% if project.model.description %}
<p>{{ project.model.description }}</p>
{% endif %}
<p>{{ project.model.metrics | length }} metrics, {{ project.model.dimensions | length }} dimensions</p>
