PRAGMA foreign_keys=OFF;
BEGIN TRANSACTION;
CREATE TABLE IF NOT EXISTS "albums"
(
    [AlbumId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [Title] NVARCHAR(160)  NOT NULL,
    [ArtistId] INTEGER  NOT NULL,
    FOREIGN KEY ([ArtistId]) REFERENCES "artists" ([ArtistId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION
);
INSERT INTO albums VALUES(1,'For Those About To Rock We Salute You',1);
INSERT INTO albums VALUES(2,'Balls to the Wall',2);
INSERT INTO albums VALUES(3,'Restless and Wild',2);
INSERT INTO albums VALUES(4,'Let There Be Rock',1);
INSERT INTO albums VALUES(5,'Big Ones',3);
INSERT INTO albums VALUES(6,'Jagged Little Pill',4);
INSERT INTO albums VALUES(7,'Facelift',5);
INSERT INTO albums VALUES(8,'Warner 25 Anos',6);
INSERT INTO albums VALUES(9,'Plays Metallica By Four Cellos',7);
INSERT INTO albums VALUES(10,'Audioslave',8);
INSERT INTO albums VALUES(11,'Out Of Exile',8);
INSERT INTO albums VALUES(12,'BackBeat Soundtrack',9);
INSERT INTO albums VALUES(13,'The Best Of Billy Cobham',10);
INSERT INTO albums VALUES(14,'Alcohol Fueled Brewtality Live! [Disc 1]',11);
INSERT INTO albums VALUES(15,'Alcohol Fueled Brewtality Live! [Disc 2]',11);
INSERT INTO albums VALUES(16,'Black Sabbath',12);
INSERT INTO albums VALUES(17,'Black Sabbath Vol. 4 (Remaster)',12);
INSERT INTO albums VALUES(18,'Body Count',13);
INSERT INTO albums VALUES(19,'Chemical Wedding',14);
INSERT INTO albums VALUES(20,'The Best Of Buddy Guy - The Millenium Collection',15);
INSERT INTO albums VALUES(21,'Prenda Minha',16);
INSERT INTO albums VALUES(22,'Sozinho Remix Ao Vivo',16);
INSERT INTO albums VALUES(23,'Minha Historia',17);
INSERT INTO albums VALUES(24,'Afrociberdelia',18);
INSERT INTO albums VALUES(25,'Da Lama Ao Caos',18);
INSERT INTO albums VALUES(26,'Acústico MTV [Live]',19);
INSERT INTO albums VALUES(27,'Cidade Negra - Hits',19);
INSERT INTO albums VALUES(28,'Na Pista',20);
INSERT INTO albums VALUES(29,'Axé Bahia 2001',21);
INSERT INTO albums VALUES(30,'BBC Sessions [Disc 1] [Live]',22);
INSERT INTO albums VALUES(31,'Bongo Fury',23);
INSERT INTO albums VALUES(32,'Carnaval 2001',21);
INSERT INTO albums VALUES(33,'Chill: Brazil (Disc 1)',24);
INSERT INTO albums VALUES(34,'Chill: Brazil (Disc 2)',6);
INSERT INTO albums VALUES(35,'Garage Inc. (Disc 1)',50);
INSERT INTO albums VALUES(36,'Greatest Hits II',51);
INSERT INTO albums VALUES(37,'Greatest Kiss',52);
INSERT INTO albums VALUES(38,'Heart of the Night',53);
INSERT INTO albums VALUES(39,'International Superhits',54);
INSERT INTO albums VALUES(40,'Into The Light',55);
INSERT INTO albums VALUES(41,'Meus Momentos',56);
INSERT INTO albums VALUES(42,'Minha História',57);
INSERT INTO albums VALUES(43,'MK III The Final Concerts [Disc 1]',58);
INSERT INTO albums VALUES(44,'Physical Graffiti [Disc 1]',22);
INSERT INTO albums VALUES(45,'Sambas De Enredo 2001',21);
INSERT INTO albums VALUES(46,'Supernatural',59);
INSERT INTO albums VALUES(47,'The Best of Ed Motta',37);
INSERT INTO albums VALUES(48,'The Essential Miles Davis [Disc 1]',68);
INSERT INTO albums VALUES(49,'The Essential Miles Davis [Disc 2]',68);
INSERT INTO albums VALUES(50,'The Final Concerts (Disc 2)',58);
INSERT INTO albums VALUES(51,'Up An'' Atom',69);
INSERT INTO albums VALUES(52,'Vinícius De Moraes - Sem Limite',70);
INSERT INTO albums VALUES(53,'Vozes do MPB',21);
INSERT INTO albums VALUES(54,'Chronicle, Vol. 1',76);
INSERT INTO albums VALUES(55,'Chronicle, Vol. 2',76);
INSERT INTO albums VALUES(56,'Cássia Eller - Coleção Sem Limite [Disc 2]',77);
INSERT INTO albums VALUES(57,'Cássia Eller - Sem Limite [Disc 1]',77);
INSERT INTO albums VALUES(58,'Come Taste The Band',58);
INSERT INTO albums VALUES(59,'Deep Purple In Rock',58);
INSERT INTO albums VALUES(60,'Fireball',58);
INSERT INTO albums VALUES(61,'Knocking at Your Back Door: The Best Of Deep Purple in the 80''s',58);
INSERT INTO albums VALUES(62,'Machine Head',58);
INSERT INTO albums VALUES(63,'Purpendicular',58);
INSERT INTO albums VALUES(64,'Slaves And Masters',58);
INSERT INTO albums VALUES(65,'Stormbringer',58);
INSERT INTO albums VALUES(66,'The Battle Rages On',58);
INSERT INTO albums VALUES(67,'Vault: Def Leppard''s Greatest Hits',78);
INSERT INTO albums VALUES(68,'Outbreak',79);
INSERT INTO albums VALUES(69,'Djavan Ao Vivo - Vol. 02',80);
INSERT INTO albums VALUES(70,'Djavan Ao Vivo - Vol. 1',80);
INSERT INTO albums VALUES(71,'Elis Regina-Minha História',41);
INSERT INTO albums VALUES(72,'The Cream Of Clapton',81);
INSERT INTO albums VALUES(73,'Unplugged',81);
INSERT INTO albums VALUES(74,'Album Of The Year',82);
INSERT INTO albums VALUES(75,'Angel Dust',82);
INSERT INTO albums VALUES(76,'King For A Day Fool For A Lifetime',82);
INSERT INTO albums VALUES(77,'The Real Thing',82);
INSERT INTO albums VALUES(78,'Deixa Entrar',83);
INSERT INTO albums VALUES(79,'In Your Honor [Disc 1]',84);
INSERT INTO albums VALUES(80,'In Your Honor [Disc 2]',84);
INSERT INTO albums VALUES(81,'One By One',84);
INSERT INTO albums VALUES(82,'The Colour And The Shape',84);
INSERT INTO albums VALUES(83,'My Way: The Best Of Frank Sinatra [Disc 1]',85);
INSERT INTO albums VALUES(84,'Roda De Funk',86);
INSERT INTO albums VALUES(85,'As Canções de Eu Tu Eles',27);
INSERT INTO albums VALUES(86,'Quanta Gente Veio Ver (Live)',27);
INSERT INTO albums VALUES(87,'Quanta Gente Veio ver--Bônus De Carnaval',27);
INSERT INTO albums VALUES(88,'Faceless',87);
INSERT INTO albums VALUES(89,'American Idiot',54);
INSERT INTO albums VALUES(90,'Appetite for Destruction',88);
INSERT INTO albums VALUES(91,'Use Your Illusion I',88);
INSERT INTO albums VALUES(92,'Use Your Illusion II',88);
INSERT INTO albums VALUES(93,'Blue Moods',89);
INSERT INTO albums VALUES(94,'A Matter of Life and Death',90);
INSERT INTO albums VALUES(95,'A Real Dead One',90);
INSERT INTO albums VALUES(96,'A Real Live One',90);
INSERT INTO albums VALUES(97,'Brave New World',90);
INSERT INTO albums VALUES(98,'Dance Of Death',90);
INSERT INTO albums VALUES(99,'Fear Of The Dark',90);
INSERT INTO albums VALUES(100,'Iron Maiden',90);
INSERT INTO albums VALUES(101,'Killers',90);
INSERT INTO albums VALUES(102,'Live After Death',90);
INSERT INTO albums VALUES(103,'Live At Donington 1992 (Disc 1)',90);
INSERT INTO albums VALUES(104,'Live At Donington 1992 (Disc 2)',90);
INSERT INTO albums VALUES(105,'No Prayer For The Dying',90);
INSERT INTO albums VALUES(106,'Piece Of Mind',90);
INSERT INTO albums VALUES(107,'Powerslave',90);
INSERT INTO albums VALUES(108,'Rock In Rio [CD1]',90);
INSERT INTO albums VALUES(109,'Rock In Rio [CD2]',90);
INSERT INTO albums VALUES(110,'Seventh Son of a Seventh Son',90);
INSERT INTO albums VALUES(111,'Somewhere in Time',90);
INSERT INTO albums VALUES(112,'The Number of The Beast',90);
INSERT INTO albums VALUES(113,'The X Factor',90);
INSERT INTO albums VALUES(114,'Virtual XI',90);
INSERT INTO albums VALUES(115,'Sex Machine',91);
INSERT INTO albums VALUES(116,'Emergency On Planet Earth',92);
INSERT INTO albums VALUES(117,'Synkronized',92);
INSERT INTO albums VALUES(118,'The Return Of The Space Cowboy',92);
INSERT INTO albums VALUES(119,'Get Born',93);
INSERT INTO albums VALUES(120,'Are You Experienced?',94);
INSERT INTO albums VALUES(121,'Surfing with the Alien (Remastered)',95);
INSERT INTO albums VALUES(122,'Jorge Ben Jor 25 Anos',46);
INSERT INTO albums VALUES(123,'Jota Quest-1995',96);
INSERT INTO albums VALUES(124,'Cafezinho',97);
INSERT INTO albums VALUES(125,'Living After Midnight',98);
INSERT INTO albums VALUES(126,'Unplugged [Live]',52);
INSERT INTO albums VALUES(127,'BBC Sessions [Disc 2] [Live]',22);
INSERT INTO albums VALUES(128,'Coda',22);
INSERT INTO albums VALUES(129,'Houses Of The Holy',22);
INSERT INTO albums VALUES(130,'In Through The Out Door',22);
INSERT INTO albums VALUES(131,'IV',22);
INSERT INTO albums VALUES(132,'Led Zeppelin I',22);
INSERT INTO albums VALUES(133,'Led Zeppelin II',22);
INSERT INTO albums VALUES(134,'Led Zeppelin III',22);
INSERT INTO albums VALUES(135,'Physical Graffiti [Disc 2]',22);
INSERT INTO albums VALUES(136,'Presence',22);
INSERT INTO albums VALUES(137,'The Song Remains The Same (Disc 1)',22);
INSERT INTO albums VALUES(138,'The Song Remains The Same (Disc 2)',22);
INSERT INTO albums VALUES(139,'A TempestadeTempestade Ou O Livro Dos Dias',99);
INSERT INTO albums VALUES(140,'Mais Do Mesmo',99);
INSERT INTO albums VALUES(141,'Greatest Hits',100);
INSERT INTO albums VALUES(142,'Lulu Santos - RCA 100 Anos De Música - Álbum 01',101);
INSERT INTO albums VALUES(143,'Lulu Santos - RCA 100 Anos De Música - Álbum 02',101);
INSERT INTO albums VALUES(144,'Misplaced Childhood',102);
INSERT INTO albums VALUES(145,'Barulhinho Bom',103);
INSERT INTO albums VALUES(146,'Seek And Shall Find: More Of The Best (1963-1981)',104);
INSERT INTO albums VALUES(147,'The Best Of Men At Work',105);
INSERT INTO albums VALUES(148,'Black Album',50);
INSERT INTO albums VALUES(149,'Garage Inc. (Disc 2)',50);
INSERT INTO albums VALUES(150,'Kill ''Em All',50);
INSERT INTO albums VALUES(151,'Load',50);
INSERT INTO albums VALUES(152,'Master Of Puppets',50);
INSERT INTO albums VALUES(153,'ReLoad',50);
INSERT INTO albums VALUES(154,'Ride The Lightning',50);
INSERT INTO albums VALUES(155,'St. Anger',50);
INSERT INTO albums VALUES(156,'...And Justice For All',50);
INSERT INTO albums VALUES(157,'Miles Ahead',68);
INSERT INTO albums VALUES(158,'Milton Nascimento Ao Vivo',42);
INSERT INTO albums VALUES(159,'Minas',42);
INSERT INTO albums VALUES(160,'Ace Of Spades',106);
INSERT INTO albums VALUES(161,'Demorou...',108);
INSERT INTO albums VALUES(162,'Motley Crue Greatest Hits',109);
INSERT INTO albums VALUES(163,'From The Muddy Banks Of The Wishkah [Live]',110);
INSERT INTO albums VALUES(164,'Nevermind',110);
INSERT INTO albums VALUES(165,'Compositores',111);
INSERT INTO albums VALUES(166,'Olodum',112);
INSERT INTO albums VALUES(167,'Acústico MTV',113);
INSERT INTO albums VALUES(168,'Arquivo II',113);
INSERT INTO albums VALUES(169,'Arquivo Os Paralamas Do Sucesso',113);
INSERT INTO albums VALUES(170,'Bark at the Moon (Remastered)',114);
INSERT INTO albums VALUES(171,'Blizzard of Ozz',114);
INSERT INTO albums VALUES(172,'Diary of a Madman (Remastered)',114);
INSERT INTO albums VALUES(173,'No More Tears (Remastered)',114);
INSERT INTO albums VALUES(174,'Tribute',114);
INSERT INTO albums VALUES(175,'Walking Into Clarksdale',115);
INSERT INTO albums VALUES(176,'Original Soundtracks 1',116);
INSERT INTO albums VALUES(177,'The Beast Live',117);
INSERT INTO albums VALUES(178,'Live On Two Legs [Live]',118);
INSERT INTO albums VALUES(179,'Pearl Jam',118);
INSERT INTO albums VALUES(180,'Riot Act',118);
INSERT INTO albums VALUES(181,'Ten',118);
INSERT INTO albums VALUES(182,'Vs.',118);
INSERT INTO albums VALUES(183,'Dark Side Of The Moon',120);
INSERT INTO albums VALUES(184,'Os Cães Ladram Mas A Caravana Não Pára',121);
INSERT INTO albums VALUES(185,'Greatest Hits I',51);
INSERT INTO albums VALUES(186,'News Of The World',51);
INSERT INTO albums VALUES(187,'Out Of Time',122);
INSERT INTO albums VALUES(188,'Green',124);
INSERT INTO albums VALUES(189,'New Adventures In Hi-Fi',124);
INSERT INTO albums VALUES(190,'The Best Of R.E.M.: The IRS Years',124);
INSERT INTO albums VALUES(191,'Cesta Básica',125);
INSERT INTO albums VALUES(192,'Raul Seixas',126);
INSERT INTO albums VALUES(193,'Blood Sugar Sex Magik',127);
INSERT INTO albums VALUES(194,'By The Way',127);
INSERT INTO albums VALUES(195,'Californication',127);
INSERT INTO albums VALUES(196,'Retrospective I (1974-1980)',128);
INSERT INTO albums VALUES(197,'Santana - As Years Go By',59);
INSERT INTO albums VALUES(198,'Santana Live',59);
INSERT INTO albums VALUES(199,'Maquinarama',130);
INSERT INTO albums VALUES(200,'O Samba Poconé',130);
INSERT INTO albums VALUES(201,'Judas 0: B-Sides and Rarities',131);
INSERT INTO albums VALUES(202,'Rotten Apples: Greatest Hits',131);
INSERT INTO albums VALUES(203,'A-Sides',132);
INSERT INTO albums VALUES(204,'Morning Dance',53);
INSERT INTO albums VALUES(205,'In Step',133);
INSERT INTO albums VALUES(206,'Core',134);
INSERT INTO albums VALUES(207,'Mezmerize',135);
INSERT INTO albums VALUES(208,'[1997] Black Light Syndrome',136);
INSERT INTO albums VALUES(209,'Live [Disc 1]',137);
INSERT INTO albums VALUES(210,'Live [Disc 2]',137);
INSERT INTO albums VALUES(211,'The Singles',138);
INSERT INTO albums VALUES(212,'Beyond Good And Evil',139);
INSERT INTO albums VALUES(213,'Pure Cult: The Best Of The Cult (For Rockers, Ravers, Lovers & Sinners) [UK]',139);
INSERT INTO albums VALUES(214,'The Doors',140);
INSERT INTO albums VALUES(215,'The Police Greatest Hits',141);
INSERT INTO albums VALUES(216,'Hot Rocks, 1964-1971 (Disc 1)',142);
INSERT INTO albums VALUES(217,'No Security',142);
INSERT INTO albums VALUES(218,'Voodoo Lounge',142);
INSERT INTO albums VALUES(219,'Tangents',143);
INSERT INTO albums VALUES(220,'Transmission',143);
INSERT INTO albums VALUES(221,'My Generation - The Very Best Of The Who',144);
INSERT INTO albums VALUES(222,'Serie Sem Limite (Disc 1)',145);
INSERT INTO albums VALUES(223,'Serie Sem Limite (Disc 2)',145);
INSERT INTO albums VALUES(224,'Acústico',146);
INSERT INTO albums VALUES(225,'Volume Dois',146);
INSERT INTO albums VALUES(226,'Battlestar Galactica: The Story So Far',147);
INSERT INTO albums VALUES(227,'Battlestar Galactica, Season 3',147);
INSERT INTO albums VALUES(228,'Heroes, Season 1',148);
INSERT INTO albums VALUES(229,'Lost, Season 3',149);
INSERT INTO albums VALUES(230,'Lost, Season 1',149);
INSERT INTO albums VALUES(231,'Lost, Season 2',149);
INSERT INTO albums VALUES(232,'Achtung Baby',150);
INSERT INTO albums VALUES(233,'All That You Can''t Leave Behind',150);
INSERT INTO albums VALUES(234,'B-Sides 1980-1990',150);
INSERT INTO albums VALUES(235,'How To Dismantle An Atomic Bomb',150);
INSERT INTO albums VALUES(236,'Pop',150);
INSERT INTO albums VALUES(237,'Rattle And Hum',150);
INSERT INTO albums VALUES(238,'The Best Of 1980-1990',150);
INSERT INTO albums VALUES(239,'War',150);
INSERT INTO albums VALUES(240,'Zooropa',150);
INSERT INTO albums VALUES(241,'UB40 The Best Of - Volume Two [UK]',151);
INSERT INTO albums VALUES(242,'Diver Down',152);
INSERT INTO albums VALUES(243,'The Best Of Van Halen, Vol. I',152);
INSERT INTO albums VALUES(244,'Van Halen',152);
INSERT INTO albums VALUES(245,'Van Halen III',152);
INSERT INTO albums VALUES(246,'Contraband',153);
INSERT INTO albums VALUES(247,'Vinicius De Moraes',72);
INSERT INTO albums VALUES(248,'Ao Vivo [IMPORT]',155);
INSERT INTO albums VALUES(249,'The Office, Season 1',156);
INSERT INTO albums VALUES(250,'The Office, Season 2',156);
INSERT INTO albums VALUES(251,'The Office, Season 3',156);
INSERT INTO albums VALUES(252,'Un-Led-Ed',157);
INSERT INTO albums VALUES(253,'Battlestar Galactica (Classic), Season 1',158);
INSERT INTO albums VALUES(254,'Aquaman',159);
INSERT INTO albums VALUES(255,'Instant Karma: The Amnesty International Campaign to Save Darfur',150);
INSERT INTO albums VALUES(256,'Speak of the Devil',114);
INSERT INTO albums VALUES(257,'20th Century Masters - The Millennium Collection: The Best of Scorpions',179);
INSERT INTO albums VALUES(258,'House of Pain',180);
INSERT INTO albums VALUES(259,'Radio Brasil (O Som da Jovem Vanguarda) - Seleccao de Henrique Amaro',36);
INSERT INTO albums VALUES(260,'Cake: B-Sides and Rarities',196);
INSERT INTO albums VALUES(261,'LOST, Season 4',149);
INSERT INTO albums VALUES(262,'Quiet Songs',197);
INSERT INTO albums VALUES(263,'Muso Ko',198);
INSERT INTO albums VALUES(264,'Realize',199);
INSERT INTO albums VALUES(265,'Every Kind of Light',200);
INSERT INTO albums VALUES(266,'Duos II',201);
INSERT INTO albums VALUES(267,'Worlds',202);
INSERT INTO albums VALUES(268,'The Best of Beethoven',203);
INSERT INTO albums VALUES(269,'Temple of the Dog',204);
INSERT INTO albums VALUES(270,'Carry On',205);
INSERT INTO albums VALUES(271,'Revelations',8);
INSERT INTO albums VALUES(272,'Adorate Deum: Gregorian Chant from the Proper of the Mass',206);
INSERT INTO albums VALUES(273,'Allegri: Miserere',207);
INSERT INTO albums VALUES(274,'Pachelbel: Canon & Gigue',208);
INSERT INTO albums VALUES(275,'Vivaldi: The Four Seasons',209);
INSERT INTO albums VALUES(276,'Bach: Violin Concertos',210);
INSERT INTO albums VALUES(277,'Bach: Goldberg Variations',211);
INSERT INTO albums VALUES(278,'Bach: The Cello Suites',212);
INSERT INTO albums VALUES(279,'Handel: The Messiah (Highlights)',213);
INSERT INTO albums VALUES(280,'The World of Classical Favourites',214);
INSERT INTO albums VALUES(281,'Sir Neville Marriner: A Celebration',215);
INSERT INTO albums VALUES(282,'Mozart: Wind Concertos',216);
INSERT INTO albums VALUES(283,'Haydn: Symphonies 99 - 104',217);
INSERT INTO albums VALUES(284,'Beethoven: Symhonies Nos. 5 & 6',218);
INSERT INTO albums VALUES(285,'A Soprano Inspired',219);
INSERT INTO albums VALUES(286,'Great Opera Choruses',220);
INSERT INTO albums VALUES(287,'Wagner: Favourite Overtures',221);
INSERT INTO albums VALUES(288,'Fauré: Requiem, Ravel: Pavane & Others',222);
INSERT INTO albums VALUES(289,'Tchaikovsky: The Nutcracker',223);
INSERT INTO albums VALUES(290,'The Last Night of the Proms',224);
INSERT INTO albums VALUES(291,'Puccini: Madama Butterfly - Highlights',225);
INSERT INTO albums VALUES(292,'Holst: The Planets, Op. 32 & Vaughan Williams: Fantasies',226);
INSERT INTO albums VALUES(293,'Pavarotti''s Opera Made Easy',227);
INSERT INTO albums VALUES(294,'Great Performances - Barber''s Adagio and Other Romantic Favorites for Strings',228);
INSERT INTO albums VALUES(295,'Carmina Burana',229);
INSERT INTO albums VALUES(296,'A Copland Celebration, Vol. I',230);
INSERT INTO albums VALUES(297,'Bach: Toccata & Fugue in D Minor',231);
INSERT INTO albums VALUES(298,'Prokofiev: Symphony No.1',232);
INSERT INTO albums VALUES(299,'Scheherazade',233);
INSERT INTO albums VALUES(300,'Bach: The Brandenburg Concertos',234);
INSERT INTO albums VALUES(301,'Chopin: Piano Concertos Nos. 1 & 2',235);
INSERT INTO albums VALUES(302,'Mascagni: Cavalleria Rusticana',236);
INSERT INTO albums VALUES(303,'Sibelius: Finlandia',237);
INSERT INTO albums VALUES(304,'Beethoven Piano Sonatas: Moonlight & Pastorale',238);
INSERT INTO albums VALUES(305,'Great Recordings of the Century - Mahler: Das Lied von der Erde',240);
INSERT INTO albums VALUES(306,'Elgar: Cello Concerto & Vaughan Williams: Fantasias',241);
INSERT INTO albums VALUES(307,'Adams, John: The Chairman Dances',242);
INSERT INTO albums VALUES(308,'Tchaikovsky: 1812 Festival Overture, Op.49, Capriccio Italien & Beethoven: Wellington''s Victory',243);
INSERT INTO albums VALUES(309,'Palestrina: Missa Papae Marcelli & Allegri: Miserere',244);
INSERT INTO albums VALUES(310,'Prokofiev: Romeo & Juliet',245);
INSERT INTO albums VALUES(311,'Strauss: Waltzes',226);
INSERT INTO albums VALUES(312,'Berlioz: Symphonie Fantastique',245);
INSERT INTO albums VALUES(313,'Bizet: Carmen Highlights',246);
INSERT INTO albums VALUES(314,'English Renaissance',247);
INSERT INTO albums VALUES(315,'Handel: Music for the Royal Fireworks (Original Version 1749)',208);
INSERT INTO albums VALUES(316,'Grieg: Peer Gynt Suites & Sibelius: Pelléas et Mélisande',248);
INSERT INTO albums VALUES(317,'Mozart Gala: Famous Arias',249);
INSERT INTO albums VALUES(318,'SCRIABIN: Vers la flamme',250);
INSERT INTO albums VALUES(319,'Armada: Music from the Courts of England and Spain',251);
INSERT INTO albums VALUES(320,'Mozart: Symphonies Nos. 40 & 41',248);
INSERT INTO albums VALUES(321,'Back to Black',252);
INSERT INTO albums VALUES(322,'Frank',252);
INSERT INTO albums VALUES(323,'Carried to Dust (Bonus Track Version)',253);
INSERT INTO albums VALUES(324,'Beethoven: Symphony No. 6 ''Pastoral'' Etc.',254);
INSERT INTO albums VALUES(325,'Bartok: Violin & Viola Concertos',255);
INSERT INTO albums VALUES(326,'Mendelssohn: A Midsummer Night''s Dream',256);
INSERT INTO albums VALUES(327,'Bach: Orchestral Suites Nos. 1 - 4',257);
INSERT INTO albums VALUES(328,'Charpentier: Divertissements, Airs & Concerts',258);
INSERT INTO albums VALUES(329,'South American Getaway',259);
INSERT INTO albums VALUES(330,'Górecki: Symphony No. 3',260);
INSERT INTO albums VALUES(331,'Purcell: The Fairy Queen',261);
INSERT INTO albums VALUES(332,'The Ultimate Relexation Album',262);
INSERT INTO albums VALUES(333,'Purcell: Music for the Queen Mary',263);
INSERT INTO albums VALUES(334,'Weill: The Seven Deadly Sins',264);
INSERT INTO albums VALUES(335,'J.S. Bach: Chaconne, Suite in E Minor, Partita in E Major & Prelude, Fugue and Allegro',265);
INSERT INTO albums VALUES(336,'Prokofiev: Symphony No.5 & Stravinksy: Le Sacre Du Printemps',248);
INSERT INTO albums VALUES(337,'Szymanowski: Piano Works, Vol. 1',266);
INSERT INTO albums VALUES(338,'Nielsen: The Six Symphonies',267);
INSERT INTO albums VALUES(339,'Great Recordings of the Century: Paganini''s 24 Caprices',268);
INSERT INTO albums VALUES(340,'Liszt - 12 Études D''Execution Transcendante',269);
INSERT INTO albums VALUES(341,'Great Recordings of the Century - Shubert: Schwanengesang, 4 Lieder',270);
INSERT INTO albums VALUES(342,'Locatelli: Concertos for Violin, Strings and Continuo, Vol. 3',271);
INSERT INTO albums VALUES(343,'Respighi:Pines of Rome',226);
INSERT INTO albums VALUES(344,'Schubert: The Late String Quartets & String Quintet (3 CD''s)',272);
INSERT INTO albums VALUES(345,'Monteverdi: L''Orfeo',273);
INSERT INTO albums VALUES(346,'Mozart: Chamber Music',274);
INSERT INTO albums VALUES(347,'Koyaanisqatsi (Soundtrack from the Motion Picture)',275);
CREATE TABLE IF NOT EXISTS "artists"
(
    [ArtistId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [Name] NVARCHAR(120)
);
INSERT INTO artists VALUES(1,'AC/DC');
INSERT INTO artists VALUES(2,'Accept');
INSERT INTO artists VALUES(3,'Aerosmith');
INSERT INTO artists VALUES(4,'Alanis Morissette');
INSERT INTO artists VALUES(5,'Alice In Chains');
INSERT INTO artists VALUES(6,'Antônio Carlos Jobim');
INSERT INTO artists VALUES(7,'Apocalyptica');
INSERT INTO artists VALUES(8,'Audioslave');
INSERT INTO artists VALUES(9,'BackBeat');
INSERT INTO artists VALUES(10,'Billy Cobham');
INSERT INTO artists VALUES(11,'Black Label Society');
INSERT INTO artists VALUES(12,'Black Sabbath');
INSERT INTO artists VALUES(13,'Body Count');
INSERT INTO artists VALUES(14,'Bruce Dickinson');
INSERT INTO artists VALUES(15,'Buddy Guy');
INSERT INTO artists VALUES(16,'Caetano Veloso');
INSERT INTO artists VALUES(17,'Chico Buarque');
INSERT INTO artists VALUES(18,'Chico Science & Nação Zumbi');
INSERT INTO artists VALUES(19,'Cidade Negra');
INSERT INTO artists VALUES(20,'Cláudio Zoli');
INSERT INTO artists VALUES(21,'Various Artists');
INSERT INTO artists VALUES(22,'Led Zeppelin');
INSERT INTO artists VALUES(23,'Frank Zappa & Captain Beefheart');
INSERT INTO artists VALUES(24,'Marcos Valle');
INSERT INTO artists VALUES(25,'Milton Nascimento & Bebeto');
INSERT INTO artists VALUES(26,'Azymuth');
INSERT INTO artists VALUES(27,'Gilberto Gil');
INSERT INTO artists VALUES(28,'João Gilberto');
INSERT INTO artists VALUES(29,'Bebel Gilberto');
INSERT INTO artists VALUES(30,'Jorge Vercilo');
INSERT INTO artists VALUES(31,'Baby Consuelo');
INSERT INTO artists VALUES(32,'Ney Matogrosso');
INSERT INTO artists VALUES(33,'Luiz Melodia');
INSERT INTO artists VALUES(34,'Nando Reis');
INSERT INTO artists VALUES(35,'Pedro Luís & A Parede');
INSERT INTO artists VALUES(36,'O Rappa');
INSERT INTO artists VALUES(37,'Ed Motta');
INSERT INTO artists VALUES(38,'Banda Black Rio');
INSERT INTO artists VALUES(39,'Fernanda Porto');
INSERT INTO artists VALUES(40,'Os Cariocas');
INSERT INTO artists VALUES(41,'Elis Regina');
INSERT INTO artists VALUES(42,'Milton Nascimento');
INSERT INTO artists VALUES(43,'A Cor Do Som');
INSERT INTO artists VALUES(44,'Kid Abelha');
INSERT INTO artists VALUES(45,'Sandra De Sá');
INSERT INTO artists VALUES(46,'Jorge Ben');
INSERT INTO artists VALUES(47,'Hermeto Pascoal');
INSERT INTO artists VALUES(48,'Barão Vermelho');
INSERT INTO artists VALUES(49,'Edson, DJ Marky & DJ Patife Featuring Fernanda Porto');
INSERT INTO artists VALUES(50,'Metallica');
INSERT INTO artists VALUES(51,'Queen');
INSERT INTO artists VALUES(52,'Kiss');
INSERT INTO artists VALUES(53,'Spyro Gyra');
INSERT INTO artists VALUES(54,'Green Day');
INSERT INTO artists VALUES(55,'David Coverdale');
INSERT INTO artists VALUES(56,'Gonzaguinha');
INSERT INTO artists VALUES(57,'Os Mutantes');
INSERT INTO artists VALUES(58,'Deep Purple');
INSERT INTO artists VALUES(59,'Santana');
INSERT INTO artists VALUES(60,'Santana Feat. Dave Matthews');
INSERT INTO artists VALUES(61,'Santana Feat. Everlast');
INSERT INTO artists VALUES(62,'Santana Feat. Rob Thomas');
INSERT INTO artists VALUES(63,'Santana Feat. Lauryn Hill & Cee-Lo');
INSERT INTO artists VALUES(64,'Santana Feat. The Project G&B');
INSERT INTO artists VALUES(65,'Santana Feat. Maná');
INSERT INTO artists VALUES(66,'Santana Feat. Eagle-Eye Cherry');
INSERT INTO artists VALUES(67,'Santana Feat. Eric Clapton');
INSERT INTO artists VALUES(68,'Miles Davis');
INSERT INTO artists VALUES(69,'Gene Krupa');
INSERT INTO artists VALUES(70,'Toquinho & Vinícius');
INSERT INTO artists VALUES(71,'Vinícius De Moraes & Baden Powell');
INSERT INTO artists VALUES(72,'Vinícius De Moraes');
INSERT INTO artists VALUES(73,'Vinícius E Qurteto Em Cy');
INSERT INTO artists VALUES(74,'Vinícius E Odette Lara');
INSERT INTO artists VALUES(75,'Vinicius, Toquinho & Quarteto Em Cy');
INSERT INTO artists VALUES(76,'Creedence Clearwater Revival');
INSERT INTO artists VALUES(77,'Cássia Eller');
INSERT INTO artists VALUES(78,'Def Leppard');
INSERT INTO artists VALUES(79,'Dennis Chambers');
INSERT INTO artists VALUES(80,'Djavan');
INSERT INTO artists VALUES(81,'Eric Clapton');
INSERT INTO artists VALUES(82,'Faith No More');
INSERT INTO artists VALUES(83,'Falamansa');
INSERT INTO artists VALUES(84,'Foo Fighters');
INSERT INTO artists VALUES(85,'Frank Sinatra');
INSERT INTO artists VALUES(86,'Funk Como Le Gusta');
INSERT INTO artists VALUES(87,'Godsmack');
INSERT INTO artists VALUES(88,'Guns N'' Roses');
INSERT INTO artists VALUES(89,'Incognito');
INSERT INTO artists VALUES(90,'Iron Maiden');
INSERT INTO artists VALUES(91,'James Brown');
INSERT INTO artists VALUES(92,'Jamiroquai');
INSERT INTO artists VALUES(93,'JET');
INSERT INTO artists VALUES(94,'Jimi Hendrix');
INSERT INTO artists VALUES(95,'Joe Satriani');
INSERT INTO artists VALUES(96,'Jota Quest');
INSERT INTO artists VALUES(97,'João Suplicy');
INSERT INTO artists VALUES(98,'Judas Priest');
INSERT INTO artists VALUES(99,'Legião Urbana');
INSERT INTO artists VALUES(100,'Lenny Kravitz');
INSERT INTO artists VALUES(101,'Lulu Santos');
INSERT INTO artists VALUES(102,'Marillion');
INSERT INTO artists VALUES(103,'Marisa Monte');
INSERT INTO artists VALUES(104,'Marvin Gaye');
INSERT INTO artists VALUES(105,'Men At Work');
INSERT INTO artists VALUES(106,'Motörhead');
INSERT INTO artists VALUES(107,'Motörhead & Girlschool');
INSERT INTO artists VALUES(108,'Mônica Marianno');
INSERT INTO artists VALUES(109,'Mötley Crüe');
INSERT INTO artists VALUES(110,'Nirvana');
INSERT INTO artists VALUES(111,'O Terço');
INSERT INTO artists VALUES(112,'Olodum');
INSERT INTO artists VALUES(113,'Os Paralamas Do Sucesso');
INSERT INTO artists VALUES(114,'Ozzy Osbourne');
INSERT INTO artists VALUES(115,'Page & Plant');
INSERT INTO artists VALUES(116,'Passengers');
INSERT INTO artists VALUES(117,'Paul D''Ianno');
INSERT INTO artists VALUES(118,'Pearl Jam');
INSERT INTO artists VALUES(119,'Peter Tosh');
INSERT INTO artists VALUES(120,'Pink Floyd');
INSERT INTO artists VALUES(121,'Planet Hemp');
INSERT INTO artists VALUES(122,'R.E.M. Feat. Kate Pearson');
INSERT INTO artists VALUES(123,'R.E.M. Feat. KRS-One');
INSERT INTO artists VALUES(124,'R.E.M.');
INSERT INTO artists VALUES(125,'Raimundos');
INSERT INTO artists VALUES(126,'Raul Seixas');
INSERT INTO artists VALUES(127,'Red Hot Chili Peppers');
INSERT INTO artists VALUES(128,'Rush');
INSERT INTO artists VALUES(129,'Simply Red');
INSERT INTO artists VALUES(130,'Skank');
INSERT INTO artists VALUES(131,'Smashing Pumpkins');
INSERT INTO artists VALUES(132,'Soundgarden');
INSERT INTO artists VALUES(133,'Stevie Ray Vaughan & Double Trouble');
INSERT INTO artists VALUES(134,'Stone Temple Pilots');
INSERT INTO artists VALUES(135,'System Of A Down');
INSERT INTO artists VALUES(136,'Terry Bozzio, Tony Levin & Steve Stevens');
INSERT INTO artists VALUES(137,'The Black Crowes');
INSERT INTO artists VALUES(138,'The Clash');
INSERT INTO artists VALUES(139,'The Cult');
INSERT INTO artists VALUES(140,'The Doors');
INSERT INTO artists VALUES(141,'The Police');
INSERT INTO artists VALUES(142,'The Rolling Stones');
INSERT INTO artists VALUES(143,'The Tea Party');
INSERT INTO artists VALUES(144,'The Who');
INSERT INTO artists VALUES(145,'Tim Maia');
INSERT INTO artists VALUES(146,'Titãs');
INSERT INTO artists VALUES(147,'Battlestar Galactica');
INSERT INTO artists VALUES(148,'Heroes');
INSERT INTO artists VALUES(149,'Lost');
INSERT INTO artists VALUES(150,'U2');
INSERT INTO artists VALUES(151,'UB40');
INSERT INTO artists VALUES(152,'Van Halen');
INSERT INTO artists VALUES(153,'Velvet Revolver');
INSERT INTO artists VALUES(154,'Whitesnake');
INSERT INTO artists VALUES(155,'Zeca Pagodinho');
INSERT INTO artists VALUES(156,'The Office');
INSERT INTO artists VALUES(157,'Dread Zeppelin');
INSERT INTO artists VALUES(158,'Battlestar Galactica (Classic)');
INSERT INTO artists VALUES(159,'Aquaman');
INSERT INTO artists VALUES(160,'Christina Aguilera featuring BigElf');
INSERT INTO artists VALUES(161,'Aerosmith & Sierra Leone''s Refugee Allstars');
INSERT INTO artists VALUES(162,'Los Lonely Boys');
INSERT INTO artists VALUES(163,'Corinne Bailey Rae');
INSERT INTO artists VALUES(164,'Dhani Harrison & Jakob Dylan');
INSERT INTO artists VALUES(165,'Jackson Browne');
INSERT INTO artists VALUES(166,'Avril Lavigne');
INSERT INTO artists VALUES(167,'Big & Rich');
INSERT INTO artists VALUES(168,'Youssou N''Dour');
INSERT INTO artists VALUES(169,'Black Eyed Peas');
INSERT INTO artists VALUES(170,'Jack Johnson');
INSERT INTO artists VALUES(171,'Ben Harper');
INSERT INTO artists VALUES(172,'Snow Patrol');
INSERT INTO artists VALUES(173,'Matisyahu');
INSERT INTO artists VALUES(174,'The Postal Service');
INSERT INTO artists VALUES(175,'Jaguares');
INSERT INTO artists VALUES(176,'The Flaming Lips');
INSERT INTO artists VALUES(177,'Jack''s Mannequin & Mick Fleetwood');
INSERT INTO artists VALUES(178,'Regina Spektor');
INSERT INTO artists VALUES(179,'Scorpions');
INSERT INTO artists VALUES(180,'House Of Pain');
INSERT INTO artists VALUES(181,'Xis');
INSERT INTO artists VALUES(182,'Nega Gizza');
INSERT INTO artists VALUES(183,'Gustavo & Andres Veiga & Salazar');
INSERT INTO artists VALUES(184,'Rodox');
INSERT INTO artists VALUES(185,'Charlie Brown Jr.');
INSERT INTO artists VALUES(186,'Pedro Luís E A Parede');
INSERT INTO artists VALUES(187,'Los Hermanos');
INSERT INTO artists VALUES(188,'Mundo Livre S/A');
INSERT INTO artists VALUES(189,'Otto');
INSERT INTO artists VALUES(190,'Instituto');
INSERT INTO artists VALUES(191,'Nação Zumbi');
INSERT INTO artists VALUES(192,'DJ Dolores & Orchestra Santa Massa');
INSERT INTO artists VALUES(193,'Seu Jorge');
INSERT INTO artists VALUES(194,'Sabotage E Instituto');
INSERT INTO artists VALUES(195,'Stereo Maracana');
INSERT INTO artists VALUES(196,'Cake');
INSERT INTO artists VALUES(197,'Aisha Duo');
INSERT INTO artists VALUES(198,'Habib Koité and Bamada');
INSERT INTO artists VALUES(199,'Karsh Kale');
INSERT INTO artists VALUES(200,'The Posies');
INSERT INTO artists VALUES(201,'Luciana Souza/Romero Lubambo');
INSERT INTO artists VALUES(202,'Aaron Goldberg');
INSERT INTO artists VALUES(203,'Nicolaus Esterhazy Sinfonia');
INSERT INTO artists VALUES(204,'Temple of the Dog');
INSERT INTO artists VALUES(205,'Chris Cornell');
INSERT INTO artists VALUES(206,'Alberto Turco & Nova Schola Gregoriana');
INSERT INTO artists VALUES(207,'Richard Marlow & The Choir of Trinity College, Cambridge');
INSERT INTO artists VALUES(208,'English Concert & Trevor Pinnock');
INSERT INTO artists VALUES(209,'Anne-Sophie Mutter, Herbert Von Karajan & Wiener Philharmoniker');
INSERT INTO artists VALUES(210,'Hilary Hahn, Jeffrey Kahane, Los Angeles Chamber Orchestra & Margaret Batjer');
INSERT INTO artists VALUES(211,'Wilhelm Kempff');
INSERT INTO artists VALUES(212,'Yo-Yo Ma');
INSERT INTO artists VALUES(213,'Scholars Baroque Ensemble');
INSERT INTO artists VALUES(214,'Academy of St. Martin in the Fields & Sir Neville Marriner');
INSERT INTO artists VALUES(215,'Academy of St. Martin in the Fields Chamber Ensemble & Sir Neville Marriner');
INSERT INTO artists VALUES(216,'Berliner Philharmoniker, Claudio Abbado & Sabine Meyer');
INSERT INTO artists VALUES(217,'Royal Philharmonic Orchestra & Sir Thomas Beecham');
INSERT INTO artists VALUES(218,'Orchestre Révolutionnaire et Romantique & John Eliot Gardiner');
INSERT INTO artists VALUES(219,'Britten Sinfonia, Ivor Bolton & Lesley Garrett');
INSERT INTO artists VALUES(220,'Chicago Symphony Chorus, Chicago Symphony Orchestra & Sir Georg Solti');
INSERT INTO artists VALUES(221,'Sir Georg Solti & Wiener Philharmoniker');
INSERT INTO artists VALUES(222,'Academy of St. Martin in the Fields, John Birch, Sir Neville Marriner & Sylvia McNair');
INSERT INTO artists VALUES(223,'London Symphony Orchestra & Sir Charles Mackerras');
INSERT INTO artists VALUES(224,'Barry Wordsworth & BBC Concert Orchestra');
INSERT INTO artists VALUES(225,'Herbert Von Karajan, Mirella Freni & Wiener Philharmoniker');
INSERT INTO artists VALUES(226,'Eugene Ormandy');
INSERT INTO artists VALUES(227,'Luciano Pavarotti');
INSERT INTO artists VALUES(228,'Leonard Bernstein & New York Philharmonic');
INSERT INTO artists VALUES(229,'Boston Symphony Orchestra & Seiji Ozawa');
INSERT INTO artists VALUES(230,'Aaron Copland & London Symphony Orchestra');
INSERT INTO artists VALUES(231,'Ton Koopman');
INSERT INTO artists VALUES(232,'Sergei Prokofiev & Yuri Temirkanov');
INSERT INTO artists VALUES(233,'Chicago Symphony Orchestra & Fritz Reiner');
INSERT INTO artists VALUES(234,'Orchestra of The Age of Enlightenment');
INSERT INTO artists VALUES(235,'Emanuel Ax, Eugene Ormandy & Philadelphia Orchestra');
INSERT INTO artists VALUES(236,'James Levine');
INSERT INTO artists VALUES(237,'Berliner Philharmoniker & Hans Rosbaud');
INSERT INTO artists VALUES(238,'Maurizio Pollini');
INSERT INTO artists VALUES(239,'Academy of St. Martin in the Fields, Sir Neville Marriner & William Bennett');
INSERT INTO artists VALUES(240,'Gustav Mahler');
INSERT INTO artists VALUES(241,'Felix Schmidt, London Symphony Orchestra & Rafael Frühbeck de Burgos');
INSERT INTO artists VALUES(242,'Edo de Waart & San Francisco Symphony');
INSERT INTO artists VALUES(243,'Antal Doráti & London Symphony Orchestra');
INSERT INTO artists VALUES(244,'Choir Of Westminster Abbey & Simon Preston');
INSERT INTO artists VALUES(245,'Michael Tilson Thomas & San Francisco Symphony');
INSERT INTO artists VALUES(246,'Chor der Wiener Staatsoper, Herbert Von Karajan & Wiener Philharmoniker');
INSERT INTO artists VALUES(247,'The King''s Singers');
INSERT INTO artists VALUES(248,'Berliner Philharmoniker & Herbert Von Karajan');
INSERT INTO artists VALUES(249,'Sir Georg Solti, Sumi Jo & Wiener Philharmoniker');
INSERT INTO artists VALUES(250,'Christopher O''Riley');
INSERT INTO artists VALUES(251,'Fretwork');
INSERT INTO artists VALUES(252,'Amy Winehouse');
INSERT INTO artists VALUES(253,'Calexico');
INSERT INTO artists VALUES(254,'Otto Klemperer & Philharmonia Orchestra');
INSERT INTO artists VALUES(255,'Yehudi Menuhin');
INSERT INTO artists VALUES(256,'Philharmonia Orchestra & Sir Neville Marriner');
INSERT INTO artists VALUES(257,'Academy of St. Martin in the Fields, Sir Neville Marriner & Thurston Dart');
INSERT INTO artists VALUES(258,'Les Arts Florissants & William Christie');
INSERT INTO artists VALUES(259,'The 12 Cellists of The Berlin Philharmonic');
INSERT INTO artists VALUES(260,'Adrian Leaper & Doreen de Feis');
INSERT INTO artists VALUES(261,'Roger Norrington, London Classical Players');
INSERT INTO artists VALUES(262,'Charles Dutoit & L''Orchestre Symphonique de Montréal');
INSERT INTO artists VALUES(263,'Equale Brass Ensemble, John Eliot Gardiner & Munich Monteverdi Orchestra and Choir');
INSERT INTO artists VALUES(264,'Kent Nagano and Orchestre de l''Opéra de Lyon');
INSERT INTO artists VALUES(265,'Julian Bream');
INSERT INTO artists VALUES(266,'Martin Roscoe');
INSERT INTO artists VALUES(267,'Göteborgs Symfoniker & Neeme Järvi');
INSERT INTO artists VALUES(268,'Itzhak Perlman');
INSERT INTO artists VALUES(269,'Michele Campanella');
INSERT INTO artists VALUES(270,'Gerald Moore');
INSERT INTO artists VALUES(271,'Mela Tenenbaum, Pro Musica Prague & Richard Kapp');
INSERT INTO artists VALUES(272,'Emerson String Quartet');
INSERT INTO artists VALUES(273,'C. Monteverdi, Nigel Rogers - Chiaroscuro; London Baroque; London Cornett & Sackbu');
INSERT INTO artists VALUES(274,'Nash Ensemble');
INSERT INTO artists VALUES(275,'Philip Glass Ensemble');
CREATE TABLE IF NOT EXISTS "customers"
(
    [CustomerId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [FirstName] NVARCHAR(40)  NOT NULL,
    [LastName] NVARCHAR(20)  NOT NULL,
    [Company] NVARCHAR(80),
    [Address] NVARCHAR(70),
    [City] NVARCHAR(40),
    [State] NVARCHAR(40),
    [Country] NVARCHAR(40),
    [PostalCode] NVARCHAR(10),
    [Phone] NVARCHAR(24),
    [Fax] NVARCHAR(24),
    [Email] NVARCHAR(60)  NOT NULL,
    [SupportRepId] INTEGER,
    FOREIGN KEY ([SupportRepId]) REFERENCES "employees" ([EmployeeId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION
);
INSERT INTO customers VALUES(1,'Luís','Gonçalves','Embraer - Empresa Brasileira de Aeronáutica S.A.','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000','+55 (12) 3923-5555','+55 (12) 3923-5566','luisg@embraer.com.br',3);
INSERT INTO customers VALUES(2,'Leonie','Köhler',NULL,'Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174','+49 0711 2842222',NULL,'leonekohler@surfeu.de',5);
INSERT INTO customers VALUES(3,'François','Tremblay',NULL,'1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7','+1 (514) 721-4711',NULL,'ftremblay@gmail.com',3);
INSERT INTO customers VALUES(4,'Bjørn','Hansen',NULL,'Ullevålsveien 14','Oslo',NULL,'Norway','0171','+47 22 44 22 22',NULL,'bjorn.hansen@yahoo.no',4);
INSERT INTO customers VALUES(5,'František','Wichterlová','JetBrains s.r.o.','Klanova 9/506','Prague',NULL,'Czech Republic','14700','+420 2 4172 5555','+420 2 4172 5555','frantisekw@jetbrains.com',4);
INSERT INTO customers VALUES(6,'Helena','Holý',NULL,'Rilská 3174/6','Prague',NULL,'Czech Republic','14300','+420 2 4177 0449',NULL,'hholy@gmail.com',5);
INSERT INTO customers VALUES(7,'Astrid','Gruber',NULL,'Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010','+43 01 5134505',NULL,'astrid.gruber@apple.at',5);
INSERT INTO customers VALUES(8,'Daan','Peeters',NULL,'Grétrystraat 63','Brussels',NULL,'Belgium','1000','+32 02 219 03 03',NULL,'daan_peeters@apple.be',4);
INSERT INTO customers VALUES(9,'Kara','Nielsen',NULL,'Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720','+453 3331 9991',NULL,'kara.nielsen@jubii.dk',4);
INSERT INTO customers VALUES(10,'Eduardo','Martins','Woodstock Discos','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010','+55 (11) 3033-5446','+55 (11) 3033-4564','eduardo@woodstock.com.br',4);
INSERT INTO customers VALUES(11,'Alexandre','Rocha','Banco do Brasil S.A.','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200','+55 (11) 3055-3278','+55 (11) 3055-8131','alero@uol.com.br',5);
INSERT INTO customers VALUES(12,'Roberto','Almeida','Riotur','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020','+55 (21) 2271-7000','+55 (21) 2271-7070','roberto.almeida@riotur.gov.br',3);
INSERT INTO customers VALUES(13,'Fernanda','Ramos',NULL,'Qe 7 Bloco G','Brasília','DF','Brazil','71020-677','+55 (61) 3363-5547','+55 (61) 3363-7855','fernadaramos4@uol.com.br',4);
INSERT INTO customers VALUES(14,'Mark','Philips','Telus','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7','+1 (780) 434-4554','+1 (780) 434-5565','mphilips12@shaw.ca',5);
INSERT INTO customers VALUES(15,'Jennifer','Peterson','Rogers Canada','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8','+1 (604) 688-2255','+1 (604) 688-8756','jenniferp@rogers.ca',3);
INSERT INTO customers VALUES(16,'Frank','Harris','Google Inc.','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351','+1 (650) 253-0000','+1 (650) 253-0000','fharris@google.com',4);
INSERT INTO customers VALUES(17,'Jack','Smith','Microsoft Corporation','1 Microsoft Way','Redmond','WA','USA','98052-8300','+1 (425) 882-8080','+1 (425) 882-8081','jacksmith@microsoft.com',5);
INSERT INTO customers VALUES(18,'Michelle','Brooks',NULL,'627 Broadway','New York','NY','USA','10012-2612','+1 (212) 221-3546','+1 (212) 221-4679','michelleb@aol.com',3);
INSERT INTO customers VALUES(19,'Tim','Goyer','Apple Inc.','1 Infinite Loop','Cupertino','CA','USA','95014','+1 (408) 996-1010','+1 (408) 996-1011','tgoyer@apple.com',3);
INSERT INTO customers VALUES(20,'Dan','Miller',NULL,'541 Del Medio Avenue','Mountain View','CA','USA','94040-111','+1 (650) 644-3358',NULL,'dmiller@comcast.com',4);
INSERT INTO customers VALUES(21,'Kathy','Chase',NULL,'801 W 4th Street','Reno','NV','USA','89503','+1 (775) 223-7665',NULL,'kachase@hotmail.com',5);
INSERT INTO customers VALUES(22,'Heather','Leacock',NULL,'120 S Orange Ave','Orlando','FL','USA','32801','+1 (407) 999-7788',NULL,'hleacock@gmail.com',4);
INSERT INTO customers VALUES(23,'John','Gordon',NULL,'69 Salem Street','Boston','MA','USA','2113','+1 (617) 522-1333',NULL,'johngordon22@yahoo.com',4);
INSERT INTO customers VALUES(24,'Frank','Ralston',NULL,'162 E Superior Street','Chicago','IL','USA','60611','+1 (312) 332-3232',NULL,'fralston@gmail.com',3);
INSERT INTO customers VALUES(25,'Victor','Stevens',NULL,'319 N. Frances Street','Madison','WI','USA','53703','+1 (608) 257-0597',NULL,'vstevens@yahoo.com',5);
INSERT INTO customers VALUES(26,'Richard','Cunningham',NULL,'2211 W Berry Street','Fort Worth','TX','USA','76110','+1 (817) 924-7272',NULL,'ricunningham@hotmail.com',4);
INSERT INTO customers VALUES(27,'Patrick','Gray',NULL,'1033 N Park Ave','Tucson','AZ','USA','85719','+1 (520) 622-4200',NULL,'patrick.gray@aol.com',4);
INSERT INTO customers VALUES(28,'Julia','Barnett',NULL,'302 S 700 E','Salt Lake City','UT','USA','84102','+1 (801) 531-7272',NULL,'jubarnett@gmail.com',5);
INSERT INTO customers VALUES(29,'Robert','Brown',NULL,'796 Dundas Street West','Toronto','ON','Canada','M6J 1V1','+1 (416) 363-8888',NULL,'robbrown@shaw.ca',3);
INSERT INTO customers VALUES(30,'Edward','Francis',NULL,'230 Elgin Street','Ottawa','ON','Canada','K2P 1L7','+1 (613) 234-3322',NULL,'edfrancis@yachoo.ca',3);
INSERT INTO customers VALUES(31,'Martha','Silk',NULL,'194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5','+1 (902) 450-0450',NULL,'marthasilk@gmail.com',5);
INSERT INTO customers VALUES(32,'Aaron','Mitchell',NULL,'696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9','+1 (204) 452-6452',NULL,'aaronmitchell@yahoo.ca',4);
INSERT INTO customers VALUES(33,'Ellie','Sullivan',NULL,'5112 48 Street','Yellowknife','NT','Canada','X1A 1N6','+1 (867) 920-2233',NULL,'ellie.sullivan@shaw.ca',3);
INSERT INTO customers VALUES(34,'João','Fernandes',NULL,'Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,'+351 (213) 466-111',NULL,'jfernandes@yahoo.pt',4);
INSERT INTO customers VALUES(35,'Madalena','Sampaio',NULL,'Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,'+351 (225) 022-448',NULL,'masampaio@sapo.pt',4);
INSERT INTO customers VALUES(36,'Hannah','Schneider',NULL,'Tauentzienstraße 8','Berlin',NULL,'Germany','10789','+49 030 26550280',NULL,'hannah.schneider@yahoo.de',5);
INSERT INTO customers VALUES(37,'Fynn','Zimmermann',NULL,'Berger Straße 10','Frankfurt',NULL,'Germany','60316','+49 069 40598889',NULL,'fzimmermann@yahoo.de',3);
INSERT INTO customers VALUES(38,'Niklas','Schröder',NULL,'Barbarossastraße 19','Berlin',NULL,'Germany','10779','+49 030 2141444',NULL,'nschroder@surfeu.de',3);
INSERT INTO customers VALUES(39,'Camille','Bernard',NULL,'4, Rue Milton','Paris',NULL,'France','75009','+33 01 49 70 65 65',NULL,'camille.bernard@yahoo.fr',4);
INSERT INTO customers VALUES(40,'Dominique','Lefebvre',NULL,'8, Rue Hanovre','Paris',NULL,'France','75002','+33 01 47 42 71 71',NULL,'dominiquelefebvre@gmail.com',4);
INSERT INTO customers VALUES(41,'Marc','Dubois',NULL,'11, Place Bellecour','Lyon',NULL,'France','69002','+33 04 78 30 30 30',NULL,'marc.dubois@hotmail.com',5);
INSERT INTO customers VALUES(42,'Wyatt','Girard',NULL,'9, Place Louis Barthou','Bordeaux',NULL,'France','33000','+33 05 56 96 96 96',NULL,'wyatt.girard@yahoo.fr',3);
INSERT INTO customers VALUES(43,'Isabelle','Mercier',NULL,'68, Rue Jouvence','Dijon',NULL,'France','21000','+33 03 80 73 66 99',NULL,'isabelle_mercier@apple.fr',3);
INSERT INTO customers VALUES(44,'Terhi','Hämäläinen',NULL,'Porthaninkatu 9','Helsinki',NULL,'Finland','00530','+358 09 870 2000',NULL,'terhi.hamalainen@apple.fi',3);
INSERT INTO customers VALUES(45,'Ladislav','Kovács',NULL,'Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',NULL,NULL,'ladislav_kovacs@apple.hu',3);
INSERT INTO customers VALUES(46,'Hugh','O''Reilly',NULL,'3 Chatham Street','Dublin','Dublin','Ireland',NULL,'+353 01 6792424',NULL,'hughoreilly@apple.ie',3);
INSERT INTO customers VALUES(47,'Lucas','Mancini',NULL,'Via Degli Scipioni, 43','Rome','RM','Italy','00192','+39 06 39733434',NULL,'lucas.mancini@yahoo.it',5);
INSERT INTO customers VALUES(48,'Johannes','Van der Berg',NULL,'Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016','+31 020 6223130',NULL,'johavanderberg@yahoo.nl',5);
INSERT INTO customers VALUES(49,'Stanisław','Wójcik',NULL,'Ordynacka 10','Warsaw',NULL,'Poland','00-358','+48 22 828 37 39',NULL,'stanisław.wójcik@wp.pl',4);
INSERT INTO customers VALUES(50,'Enrique','Muñoz',NULL,'C/ San Bernardo 85','Madrid',NULL,'Spain','28015','+34 914 454 454',NULL,'enrique_munoz@yahoo.es',5);
INSERT INTO customers VALUES(51,'Joakim','Johansson',NULL,'Celsiusg. 9','Stockholm',NULL,'Sweden','11230','+46 08-651 52 52',NULL,'joakim.johansson@yahoo.se',5);
INSERT INTO customers VALUES(52,'Emma','Jones',NULL,'202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH','+44 020 7707 0707',NULL,'emma_jones@hotmail.com',3);
INSERT INTO customers VALUES(53,'Phil','Hughes',NULL,'113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN','+44 020 7976 5722',NULL,'phil.hughes@gmail.com',3);
INSERT INTO customers VALUES(54,'Steve','Murray',NULL,'110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH','+44 0131 315 3300',NULL,'steve.murray@yahoo.uk',5);
INSERT INTO customers VALUES(55,'Mark','Taylor',NULL,'421 Bourke Street','Sidney','NSW','Australia','2010','+61 (02) 9332 3633',NULL,'mark.taylor@yahoo.au',4);
INSERT INTO customers VALUES(56,'Diego','Gutiérrez',NULL,'307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106','+54 (0)11 4311 4333',NULL,'diego.gutierrez@yahoo.ar',4);
INSERT INTO customers VALUES(57,'Luis','Rojas',NULL,'Calle Lira, 198','Santiago',NULL,'Chile',NULL,'+56 (0)2 635 4444',NULL,'luisrojas@yahoo.cl',5);
INSERT INTO customers VALUES(58,'Manoj','Pareek',NULL,'12,Community Centre','Delhi',NULL,'India','110017','+91 0124 39883988',NULL,'manoj.pareek@rediff.com',3);
INSERT INTO customers VALUES(59,'Puja','Srivastava',NULL,'3,Raj Bhavan Road','Bangalore',NULL,'India','560001','+91 080 22289999',NULL,'puja_srivastava@yahoo.in',3);
CREATE TABLE IF NOT EXISTS "employees"
(
    [EmployeeId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [LastName] NVARCHAR(20)  NOT NULL,
    [FirstName] NVARCHAR(20)  NOT NULL,
    [Title] NVARCHAR(30),
    [ReportsTo] INTEGER,
    [BirthDate] DATETIME,
    [HireDate] DATETIME,
    [Address] NVARCHAR(70),
    [City] NVARCHAR(40),
    [State] NVARCHAR(40),
    [Country] NVARCHAR(40),
    [PostalCode] NVARCHAR(10),
    [Phone] NVARCHAR(24),
    [Fax] NVARCHAR(24),
    [Email] NVARCHAR(60),
    FOREIGN KEY ([ReportsTo]) REFERENCES "employees" ([EmployeeId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION
);
INSERT INTO employees VALUES(1,'Adams','Andrew','General Manager',NULL,'1962-02-18 00:00:00','2002-08-14 00:00:00','11120 Jasper Ave NW','Edmonton','AB','Canada','T5K 2N1','+1 (780) 428-9482','+1 (780) 428-3457','andrew@chinookcorp.com');
INSERT INTO employees VALUES(2,'Edwards','Nancy','Sales Manager',1,'1958-12-08 00:00:00','2002-05-01 00:00:00','825 8 Ave SW','Calgary','AB','Canada','T2P 2T3','+1 (403) 262-3443','+1 (403) 262-3322','nancy@chinookcorp.com');
INSERT INTO employees VALUES(3,'Peacock','Jane','Sales Support Agent',2,'1973-08-29 00:00:00','2002-04-01 00:00:00','1111 6 Ave SW','Calgary','AB','Canada','T2P 5M5','+1 (403) 262-3443','+1 (403) 262-6712','jane@chinookcorp.com');
INSERT INTO employees VALUES(4,'Park','Margaret','Sales Support Agent',2,'1947-09-19 00:00:00','2003-05-03 00:00:00','683 10 Street SW','Calgary','AB','Canada','T2P 5G3','+1 (403) 263-4423','+1 (403) 263-4289','margaret@chinookcorp.com');
INSERT INTO employees VALUES(5,'Johnson','Steve','Sales Support Agent',2,'1965-03-03 00:00:00','2003-10-17 00:00:00','7727B 41 Ave','Calgary','AB','Canada','T3B 1Y7','1 (780) 836-9987','1 (780) 836-9543','steve@chinookcorp.com');
INSERT INTO employees VALUES(6,'Mitchell','Michael','IT Manager',1,'1973-07-01 00:00:00','2003-10-17 00:00:00','5827 Bowness Road NW','Calgary','AB','Canada','T3B 0C5','+1 (403) 246-9887','+1 (403) 246-9899','michael@chinookcorp.com');
INSERT INTO employees VALUES(7,'King','Robert','IT Staff',6,'1970-05-29 00:00:00','2004-01-02 00:00:00','590 Columbia Boulevard West','Lethbridge','AB','Canada','T1K 5N8','+1 (403) 456-9986','+1 (403) 456-8485','robert@chinookcorp.com');
INSERT INTO employees VALUES(8,'Callahan','Laura','IT Staff',6,'1968-01-09 00:00:00','2004-03-04 00:00:00','923 7 ST NW','Lethbridge','AB','Canada','T1H 1Y8','+1 (403) 467-3351','+1 (403) 467-8772','laura@chinookcorp.com');
CREATE TABLE IF NOT EXISTS "genres"
(
    [GenreId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [Name] NVARCHAR(120)
);
INSERT INTO genres VALUES(1,'Rock');
INSERT INTO genres VALUES(2,'Jazz');
INSERT INTO genres VALUES(3,'Metal');
INSERT INTO genres VALUES(4,'Alternative & Punk');
INSERT INTO genres VALUES(5,'Rock And Roll');
INSERT INTO genres VALUES(6,'Blues');
INSERT INTO genres VALUES(7,'Latin');
INSERT INTO genres VALUES(8,'Reggae');
INSERT INTO genres VALUES(9,'Pop');
INSERT INTO genres VALUES(10,'Soundtrack');
INSERT INTO genres VALUES(11,'Bossa Nova');
INSERT INTO genres VALUES(12,'Easy Listening');
INSERT INTO genres VALUES(13,'Heavy Metal');
INSERT INTO genres VALUES(14,'R&B/Soul');
INSERT INTO genres VALUES(15,'Electronica/Dance');
INSERT INTO genres VALUES(16,'World');
INSERT INTO genres VALUES(17,'Hip Hop/Rap');
INSERT INTO genres VALUES(18,'Science Fiction');
INSERT INTO genres VALUES(19,'TV Shows');
INSERT INTO genres VALUES(20,'Sci Fi & Fantasy');
INSERT INTO genres VALUES(21,'Drama');
INSERT INTO genres VALUES(22,'Comedy');
INSERT INTO genres VALUES(23,'Alternative');
INSERT INTO genres VALUES(24,'Classical');
INSERT INTO genres VALUES(25,'Opera');
CREATE TABLE IF NOT EXISTS "invoices"
(
    [InvoiceId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [CustomerId] INTEGER  NOT NULL,
    [InvoiceDate] DATETIME  NOT NULL,
    [BillingAddress] NVARCHAR(70),
    [BillingCity] NVARCHAR(40),
    [BillingState] NVARCHAR(40),
    [BillingCountry] NVARCHAR(40),
    [BillingPostalCode] NVARCHAR(10),
    [Total] NUMERIC(10,2)  NOT NULL,
    FOREIGN KEY ([CustomerId]) REFERENCES "customers" ([CustomerId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION
);
INSERT INTO invoices VALUES(1,2,'2009-01-01 00:00:00','Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174',1.9799999999999999822);
INSERT INTO invoices VALUES(2,4,'2009-01-02 00:00:00','Ullevålsveien 14','Oslo',NULL,'Norway','0171',3.9599999999999999644);
INSERT INTO invoices VALUES(3,8,'2009-01-03 00:00:00','Grétrystraat 63','Brussels',NULL,'Belgium','1000',5.9400000000000003907);
INSERT INTO invoices VALUES(4,14,'2009-01-06 00:00:00','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7',8.9100000000000001421);
INSERT INTO invoices VALUES(5,23,'2009-01-11 00:00:00','69 Salem Street','Boston','MA','USA','2113',13.859999999999999431);
INSERT INTO invoices VALUES(6,37,'2009-01-19 00:00:00','Berger Straße 10','Frankfurt',NULL,'Germany','60316',0.98999999999999999111);
INSERT INTO invoices VALUES(7,38,'2009-02-01 00:00:00','Barbarossastraße 19','Berlin',NULL,'Germany','10779',1.9799999999999999822);
INSERT INTO invoices VALUES(8,40,'2009-02-01 00:00:00','8, Rue Hanovre','Paris',NULL,'France','75002',1.9799999999999999822);
INSERT INTO invoices VALUES(9,42,'2009-02-02 00:00:00','9, Place Louis Barthou','Bordeaux',NULL,'France','33000',3.9599999999999999644);
INSERT INTO invoices VALUES(10,46,'2009-02-03 00:00:00','3 Chatham Street','Dublin','Dublin','Ireland',NULL,5.9400000000000003907);
INSERT INTO invoices VALUES(11,52,'2009-02-06 00:00:00','202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH',8.9100000000000001421);
INSERT INTO invoices VALUES(12,2,'2009-02-11 00:00:00','Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174',13.859999999999999431);
INSERT INTO invoices VALUES(13,16,'2009-02-19 00:00:00','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351',0.98999999999999999111);
INSERT INTO invoices VALUES(14,17,'2009-03-04 00:00:00','1 Microsoft Way','Redmond','WA','USA','98052-8300',1.9799999999999999822);
INSERT INTO invoices VALUES(15,19,'2009-03-04 00:00:00','1 Infinite Loop','Cupertino','CA','USA','95014',1.9799999999999999822);
INSERT INTO invoices VALUES(16,21,'2009-03-05 00:00:00','801 W 4th Street','Reno','NV','USA','89503',3.9599999999999999644);
INSERT INTO invoices VALUES(17,25,'2009-03-06 00:00:00','319 N. Frances Street','Madison','WI','USA','53703',5.9400000000000003907);
INSERT INTO invoices VALUES(18,31,'2009-03-09 00:00:00','194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5',8.9100000000000001421);
INSERT INTO invoices VALUES(19,40,'2009-03-14 00:00:00','8, Rue Hanovre','Paris',NULL,'France','75002',13.859999999999999431);
INSERT INTO invoices VALUES(20,54,'2009-03-22 00:00:00','110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH',0.98999999999999999111);
INSERT INTO invoices VALUES(21,55,'2009-04-04 00:00:00','421 Bourke Street','Sidney','NSW','Australia','2010',1.9799999999999999822);
INSERT INTO invoices VALUES(22,57,'2009-04-04 00:00:00','Calle Lira, 198','Santiago',NULL,'Chile',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(23,59,'2009-04-05 00:00:00','3,Raj Bhavan Road','Bangalore',NULL,'India','560001',3.9599999999999999644);
INSERT INTO invoices VALUES(24,4,'2009-04-06 00:00:00','Ullevålsveien 14','Oslo',NULL,'Norway','0171',5.9400000000000003907);
INSERT INTO invoices VALUES(25,10,'2009-04-09 00:00:00','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010',8.9100000000000001421);
INSERT INTO invoices VALUES(26,19,'2009-04-14 00:00:00','1 Infinite Loop','Cupertino','CA','USA','95014',13.859999999999999431);
INSERT INTO invoices VALUES(27,33,'2009-04-22 00:00:00','5112 48 Street','Yellowknife','NT','Canada','X1A 1N6',0.98999999999999999111);
INSERT INTO invoices VALUES(28,34,'2009-05-05 00:00:00','Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(29,36,'2009-05-05 00:00:00','Tauentzienstraße 8','Berlin',NULL,'Germany','10789',1.9799999999999999822);
INSERT INTO invoices VALUES(30,38,'2009-05-06 00:00:00','Barbarossastraße 19','Berlin',NULL,'Germany','10779',3.9599999999999999644);
INSERT INTO invoices VALUES(31,42,'2009-05-07 00:00:00','9, Place Louis Barthou','Bordeaux',NULL,'France','33000',5.9400000000000003907);
INSERT INTO invoices VALUES(32,48,'2009-05-10 00:00:00','Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016',8.9100000000000001421);
INSERT INTO invoices VALUES(33,57,'2009-05-15 00:00:00','Calle Lira, 198','Santiago',NULL,'Chile',NULL,13.859999999999999431);
INSERT INTO invoices VALUES(34,12,'2009-05-23 00:00:00','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020',0.98999999999999999111);
INSERT INTO invoices VALUES(35,13,'2009-06-05 00:00:00','Qe 7 Bloco G','Brasília','DF','Brazil','71020-677',1.9799999999999999822);
INSERT INTO invoices VALUES(36,15,'2009-06-05 00:00:00','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8',1.9799999999999999822);
INSERT INTO invoices VALUES(37,17,'2009-06-06 00:00:00','1 Microsoft Way','Redmond','WA','USA','98052-8300',3.9599999999999999644);
INSERT INTO invoices VALUES(38,21,'2009-06-07 00:00:00','801 W 4th Street','Reno','NV','USA','89503',5.9400000000000003907);
INSERT INTO invoices VALUES(39,27,'2009-06-10 00:00:00','1033 N Park Ave','Tucson','AZ','USA','85719',8.9100000000000001421);
INSERT INTO invoices VALUES(40,36,'2009-06-15 00:00:00','Tauentzienstraße 8','Berlin',NULL,'Germany','10789',13.859999999999999431);
INSERT INTO invoices VALUES(41,50,'2009-06-23 00:00:00','C/ San Bernardo 85','Madrid',NULL,'Spain','28015',0.98999999999999999111);
INSERT INTO invoices VALUES(42,51,'2009-07-06 00:00:00','Celsiusg. 9','Stockholm',NULL,'Sweden','11230',1.9799999999999999822);
INSERT INTO invoices VALUES(43,53,'2009-07-06 00:00:00','113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN',1.9799999999999999822);
INSERT INTO invoices VALUES(44,55,'2009-07-07 00:00:00','421 Bourke Street','Sidney','NSW','Australia','2010',3.9599999999999999644);
INSERT INTO invoices VALUES(45,59,'2009-07-08 00:00:00','3,Raj Bhavan Road','Bangalore',NULL,'India','560001',5.9400000000000003907);
INSERT INTO invoices VALUES(46,6,'2009-07-11 00:00:00','Rilská 3174/6','Prague',NULL,'Czech Republic','14300',8.9100000000000001421);
INSERT INTO invoices VALUES(47,15,'2009-07-16 00:00:00','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8',13.859999999999999431);
INSERT INTO invoices VALUES(48,29,'2009-07-24 00:00:00','796 Dundas Street West','Toronto','ON','Canada','M6J 1V1',0.98999999999999999111);
INSERT INTO invoices VALUES(49,30,'2009-08-06 00:00:00','230 Elgin Street','Ottawa','ON','Canada','K2P 1L7',1.9799999999999999822);
INSERT INTO invoices VALUES(50,32,'2009-08-06 00:00:00','696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9',1.9799999999999999822);
INSERT INTO invoices VALUES(51,34,'2009-08-07 00:00:00','Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,3.9599999999999999644);
INSERT INTO invoices VALUES(52,38,'2009-08-08 00:00:00','Barbarossastraße 19','Berlin',NULL,'Germany','10779',5.9400000000000003907);
INSERT INTO invoices VALUES(53,44,'2009-08-11 00:00:00','Porthaninkatu 9','Helsinki',NULL,'Finland','00530',8.9100000000000001421);
INSERT INTO invoices VALUES(54,53,'2009-08-16 00:00:00','113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN',13.859999999999999431);
INSERT INTO invoices VALUES(55,8,'2009-08-24 00:00:00','Grétrystraat 63','Brussels',NULL,'Belgium','1000',0.98999999999999999111);
INSERT INTO invoices VALUES(56,9,'2009-09-06 00:00:00','Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720',1.9799999999999999822);
INSERT INTO invoices VALUES(57,11,'2009-09-06 00:00:00','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200',1.9799999999999999822);
INSERT INTO invoices VALUES(58,13,'2009-09-07 00:00:00','Qe 7 Bloco G','Brasília','DF','Brazil','71020-677',3.9599999999999999644);
INSERT INTO invoices VALUES(59,17,'2009-09-08 00:00:00','1 Microsoft Way','Redmond','WA','USA','98052-8300',5.9400000000000003907);
INSERT INTO invoices VALUES(60,23,'2009-09-11 00:00:00','69 Salem Street','Boston','MA','USA','2113',8.9100000000000001421);
INSERT INTO invoices VALUES(61,32,'2009-09-16 00:00:00','696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9',13.859999999999999431);
INSERT INTO invoices VALUES(62,46,'2009-09-24 00:00:00','3 Chatham Street','Dublin','Dublin','Ireland',NULL,0.98999999999999999111);
INSERT INTO invoices VALUES(63,47,'2009-10-07 00:00:00','Via Degli Scipioni, 43','Rome','RM','Italy','00192',1.9799999999999999822);
INSERT INTO invoices VALUES(64,49,'2009-10-07 00:00:00','Ordynacka 10','Warsaw',NULL,'Poland','00-358',1.9799999999999999822);
INSERT INTO invoices VALUES(65,51,'2009-10-08 00:00:00','Celsiusg. 9','Stockholm',NULL,'Sweden','11230',3.9599999999999999644);
INSERT INTO invoices VALUES(66,55,'2009-10-09 00:00:00','421 Bourke Street','Sidney','NSW','Australia','2010',5.9400000000000003907);
INSERT INTO invoices VALUES(67,2,'2009-10-12 00:00:00','Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174',8.9100000000000001421);
INSERT INTO invoices VALUES(68,11,'2009-10-17 00:00:00','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200',13.859999999999999431);
INSERT INTO invoices VALUES(69,25,'2009-10-25 00:00:00','319 N. Frances Street','Madison','WI','USA','53703',0.98999999999999999111);
INSERT INTO invoices VALUES(70,26,'2009-11-07 00:00:00','2211 W Berry Street','Fort Worth','TX','USA','76110',1.9799999999999999822);
INSERT INTO invoices VALUES(71,28,'2009-11-07 00:00:00','302 S 700 E','Salt Lake City','UT','USA','84102',1.9799999999999999822);
INSERT INTO invoices VALUES(72,30,'2009-11-08 00:00:00','230 Elgin Street','Ottawa','ON','Canada','K2P 1L7',3.9599999999999999644);
INSERT INTO invoices VALUES(73,34,'2009-11-09 00:00:00','Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,5.9400000000000003907);
INSERT INTO invoices VALUES(74,40,'2009-11-12 00:00:00','8, Rue Hanovre','Paris',NULL,'France','75002',8.9100000000000001421);
INSERT INTO invoices VALUES(75,49,'2009-11-17 00:00:00','Ordynacka 10','Warsaw',NULL,'Poland','00-358',13.859999999999999431);
INSERT INTO invoices VALUES(76,4,'2009-11-25 00:00:00','Ullevålsveien 14','Oslo',NULL,'Norway','0171',0.98999999999999999111);
INSERT INTO invoices VALUES(77,5,'2009-12-08 00:00:00','Klanova 9/506','Prague',NULL,'Czech Republic','14700',1.9799999999999999822);
INSERT INTO invoices VALUES(78,7,'2009-12-08 00:00:00','Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010',1.9799999999999999822);
INSERT INTO invoices VALUES(79,9,'2009-12-09 00:00:00','Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720',3.9599999999999999644);
INSERT INTO invoices VALUES(80,13,'2009-12-10 00:00:00','Qe 7 Bloco G','Brasília','DF','Brazil','71020-677',5.9400000000000003907);
INSERT INTO invoices VALUES(81,19,'2009-12-13 00:00:00','1 Infinite Loop','Cupertino','CA','USA','95014',8.9100000000000001421);
INSERT INTO invoices VALUES(82,28,'2009-12-18 00:00:00','302 S 700 E','Salt Lake City','UT','USA','84102',13.859999999999999431);
INSERT INTO invoices VALUES(83,42,'2009-12-26 00:00:00','9, Place Louis Barthou','Bordeaux',NULL,'France','33000',0.98999999999999999111);
INSERT INTO invoices VALUES(84,43,'2010-01-08 00:00:00','68, Rue Jouvence','Dijon',NULL,'France','21000',1.9799999999999999822);
INSERT INTO invoices VALUES(85,45,'2010-01-08 00:00:00','Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',1.9799999999999999822);
INSERT INTO invoices VALUES(86,47,'2010-01-09 00:00:00','Via Degli Scipioni, 43','Rome','RM','Italy','00192',3.9599999999999999644);
INSERT INTO invoices VALUES(87,51,'2010-01-10 00:00:00','Celsiusg. 9','Stockholm',NULL,'Sweden','11230',6.9400000000000003907);
INSERT INTO invoices VALUES(88,57,'2010-01-13 00:00:00','Calle Lira, 198','Santiago',NULL,'Chile',NULL,17.910000000000000142);
INSERT INTO invoices VALUES(89,7,'2010-01-18 00:00:00','Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010',18.859999999999999431);
INSERT INTO invoices VALUES(90,21,'2010-01-26 00:00:00','801 W 4th Street','Reno','NV','USA','89503',0.98999999999999999111);
INSERT INTO invoices VALUES(91,22,'2010-02-08 00:00:00','120 S Orange Ave','Orlando','FL','USA','32801',1.9799999999999999822);
INSERT INTO invoices VALUES(92,24,'2010-02-08 00:00:00','162 E Superior Street','Chicago','IL','USA','60611',1.9799999999999999822);
INSERT INTO invoices VALUES(93,26,'2010-02-09 00:00:00','2211 W Berry Street','Fort Worth','TX','USA','76110',3.9599999999999999644);
INSERT INTO invoices VALUES(94,30,'2010-02-10 00:00:00','230 Elgin Street','Ottawa','ON','Canada','K2P 1L7',5.9400000000000003907);
INSERT INTO invoices VALUES(95,36,'2010-02-13 00:00:00','Tauentzienstraße 8','Berlin',NULL,'Germany','10789',8.9100000000000001421);
INSERT INTO invoices VALUES(96,45,'2010-02-18 00:00:00','Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',21.859999999999999431);
INSERT INTO invoices VALUES(97,59,'2010-02-26 00:00:00','3,Raj Bhavan Road','Bangalore',NULL,'India','560001',1.9899999999999999911);
INSERT INTO invoices VALUES(98,1,'2010-03-11 00:00:00','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000',3.9799999999999999822);
INSERT INTO invoices VALUES(99,3,'2010-03-11 00:00:00','1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7',3.9799999999999999822);
INSERT INTO invoices VALUES(100,5,'2010-03-12 00:00:00','Klanova 9/506','Prague',NULL,'Czech Republic','14700',3.9599999999999999644);
INSERT INTO invoices VALUES(101,9,'2010-03-13 00:00:00','Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720',5.9400000000000003907);
INSERT INTO invoices VALUES(102,15,'2010-03-16 00:00:00','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8',9.9100000000000001421);
INSERT INTO invoices VALUES(103,24,'2010-03-21 00:00:00','162 E Superior Street','Chicago','IL','USA','60611',15.859999999999999431);
INSERT INTO invoices VALUES(104,38,'2010-03-29 00:00:00','Barbarossastraße 19','Berlin',NULL,'Germany','10779',0.98999999999999999111);
INSERT INTO invoices VALUES(105,39,'2010-04-11 00:00:00','4, Rue Milton','Paris',NULL,'France','75009',1.9799999999999999822);
INSERT INTO invoices VALUES(106,41,'2010-04-11 00:00:00','11, Place Bellecour','Lyon',NULL,'France','69002',1.9799999999999999822);
INSERT INTO invoices VALUES(107,43,'2010-04-12 00:00:00','68, Rue Jouvence','Dijon',NULL,'France','21000',3.9599999999999999644);
INSERT INTO invoices VALUES(108,47,'2010-04-13 00:00:00','Via Degli Scipioni, 43','Rome','RM','Italy','00192',5.9400000000000003907);
INSERT INTO invoices VALUES(109,53,'2010-04-16 00:00:00','113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN',8.9100000000000001421);
INSERT INTO invoices VALUES(110,3,'2010-04-21 00:00:00','1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7',13.859999999999999431);
INSERT INTO invoices VALUES(111,17,'2010-04-29 00:00:00','1 Microsoft Way','Redmond','WA','USA','98052-8300',0.98999999999999999111);
INSERT INTO invoices VALUES(112,18,'2010-05-12 00:00:00','627 Broadway','New York','NY','USA','10012-2612',1.9799999999999999822);
INSERT INTO invoices VALUES(113,20,'2010-05-12 00:00:00','541 Del Medio Avenue','Mountain View','CA','USA','94040-111',1.9799999999999999822);
INSERT INTO invoices VALUES(114,22,'2010-05-13 00:00:00','120 S Orange Ave','Orlando','FL','USA','32801',3.9599999999999999644);
INSERT INTO invoices VALUES(115,26,'2010-05-14 00:00:00','2211 W Berry Street','Fort Worth','TX','USA','76110',5.9400000000000003907);
INSERT INTO invoices VALUES(116,32,'2010-05-17 00:00:00','696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9',8.9100000000000001421);
INSERT INTO invoices VALUES(117,41,'2010-05-22 00:00:00','11, Place Bellecour','Lyon',NULL,'France','69002',13.859999999999999431);
INSERT INTO invoices VALUES(118,55,'2010-05-30 00:00:00','421 Bourke Street','Sidney','NSW','Australia','2010',0.98999999999999999111);
INSERT INTO invoices VALUES(119,56,'2010-06-12 00:00:00','307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106',1.9799999999999999822);
INSERT INTO invoices VALUES(120,58,'2010-06-12 00:00:00','12,Community Centre','Delhi',NULL,'India','110017',1.9799999999999999822);
INSERT INTO invoices VALUES(121,1,'2010-06-13 00:00:00','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000',3.9599999999999999644);
INSERT INTO invoices VALUES(122,5,'2010-06-14 00:00:00','Klanova 9/506','Prague',NULL,'Czech Republic','14700',5.9400000000000003907);
INSERT INTO invoices VALUES(123,11,'2010-06-17 00:00:00','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200',8.9100000000000001421);
INSERT INTO invoices VALUES(124,20,'2010-06-22 00:00:00','541 Del Medio Avenue','Mountain View','CA','USA','94040-111',13.859999999999999431);
INSERT INTO invoices VALUES(125,34,'2010-06-30 00:00:00','Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,0.98999999999999999111);
INSERT INTO invoices VALUES(126,35,'2010-07-13 00:00:00','Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(127,37,'2010-07-13 00:00:00','Berger Straße 10','Frankfurt',NULL,'Germany','60316',1.9799999999999999822);
INSERT INTO invoices VALUES(128,39,'2010-07-14 00:00:00','4, Rue Milton','Paris',NULL,'France','75009',3.9599999999999999644);
INSERT INTO invoices VALUES(129,43,'2010-07-15 00:00:00','68, Rue Jouvence','Dijon',NULL,'France','21000',5.9400000000000003907);
INSERT INTO invoices VALUES(130,49,'2010-07-18 00:00:00','Ordynacka 10','Warsaw',NULL,'Poland','00-358',8.9100000000000001421);
INSERT INTO invoices VALUES(131,58,'2010-07-23 00:00:00','12,Community Centre','Delhi',NULL,'India','110017',13.859999999999999431);
INSERT INTO invoices VALUES(132,13,'2010-07-31 00:00:00','Qe 7 Bloco G','Brasília','DF','Brazil','71020-677',0.98999999999999999111);
INSERT INTO invoices VALUES(133,14,'2010-08-13 00:00:00','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7',1.9799999999999999822);
INSERT INTO invoices VALUES(134,16,'2010-08-13 00:00:00','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351',1.9799999999999999822);
INSERT INTO invoices VALUES(135,18,'2010-08-14 00:00:00','627 Broadway','New York','NY','USA','10012-2612',3.9599999999999999644);
INSERT INTO invoices VALUES(136,22,'2010-08-15 00:00:00','120 S Orange Ave','Orlando','FL','USA','32801',5.9400000000000003907);
INSERT INTO invoices VALUES(137,28,'2010-08-18 00:00:00','302 S 700 E','Salt Lake City','UT','USA','84102',8.9100000000000001421);
INSERT INTO invoices VALUES(138,37,'2010-08-23 00:00:00','Berger Straße 10','Frankfurt',NULL,'Germany','60316',13.859999999999999431);
INSERT INTO invoices VALUES(139,51,'2010-08-31 00:00:00','Celsiusg. 9','Stockholm',NULL,'Sweden','11230',0.98999999999999999111);
INSERT INTO invoices VALUES(140,52,'2010-09-13 00:00:00','202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH',1.9799999999999999822);
INSERT INTO invoices VALUES(141,54,'2010-09-13 00:00:00','110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH',1.9799999999999999822);
INSERT INTO invoices VALUES(142,56,'2010-09-14 00:00:00','307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106',3.9599999999999999644);
INSERT INTO invoices VALUES(143,1,'2010-09-15 00:00:00','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000',5.9400000000000003907);
INSERT INTO invoices VALUES(144,7,'2010-09-18 00:00:00','Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010',8.9100000000000001421);
INSERT INTO invoices VALUES(145,16,'2010-09-23 00:00:00','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351',13.859999999999999431);
INSERT INTO invoices VALUES(146,30,'2010-10-01 00:00:00','230 Elgin Street','Ottawa','ON','Canada','K2P 1L7',0.98999999999999999111);
INSERT INTO invoices VALUES(147,31,'2010-10-14 00:00:00','194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5',1.9799999999999999822);
INSERT INTO invoices VALUES(148,33,'2010-10-14 00:00:00','5112 48 Street','Yellowknife','NT','Canada','X1A 1N6',1.9799999999999999822);
INSERT INTO invoices VALUES(149,35,'2010-10-15 00:00:00','Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,3.9599999999999999644);
INSERT INTO invoices VALUES(150,39,'2010-10-16 00:00:00','4, Rue Milton','Paris',NULL,'France','75009',5.9400000000000003907);
INSERT INTO invoices VALUES(151,45,'2010-10-19 00:00:00','Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',8.9100000000000001421);
INSERT INTO invoices VALUES(152,54,'2010-10-24 00:00:00','110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH',13.859999999999999431);
INSERT INTO invoices VALUES(153,9,'2010-11-01 00:00:00','Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720',0.98999999999999999111);
INSERT INTO invoices VALUES(154,10,'2010-11-14 00:00:00','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010',1.9799999999999999822);
INSERT INTO invoices VALUES(155,12,'2010-11-14 00:00:00','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020',1.9799999999999999822);
INSERT INTO invoices VALUES(156,14,'2010-11-15 00:00:00','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7',3.9599999999999999644);
INSERT INTO invoices VALUES(157,18,'2010-11-16 00:00:00','627 Broadway','New York','NY','USA','10012-2612',5.9400000000000003907);
INSERT INTO invoices VALUES(158,24,'2010-11-19 00:00:00','162 E Superior Street','Chicago','IL','USA','60611',8.9100000000000001421);
INSERT INTO invoices VALUES(159,33,'2010-11-24 00:00:00','5112 48 Street','Yellowknife','NT','Canada','X1A 1N6',13.859999999999999431);
INSERT INTO invoices VALUES(160,47,'2010-12-02 00:00:00','Via Degli Scipioni, 43','Rome','RM','Italy','00192',0.98999999999999999111);
INSERT INTO invoices VALUES(161,48,'2010-12-15 00:00:00','Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016',1.9799999999999999822);
INSERT INTO invoices VALUES(162,50,'2010-12-15 00:00:00','C/ San Bernardo 85','Madrid',NULL,'Spain','28015',1.9799999999999999822);
INSERT INTO invoices VALUES(163,52,'2010-12-16 00:00:00','202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH',3.9599999999999999644);
INSERT INTO invoices VALUES(164,56,'2010-12-17 00:00:00','307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106',5.9400000000000003907);
INSERT INTO invoices VALUES(165,3,'2010-12-20 00:00:00','1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7',8.9100000000000001421);
INSERT INTO invoices VALUES(166,12,'2010-12-25 00:00:00','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020',13.859999999999999431);
INSERT INTO invoices VALUES(167,26,'2011-01-02 00:00:00','2211 W Berry Street','Fort Worth','TX','USA','76110',0.98999999999999999111);
INSERT INTO invoices VALUES(168,27,'2011-01-15 00:00:00','1033 N Park Ave','Tucson','AZ','USA','85719',1.9799999999999999822);
INSERT INTO invoices VALUES(169,29,'2011-01-15 00:00:00','796 Dundas Street West','Toronto','ON','Canada','M6J 1V1',1.9799999999999999822);
INSERT INTO invoices VALUES(170,31,'2011-01-16 00:00:00','194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5',3.9599999999999999644);
INSERT INTO invoices VALUES(171,35,'2011-01-17 00:00:00','Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,5.9400000000000003907);
INSERT INTO invoices VALUES(172,41,'2011-01-20 00:00:00','11, Place Bellecour','Lyon',NULL,'France','69002',8.9100000000000001421);
INSERT INTO invoices VALUES(173,50,'2011-01-25 00:00:00','C/ San Bernardo 85','Madrid',NULL,'Spain','28015',13.859999999999999431);
INSERT INTO invoices VALUES(174,5,'2011-02-02 00:00:00','Klanova 9/506','Prague',NULL,'Czech Republic','14700',0.98999999999999999111);
INSERT INTO invoices VALUES(175,6,'2011-02-15 00:00:00','Rilská 3174/6','Prague',NULL,'Czech Republic','14300',1.9799999999999999822);
INSERT INTO invoices VALUES(176,8,'2011-02-15 00:00:00','Grétrystraat 63','Brussels',NULL,'Belgium','1000',1.9799999999999999822);
INSERT INTO invoices VALUES(177,10,'2011-02-16 00:00:00','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010',3.9599999999999999644);
INSERT INTO invoices VALUES(178,14,'2011-02-17 00:00:00','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7',5.9400000000000003907);
INSERT INTO invoices VALUES(179,20,'2011-02-20 00:00:00','541 Del Medio Avenue','Mountain View','CA','USA','94040-111',8.9100000000000001421);
INSERT INTO invoices VALUES(180,29,'2011-02-25 00:00:00','796 Dundas Street West','Toronto','ON','Canada','M6J 1V1',13.859999999999999431);
INSERT INTO invoices VALUES(181,43,'2011-03-05 00:00:00','68, Rue Jouvence','Dijon',NULL,'France','21000',0.98999999999999999111);
INSERT INTO invoices VALUES(182,44,'2011-03-18 00:00:00','Porthaninkatu 9','Helsinki',NULL,'Finland','00530',1.9799999999999999822);
INSERT INTO invoices VALUES(183,46,'2011-03-18 00:00:00','3 Chatham Street','Dublin','Dublin','Ireland',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(184,48,'2011-03-19 00:00:00','Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016',3.9599999999999999644);
INSERT INTO invoices VALUES(185,52,'2011-03-20 00:00:00','202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH',5.9400000000000003907);
INSERT INTO invoices VALUES(186,58,'2011-03-23 00:00:00','12,Community Centre','Delhi',NULL,'India','110017',8.9100000000000001421);
INSERT INTO invoices VALUES(187,8,'2011-03-28 00:00:00','Grétrystraat 63','Brussels',NULL,'Belgium','1000',13.859999999999999431);
INSERT INTO invoices VALUES(188,22,'2011-04-05 00:00:00','120 S Orange Ave','Orlando','FL','USA','32801',0.98999999999999999111);
INSERT INTO invoices VALUES(189,23,'2011-04-18 00:00:00','69 Salem Street','Boston','MA','USA','2113',1.9799999999999999822);
INSERT INTO invoices VALUES(190,25,'2011-04-18 00:00:00','319 N. Frances Street','Madison','WI','USA','53703',1.9799999999999999822);
INSERT INTO invoices VALUES(191,27,'2011-04-19 00:00:00','1033 N Park Ave','Tucson','AZ','USA','85719',3.9599999999999999644);
INSERT INTO invoices VALUES(192,31,'2011-04-20 00:00:00','194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5',5.9400000000000003907);
INSERT INTO invoices VALUES(193,37,'2011-04-23 00:00:00','Berger Straße 10','Frankfurt',NULL,'Germany','60316',14.910000000000000142);
INSERT INTO invoices VALUES(194,46,'2011-04-28 00:00:00','3 Chatham Street','Dublin','Dublin','Ireland',NULL,21.859999999999999431);
INSERT INTO invoices VALUES(195,1,'2011-05-06 00:00:00','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000',0.98999999999999999111);
INSERT INTO invoices VALUES(196,2,'2011-05-19 00:00:00','Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174',1.9799999999999999822);
INSERT INTO invoices VALUES(197,4,'2011-05-19 00:00:00','Ullevålsveien 14','Oslo',NULL,'Norway','0171',1.9799999999999999822);
INSERT INTO invoices VALUES(198,6,'2011-05-20 00:00:00','Rilská 3174/6','Prague',NULL,'Czech Republic','14300',3.9599999999999999644);
INSERT INTO invoices VALUES(199,10,'2011-05-21 00:00:00','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010',5.9400000000000003907);
INSERT INTO invoices VALUES(200,16,'2011-05-24 00:00:00','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351',8.9100000000000001421);
INSERT INTO invoices VALUES(201,25,'2011-05-29 00:00:00','319 N. Frances Street','Madison','WI','USA','53703',18.859999999999999431);
INSERT INTO invoices VALUES(202,39,'2011-06-06 00:00:00','4, Rue Milton','Paris',NULL,'France','75009',1.9899999999999999911);
INSERT INTO invoices VALUES(203,40,'2011-06-19 00:00:00','8, Rue Hanovre','Paris',NULL,'France','75002',2.9799999999999999822);
INSERT INTO invoices VALUES(204,42,'2011-06-19 00:00:00','9, Place Louis Barthou','Bordeaux',NULL,'France','33000',3.9799999999999999822);
INSERT INTO invoices VALUES(205,44,'2011-06-20 00:00:00','Porthaninkatu 9','Helsinki',NULL,'Finland','00530',7.9599999999999999644);
INSERT INTO invoices VALUES(206,48,'2011-06-21 00:00:00','Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016',8.9399999999999995026);
INSERT INTO invoices VALUES(207,54,'2011-06-24 00:00:00','110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH',8.9100000000000001421);
INSERT INTO invoices VALUES(208,4,'2011-06-29 00:00:00','Ullevålsveien 14','Oslo',NULL,'Norway','0171',15.859999999999999431);
INSERT INTO invoices VALUES(209,18,'2011-07-07 00:00:00','627 Broadway','New York','NY','USA','10012-2612',0.98999999999999999111);
INSERT INTO invoices VALUES(210,19,'2011-07-20 00:00:00','1 Infinite Loop','Cupertino','CA','USA','95014',1.9799999999999999822);
INSERT INTO invoices VALUES(211,21,'2011-07-20 00:00:00','801 W 4th Street','Reno','NV','USA','89503',1.9799999999999999822);
INSERT INTO invoices VALUES(212,23,'2011-07-21 00:00:00','69 Salem Street','Boston','MA','USA','2113',3.9599999999999999644);
INSERT INTO invoices VALUES(213,27,'2011-07-22 00:00:00','1033 N Park Ave','Tucson','AZ','USA','85719',5.9400000000000003907);
INSERT INTO invoices VALUES(214,33,'2011-07-25 00:00:00','5112 48 Street','Yellowknife','NT','Canada','X1A 1N6',8.9100000000000001421);
INSERT INTO invoices VALUES(215,42,'2011-07-30 00:00:00','9, Place Louis Barthou','Bordeaux',NULL,'France','33000',13.859999999999999431);
INSERT INTO invoices VALUES(216,56,'2011-08-07 00:00:00','307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106',0.98999999999999999111);
INSERT INTO invoices VALUES(217,57,'2011-08-20 00:00:00','Calle Lira, 198','Santiago',NULL,'Chile',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(218,59,'2011-08-20 00:00:00','3,Raj Bhavan Road','Bangalore',NULL,'India','560001',1.9799999999999999822);
INSERT INTO invoices VALUES(219,2,'2011-08-21 00:00:00','Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174',3.9599999999999999644);
INSERT INTO invoices VALUES(220,6,'2011-08-22 00:00:00','Rilská 3174/6','Prague',NULL,'Czech Republic','14300',5.9400000000000003907);
INSERT INTO invoices VALUES(221,12,'2011-08-25 00:00:00','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020',8.9100000000000001421);
INSERT INTO invoices VALUES(222,21,'2011-08-30 00:00:00','801 W 4th Street','Reno','NV','USA','89503',13.859999999999999431);
INSERT INTO invoices VALUES(223,35,'2011-09-07 00:00:00','Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,0.98999999999999999111);
INSERT INTO invoices VALUES(224,36,'2011-09-20 00:00:00','Tauentzienstraße 8','Berlin',NULL,'Germany','10789',1.9799999999999999822);
INSERT INTO invoices VALUES(225,38,'2011-09-20 00:00:00','Barbarossastraße 19','Berlin',NULL,'Germany','10779',1.9799999999999999822);
INSERT INTO invoices VALUES(226,40,'2011-09-21 00:00:00','8, Rue Hanovre','Paris',NULL,'France','75002',3.9599999999999999644);
INSERT INTO invoices VALUES(227,44,'2011-09-22 00:00:00','Porthaninkatu 9','Helsinki',NULL,'Finland','00530',5.9400000000000003907);
INSERT INTO invoices VALUES(228,50,'2011-09-25 00:00:00','C/ San Bernardo 85','Madrid',NULL,'Spain','28015',8.9100000000000001421);
INSERT INTO invoices VALUES(229,59,'2011-09-30 00:00:00','3,Raj Bhavan Road','Bangalore',NULL,'India','560001',13.859999999999999431);
INSERT INTO invoices VALUES(230,14,'2011-10-08 00:00:00','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7',0.98999999999999999111);
INSERT INTO invoices VALUES(231,15,'2011-10-21 00:00:00','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8',1.9799999999999999822);
INSERT INTO invoices VALUES(232,17,'2011-10-21 00:00:00','1 Microsoft Way','Redmond','WA','USA','98052-8300',1.9799999999999999822);
INSERT INTO invoices VALUES(233,19,'2011-10-22 00:00:00','1 Infinite Loop','Cupertino','CA','USA','95014',3.9599999999999999644);
INSERT INTO invoices VALUES(234,23,'2011-10-23 00:00:00','69 Salem Street','Boston','MA','USA','2113',5.9400000000000003907);
INSERT INTO invoices VALUES(235,29,'2011-10-26 00:00:00','796 Dundas Street West','Toronto','ON','Canada','M6J 1V1',8.9100000000000001421);
INSERT INTO invoices VALUES(236,38,'2011-10-31 00:00:00','Barbarossastraße 19','Berlin',NULL,'Germany','10779',13.859999999999999431);
INSERT INTO invoices VALUES(237,52,'2011-11-08 00:00:00','202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH',0.98999999999999999111);
INSERT INTO invoices VALUES(238,53,'2011-11-21 00:00:00','113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN',1.9799999999999999822);
INSERT INTO invoices VALUES(239,55,'2011-11-21 00:00:00','421 Bourke Street','Sidney','NSW','Australia','2010',1.9799999999999999822);
INSERT INTO invoices VALUES(240,57,'2011-11-22 00:00:00','Calle Lira, 198','Santiago',NULL,'Chile',NULL,3.9599999999999999644);
INSERT INTO invoices VALUES(241,2,'2011-11-23 00:00:00','Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174',5.9400000000000003907);
INSERT INTO invoices VALUES(242,8,'2011-11-26 00:00:00','Grétrystraat 63','Brussels',NULL,'Belgium','1000',8.9100000000000001421);
INSERT INTO invoices VALUES(243,17,'2011-12-01 00:00:00','1 Microsoft Way','Redmond','WA','USA','98052-8300',13.859999999999999431);
INSERT INTO invoices VALUES(244,31,'2011-12-09 00:00:00','194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5',0.98999999999999999111);
INSERT INTO invoices VALUES(245,32,'2011-12-22 00:00:00','696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9',1.9799999999999999822);
INSERT INTO invoices VALUES(246,34,'2011-12-22 00:00:00','Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(247,36,'2011-12-23 00:00:00','Tauentzienstraße 8','Berlin',NULL,'Germany','10789',3.9599999999999999644);
INSERT INTO invoices VALUES(248,40,'2011-12-24 00:00:00','8, Rue Hanovre','Paris',NULL,'France','75002',5.9400000000000003907);
INSERT INTO invoices VALUES(249,46,'2011-12-27 00:00:00','3 Chatham Street','Dublin','Dublin','Ireland',NULL,8.9100000000000001421);
INSERT INTO invoices VALUES(250,55,'2012-01-01 00:00:00','421 Bourke Street','Sidney','NSW','Australia','2010',13.859999999999999431);
INSERT INTO invoices VALUES(251,10,'2012-01-09 00:00:00','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010',0.98999999999999999111);
INSERT INTO invoices VALUES(252,11,'2012-01-22 00:00:00','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200',1.9799999999999999822);
INSERT INTO invoices VALUES(253,13,'2012-01-22 00:00:00','Qe 7 Bloco G','Brasília','DF','Brazil','71020-677',1.9799999999999999822);
INSERT INTO invoices VALUES(254,15,'2012-01-23 00:00:00','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8',3.9599999999999999644);
INSERT INTO invoices VALUES(255,19,'2012-01-24 00:00:00','1 Infinite Loop','Cupertino','CA','USA','95014',5.9400000000000003907);
INSERT INTO invoices VALUES(256,25,'2012-01-27 00:00:00','319 N. Frances Street','Madison','WI','USA','53703',8.9100000000000001421);
INSERT INTO invoices VALUES(257,34,'2012-02-01 00:00:00','Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,13.859999999999999431);
INSERT INTO invoices VALUES(258,48,'2012-02-09 00:00:00','Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016',0.98999999999999999111);
INSERT INTO invoices VALUES(259,49,'2012-02-22 00:00:00','Ordynacka 10','Warsaw',NULL,'Poland','00-358',1.9799999999999999822);
INSERT INTO invoices VALUES(260,51,'2012-02-22 00:00:00','Celsiusg. 9','Stockholm',NULL,'Sweden','11230',1.9799999999999999822);
INSERT INTO invoices VALUES(261,53,'2012-02-23 00:00:00','113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN',3.9599999999999999644);
INSERT INTO invoices VALUES(262,57,'2012-02-24 00:00:00','Calle Lira, 198','Santiago',NULL,'Chile',NULL,5.9400000000000003907);
INSERT INTO invoices VALUES(263,4,'2012-02-27 00:00:00','Ullevålsveien 14','Oslo',NULL,'Norway','0171',8.9100000000000001421);
INSERT INTO invoices VALUES(264,13,'2012-03-03 00:00:00','Qe 7 Bloco G','Brasília','DF','Brazil','71020-677',13.859999999999999431);
INSERT INTO invoices VALUES(265,27,'2012-03-11 00:00:00','1033 N Park Ave','Tucson','AZ','USA','85719',0.98999999999999999111);
INSERT INTO invoices VALUES(266,28,'2012-03-24 00:00:00','302 S 700 E','Salt Lake City','UT','USA','84102',1.9799999999999999822);
INSERT INTO invoices VALUES(267,30,'2012-03-24 00:00:00','230 Elgin Street','Ottawa','ON','Canada','K2P 1L7',1.9799999999999999822);
INSERT INTO invoices VALUES(268,32,'2012-03-25 00:00:00','696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9',3.9599999999999999644);
INSERT INTO invoices VALUES(269,36,'2012-03-26 00:00:00','Tauentzienstraße 8','Berlin',NULL,'Germany','10789',5.9400000000000003907);
INSERT INTO invoices VALUES(270,42,'2012-03-29 00:00:00','9, Place Louis Barthou','Bordeaux',NULL,'France','33000',8.9100000000000001421);
INSERT INTO invoices VALUES(271,51,'2012-04-03 00:00:00','Celsiusg. 9','Stockholm',NULL,'Sweden','11230',13.859999999999999431);
INSERT INTO invoices VALUES(272,6,'2012-04-11 00:00:00','Rilská 3174/6','Prague',NULL,'Czech Republic','14300',0.98999999999999999111);
INSERT INTO invoices VALUES(273,7,'2012-04-24 00:00:00','Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010',1.9799999999999999822);
INSERT INTO invoices VALUES(274,9,'2012-04-24 00:00:00','Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720',1.9799999999999999822);
INSERT INTO invoices VALUES(275,11,'2012-04-25 00:00:00','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200',3.9599999999999999644);
INSERT INTO invoices VALUES(276,15,'2012-04-26 00:00:00','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8',5.9400000000000003907);
INSERT INTO invoices VALUES(277,21,'2012-04-29 00:00:00','801 W 4th Street','Reno','NV','USA','89503',8.9100000000000001421);
INSERT INTO invoices VALUES(278,30,'2012-05-04 00:00:00','230 Elgin Street','Ottawa','ON','Canada','K2P 1L7',13.859999999999999431);
INSERT INTO invoices VALUES(279,44,'2012-05-12 00:00:00','Porthaninkatu 9','Helsinki',NULL,'Finland','00530',0.98999999999999999111);
INSERT INTO invoices VALUES(280,45,'2012-05-25 00:00:00','Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',1.9799999999999999822);
INSERT INTO invoices VALUES(281,47,'2012-05-25 00:00:00','Via Degli Scipioni, 43','Rome','RM','Italy','00192',1.9799999999999999822);
INSERT INTO invoices VALUES(282,49,'2012-05-26 00:00:00','Ordynacka 10','Warsaw',NULL,'Poland','00-358',3.9599999999999999644);
INSERT INTO invoices VALUES(283,53,'2012-05-27 00:00:00','113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN',5.9400000000000003907);
INSERT INTO invoices VALUES(284,59,'2012-05-30 00:00:00','3,Raj Bhavan Road','Bangalore',NULL,'India','560001',8.9100000000000001421);
INSERT INTO invoices VALUES(285,9,'2012-06-04 00:00:00','Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720',13.859999999999999431);
INSERT INTO invoices VALUES(286,23,'2012-06-12 00:00:00','69 Salem Street','Boston','MA','USA','2113',0.98999999999999999111);
INSERT INTO invoices VALUES(287,24,'2012-06-25 00:00:00','162 E Superior Street','Chicago','IL','USA','60611',1.9799999999999999822);
INSERT INTO invoices VALUES(288,26,'2012-06-25 00:00:00','2211 W Berry Street','Fort Worth','TX','USA','76110',1.9799999999999999822);
INSERT INTO invoices VALUES(289,28,'2012-06-26 00:00:00','302 S 700 E','Salt Lake City','UT','USA','84102',3.9599999999999999644);
INSERT INTO invoices VALUES(290,32,'2012-06-27 00:00:00','696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9',5.9400000000000003907);
INSERT INTO invoices VALUES(291,38,'2012-06-30 00:00:00','Barbarossastraße 19','Berlin',NULL,'Germany','10779',8.9100000000000001421);
INSERT INTO invoices VALUES(292,47,'2012-07-05 00:00:00','Via Degli Scipioni, 43','Rome','RM','Italy','00192',13.859999999999999431);
INSERT INTO invoices VALUES(293,2,'2012-07-13 00:00:00','Theodor-Heuss-Straße 34','Stuttgart',NULL,'Germany','70174',0.98999999999999999111);
INSERT INTO invoices VALUES(294,3,'2012-07-26 00:00:00','1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7',1.9799999999999999822);
INSERT INTO invoices VALUES(295,5,'2012-07-26 00:00:00','Klanova 9/506','Prague',NULL,'Czech Republic','14700',1.9799999999999999822);
INSERT INTO invoices VALUES(296,7,'2012-07-27 00:00:00','Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010',3.9599999999999999644);
INSERT INTO invoices VALUES(297,11,'2012-07-28 00:00:00','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200',5.9400000000000003907);
INSERT INTO invoices VALUES(298,17,'2012-07-31 00:00:00','1 Microsoft Way','Redmond','WA','USA','98052-8300',10.910000000000000142);
INSERT INTO invoices VALUES(299,26,'2012-08-05 00:00:00','2211 W Berry Street','Fort Worth','TX','USA','76110',23.859999999999999431);
INSERT INTO invoices VALUES(300,40,'2012-08-13 00:00:00','8, Rue Hanovre','Paris',NULL,'France','75002',0.98999999999999999111);
INSERT INTO invoices VALUES(301,41,'2012-08-26 00:00:00','11, Place Bellecour','Lyon',NULL,'France','69002',1.9799999999999999822);
INSERT INTO invoices VALUES(302,43,'2012-08-26 00:00:00','68, Rue Jouvence','Dijon',NULL,'France','21000',1.9799999999999999822);
INSERT INTO invoices VALUES(303,45,'2012-08-27 00:00:00','Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',3.9599999999999999644);
INSERT INTO invoices VALUES(304,49,'2012-08-28 00:00:00','Ordynacka 10','Warsaw',NULL,'Poland','00-358',5.9400000000000003907);
INSERT INTO invoices VALUES(305,55,'2012-08-31 00:00:00','421 Bourke Street','Sidney','NSW','Australia','2010',8.9100000000000001421);
INSERT INTO invoices VALUES(306,5,'2012-09-05 00:00:00','Klanova 9/506','Prague',NULL,'Czech Republic','14700',16.859999999999999431);
INSERT INTO invoices VALUES(307,19,'2012-09-13 00:00:00','1 Infinite Loop','Cupertino','CA','USA','95014',1.9899999999999999911);
INSERT INTO invoices VALUES(308,20,'2012-09-26 00:00:00','541 Del Medio Avenue','Mountain View','CA','USA','94040-111',3.9799999999999999822);
INSERT INTO invoices VALUES(309,22,'2012-09-26 00:00:00','120 S Orange Ave','Orlando','FL','USA','32801',3.9799999999999999822);
INSERT INTO invoices VALUES(310,24,'2012-09-27 00:00:00','162 E Superior Street','Chicago','IL','USA','60611',7.9599999999999999644);
INSERT INTO invoices VALUES(311,28,'2012-09-28 00:00:00','302 S 700 E','Salt Lake City','UT','USA','84102',11.939999999999999502);
INSERT INTO invoices VALUES(312,34,'2012-10-01 00:00:00','Rua da Assunção 53','Lisbon',NULL,'Portugal',NULL,10.910000000000000142);
INSERT INTO invoices VALUES(313,43,'2012-10-06 00:00:00','68, Rue Jouvence','Dijon',NULL,'France','21000',16.859999999999999431);
INSERT INTO invoices VALUES(314,57,'2012-10-14 00:00:00','Calle Lira, 198','Santiago',NULL,'Chile',NULL,0.98999999999999999111);
INSERT INTO invoices VALUES(315,58,'2012-10-27 00:00:00','12,Community Centre','Delhi',NULL,'India','110017',1.9799999999999999822);
INSERT INTO invoices VALUES(316,1,'2012-10-27 00:00:00','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000',1.9799999999999999822);
INSERT INTO invoices VALUES(317,3,'2012-10-28 00:00:00','1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7',3.9599999999999999644);
INSERT INTO invoices VALUES(318,7,'2012-10-29 00:00:00','Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010',5.9400000000000003907);
INSERT INTO invoices VALUES(319,13,'2012-11-01 00:00:00','Qe 7 Bloco G','Brasília','DF','Brazil','71020-677',8.9100000000000001421);
INSERT INTO invoices VALUES(320,22,'2012-11-06 00:00:00','120 S Orange Ave','Orlando','FL','USA','32801',13.859999999999999431);
INSERT INTO invoices VALUES(321,36,'2012-11-14 00:00:00','Tauentzienstraße 8','Berlin',NULL,'Germany','10789',0.98999999999999999111);
INSERT INTO invoices VALUES(322,37,'2012-11-27 00:00:00','Berger Straße 10','Frankfurt',NULL,'Germany','60316',1.9799999999999999822);
INSERT INTO invoices VALUES(323,39,'2012-11-27 00:00:00','4, Rue Milton','Paris',NULL,'France','75009',1.9799999999999999822);
INSERT INTO invoices VALUES(324,41,'2012-11-28 00:00:00','11, Place Bellecour','Lyon',NULL,'France','69002',3.9599999999999999644);
INSERT INTO invoices VALUES(325,45,'2012-11-29 00:00:00','Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',5.9400000000000003907);
INSERT INTO invoices VALUES(326,51,'2012-12-02 00:00:00','Celsiusg. 9','Stockholm',NULL,'Sweden','11230',8.9100000000000001421);
INSERT INTO invoices VALUES(327,1,'2012-12-07 00:00:00','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000',13.859999999999999431);
INSERT INTO invoices VALUES(328,15,'2012-12-15 00:00:00','700 W Pender Street','Vancouver','BC','Canada','V6C 1G8',0.98999999999999999111);
INSERT INTO invoices VALUES(329,16,'2012-12-28 00:00:00','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351',1.9799999999999999822);
INSERT INTO invoices VALUES(330,18,'2012-12-28 00:00:00','627 Broadway','New York','NY','USA','10012-2612',1.9799999999999999822);
INSERT INTO invoices VALUES(331,20,'2012-12-29 00:00:00','541 Del Medio Avenue','Mountain View','CA','USA','94040-111',3.9599999999999999644);
INSERT INTO invoices VALUES(332,24,'2012-12-30 00:00:00','162 E Superior Street','Chicago','IL','USA','60611',5.9400000000000003907);
INSERT INTO invoices VALUES(333,30,'2013-01-02 00:00:00','230 Elgin Street','Ottawa','ON','Canada','K2P 1L7',8.9100000000000001421);
INSERT INTO invoices VALUES(334,39,'2013-01-07 00:00:00','4, Rue Milton','Paris',NULL,'France','75009',13.859999999999999431);
INSERT INTO invoices VALUES(335,53,'2013-01-15 00:00:00','113 Lupus St','London',NULL,'United Kingdom','SW1V 3EN',0.98999999999999999111);
INSERT INTO invoices VALUES(336,54,'2013-01-28 00:00:00','110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH',1.9799999999999999822);
INSERT INTO invoices VALUES(337,56,'2013-01-28 00:00:00','307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106',1.9799999999999999822);
INSERT INTO invoices VALUES(338,58,'2013-01-29 00:00:00','12,Community Centre','Delhi',NULL,'India','110017',3.9599999999999999644);
INSERT INTO invoices VALUES(339,3,'2013-01-30 00:00:00','1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7',5.9400000000000003907);
INSERT INTO invoices VALUES(340,9,'2013-02-02 00:00:00','Sønder Boulevard 51','Copenhagen',NULL,'Denmark','1720',8.9100000000000001421);
INSERT INTO invoices VALUES(341,18,'2013-02-07 00:00:00','627 Broadway','New York','NY','USA','10012-2612',13.859999999999999431);
INSERT INTO invoices VALUES(342,32,'2013-02-15 00:00:00','696 Osborne Street','Winnipeg','MB','Canada','R3L 2B9',0.98999999999999999111);
INSERT INTO invoices VALUES(343,33,'2013-02-28 00:00:00','5112 48 Street','Yellowknife','NT','Canada','X1A 1N6',1.9799999999999999822);
INSERT INTO invoices VALUES(344,35,'2013-02-28 00:00:00','Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(345,37,'2013-03-01 00:00:00','Berger Straße 10','Frankfurt',NULL,'Germany','60316',3.9599999999999999644);
INSERT INTO invoices VALUES(346,41,'2013-03-02 00:00:00','11, Place Bellecour','Lyon',NULL,'France','69002',5.9400000000000003907);
INSERT INTO invoices VALUES(347,47,'2013-03-05 00:00:00','Via Degli Scipioni, 43','Rome','RM','Italy','00192',8.9100000000000001421);
INSERT INTO invoices VALUES(348,56,'2013-03-10 00:00:00','307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106',13.859999999999999431);
INSERT INTO invoices VALUES(349,11,'2013-03-18 00:00:00','Av. Paulista, 2022','São Paulo','SP','Brazil','01310-200',0.98999999999999999111);
INSERT INTO invoices VALUES(350,12,'2013-03-31 00:00:00','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020',1.9799999999999999822);
INSERT INTO invoices VALUES(351,14,'2013-03-31 00:00:00','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7',1.9799999999999999822);
INSERT INTO invoices VALUES(352,16,'2013-04-01 00:00:00','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351',3.9599999999999999644);
INSERT INTO invoices VALUES(353,20,'2013-04-02 00:00:00','541 Del Medio Avenue','Mountain View','CA','USA','94040-111',5.9400000000000003907);
INSERT INTO invoices VALUES(354,26,'2013-04-05 00:00:00','2211 W Berry Street','Fort Worth','TX','USA','76110',8.9100000000000001421);
INSERT INTO invoices VALUES(355,35,'2013-04-10 00:00:00','Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,13.859999999999999431);
INSERT INTO invoices VALUES(356,49,'2013-04-18 00:00:00','Ordynacka 10','Warsaw',NULL,'Poland','00-358',0.98999999999999999111);
INSERT INTO invoices VALUES(357,50,'2013-05-01 00:00:00','C/ San Bernardo 85','Madrid',NULL,'Spain','28015',1.9799999999999999822);
INSERT INTO invoices VALUES(358,52,'2013-05-01 00:00:00','202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH',1.9799999999999999822);
INSERT INTO invoices VALUES(359,54,'2013-05-02 00:00:00','110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH',3.9599999999999999644);
INSERT INTO invoices VALUES(360,58,'2013-05-03 00:00:00','12,Community Centre','Delhi',NULL,'India','110017',5.9400000000000003907);
INSERT INTO invoices VALUES(361,5,'2013-05-06 00:00:00','Klanova 9/506','Prague',NULL,'Czech Republic','14700',8.9100000000000001421);
INSERT INTO invoices VALUES(362,14,'2013-05-11 00:00:00','8210 111 ST NW','Edmonton','AB','Canada','T6G 2C7',13.859999999999999431);
INSERT INTO invoices VALUES(363,28,'2013-05-19 00:00:00','302 S 700 E','Salt Lake City','UT','USA','84102',0.98999999999999999111);
INSERT INTO invoices VALUES(364,29,'2013-06-01 00:00:00','796 Dundas Street West','Toronto','ON','Canada','M6J 1V1',1.9799999999999999822);
INSERT INTO invoices VALUES(365,31,'2013-06-01 00:00:00','194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5',1.9799999999999999822);
INSERT INTO invoices VALUES(366,33,'2013-06-02 00:00:00','5112 48 Street','Yellowknife','NT','Canada','X1A 1N6',3.9599999999999999644);
INSERT INTO invoices VALUES(367,37,'2013-06-03 00:00:00','Berger Straße 10','Frankfurt',NULL,'Germany','60316',5.9400000000000003907);
INSERT INTO invoices VALUES(368,43,'2013-06-06 00:00:00','68, Rue Jouvence','Dijon',NULL,'France','21000',8.9100000000000001421);
INSERT INTO invoices VALUES(369,52,'2013-06-11 00:00:00','202 Hoxton Street','London',NULL,'United Kingdom','N1 5LH',13.859999999999999431);
INSERT INTO invoices VALUES(370,7,'2013-06-19 00:00:00','Rotenturmstraße 4, 1010 Innere Stadt','Vienne',NULL,'Austria','1010',0.98999999999999999111);
INSERT INTO invoices VALUES(371,8,'2013-07-02 00:00:00','Grétrystraat 63','Brussels',NULL,'Belgium','1000',1.9799999999999999822);
INSERT INTO invoices VALUES(372,10,'2013-07-02 00:00:00','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010',1.9799999999999999822);
INSERT INTO invoices VALUES(373,12,'2013-07-03 00:00:00','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020',3.9599999999999999644);
INSERT INTO invoices VALUES(374,16,'2013-07-04 00:00:00','1600 Amphitheatre Parkway','Mountain View','CA','USA','94043-1351',5.9400000000000003907);
INSERT INTO invoices VALUES(375,22,'2013-07-07 00:00:00','120 S Orange Ave','Orlando','FL','USA','32801',8.9100000000000001421);
INSERT INTO invoices VALUES(376,31,'2013-07-12 00:00:00','194A Chain Lake Drive','Halifax','NS','Canada','B3S 1C5',13.859999999999999431);
INSERT INTO invoices VALUES(377,45,'2013-07-20 00:00:00','Erzsébet krt. 58.','Budapest',NULL,'Hungary','H-1073',0.98999999999999999111);
INSERT INTO invoices VALUES(378,46,'2013-08-02 00:00:00','3 Chatham Street','Dublin','Dublin','Ireland',NULL,1.9799999999999999822);
INSERT INTO invoices VALUES(379,48,'2013-08-02 00:00:00','Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016',1.9799999999999999822);
INSERT INTO invoices VALUES(380,50,'2013-08-03 00:00:00','C/ San Bernardo 85','Madrid',NULL,'Spain','28015',3.9599999999999999644);
INSERT INTO invoices VALUES(381,54,'2013-08-04 00:00:00','110 Raeburn Pl','Edinburgh ',NULL,'United Kingdom','EH4 1HH',5.9400000000000003907);
INSERT INTO invoices VALUES(382,1,'2013-08-07 00:00:00','Av. Brigadeiro Faria Lima, 2170','São José dos Campos','SP','Brazil','12227-000',8.9100000000000001421);
INSERT INTO invoices VALUES(383,10,'2013-08-12 00:00:00','Rua Dr. Falcão Filho, 155','São Paulo','SP','Brazil','01007-010',13.859999999999999431);
INSERT INTO invoices VALUES(384,24,'2013-08-20 00:00:00','162 E Superior Street','Chicago','IL','USA','60611',0.98999999999999999111);
INSERT INTO invoices VALUES(385,25,'2013-09-02 00:00:00','319 N. Frances Street','Madison','WI','USA','53703',1.9799999999999999822);
INSERT INTO invoices VALUES(386,27,'2013-09-02 00:00:00','1033 N Park Ave','Tucson','AZ','USA','85719',1.9799999999999999822);
INSERT INTO invoices VALUES(387,29,'2013-09-03 00:00:00','796 Dundas Street West','Toronto','ON','Canada','M6J 1V1',3.9599999999999999644);
INSERT INTO invoices VALUES(388,33,'2013-09-04 00:00:00','5112 48 Street','Yellowknife','NT','Canada','X1A 1N6',5.9400000000000003907);
INSERT INTO invoices VALUES(389,39,'2013-09-07 00:00:00','4, Rue Milton','Paris',NULL,'France','75009',8.9100000000000001421);
INSERT INTO invoices VALUES(390,48,'2013-09-12 00:00:00','Lijnbaansgracht 120bg','Amsterdam','VV','Netherlands','1016',13.859999999999999431);
INSERT INTO invoices VALUES(391,3,'2013-09-20 00:00:00','1498 rue Bélanger','Montréal','QC','Canada','H2G 1A7',0.98999999999999999111);
INSERT INTO invoices VALUES(392,4,'2013-10-03 00:00:00','Ullevålsveien 14','Oslo',NULL,'Norway','0171',1.9799999999999999822);
INSERT INTO invoices VALUES(393,6,'2013-10-03 00:00:00','Rilská 3174/6','Prague',NULL,'Czech Republic','14300',1.9799999999999999822);
INSERT INTO invoices VALUES(394,8,'2013-10-04 00:00:00','Grétrystraat 63','Brussels',NULL,'Belgium','1000',3.9599999999999999644);
INSERT INTO invoices VALUES(395,12,'2013-10-05 00:00:00','Praça Pio X, 119','Rio de Janeiro','RJ','Brazil','20040-020',5.9400000000000003907);
INSERT INTO invoices VALUES(396,18,'2013-10-08 00:00:00','627 Broadway','New York','NY','USA','10012-2612',8.9100000000000001421);
INSERT INTO invoices VALUES(397,27,'2013-10-13 00:00:00','1033 N Park Ave','Tucson','AZ','USA','85719',13.859999999999999431);
INSERT INTO invoices VALUES(398,41,'2013-10-21 00:00:00','11, Place Bellecour','Lyon',NULL,'France','69002',0.98999999999999999111);
INSERT INTO invoices VALUES(399,42,'2013-11-03 00:00:00','9, Place Louis Barthou','Bordeaux',NULL,'France','33000',1.9799999999999999822);
INSERT INTO invoices VALUES(400,44,'2013-11-03 00:00:00','Porthaninkatu 9','Helsinki',NULL,'Finland','00530',1.9799999999999999822);
INSERT INTO invoices VALUES(401,46,'2013-11-04 00:00:00','3 Chatham Street','Dublin','Dublin','Ireland',NULL,3.9599999999999999644);
INSERT INTO invoices VALUES(402,50,'2013-11-05 00:00:00','C/ San Bernardo 85','Madrid',NULL,'Spain','28015',5.9400000000000003907);
INSERT INTO invoices VALUES(403,56,'2013-11-08 00:00:00','307 Macacha Güemes','Buenos Aires',NULL,'Argentina','1106',8.9100000000000001421);
INSERT INTO invoices VALUES(404,6,'2013-11-13 00:00:00','Rilská 3174/6','Prague',NULL,'Czech Republic','14300',25.859999999999999432);
INSERT INTO invoices VALUES(405,20,'2013-11-21 00:00:00','541 Del Medio Avenue','Mountain View','CA','USA','94040-111',0.98999999999999999111);
INSERT INTO invoices VALUES(406,21,'2013-12-04 00:00:00','801 W 4th Street','Reno','NV','USA','89503',1.9799999999999999822);
INSERT INTO invoices VALUES(407,23,'2013-12-04 00:00:00','69 Salem Street','Boston','MA','USA','2113',1.9799999999999999822);
INSERT INTO invoices VALUES(408,25,'2013-12-05 00:00:00','319 N. Frances Street','Madison','WI','USA','53703',3.9599999999999999644);
INSERT INTO invoices VALUES(409,29,'2013-12-06 00:00:00','796 Dundas Street West','Toronto','ON','Canada','M6J 1V1',5.9400000000000003907);
INSERT INTO invoices VALUES(410,35,'2013-12-09 00:00:00','Rua dos Campeões Europeus de Viena, 4350','Porto',NULL,'Portugal',NULL,8.9100000000000001421);
INSERT INTO invoices VALUES(411,44,'2013-12-14 00:00:00','Porthaninkatu 9','Helsinki',NULL,'Finland','00530',13.859999999999999431);
INSERT INTO invoices VALUES(412,58,'2013-12-22 00:00:00','12,Community Centre','Delhi',NULL,'India','110017',1.9899999999999999911);
CREATE TABLE IF NOT EXISTS "invoice_items"
(
    [InvoiceLineId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [InvoiceId] INTEGER  NOT NULL,
    [TrackId] INTEGER  NOT NULL,
    [UnitPrice] NUMERIC(10,2)  NOT NULL,
    [Quantity] INTEGER  NOT NULL,
    FOREIGN KEY ([InvoiceId]) REFERENCES "invoices" ([InvoiceId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION,
    FOREIGN KEY ([TrackId]) REFERENCES "tracks" ([TrackId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION
);
INSERT INTO invoice_items VALUES(1,1,2,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2,1,4,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(3,2,6,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(4,2,8,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(5,2,10,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(6,2,12,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(7,3,16,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(8,3,20,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(9,3,24,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(10,3,28,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(11,3,32,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(12,3,36,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(13,4,42,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(14,4,48,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(15,4,54,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(16,4,60,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(17,4,66,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(18,4,72,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(19,4,78,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(20,4,84,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(21,4,90,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(22,5,99,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(23,5,108,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(24,5,117,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(25,5,126,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(26,5,135,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(27,5,144,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(28,5,153,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(29,5,162,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(30,5,171,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(31,5,180,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(32,5,189,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(33,5,198,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(34,5,207,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(35,5,216,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(36,6,230,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(37,7,231,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(38,7,232,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(39,8,234,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(40,8,236,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(41,9,238,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(42,9,240,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(43,9,242,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(44,9,244,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(45,10,248,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(46,10,252,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(47,10,256,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(48,10,260,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(49,10,264,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(50,10,268,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(51,11,274,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(52,11,280,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(53,11,286,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(54,11,292,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(55,11,298,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(56,11,304,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(57,11,310,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(58,11,316,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(59,11,322,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(60,12,331,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(61,12,340,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(62,12,349,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(63,12,358,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(64,12,367,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(65,12,376,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(66,12,385,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(67,12,394,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(68,12,403,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(69,12,412,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(70,12,421,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(71,12,430,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(72,12,439,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(73,12,448,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(74,13,462,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(75,14,463,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(76,14,464,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(77,15,466,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(78,15,468,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(79,16,470,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(80,16,472,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(81,16,474,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(82,16,476,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(83,17,480,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(84,17,484,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(85,17,488,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(86,17,492,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(87,17,496,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(88,17,500,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(89,18,506,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(90,18,512,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(91,18,518,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(92,18,524,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(93,18,530,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(94,18,536,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(95,18,542,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(96,18,548,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(97,18,554,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(98,19,563,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(99,19,572,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(100,19,581,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(101,19,590,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(102,19,599,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(103,19,608,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(104,19,617,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(105,19,626,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(106,19,635,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(107,19,644,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(108,19,653,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(109,19,662,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(110,19,671,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(111,19,680,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(112,20,694,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(113,21,695,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(114,21,696,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(115,22,698,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(116,22,700,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(117,23,702,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(118,23,704,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(119,23,706,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(120,23,708,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(121,24,712,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(122,24,716,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(123,24,720,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(124,24,724,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(125,24,728,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(126,24,732,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(127,25,738,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(128,25,744,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(129,25,750,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(130,25,756,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(131,25,762,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(132,25,768,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(133,25,774,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(134,25,780,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(135,25,786,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(136,26,795,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(137,26,804,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(138,26,813,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(139,26,822,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(140,26,831,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(141,26,840,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(142,26,849,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(143,26,858,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(144,26,867,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(145,26,876,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(146,26,885,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(147,26,894,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(148,26,903,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(149,26,912,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(150,27,926,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(151,28,927,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(152,28,928,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(153,29,930,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(154,29,932,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(155,30,934,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(156,30,936,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(157,30,938,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(158,30,940,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(159,31,944,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(160,31,948,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(161,31,952,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(162,31,956,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(163,31,960,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(164,31,964,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(165,32,970,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(166,32,976,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(167,32,982,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(168,32,988,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(169,32,994,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(170,32,1000,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(171,32,1006,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(172,32,1012,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(173,32,1018,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(174,33,1027,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(175,33,1036,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(176,33,1045,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(177,33,1054,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(178,33,1063,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(179,33,1072,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(180,33,1081,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(181,33,1090,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(182,33,1099,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(183,33,1108,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(184,33,1117,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(185,33,1126,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(186,33,1135,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(187,33,1144,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(188,34,1158,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(189,35,1159,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(190,35,1160,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(191,36,1162,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(192,36,1164,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(193,37,1166,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(194,37,1168,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(195,37,1170,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(196,37,1172,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(197,38,1176,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(198,38,1180,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(199,38,1184,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(200,38,1188,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(201,38,1192,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(202,38,1196,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(203,39,1202,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(204,39,1208,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(205,39,1214,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(206,39,1220,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(207,39,1226,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(208,39,1232,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(209,39,1238,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(210,39,1244,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(211,39,1250,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(212,40,1259,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(213,40,1268,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(214,40,1277,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(215,40,1286,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(216,40,1295,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(217,40,1304,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(218,40,1313,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(219,40,1322,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(220,40,1331,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(221,40,1340,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(222,40,1349,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(223,40,1358,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(224,40,1367,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(225,40,1376,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(226,41,1390,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(227,42,1391,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(228,42,1392,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(229,43,1394,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(230,43,1396,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(231,44,1398,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(232,44,1400,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(233,44,1402,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(234,44,1404,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(235,45,1408,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(236,45,1412,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(237,45,1416,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(238,45,1420,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(239,45,1424,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(240,45,1428,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(241,46,1434,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(242,46,1440,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(243,46,1446,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(244,46,1452,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(245,46,1458,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(246,46,1464,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(247,46,1470,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(248,46,1476,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(249,46,1482,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(250,47,1491,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(251,47,1500,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(252,47,1509,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(253,47,1518,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(254,47,1527,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(255,47,1536,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(256,47,1545,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(257,47,1554,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(258,47,1563,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(259,47,1572,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(260,47,1581,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(261,47,1590,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(262,47,1599,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(263,47,1608,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(264,48,1622,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(265,49,1623,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(266,49,1624,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(267,50,1626,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(268,50,1628,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(269,51,1630,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(270,51,1632,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(271,51,1634,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(272,51,1636,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(273,52,1640,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(274,52,1644,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(275,52,1648,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(276,52,1652,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(277,52,1656,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(278,52,1660,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(279,53,1666,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(280,53,1672,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(281,53,1678,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(282,53,1684,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(283,53,1690,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(284,53,1696,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(285,53,1702,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(286,53,1708,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(287,53,1714,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(288,54,1723,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(289,54,1732,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(290,54,1741,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(291,54,1750,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(292,54,1759,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(293,54,1768,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(294,54,1777,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(295,54,1786,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(296,54,1795,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(297,54,1804,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(298,54,1813,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(299,54,1822,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(300,54,1831,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(301,54,1840,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(302,55,1854,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(303,56,1855,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(304,56,1856,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(305,57,1858,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(306,57,1860,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(307,58,1862,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(308,58,1864,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(309,58,1866,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(310,58,1868,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(311,59,1872,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(312,59,1876,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(313,59,1880,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(314,59,1884,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(315,59,1888,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(316,59,1892,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(317,60,1898,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(318,60,1904,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(319,60,1910,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(320,60,1916,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(321,60,1922,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(322,60,1928,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(323,60,1934,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(324,60,1940,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(325,60,1946,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(326,61,1955,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(327,61,1964,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(328,61,1973,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(329,61,1982,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(330,61,1991,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(331,61,2000,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(332,61,2009,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(333,61,2018,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(334,61,2027,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(335,61,2036,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(336,61,2045,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(337,61,2054,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(338,61,2063,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(339,61,2072,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(340,62,2086,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(341,63,2087,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(342,63,2088,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(343,64,2090,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(344,64,2092,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(345,65,2094,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(346,65,2096,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(347,65,2098,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(348,65,2100,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(349,66,2104,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(350,66,2108,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(351,66,2112,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(352,66,2116,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(353,66,2120,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(354,66,2124,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(355,67,2130,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(356,67,2136,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(357,67,2142,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(358,67,2148,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(359,67,2154,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(360,67,2160,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(361,67,2166,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(362,67,2172,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(363,67,2178,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(364,68,2187,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(365,68,2196,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(366,68,2205,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(367,68,2214,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(368,68,2223,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(369,68,2232,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(370,68,2241,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(371,68,2250,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(372,68,2259,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(373,68,2268,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(374,68,2277,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(375,68,2286,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(376,68,2295,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(377,68,2304,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(378,69,2318,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(379,70,2319,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(380,70,2320,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(381,71,2322,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(382,71,2324,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(383,72,2326,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(384,72,2328,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(385,72,2330,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(386,72,2332,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(387,73,2336,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(388,73,2340,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(389,73,2344,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(390,73,2348,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(391,73,2352,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(392,73,2356,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(393,74,2362,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(394,74,2368,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(395,74,2374,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(396,74,2380,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(397,74,2386,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(398,74,2392,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(399,74,2398,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(400,74,2404,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(401,74,2410,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(402,75,2419,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(403,75,2428,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(404,75,2437,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(405,75,2446,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(406,75,2455,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(407,75,2464,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(408,75,2473,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(409,75,2482,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(410,75,2491,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(411,75,2500,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(412,75,2509,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(413,75,2518,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(414,75,2527,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(415,75,2536,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(416,76,2550,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(417,77,2551,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(418,77,2552,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(419,78,2554,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(420,78,2556,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(421,79,2558,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(422,79,2560,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(423,79,2562,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(424,79,2564,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(425,80,2568,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(426,80,2572,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(427,80,2576,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(428,80,2580,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(429,80,2584,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(430,80,2588,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(431,81,2594,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(432,81,2600,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(433,81,2606,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(434,81,2612,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(435,81,2618,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(436,81,2624,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(437,81,2630,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(438,81,2636,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(439,81,2642,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(440,82,2651,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(441,82,2660,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(442,82,2669,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(443,82,2678,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(444,82,2687,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(445,82,2696,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(446,82,2705,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(447,82,2714,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(448,82,2723,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(449,82,2732,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(450,82,2741,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(451,82,2750,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(452,82,2759,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(453,82,2768,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(454,83,2782,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(455,84,2783,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(456,84,2784,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(457,85,2786,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(458,85,2788,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(459,86,2790,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(460,86,2792,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(461,86,2794,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(462,86,2796,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(463,87,2800,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(464,87,2804,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(465,87,2808,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(466,87,2812,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(467,87,2816,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(468,87,2820,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(469,88,2826,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(470,88,2832,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(471,88,2838,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(472,88,2844,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(473,88,2850,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(474,88,2856,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(475,88,2862,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(476,88,2868,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(477,88,2874,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(478,89,2883,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(479,89,2892,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(480,89,2901,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(481,89,2910,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(482,89,2919,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(483,89,2928,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(484,89,2937,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(485,89,2946,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(486,89,2955,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(487,89,2964,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(488,89,2973,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(489,89,2982,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(490,89,2991,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(491,89,3000,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(492,90,3014,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(493,91,3015,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(494,91,3016,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(495,92,3018,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(496,92,3020,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(497,93,3022,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(498,93,3024,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(499,93,3026,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(500,93,3028,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(501,94,3032,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(502,94,3036,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(503,94,3040,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(504,94,3044,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(505,94,3048,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(506,94,3052,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(507,95,3058,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(508,95,3064,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(509,95,3070,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(510,95,3076,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(511,95,3082,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(512,95,3088,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(513,95,3094,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(514,95,3100,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(515,95,3106,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(516,96,3115,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(517,96,3124,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(518,96,3133,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(519,96,3142,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(520,96,3151,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(521,96,3160,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(522,96,3169,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(523,96,3178,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(524,96,3187,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(525,96,3196,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(526,96,3205,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(527,96,3214,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(528,96,3223,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(529,96,3232,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(530,97,3246,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(531,98,3247,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(532,98,3248,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(533,99,3250,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(534,99,3252,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(535,100,3254,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(536,100,3256,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(537,100,3258,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(538,100,3260,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(539,101,3264,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(540,101,3268,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(541,101,3272,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(542,101,3276,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(543,101,3280,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(544,101,3284,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(545,102,3290,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(546,102,3296,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(547,102,3302,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(548,102,3308,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(549,102,3314,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(550,102,3320,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(551,102,3326,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(552,102,3332,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(553,102,3338,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(554,103,3347,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(555,103,3356,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(556,103,3365,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(557,103,3374,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(558,103,3383,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(559,103,3392,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(560,103,3401,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(561,103,3410,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(562,103,3419,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(563,103,3428,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(564,103,3437,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(565,103,3446,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(566,103,3455,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(567,103,3464,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(568,104,3478,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(569,105,3479,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(570,105,3480,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(571,106,3482,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(572,106,3484,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(573,107,3486,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(574,107,3488,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(575,107,3490,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(576,107,3492,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(577,108,3496,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(578,108,3500,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(579,108,1,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(580,108,5,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(581,108,9,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(582,108,13,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(583,109,19,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(584,109,25,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(585,109,31,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(586,109,37,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(587,109,43,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(588,109,49,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(589,109,55,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(590,109,61,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(591,109,67,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(592,110,76,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(593,110,85,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(594,110,94,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(595,110,103,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(596,110,112,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(597,110,121,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(598,110,130,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(599,110,139,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(600,110,148,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(601,110,157,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(602,110,166,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(603,110,175,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(604,110,184,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(605,110,193,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(606,111,207,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(607,112,208,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(608,112,209,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(609,113,211,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(610,113,213,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(611,114,215,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(612,114,217,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(613,114,219,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(614,114,221,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(615,115,225,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(616,115,229,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(617,115,233,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(618,115,237,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(619,115,241,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(620,115,245,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(621,116,251,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(622,116,257,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(623,116,263,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(624,116,269,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(625,116,275,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(626,116,281,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(627,116,287,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(628,116,293,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(629,116,299,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(630,117,308,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(631,117,317,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(632,117,326,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(633,117,335,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(634,117,344,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(635,117,353,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(636,117,362,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(637,117,371,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(638,117,380,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(639,117,389,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(640,117,398,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(641,117,407,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(642,117,416,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(643,117,425,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(644,118,439,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(645,119,440,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(646,119,441,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(647,120,443,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(648,120,445,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(649,121,447,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(650,121,449,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(651,121,451,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(652,121,453,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(653,122,457,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(654,122,461,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(655,122,465,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(656,122,469,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(657,122,473,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(658,122,477,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(659,123,483,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(660,123,489,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(661,123,495,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(662,123,501,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(663,123,507,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(664,123,513,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(665,123,519,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(666,123,525,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(667,123,531,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(668,124,540,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(669,124,549,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(670,124,558,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(671,124,567,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(672,124,576,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(673,124,585,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(674,124,594,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(675,124,603,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(676,124,612,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(677,124,621,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(678,124,630,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(679,124,639,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(680,124,648,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(681,124,657,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(682,125,671,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(683,126,672,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(684,126,673,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(685,127,675,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(686,127,677,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(687,128,679,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(688,128,681,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(689,128,683,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(690,128,685,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(691,129,689,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(692,129,693,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(693,129,697,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(694,129,701,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(695,129,705,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(696,129,709,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(697,130,715,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(698,130,721,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(699,130,727,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(700,130,733,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(701,130,739,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(702,130,745,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(703,130,751,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(704,130,757,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(705,130,763,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(706,131,772,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(707,131,781,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(708,131,790,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(709,131,799,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(710,131,808,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(711,131,817,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(712,131,826,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(713,131,835,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(714,131,844,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(715,131,853,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(716,131,862,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(717,131,871,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(718,131,880,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(719,131,889,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(720,132,903,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(721,133,904,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(722,133,905,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(723,134,907,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(724,134,909,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(725,135,911,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(726,135,913,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(727,135,915,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(728,135,917,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(729,136,921,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(730,136,925,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(731,136,929,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(732,136,933,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(733,136,937,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(734,136,941,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(735,137,947,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(736,137,953,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(737,137,959,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(738,137,965,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(739,137,971,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(740,137,977,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(741,137,983,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(742,137,989,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(743,137,995,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(744,138,1004,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(745,138,1013,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(746,138,1022,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(747,138,1031,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(748,138,1040,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(749,138,1049,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(750,138,1058,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(751,138,1067,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(752,138,1076,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(753,138,1085,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(754,138,1094,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(755,138,1103,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(756,138,1112,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(757,138,1121,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(758,139,1135,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(759,140,1136,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(760,140,1137,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(761,141,1139,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(762,141,1141,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(763,142,1143,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(764,142,1145,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(765,142,1147,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(766,142,1149,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(767,143,1153,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(768,143,1157,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(769,143,1161,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(770,143,1165,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(771,143,1169,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(772,143,1173,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(773,144,1179,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(774,144,1185,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(775,144,1191,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(776,144,1197,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(777,144,1203,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(778,144,1209,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(779,144,1215,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(780,144,1221,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(781,144,1227,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(782,145,1236,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(783,145,1245,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(784,145,1254,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(785,145,1263,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(786,145,1272,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(787,145,1281,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(788,145,1290,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(789,145,1299,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(790,145,1308,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(791,145,1317,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(792,145,1326,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(793,145,1335,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(794,145,1344,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(795,145,1353,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(796,146,1367,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(797,147,1368,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(798,147,1369,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(799,148,1371,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(800,148,1373,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(801,149,1375,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(802,149,1377,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(803,149,1379,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(804,149,1381,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(805,150,1385,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(806,150,1389,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(807,150,1393,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(808,150,1397,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(809,150,1401,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(810,150,1405,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(811,151,1411,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(812,151,1417,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(813,151,1423,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(814,151,1429,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(815,151,1435,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(816,151,1441,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(817,151,1447,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(818,151,1453,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(819,151,1459,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(820,152,1468,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(821,152,1477,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(822,152,1486,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(823,152,1495,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(824,152,1504,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(825,152,1513,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(826,152,1522,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(827,152,1531,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(828,152,1540,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(829,152,1549,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(830,152,1558,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(831,152,1567,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(832,152,1576,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(833,152,1585,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(834,153,1599,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(835,154,1600,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(836,154,1601,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(837,155,1603,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(838,155,1605,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(839,156,1607,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(840,156,1609,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(841,156,1611,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(842,156,1613,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(843,157,1617,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(844,157,1621,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(845,157,1625,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(846,157,1629,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(847,157,1633,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(848,157,1637,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(849,158,1643,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(850,158,1649,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(851,158,1655,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(852,158,1661,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(853,158,1667,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(854,158,1673,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(855,158,1679,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(856,158,1685,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(857,158,1691,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(858,159,1700,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(859,159,1709,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(860,159,1718,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(861,159,1727,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(862,159,1736,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(863,159,1745,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(864,159,1754,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(865,159,1763,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(866,159,1772,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(867,159,1781,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(868,159,1790,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(869,159,1799,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(870,159,1808,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(871,159,1817,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(872,160,1831,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(873,161,1832,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(874,161,1833,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(875,162,1835,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(876,162,1837,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(877,163,1839,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(878,163,1841,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(879,163,1843,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(880,163,1845,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(881,164,1849,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(882,164,1853,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(883,164,1857,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(884,164,1861,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(885,164,1865,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(886,164,1869,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(887,165,1875,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(888,165,1881,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(889,165,1887,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(890,165,1893,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(891,165,1899,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(892,165,1905,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(893,165,1911,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(894,165,1917,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(895,165,1923,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(896,166,1932,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(897,166,1941,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(898,166,1950,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(899,166,1959,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(900,166,1968,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(901,166,1977,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(902,166,1986,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(903,166,1995,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(904,166,2004,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(905,166,2013,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(906,166,2022,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(907,166,2031,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(908,166,2040,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(909,166,2049,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(910,167,2063,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(911,168,2064,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(912,168,2065,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(913,169,2067,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(914,169,2069,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(915,170,2071,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(916,170,2073,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(917,170,2075,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(918,170,2077,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(919,171,2081,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(920,171,2085,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(921,171,2089,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(922,171,2093,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(923,171,2097,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(924,171,2101,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(925,172,2107,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(926,172,2113,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(927,172,2119,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(928,172,2125,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(929,172,2131,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(930,172,2137,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(931,172,2143,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(932,172,2149,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(933,172,2155,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(934,173,2164,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(935,173,2173,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(936,173,2182,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(937,173,2191,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(938,173,2200,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(939,173,2209,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(940,173,2218,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(941,173,2227,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(942,173,2236,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(943,173,2245,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(944,173,2254,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(945,173,2263,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(946,173,2272,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(947,173,2281,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(948,174,2295,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(949,175,2296,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(950,175,2297,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(951,176,2299,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(952,176,2301,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(953,177,2303,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(954,177,2305,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(955,177,2307,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(956,177,2309,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(957,178,2313,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(958,178,2317,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(959,178,2321,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(960,178,2325,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(961,178,2329,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(962,178,2333,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(963,179,2339,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(964,179,2345,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(965,179,2351,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(966,179,2357,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(967,179,2363,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(968,179,2369,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(969,179,2375,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(970,179,2381,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(971,179,2387,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(972,180,2396,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(973,180,2405,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(974,180,2414,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(975,180,2423,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(976,180,2432,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(977,180,2441,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(978,180,2450,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(979,180,2459,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(980,180,2468,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(981,180,2477,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(982,180,2486,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(983,180,2495,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(984,180,2504,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(985,180,2513,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(986,181,2527,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(987,182,2528,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(988,182,2529,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(989,183,2531,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(990,183,2533,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(991,184,2535,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(992,184,2537,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(993,184,2539,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(994,184,2541,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(995,185,2545,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(996,185,2549,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(997,185,2553,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(998,185,2557,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(999,185,2561,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1000,185,2565,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1001,186,2571,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1002,186,2577,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1003,186,2583,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1004,186,2589,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1005,186,2595,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1006,186,2601,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1007,186,2607,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1008,186,2613,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1009,186,2619,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1010,187,2628,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1011,187,2637,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1012,187,2646,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1013,187,2655,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1014,187,2664,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1015,187,2673,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1016,187,2682,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1017,187,2691,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1018,187,2700,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1019,187,2709,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1020,187,2718,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1021,187,2727,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1022,187,2736,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1023,187,2745,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1024,188,2759,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1025,189,2760,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1026,189,2761,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1027,190,2763,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1028,190,2765,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1029,191,2767,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1030,191,2769,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1031,191,2771,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1032,191,2773,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1033,192,2777,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1034,192,2781,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1035,192,2785,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1036,192,2789,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1037,192,2793,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1038,192,2797,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1039,193,2803,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1040,193,2809,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1041,193,2815,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1042,193,2821,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1043,193,2827,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1044,193,2833,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1045,193,2839,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1046,193,2845,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1047,193,2851,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1048,194,2860,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1049,194,2869,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1050,194,2878,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1051,194,2887,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1052,194,2896,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1053,194,2905,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1054,194,2914,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1055,194,2923,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1056,194,2932,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1057,194,2941,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1058,194,2950,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1059,194,2959,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1060,194,2968,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1061,194,2977,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1062,195,2991,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1063,196,2992,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1064,196,2993,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1065,197,2995,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1066,197,2997,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1067,198,2999,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1068,198,3001,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1069,198,3003,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1070,198,3005,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1071,199,3009,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1072,199,3013,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1073,199,3017,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1074,199,3021,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1075,199,3025,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1076,199,3029,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1077,200,3035,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1078,200,3041,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1079,200,3047,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1080,200,3053,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1081,200,3059,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1082,200,3065,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1083,200,3071,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1084,200,3077,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1085,200,3083,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1086,201,3092,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1087,201,3101,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1088,201,3110,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1089,201,3119,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1090,201,3128,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1091,201,3137,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1092,201,3146,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1093,201,3155,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1094,201,3164,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1095,201,3173,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1096,201,3182,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1097,201,3191,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1098,201,3200,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1099,201,3209,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1100,202,3223,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1101,203,3224,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1102,203,3225,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1103,204,3227,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1104,204,3229,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1105,205,3231,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1106,205,3233,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1107,205,3235,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1108,205,3237,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1109,206,3241,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1110,206,3245,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1111,206,3249,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1112,206,3253,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1113,206,3257,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1114,206,3261,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1115,207,3267,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1116,207,3273,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1117,207,3279,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1118,207,3285,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1119,207,3291,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1120,207,3297,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1121,207,3303,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1122,207,3309,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1123,207,3315,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1124,208,3324,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1125,208,3333,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1126,208,3342,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1127,208,3351,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1128,208,3360,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1129,208,3369,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1130,208,3378,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1131,208,3387,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1132,208,3396,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1133,208,3405,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1134,208,3414,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1135,208,3423,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1136,208,3432,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1137,208,3441,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1138,209,3455,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1139,210,3456,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1140,210,3457,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1141,211,3459,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1142,211,3461,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1143,212,3463,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1144,212,3465,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1145,212,3467,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1146,212,3469,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1147,213,3473,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1148,213,3477,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1149,213,3481,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1150,213,3485,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1151,213,3489,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1152,213,3493,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1153,214,3499,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1154,214,2,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1155,214,8,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1156,214,14,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1157,214,20,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1158,214,26,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1159,214,32,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1160,214,38,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1161,214,44,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1162,215,53,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1163,215,62,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1164,215,71,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1165,215,80,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1166,215,89,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1167,215,98,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1168,215,107,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1169,215,116,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1170,215,125,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1171,215,134,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1172,215,143,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1173,215,152,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1174,215,161,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1175,215,170,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1176,216,184,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1177,217,185,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1178,217,186,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1179,218,188,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1180,218,190,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1181,219,192,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1182,219,194,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1183,219,196,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1184,219,198,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1185,220,202,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1186,220,206,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1187,220,210,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1188,220,214,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1189,220,218,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1190,220,222,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1191,221,228,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1192,221,234,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1193,221,240,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1194,221,246,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1195,221,252,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1196,221,258,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1197,221,264,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1198,221,270,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1199,221,276,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1200,222,285,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1201,222,294,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1202,222,303,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1203,222,312,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1204,222,321,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1205,222,330,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1206,222,339,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1207,222,348,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1208,222,357,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1209,222,366,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1210,222,375,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1211,222,384,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1212,222,393,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1213,222,402,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1214,223,416,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1215,224,417,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1216,224,418,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1217,225,420,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1218,225,422,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1219,226,424,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1220,226,426,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1221,226,428,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1222,226,430,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1223,227,434,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1224,227,438,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1225,227,442,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1226,227,446,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1227,227,450,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1228,227,454,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1229,228,460,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1230,228,466,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1231,228,472,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1232,228,478,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1233,228,484,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1234,228,490,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1235,228,496,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1236,228,502,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1237,228,508,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1238,229,517,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1239,229,526,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1240,229,535,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1241,229,544,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1242,229,553,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1243,229,562,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1244,229,571,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1245,229,580,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1246,229,589,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1247,229,598,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1248,229,607,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1249,229,616,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1250,229,625,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1251,229,634,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1252,230,648,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1253,231,649,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1254,231,650,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1255,232,652,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1256,232,654,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1257,233,656,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1258,233,658,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1259,233,660,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1260,233,662,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1261,234,666,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1262,234,670,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1263,234,674,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1264,234,678,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1265,234,682,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1266,234,686,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1267,235,692,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1268,235,698,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1269,235,704,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1270,235,710,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1271,235,716,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1272,235,722,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1273,235,728,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1274,235,734,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1275,235,740,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1276,236,749,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1277,236,758,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1278,236,767,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1279,236,776,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1280,236,785,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1281,236,794,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1282,236,803,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1283,236,812,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1284,236,821,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1285,236,830,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1286,236,839,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1287,236,848,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1288,236,857,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1289,236,866,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1290,237,880,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1291,238,881,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1292,238,882,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1293,239,884,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1294,239,886,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1295,240,888,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1296,240,890,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1297,240,892,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1298,240,894,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1299,241,898,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1300,241,902,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1301,241,906,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1302,241,910,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1303,241,914,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1304,241,918,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1305,242,924,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1306,242,930,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1307,242,936,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1308,242,942,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1309,242,948,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1310,242,954,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1311,242,960,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1312,242,966,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1313,242,972,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1314,243,981,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1315,243,990,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1316,243,999,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1317,243,1008,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1318,243,1017,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1319,243,1026,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1320,243,1035,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1321,243,1044,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1322,243,1053,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1323,243,1062,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1324,243,1071,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1325,243,1080,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1326,243,1089,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1327,243,1098,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1328,244,1112,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1329,245,1113,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1330,245,1114,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1331,246,1116,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1332,246,1118,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1333,247,1120,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1334,247,1122,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1335,247,1124,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1336,247,1126,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1337,248,1130,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1338,248,1134,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1339,248,1138,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1340,248,1142,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1341,248,1146,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1342,248,1150,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1343,249,1156,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1344,249,1162,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1345,249,1168,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1346,249,1174,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1347,249,1180,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1348,249,1186,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1349,249,1192,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1350,249,1198,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1351,249,1204,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1352,250,1213,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1353,250,1222,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1354,250,1231,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1355,250,1240,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1356,250,1249,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1357,250,1258,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1358,250,1267,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1359,250,1276,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1360,250,1285,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1361,250,1294,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1362,250,1303,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1363,250,1312,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1364,250,1321,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1365,250,1330,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1366,251,1344,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1367,252,1345,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1368,252,1346,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1369,253,1348,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1370,253,1350,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1371,254,1352,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1372,254,1354,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1373,254,1356,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1374,254,1358,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1375,255,1362,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1376,255,1366,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1377,255,1370,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1378,255,1374,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1379,255,1378,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1380,255,1382,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1381,256,1388,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1382,256,1394,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1383,256,1400,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1384,256,1406,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1385,256,1412,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1386,256,1418,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1387,256,1424,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1388,256,1430,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1389,256,1436,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1390,257,1445,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1391,257,1454,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1392,257,1463,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1393,257,1472,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1394,257,1481,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1395,257,1490,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1396,257,1499,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1397,257,1508,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1398,257,1517,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1399,257,1526,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1400,257,1535,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1401,257,1544,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1402,257,1553,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1403,257,1562,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1404,258,1576,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1405,259,1577,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1406,259,1578,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1407,260,1580,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1408,260,1582,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1409,261,1584,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1410,261,1586,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1411,261,1588,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1412,261,1590,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1413,262,1594,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1414,262,1598,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1415,262,1602,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1416,262,1606,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1417,262,1610,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1418,262,1614,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1419,263,1620,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1420,263,1626,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1421,263,1632,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1422,263,1638,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1423,263,1644,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1424,263,1650,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1425,263,1656,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1426,263,1662,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1427,263,1668,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1428,264,1677,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1429,264,1686,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1430,264,1695,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1431,264,1704,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1432,264,1713,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1433,264,1722,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1434,264,1731,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1435,264,1740,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1436,264,1749,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1437,264,1758,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1438,264,1767,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1439,264,1776,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1440,264,1785,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1441,264,1794,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1442,265,1808,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1443,266,1809,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1444,266,1810,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1445,267,1812,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1446,267,1814,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1447,268,1816,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1448,268,1818,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1449,268,1820,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1450,268,1822,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1451,269,1826,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1452,269,1830,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1453,269,1834,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1454,269,1838,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1455,269,1842,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1456,269,1846,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1457,270,1852,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1458,270,1858,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1459,270,1864,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1460,270,1870,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1461,270,1876,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1462,270,1882,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1463,270,1888,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1464,270,1894,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1465,270,1900,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1466,271,1909,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1467,271,1918,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1468,271,1927,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1469,271,1936,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1470,271,1945,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1471,271,1954,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1472,271,1963,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1473,271,1972,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1474,271,1981,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1475,271,1990,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1476,271,1999,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1477,271,2008,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1478,271,2017,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1479,271,2026,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1480,272,2040,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1481,273,2041,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1482,273,2042,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1483,274,2044,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1484,274,2046,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1485,275,2048,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1486,275,2050,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1487,275,2052,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1488,275,2054,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1489,276,2058,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1490,276,2062,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1491,276,2066,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1492,276,2070,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1493,276,2074,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1494,276,2078,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1495,277,2084,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1496,277,2090,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1497,277,2096,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1498,277,2102,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1499,277,2108,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1500,277,2114,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1501,277,2120,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1502,277,2126,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1503,277,2132,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1504,278,2141,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1505,278,2150,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1506,278,2159,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1507,278,2168,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1508,278,2177,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1509,278,2186,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1510,278,2195,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1511,278,2204,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1512,278,2213,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1513,278,2222,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1514,278,2231,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1515,278,2240,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1516,278,2249,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1517,278,2258,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1518,279,2272,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1519,280,2273,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1520,280,2274,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1521,281,2276,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1522,281,2278,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1523,282,2280,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1524,282,2282,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1525,282,2284,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1526,282,2286,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1527,283,2290,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1528,283,2294,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1529,283,2298,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1530,283,2302,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1531,283,2306,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1532,283,2310,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1533,284,2316,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1534,284,2322,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1535,284,2328,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1536,284,2334,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1537,284,2340,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1538,284,2346,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1539,284,2352,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1540,284,2358,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1541,284,2364,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1542,285,2373,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1543,285,2382,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1544,285,2391,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1545,285,2400,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1546,285,2409,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1547,285,2418,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1548,285,2427,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1549,285,2436,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1550,285,2445,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1551,285,2454,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1552,285,2463,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1553,285,2472,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1554,285,2481,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1555,285,2490,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1556,286,2504,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1557,287,2505,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1558,287,2506,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1559,288,2508,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1560,288,2510,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1561,289,2512,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1562,289,2514,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1563,289,2516,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1564,289,2518,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1565,290,2522,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1566,290,2526,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1567,290,2530,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1568,290,2534,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1569,290,2538,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1570,290,2542,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1571,291,2548,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1572,291,2554,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1573,291,2560,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1574,291,2566,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1575,291,2572,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1576,291,2578,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1577,291,2584,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1578,291,2590,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1579,291,2596,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1580,292,2605,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1581,292,2614,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1582,292,2623,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1583,292,2632,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1584,292,2641,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1585,292,2650,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1586,292,2659,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1587,292,2668,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1588,292,2677,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1589,292,2686,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1590,292,2695,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1591,292,2704,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1592,292,2713,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1593,292,2722,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1594,293,2736,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1595,294,2737,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1596,294,2738,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1597,295,2740,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1598,295,2742,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1599,296,2744,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1600,296,2746,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1601,296,2748,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1602,296,2750,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1603,297,2754,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1604,297,2758,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1605,297,2762,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1606,297,2766,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1607,297,2770,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1608,297,2774,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1609,298,2780,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1610,298,2786,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1611,298,2792,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1612,298,2798,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1613,298,2804,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1614,298,2810,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1615,298,2816,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1616,298,2822,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1617,298,2828,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1618,299,2837,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1619,299,2846,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1620,299,2855,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1621,299,2864,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1622,299,2873,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1623,299,2882,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1624,299,2891,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1625,299,2900,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1626,299,2909,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1627,299,2918,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1628,299,2927,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1629,299,2936,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1630,299,2945,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1631,299,2954,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1632,300,2968,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1633,301,2969,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1634,301,2970,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1635,302,2972,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1636,302,2974,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1637,303,2976,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1638,303,2978,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1639,303,2980,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1640,303,2982,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1641,304,2986,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1642,304,2990,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1643,304,2994,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1644,304,2998,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1645,304,3002,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1646,304,3006,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1647,305,3012,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1648,305,3018,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1649,305,3024,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1650,305,3030,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1651,305,3036,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1652,305,3042,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1653,305,3048,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1654,305,3054,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1655,305,3060,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1656,306,3069,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1657,306,3078,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1658,306,3087,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1659,306,3096,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1660,306,3105,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1661,306,3114,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1662,306,3123,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1663,306,3132,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1664,306,3141,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1665,306,3150,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1666,306,3159,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1667,306,3168,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1668,306,3177,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1669,306,3186,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1670,307,3200,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1671,308,3201,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1672,308,3202,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1673,309,3204,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1674,309,3206,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1675,310,3208,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1676,310,3210,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1677,310,3212,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1678,310,3214,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1679,311,3218,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1680,311,3222,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1681,311,3226,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1682,311,3230,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1683,311,3234,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1684,311,3238,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1685,312,3244,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1686,312,3250,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1687,312,3256,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1688,312,3262,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1689,312,3268,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1690,312,3274,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1691,312,3280,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1692,312,3286,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1693,312,3292,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1694,313,3301,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1695,313,3310,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1696,313,3319,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1697,313,3328,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1698,313,3337,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1699,313,3346,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1700,313,3355,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1701,313,3364,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(1702,313,3373,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1703,313,3382,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1704,313,3391,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1705,313,3400,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1706,313,3409,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1707,313,3418,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1708,314,3432,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1709,315,3433,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1710,315,3434,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1711,316,3436,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1712,316,3438,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1713,317,3440,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1714,317,3442,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1715,317,3444,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1716,317,3446,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1717,318,3450,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1718,318,3454,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1719,318,3458,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1720,318,3462,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1721,318,3466,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1722,318,3470,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1723,319,3476,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1724,319,3482,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1725,319,3488,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1726,319,3494,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1727,319,3500,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1728,319,3,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1729,319,9,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1730,319,15,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1731,319,21,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1732,320,30,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1733,320,39,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1734,320,48,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1735,320,57,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1736,320,66,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1737,320,75,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1738,320,84,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1739,320,93,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1740,320,102,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1741,320,111,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1742,320,120,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1743,320,129,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1744,320,138,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1745,320,147,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1746,321,161,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1747,322,162,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1748,322,163,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1749,323,165,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1750,323,167,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1751,324,169,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1752,324,171,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1753,324,173,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1754,324,175,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1755,325,179,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1756,325,183,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1757,325,187,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1758,325,191,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1759,325,195,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1760,325,199,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1761,326,205,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1762,326,211,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1763,326,217,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1764,326,223,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1765,326,229,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1766,326,235,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1767,326,241,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1768,326,247,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1769,326,253,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1770,327,262,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1771,327,271,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1772,327,280,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1773,327,289,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1774,327,298,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1775,327,307,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1776,327,316,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1777,327,325,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1778,327,334,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1779,327,343,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1780,327,352,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1781,327,361,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1782,327,370,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1783,327,379,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1784,328,393,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1785,329,394,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1786,329,395,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1787,330,397,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1788,330,399,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1789,331,401,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1790,331,403,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1791,331,405,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1792,331,407,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1793,332,411,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1794,332,415,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1795,332,419,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1796,332,423,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1797,332,427,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1798,332,431,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1799,333,437,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1800,333,443,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1801,333,449,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1802,333,455,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1803,333,461,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1804,333,467,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1805,333,473,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1806,333,479,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1807,333,485,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1808,334,494,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1809,334,503,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1810,334,512,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1811,334,521,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1812,334,530,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1813,334,539,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1814,334,548,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1815,334,557,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1816,334,566,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1817,334,575,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1818,334,584,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1819,334,593,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1820,334,602,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1821,334,611,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1822,335,625,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1823,336,626,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1824,336,627,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1825,337,629,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1826,337,631,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1827,338,633,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1828,338,635,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1829,338,637,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1830,338,639,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1831,339,643,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1832,339,647,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1833,339,651,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1834,339,655,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1835,339,659,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1836,339,663,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1837,340,669,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1838,340,675,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1839,340,681,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1840,340,687,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1841,340,693,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1842,340,699,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1843,340,705,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1844,340,711,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1845,340,717,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1846,341,726,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1847,341,735,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1848,341,744,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1849,341,753,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1850,341,762,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1851,341,771,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1852,341,780,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1853,341,789,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1854,341,798,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1855,341,807,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1856,341,816,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1857,341,825,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1858,341,834,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1859,341,843,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1860,342,857,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1861,343,858,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1862,343,859,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1863,344,861,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1864,344,863,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1865,345,865,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1866,345,867,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1867,345,869,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1868,345,871,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1869,346,875,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1870,346,879,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1871,346,883,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1872,346,887,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1873,346,891,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1874,346,895,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1875,347,901,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1876,347,907,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1877,347,913,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1878,347,919,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1879,347,925,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1880,347,931,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1881,347,937,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1882,347,943,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1883,347,949,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1884,348,958,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1885,348,967,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1886,348,976,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1887,348,985,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1888,348,994,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1889,348,1003,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1890,348,1012,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1891,348,1021,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1892,348,1030,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1893,348,1039,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1894,348,1048,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1895,348,1057,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1896,348,1066,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1897,348,1075,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1898,349,1089,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1899,350,1090,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1900,350,1091,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1901,351,1093,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1902,351,1095,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1903,352,1097,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1904,352,1099,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1905,352,1101,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1906,352,1103,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1907,353,1107,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1908,353,1111,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1909,353,1115,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1910,353,1119,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1911,353,1123,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1912,353,1127,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1913,354,1133,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1914,354,1139,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1915,354,1145,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1916,354,1151,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1917,354,1157,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1918,354,1163,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1919,354,1169,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1920,354,1175,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1921,354,1181,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1922,355,1190,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1923,355,1199,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1924,355,1208,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1925,355,1217,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1926,355,1226,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1927,355,1235,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1928,355,1244,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1929,355,1253,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1930,355,1262,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1931,355,1271,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1932,355,1280,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1933,355,1289,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1934,355,1298,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1935,355,1307,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1936,356,1321,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1937,357,1322,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1938,357,1323,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1939,358,1325,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1940,358,1327,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1941,359,1329,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1942,359,1331,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1943,359,1333,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1944,359,1335,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1945,360,1339,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1946,360,1343,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1947,360,1347,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1948,360,1351,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1949,360,1355,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1950,360,1359,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1951,361,1365,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1952,361,1371,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1953,361,1377,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1954,361,1383,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1955,361,1389,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1956,361,1395,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1957,361,1401,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1958,361,1407,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1959,361,1413,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1960,362,1422,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1961,362,1431,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1962,362,1440,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1963,362,1449,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1964,362,1458,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1965,362,1467,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1966,362,1476,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1967,362,1485,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1968,362,1494,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1969,362,1503,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1970,362,1512,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1971,362,1521,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1972,362,1530,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1973,362,1539,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1974,363,1553,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1975,364,1554,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1976,364,1555,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1977,365,1557,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1978,365,1559,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1979,366,1561,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1980,366,1563,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1981,366,1565,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1982,366,1567,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1983,367,1571,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1984,367,1575,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1985,367,1579,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1986,367,1583,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1987,367,1587,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1988,367,1591,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1989,368,1597,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1990,368,1603,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1991,368,1609,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1992,368,1615,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1993,368,1621,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1994,368,1627,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1995,368,1633,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1996,368,1639,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1997,368,1645,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1998,369,1654,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(1999,369,1663,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2000,369,1672,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2001,369,1681,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2002,369,1690,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2003,369,1699,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2004,369,1708,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2005,369,1717,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2006,369,1726,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2007,369,1735,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2008,369,1744,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2009,369,1753,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2010,369,1762,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2011,369,1771,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2012,370,1785,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2013,371,1786,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2014,371,1787,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2015,372,1789,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2016,372,1791,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2017,373,1793,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2018,373,1795,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2019,373,1797,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2020,373,1799,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2021,374,1803,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2022,374,1807,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2023,374,1811,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2024,374,1815,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2025,374,1819,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2026,374,1823,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2027,375,1829,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2028,375,1835,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2029,375,1841,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2030,375,1847,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2031,375,1853,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2032,375,1859,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2033,375,1865,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2034,375,1871,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2035,375,1877,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2036,376,1886,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2037,376,1895,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2038,376,1904,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2039,376,1913,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2040,376,1922,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2041,376,1931,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2042,376,1940,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2043,376,1949,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2044,376,1958,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2045,376,1967,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2046,376,1976,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2047,376,1985,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2048,376,1994,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2049,376,2003,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2050,377,2017,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2051,378,2018,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2052,378,2019,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2053,379,2021,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2054,379,2023,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2055,380,2025,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2056,380,2027,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2057,380,2029,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2058,380,2031,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2059,381,2035,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2060,381,2039,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2061,381,2043,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2062,381,2047,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2063,381,2051,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2064,381,2055,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2065,382,2061,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2066,382,2067,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2067,382,2073,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2068,382,2079,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2069,382,2085,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2070,382,2091,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2071,382,2097,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2072,382,2103,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2073,382,2109,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2074,383,2118,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2075,383,2127,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2076,383,2136,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2077,383,2145,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2078,383,2154,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2079,383,2163,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2080,383,2172,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2081,383,2181,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2082,383,2190,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2083,383,2199,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2084,383,2208,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2085,383,2217,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2086,383,2226,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2087,383,2235,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2088,384,2249,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2089,385,2250,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2090,385,2251,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2091,386,2253,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2092,386,2255,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2093,387,2257,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2094,387,2259,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2095,387,2261,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2096,387,2263,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2097,388,2267,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2098,388,2271,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2099,388,2275,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2100,388,2279,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2101,388,2283,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2102,388,2287,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2103,389,2293,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2104,389,2299,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2105,389,2305,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2106,389,2311,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2107,389,2317,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2108,389,2323,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2109,389,2329,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2110,389,2335,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2111,389,2341,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2112,390,2350,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2113,390,2359,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2114,390,2368,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2115,390,2377,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2116,390,2386,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2117,390,2395,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2118,390,2404,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2119,390,2413,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2120,390,2422,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2121,390,2431,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2122,390,2440,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2123,390,2449,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2124,390,2458,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2125,390,2467,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2126,391,2481,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2127,392,2482,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2128,392,2483,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2129,393,2485,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2130,393,2487,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2131,394,2489,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2132,394,2491,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2133,394,2493,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2134,394,2495,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2135,395,2499,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2136,395,2503,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2137,395,2507,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2138,395,2511,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2139,395,2515,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2140,395,2519,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2141,396,2525,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2142,396,2531,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2143,396,2537,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2144,396,2543,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2145,396,2549,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2146,396,2555,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2147,396,2561,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2148,396,2567,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2149,396,2573,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2150,397,2582,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2151,397,2591,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2152,397,2600,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2153,397,2609,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2154,397,2618,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2155,397,2627,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2156,397,2636,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2157,397,2645,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2158,397,2654,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2159,397,2663,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2160,397,2672,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2161,397,2681,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2162,397,2690,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2163,397,2699,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2164,398,2713,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2165,399,2714,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2166,399,2715,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2167,400,2717,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2168,400,2719,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2169,401,2721,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2170,401,2723,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2171,401,2725,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2172,401,2727,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2173,402,2731,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2174,402,2735,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2175,402,2739,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2176,402,2743,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2177,402,2747,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2178,402,2751,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2179,403,2757,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2180,403,2763,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2181,403,2769,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2182,403,2775,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2183,403,2781,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2184,403,2787,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2185,403,2793,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2186,403,2799,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2187,403,2805,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2188,404,2814,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2189,404,2823,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2190,404,2832,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2191,404,2841,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2192,404,2850,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2193,404,2859,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2194,404,2868,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2195,404,2877,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2196,404,2886,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2197,404,2895,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2198,404,2904,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2199,404,2913,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2200,404,2922,1.9899999999999999911,1);
INSERT INTO invoice_items VALUES(2201,404,2931,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2202,405,2945,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2203,406,2946,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2204,406,2947,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2205,407,2949,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2206,407,2951,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2207,408,2953,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2208,408,2955,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2209,408,2957,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2210,408,2959,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2211,409,2963,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2212,409,2967,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2213,409,2971,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2214,409,2975,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2215,409,2979,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2216,409,2983,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2217,410,2989,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2218,410,2995,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2219,410,3001,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2220,410,3007,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2221,410,3013,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2222,410,3019,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2223,410,3025,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2224,410,3031,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2225,410,3037,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2226,411,3046,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2227,411,3055,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2228,411,3064,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2229,411,3073,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2230,411,3082,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2231,411,3091,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2232,411,3100,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2233,411,3109,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2234,411,3118,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2235,411,3127,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2236,411,3136,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2237,411,3145,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2238,411,3154,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2239,411,3163,0.98999999999999999111,1);
INSERT INTO invoice_items VALUES(2240,412,3177,1.9899999999999999911,1);
CREATE TABLE IF NOT EXISTS "media_types"
(
    [MediaTypeId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [Name] NVARCHAR(120)
);
INSERT INTO media_types VALUES(1,'MPEG audio file');
INSERT INTO media_types VALUES(2,'Protected AAC audio file');
INSERT INTO media_types VALUES(3,'Protected MPEG-4 video file');
INSERT INTO media_types VALUES(4,'Purchased AAC audio file');
INSERT INTO media_types VALUES(5,'AAC audio file');
CREATE TABLE IF NOT EXISTS "playlists"
(
    [PlaylistId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [Name] NVARCHAR(120)
);
INSERT INTO playlists VALUES(1,'Music');
INSERT INTO playlists VALUES(2,'Movies');
INSERT INTO playlists VALUES(3,'TV Shows');
INSERT INTO playlists VALUES(4,'Audiobooks');
INSERT INTO playlists VALUES(5,'90’s Music');
INSERT INTO playlists VALUES(6,'Audiobooks');
INSERT INTO playlists VALUES(7,'Movies');
INSERT INTO playlists VALUES(8,'Music');
INSERT INTO playlists VALUES(9,'Music Videos');
INSERT INTO playlists VALUES(10,'TV Shows');
INSERT INTO playlists VALUES(11,'Brazilian Music');
INSERT INTO playlists VALUES(12,'Classical');
INSERT INTO playlists VALUES(13,'Classical 101 - Deep Cuts');
INSERT INTO playlists VALUES(14,'Classical 101 - Next Steps');
INSERT INTO playlists VALUES(15,'Classical 101 - The Basics');
INSERT INTO playlists VALUES(16,'Grunge');
INSERT INTO playlists VALUES(17,'Heavy Metal Classic');
INSERT INTO playlists VALUES(18,'On-The-Go 1');
CREATE TABLE IF NOT EXISTS "playlist_track"
(
    [PlaylistId] INTEGER  NOT NULL,
    [TrackId] INTEGER  NOT NULL,
    CONSTRAINT [PK_PlaylistTrack] PRIMARY KEY  ([PlaylistId], [TrackId]),
    FOREIGN KEY ([PlaylistId]) REFERENCES "playlists" ([PlaylistId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION,
    FOREIGN KEY ([TrackId]) REFERENCES "tracks" ([TrackId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION
);
INSERT INTO playlist_track VALUES(1,3402);
INSERT INTO playlist_track VALUES(1,3389);
INSERT INTO playlist_track VALUES(1,3390);
INSERT INTO playlist_track VALUES(1,3391);
INSERT INTO playlist_track VALUES(1,3392);
INSERT INTO playlist_track VALUES(1,3393);
INSERT INTO playlist_track VALUES(1,3394);
INSERT INTO playlist_track VALUES(1,3395);
INSERT INTO playlist_track VALUES(1,3396);
INSERT INTO playlist_track VALUES(1,3397);
INSERT INTO playlist_track VALUES(1,3398);
INSERT INTO playlist_track VALUES(1,3399);
INSERT INTO playlist_track VALUES(1,3400);
INSERT INTO playlist_track VALUES(1,3401);
INSERT INTO playlist_track VALUES(1,3336);
INSERT INTO playlist_track VALUES(1,3478);
INSERT INTO playlist_track VALUES(1,3375);
INSERT INTO playlist_track VALUES(1,3376);
INSERT INTO playlist_track VALUES(1,3377);
INSERT INTO playlist_track VALUES(1,3378);
INSERT INTO playlist_track VALUES(1,3379);
INSERT INTO playlist_track VALUES(1,3380);
INSERT INTO playlist_track VALUES(1,3381);
INSERT INTO playlist_track VALUES(1,3382);
INSERT INTO playlist_track VALUES(1,3383);
INSERT INTO playlist_track VALUES(1,3384);
INSERT INTO playlist_track VALUES(1,3385);
INSERT INTO playlist_track VALUES(1,3386);
INSERT INTO playlist_track VALUES(1,3387);
INSERT INTO playlist_track VALUES(1,3388);
INSERT INTO playlist_track VALUES(1,3365);
INSERT INTO playlist_track VALUES(1,3366);
INSERT INTO playlist_track VALUES(1,3367);
INSERT INTO playlist_track VALUES(1,3368);
INSERT INTO playlist_track VALUES(1,3369);
INSERT INTO playlist_track VALUES(1,3370);
INSERT INTO playlist_track VALUES(1,3371);
INSERT INTO playlist_track VALUES(1,3372);
INSERT INTO playlist_track VALUES(1,3373);
INSERT INTO playlist_track VALUES(1,3374);
INSERT INTO playlist_track VALUES(1,99);
INSERT INTO playlist_track VALUES(1,100);
INSERT INTO playlist_track VALUES(1,101);
INSERT INTO playlist_track VALUES(1,102);
INSERT INTO playlist_track VALUES(1,103);
INSERT INTO playlist_track VALUES(1,104);
INSERT INTO playlist_track VALUES(1,105);
INSERT INTO playlist_track VALUES(1,106);
INSERT INTO playlist_track VALUES(1,107);
INSERT INTO playlist_track VALUES(1,108);
INSERT INTO playlist_track VALUES(1,109);
INSERT INTO playlist_track VALUES(1,110);
INSERT INTO playlist_track VALUES(1,166);
INSERT INTO playlist_track VALUES(1,167);
INSERT INTO playlist_track VALUES(1,168);
INSERT INTO playlist_track VALUES(1,169);
INSERT INTO playlist_track VALUES(1,170);
INSERT INTO playlist_track VALUES(1,171);
INSERT INTO playlist_track VALUES(1,172);
INSERT INTO playlist_track VALUES(1,173);
INSERT INTO playlist_track VALUES(1,174);
INSERT INTO playlist_track VALUES(1,175);
INSERT INTO playlist_track VALUES(1,176);
INSERT INTO playlist_track VALUES(1,177);
INSERT INTO playlist_track VALUES(1,178);
INSERT INTO playlist_track VALUES(1,179);
INSERT INTO playlist_track VALUES(1,180);
INSERT INTO playlist_track VALUES(1,181);
INSERT INTO playlist_track VALUES(1,182);
INSERT INTO playlist_track VALUES(1,2591);
INSERT INTO playlist_track VALUES(1,2592);
INSERT INTO playlist_track VALUES(1,2593);
INSERT INTO playlist_track VALUES(1,2594);
INSERT INTO playlist_track VALUES(1,2595);
INSERT INTO playlist_track VALUES(1,2596);
INSERT INTO playlist_track VALUES(1,2597);
INSERT INTO playlist_track VALUES(1,2598);
INSERT INTO playlist_track VALUES(1,2599);
INSERT INTO playlist_track VALUES(1,2600);
INSERT INTO playlist_track VALUES(1,2601);
INSERT INTO playlist_track VALUES(1,2602);
INSERT INTO playlist_track VALUES(1,2603);
INSERT INTO playlist_track VALUES(1,2604);
INSERT INTO playlist_track VALUES(1,2605);
INSERT INTO playlist_track VALUES(1,2606);
INSERT INTO playlist_track VALUES(1,2607);
INSERT INTO playlist_track VALUES(1,2608);
INSERT INTO playlist_track VALUES(1,923);
INSERT INTO playlist_track VALUES(1,924);
INSERT INTO playlist_track VALUES(1,925);
INSERT INTO playlist_track VALUES(1,926);
INSERT INTO playlist_track VALUES(1,927);
INSERT INTO playlist_track VALUES(1,928);
INSERT INTO playlist_track VALUES(1,929);
INSERT INTO playlist_track VALUES(1,930);
INSERT INTO playlist_track VALUES(1,931);
INSERT INTO playlist_track VALUES(1,932);
INSERT INTO playlist_track VALUES(1,933);
INSERT INTO playlist_track VALUES(1,934);
INSERT INTO playlist_track VALUES(1,935);
INSERT INTO playlist_track VALUES(1,936);
INSERT INTO playlist_track VALUES(1,937);
INSERT INTO playlist_track VALUES(1,938);
INSERT INTO playlist_track VALUES(1,939);
INSERT INTO playlist_track VALUES(1,940);
INSERT INTO playlist_track VALUES(1,941);
INSERT INTO playlist_track VALUES(1,942);
INSERT INTO playlist_track VALUES(1,943);
INSERT INTO playlist_track VALUES(1,944);
INSERT INTO playlist_track VALUES(1,945);
INSERT INTO playlist_track VALUES(1,946);
INSERT INTO playlist_track VALUES(1,947);
INSERT INTO playlist_track VALUES(1,948);
INSERT INTO playlist_track VALUES(1,964);
INSERT INTO playlist_track VALUES(1,965);
INSERT INTO playlist_track VALUES(1,966);
INSERT INTO playlist_track VALUES(1,967);
INSERT INTO playlist_track VALUES(1,968);
INSERT INTO playlist_track VALUES(1,969);
INSERT INTO playlist_track VALUES(1,970);
INSERT INTO playlist_track VALUES(1,971);
INSERT INTO playlist_track VALUES(1,972);
INSERT INTO playlist_track VALUES(1,973);
INSERT INTO playlist_track VALUES(1,974);
INSERT INTO playlist_track VALUES(1,1009);
INSERT INTO playlist_track VALUES(1,1010);
INSERT INTO playlist_track VALUES(1,1011);
INSERT INTO playlist_track VALUES(1,1012);
INSERT INTO playlist_track VALUES(1,1013);
INSERT INTO playlist_track VALUES(1,1014);
INSERT INTO playlist_track VALUES(1,1015);
INSERT INTO playlist_track VALUES(1,1016);
INSERT INTO playlist_track VALUES(1,1017);
INSERT INTO playlist_track VALUES(1,1018);
INSERT INTO playlist_track VALUES(1,1019);
INSERT INTO playlist_track VALUES(1,1133);
INSERT INTO playlist_track VALUES(1,1134);
INSERT INTO playlist_track VALUES(1,1135);
INSERT INTO playlist_track VALUES(1,1136);
INSERT INTO playlist_track VALUES(1,1137);
INSERT INTO playlist_track VALUES(1,1138);
INSERT INTO playlist_track VALUES(1,1139);
INSERT INTO playlist_track VALUES(1,1140);
INSERT INTO playlist_track VALUES(1,1141);
INSERT INTO playlist_track VALUES(1,1142);
INSERT INTO playlist_track VALUES(1,1143);
INSERT INTO playlist_track VALUES(1,1144);
INSERT INTO playlist_track VALUES(1,1145);
INSERT INTO playlist_track VALUES(1,468);
INSERT INTO playlist_track VALUES(1,469);
INSERT INTO playlist_track VALUES(1,470);
INSERT INTO playlist_track VALUES(1,471);
INSERT INTO playlist_track VALUES(1,472);
INSERT INTO playlist_track VALUES(1,473);
INSERT INTO playlist_track VALUES(1,474);
INSERT INTO playlist_track VALUES(1,475);
INSERT INTO playlist_track VALUES(1,476);
INSERT INTO playlist_track VALUES(1,477);
INSERT INTO playlist_track VALUES(1,478);
INSERT INTO playlist_track VALUES(1,479);
INSERT INTO playlist_track VALUES(1,480);
INSERT INTO playlist_track VALUES(1,481);
INSERT INTO playlist_track VALUES(1,482);
INSERT INTO playlist_track VALUES(1,483);
INSERT INTO playlist_track VALUES(1,484);
INSERT INTO playlist_track VALUES(1,485);
INSERT INTO playlist_track VALUES(1,486);
INSERT INTO playlist_track VALUES(1,487);
INSERT INTO playlist_track VALUES(1,488);
INSERT INTO playlist_track VALUES(1,1466);
INSERT INTO playlist_track VALUES(1,1467);
INSERT INTO playlist_track VALUES(1,1468);
INSERT INTO playlist_track VALUES(1,1469);
INSERT INTO playlist_track VALUES(1,1470);
INSERT INTO playlist_track VALUES(1,1471);
INSERT INTO playlist_track VALUES(1,1472);
INSERT INTO playlist_track VALUES(1,1473);
INSERT INTO playlist_track VALUES(1,1474);
INSERT INTO playlist_track VALUES(1,1475);
INSERT INTO playlist_track VALUES(1,1476);
INSERT INTO playlist_track VALUES(1,1477);
INSERT INTO playlist_track VALUES(1,1478);
INSERT INTO playlist_track VALUES(1,529);
INSERT INTO playlist_track VALUES(1,530);
INSERT INTO playlist_track VALUES(1,531);
INSERT INTO playlist_track VALUES(1,532);
INSERT INTO playlist_track VALUES(1,533);
INSERT INTO playlist_track VALUES(1,534);
INSERT INTO playlist_track VALUES(1,535);
INSERT INTO playlist_track VALUES(1,536);
INSERT INTO playlist_track VALUES(1,537);
INSERT INTO playlist_track VALUES(1,538);
INSERT INTO playlist_track VALUES(1,539);
INSERT INTO playlist_track VALUES(1,540);
INSERT INTO playlist_track VALUES(1,541);
INSERT INTO playlist_track VALUES(1,542);
INSERT INTO playlist_track VALUES(1,2165);
INSERT INTO playlist_track VALUES(1,2166);
INSERT INTO playlist_track VALUES(1,2167);
INSERT INTO playlist_track VALUES(1,2168);
INSERT INTO playlist_track VALUES(1,2169);
INSERT INTO playlist_track VALUES(1,2170);
INSERT INTO playlist_track VALUES(1,2171);
INSERT INTO playlist_track VALUES(1,2172);
INSERT INTO playlist_track VALUES(1,2173);
INSERT INTO playlist_track VALUES(1,2174);
INSERT INTO playlist_track VALUES(1,2175);
INSERT INTO playlist_track VALUES(1,2176);
INSERT INTO playlist_track VALUES(1,2177);
INSERT INTO playlist_track VALUES(1,2318);
INSERT INTO playlist_track VALUES(1,2319);
INSERT INTO playlist_track VALUES(1,2320);
INSERT INTO playlist_track VALUES(1,2321);
INSERT INTO playlist_track VALUES(1,2322);
INSERT INTO playlist_track VALUES(1,2323);
INSERT INTO playlist_track VALUES(1,2324);
INSERT INTO playlist_track VALUES(1,2325);
INSERT INTO playlist_track VALUES(1,2326);
INSERT INTO playlist_track VALUES(1,2327);
INSERT INTO playlist_track VALUES(1,2328);
INSERT INTO playlist_track VALUES(1,2329);
INSERT INTO playlist_track VALUES(1,2330);
INSERT INTO playlist_track VALUES(1,2331);
INSERT INTO playlist_track VALUES(1,2332);
INSERT INTO playlist_track VALUES(1,2333);
INSERT INTO playlist_track VALUES(1,2285);
INSERT INTO playlist_track VALUES(1,2286);
INSERT INTO playlist_track VALUES(1,2287);
INSERT INTO playlist_track VALUES(1,2288);
INSERT INTO playlist_track VALUES(1,2289);
INSERT INTO playlist_track VALUES(1,2290);
INSERT INTO playlist_track VALUES(1,2291);
INSERT INTO playlist_track VALUES(1,2292);
INSERT INTO playlist_track VALUES(1,2293);
INSERT INTO playlist_track VALUES(1,2294);
INSERT INTO playlist_track VALUES(1,2295);
INSERT INTO playlist_track VALUES(1,2310);
INSERT INTO playlist_track VALUES(1,2311);
INSERT INTO playlist_track VALUES(1,2312);
INSERT INTO playlist_track VALUES(1,2313);
INSERT INTO playlist_track VALUES(1,2314);
INSERT INTO playlist_track VALUES(1,2315);
INSERT INTO playlist_track VALUES(1,2316);
INSERT INTO playlist_track VALUES(1,2317);
INSERT INTO playlist_track VALUES(1,2282);
INSERT INTO playlist_track VALUES(1,2283);
INSERT INTO playlist_track VALUES(1,2284);
INSERT INTO playlist_track VALUES(1,2334);
INSERT INTO playlist_track VALUES(1,2335);
INSERT INTO playlist_track VALUES(1,2336);
INSERT INTO playlist_track VALUES(1,2337);
INSERT INTO playlist_track VALUES(1,2338);
INSERT INTO playlist_track VALUES(1,2339);
INSERT INTO playlist_track VALUES(1,2340);
INSERT INTO playlist_track VALUES(1,2341);
INSERT INTO playlist_track VALUES(1,2342);
INSERT INTO playlist_track VALUES(1,2343);
INSERT INTO playlist_track VALUES(1,2358);
INSERT INTO playlist_track VALUES(1,2359);
INSERT INTO playlist_track VALUES(1,2360);
INSERT INTO playlist_track VALUES(1,2361);
INSERT INTO playlist_track VALUES(1,2362);
INSERT INTO playlist_track VALUES(1,2363);
INSERT INTO playlist_track VALUES(1,2364);
INSERT INTO playlist_track VALUES(1,2365);
INSERT INTO playlist_track VALUES(1,2366);
INSERT INTO playlist_track VALUES(1,2367);
INSERT INTO playlist_track VALUES(1,2368);
INSERT INTO playlist_track VALUES(1,2369);
INSERT INTO playlist_track VALUES(1,2370);
INSERT INTO playlist_track VALUES(1,2371);
INSERT INTO playlist_track VALUES(1,2372);
INSERT INTO playlist_track VALUES(1,2373);
INSERT INTO playlist_track VALUES(1,2374);
INSERT INTO playlist_track VALUES(1,2472);
INSERT INTO playlist_track VALUES(1,2473);
INSERT INTO playlist_track VALUES(1,2474);
INSERT INTO playlist_track VALUES(1,2475);
INSERT INTO playlist_track VALUES(1,2476);
INSERT INTO playlist_track VALUES(1,2477);
INSERT INTO playlist_track VALUES(1,2478);
INSERT INTO playlist_track VALUES(1,2479);
INSERT INTO playlist_track VALUES(1,2480);
INSERT INTO playlist_track VALUES(1,2481);
INSERT INTO playlist_track VALUES(1,2482);
INSERT INTO playlist_track VALUES(1,2483);
INSERT INTO playlist_track VALUES(1,2484);
INSERT INTO playlist_track VALUES(1,2485);
INSERT INTO playlist_track VALUES(1,2486);
INSERT INTO playlist_track VALUES(1,2487);
INSERT INTO playlist_track VALUES(1,2488);
INSERT INTO playlist_track VALUES(1,2489);
INSERT INTO playlist_track VALUES(1,2490);
INSERT INTO playlist_track VALUES(1,2491);
INSERT INTO playlist_track VALUES(1,2492);
INSERT INTO playlist_track VALUES(1,2493);
INSERT INTO playlist_track VALUES(1,2494);
INSERT INTO playlist_track VALUES(1,2495);
INSERT INTO playlist_track VALUES(1,2496);
INSERT INTO playlist_track VALUES(1,2497);
INSERT INTO playlist_track VALUES(1,2498);
INSERT INTO playlist_track VALUES(1,2499);
INSERT INTO playlist_track VALUES(1,2500);
INSERT INTO playlist_track VALUES(1,2501);
INSERT INTO playlist_track VALUES(1,2502);
INSERT INTO playlist_track VALUES(1,2503);
INSERT INTO playlist_track VALUES(1,2504);
INSERT INTO playlist_track VALUES(1,2505);
INSERT INTO playlist_track VALUES(1,2705);
INSERT INTO playlist_track VALUES(1,2706);
INSERT INTO playlist_track VALUES(1,2707);
INSERT INTO playlist_track VALUES(1,2708);
INSERT INTO playlist_track VALUES(1,2709);
INSERT INTO playlist_track VALUES(1,2710);
INSERT INTO playlist_track VALUES(1,2711);
INSERT INTO playlist_track VALUES(1,2712);
INSERT INTO playlist_track VALUES(1,2713);
INSERT INTO playlist_track VALUES(1,2714);
INSERT INTO playlist_track VALUES(1,2715);
INSERT INTO playlist_track VALUES(1,2716);
INSERT INTO playlist_track VALUES(1,2717);
INSERT INTO playlist_track VALUES(1,2718);
INSERT INTO playlist_track VALUES(1,2719);
INSERT INTO playlist_track VALUES(1,2720);
INSERT INTO playlist_track VALUES(1,2721);
INSERT INTO playlist_track VALUES(1,2722);
INSERT INTO playlist_track VALUES(1,2723);
INSERT INTO playlist_track VALUES(1,2724);
INSERT INTO playlist_track VALUES(1,2725);
INSERT INTO playlist_track VALUES(1,2726);
INSERT INTO playlist_track VALUES(1,2727);
INSERT INTO playlist_track VALUES(1,2728);
INSERT INTO playlist_track VALUES(1,2729);
INSERT INTO playlist_track VALUES(1,2730);
INSERT INTO playlist_track VALUES(1,2781);
INSERT INTO playlist_track VALUES(1,2782);
INSERT INTO playlist_track VALUES(1,2783);
INSERT INTO playlist_track VALUES(1,2784);
INSERT INTO playlist_track VALUES(1,2785);
INSERT INTO playlist_track VALUES(1,2786);
INSERT INTO playlist_track VALUES(1,2787);
INSERT INTO playlist_track VALUES(1,2788);
INSERT INTO playlist_track VALUES(1,2789);
INSERT INTO playlist_track VALUES(1,2790);
INSERT INTO playlist_track VALUES(1,2791);
INSERT INTO playlist_track VALUES(1,2792);
INSERT INTO playlist_track VALUES(1,2793);
INSERT INTO playlist_track VALUES(1,2794);
INSERT INTO playlist_track VALUES(1,2795);
INSERT INTO playlist_track VALUES(1,2796);
INSERT INTO playlist_track VALUES(1,2797);
INSERT INTO playlist_track VALUES(1,2798);
INSERT INTO playlist_track VALUES(1,2799);
INSERT INTO playlist_track VALUES(1,2800);
INSERT INTO playlist_track VALUES(1,2801);
INSERT INTO playlist_track VALUES(1,2802);
INSERT INTO playlist_track VALUES(1,2803);
INSERT INTO playlist_track VALUES(1,2804);
INSERT INTO playlist_track VALUES(1,2805);
INSERT INTO playlist_track VALUES(1,2806);
INSERT INTO playlist_track VALUES(1,2807);
INSERT INTO playlist_track VALUES(1,2808);
INSERT INTO playlist_track VALUES(1,2809);
INSERT INTO playlist_track VALUES(1,2810);
INSERT INTO playlist_track VALUES(1,2811);
INSERT INTO playlist_track VALUES(1,2812);
INSERT INTO playlist_track VALUES(1,2813);
INSERT INTO playlist_track VALUES(1,2814);
INSERT INTO playlist_track VALUES(1,2815);
INSERT INTO playlist_track VALUES(1,2816);
INSERT INTO playlist_track VALUES(1,2817);
INSERT INTO playlist_track VALUES(1,2818);
INSERT INTO playlist_track VALUES(1,2572);
INSERT INTO playlist_track VALUES(1,2573);
INSERT INTO playlist_track VALUES(1,2574);
INSERT INTO playlist_track VALUES(1,2575);
INSERT INTO playlist_track VALUES(1,2576);
INSERT INTO playlist_track VALUES(1,2577);
INSERT INTO playlist_track VALUES(1,2578);
INSERT INTO playlist_track VALUES(1,2579);
INSERT INTO playlist_track VALUES(1,2580);
INSERT INTO playlist_track VALUES(1,2581);
INSERT INTO playlist_track VALUES(1,2582);
INSERT INTO playlist_track VALUES(1,2583);
INSERT INTO playlist_track VALUES(1,2584);
INSERT INTO playlist_track VALUES(1,2585);
INSERT INTO playlist_track VALUES(1,2586);
INSERT INTO playlist_track VALUES(1,2587);
INSERT INTO playlist_track VALUES(1,2588);
INSERT INTO playlist_track VALUES(1,2589);
INSERT INTO playlist_track VALUES(1,2590);
INSERT INTO playlist_track VALUES(1,194);
INSERT INTO playlist_track VALUES(1,195);
INSERT INTO playlist_track VALUES(1,196);
INSERT INTO playlist_track VALUES(1,197);
INSERT INTO playlist_track VALUES(1,198);
INSERT INTO playlist_track VALUES(1,199);
INSERT INTO playlist_track VALUES(1,200);
INSERT INTO playlist_track VALUES(1,201);
INSERT INTO playlist_track VALUES(1,202);
INSERT INTO playlist_track VALUES(1,203);
INSERT INTO playlist_track VALUES(1,204);
INSERT INTO playlist_track VALUES(1,891);
INSERT INTO playlist_track VALUES(1,892);
INSERT INTO playlist_track VALUES(1,893);
INSERT INTO playlist_track VALUES(1,894);
INSERT INTO playlist_track VALUES(1,895);
INSERT INTO playlist_track VALUES(1,896);
INSERT INTO playlist_track VALUES(1,897);
INSERT INTO playlist_track VALUES(1,898);
INSERT INTO playlist_track VALUES(1,899);
INSERT INTO playlist_track VALUES(1,900);
INSERT INTO playlist_track VALUES(1,901);
INSERT INTO playlist_track VALUES(1,902);
INSERT INTO playlist_track VALUES(1,903);
INSERT INTO playlist_track VALUES(1,904);
INSERT INTO playlist_track VALUES(1,905);
INSERT INTO playlist_track VALUES(1,906);
INSERT INTO playlist_track VALUES(1,907);
INSERT INTO playlist_track VALUES(1,908);
INSERT INTO playlist_track VALUES(1,909);
INSERT INTO playlist_track VALUES(1,910);
INSERT INTO playlist_track VALUES(1,911);
INSERT INTO playlist_track VALUES(1,912);
INSERT INTO playlist_track VALUES(1,913);
INSERT INTO playlist_track VALUES(1,914);
INSERT INTO playlist_track VALUES(1,915);
INSERT INTO playlist_track VALUES(1,916);
INSERT INTO playlist_track VALUES(1,917);
INSERT INTO playlist_track VALUES(1,918);
INSERT INTO playlist_track VALUES(1,919);
INSERT INTO playlist_track VALUES(1,920);
INSERT INTO playlist_track VALUES(1,921);
INSERT INTO playlist_track VALUES(1,922);
INSERT INTO playlist_track VALUES(1,1268);
INSERT INTO playlist_track VALUES(1,1269);
INSERT INTO playlist_track VALUES(1,1270);
INSERT INTO playlist_track VALUES(1,1271);
INSERT INTO playlist_track VALUES(1,1272);
INSERT INTO playlist_track VALUES(1,1273);
INSERT INTO playlist_track VALUES(1,1274);
INSERT INTO playlist_track VALUES(1,1275);
INSERT INTO playlist_track VALUES(1,1276);
INSERT INTO playlist_track VALUES(1,2532);
INSERT INTO playlist_track VALUES(1,2533);
INSERT INTO playlist_track VALUES(1,2534);
INSERT INTO playlist_track VALUES(1,2535);
INSERT INTO playlist_track VALUES(1,2536);
INSERT INTO playlist_track VALUES(1,2537);
INSERT INTO playlist_track VALUES(1,2538);
INSERT INTO playlist_track VALUES(1,2539);
INSERT INTO playlist_track VALUES(1,2540);
INSERT INTO playlist_track VALUES(1,2541);
INSERT INTO playlist_track VALUES(1,646);
INSERT INTO playlist_track VALUES(1,647);
INSERT INTO playlist_track VALUES(1,648);
INSERT INTO playlist_track VALUES(1,649);
INSERT INTO playlist_track VALUES(1,651);
INSERT INTO playlist_track VALUES(1,653);
INSERT INTO playlist_track VALUES(1,655);
INSERT INTO playlist_track VALUES(1,658);
INSERT INTO playlist_track VALUES(1,652);
INSERT INTO playlist_track VALUES(1,656);
INSERT INTO playlist_track VALUES(1,657);
INSERT INTO playlist_track VALUES(1,650);
INSERT INTO playlist_track VALUES(1,659);
INSERT INTO playlist_track VALUES(1,654);
INSERT INTO playlist_track VALUES(1,660);
INSERT INTO playlist_track VALUES(1,3427);
INSERT INTO playlist_track VALUES(1,3411);
INSERT INTO playlist_track VALUES(1,3412);
INSERT INTO playlist_track VALUES(1,3419);
INSERT INTO playlist_track VALUES(1,3482);
INSERT INTO playlist_track VALUES(1,3438);
INSERT INTO playlist_track VALUES(1,3485);
INSERT INTO playlist_track VALUES(1,3403);
INSERT INTO playlist_track VALUES(1,3406);
INSERT INTO playlist_track VALUES(1,3442);
INSERT INTO playlist_track VALUES(1,3421);
INSERT INTO playlist_track VALUES(1,3436);
INSERT INTO playlist_track VALUES(1,3450);
INSERT INTO playlist_track VALUES(1,3454);
INSERT INTO playlist_track VALUES(1,3491);
INSERT INTO playlist_track VALUES(1,3413);
INSERT INTO playlist_track VALUES(1,3426);
INSERT INTO playlist_track VALUES(1,3416);
INSERT INTO playlist_track VALUES(1,3501);
INSERT INTO playlist_track VALUES(1,3487);
INSERT INTO playlist_track VALUES(1,3417);
INSERT INTO playlist_track VALUES(1,3432);
INSERT INTO playlist_track VALUES(1,3443);
INSERT INTO playlist_track VALUES(1,3447);
INSERT INTO playlist_track VALUES(1,3452);
INSERT INTO playlist_track VALUES(1,3441);
INSERT INTO playlist_track VALUES(1,3434);
INSERT INTO playlist_track VALUES(1,3500);
INSERT INTO playlist_track VALUES(1,3449);
INSERT INTO playlist_track VALUES(1,3405);
INSERT INTO playlist_track VALUES(1,3488);
INSERT INTO playlist_track VALUES(1,3423);
INSERT INTO playlist_track VALUES(1,3499);
INSERT INTO playlist_track VALUES(1,3445);
INSERT INTO playlist_track VALUES(1,3440);
INSERT INTO playlist_track VALUES(1,3453);
INSERT INTO playlist_track VALUES(1,3497);
INSERT INTO playlist_track VALUES(1,3494);
INSERT INTO playlist_track VALUES(1,3439);
INSERT INTO playlist_track VALUES(1,3422);
INSERT INTO playlist_track VALUES(1,3407);
INSERT INTO playlist_track VALUES(1,3495);
INSERT INTO playlist_track VALUES(1,3435);
INSERT INTO playlist_track VALUES(1,3490);
INSERT INTO playlist_track VALUES(1,3489);
INSERT INTO playlist_track VALUES(1,3448);
INSERT INTO playlist_track VALUES(1,3492);
INSERT INTO playlist_track VALUES(1,3425);
INSERT INTO playlist_track VALUES(1,3483);
INSERT INTO playlist_track VALUES(1,3420);
INSERT INTO playlist_track VALUES(1,3424);
INSERT INTO playlist_track VALUES(1,3493);
INSERT INTO playlist_track VALUES(1,3437);
INSERT INTO playlist_track VALUES(1,3498);
INSERT INTO playlist_track VALUES(1,3446);
INSERT INTO playlist_track VALUES(1,3444);
INSERT INTO playlist_track VALUES(1,3496);
INSERT INTO playlist_track VALUES(1,3502);
INSERT INTO playlist_track VALUES(1,3359);
INSERT INTO playlist_track VALUES(1,3433);
INSERT INTO playlist_track VALUES(1,3415);
INSERT INTO playlist_track VALUES(1,3479);
INSERT INTO playlist_track VALUES(1,3481);
INSERT INTO playlist_track VALUES(1,3404);
INSERT INTO playlist_track VALUES(1,3486);
INSERT INTO playlist_track VALUES(1,3414);
INSERT INTO playlist_track VALUES(1,3410);
INSERT INTO playlist_track VALUES(1,3431);
INSERT INTO playlist_track VALUES(1,3418);
INSERT INTO playlist_track VALUES(1,3430);
INSERT INTO playlist_track VALUES(1,3408);
INSERT INTO playlist_track VALUES(1,3480);
INSERT INTO playlist_track VALUES(1,3409);
INSERT INTO playlist_track VALUES(1,3484);
INSERT INTO playlist_track VALUES(1,1033);
INSERT INTO playlist_track VALUES(1,1034);
INSERT INTO playlist_track VALUES(1,1035);
INSERT INTO playlist_track VALUES(1,1036);
INSERT INTO playlist_track VALUES(1,1037);
INSERT INTO playlist_track VALUES(1,1038);
INSERT INTO playlist_track VALUES(1,1039);
INSERT INTO playlist_track VALUES(1,1040);
INSERT INTO playlist_track VALUES(1,1041);
INSERT INTO playlist_track VALUES(1,1042);
INSERT INTO playlist_track VALUES(1,1043);
INSERT INTO playlist_track VALUES(1,1044);
INSERT INTO playlist_track VALUES(1,1045);
INSERT INTO playlist_track VALUES(1,1046);
INSERT INTO playlist_track VALUES(1,1047);
INSERT INTO playlist_track VALUES(1,1048);
INSERT INTO playlist_track VALUES(1,1049);
INSERT INTO playlist_track VALUES(1,1050);
INSERT INTO playlist_track VALUES(1,1051);
INSERT INTO playlist_track VALUES(1,1052);
INSERT INTO playlist_track VALUES(1,1053);
INSERT INTO playlist_track VALUES(1,1054);
INSERT INTO playlist_track VALUES(1,1055);
INSERT INTO playlist_track VALUES(1,1056);
INSERT INTO playlist_track VALUES(1,3324);
INSERT INTO playlist_track VALUES(1,3331);
INSERT INTO playlist_track VALUES(1,3332);
INSERT INTO playlist_track VALUES(1,3322);
INSERT INTO playlist_track VALUES(1,3329);
INSERT INTO playlist_track VALUES(1,1455);
INSERT INTO playlist_track VALUES(1,1456);
INSERT INTO playlist_track VALUES(1,1457);
INSERT INTO playlist_track VALUES(1,1458);
INSERT INTO playlist_track VALUES(1,1459);
INSERT INTO playlist_track VALUES(1,1460);
INSERT INTO playlist_track VALUES(1,1461);
INSERT INTO playlist_track VALUES(1,1462);
INSERT INTO playlist_track VALUES(1,1463);
INSERT INTO playlist_track VALUES(1,1464);
INSERT INTO playlist_track VALUES(1,1465);
INSERT INTO playlist_track VALUES(1,3352);
INSERT INTO playlist_track VALUES(1,3358);
INSERT INTO playlist_track VALUES(1,3326);
INSERT INTO playlist_track VALUES(1,3327);
INSERT INTO playlist_track VALUES(1,3330);
INSERT INTO playlist_track VALUES(1,3321);
INSERT INTO playlist_track VALUES(1,3319);
INSERT INTO playlist_track VALUES(1,3328);
INSERT INTO playlist_track VALUES(1,3325);
INSERT INTO playlist_track VALUES(1,3323);
INSERT INTO playlist_track VALUES(1,3334);
INSERT INTO playlist_track VALUES(1,3333);
INSERT INTO playlist_track VALUES(1,3335);
INSERT INTO playlist_track VALUES(1,3320);
INSERT INTO playlist_track VALUES(1,1245);
INSERT INTO playlist_track VALUES(1,1246);
INSERT INTO playlist_track VALUES(1,1247);
INSERT INTO playlist_track VALUES(1,1248);
INSERT INTO playlist_track VALUES(1,1249);
INSERT INTO playlist_track VALUES(1,1250);
INSERT INTO playlist_track VALUES(1,1251);
INSERT INTO playlist_track VALUES(1,1252);
INSERT INTO playlist_track VALUES(1,1253);
INSERT INTO playlist_track VALUES(1,1254);
INSERT INTO playlist_track VALUES(1,1255);
INSERT INTO playlist_track VALUES(1,1277);
INSERT INTO playlist_track VALUES(1,1278);
INSERT INTO playlist_track VALUES(1,1279);
INSERT INTO playlist_track VALUES(1,1280);
INSERT INTO playlist_track VALUES(1,1281);
INSERT INTO playlist_track VALUES(1,1282);
INSERT INTO playlist_track VALUES(1,1283);
INSERT INTO playlist_track VALUES(1,1284);
INSERT INTO playlist_track VALUES(1,1285);
INSERT INTO playlist_track VALUES(1,1286);
INSERT INTO playlist_track VALUES(1,1287);
INSERT INTO playlist_track VALUES(1,1288);
INSERT INTO playlist_track VALUES(1,1300);
INSERT INTO playlist_track VALUES(1,1301);
INSERT INTO playlist_track VALUES(1,1302);
INSERT INTO playlist_track VALUES(1,1303);
INSERT INTO playlist_track VALUES(1,1304);
INSERT INTO playlist_track VALUES(1,3301);
INSERT INTO playlist_track VALUES(1,3300);
INSERT INTO playlist_track VALUES(1,3302);
INSERT INTO playlist_track VALUES(1,3303);
INSERT INTO playlist_track VALUES(1,3304);
INSERT INTO playlist_track VALUES(1,3305);
INSERT INTO playlist_track VALUES(1,3306);
INSERT INTO playlist_track VALUES(1,3307);
INSERT INTO playlist_track VALUES(1,3308);
INSERT INTO playlist_track VALUES(1,3309);
INSERT INTO playlist_track VALUES(1,3310);
INSERT INTO playlist_track VALUES(1,3311);
INSERT INTO playlist_track VALUES(1,3312);
INSERT INTO playlist_track VALUES(1,3313);
INSERT INTO playlist_track VALUES(1,3314);
INSERT INTO playlist_track VALUES(1,3315);
INSERT INTO playlist_track VALUES(1,3316);
INSERT INTO playlist_track VALUES(1,3317);
INSERT INTO playlist_track VALUES(1,3318);
INSERT INTO playlist_track VALUES(1,2238);
INSERT INTO playlist_track VALUES(1,2239);
INSERT INTO playlist_track VALUES(1,2240);
INSERT INTO playlist_track VALUES(1,2241);
INSERT INTO playlist_track VALUES(1,2242);
INSERT INTO playlist_track VALUES(1,2243);
INSERT INTO playlist_track VALUES(1,2244);
INSERT INTO playlist_track VALUES(1,2245);
INSERT INTO playlist_track VALUES(1,2246);
INSERT INTO playlist_track VALUES(1,2247);
INSERT INTO playlist_track VALUES(1,2248);
INSERT INTO playlist_track VALUES(1,2249);
INSERT INTO playlist_track VALUES(1,2250);
INSERT INTO playlist_track VALUES(1,2251);
INSERT INTO playlist_track VALUES(1,2252);
INSERT INTO playlist_track VALUES(1,2253);
INSERT INTO playlist_track VALUES(1,3357);
INSERT INTO playlist_track VALUES(1,3350);
INSERT INTO playlist_track VALUES(1,3349);
INSERT INTO playlist_track VALUES(1,63);
INSERT INTO playlist_track VALUES(1,64);
INSERT INTO playlist_track VALUES(1,65);
INSERT INTO playlist_track VALUES(1,66);
INSERT INTO playlist_track VALUES(1,67);
INSERT INTO playlist_track VALUES(1,68);
INSERT INTO playlist_track VALUES(1,69);
INSERT INTO playlist_track VALUES(1,70);
INSERT INTO playlist_track VALUES(1,71);
INSERT INTO playlist_track VALUES(1,72);
INSERT INTO playlist_track VALUES(1,73);
INSERT INTO playlist_track VALUES(1,74);
INSERT INTO playlist_track VALUES(1,75);
INSERT INTO playlist_track VALUES(1,76);
INSERT INTO playlist_track VALUES(1,123);
INSERT INTO playlist_track VALUES(1,124);
INSERT INTO playlist_track VALUES(1,125);
INSERT INTO playlist_track VALUES(1,126);
INSERT INTO playlist_track VALUES(1,127);
INSERT INTO playlist_track VALUES(1,128);
INSERT INTO playlist_track VALUES(1,129);
INSERT INTO playlist_track VALUES(1,130);
INSERT INTO playlist_track VALUES(1,842);
INSERT INTO playlist_track VALUES(1,843);
INSERT INTO playlist_track VALUES(1,844);
INSERT INTO playlist_track VALUES(1,845);
INSERT INTO playlist_track VALUES(1,846);
INSERT INTO playlist_track VALUES(1,847);
INSERT INTO playlist_track VALUES(1,848);
INSERT INTO playlist_track VALUES(1,849);
INSERT INTO playlist_track VALUES(1,850);
INSERT INTO playlist_track VALUES(1,624);
INSERT INTO playlist_track VALUES(1,625);
INSERT INTO playlist_track VALUES(1,626);
INSERT INTO playlist_track VALUES(1,627);
INSERT INTO playlist_track VALUES(1,628);
INSERT INTO playlist_track VALUES(1,629);
INSERT INTO playlist_track VALUES(1,630);
INSERT INTO playlist_track VALUES(1,631);
INSERT INTO playlist_track VALUES(1,632);
INSERT INTO playlist_track VALUES(1,633);
INSERT INTO playlist_track VALUES(1,634);
INSERT INTO playlist_track VALUES(1,635);
INSERT INTO playlist_track VALUES(1,636);
INSERT INTO playlist_track VALUES(1,637);
INSERT INTO playlist_track VALUES(1,638);
INSERT INTO playlist_track VALUES(1,639);
INSERT INTO playlist_track VALUES(1,640);
INSERT INTO playlist_track VALUES(1,641);
INSERT INTO playlist_track VALUES(1,642);
INSERT INTO playlist_track VALUES(1,643);
INSERT INTO playlist_track VALUES(1,644);
INSERT INTO playlist_track VALUES(1,645);
INSERT INTO playlist_track VALUES(1,1102);
INSERT INTO playlist_track VALUES(1,1103);
INSERT INTO playlist_track VALUES(1,1104);
INSERT INTO playlist_track VALUES(1,1188);
INSERT INTO playlist_track VALUES(1,1189);
INSERT INTO playlist_track VALUES(1,1190);
INSERT INTO playlist_track VALUES(1,1191);
INSERT INTO playlist_track VALUES(1,1192);
INSERT INTO playlist_track VALUES(1,1193);
INSERT INTO playlist_track VALUES(1,1194);
INSERT INTO playlist_track VALUES(1,1195);
INSERT INTO playlist_track VALUES(1,1196);
INSERT INTO playlist_track VALUES(1,1197);
INSERT INTO playlist_track VALUES(1,1198);
INSERT INTO playlist_track VALUES(1,1199);
INSERT INTO playlist_track VALUES(1,1200);
INSERT INTO playlist_track VALUES(1,597);
INSERT INTO playlist_track VALUES(1,598);
INSERT INTO playlist_track VALUES(1,599);
INSERT INTO playlist_track VALUES(1,600);
INSERT INTO playlist_track VALUES(1,601);
INSERT INTO playlist_track VALUES(1,602);
INSERT INTO playlist_track VALUES(1,603);
INSERT INTO playlist_track VALUES(1,604);
INSERT INTO playlist_track VALUES(1,605);
INSERT INTO playlist_track VALUES(1,606);
INSERT INTO playlist_track VALUES(1,607);
INSERT INTO playlist_track VALUES(1,608);
INSERT INTO playlist_track VALUES(1,609);
INSERT INTO playlist_track VALUES(1,610);
INSERT INTO playlist_track VALUES(1,611);
INSERT INTO playlist_track VALUES(1,612);
INSERT INTO playlist_track VALUES(1,613);
INSERT INTO playlist_track VALUES(1,614);
INSERT INTO playlist_track VALUES(1,615);
INSERT INTO playlist_track VALUES(1,616);
INSERT INTO playlist_track VALUES(1,617);
INSERT INTO playlist_track VALUES(1,618);
INSERT INTO playlist_track VALUES(1,619);
INSERT INTO playlist_track VALUES(1,1902);
INSERT INTO playlist_track VALUES(1,1903);
INSERT INTO playlist_track VALUES(1,1904);
INSERT INTO playlist_track VALUES(1,1905);
INSERT INTO playlist_track VALUES(1,1906);
INSERT INTO playlist_track VALUES(1,1907);
INSERT INTO playlist_track VALUES(1,1908);
INSERT INTO playlist_track VALUES(1,1909);
INSERT INTO playlist_track VALUES(1,1910);
INSERT INTO playlist_track VALUES(1,1911);
INSERT INTO playlist_track VALUES(1,1912);
INSERT INTO playlist_track VALUES(1,1913);
INSERT INTO playlist_track VALUES(1,1914);
INSERT INTO playlist_track VALUES(1,1915);
INSERT INTO playlist_track VALUES(1,456);
INSERT INTO playlist_track VALUES(1,457);
INSERT INTO playlist_track VALUES(1,458);
INSERT INTO playlist_track VALUES(1,459);
INSERT INTO playlist_track VALUES(1,460);
INSERT INTO playlist_track VALUES(1,461);
INSERT INTO playlist_track VALUES(1,462);
INSERT INTO playlist_track VALUES(1,463);
INSERT INTO playlist_track VALUES(1,464);
INSERT INTO playlist_track VALUES(1,465);
INSERT INTO playlist_track VALUES(1,466);
INSERT INTO playlist_track VALUES(1,467);
INSERT INTO playlist_track VALUES(1,2523);
INSERT INTO playlist_track VALUES(1,2524);
INSERT INTO playlist_track VALUES(1,2525);
INSERT INTO playlist_track VALUES(1,2526);
INSERT INTO playlist_track VALUES(1,2527);
INSERT INTO playlist_track VALUES(1,2528);
INSERT INTO playlist_track VALUES(1,2529);
INSERT INTO playlist_track VALUES(1,2530);
INSERT INTO playlist_track VALUES(1,2531);
INSERT INTO playlist_track VALUES(1,379);
INSERT INTO playlist_track VALUES(1,391);
INSERT INTO playlist_track VALUES(1,376);
INSERT INTO playlist_track VALUES(1,397);
INSERT INTO playlist_track VALUES(1,382);
INSERT INTO playlist_track VALUES(1,389);
INSERT INTO playlist_track VALUES(1,404);
INSERT INTO playlist_track VALUES(1,406);
INSERT INTO playlist_track VALUES(1,380);
INSERT INTO playlist_track VALUES(1,394);
INSERT INTO playlist_track VALUES(1,515);
INSERT INTO playlist_track VALUES(1,516);
INSERT INTO playlist_track VALUES(1,517);
INSERT INTO playlist_track VALUES(1,518);
INSERT INTO playlist_track VALUES(1,519);
INSERT INTO playlist_track VALUES(1,520);
INSERT INTO playlist_track VALUES(1,521);
INSERT INTO playlist_track VALUES(1,522);
INSERT INTO playlist_track VALUES(1,523);
INSERT INTO playlist_track VALUES(1,524);
INSERT INTO playlist_track VALUES(1,525);
INSERT INTO playlist_track VALUES(1,526);
INSERT INTO playlist_track VALUES(1,527);
INSERT INTO playlist_track VALUES(1,528);
INSERT INTO playlist_track VALUES(1,205);
INSERT INTO playlist_track VALUES(1,206);
INSERT INTO playlist_track VALUES(1,207);
INSERT INTO playlist_track VALUES(1,208);
INSERT INTO playlist_track VALUES(1,209);
INSERT INTO playlist_track VALUES(1,210);
INSERT INTO playlist_track VALUES(1,211);
INSERT INTO playlist_track VALUES(1,212);
INSERT INTO playlist_track VALUES(1,213);
INSERT INTO playlist_track VALUES(1,214);
INSERT INTO playlist_track VALUES(1,215);
INSERT INTO playlist_track VALUES(1,216);
INSERT INTO playlist_track VALUES(1,217);
INSERT INTO playlist_track VALUES(1,218);
INSERT INTO playlist_track VALUES(1,219);
INSERT INTO playlist_track VALUES(1,220);
INSERT INTO playlist_track VALUES(1,221);
INSERT INTO playlist_track VALUES(1,222);
INSERT INTO playlist_track VALUES(1,223);
INSERT INTO playlist_track VALUES(1,224);
INSERT INTO playlist_track VALUES(1,225);
INSERT INTO playlist_track VALUES(1,715);
INSERT INTO playlist_track VALUES(1,716);
INSERT INTO playlist_track VALUES(1,717);
INSERT INTO playlist_track VALUES(1,718);
INSERT INTO playlist_track VALUES(1,719);
INSERT INTO playlist_track VALUES(1,720);
INSERT INTO playlist_track VALUES(1,721);
INSERT INTO playlist_track VALUES(1,722);
INSERT INTO playlist_track VALUES(1,723);
INSERT INTO playlist_track VALUES(1,724);
INSERT INTO playlist_track VALUES(1,725);
INSERT INTO playlist_track VALUES(1,726);
INSERT INTO playlist_track VALUES(1,727);
INSERT INTO playlist_track VALUES(1,728);
INSERT INTO playlist_track VALUES(1,729);
INSERT INTO playlist_track VALUES(1,730);
INSERT INTO playlist_track VALUES(1,731);
INSERT INTO playlist_track VALUES(1,732);
INSERT INTO playlist_track VALUES(1,733);
INSERT INTO playlist_track VALUES(1,734);
INSERT INTO playlist_track VALUES(1,735);
INSERT INTO playlist_track VALUES(1,736);
INSERT INTO playlist_track VALUES(1,737);
INSERT INTO playlist_track VALUES(1,738);
INSERT INTO playlist_track VALUES(1,739);
INSERT INTO playlist_track VALUES(1,740);
INSERT INTO playlist_track VALUES(1,741);
INSERT INTO playlist_track VALUES(1,742);
INSERT INTO playlist_track VALUES(1,743);
INSERT INTO playlist_track VALUES(1,744);
INSERT INTO playlist_track VALUES(1,226);
INSERT INTO playlist_track VALUES(1,227);
INSERT INTO playlist_track VALUES(1,228);
INSERT INTO playlist_track VALUES(1,229);
INSERT INTO playlist_track VALUES(1,230);
INSERT INTO playlist_track VALUES(1,231);
INSERT INTO playlist_track VALUES(1,232);
INSERT INTO playlist_track VALUES(1,233);
INSERT INTO playlist_track VALUES(1,234);
INSERT INTO playlist_track VALUES(1,235);
INSERT INTO playlist_track VALUES(1,236);
INSERT INTO playlist_track VALUES(1,237);
INSERT INTO playlist_track VALUES(1,238);
INSERT INTO playlist_track VALUES(1,239);
INSERT INTO playlist_track VALUES(1,240);
INSERT INTO playlist_track VALUES(1,241);
INSERT INTO playlist_track VALUES(1,242);
INSERT INTO playlist_track VALUES(1,243);
INSERT INTO playlist_track VALUES(1,244);
INSERT INTO playlist_track VALUES(1,245);
INSERT INTO playlist_track VALUES(1,246);
INSERT INTO playlist_track VALUES(1,247);
INSERT INTO playlist_track VALUES(1,248);
INSERT INTO playlist_track VALUES(1,249);
INSERT INTO playlist_track VALUES(1,250);
INSERT INTO playlist_track VALUES(1,251);
INSERT INTO playlist_track VALUES(1,252);
INSERT INTO playlist_track VALUES(1,253);
INSERT INTO playlist_track VALUES(1,254);
INSERT INTO playlist_track VALUES(1,255);
INSERT INTO playlist_track VALUES(1,256);
INSERT INTO playlist_track VALUES(1,257);
INSERT INTO playlist_track VALUES(1,258);
INSERT INTO playlist_track VALUES(1,259);
INSERT INTO playlist_track VALUES(1,260);
INSERT INTO playlist_track VALUES(1,261);
INSERT INTO playlist_track VALUES(1,262);
INSERT INTO playlist_track VALUES(1,263);
INSERT INTO playlist_track VALUES(1,264);
INSERT INTO playlist_track VALUES(1,265);
INSERT INTO playlist_track VALUES(1,266);
INSERT INTO playlist_track VALUES(1,267);
INSERT INTO playlist_track VALUES(1,268);
INSERT INTO playlist_track VALUES(1,269);
INSERT INTO playlist_track VALUES(1,270);
INSERT INTO playlist_track VALUES(1,271);
INSERT INTO playlist_track VALUES(1,272);
INSERT INTO playlist_track VALUES(1,273);
INSERT INTO playlist_track VALUES(1,274);
INSERT INTO playlist_track VALUES(1,275);
INSERT INTO playlist_track VALUES(1,276);
INSERT INTO playlist_track VALUES(1,277);
INSERT INTO playlist_track VALUES(1,278);
INSERT INTO playlist_track VALUES(1,279);
INSERT INTO playlist_track VALUES(1,280);
INSERT INTO playlist_track VALUES(1,281);
INSERT INTO playlist_track VALUES(1,313);
INSERT INTO playlist_track VALUES(1,314);
INSERT INTO playlist_track VALUES(1,315);
INSERT INTO playlist_track VALUES(1,316);
INSERT INTO playlist_track VALUES(1,317);
INSERT INTO playlist_track VALUES(1,318);
INSERT INTO playlist_track VALUES(1,319);
INSERT INTO playlist_track VALUES(1,320);
INSERT INTO playlist_track VALUES(1,321);
INSERT INTO playlist_track VALUES(1,322);
INSERT INTO playlist_track VALUES(1,399);
INSERT INTO playlist_track VALUES(1,851);
INSERT INTO playlist_track VALUES(1,852);
INSERT INTO playlist_track VALUES(1,853);
INSERT INTO playlist_track VALUES(1,854);
INSERT INTO playlist_track VALUES(1,855);
INSERT INTO playlist_track VALUES(1,856);
INSERT INTO playlist_track VALUES(1,857);
INSERT INTO playlist_track VALUES(1,858);
INSERT INTO playlist_track VALUES(1,859);
INSERT INTO playlist_track VALUES(1,860);
INSERT INTO playlist_track VALUES(1,861);
INSERT INTO playlist_track VALUES(1,862);
INSERT INTO playlist_track VALUES(1,863);
INSERT INTO playlist_track VALUES(1,864);
INSERT INTO playlist_track VALUES(1,865);
INSERT INTO playlist_track VALUES(1,866);
INSERT INTO playlist_track VALUES(1,867);
INSERT INTO playlist_track VALUES(1,868);
INSERT INTO playlist_track VALUES(1,869);
INSERT INTO playlist_track VALUES(1,870);
INSERT INTO playlist_track VALUES(1,871);
INSERT INTO playlist_track VALUES(1,872);
INSERT INTO playlist_track VALUES(1,873);
INSERT INTO playlist_track VALUES(1,874);
INSERT INTO playlist_track VALUES(1,875);
INSERT INTO playlist_track VALUES(1,876);
INSERT INTO playlist_track VALUES(1,583);
INSERT INTO playlist_track VALUES(1,584);
INSERT INTO playlist_track VALUES(1,585);
INSERT INTO playlist_track VALUES(1,586);
INSERT INTO playlist_track VALUES(1,587);
INSERT INTO playlist_track VALUES(1,588);
INSERT INTO playlist_track VALUES(1,589);
INSERT INTO playlist_track VALUES(1,590);
INSERT INTO playlist_track VALUES(1,591);
INSERT INTO playlist_track VALUES(1,592);
INSERT INTO playlist_track VALUES(1,593);
INSERT INTO playlist_track VALUES(1,594);
INSERT INTO playlist_track VALUES(1,595);
INSERT INTO playlist_track VALUES(1,596);
INSERT INTO playlist_track VALUES(1,388);
INSERT INTO playlist_track VALUES(1,402);
INSERT INTO playlist_track VALUES(1,407);
INSERT INTO playlist_track VALUES(1,396);
INSERT INTO playlist_track VALUES(1,877);
INSERT INTO playlist_track VALUES(1,878);
INSERT INTO playlist_track VALUES(1,879);
INSERT INTO playlist_track VALUES(1,880);
INSERT INTO playlist_track VALUES(1,881);
INSERT INTO playlist_track VALUES(1,882);
INSERT INTO playlist_track VALUES(1,883);
INSERT INTO playlist_track VALUES(1,884);
INSERT INTO playlist_track VALUES(1,885);
INSERT INTO playlist_track VALUES(1,886);
INSERT INTO playlist_track VALUES(1,887);
INSERT INTO playlist_track VALUES(1,888);
INSERT INTO playlist_track VALUES(1,889);
INSERT INTO playlist_track VALUES(1,890);
INSERT INTO playlist_track VALUES(1,975);
INSERT INTO playlist_track VALUES(1,976);
INSERT INTO playlist_track VALUES(1,977);
INSERT INTO playlist_track VALUES(1,978);
INSERT INTO playlist_track VALUES(1,979);
INSERT INTO playlist_track VALUES(1,980);
INSERT INTO playlist_track VALUES(1,981);
INSERT INTO playlist_track VALUES(1,982);
INSERT INTO playlist_track VALUES(1,983);
INSERT INTO playlist_track VALUES(1,984);
INSERT INTO playlist_track VALUES(1,985);
INSERT INTO playlist_track VALUES(1,986);
INSERT INTO playlist_track VALUES(1,987);
INSERT INTO playlist_track VALUES(1,988);
INSERT INTO playlist_track VALUES(1,390);
INSERT INTO playlist_track VALUES(1,1057);
INSERT INTO playlist_track VALUES(1,1058);
INSERT INTO playlist_track VALUES(1,1059);
INSERT INTO playlist_track VALUES(1,1060);
INSERT INTO playlist_track VALUES(1,1061);
INSERT INTO playlist_track VALUES(1,1062);
INSERT INTO playlist_track VALUES(1,1063);
INSERT INTO playlist_track VALUES(1,1064);
INSERT INTO playlist_track VALUES(1,1065);
INSERT INTO playlist_track VALUES(1,1066);
INSERT INTO playlist_track VALUES(1,1067);
INSERT INTO playlist_track VALUES(1,1068);
INSERT INTO playlist_track VALUES(1,1069);
INSERT INTO playlist_track VALUES(1,1070);
INSERT INTO playlist_track VALUES(1,1071);
INSERT INTO playlist_track VALUES(1,1072);
INSERT INTO playlist_track VALUES(1,377);
INSERT INTO playlist_track VALUES(1,395);
INSERT INTO playlist_track VALUES(1,1087);
INSERT INTO playlist_track VALUES(1,1088);
INSERT INTO playlist_track VALUES(1,1089);
INSERT INTO playlist_track VALUES(1,1090);
INSERT INTO playlist_track VALUES(1,1091);
INSERT INTO playlist_track VALUES(1,1092);
INSERT INTO playlist_track VALUES(1,1093);
INSERT INTO playlist_track VALUES(1,1094);
INSERT INTO playlist_track VALUES(1,1095);
INSERT INTO playlist_track VALUES(1,1096);
INSERT INTO playlist_track VALUES(1,1097);
INSERT INTO playlist_track VALUES(1,1098);
INSERT INTO playlist_track VALUES(1,1099);
INSERT INTO playlist_track VALUES(1,1100);
INSERT INTO playlist_track VALUES(1,1101);
INSERT INTO playlist_track VALUES(1,1105);
INSERT INTO playlist_track VALUES(1,1106);
INSERT INTO playlist_track VALUES(1,1107);
INSERT INTO playlist_track VALUES(1,1108);
INSERT INTO playlist_track VALUES(1,1109);
INSERT INTO playlist_track VALUES(1,1110);
INSERT INTO playlist_track VALUES(1,1111);
INSERT INTO playlist_track VALUES(1,1112);
INSERT INTO playlist_track VALUES(1,1113);
INSERT INTO playlist_track VALUES(1,1114);
INSERT INTO playlist_track VALUES(1,1115);
INSERT INTO playlist_track VALUES(1,1116);
INSERT INTO playlist_track VALUES(1,1117);
INSERT INTO playlist_track VALUES(1,1118);
INSERT INTO playlist_track VALUES(1,1119);
INSERT INTO playlist_track VALUES(1,1120);
INSERT INTO playlist_track VALUES(1,501);
INSERT INTO playlist_track VALUES(1,502);
INSERT INTO playlist_track VALUES(1,503);
INSERT INTO playlist_track VALUES(1,504);
INSERT INTO playlist_track VALUES(1,505);
INSERT INTO playlist_track VALUES(1,506);
INSERT INTO playlist_track VALUES(1,507);
INSERT INTO playlist_track VALUES(1,508);
INSERT INTO playlist_track VALUES(1,509);
INSERT INTO playlist_track VALUES(1,510);
INSERT INTO playlist_track VALUES(1,511);
INSERT INTO playlist_track VALUES(1,512);
INSERT INTO playlist_track VALUES(1,513);
INSERT INTO playlist_track VALUES(1,514);
INSERT INTO playlist_track VALUES(1,405);
INSERT INTO playlist_track VALUES(1,378);
INSERT INTO playlist_track VALUES(1,392);
INSERT INTO playlist_track VALUES(1,403);
INSERT INTO playlist_track VALUES(1,1506);
INSERT INTO playlist_track VALUES(1,1507);
INSERT INTO playlist_track VALUES(1,1508);
INSERT INTO playlist_track VALUES(1,1509);
INSERT INTO playlist_track VALUES(1,1510);
INSERT INTO playlist_track VALUES(1,1511);
INSERT INTO playlist_track VALUES(1,1512);
INSERT INTO playlist_track VALUES(1,1513);
INSERT INTO playlist_track VALUES(1,1514);
INSERT INTO playlist_track VALUES(1,1515);
INSERT INTO playlist_track VALUES(1,1516);
INSERT INTO playlist_track VALUES(1,1517);
INSERT INTO playlist_track VALUES(1,1518);
INSERT INTO playlist_track VALUES(1,1519);
INSERT INTO playlist_track VALUES(1,381);
INSERT INTO playlist_track VALUES(1,1520);
INSERT INTO playlist_track VALUES(1,1521);
INSERT INTO playlist_track VALUES(1,1522);
INSERT INTO playlist_track VALUES(1,1523);
INSERT INTO playlist_track VALUES(1,1524);
INSERT INTO playlist_track VALUES(1,1525);
INSERT INTO playlist_track VALUES(1,1526);
INSERT INTO playlist_track VALUES(1,1527);
INSERT INTO playlist_track VALUES(1,1528);
INSERT INTO playlist_track VALUES(1,1529);
INSERT INTO playlist_track VALUES(1,1530);
INSERT INTO playlist_track VALUES(1,1531);
INSERT INTO playlist_track VALUES(1,400);
INSERT INTO playlist_track VALUES(1,1686);
INSERT INTO playlist_track VALUES(1,1687);
INSERT INTO playlist_track VALUES(1,1688);
INSERT INTO playlist_track VALUES(1,1689);
INSERT INTO playlist_track VALUES(1,1690);
INSERT INTO playlist_track VALUES(1,1691);
INSERT INTO playlist_track VALUES(1,1692);
INSERT INTO playlist_track VALUES(1,1693);
INSERT INTO playlist_track VALUES(1,1694);
INSERT INTO playlist_track VALUES(1,1695);
INSERT INTO playlist_track VALUES(1,1696);
INSERT INTO playlist_track VALUES(1,1697);
INSERT INTO playlist_track VALUES(1,1698);
INSERT INTO playlist_track VALUES(1,1699);
INSERT INTO playlist_track VALUES(1,1700);
INSERT INTO playlist_track VALUES(1,1701);
INSERT INTO playlist_track VALUES(1,1671);
INSERT INTO playlist_track VALUES(1,1672);
INSERT INTO playlist_track VALUES(1,1673);
INSERT INTO playlist_track VALUES(1,1674);
INSERT INTO playlist_track VALUES(1,1675);
INSERT INTO playlist_track VALUES(1,1676);
INSERT INTO playlist_track VALUES(1,1677);
INSERT INTO playlist_track VALUES(1,1678);
INSERT INTO playlist_track VALUES(1,1679);
INSERT INTO playlist_track VALUES(1,1680);
INSERT INTO playlist_track VALUES(1,1681);
INSERT INTO playlist_track VALUES(1,1682);
INSERT INTO playlist_track VALUES(1,1683);
INSERT INTO playlist_track VALUES(1,1684);
INSERT INTO playlist_track VALUES(1,1685);
INSERT INTO playlist_track VALUES(1,3356);
INSERT INTO playlist_track VALUES(1,384);
INSERT INTO playlist_track VALUES(1,1717);
INSERT INTO playlist_track VALUES(1,1720);
INSERT INTO playlist_track VALUES(1,1722);
INSERT INTO playlist_track VALUES(1,1723);
INSERT INTO playlist_track VALUES(1,1726);
INSERT INTO playlist_track VALUES(1,1727);
INSERT INTO playlist_track VALUES(1,1730);
INSERT INTO playlist_track VALUES(1,1731);
INSERT INTO playlist_track VALUES(1,1733);
INSERT INTO playlist_track VALUES(1,1736);
INSERT INTO playlist_track VALUES(1,1737);
INSERT INTO playlist_track VALUES(1,1740);
INSERT INTO playlist_track VALUES(1,1742);
INSERT INTO playlist_track VALUES(1,1743);
INSERT INTO playlist_track VALUES(1,1718);
INSERT INTO playlist_track VALUES(1,1719);
INSERT INTO playlist_track VALUES(1,1721);
INSERT INTO playlist_track VALUES(1,1724);
INSERT INTO playlist_track VALUES(1,1725);
INSERT INTO playlist_track VALUES(1,1728);
INSERT INTO playlist_track VALUES(1,1729);
INSERT INTO playlist_track VALUES(1,1732);
INSERT INTO playlist_track VALUES(1,1734);
INSERT INTO playlist_track VALUES(1,1735);
INSERT INTO playlist_track VALUES(1,1738);
INSERT INTO playlist_track VALUES(1,1739);
INSERT INTO playlist_track VALUES(1,1741);
INSERT INTO playlist_track VALUES(1,1744);
INSERT INTO playlist_track VALUES(1,374);
INSERT INTO playlist_track VALUES(1,1755);
INSERT INTO playlist_track VALUES(1,1762);
INSERT INTO playlist_track VALUES(1,1763);
INSERT INTO playlist_track VALUES(1,1756);
INSERT INTO playlist_track VALUES(1,1764);
INSERT INTO playlist_track VALUES(1,1757);
INSERT INTO playlist_track VALUES(1,1758);
INSERT INTO playlist_track VALUES(1,1765);
INSERT INTO playlist_track VALUES(1,1766);
INSERT INTO playlist_track VALUES(1,1759);
INSERT INTO playlist_track VALUES(1,1760);
INSERT INTO playlist_track VALUES(1,1767);
INSERT INTO playlist_track VALUES(1,1761);
INSERT INTO playlist_track VALUES(1,1768);
INSERT INTO playlist_track VALUES(1,1769);
INSERT INTO playlist_track VALUES(1,1770);
INSERT INTO playlist_track VALUES(1,1771);
INSERT INTO playlist_track VALUES(1,1772);
INSERT INTO playlist_track VALUES(1,398);
INSERT INTO playlist_track VALUES(1,1916);
INSERT INTO playlist_track VALUES(1,1917);
INSERT INTO playlist_track VALUES(1,1918);
INSERT INTO playlist_track VALUES(1,1919);
INSERT INTO playlist_track VALUES(1,1920);
INSERT INTO playlist_track VALUES(1,1921);
INSERT INTO playlist_track VALUES(1,1922);
INSERT INTO playlist_track VALUES(1,1923);
INSERT INTO playlist_track VALUES(1,1924);
INSERT INTO playlist_track VALUES(1,1925);
INSERT INTO playlist_track VALUES(1,1926);
INSERT INTO playlist_track VALUES(1,1927);
INSERT INTO playlist_track VALUES(1,1928);
INSERT INTO playlist_track VALUES(1,1929);
INSERT INTO playlist_track VALUES(1,1930);
INSERT INTO playlist_track VALUES(1,1931);
INSERT INTO playlist_track VALUES(1,1932);
INSERT INTO playlist_track VALUES(1,1933);
INSERT INTO playlist_track VALUES(1,1934);
INSERT INTO playlist_track VALUES(1,1935);
INSERT INTO playlist_track VALUES(1,1936);
INSERT INTO playlist_track VALUES(1,1937);
INSERT INTO playlist_track VALUES(1,1938);
INSERT INTO playlist_track VALUES(1,1939);
INSERT INTO playlist_track VALUES(1,1940);
INSERT INTO playlist_track VALUES(1,1941);
INSERT INTO playlist_track VALUES(1,375);
INSERT INTO playlist_track VALUES(1,385);
INSERT INTO playlist_track VALUES(1,383);
INSERT INTO playlist_track VALUES(1,387);
INSERT INTO playlist_track VALUES(1,2030);
INSERT INTO playlist_track VALUES(1,2031);
INSERT INTO playlist_track VALUES(1,2032);
INSERT INTO playlist_track VALUES(1,2033);
INSERT INTO playlist_track VALUES(1,2034);
INSERT INTO playlist_track VALUES(1,2035);
INSERT INTO playlist_track VALUES(1,2036);
INSERT INTO playlist_track VALUES(1,2037);
INSERT INTO playlist_track VALUES(1,2038);
INSERT INTO playlist_track VALUES(1,2039);
INSERT INTO playlist_track VALUES(1,2040);
INSERT INTO playlist_track VALUES(1,2041);
INSERT INTO playlist_track VALUES(1,2042);
INSERT INTO playlist_track VALUES(1,2043);
INSERT INTO playlist_track VALUES(1,393);
INSERT INTO playlist_track VALUES(1,2044);
INSERT INTO playlist_track VALUES(1,2045);
INSERT INTO playlist_track VALUES(1,2046);
INSERT INTO playlist_track VALUES(1,2047);
INSERT INTO playlist_track VALUES(1,2048);
INSERT INTO playlist_track VALUES(1,2049);
INSERT INTO playlist_track VALUES(1,2050);
INSERT INTO playlist_track VALUES(1,2051);
INSERT INTO playlist_track VALUES(1,2052);
INSERT INTO playlist_track VALUES(1,2053);
INSERT INTO playlist_track VALUES(1,2054);
INSERT INTO playlist_track VALUES(1,2055);
INSERT INTO playlist_track VALUES(1,2056);
INSERT INTO playlist_track VALUES(1,2057);
INSERT INTO playlist_track VALUES(1,2058);
INSERT INTO playlist_track VALUES(1,2059);
INSERT INTO playlist_track VALUES(1,2060);
INSERT INTO playlist_track VALUES(1,2061);
INSERT INTO playlist_track VALUES(1,2062);
INSERT INTO playlist_track VALUES(1,2063);
INSERT INTO playlist_track VALUES(1,2064);
INSERT INTO playlist_track VALUES(1,2065);
INSERT INTO playlist_track VALUES(1,2066);
INSERT INTO playlist_track VALUES(1,2067);
INSERT INTO playlist_track VALUES(1,2068);
INSERT INTO playlist_track VALUES(1,2069);
INSERT INTO playlist_track VALUES(1,2070);
INSERT INTO playlist_track VALUES(1,2071);
INSERT INTO playlist_track VALUES(1,2072);
INSERT INTO playlist_track VALUES(1,2073);
INSERT INTO playlist_track VALUES(1,2074);
INSERT INTO playlist_track VALUES(1,2075);
INSERT INTO playlist_track VALUES(1,2076);
INSERT INTO playlist_track VALUES(1,2077);
INSERT INTO playlist_track VALUES(1,2078);
INSERT INTO playlist_track VALUES(1,2079);
INSERT INTO playlist_track VALUES(1,2080);
INSERT INTO playlist_track VALUES(1,2081);
INSERT INTO playlist_track VALUES(1,2082);
INSERT INTO playlist_track VALUES(1,2083);
INSERT INTO playlist_track VALUES(1,2084);
INSERT INTO playlist_track VALUES(1,2085);
INSERT INTO playlist_track VALUES(1,2086);
INSERT INTO playlist_track VALUES(1,2087);
INSERT INTO playlist_track VALUES(1,2088);
INSERT INTO playlist_track VALUES(1,2089);
INSERT INTO playlist_track VALUES(1,2090);
INSERT INTO playlist_track VALUES(1,2091);
INSERT INTO playlist_track VALUES(1,2092);
INSERT INTO playlist_track VALUES(1,386);
INSERT INTO playlist_track VALUES(1,401);
INSERT INTO playlist_track VALUES(1,2751);
INSERT INTO playlist_track VALUES(1,2752);
INSERT INTO playlist_track VALUES(1,2753);
INSERT INTO playlist_track VALUES(1,2754);
INSERT INTO playlist_track VALUES(1,2755);
INSERT INTO playlist_track VALUES(1,2756);
INSERT INTO playlist_track VALUES(1,2757);
INSERT INTO playlist_track VALUES(1,2758);
INSERT INTO playlist_track VALUES(1,2759);
INSERT INTO playlist_track VALUES(1,2760);
INSERT INTO playlist_track VALUES(1,2761);
INSERT INTO playlist_track VALUES(1,2762);
INSERT INTO playlist_track VALUES(1,2763);
INSERT INTO playlist_track VALUES(1,2764);
INSERT INTO playlist_track VALUES(1,2765);
INSERT INTO playlist_track VALUES(1,2766);
INSERT INTO playlist_track VALUES(1,2767);
INSERT INTO playlist_track VALUES(1,2768);
INSERT INTO playlist_track VALUES(1,2769);
INSERT INTO playlist_track VALUES(1,2770);
INSERT INTO playlist_track VALUES(1,2771);
INSERT INTO playlist_track VALUES(1,2772);
INSERT INTO playlist_track VALUES(1,2773);
INSERT INTO playlist_track VALUES(1,2774);
INSERT INTO playlist_track VALUES(1,2775);
INSERT INTO playlist_track VALUES(1,2776);
INSERT INTO playlist_track VALUES(1,2777);
INSERT INTO playlist_track VALUES(1,2778);
INSERT INTO playlist_track VALUES(1,2779);
INSERT INTO playlist_track VALUES(1,2780);
INSERT INTO playlist_track VALUES(1,556);
INSERT INTO playlist_track VALUES(1,557);
INSERT INTO playlist_track VALUES(1,558);
INSERT INTO playlist_track VALUES(1,559);
INSERT INTO playlist_track VALUES(1,560);
INSERT INTO playlist_track VALUES(1,561);
INSERT INTO playlist_track VALUES(1,562);
INSERT INTO playlist_track VALUES(1,563);
INSERT INTO playlist_track VALUES(1,564);
INSERT INTO playlist_track VALUES(1,565);
INSERT INTO playlist_track VALUES(1,566);
INSERT INTO playlist_track VALUES(1,567);
INSERT INTO playlist_track VALUES(1,568);
INSERT INTO playlist_track VALUES(1,569);
INSERT INTO playlist_track VALUES(1,661);
INSERT INTO playlist_track VALUES(1,662);
INSERT INTO playlist_track VALUES(1,663);
INSERT INTO playlist_track VALUES(1,664);
INSERT INTO playlist_track VALUES(1,665);
INSERT INTO playlist_track VALUES(1,666);
INSERT INTO playlist_track VALUES(1,667);
INSERT INTO playlist_track VALUES(1,668);
INSERT INTO playlist_track VALUES(1,669);
INSERT INTO playlist_track VALUES(1,670);
INSERT INTO playlist_track VALUES(1,671);
INSERT INTO playlist_track VALUES(1,672);
INSERT INTO playlist_track VALUES(1,673);
INSERT INTO playlist_track VALUES(1,674);
INSERT INTO playlist_track VALUES(1,3117);
INSERT INTO playlist_track VALUES(1,3118);
INSERT INTO playlist_track VALUES(1,3119);
INSERT INTO playlist_track VALUES(1,3120);
INSERT INTO playlist_track VALUES(1,3121);
INSERT INTO playlist_track VALUES(1,3122);
INSERT INTO playlist_track VALUES(1,3123);
INSERT INTO playlist_track VALUES(1,3124);
INSERT INTO playlist_track VALUES(1,3125);
INSERT INTO playlist_track VALUES(1,3126);
INSERT INTO playlist_track VALUES(1,3127);
INSERT INTO playlist_track VALUES(1,3128);
INSERT INTO playlist_track VALUES(1,3129);
INSERT INTO playlist_track VALUES(1,3130);
INSERT INTO playlist_track VALUES(1,3131);
INSERT INTO playlist_track VALUES(1,3146);
INSERT INTO playlist_track VALUES(1,3147);
INSERT INTO playlist_track VALUES(1,3148);
INSERT INTO playlist_track VALUES(1,3149);
INSERT INTO playlist_track VALUES(1,3150);
INSERT INTO playlist_track VALUES(1,3151);
INSERT INTO playlist_track VALUES(1,3152);
INSERT INTO playlist_track VALUES(1,3153);
INSERT INTO playlist_track VALUES(1,3154);
INSERT INTO playlist_track VALUES(1,3155);
INSERT INTO playlist_track VALUES(1,3156);
INSERT INTO playlist_track VALUES(1,3157);
INSERT INTO playlist_track VALUES(1,3158);
INSERT INTO playlist_track VALUES(1,3159);
INSERT INTO playlist_track VALUES(1,3160);
INSERT INTO playlist_track VALUES(1,3161);
INSERT INTO playlist_track VALUES(1,3162);
INSERT INTO playlist_track VALUES(1,3163);
INSERT INTO playlist_track VALUES(1,3164);
INSERT INTO playlist_track VALUES(1,77);
INSERT INTO playlist_track VALUES(1,78);
INSERT INTO playlist_track VALUES(1,79);
INSERT INTO playlist_track VALUES(1,80);
INSERT INTO playlist_track VALUES(1,81);
INSERT INTO playlist_track VALUES(1,82);
INSERT INTO playlist_track VALUES(1,83);
INSERT INTO playlist_track VALUES(1,84);
INSERT INTO playlist_track VALUES(1,131);
INSERT INTO playlist_track VALUES(1,132);
INSERT INTO playlist_track VALUES(1,133);
INSERT INTO playlist_track VALUES(1,134);
INSERT INTO playlist_track VALUES(1,135);
INSERT INTO playlist_track VALUES(1,136);
INSERT INTO playlist_track VALUES(1,137);
INSERT INTO playlist_track VALUES(1,138);
INSERT INTO playlist_track VALUES(1,139);
INSERT INTO playlist_track VALUES(1,140);
INSERT INTO playlist_track VALUES(1,141);
INSERT INTO playlist_track VALUES(1,142);
INSERT INTO playlist_track VALUES(1,143);
INSERT INTO playlist_track VALUES(1,144);
INSERT INTO playlist_track VALUES(1,145);
INSERT INTO playlist_track VALUES(1,146);
INSERT INTO playlist_track VALUES(1,147);
INSERT INTO playlist_track VALUES(1,148);
INSERT INTO playlist_track VALUES(1,149);
INSERT INTO playlist_track VALUES(1,150);
INSERT INTO playlist_track VALUES(1,151);
INSERT INTO playlist_track VALUES(1,152);
INSERT INTO playlist_track VALUES(1,153);
INSERT INTO playlist_track VALUES(1,154);
INSERT INTO playlist_track VALUES(1,155);
INSERT INTO playlist_track VALUES(1,156);
INSERT INTO playlist_track VALUES(1,157);
INSERT INTO playlist_track VALUES(1,158);
INSERT INTO playlist_track VALUES(1,159);
INSERT INTO playlist_track VALUES(1,160);
INSERT INTO playlist_track VALUES(1,161);
INSERT INTO playlist_track VALUES(1,162);
INSERT INTO playlist_track VALUES(1,163);
INSERT INTO playlist_track VALUES(1,164);
INSERT INTO playlist_track VALUES(1,165);
INSERT INTO playlist_track VALUES(1,183);
INSERT INTO playlist_track VALUES(1,184);
INSERT INTO playlist_track VALUES(1,185);
INSERT INTO playlist_track VALUES(1,186);
INSERT INTO playlist_track VALUES(1,187);
INSERT INTO playlist_track VALUES(1,188);
INSERT INTO playlist_track VALUES(1,189);
INSERT INTO playlist_track VALUES(1,190);
INSERT INTO playlist_track VALUES(1,191);
INSERT INTO playlist_track VALUES(1,192);
INSERT INTO playlist_track VALUES(1,193);
INSERT INTO playlist_track VALUES(1,1121);
INSERT INTO playlist_track VALUES(1,1122);
INSERT INTO playlist_track VALUES(1,1123);
INSERT INTO playlist_track VALUES(1,1124);
INSERT INTO playlist_track VALUES(1,1125);
INSERT INTO playlist_track VALUES(1,1126);
INSERT INTO playlist_track VALUES(1,1127);
INSERT INTO playlist_track VALUES(1,1128);
INSERT INTO playlist_track VALUES(1,1129);
INSERT INTO playlist_track VALUES(1,1130);
INSERT INTO playlist_track VALUES(1,1131);
INSERT INTO playlist_track VALUES(1,1132);
INSERT INTO playlist_track VALUES(1,1174);
INSERT INTO playlist_track VALUES(1,1175);
INSERT INTO playlist_track VALUES(1,1176);
INSERT INTO playlist_track VALUES(1,1177);
INSERT INTO playlist_track VALUES(1,1178);
INSERT INTO playlist_track VALUES(1,1179);
INSERT INTO playlist_track VALUES(1,1180);
INSERT INTO playlist_track VALUES(1,1181);
INSERT INTO playlist_track VALUES(1,1182);
INSERT INTO playlist_track VALUES(1,1183);
INSERT INTO playlist_track VALUES(1,1184);
INSERT INTO playlist_track VALUES(1,1185);
INSERT INTO playlist_track VALUES(1,1186);
INSERT INTO playlist_track VALUES(1,1187);
INSERT INTO playlist_track VALUES(1,1289);
INSERT INTO playlist_track VALUES(1,1290);
INSERT INTO playlist_track VALUES(1,1291);
INSERT INTO playlist_track VALUES(1,1292);
INSERT INTO playlist_track VALUES(1,1293);
INSERT INTO playlist_track VALUES(1,1294);
INSERT INTO playlist_track VALUES(1,1295);
INSERT INTO playlist_track VALUES(1,1296);
INSERT INTO playlist_track VALUES(1,1297);
INSERT INTO playlist_track VALUES(1,1298);
INSERT INTO playlist_track VALUES(1,1299);
INSERT INTO playlist_track VALUES(1,1325);
INSERT INTO playlist_track VALUES(1,1326);
INSERT INTO playlist_track VALUES(1,1327);
INSERT INTO playlist_track VALUES(1,1328);
INSERT INTO playlist_track VALUES(1,1329);
INSERT INTO playlist_track VALUES(1,1330);
INSERT INTO playlist_track VALUES(1,1331);
INSERT INTO playlist_track VALUES(1,1332);
INSERT INTO playlist_track VALUES(1,1333);
INSERT INTO playlist_track VALUES(1,1334);
INSERT INTO playlist_track VALUES(1,1391);
INSERT INTO playlist_track VALUES(1,1388);
INSERT INTO playlist_track VALUES(1,1394);
INSERT INTO playlist_track VALUES(1,1387);
INSERT INTO playlist_track VALUES(1,1392);
INSERT INTO playlist_track VALUES(1,1389);
INSERT INTO playlist_track VALUES(1,1390);
INSERT INTO playlist_track VALUES(1,1335);
INSERT INTO playlist_track VALUES(1,1336);
INSERT INTO playlist_track VALUES(1,1337);
INSERT INTO playlist_track VALUES(1,1338);
INSERT INTO playlist_track VALUES(1,1339);
INSERT INTO playlist_track VALUES(1,1340);
INSERT INTO playlist_track VALUES(1,1341);
INSERT INTO playlist_track VALUES(1,1342);
INSERT INTO playlist_track VALUES(1,1343);
INSERT INTO playlist_track VALUES(1,1344);
INSERT INTO playlist_track VALUES(1,1345);
INSERT INTO playlist_track VALUES(1,1346);
INSERT INTO playlist_track VALUES(1,1347);
INSERT INTO playlist_track VALUES(1,1348);
INSERT INTO playlist_track VALUES(1,1349);
INSERT INTO playlist_track VALUES(1,1350);
INSERT INTO playlist_track VALUES(1,1351);
INSERT INTO playlist_track VALUES(1,1212);
INSERT INTO playlist_track VALUES(1,1213);
INSERT INTO playlist_track VALUES(1,1214);
INSERT INTO playlist_track VALUES(1,1215);
INSERT INTO playlist_track VALUES(1,1216);
INSERT INTO playlist_track VALUES(1,1217);
INSERT INTO playlist_track VALUES(1,1218);
INSERT INTO playlist_track VALUES(1,1219);
INSERT INTO playlist_track VALUES(1,1220);
INSERT INTO playlist_track VALUES(1,1221);
INSERT INTO playlist_track VALUES(1,1222);
INSERT INTO playlist_track VALUES(1,1223);
INSERT INTO playlist_track VALUES(1,1224);
INSERT INTO playlist_track VALUES(1,1225);
INSERT INTO playlist_track VALUES(1,1226);
INSERT INTO playlist_track VALUES(1,1227);
INSERT INTO playlist_track VALUES(1,1228);
INSERT INTO playlist_track VALUES(1,1229);
INSERT INTO playlist_track VALUES(1,1230);
INSERT INTO playlist_track VALUES(1,1231);
INSERT INTO playlist_track VALUES(1,1232);
INSERT INTO playlist_track VALUES(1,1233);
INSERT INTO playlist_track VALUES(1,1234);
INSERT INTO playlist_track VALUES(1,1352);
INSERT INTO playlist_track VALUES(1,1353);
INSERT INTO playlist_track VALUES(1,1354);
INSERT INTO playlist_track VALUES(1,1355);
INSERT INTO playlist_track VALUES(1,1356);
INSERT INTO playlist_track VALUES(1,1357);
INSERT INTO playlist_track VALUES(1,1358);
INSERT INTO playlist_track VALUES(1,1359);
INSERT INTO playlist_track VALUES(1,1360);
INSERT INTO playlist_track VALUES(1,1361);
INSERT INTO playlist_track VALUES(1,1364);
INSERT INTO playlist_track VALUES(1,1371);
INSERT INTO playlist_track VALUES(1,1372);
INSERT INTO playlist_track VALUES(1,1373);
INSERT INTO playlist_track VALUES(1,1374);
INSERT INTO playlist_track VALUES(1,1375);
INSERT INTO playlist_track VALUES(1,1376);
INSERT INTO playlist_track VALUES(1,1377);
INSERT INTO playlist_track VALUES(1,1378);
INSERT INTO playlist_track VALUES(1,1379);
INSERT INTO playlist_track VALUES(1,1380);
INSERT INTO playlist_track VALUES(1,1381);
INSERT INTO playlist_track VALUES(1,1382);
INSERT INTO playlist_track VALUES(1,1386);
INSERT INTO playlist_track VALUES(1,1383);
INSERT INTO playlist_track VALUES(1,1385);
INSERT INTO playlist_track VALUES(1,1384);
INSERT INTO playlist_track VALUES(1,1546);
INSERT INTO playlist_track VALUES(1,1547);
INSERT INTO playlist_track VALUES(1,1548);
INSERT INTO playlist_track VALUES(1,1549);
INSERT INTO playlist_track VALUES(1,1550);
INSERT INTO playlist_track VALUES(1,1551);
INSERT INTO playlist_track VALUES(1,1552);
INSERT INTO playlist_track VALUES(1,1553);
INSERT INTO playlist_track VALUES(1,1554);
INSERT INTO playlist_track VALUES(1,1555);
INSERT INTO playlist_track VALUES(1,1556);
INSERT INTO playlist_track VALUES(1,1557);
INSERT INTO playlist_track VALUES(1,1558);
INSERT INTO playlist_track VALUES(1,1559);
INSERT INTO playlist_track VALUES(1,1560);
INSERT INTO playlist_track VALUES(1,1561);
INSERT INTO playlist_track VALUES(1,1893);
INSERT INTO playlist_track VALUES(1,1894);
INSERT INTO playlist_track VALUES(1,1895);
INSERT INTO playlist_track VALUES(1,1896);
INSERT INTO playlist_track VALUES(1,1897);
INSERT INTO playlist_track VALUES(1,1898);
INSERT INTO playlist_track VALUES(1,1899);
INSERT INTO playlist_track VALUES(1,1900);
INSERT INTO playlist_track VALUES(1,1901);
INSERT INTO playlist_track VALUES(1,1801);
INSERT INTO playlist_track VALUES(1,1802);
INSERT INTO playlist_track VALUES(1,1803);
INSERT INTO playlist_track VALUES(1,1804);
INSERT INTO playlist_track VALUES(1,1805);
INSERT INTO playlist_track VALUES(1,1806);
INSERT INTO playlist_track VALUES(1,1807);
INSERT INTO playlist_track VALUES(1,1808);
INSERT INTO playlist_track VALUES(1,1809);
INSERT INTO playlist_track VALUES(1,1810);
INSERT INTO playlist_track VALUES(1,1811);
INSERT INTO playlist_track VALUES(1,1812);
INSERT INTO playlist_track VALUES(1,408);
INSERT INTO playlist_track VALUES(1,409);
INSERT INTO playlist_track VALUES(1,410);
INSERT INTO playlist_track VALUES(1,411);
INSERT INTO playlist_track VALUES(1,412);
INSERT INTO playlist_track VALUES(1,413);
INSERT INTO playlist_track VALUES(1,414);
INSERT INTO playlist_track VALUES(1,415);
INSERT INTO playlist_track VALUES(1,416);
INSERT INTO playlist_track VALUES(1,417);
INSERT INTO playlist_track VALUES(1,418);
INSERT INTO playlist_track VALUES(1,1813);
INSERT INTO playlist_track VALUES(1,1814);
INSERT INTO playlist_track VALUES(1,1815);
INSERT INTO playlist_track VALUES(1,1816);
INSERT INTO playlist_track VALUES(1,1817);
INSERT INTO playlist_track VALUES(1,1818);
INSERT INTO playlist_track VALUES(1,1819);
INSERT INTO playlist_track VALUES(1,1820);
INSERT INTO playlist_track VALUES(1,1821);
INSERT INTO playlist_track VALUES(1,1822);
INSERT INTO playlist_track VALUES(1,1823);
INSERT INTO playlist_track VALUES(1,1824);
INSERT INTO playlist_track VALUES(1,1825);
INSERT INTO playlist_track VALUES(1,1826);
INSERT INTO playlist_track VALUES(1,1827);
INSERT INTO playlist_track VALUES(1,1828);
INSERT INTO playlist_track VALUES(1,1829);
INSERT INTO playlist_track VALUES(1,1830);
INSERT INTO playlist_track VALUES(1,1831);
INSERT INTO playlist_track VALUES(1,1832);
INSERT INTO playlist_track VALUES(1,1833);
INSERT INTO playlist_track VALUES(1,1834);
INSERT INTO playlist_track VALUES(1,1835);
INSERT INTO playlist_track VALUES(1,1836);
INSERT INTO playlist_track VALUES(1,1837);
INSERT INTO playlist_track VALUES(1,1838);
INSERT INTO playlist_track VALUES(1,1839);
INSERT INTO playlist_track VALUES(1,1840);
INSERT INTO playlist_track VALUES(1,1841);
INSERT INTO playlist_track VALUES(1,1842);
INSERT INTO playlist_track VALUES(1,1843);
INSERT INTO playlist_track VALUES(1,1844);
INSERT INTO playlist_track VALUES(1,1845);
INSERT INTO playlist_track VALUES(1,1846);
INSERT INTO playlist_track VALUES(1,1847);
INSERT INTO playlist_track VALUES(1,1848);
INSERT INTO playlist_track VALUES(1,1849);
INSERT INTO playlist_track VALUES(1,1850);
INSERT INTO playlist_track VALUES(1,1851);
INSERT INTO playlist_track VALUES(1,1852);
INSERT INTO playlist_track VALUES(1,1853);
INSERT INTO playlist_track VALUES(1,1854);
INSERT INTO playlist_track VALUES(1,1855);
INSERT INTO playlist_track VALUES(1,1856);
INSERT INTO playlist_track VALUES(1,1857);
INSERT INTO playlist_track VALUES(1,1858);
INSERT INTO playlist_track VALUES(1,1859);
INSERT INTO playlist_track VALUES(1,1860);
INSERT INTO playlist_track VALUES(1,1861);
INSERT INTO playlist_track VALUES(1,1862);
INSERT INTO playlist_track VALUES(1,1863);
INSERT INTO playlist_track VALUES(1,1864);
INSERT INTO playlist_track VALUES(1,1865);
INSERT INTO playlist_track VALUES(1,1866);
INSERT INTO playlist_track VALUES(1,1867);
INSERT INTO playlist_track VALUES(1,1868);
INSERT INTO playlist_track VALUES(1,1869);
INSERT INTO playlist_track VALUES(1,1870);
INSERT INTO playlist_track VALUES(1,1871);
INSERT INTO playlist_track VALUES(1,1872);
INSERT INTO playlist_track VALUES(1,1873);
INSERT INTO playlist_track VALUES(1,1874);
INSERT INTO playlist_track VALUES(1,1875);
INSERT INTO playlist_track VALUES(1,1876);
INSERT INTO playlist_track VALUES(1,1877);
INSERT INTO playlist_track VALUES(1,1878);
INSERT INTO playlist_track VALUES(1,1879);
INSERT INTO playlist_track VALUES(1,1880);
INSERT INTO playlist_track VALUES(1,1881);
INSERT INTO playlist_track VALUES(1,1882);
INSERT INTO playlist_track VALUES(1,1883);
INSERT INTO playlist_track VALUES(1,1884);
INSERT INTO playlist_track VALUES(1,1885);
INSERT INTO playlist_track VALUES(1,1886);
INSERT INTO playlist_track VALUES(1,1887);
INSERT INTO playlist_track VALUES(1,1888);
INSERT INTO playlist_track VALUES(1,1889);
INSERT INTO playlist_track VALUES(1,1890);
INSERT INTO playlist_track VALUES(1,1891);
INSERT INTO playlist_track VALUES(1,1892);
INSERT INTO playlist_track VALUES(1,1969);
INSERT INTO playlist_track VALUES(1,1970);
INSERT INTO playlist_track VALUES(1,1971);
INSERT INTO playlist_track VALUES(1,1972);
INSERT INTO playlist_track VALUES(1,1973);
INSERT INTO playlist_track VALUES(1,1974);
INSERT INTO playlist_track VALUES(1,1975);
INSERT INTO playlist_track VALUES(1,1976);
INSERT INTO playlist_track VALUES(1,1977);
INSERT INTO playlist_track VALUES(1,1978);
INSERT INTO playlist_track VALUES(1,1979);
INSERT INTO playlist_track VALUES(1,1980);
INSERT INTO playlist_track VALUES(1,1981);
INSERT INTO playlist_track VALUES(1,1982);
INSERT INTO playlist_track VALUES(1,1983);
INSERT INTO playlist_track VALUES(1,1984);
INSERT INTO playlist_track VALUES(1,1985);
INSERT INTO playlist_track VALUES(1,1942);
INSERT INTO playlist_track VALUES(1,1943);
INSERT INTO playlist_track VALUES(1,1944);
INSERT INTO playlist_track VALUES(1,1945);
INSERT INTO playlist_track VALUES(1,1946);
INSERT INTO playlist_track VALUES(1,1947);
INSERT INTO playlist_track VALUES(1,1948);
INSERT INTO playlist_track VALUES(1,1949);
INSERT INTO playlist_track VALUES(1,1950);
INSERT INTO playlist_track VALUES(1,1951);
INSERT INTO playlist_track VALUES(1,1952);
INSERT INTO playlist_track VALUES(1,1953);
INSERT INTO playlist_track VALUES(1,1954);
INSERT INTO playlist_track VALUES(1,1955);
INSERT INTO playlist_track VALUES(1,1956);
INSERT INTO playlist_track VALUES(1,2099);
INSERT INTO playlist_track VALUES(1,2100);
INSERT INTO playlist_track VALUES(1,2101);
INSERT INTO playlist_track VALUES(1,2102);
INSERT INTO playlist_track VALUES(1,2103);
INSERT INTO playlist_track VALUES(1,2104);
INSERT INTO playlist_track VALUES(1,2105);
INSERT INTO playlist_track VALUES(1,2106);
INSERT INTO playlist_track VALUES(1,2107);
INSERT INTO playlist_track VALUES(1,2108);
INSERT INTO playlist_track VALUES(1,2109);
INSERT INTO playlist_track VALUES(1,2110);
INSERT INTO playlist_track VALUES(1,2111);
INSERT INTO playlist_track VALUES(1,2112);
INSERT INTO playlist_track VALUES(1,2554);
INSERT INTO playlist_track VALUES(1,2555);
INSERT INTO playlist_track VALUES(1,2556);
INSERT INTO playlist_track VALUES(1,2557);
INSERT INTO playlist_track VALUES(1,2558);
INSERT INTO playlist_track VALUES(1,2559);
INSERT INTO playlist_track VALUES(1,2560);
INSERT INTO playlist_track VALUES(1,2561);
INSERT INTO playlist_track VALUES(1,2562);
INSERT INTO playlist_track VALUES(1,2563);
INSERT INTO playlist_track VALUES(1,2564);
INSERT INTO playlist_track VALUES(1,3132);
INSERT INTO playlist_track VALUES(1,3133);
INSERT INTO playlist_track VALUES(1,3134);
INSERT INTO playlist_track VALUES(1,3135);
INSERT INTO playlist_track VALUES(1,3136);
INSERT INTO playlist_track VALUES(1,3137);
INSERT INTO playlist_track VALUES(1,3138);
INSERT INTO playlist_track VALUES(1,3139);
INSERT INTO playlist_track VALUES(1,3140);
INSERT INTO playlist_track VALUES(1,3141);
INSERT INTO playlist_track VALUES(1,3142);
INSERT INTO playlist_track VALUES(1,3143);
INSERT INTO playlist_track VALUES(1,3144);
INSERT INTO playlist_track VALUES(1,3145);
INSERT INTO playlist_track VALUES(1,3451);
INSERT INTO playlist_track VALUES(1,3256);
INSERT INTO playlist_track VALUES(1,3467);
INSERT INTO playlist_track VALUES(1,3468);
INSERT INTO playlist_track VALUES(1,3469);
INSERT INTO playlist_track VALUES(1,3470);
INSERT INTO playlist_track VALUES(1,3471);
INSERT INTO playlist_track VALUES(1,3472);
INSERT INTO playlist_track VALUES(1,3473);
INSERT INTO playlist_track VALUES(1,3474);
INSERT INTO playlist_track VALUES(1,3475);
INSERT INTO playlist_track VALUES(1,3476);
INSERT INTO playlist_track VALUES(1,3477);
INSERT INTO playlist_track VALUES(1,3262);
INSERT INTO playlist_track VALUES(1,3268);
INSERT INTO playlist_track VALUES(1,3263);
INSERT INTO playlist_track VALUES(1,3266);
INSERT INTO playlist_track VALUES(1,3255);
INSERT INTO playlist_track VALUES(1,3259);
INSERT INTO playlist_track VALUES(1,3260);
INSERT INTO playlist_track VALUES(1,3273);
INSERT INTO playlist_track VALUES(1,3265);
INSERT INTO playlist_track VALUES(1,3274);
INSERT INTO playlist_track VALUES(1,3267);
INSERT INTO playlist_track VALUES(1,3261);
INSERT INTO playlist_track VALUES(1,3272);
INSERT INTO playlist_track VALUES(1,3257);
INSERT INTO playlist_track VALUES(1,3258);
INSERT INTO playlist_track VALUES(1,3270);
INSERT INTO playlist_track VALUES(1,3271);
INSERT INTO playlist_track VALUES(1,3254);
INSERT INTO playlist_track VALUES(1,3275);
INSERT INTO playlist_track VALUES(1,3269);
INSERT INTO playlist_track VALUES(1,3253);
INSERT INTO playlist_track VALUES(1,323);
INSERT INTO playlist_track VALUES(1,324);
INSERT INTO playlist_track VALUES(1,325);
INSERT INTO playlist_track VALUES(1,326);
INSERT INTO playlist_track VALUES(1,327);
INSERT INTO playlist_track VALUES(1,328);
INSERT INTO playlist_track VALUES(1,329);
INSERT INTO playlist_track VALUES(1,330);
INSERT INTO playlist_track VALUES(1,331);
INSERT INTO playlist_track VALUES(1,332);
INSERT INTO playlist_track VALUES(1,333);
INSERT INTO playlist_track VALUES(1,334);
INSERT INTO playlist_track VALUES(1,335);
INSERT INTO playlist_track VALUES(1,336);
INSERT INTO playlist_track VALUES(1,3264);
INSERT INTO playlist_track VALUES(1,3455);
INSERT INTO playlist_track VALUES(1,3456);
INSERT INTO playlist_track VALUES(1,3457);
INSERT INTO playlist_track VALUES(1,3458);
INSERT INTO playlist_track VALUES(1,3459);
INSERT INTO playlist_track VALUES(1,3460);
INSERT INTO playlist_track VALUES(1,3461);
INSERT INTO playlist_track VALUES(1,3462);
INSERT INTO playlist_track VALUES(1,3463);
INSERT INTO playlist_track VALUES(1,3464);
INSERT INTO playlist_track VALUES(1,3465);
INSERT INTO playlist_track VALUES(1,3466);
INSERT INTO playlist_track VALUES(1,1414);
INSERT INTO playlist_track VALUES(1,1415);
INSERT INTO playlist_track VALUES(1,1416);
INSERT INTO playlist_track VALUES(1,1417);
INSERT INTO playlist_track VALUES(1,1418);
INSERT INTO playlist_track VALUES(1,1419);
INSERT INTO playlist_track VALUES(1,1420);
INSERT INTO playlist_track VALUES(1,1421);
INSERT INTO playlist_track VALUES(1,1422);
INSERT INTO playlist_track VALUES(1,1423);
INSERT INTO playlist_track VALUES(1,1424);
INSERT INTO playlist_track VALUES(1,1425);
INSERT INTO playlist_track VALUES(1,1426);
INSERT INTO playlist_track VALUES(1,1427);
INSERT INTO playlist_track VALUES(1,1428);
INSERT INTO playlist_track VALUES(1,1429);
INSERT INTO playlist_track VALUES(1,1430);
INSERT INTO playlist_track VALUES(1,1431);
INSERT INTO playlist_track VALUES(1,1432);
INSERT INTO playlist_track VALUES(1,1433);
INSERT INTO playlist_track VALUES(1,1444);
INSERT INTO playlist_track VALUES(1,1445);
INSERT INTO playlist_track VALUES(1,1446);
INSERT INTO playlist_track VALUES(1,1447);
INSERT INTO playlist_track VALUES(1,1448);
INSERT INTO playlist_track VALUES(1,1449);
INSERT INTO playlist_track VALUES(1,1450);
INSERT INTO playlist_track VALUES(1,1451);
INSERT INTO playlist_track VALUES(1,1452);
INSERT INTO playlist_track VALUES(1,1453);
INSERT INTO playlist_track VALUES(1,1454);
INSERT INTO playlist_track VALUES(1,1773);
INSERT INTO playlist_track VALUES(1,1774);
INSERT INTO playlist_track VALUES(1,1775);
INSERT INTO playlist_track VALUES(1,1776);
INSERT INTO playlist_track VALUES(1,1777);
INSERT INTO playlist_track VALUES(1,1778);
INSERT INTO playlist_track VALUES(1,1779);
INSERT INTO playlist_track VALUES(1,1780);
INSERT INTO playlist_track VALUES(1,1781);
INSERT INTO playlist_track VALUES(1,1782);
INSERT INTO playlist_track VALUES(1,1783);
INSERT INTO playlist_track VALUES(1,1784);
INSERT INTO playlist_track VALUES(1,1785);
INSERT INTO playlist_track VALUES(1,1786);
INSERT INTO playlist_track VALUES(1,1787);
INSERT INTO playlist_track VALUES(1,1788);
INSERT INTO playlist_track VALUES(1,1789);
INSERT INTO playlist_track VALUES(1,1790);
INSERT INTO playlist_track VALUES(1,282);
INSERT INTO playlist_track VALUES(1,283);
INSERT INTO playlist_track VALUES(1,284);
INSERT INTO playlist_track VALUES(1,285);
INSERT INTO playlist_track VALUES(1,286);
INSERT INTO playlist_track VALUES(1,287);
INSERT INTO playlist_track VALUES(1,288);
INSERT INTO playlist_track VALUES(1,289);
INSERT INTO playlist_track VALUES(1,290);
INSERT INTO playlist_track VALUES(1,291);
INSERT INTO playlist_track VALUES(1,292);
INSERT INTO playlist_track VALUES(1,293);
INSERT INTO playlist_track VALUES(1,294);
INSERT INTO playlist_track VALUES(1,295);
INSERT INTO playlist_track VALUES(1,296);
INSERT INTO playlist_track VALUES(1,297);
INSERT INTO playlist_track VALUES(1,298);
INSERT INTO playlist_track VALUES(1,299);
INSERT INTO playlist_track VALUES(1,300);
INSERT INTO playlist_track VALUES(1,301);
INSERT INTO playlist_track VALUES(1,302);
INSERT INTO playlist_track VALUES(1,303);
INSERT INTO playlist_track VALUES(1,304);
INSERT INTO playlist_track VALUES(1,305);
INSERT INTO playlist_track VALUES(1,306);
INSERT INTO playlist_track VALUES(1,307);
INSERT INTO playlist_track VALUES(1,308);
INSERT INTO playlist_track VALUES(1,309);
INSERT INTO playlist_track VALUES(1,310);
INSERT INTO playlist_track VALUES(1,311);
INSERT INTO playlist_track VALUES(1,312);
INSERT INTO playlist_track VALUES(1,2216);
INSERT INTO playlist_track VALUES(1,2217);
INSERT INTO playlist_track VALUES(1,2218);
INSERT INTO playlist_track VALUES(1,2219);
INSERT INTO playlist_track VALUES(1,2220);
INSERT INTO playlist_track VALUES(1,2221);
INSERT INTO playlist_track VALUES(1,2222);
INSERT INTO playlist_track VALUES(1,2223);
INSERT INTO playlist_track VALUES(1,2224);
INSERT INTO playlist_track VALUES(1,2225);
INSERT INTO playlist_track VALUES(1,2226);
INSERT INTO playlist_track VALUES(1,2227);
INSERT INTO playlist_track VALUES(1,2228);
INSERT INTO playlist_track VALUES(1,3038);
INSERT INTO playlist_track VALUES(1,3039);
INSERT INTO playlist_track VALUES(1,3040);
INSERT INTO playlist_track VALUES(1,3041);
INSERT INTO playlist_track VALUES(1,3042);
INSERT INTO playlist_track VALUES(1,3043);
INSERT INTO playlist_track VALUES(1,3044);
INSERT INTO playlist_track VALUES(1,3045);
INSERT INTO playlist_track VALUES(1,3046);
INSERT INTO playlist_track VALUES(1,3047);
INSERT INTO playlist_track VALUES(1,3048);
INSERT INTO playlist_track VALUES(1,3049);
INSERT INTO playlist_track VALUES(1,3050);
INSERT INTO playlist_track VALUES(1,3051);
INSERT INTO playlist_track VALUES(1,1);
INSERT INTO playlist_track VALUES(1,6);
INSERT INTO playlist_track VALUES(1,7);
INSERT INTO playlist_track VALUES(1,8);
INSERT INTO playlist_track VALUES(1,9);
INSERT INTO playlist_track VALUES(1,10);
INSERT INTO playlist_track VALUES(1,11);
INSERT INTO playlist_track VALUES(1,12);
INSERT INTO playlist_track VALUES(1,13);
INSERT INTO playlist_track VALUES(1,14);
INSERT INTO playlist_track VALUES(1,15);
INSERT INTO playlist_track VALUES(1,16);
INSERT INTO playlist_track VALUES(1,17);
INSERT INTO playlist_track VALUES(1,18);
INSERT INTO playlist_track VALUES(1,19);
INSERT INTO playlist_track VALUES(1,20);
INSERT INTO playlist_track VALUES(1,21);
INSERT INTO playlist_track VALUES(1,22);
INSERT INTO playlist_track VALUES(1,2);
INSERT INTO playlist_track VALUES(1,3);
INSERT INTO playlist_track VALUES(1,4);
INSERT INTO playlist_track VALUES(1,5);
INSERT INTO playlist_track VALUES(1,23);
INSERT INTO playlist_track VALUES(1,24);
INSERT INTO playlist_track VALUES(1,25);
INSERT INTO playlist_track VALUES(1,26);
INSERT INTO playlist_track VALUES(1,27);
INSERT INTO playlist_track VALUES(1,28);
INSERT INTO playlist_track VALUES(1,29);
INSERT INTO playlist_track VALUES(1,30);
INSERT INTO playlist_track VALUES(1,31);
INSERT INTO playlist_track VALUES(1,32);
INSERT INTO playlist_track VALUES(1,33);
INSERT INTO playlist_track VALUES(1,34);
INSERT INTO playlist_track VALUES(1,35);
INSERT INTO playlist_track VALUES(1,36);
INSERT INTO playlist_track VALUES(1,37);
INSERT INTO playlist_track VALUES(1,38);
INSERT INTO playlist_track VALUES(1,39);
INSERT INTO playlist_track VALUES(1,40);
INSERT INTO playlist_track VALUES(1,41);
INSERT INTO playlist_track VALUES(1,42);
INSERT INTO playlist_track VALUES(1,43);
INSERT INTO playlist_track VALUES(1,44);
INSERT INTO playlist_track VALUES(1,45);
INSERT INTO playlist_track VALUES(1,46);
INSERT INTO playlist_track VALUES(1,47);
INSERT INTO playlist_track VALUES(1,48);
INSERT INTO playlist_track VALUES(1,49);
INSERT INTO playlist_track VALUES(1,50);
INSERT INTO playlist_track VALUES(1,51);
INSERT INTO playlist_track VALUES(1,52);
INSERT INTO playlist_track VALUES(1,53);
INSERT INTO playlist_track VALUES(1,54);
INSERT INTO playlist_track VALUES(1,55);
INSERT INTO playlist_track VALUES(1,56);
INSERT INTO playlist_track VALUES(1,57);
INSERT INTO playlist_track VALUES(1,58);
INSERT INTO playlist_track VALUES(1,59);
INSERT INTO playlist_track VALUES(1,60);
INSERT INTO playlist_track VALUES(1,61);
INSERT INTO playlist_track VALUES(1,62);
INSERT INTO playlist_track VALUES(1,85);
INSERT INTO playlist_track VALUES(1,86);
INSERT INTO playlist_track VALUES(1,87);
INSERT INTO playlist_track VALUES(1,88);
INSERT INTO playlist_track VALUES(1,89);
INSERT INTO playlist_track VALUES(1,90);
INSERT INTO playlist_track VALUES(1,91);
INSERT INTO playlist_track VALUES(1,92);
INSERT INTO playlist_track VALUES(1,93);
INSERT INTO playlist_track VALUES(1,94);
INSERT INTO playlist_track VALUES(1,95);
INSERT INTO playlist_track VALUES(1,96);
INSERT INTO playlist_track VALUES(1,97);
INSERT INTO playlist_track VALUES(1,98);
INSERT INTO playlist_track VALUES(1,675);
INSERT INTO playlist_track VALUES(1,676);
INSERT INTO playlist_track VALUES(1,677);
INSERT INTO playlist_track VALUES(1,678);
INSERT INTO playlist_track VALUES(1,679);
INSERT INTO playlist_track VALUES(1,680);
INSERT INTO playlist_track VALUES(1,681);
INSERT INTO playlist_track VALUES(1,682);
INSERT INTO playlist_track VALUES(1,683);
INSERT INTO playlist_track VALUES(1,684);
INSERT INTO playlist_track VALUES(1,685);
INSERT INTO playlist_track VALUES(1,686);
INSERT INTO playlist_track VALUES(1,687);
INSERT INTO playlist_track VALUES(1,688);
INSERT INTO playlist_track VALUES(1,689);
INSERT INTO playlist_track VALUES(1,690);
INSERT INTO playlist_track VALUES(1,691);
INSERT INTO playlist_track VALUES(1,692);
INSERT INTO playlist_track VALUES(1,693);
INSERT INTO playlist_track VALUES(1,694);
INSERT INTO playlist_track VALUES(1,695);
INSERT INTO playlist_track VALUES(1,696);
INSERT INTO playlist_track VALUES(1,697);
INSERT INTO playlist_track VALUES(1,698);
INSERT INTO playlist_track VALUES(1,699);
INSERT INTO playlist_track VALUES(1,700);
INSERT INTO playlist_track VALUES(1,701);
INSERT INTO playlist_track VALUES(1,702);
INSERT INTO playlist_track VALUES(1,703);
INSERT INTO playlist_track VALUES(1,704);
INSERT INTO playlist_track VALUES(1,705);
INSERT INTO playlist_track VALUES(1,706);
INSERT INTO playlist_track VALUES(1,707);
INSERT INTO playlist_track VALUES(1,708);
INSERT INTO playlist_track VALUES(1,709);
INSERT INTO playlist_track VALUES(1,710);
INSERT INTO playlist_track VALUES(1,711);
INSERT INTO playlist_track VALUES(1,712);
INSERT INTO playlist_track VALUES(1,713);
INSERT INTO playlist_track VALUES(1,714);
INSERT INTO playlist_track VALUES(1,2609);
INSERT INTO playlist_track VALUES(1,2610);
INSERT INTO playlist_track VALUES(1,2611);
INSERT INTO playlist_track VALUES(1,2612);
INSERT INTO playlist_track VALUES(1,2613);
INSERT INTO playlist_track VALUES(1,2614);
INSERT INTO playlist_track VALUES(1,2615);
INSERT INTO playlist_track VALUES(1,2616);
INSERT INTO playlist_track VALUES(1,2617);
INSERT INTO playlist_track VALUES(1,2618);
INSERT INTO playlist_track VALUES(1,2619);
INSERT INTO playlist_track VALUES(1,2620);
INSERT INTO playlist_track VALUES(1,2621);
INSERT INTO playlist_track VALUES(1,2622);
INSERT INTO playlist_track VALUES(1,2623);
INSERT INTO playlist_track VALUES(1,2624);
INSERT INTO playlist_track VALUES(1,2625);
INSERT INTO playlist_track VALUES(1,2626);
INSERT INTO playlist_track VALUES(1,2627);
INSERT INTO playlist_track VALUES(1,2628);
INSERT INTO playlist_track VALUES(1,2629);
INSERT INTO playlist_track VALUES(1,2630);
INSERT INTO playlist_track VALUES(1,2631);
INSERT INTO playlist_track VALUES(1,2632);
INSERT INTO playlist_track VALUES(1,2633);
INSERT INTO playlist_track VALUES(1,2634);
INSERT INTO playlist_track VALUES(1,2635);
INSERT INTO playlist_track VALUES(1,2636);
INSERT INTO playlist_track VALUES(1,2637);
INSERT INTO playlist_track VALUES(1,2638);
INSERT INTO playlist_track VALUES(1,489);
INSERT INTO playlist_track VALUES(1,490);
INSERT INTO playlist_track VALUES(1,491);
INSERT INTO playlist_track VALUES(1,492);
INSERT INTO playlist_track VALUES(1,493);
INSERT INTO playlist_track VALUES(1,494);
INSERT INTO playlist_track VALUES(1,495);
INSERT INTO playlist_track VALUES(1,496);
INSERT INTO playlist_track VALUES(1,497);
INSERT INTO playlist_track VALUES(1,498);
INSERT INTO playlist_track VALUES(1,499);
INSERT INTO playlist_track VALUES(1,500);
INSERT INTO playlist_track VALUES(1,816);
INSERT INTO playlist_track VALUES(1,817);
INSERT INTO playlist_track VALUES(1,818);
INSERT INTO playlist_track VALUES(1,819);
INSERT INTO playlist_track VALUES(1,820);
INSERT INTO playlist_track VALUES(1,821);
INSERT INTO playlist_track VALUES(1,822);
INSERT INTO playlist_track VALUES(1,823);
INSERT INTO playlist_track VALUES(1,824);
INSERT INTO playlist_track VALUES(1,825);
INSERT INTO playlist_track VALUES(1,745);
INSERT INTO playlist_track VALUES(1,746);
INSERT INTO playlist_track VALUES(1,747);
INSERT INTO playlist_track VALUES(1,748);
INSERT INTO playlist_track VALUES(1,749);
INSERT INTO playlist_track VALUES(1,750);
INSERT INTO playlist_track VALUES(1,751);
INSERT INTO playlist_track VALUES(1,752);
INSERT INTO playlist_track VALUES(1,753);
INSERT INTO playlist_track VALUES(1,754);
INSERT INTO playlist_track VALUES(1,755);
INSERT INTO playlist_track VALUES(1,756);
INSERT INTO playlist_track VALUES(1,757);
INSERT INTO playlist_track VALUES(1,758);
INSERT INTO playlist_track VALUES(1,759);
INSERT INTO playlist_track VALUES(1,760);
INSERT INTO playlist_track VALUES(1,620);
INSERT INTO playlist_track VALUES(1,621);
INSERT INTO playlist_track VALUES(1,622);
INSERT INTO playlist_track VALUES(1,623);
INSERT INTO playlist_track VALUES(1,761);
INSERT INTO playlist_track VALUES(1,762);
INSERT INTO playlist_track VALUES(1,763);
INSERT INTO playlist_track VALUES(1,764);
INSERT INTO playlist_track VALUES(1,765);
INSERT INTO playlist_track VALUES(1,766);
INSERT INTO playlist_track VALUES(1,767);
INSERT INTO playlist_track VALUES(1,768);
INSERT INTO playlist_track VALUES(1,769);
INSERT INTO playlist_track VALUES(1,770);
INSERT INTO playlist_track VALUES(1,771);
INSERT INTO playlist_track VALUES(1,772);
INSERT INTO playlist_track VALUES(1,773);
INSERT INTO playlist_track VALUES(1,774);
INSERT INTO playlist_track VALUES(1,775);
INSERT INTO playlist_track VALUES(1,776);
INSERT INTO playlist_track VALUES(1,777);
INSERT INTO playlist_track VALUES(1,778);
INSERT INTO playlist_track VALUES(1,779);
INSERT INTO playlist_track VALUES(1,780);
INSERT INTO playlist_track VALUES(1,781);
INSERT INTO playlist_track VALUES(1,782);
INSERT INTO playlist_track VALUES(1,783);
INSERT INTO playlist_track VALUES(1,784);
INSERT INTO playlist_track VALUES(1,785);
INSERT INTO playlist_track VALUES(1,543);
INSERT INTO playlist_track VALUES(1,544);
INSERT INTO playlist_track VALUES(1,545);
INSERT INTO playlist_track VALUES(1,546);
INSERT INTO playlist_track VALUES(1,547);
INSERT INTO playlist_track VALUES(1,548);
INSERT INTO playlist_track VALUES(1,549);
INSERT INTO playlist_track VALUES(1,786);
INSERT INTO playlist_track VALUES(1,787);
INSERT INTO playlist_track VALUES(1,788);
INSERT INTO playlist_track VALUES(1,789);
INSERT INTO playlist_track VALUES(1,790);
INSERT INTO playlist_track VALUES(1,791);
INSERT INTO playlist_track VALUES(1,792);
INSERT INTO playlist_track VALUES(1,793);
INSERT INTO playlist_track VALUES(1,794);
INSERT INTO playlist_track VALUES(1,795);
INSERT INTO playlist_track VALUES(1,796);
INSERT INTO playlist_track VALUES(1,797);
INSERT INTO playlist_track VALUES(1,798);
INSERT INTO playlist_track VALUES(1,799);
INSERT INTO playlist_track VALUES(1,800);
INSERT INTO playlist_track VALUES(1,801);
INSERT INTO playlist_track VALUES(1,802);
INSERT INTO playlist_track VALUES(1,803);
INSERT INTO playlist_track VALUES(1,804);
INSERT INTO playlist_track VALUES(1,805);
INSERT INTO playlist_track VALUES(1,806);
INSERT INTO playlist_track VALUES(1,807);
INSERT INTO playlist_track VALUES(1,808);
INSERT INTO playlist_track VALUES(1,809);
INSERT INTO playlist_track VALUES(1,810);
INSERT INTO playlist_track VALUES(1,811);
INSERT INTO playlist_track VALUES(1,812);
INSERT INTO playlist_track VALUES(1,813);
INSERT INTO playlist_track VALUES(1,814);
INSERT INTO playlist_track VALUES(1,815);
INSERT INTO playlist_track VALUES(1,826);
INSERT INTO playlist_track VALUES(1,827);
INSERT INTO playlist_track VALUES(1,828);
INSERT INTO playlist_track VALUES(1,829);
INSERT INTO playlist_track VALUES(1,830);
INSERT INTO playlist_track VALUES(1,831);
INSERT INTO playlist_track VALUES(1,832);
INSERT INTO playlist_track VALUES(1,833);
INSERT INTO playlist_track VALUES(1,834);
INSERT INTO playlist_track VALUES(1,835);
INSERT INTO playlist_track VALUES(1,836);
INSERT INTO playlist_track VALUES(1,837);
INSERT INTO playlist_track VALUES(1,838);
INSERT INTO playlist_track VALUES(1,839);
INSERT INTO playlist_track VALUES(1,840);
INSERT INTO playlist_track VALUES(1,841);
INSERT INTO playlist_track VALUES(1,2639);
INSERT INTO playlist_track VALUES(1,2640);
INSERT INTO playlist_track VALUES(1,2641);
INSERT INTO playlist_track VALUES(1,2642);
INSERT INTO playlist_track VALUES(1,2643);
INSERT INTO playlist_track VALUES(1,2644);
INSERT INTO playlist_track VALUES(1,2645);
INSERT INTO playlist_track VALUES(1,2646);
INSERT INTO playlist_track VALUES(1,2647);
INSERT INTO playlist_track VALUES(1,2648);
INSERT INTO playlist_track VALUES(1,2649);
INSERT INTO playlist_track VALUES(1,3225);
INSERT INTO playlist_track VALUES(1,949);
INSERT INTO playlist_track VALUES(1,950);
INSERT INTO playlist_track VALUES(1,951);
INSERT INTO playlist_track VALUES(1,952);
INSERT INTO playlist_track VALUES(1,953);
INSERT INTO playlist_track VALUES(1,954);
INSERT INTO playlist_track VALUES(1,955);
INSERT INTO playlist_track VALUES(1,956);
INSERT INTO playlist_track VALUES(1,957);
INSERT INTO playlist_track VALUES(1,958);
INSERT INTO playlist_track VALUES(1,959);
INSERT INTO playlist_track VALUES(1,960);
INSERT INTO playlist_track VALUES(1,961);
INSERT INTO playlist_track VALUES(1,962);
INSERT INTO playlist_track VALUES(1,963);
INSERT INTO playlist_track VALUES(1,1020);
INSERT INTO playlist_track VALUES(1,1021);
INSERT INTO playlist_track VALUES(1,1022);
INSERT INTO playlist_track VALUES(1,1023);
INSERT INTO playlist_track VALUES(1,1024);
INSERT INTO playlist_track VALUES(1,1025);
INSERT INTO playlist_track VALUES(1,1026);
INSERT INTO playlist_track VALUES(1,1027);
INSERT INTO playlist_track VALUES(1,1028);
INSERT INTO playlist_track VALUES(1,1029);
INSERT INTO playlist_track VALUES(1,1030);
INSERT INTO playlist_track VALUES(1,1031);
INSERT INTO playlist_track VALUES(1,1032);
INSERT INTO playlist_track VALUES(1,989);
INSERT INTO playlist_track VALUES(1,990);
INSERT INTO playlist_track VALUES(1,991);
INSERT INTO playlist_track VALUES(1,992);
INSERT INTO playlist_track VALUES(1,993);
INSERT INTO playlist_track VALUES(1,994);
INSERT INTO playlist_track VALUES(1,995);
INSERT INTO playlist_track VALUES(1,996);
INSERT INTO playlist_track VALUES(1,997);
INSERT INTO playlist_track VALUES(1,998);
INSERT INTO playlist_track VALUES(1,999);
INSERT INTO playlist_track VALUES(1,1000);
INSERT INTO playlist_track VALUES(1,1001);
INSERT INTO playlist_track VALUES(1,1002);
INSERT INTO playlist_track VALUES(1,1003);
INSERT INTO playlist_track VALUES(1,1004);
INSERT INTO playlist_track VALUES(1,1005);
INSERT INTO playlist_track VALUES(1,1006);
INSERT INTO playlist_track VALUES(1,1007);
INSERT INTO playlist_track VALUES(1,1008);
INSERT INTO playlist_track VALUES(1,351);
INSERT INTO playlist_track VALUES(1,352);
INSERT INTO playlist_track VALUES(1,353);
INSERT INTO playlist_track VALUES(1,354);
INSERT INTO playlist_track VALUES(1,355);
INSERT INTO playlist_track VALUES(1,356);
INSERT INTO playlist_track VALUES(1,357);
INSERT INTO playlist_track VALUES(1,358);
INSERT INTO playlist_track VALUES(1,359);
INSERT INTO playlist_track VALUES(1,1146);
INSERT INTO playlist_track VALUES(1,1147);
INSERT INTO playlist_track VALUES(1,1148);
INSERT INTO playlist_track VALUES(1,1149);
INSERT INTO playlist_track VALUES(1,1150);
INSERT INTO playlist_track VALUES(1,1151);
INSERT INTO playlist_track VALUES(1,1152);
INSERT INTO playlist_track VALUES(1,1153);
INSERT INTO playlist_track VALUES(1,1154);
INSERT INTO playlist_track VALUES(1,1155);
INSERT INTO playlist_track VALUES(1,1156);
INSERT INTO playlist_track VALUES(1,1157);
INSERT INTO playlist_track VALUES(1,1158);
INSERT INTO playlist_track VALUES(1,1159);
INSERT INTO playlist_track VALUES(1,1160);
INSERT INTO playlist_track VALUES(1,1161);
INSERT INTO playlist_track VALUES(1,1162);
INSERT INTO playlist_track VALUES(1,1163);
INSERT INTO playlist_track VALUES(1,1164);
INSERT INTO playlist_track VALUES(1,1165);
INSERT INTO playlist_track VALUES(1,1166);
INSERT INTO playlist_track VALUES(1,1167);
INSERT INTO playlist_track VALUES(1,1168);
INSERT INTO playlist_track VALUES(1,1169);
INSERT INTO playlist_track VALUES(1,1170);
INSERT INTO playlist_track VALUES(1,1171);
INSERT INTO playlist_track VALUES(1,1172);
INSERT INTO playlist_track VALUES(1,1173);
INSERT INTO playlist_track VALUES(1,1235);
INSERT INTO playlist_track VALUES(1,1236);
INSERT INTO playlist_track VALUES(1,1237);
INSERT INTO playlist_track VALUES(1,1238);
INSERT INTO playlist_track VALUES(1,1239);
INSERT INTO playlist_track VALUES(1,1240);
INSERT INTO playlist_track VALUES(1,1241);
INSERT INTO playlist_track VALUES(1,1242);
INSERT INTO playlist_track VALUES(1,1243);
INSERT INTO playlist_track VALUES(1,1244);
INSERT INTO playlist_track VALUES(1,1256);
INSERT INTO playlist_track VALUES(1,1257);
INSERT INTO playlist_track VALUES(1,1258);
INSERT INTO playlist_track VALUES(1,1259);
INSERT INTO playlist_track VALUES(1,1260);
INSERT INTO playlist_track VALUES(1,1261);
INSERT INTO playlist_track VALUES(1,1262);
INSERT INTO playlist_track VALUES(1,1263);
INSERT INTO playlist_track VALUES(1,1264);
INSERT INTO playlist_track VALUES(1,1265);
INSERT INTO playlist_track VALUES(1,1266);
INSERT INTO playlist_track VALUES(1,1267);
INSERT INTO playlist_track VALUES(1,1305);
INSERT INTO playlist_track VALUES(1,1306);
INSERT INTO playlist_track VALUES(1,1307);
INSERT INTO playlist_track VALUES(1,1308);
INSERT INTO playlist_track VALUES(1,1309);
INSERT INTO playlist_track VALUES(1,1310);
INSERT INTO playlist_track VALUES(1,1311);
INSERT INTO playlist_track VALUES(1,1312);
INSERT INTO playlist_track VALUES(1,1313);
INSERT INTO playlist_track VALUES(1,1314);
INSERT INTO playlist_track VALUES(1,1315);
INSERT INTO playlist_track VALUES(1,1316);
INSERT INTO playlist_track VALUES(1,1317);
INSERT INTO playlist_track VALUES(1,1318);
INSERT INTO playlist_track VALUES(1,1319);
INSERT INTO playlist_track VALUES(1,1320);
INSERT INTO playlist_track VALUES(1,1321);
INSERT INTO playlist_track VALUES(1,1322);
INSERT INTO playlist_track VALUES(1,1323);
INSERT INTO playlist_track VALUES(1,1324);
INSERT INTO playlist_track VALUES(1,1201);
INSERT INTO playlist_track VALUES(1,1202);
INSERT INTO playlist_track VALUES(1,1203);
INSERT INTO playlist_track VALUES(1,1204);
INSERT INTO playlist_track VALUES(1,1205);
INSERT INTO playlist_track VALUES(1,1206);
INSERT INTO playlist_track VALUES(1,1207);
INSERT INTO playlist_track VALUES(1,1208);
INSERT INTO playlist_track VALUES(1,1209);
INSERT INTO playlist_track VALUES(1,1210);
INSERT INTO playlist_track VALUES(1,1211);
INSERT INTO playlist_track VALUES(1,1393);
INSERT INTO playlist_track VALUES(1,1362);
INSERT INTO playlist_track VALUES(1,1363);
INSERT INTO playlist_track VALUES(1,1365);
INSERT INTO playlist_track VALUES(1,1366);
INSERT INTO playlist_track VALUES(1,1367);
INSERT INTO playlist_track VALUES(1,1368);
INSERT INTO playlist_track VALUES(1,1369);
INSERT INTO playlist_track VALUES(1,1370);
INSERT INTO playlist_track VALUES(1,1406);
INSERT INTO playlist_track VALUES(1,1407);
INSERT INTO playlist_track VALUES(1,1408);
INSERT INTO playlist_track VALUES(1,1409);
INSERT INTO playlist_track VALUES(1,1410);
INSERT INTO playlist_track VALUES(1,1411);
INSERT INTO playlist_track VALUES(1,1412);
INSERT INTO playlist_track VALUES(1,1413);
INSERT INTO playlist_track VALUES(1,1395);
INSERT INTO playlist_track VALUES(1,1396);
INSERT INTO playlist_track VALUES(1,1397);
INSERT INTO playlist_track VALUES(1,1398);
INSERT INTO playlist_track VALUES(1,1399);
INSERT INTO playlist_track VALUES(1,1400);
INSERT INTO playlist_track VALUES(1,1401);
INSERT INTO playlist_track VALUES(1,1402);
INSERT INTO playlist_track VALUES(1,1403);
INSERT INTO playlist_track VALUES(1,1404);
INSERT INTO playlist_track VALUES(1,1405);
INSERT INTO playlist_track VALUES(1,1434);
INSERT INTO playlist_track VALUES(1,1435);
INSERT INTO playlist_track VALUES(1,1436);
INSERT INTO playlist_track VALUES(1,1437);
INSERT INTO playlist_track VALUES(1,1438);
INSERT INTO playlist_track VALUES(1,1439);
INSERT INTO playlist_track VALUES(1,1440);
INSERT INTO playlist_track VALUES(1,1441);
INSERT INTO playlist_track VALUES(1,1442);
INSERT INTO playlist_track VALUES(1,1443);
INSERT INTO playlist_track VALUES(1,1479);
INSERT INTO playlist_track VALUES(1,1480);
INSERT INTO playlist_track VALUES(1,1481);
INSERT INTO playlist_track VALUES(1,1482);
INSERT INTO playlist_track VALUES(1,1483);
INSERT INTO playlist_track VALUES(1,1484);
INSERT INTO playlist_track VALUES(1,1485);
INSERT INTO playlist_track VALUES(1,1486);
INSERT INTO playlist_track VALUES(1,1487);
INSERT INTO playlist_track VALUES(1,1488);
INSERT INTO playlist_track VALUES(1,1489);
INSERT INTO playlist_track VALUES(1,1490);
INSERT INTO playlist_track VALUES(1,1491);
INSERT INTO playlist_track VALUES(1,1492);
INSERT INTO playlist_track VALUES(1,1493);
INSERT INTO playlist_track VALUES(1,1494);
INSERT INTO playlist_track VALUES(1,1495);
INSERT INTO playlist_track VALUES(1,1496);
INSERT INTO playlist_track VALUES(1,1497);
INSERT INTO playlist_track VALUES(1,1498);
INSERT INTO playlist_track VALUES(1,1499);
INSERT INTO playlist_track VALUES(1,1500);
INSERT INTO playlist_track VALUES(1,1501);
INSERT INTO playlist_track VALUES(1,1502);
INSERT INTO playlist_track VALUES(1,1503);
INSERT INTO playlist_track VALUES(1,1504);
INSERT INTO playlist_track VALUES(1,1505);
INSERT INTO playlist_track VALUES(1,436);
INSERT INTO playlist_track VALUES(1,437);
INSERT INTO playlist_track VALUES(1,438);
INSERT INTO playlist_track VALUES(1,439);
INSERT INTO playlist_track VALUES(1,440);
INSERT INTO playlist_track VALUES(1,441);
INSERT INTO playlist_track VALUES(1,442);
INSERT INTO playlist_track VALUES(1,443);
INSERT INTO playlist_track VALUES(1,444);
INSERT INTO playlist_track VALUES(1,445);
INSERT INTO playlist_track VALUES(1,446);
INSERT INTO playlist_track VALUES(1,447);
INSERT INTO playlist_track VALUES(1,448);
INSERT INTO playlist_track VALUES(1,449);
INSERT INTO playlist_track VALUES(1,450);
INSERT INTO playlist_track VALUES(1,451);
INSERT INTO playlist_track VALUES(1,452);
INSERT INTO playlist_track VALUES(1,453);
INSERT INTO playlist_track VALUES(1,454);
INSERT INTO playlist_track VALUES(1,455);
INSERT INTO playlist_track VALUES(1,1562);
INSERT INTO playlist_track VALUES(1,1563);
INSERT INTO playlist_track VALUES(1,1564);
INSERT INTO playlist_track VALUES(1,1565);
INSERT INTO playlist_track VALUES(1,1566);
INSERT INTO playlist_track VALUES(1,1567);
INSERT INTO playlist_track VALUES(1,1568);
INSERT INTO playlist_track VALUES(1,1569);
INSERT INTO playlist_track VALUES(1,1570);
INSERT INTO playlist_track VALUES(1,1571);
INSERT INTO playlist_track VALUES(1,1572);
INSERT INTO playlist_track VALUES(1,1573);
INSERT INTO playlist_track VALUES(1,1574);
INSERT INTO playlist_track VALUES(1,1575);
INSERT INTO playlist_track VALUES(1,1576);
INSERT INTO playlist_track VALUES(1,337);
INSERT INTO playlist_track VALUES(1,338);
INSERT INTO playlist_track VALUES(1,339);
INSERT INTO playlist_track VALUES(1,340);
INSERT INTO playlist_track VALUES(1,341);
INSERT INTO playlist_track VALUES(1,342);
INSERT INTO playlist_track VALUES(1,343);
INSERT INTO playlist_track VALUES(1,344);
INSERT INTO playlist_track VALUES(1,345);
INSERT INTO playlist_track VALUES(1,346);
INSERT INTO playlist_track VALUES(1,347);
INSERT INTO playlist_track VALUES(1,348);
INSERT INTO playlist_track VALUES(1,349);
INSERT INTO playlist_track VALUES(1,350);
INSERT INTO playlist_track VALUES(1,1577);
INSERT INTO playlist_track VALUES(1,1578);
INSERT INTO playlist_track VALUES(1,1579);
INSERT INTO playlist_track VALUES(1,1580);
INSERT INTO playlist_track VALUES(1,1581);
INSERT INTO playlist_track VALUES(1,1582);
INSERT INTO playlist_track VALUES(1,1583);
INSERT INTO playlist_track VALUES(1,1584);
INSERT INTO playlist_track VALUES(1,1585);
INSERT INTO playlist_track VALUES(1,1586);
INSERT INTO playlist_track VALUES(1,1587);
INSERT INTO playlist_track VALUES(1,1588);
INSERT INTO playlist_track VALUES(1,1589);
INSERT INTO playlist_track VALUES(1,1590);
INSERT INTO playlist_track VALUES(1,1591);
INSERT INTO playlist_track VALUES(1,1592);
INSERT INTO playlist_track VALUES(1,1593);
INSERT INTO playlist_track VALUES(1,1594);
INSERT INTO playlist_track VALUES(1,1595);
INSERT INTO playlist_track VALUES(1,1596);
INSERT INTO playlist_track VALUES(1,1597);
INSERT INTO playlist_track VALUES(1,1598);
INSERT INTO playlist_track VALUES(1,1599);
INSERT INTO playlist_track VALUES(1,1600);
INSERT INTO playlist_track VALUES(1,1601);
INSERT INTO playlist_track VALUES(1,1602);
INSERT INTO playlist_track VALUES(1,1603);
INSERT INTO playlist_track VALUES(1,1604);
INSERT INTO playlist_track VALUES(1,1605);
INSERT INTO playlist_track VALUES(1,1606);
INSERT INTO playlist_track VALUES(1,1607);
INSERT INTO playlist_track VALUES(1,1608);
INSERT INTO playlist_track VALUES(1,1609);
INSERT INTO playlist_track VALUES(1,1610);
INSERT INTO playlist_track VALUES(1,1611);
INSERT INTO playlist_track VALUES(1,1612);
INSERT INTO playlist_track VALUES(1,1613);
INSERT INTO playlist_track VALUES(1,1614);
INSERT INTO playlist_track VALUES(1,1615);
INSERT INTO playlist_track VALUES(1,1616);
INSERT INTO playlist_track VALUES(1,1617);
INSERT INTO playlist_track VALUES(1,1618);
INSERT INTO playlist_track VALUES(1,1619);
INSERT INTO playlist_track VALUES(1,1620);
INSERT INTO playlist_track VALUES(1,1621);
INSERT INTO playlist_track VALUES(1,1622);
INSERT INTO playlist_track VALUES(1,1623);
INSERT INTO playlist_track VALUES(1,1624);
INSERT INTO playlist_track VALUES(1,1625);
INSERT INTO playlist_track VALUES(1,1626);
INSERT INTO playlist_track VALUES(1,1627);
INSERT INTO playlist_track VALUES(1,1628);
INSERT INTO playlist_track VALUES(1,1629);
INSERT INTO playlist_track VALUES(1,1630);
INSERT INTO playlist_track VALUES(1,1631);
INSERT INTO playlist_track VALUES(1,1632);
INSERT INTO playlist_track VALUES(1,1633);
INSERT INTO playlist_track VALUES(1,1634);
INSERT INTO playlist_track VALUES(1,1635);
INSERT INTO playlist_track VALUES(1,1636);
INSERT INTO playlist_track VALUES(1,1637);
INSERT INTO playlist_track VALUES(1,1638);
INSERT INTO playlist_track VALUES(1,1639);
INSERT INTO playlist_track VALUES(1,1640);
INSERT INTO playlist_track VALUES(1,1641);
INSERT INTO playlist_track VALUES(1,1642);
INSERT INTO playlist_track VALUES(1,1643);
INSERT INTO playlist_track VALUES(1,1644);
INSERT INTO playlist_track VALUES(1,1645);
INSERT INTO playlist_track VALUES(1,550);
INSERT INTO playlist_track VALUES(1,551);
INSERT INTO playlist_track VALUES(1,552);
INSERT INTO playlist_track VALUES(1,553);
INSERT INTO playlist_track VALUES(1,554);
INSERT INTO playlist_track VALUES(1,555);
INSERT INTO playlist_track VALUES(1,1646);
INSERT INTO playlist_track VALUES(1,1647);
INSERT INTO playlist_track VALUES(1,1648);
INSERT INTO playlist_track VALUES(1,1649);
INSERT INTO playlist_track VALUES(1,1650);
INSERT INTO playlist_track VALUES(1,1651);
INSERT INTO playlist_track VALUES(1,1652);
INSERT INTO playlist_track VALUES(1,1653);
INSERT INTO playlist_track VALUES(1,1654);
INSERT INTO playlist_track VALUES(1,1655);
INSERT INTO playlist_track VALUES(1,1656);
INSERT INTO playlist_track VALUES(1,1657);
INSERT INTO playlist_track VALUES(1,1658);
INSERT INTO playlist_track VALUES(1,1659);
INSERT INTO playlist_track VALUES(1,1660);
INSERT INTO playlist_track VALUES(1,1661);
INSERT INTO playlist_track VALUES(1,1662);
INSERT INTO playlist_track VALUES(1,1663);
INSERT INTO playlist_track VALUES(1,1664);
INSERT INTO playlist_track VALUES(1,1665);
INSERT INTO playlist_track VALUES(1,1666);
INSERT INTO playlist_track VALUES(1,1667);
INSERT INTO playlist_track VALUES(1,1668);
INSERT INTO playlist_track VALUES(1,1669);
INSERT INTO playlist_track VALUES(1,1670);
INSERT INTO playlist_track VALUES(1,1702);
INSERT INTO playlist_track VALUES(1,1703);
INSERT INTO playlist_track VALUES(1,1704);
INSERT INTO playlist_track VALUES(1,1705);
INSERT INTO playlist_track VALUES(1,1706);
INSERT INTO playlist_track VALUES(1,1707);
INSERT INTO playlist_track VALUES(1,1708);
INSERT INTO playlist_track VALUES(1,1709);
INSERT INTO playlist_track VALUES(1,1710);
INSERT INTO playlist_track VALUES(1,1711);
INSERT INTO playlist_track VALUES(1,1712);
INSERT INTO playlist_track VALUES(1,1713);
INSERT INTO playlist_track VALUES(1,1714);
INSERT INTO playlist_track VALUES(1,1715);
INSERT INTO playlist_track VALUES(1,1716);
INSERT INTO playlist_track VALUES(1,1745);
INSERT INTO playlist_track VALUES(1,1746);
INSERT INTO playlist_track VALUES(1,1747);
INSERT INTO playlist_track VALUES(1,1748);
INSERT INTO playlist_track VALUES(1,1749);
INSERT INTO playlist_track VALUES(1,1750);
INSERT INTO playlist_track VALUES(1,1751);
INSERT INTO playlist_track VALUES(1,1752);
INSERT INTO playlist_track VALUES(1,1753);
INSERT INTO playlist_track VALUES(1,1754);
INSERT INTO playlist_track VALUES(1,1791);
INSERT INTO playlist_track VALUES(1,1792);
INSERT INTO playlist_track VALUES(1,1793);
INSERT INTO playlist_track VALUES(1,1794);
INSERT INTO playlist_track VALUES(1,1795);
INSERT INTO playlist_track VALUES(1,1796);
INSERT INTO playlist_track VALUES(1,1797);
INSERT INTO playlist_track VALUES(1,1798);
INSERT INTO playlist_track VALUES(1,1799);
INSERT INTO playlist_track VALUES(1,1800);
INSERT INTO playlist_track VALUES(1,1986);
INSERT INTO playlist_track VALUES(1,1987);
INSERT INTO playlist_track VALUES(1,1988);
INSERT INTO playlist_track VALUES(1,1989);
INSERT INTO playlist_track VALUES(1,1990);
INSERT INTO playlist_track VALUES(1,1991);
INSERT INTO playlist_track VALUES(1,1992);
INSERT INTO playlist_track VALUES(1,1993);
INSERT INTO playlist_track VALUES(1,1994);
INSERT INTO playlist_track VALUES(1,1995);
INSERT INTO playlist_track VALUES(1,1996);
INSERT INTO playlist_track VALUES(1,1997);
INSERT INTO playlist_track VALUES(1,1998);
INSERT INTO playlist_track VALUES(1,1999);
INSERT INTO playlist_track VALUES(1,2000);
INSERT INTO playlist_track VALUES(1,2001);
INSERT INTO playlist_track VALUES(1,2002);
INSERT INTO playlist_track VALUES(1,2003);
INSERT INTO playlist_track VALUES(1,2004);
INSERT INTO playlist_track VALUES(1,2005);
INSERT INTO playlist_track VALUES(1,2006);
INSERT INTO playlist_track VALUES(1,2007);
INSERT INTO playlist_track VALUES(1,2008);
INSERT INTO playlist_track VALUES(1,2009);
INSERT INTO playlist_track VALUES(1,2010);
INSERT INTO playlist_track VALUES(1,2011);
INSERT INTO playlist_track VALUES(1,2012);
INSERT INTO playlist_track VALUES(1,2013);
INSERT INTO playlist_track VALUES(1,2014);
INSERT INTO playlist_track VALUES(1,2015);
INSERT INTO playlist_track VALUES(1,2016);
INSERT INTO playlist_track VALUES(1,2017);
INSERT INTO playlist_track VALUES(1,2018);
INSERT INTO playlist_track VALUES(1,2019);
INSERT INTO playlist_track VALUES(1,2020);
INSERT INTO playlist_track VALUES(1,2021);
INSERT INTO playlist_track VALUES(1,2022);
INSERT INTO playlist_track VALUES(1,2023);
INSERT INTO playlist_track VALUES(1,2024);
INSERT INTO playlist_track VALUES(1,2025);
INSERT INTO playlist_track VALUES(1,2026);
INSERT INTO playlist_track VALUES(1,2027);
INSERT INTO playlist_track VALUES(1,2028);
INSERT INTO playlist_track VALUES(1,2029);
INSERT INTO playlist_track VALUES(1,2093);
INSERT INTO playlist_track VALUES(1,2094);
INSERT INTO playlist_track VALUES(1,2095);
INSERT INTO playlist_track VALUES(1,2096);
INSERT INTO playlist_track VALUES(1,2097);
INSERT INTO playlist_track VALUES(1,2098);
INSERT INTO playlist_track VALUES(1,3276);
INSERT INTO playlist_track VALUES(1,3277);
INSERT INTO playlist_track VALUES(1,3278);
INSERT INTO playlist_track VALUES(1,3279);
INSERT INTO playlist_track VALUES(1,3280);
INSERT INTO playlist_track VALUES(1,3281);
INSERT INTO playlist_track VALUES(1,3282);
INSERT INTO playlist_track VALUES(1,3283);
INSERT INTO playlist_track VALUES(1,3284);
INSERT INTO playlist_track VALUES(1,3285);
INSERT INTO playlist_track VALUES(1,3286);
INSERT INTO playlist_track VALUES(1,3287);
INSERT INTO playlist_track VALUES(1,2113);
INSERT INTO playlist_track VALUES(1,2114);
INSERT INTO playlist_track VALUES(1,2115);
INSERT INTO playlist_track VALUES(1,2116);
INSERT INTO playlist_track VALUES(1,2117);
INSERT INTO playlist_track VALUES(1,2118);
INSERT INTO playlist_track VALUES(1,2119);
INSERT INTO playlist_track VALUES(1,2120);
INSERT INTO playlist_track VALUES(1,2121);
INSERT INTO playlist_track VALUES(1,2122);
INSERT INTO playlist_track VALUES(1,2123);
INSERT INTO playlist_track VALUES(1,2124);
INSERT INTO playlist_track VALUES(1,2139);
INSERT INTO playlist_track VALUES(1,2140);
INSERT INTO playlist_track VALUES(1,2141);
INSERT INTO playlist_track VALUES(1,2142);
INSERT INTO playlist_track VALUES(1,2143);
INSERT INTO playlist_track VALUES(1,2144);
INSERT INTO playlist_track VALUES(1,2145);
INSERT INTO playlist_track VALUES(1,2146);
INSERT INTO playlist_track VALUES(1,2147);
INSERT INTO playlist_track VALUES(1,2148);
INSERT INTO playlist_track VALUES(1,2149);
INSERT INTO playlist_track VALUES(1,2150);
INSERT INTO playlist_track VALUES(1,2151);
INSERT INTO playlist_track VALUES(1,2152);
INSERT INTO playlist_track VALUES(1,2153);
INSERT INTO playlist_track VALUES(1,2154);
INSERT INTO playlist_track VALUES(1,2155);
INSERT INTO playlist_track VALUES(1,2156);
INSERT INTO playlist_track VALUES(1,2157);
INSERT INTO playlist_track VALUES(1,2158);
INSERT INTO playlist_track VALUES(1,2159);
INSERT INTO playlist_track VALUES(1,2160);
INSERT INTO playlist_track VALUES(1,2161);
INSERT INTO playlist_track VALUES(1,2162);
INSERT INTO playlist_track VALUES(1,2163);
INSERT INTO playlist_track VALUES(1,2164);
INSERT INTO playlist_track VALUES(1,2178);
INSERT INTO playlist_track VALUES(1,2179);
INSERT INTO playlist_track VALUES(1,2180);
INSERT INTO playlist_track VALUES(1,2181);
INSERT INTO playlist_track VALUES(1,2182);
INSERT INTO playlist_track VALUES(1,2183);
INSERT INTO playlist_track VALUES(1,2184);
INSERT INTO playlist_track VALUES(1,2185);
INSERT INTO playlist_track VALUES(1,2186);
INSERT INTO playlist_track VALUES(1,2187);
INSERT INTO playlist_track VALUES(1,2188);
INSERT INTO playlist_track VALUES(1,2189);
INSERT INTO playlist_track VALUES(1,2190);
INSERT INTO playlist_track VALUES(1,2191);
INSERT INTO playlist_track VALUES(1,2192);
INSERT INTO playlist_track VALUES(1,2193);
INSERT INTO playlist_track VALUES(1,2194);
INSERT INTO playlist_track VALUES(1,2195);
INSERT INTO playlist_track VALUES(1,2196);
INSERT INTO playlist_track VALUES(1,2197);
INSERT INTO playlist_track VALUES(1,2198);
INSERT INTO playlist_track VALUES(1,2199);
INSERT INTO playlist_track VALUES(1,2200);
INSERT INTO playlist_track VALUES(1,2201);
INSERT INTO playlist_track VALUES(1,2202);
INSERT INTO playlist_track VALUES(1,2203);
INSERT INTO playlist_track VALUES(1,2204);
INSERT INTO playlist_track VALUES(1,2205);
INSERT INTO playlist_track VALUES(1,2206);
INSERT INTO playlist_track VALUES(1,2207);
INSERT INTO playlist_track VALUES(1,2208);
INSERT INTO playlist_track VALUES(1,2209);
INSERT INTO playlist_track VALUES(1,2210);
INSERT INTO playlist_track VALUES(1,2211);
INSERT INTO playlist_track VALUES(1,2212);
INSERT INTO playlist_track VALUES(1,2213);
INSERT INTO playlist_track VALUES(1,2214);
INSERT INTO playlist_track VALUES(1,2215);
INSERT INTO playlist_track VALUES(1,2229);
INSERT INTO playlist_track VALUES(1,2230);
INSERT INTO playlist_track VALUES(1,2231);
INSERT INTO playlist_track VALUES(1,2232);
INSERT INTO playlist_track VALUES(1,2233);
INSERT INTO playlist_track VALUES(1,2234);
INSERT INTO playlist_track VALUES(1,2235);
INSERT INTO playlist_track VALUES(1,2236);
INSERT INTO playlist_track VALUES(1,2237);
INSERT INTO playlist_track VALUES(1,2650);
INSERT INTO playlist_track VALUES(1,2651);
INSERT INTO playlist_track VALUES(1,2652);
INSERT INTO playlist_track VALUES(1,2653);
INSERT INTO playlist_track VALUES(1,2654);
INSERT INTO playlist_track VALUES(1,2655);
INSERT INTO playlist_track VALUES(1,2656);
INSERT INTO playlist_track VALUES(1,2657);
INSERT INTO playlist_track VALUES(1,2658);
INSERT INTO playlist_track VALUES(1,2659);
INSERT INTO playlist_track VALUES(1,2660);
INSERT INTO playlist_track VALUES(1,2661);
INSERT INTO playlist_track VALUES(1,2662);
INSERT INTO playlist_track VALUES(1,2663);
INSERT INTO playlist_track VALUES(1,3353);
INSERT INTO playlist_track VALUES(1,3355);
INSERT INTO playlist_track VALUES(1,2254);
INSERT INTO playlist_track VALUES(1,2255);
INSERT INTO playlist_track VALUES(1,2256);
INSERT INTO playlist_track VALUES(1,2257);
INSERT INTO playlist_track VALUES(1,2258);
INSERT INTO playlist_track VALUES(1,2259);
INSERT INTO playlist_track VALUES(1,2260);
INSERT INTO playlist_track VALUES(1,2261);
INSERT INTO playlist_track VALUES(1,2262);
INSERT INTO playlist_track VALUES(1,2263);
INSERT INTO playlist_track VALUES(1,2264);
INSERT INTO playlist_track VALUES(1,2265);
INSERT INTO playlist_track VALUES(1,2266);
INSERT INTO playlist_track VALUES(1,2267);
INSERT INTO playlist_track VALUES(1,2268);
INSERT INTO playlist_track VALUES(1,2269);
INSERT INTO playlist_track VALUES(1,2270);
INSERT INTO playlist_track VALUES(1,419);
INSERT INTO playlist_track VALUES(1,420);
INSERT INTO playlist_track VALUES(1,421);
INSERT INTO playlist_track VALUES(1,422);
INSERT INTO playlist_track VALUES(1,423);
INSERT INTO playlist_track VALUES(1,424);
INSERT INTO playlist_track VALUES(1,425);
INSERT INTO playlist_track VALUES(1,426);
INSERT INTO playlist_track VALUES(1,427);
INSERT INTO playlist_track VALUES(1,428);
INSERT INTO playlist_track VALUES(1,429);
INSERT INTO playlist_track VALUES(1,430);
INSERT INTO playlist_track VALUES(1,431);
INSERT INTO playlist_track VALUES(1,432);
INSERT INTO playlist_track VALUES(1,433);
INSERT INTO playlist_track VALUES(1,434);
INSERT INTO playlist_track VALUES(1,435);
INSERT INTO playlist_track VALUES(1,2271);
INSERT INTO playlist_track VALUES(1,2272);
INSERT INTO playlist_track VALUES(1,2273);
INSERT INTO playlist_track VALUES(1,2274);
INSERT INTO playlist_track VALUES(1,2275);
INSERT INTO playlist_track VALUES(1,2276);
INSERT INTO playlist_track VALUES(1,2277);
INSERT INTO playlist_track VALUES(1,2278);
INSERT INTO playlist_track VALUES(1,2279);
INSERT INTO playlist_track VALUES(1,2280);
INSERT INTO playlist_track VALUES(1,2281);
INSERT INTO playlist_track VALUES(1,2296);
INSERT INTO playlist_track VALUES(1,2297);
INSERT INTO playlist_track VALUES(1,2298);
INSERT INTO playlist_track VALUES(1,2299);
INSERT INTO playlist_track VALUES(1,2300);
INSERT INTO playlist_track VALUES(1,2301);
INSERT INTO playlist_track VALUES(1,2302);
INSERT INTO playlist_track VALUES(1,2303);
INSERT INTO playlist_track VALUES(1,2304);
INSERT INTO playlist_track VALUES(1,2305);
INSERT INTO playlist_track VALUES(1,2306);
INSERT INTO playlist_track VALUES(1,2307);
INSERT INTO playlist_track VALUES(1,2308);
INSERT INTO playlist_track VALUES(1,2309);
INSERT INTO playlist_track VALUES(1,2344);
INSERT INTO playlist_track VALUES(1,2345);
INSERT INTO playlist_track VALUES(1,2346);
INSERT INTO playlist_track VALUES(1,2347);
INSERT INTO playlist_track VALUES(1,2348);
INSERT INTO playlist_track VALUES(1,2349);
INSERT INTO playlist_track VALUES(1,2350);
INSERT INTO playlist_track VALUES(1,2351);
INSERT INTO playlist_track VALUES(1,2352);
INSERT INTO playlist_track VALUES(1,2353);
INSERT INTO playlist_track VALUES(1,2354);
INSERT INTO playlist_track VALUES(1,2355);
INSERT INTO playlist_track VALUES(1,2356);
INSERT INTO playlist_track VALUES(1,2357);
INSERT INTO playlist_track VALUES(1,2375);
INSERT INTO playlist_track VALUES(1,2376);
INSERT INTO playlist_track VALUES(1,2377);
INSERT INTO playlist_track VALUES(1,2378);
INSERT INTO playlist_track VALUES(1,2379);
INSERT INTO playlist_track VALUES(1,2380);
INSERT INTO playlist_track VALUES(1,2381);
INSERT INTO playlist_track VALUES(1,2382);
INSERT INTO playlist_track VALUES(1,2383);
INSERT INTO playlist_track VALUES(1,2384);
INSERT INTO playlist_track VALUES(1,2385);
INSERT INTO playlist_track VALUES(1,2386);
INSERT INTO playlist_track VALUES(1,2387);
INSERT INTO playlist_track VALUES(1,2388);
INSERT INTO playlist_track VALUES(1,2389);
INSERT INTO playlist_track VALUES(1,2390);
INSERT INTO playlist_track VALUES(1,2391);
INSERT INTO playlist_track VALUES(1,2392);
INSERT INTO playlist_track VALUES(1,2393);
INSERT INTO playlist_track VALUES(1,2394);
INSERT INTO playlist_track VALUES(1,2395);
INSERT INTO playlist_track VALUES(1,2396);
INSERT INTO playlist_track VALUES(1,2397);
INSERT INTO playlist_track VALUES(1,2398);
INSERT INTO playlist_track VALUES(1,2399);
INSERT INTO playlist_track VALUES(1,2400);
INSERT INTO playlist_track VALUES(1,2401);
INSERT INTO playlist_track VALUES(1,2402);
INSERT INTO playlist_track VALUES(1,2403);
INSERT INTO playlist_track VALUES(1,2404);
INSERT INTO playlist_track VALUES(1,2405);
INSERT INTO playlist_track VALUES(1,2664);
INSERT INTO playlist_track VALUES(1,2665);
INSERT INTO playlist_track VALUES(1,2666);
INSERT INTO playlist_track VALUES(1,2667);
INSERT INTO playlist_track VALUES(1,2668);
INSERT INTO playlist_track VALUES(1,2669);
INSERT INTO playlist_track VALUES(1,2670);
INSERT INTO playlist_track VALUES(1,2671);
INSERT INTO playlist_track VALUES(1,2672);
INSERT INTO playlist_track VALUES(1,2673);
INSERT INTO playlist_track VALUES(1,2674);
INSERT INTO playlist_track VALUES(1,2675);
INSERT INTO playlist_track VALUES(1,2676);
INSERT INTO playlist_track VALUES(1,2677);
INSERT INTO playlist_track VALUES(1,2678);
INSERT INTO playlist_track VALUES(1,2679);
INSERT INTO playlist_track VALUES(1,2680);
INSERT INTO playlist_track VALUES(1,2681);
INSERT INTO playlist_track VALUES(1,2682);
INSERT INTO playlist_track VALUES(1,2683);
INSERT INTO playlist_track VALUES(1,2684);
INSERT INTO playlist_track VALUES(1,2685);
INSERT INTO playlist_track VALUES(1,2686);
INSERT INTO playlist_track VALUES(1,2687);
INSERT INTO playlist_track VALUES(1,2688);
INSERT INTO playlist_track VALUES(1,2689);
INSERT INTO playlist_track VALUES(1,2690);
INSERT INTO playlist_track VALUES(1,2691);
INSERT INTO playlist_track VALUES(1,2692);
INSERT INTO playlist_track VALUES(1,2693);
INSERT INTO playlist_track VALUES(1,2694);
INSERT INTO playlist_track VALUES(1,2695);
INSERT INTO playlist_track VALUES(1,2696);
INSERT INTO playlist_track VALUES(1,2697);
INSERT INTO playlist_track VALUES(1,2698);
INSERT INTO playlist_track VALUES(1,2699);
INSERT INTO playlist_track VALUES(1,2700);
INSERT INTO playlist_track VALUES(1,2701);
INSERT INTO playlist_track VALUES(1,2702);
INSERT INTO playlist_track VALUES(1,2703);
INSERT INTO playlist_track VALUES(1,2704);
INSERT INTO playlist_track VALUES(1,2406);
INSERT INTO playlist_track VALUES(1,2407);
INSERT INTO playlist_track VALUES(1,2408);
INSERT INTO playlist_track VALUES(1,2409);
INSERT INTO playlist_track VALUES(1,2410);
INSERT INTO playlist_track VALUES(1,2411);
INSERT INTO playlist_track VALUES(1,2412);
INSERT INTO playlist_track VALUES(1,2413);
INSERT INTO playlist_track VALUES(1,2414);
INSERT INTO playlist_track VALUES(1,2415);
INSERT INTO playlist_track VALUES(1,2416);
INSERT INTO playlist_track VALUES(1,2417);
INSERT INTO playlist_track VALUES(1,2418);
INSERT INTO playlist_track VALUES(1,2419);
INSERT INTO playlist_track VALUES(1,2420);
INSERT INTO playlist_track VALUES(1,2421);
INSERT INTO playlist_track VALUES(1,2422);
INSERT INTO playlist_track VALUES(1,2423);
INSERT INTO playlist_track VALUES(1,2424);
INSERT INTO playlist_track VALUES(1,2425);
INSERT INTO playlist_track VALUES(1,2426);
INSERT INTO playlist_track VALUES(1,2427);
INSERT INTO playlist_track VALUES(1,2428);
INSERT INTO playlist_track VALUES(1,2429);
INSERT INTO playlist_track VALUES(1,2430);
INSERT INTO playlist_track VALUES(1,2431);
INSERT INTO playlist_track VALUES(1,2432);
INSERT INTO playlist_track VALUES(1,2433);
INSERT INTO playlist_track VALUES(1,570);
INSERT INTO playlist_track VALUES(1,573);
INSERT INTO playlist_track VALUES(1,577);
INSERT INTO playlist_track VALUES(1,580);
INSERT INTO playlist_track VALUES(1,581);
INSERT INTO playlist_track VALUES(1,571);
INSERT INTO playlist_track VALUES(1,579);
INSERT INTO playlist_track VALUES(1,582);
INSERT INTO playlist_track VALUES(1,572);
INSERT INTO playlist_track VALUES(1,575);
INSERT INTO playlist_track VALUES(1,578);
INSERT INTO playlist_track VALUES(1,574);
INSERT INTO playlist_track VALUES(1,576);
INSERT INTO playlist_track VALUES(1,3288);
INSERT INTO playlist_track VALUES(1,3289);
INSERT INTO playlist_track VALUES(1,3290);
INSERT INTO playlist_track VALUES(1,3291);
INSERT INTO playlist_track VALUES(1,3292);
INSERT INTO playlist_track VALUES(1,3293);
INSERT INTO playlist_track VALUES(1,3294);
INSERT INTO playlist_track VALUES(1,3295);
INSERT INTO playlist_track VALUES(1,3296);
INSERT INTO playlist_track VALUES(1,3297);
INSERT INTO playlist_track VALUES(1,3298);
INSERT INTO playlist_track VALUES(1,3299);
INSERT INTO playlist_track VALUES(1,2434);
INSERT INTO playlist_track VALUES(1,2435);
INSERT INTO playlist_track VALUES(1,2436);
INSERT INTO playlist_track VALUES(1,2437);
INSERT INTO playlist_track VALUES(1,2438);
INSERT INTO playlist_track VALUES(1,2439);
INSERT INTO playlist_track VALUES(1,2440);
INSERT INTO playlist_track VALUES(1,2441);
INSERT INTO playlist_track VALUES(1,2442);
INSERT INTO playlist_track VALUES(1,2443);
INSERT INTO playlist_track VALUES(1,2444);
INSERT INTO playlist_track VALUES(1,2445);
INSERT INTO playlist_track VALUES(1,2446);
INSERT INTO playlist_track VALUES(1,2447);
INSERT INTO playlist_track VALUES(1,2448);
INSERT INTO playlist_track VALUES(1,2449);
INSERT INTO playlist_track VALUES(1,2450);
INSERT INTO playlist_track VALUES(1,2451);
INSERT INTO playlist_track VALUES(1,2452);
INSERT INTO playlist_track VALUES(1,2453);
INSERT INTO playlist_track VALUES(1,2454);
INSERT INTO playlist_track VALUES(1,2455);
INSERT INTO playlist_track VALUES(1,2456);
INSERT INTO playlist_track VALUES(1,2457);
INSERT INTO playlist_track VALUES(1,2458);
INSERT INTO playlist_track VALUES(1,2459);
INSERT INTO playlist_track VALUES(1,2460);
INSERT INTO playlist_track VALUES(1,2461);
INSERT INTO playlist_track VALUES(1,2462);
INSERT INTO playlist_track VALUES(1,2463);
INSERT INTO playlist_track VALUES(1,2464);
INSERT INTO playlist_track VALUES(1,2465);
INSERT INTO playlist_track VALUES(1,2466);
INSERT INTO playlist_track VALUES(1,2467);
INSERT INTO playlist_track VALUES(1,2468);
INSERT INTO playlist_track VALUES(1,2469);
INSERT INTO playlist_track VALUES(1,2470);
INSERT INTO playlist_track VALUES(1,2471);
INSERT INTO playlist_track VALUES(1,2506);
INSERT INTO playlist_track VALUES(1,2507);
INSERT INTO playlist_track VALUES(1,2508);
INSERT INTO playlist_track VALUES(1,2509);
INSERT INTO playlist_track VALUES(1,2510);
INSERT INTO playlist_track VALUES(1,2511);
INSERT INTO playlist_track VALUES(1,2512);
INSERT INTO playlist_track VALUES(1,2513);
INSERT INTO playlist_track VALUES(1,2514);
INSERT INTO playlist_track VALUES(1,2515);
INSERT INTO playlist_track VALUES(1,2516);
INSERT INTO playlist_track VALUES(1,2517);
INSERT INTO playlist_track VALUES(1,2518);
INSERT INTO playlist_track VALUES(1,2519);
INSERT INTO playlist_track VALUES(1,2520);
INSERT INTO playlist_track VALUES(1,2521);
INSERT INTO playlist_track VALUES(1,2522);
INSERT INTO playlist_track VALUES(1,2542);
INSERT INTO playlist_track VALUES(1,2543);
INSERT INTO playlist_track VALUES(1,2544);
INSERT INTO playlist_track VALUES(1,2545);
INSERT INTO playlist_track VALUES(1,2546);
INSERT INTO playlist_track VALUES(1,2547);
INSERT INTO playlist_track VALUES(1,2548);
INSERT INTO playlist_track VALUES(1,2549);
INSERT INTO playlist_track VALUES(1,2550);
INSERT INTO playlist_track VALUES(1,2551);
INSERT INTO playlist_track VALUES(1,2552);
INSERT INTO playlist_track VALUES(1,2553);
INSERT INTO playlist_track VALUES(1,2565);
INSERT INTO playlist_track VALUES(1,2566);
INSERT INTO playlist_track VALUES(1,2567);
INSERT INTO playlist_track VALUES(1,2568);
INSERT INTO playlist_track VALUES(1,2569);
INSERT INTO playlist_track VALUES(1,2570);
INSERT INTO playlist_track VALUES(1,2571);
INSERT INTO playlist_track VALUES(1,2926);
INSERT INTO playlist_track VALUES(1,2927);
INSERT INTO playlist_track VALUES(1,2928);
INSERT INTO playlist_track VALUES(1,2929);
INSERT INTO playlist_track VALUES(1,2930);
INSERT INTO playlist_track VALUES(1,2931);
INSERT INTO playlist_track VALUES(1,2932);
INSERT INTO playlist_track VALUES(1,2933);
INSERT INTO playlist_track VALUES(1,2934);
INSERT INTO playlist_track VALUES(1,2935);
INSERT INTO playlist_track VALUES(1,2936);
INSERT INTO playlist_track VALUES(1,2937);
INSERT INTO playlist_track VALUES(1,2938);
INSERT INTO playlist_track VALUES(1,2939);
INSERT INTO playlist_track VALUES(1,2940);
INSERT INTO playlist_track VALUES(1,2941);
INSERT INTO playlist_track VALUES(1,2942);
INSERT INTO playlist_track VALUES(1,2943);
INSERT INTO playlist_track VALUES(1,2944);
INSERT INTO playlist_track VALUES(1,2945);
INSERT INTO playlist_track VALUES(1,2946);
INSERT INTO playlist_track VALUES(1,2947);
INSERT INTO playlist_track VALUES(1,2948);
INSERT INTO playlist_track VALUES(1,2949);
INSERT INTO playlist_track VALUES(1,2950);
INSERT INTO playlist_track VALUES(1,2951);
INSERT INTO playlist_track VALUES(1,2952);
INSERT INTO playlist_track VALUES(1,2953);
INSERT INTO playlist_track VALUES(1,2954);
INSERT INTO playlist_track VALUES(1,2955);
INSERT INTO playlist_track VALUES(1,2956);
INSERT INTO playlist_track VALUES(1,2957);
INSERT INTO playlist_track VALUES(1,2958);
INSERT INTO playlist_track VALUES(1,2959);
INSERT INTO playlist_track VALUES(1,2960);
INSERT INTO playlist_track VALUES(1,2961);
INSERT INTO playlist_track VALUES(1,2962);
INSERT INTO playlist_track VALUES(1,2963);
INSERT INTO playlist_track VALUES(1,3004);
INSERT INTO playlist_track VALUES(1,3005);
INSERT INTO playlist_track VALUES(1,3006);
INSERT INTO playlist_track VALUES(1,3007);
INSERT INTO playlist_track VALUES(1,3008);
INSERT INTO playlist_track VALUES(1,3009);
INSERT INTO playlist_track VALUES(1,3010);
INSERT INTO playlist_track VALUES(1,3011);
INSERT INTO playlist_track VALUES(1,3012);
INSERT INTO playlist_track VALUES(1,3013);
INSERT INTO playlist_track VALUES(1,3014);
INSERT INTO playlist_track VALUES(1,3015);
INSERT INTO playlist_track VALUES(1,3016);
INSERT INTO playlist_track VALUES(1,3017);
INSERT INTO playlist_track VALUES(1,2964);
INSERT INTO playlist_track VALUES(1,2965);
INSERT INTO playlist_track VALUES(1,2966);
INSERT INTO playlist_track VALUES(1,2967);
INSERT INTO playlist_track VALUES(1,2968);
INSERT INTO playlist_track VALUES(1,2969);
INSERT INTO playlist_track VALUES(1,2970);
INSERT INTO playlist_track VALUES(1,2971);
INSERT INTO playlist_track VALUES(1,2972);
INSERT INTO playlist_track VALUES(1,2973);
INSERT INTO playlist_track VALUES(1,2974);
INSERT INTO playlist_track VALUES(1,2975);
INSERT INTO playlist_track VALUES(1,2976);
INSERT INTO playlist_track VALUES(1,2977);
INSERT INTO playlist_track VALUES(1,2978);
INSERT INTO playlist_track VALUES(1,2979);
INSERT INTO playlist_track VALUES(1,2980);
INSERT INTO playlist_track VALUES(1,2981);
INSERT INTO playlist_track VALUES(1,2982);
INSERT INTO playlist_track VALUES(1,2983);
INSERT INTO playlist_track VALUES(1,2984);
INSERT INTO playlist_track VALUES(1,2985);
INSERT INTO playlist_track VALUES(1,2986);
INSERT INTO playlist_track VALUES(1,2987);
INSERT INTO playlist_track VALUES(1,2988);
INSERT INTO playlist_track VALUES(1,2989);
INSERT INTO playlist_track VALUES(1,2990);
INSERT INTO playlist_track VALUES(1,2991);
INSERT INTO playlist_track VALUES(1,2992);
INSERT INTO playlist_track VALUES(1,2993);
INSERT INTO playlist_track VALUES(1,2994);
INSERT INTO playlist_track VALUES(1,2995);
INSERT INTO playlist_track VALUES(1,2996);
INSERT INTO playlist_track VALUES(1,2997);
INSERT INTO playlist_track VALUES(1,2998);
INSERT INTO playlist_track VALUES(1,2999);
INSERT INTO playlist_track VALUES(1,3000);
INSERT INTO playlist_track VALUES(1,3001);
INSERT INTO playlist_track VALUES(1,3002);
INSERT INTO playlist_track VALUES(1,3003);
INSERT INTO playlist_track VALUES(1,3018);
INSERT INTO playlist_track VALUES(1,3019);
INSERT INTO playlist_track VALUES(1,3020);
INSERT INTO playlist_track VALUES(1,3021);
INSERT INTO playlist_track VALUES(1,3022);
INSERT INTO playlist_track VALUES(1,3023);
INSERT INTO playlist_track VALUES(1,3024);
INSERT INTO playlist_track VALUES(1,3025);
INSERT INTO playlist_track VALUES(1,3026);
INSERT INTO playlist_track VALUES(1,3027);
INSERT INTO playlist_track VALUES(1,3028);
INSERT INTO playlist_track VALUES(1,3029);
INSERT INTO playlist_track VALUES(1,3030);
INSERT INTO playlist_track VALUES(1,3031);
INSERT INTO playlist_track VALUES(1,3032);
INSERT INTO playlist_track VALUES(1,3033);
INSERT INTO playlist_track VALUES(1,3034);
INSERT INTO playlist_track VALUES(1,3035);
INSERT INTO playlist_track VALUES(1,3036);
INSERT INTO playlist_track VALUES(1,3037);
INSERT INTO playlist_track VALUES(1,3064);
INSERT INTO playlist_track VALUES(1,3065);
INSERT INTO playlist_track VALUES(1,3066);
INSERT INTO playlist_track VALUES(1,3067);
INSERT INTO playlist_track VALUES(1,3068);
INSERT INTO playlist_track VALUES(1,3069);
INSERT INTO playlist_track VALUES(1,3070);
INSERT INTO playlist_track VALUES(1,3071);
INSERT INTO playlist_track VALUES(1,3072);
INSERT INTO playlist_track VALUES(1,3073);
INSERT INTO playlist_track VALUES(1,3074);
INSERT INTO playlist_track VALUES(1,3075);
INSERT INTO playlist_track VALUES(1,3076);
INSERT INTO playlist_track VALUES(1,3077);
INSERT INTO playlist_track VALUES(1,3078);
INSERT INTO playlist_track VALUES(1,3079);
INSERT INTO playlist_track VALUES(1,3080);
INSERT INTO playlist_track VALUES(1,3052);
INSERT INTO playlist_track VALUES(1,3053);
INSERT INTO playlist_track VALUES(1,3054);
INSERT INTO playlist_track VALUES(1,3055);
INSERT INTO playlist_track VALUES(1,3056);
INSERT INTO playlist_track VALUES(1,3057);
INSERT INTO playlist_track VALUES(1,3058);
INSERT INTO playlist_track VALUES(1,3059);
INSERT INTO playlist_track VALUES(1,3060);
INSERT INTO playlist_track VALUES(1,3061);
INSERT INTO playlist_track VALUES(1,3062);
INSERT INTO playlist_track VALUES(1,3063);
INSERT INTO playlist_track VALUES(1,3081);
INSERT INTO playlist_track VALUES(1,3082);
INSERT INTO playlist_track VALUES(1,3083);
INSERT INTO playlist_track VALUES(1,3084);
INSERT INTO playlist_track VALUES(1,3085);
INSERT INTO playlist_track VALUES(1,3086);
INSERT INTO playlist_track VALUES(1,3087);
INSERT INTO playlist_track VALUES(1,3088);
INSERT INTO playlist_track VALUES(1,3089);
INSERT INTO playlist_track VALUES(1,3090);
INSERT INTO playlist_track VALUES(1,3091);
INSERT INTO playlist_track VALUES(1,3092);
INSERT INTO playlist_track VALUES(1,3093);
INSERT INTO playlist_track VALUES(1,3094);
INSERT INTO playlist_track VALUES(1,3095);
INSERT INTO playlist_track VALUES(1,3096);
INSERT INTO playlist_track VALUES(1,3097);
INSERT INTO playlist_track VALUES(1,3098);
INSERT INTO playlist_track VALUES(1,3099);
INSERT INTO playlist_track VALUES(1,3100);
INSERT INTO playlist_track VALUES(1,3101);
INSERT INTO playlist_track VALUES(1,3102);
INSERT INTO playlist_track VALUES(1,3103);
INSERT INTO playlist_track VALUES(1,3104);
INSERT INTO playlist_track VALUES(1,3105);
INSERT INTO playlist_track VALUES(1,3106);
INSERT INTO playlist_track VALUES(1,3107);
INSERT INTO playlist_track VALUES(1,3108);
INSERT INTO playlist_track VALUES(1,3109);
INSERT INTO playlist_track VALUES(1,3110);
INSERT INTO playlist_track VALUES(1,3111);
INSERT INTO playlist_track VALUES(1,3112);
INSERT INTO playlist_track VALUES(1,3113);
INSERT INTO playlist_track VALUES(1,3114);
INSERT INTO playlist_track VALUES(1,3115);
INSERT INTO playlist_track VALUES(1,3116);
INSERT INTO playlist_track VALUES(1,2731);
INSERT INTO playlist_track VALUES(1,2732);
INSERT INTO playlist_track VALUES(1,2733);
INSERT INTO playlist_track VALUES(1,2734);
INSERT INTO playlist_track VALUES(1,2735);
INSERT INTO playlist_track VALUES(1,2736);
INSERT INTO playlist_track VALUES(1,2737);
INSERT INTO playlist_track VALUES(1,2738);
INSERT INTO playlist_track VALUES(1,2739);
INSERT INTO playlist_track VALUES(1,2740);
INSERT INTO playlist_track VALUES(1,2741);
INSERT INTO playlist_track VALUES(1,2742);
INSERT INTO playlist_track VALUES(1,2743);
INSERT INTO playlist_track VALUES(1,2744);
INSERT INTO playlist_track VALUES(1,2745);
INSERT INTO playlist_track VALUES(1,2746);
INSERT INTO playlist_track VALUES(1,2747);
INSERT INTO playlist_track VALUES(1,2748);
INSERT INTO playlist_track VALUES(1,2749);
INSERT INTO playlist_track VALUES(1,2750);
INSERT INTO playlist_track VALUES(1,111);
INSERT INTO playlist_track VALUES(1,112);
INSERT INTO playlist_track VALUES(1,113);
INSERT INTO playlist_track VALUES(1,114);
INSERT INTO playlist_track VALUES(1,115);
INSERT INTO playlist_track VALUES(1,116);
INSERT INTO playlist_track VALUES(1,117);
INSERT INTO playlist_track VALUES(1,118);
INSERT INTO playlist_track VALUES(1,119);
INSERT INTO playlist_track VALUES(1,120);
INSERT INTO playlist_track VALUES(1,121);
INSERT INTO playlist_track VALUES(1,122);
INSERT INTO playlist_track VALUES(1,1073);
INSERT INTO playlist_track VALUES(1,1074);
INSERT INTO playlist_track VALUES(1,1075);
INSERT INTO playlist_track VALUES(1,1076);
INSERT INTO playlist_track VALUES(1,1077);
INSERT INTO playlist_track VALUES(1,1078);
INSERT INTO playlist_track VALUES(1,1079);
INSERT INTO playlist_track VALUES(1,1080);
INSERT INTO playlist_track VALUES(1,1081);
INSERT INTO playlist_track VALUES(1,1082);
INSERT INTO playlist_track VALUES(1,1083);
INSERT INTO playlist_track VALUES(1,1084);
INSERT INTO playlist_track VALUES(1,1085);
INSERT INTO playlist_track VALUES(1,1086);
INSERT INTO playlist_track VALUES(1,2125);
INSERT INTO playlist_track VALUES(1,2126);
INSERT INTO playlist_track VALUES(1,2127);
INSERT INTO playlist_track VALUES(1,2128);
INSERT INTO playlist_track VALUES(1,2129);
INSERT INTO playlist_track VALUES(1,2130);
INSERT INTO playlist_track VALUES(1,2131);
INSERT INTO playlist_track VALUES(1,2132);
INSERT INTO playlist_track VALUES(1,2133);
INSERT INTO playlist_track VALUES(1,2134);
INSERT INTO playlist_track VALUES(1,2135);
INSERT INTO playlist_track VALUES(1,2136);
INSERT INTO playlist_track VALUES(1,2137);
INSERT INTO playlist_track VALUES(1,2138);
INSERT INTO playlist_track VALUES(1,3503);
INSERT INTO playlist_track VALUES(1,360);
INSERT INTO playlist_track VALUES(1,361);
INSERT INTO playlist_track VALUES(1,362);
INSERT INTO playlist_track VALUES(1,363);
INSERT INTO playlist_track VALUES(1,364);
INSERT INTO playlist_track VALUES(1,365);
INSERT INTO playlist_track VALUES(1,366);
INSERT INTO playlist_track VALUES(1,367);
INSERT INTO playlist_track VALUES(1,368);
INSERT INTO playlist_track VALUES(1,369);
INSERT INTO playlist_track VALUES(1,370);
INSERT INTO playlist_track VALUES(1,371);
INSERT INTO playlist_track VALUES(1,372);
INSERT INTO playlist_track VALUES(1,373);
INSERT INTO playlist_track VALUES(1,3354);
INSERT INTO playlist_track VALUES(1,3351);
INSERT INTO playlist_track VALUES(1,1532);
INSERT INTO playlist_track VALUES(1,1533);
INSERT INTO playlist_track VALUES(1,1534);
INSERT INTO playlist_track VALUES(1,1535);
INSERT INTO playlist_track VALUES(1,1536);
INSERT INTO playlist_track VALUES(1,1537);
INSERT INTO playlist_track VALUES(1,1538);
INSERT INTO playlist_track VALUES(1,1539);
INSERT INTO playlist_track VALUES(1,1540);
INSERT INTO playlist_track VALUES(1,1541);
INSERT INTO playlist_track VALUES(1,1542);
INSERT INTO playlist_track VALUES(1,1543);
INSERT INTO playlist_track VALUES(1,1544);
INSERT INTO playlist_track VALUES(1,1545);
INSERT INTO playlist_track VALUES(1,1957);
INSERT INTO playlist_track VALUES(1,1958);
INSERT INTO playlist_track VALUES(1,1959);
INSERT INTO playlist_track VALUES(1,1960);
INSERT INTO playlist_track VALUES(1,1961);
INSERT INTO playlist_track VALUES(1,1962);
INSERT INTO playlist_track VALUES(1,1963);
INSERT INTO playlist_track VALUES(1,1964);
INSERT INTO playlist_track VALUES(1,1965);
INSERT INTO playlist_track VALUES(1,1966);
INSERT INTO playlist_track VALUES(1,1967);
INSERT INTO playlist_track VALUES(1,1968);
INSERT INTO playlist_track VALUES(3,3250);
INSERT INTO playlist_track VALUES(3,2819);
INSERT INTO playlist_track VALUES(3,2820);
INSERT INTO playlist_track VALUES(3,2821);
INSERT INTO playlist_track VALUES(3,2822);
INSERT INTO playlist_track VALUES(3,2823);
INSERT INTO playlist_track VALUES(3,2824);
INSERT INTO playlist_track VALUES(3,2825);
INSERT INTO playlist_track VALUES(3,2826);
INSERT INTO playlist_track VALUES(3,2827);
INSERT INTO playlist_track VALUES(3,2828);
INSERT INTO playlist_track VALUES(3,2829);
INSERT INTO playlist_track VALUES(3,2830);
INSERT INTO playlist_track VALUES(3,2831);
INSERT INTO playlist_track VALUES(3,2832);
INSERT INTO playlist_track VALUES(3,2833);
INSERT INTO playlist_track VALUES(3,2834);
INSERT INTO playlist_track VALUES(3,2835);
INSERT INTO playlist_track VALUES(3,2836);
INSERT INTO playlist_track VALUES(3,2837);
INSERT INTO playlist_track VALUES(3,2838);
INSERT INTO playlist_track VALUES(3,3226);
INSERT INTO playlist_track VALUES(3,3227);
INSERT INTO playlist_track VALUES(3,3228);
INSERT INTO playlist_track VALUES(3,3229);
INSERT INTO playlist_track VALUES(3,3230);
INSERT INTO playlist_track VALUES(3,3231);
INSERT INTO playlist_track VALUES(3,3232);
INSERT INTO playlist_track VALUES(3,3233);
INSERT INTO playlist_track VALUES(3,3234);
INSERT INTO playlist_track VALUES(3,3235);
INSERT INTO playlist_track VALUES(3,3236);
INSERT INTO playlist_track VALUES(3,3237);
INSERT INTO playlist_track VALUES(3,3238);
INSERT INTO playlist_track VALUES(3,3239);
INSERT INTO playlist_track VALUES(3,3240);
INSERT INTO playlist_track VALUES(3,3241);
INSERT INTO playlist_track VALUES(3,3242);
INSERT INTO playlist_track VALUES(3,3243);
INSERT INTO playlist_track VALUES(3,3244);
INSERT INTO playlist_track VALUES(3,3245);
INSERT INTO playlist_track VALUES(3,3246);
INSERT INTO playlist_track VALUES(3,3247);
INSERT INTO playlist_track VALUES(3,3248);
INSERT INTO playlist_track VALUES(3,3249);
INSERT INTO playlist_track VALUES(3,2839);
INSERT INTO playlist_track VALUES(3,2840);
INSERT INTO playlist_track VALUES(3,2841);
INSERT INTO playlist_track VALUES(3,2842);
INSERT INTO playlist_track VALUES(3,2843);
INSERT INTO playlist_track VALUES(3,2844);
INSERT INTO playlist_track VALUES(3,2845);
INSERT INTO playlist_track VALUES(3,2846);
INSERT INTO playlist_track VALUES(3,2847);
INSERT INTO playlist_track VALUES(3,2848);
INSERT INTO playlist_track VALUES(3,2849);
INSERT INTO playlist_track VALUES(3,2850);
INSERT INTO playlist_track VALUES(3,2851);
INSERT INTO playlist_track VALUES(3,2852);
INSERT INTO playlist_track VALUES(3,2853);
INSERT INTO playlist_track VALUES(3,2854);
INSERT INTO playlist_track VALUES(3,2855);
INSERT INTO playlist_track VALUES(3,2856);
INSERT INTO playlist_track VALUES(3,3166);
INSERT INTO playlist_track VALUES(3,3167);
INSERT INTO playlist_track VALUES(3,3168);
INSERT INTO playlist_track VALUES(3,3171);
INSERT INTO playlist_track VALUES(3,3223);
INSERT INTO playlist_track VALUES(3,2858);
INSERT INTO playlist_track VALUES(3,2861);
INSERT INTO playlist_track VALUES(3,2865);
INSERT INTO playlist_track VALUES(3,2868);
INSERT INTO playlist_track VALUES(3,2871);
INSERT INTO playlist_track VALUES(3,2873);
INSERT INTO playlist_track VALUES(3,2877);
INSERT INTO playlist_track VALUES(3,2880);
INSERT INTO playlist_track VALUES(3,2883);
INSERT INTO playlist_track VALUES(3,2885);
INSERT INTO playlist_track VALUES(3,2888);
INSERT INTO playlist_track VALUES(3,2893);
INSERT INTO playlist_track VALUES(3,2894);
INSERT INTO playlist_track VALUES(3,2898);
INSERT INTO playlist_track VALUES(3,2901);
INSERT INTO playlist_track VALUES(3,2904);
INSERT INTO playlist_track VALUES(3,2906);
INSERT INTO playlist_track VALUES(3,2911);
INSERT INTO playlist_track VALUES(3,2913);
INSERT INTO playlist_track VALUES(3,2915);
INSERT INTO playlist_track VALUES(3,2917);
INSERT INTO playlist_track VALUES(3,2919);
INSERT INTO playlist_track VALUES(3,2921);
INSERT INTO playlist_track VALUES(3,2923);
INSERT INTO playlist_track VALUES(3,2925);
INSERT INTO playlist_track VALUES(3,2859);
INSERT INTO playlist_track VALUES(3,2860);
INSERT INTO playlist_track VALUES(3,2864);
INSERT INTO playlist_track VALUES(3,2867);
INSERT INTO playlist_track VALUES(3,2869);
INSERT INTO playlist_track VALUES(3,2872);
INSERT INTO playlist_track VALUES(3,2878);
INSERT INTO playlist_track VALUES(3,2879);
INSERT INTO playlist_track VALUES(3,2884);
INSERT INTO playlist_track VALUES(3,2887);
INSERT INTO playlist_track VALUES(3,2889);
INSERT INTO playlist_track VALUES(3,2892);
INSERT INTO playlist_track VALUES(3,2896);
INSERT INTO playlist_track VALUES(3,2897);
INSERT INTO playlist_track VALUES(3,2902);
INSERT INTO playlist_track VALUES(3,2905);
INSERT INTO playlist_track VALUES(3,2907);
INSERT INTO playlist_track VALUES(3,2910);
INSERT INTO playlist_track VALUES(3,2914);
INSERT INTO playlist_track VALUES(3,2916);
INSERT INTO playlist_track VALUES(3,2918);
INSERT INTO playlist_track VALUES(3,2920);
INSERT INTO playlist_track VALUES(3,2922);
INSERT INTO playlist_track VALUES(3,2924);
INSERT INTO playlist_track VALUES(3,2857);
INSERT INTO playlist_track VALUES(3,2862);
INSERT INTO playlist_track VALUES(3,2863);
INSERT INTO playlist_track VALUES(3,2866);
INSERT INTO playlist_track VALUES(3,2870);
INSERT INTO playlist_track VALUES(3,2874);
INSERT INTO playlist_track VALUES(3,2875);
INSERT INTO playlist_track VALUES(3,2876);
INSERT INTO playlist_track VALUES(3,2881);
INSERT INTO playlist_track VALUES(3,2882);
INSERT INTO playlist_track VALUES(3,2886);
INSERT INTO playlist_track VALUES(3,2890);
INSERT INTO playlist_track VALUES(3,2891);
INSERT INTO playlist_track VALUES(3,2895);
INSERT INTO playlist_track VALUES(3,2899);
INSERT INTO playlist_track VALUES(3,2900);
INSERT INTO playlist_track VALUES(3,2903);
INSERT INTO playlist_track VALUES(3,2908);
INSERT INTO playlist_track VALUES(3,2909);
INSERT INTO playlist_track VALUES(3,2912);
INSERT INTO playlist_track VALUES(3,3165);
INSERT INTO playlist_track VALUES(3,3169);
INSERT INTO playlist_track VALUES(3,3170);
INSERT INTO playlist_track VALUES(3,3252);
INSERT INTO playlist_track VALUES(3,3224);
INSERT INTO playlist_track VALUES(3,3251);
INSERT INTO playlist_track VALUES(3,3340);
INSERT INTO playlist_track VALUES(3,3339);
INSERT INTO playlist_track VALUES(3,3338);
INSERT INTO playlist_track VALUES(3,3337);
INSERT INTO playlist_track VALUES(3,3341);
INSERT INTO playlist_track VALUES(3,3345);
INSERT INTO playlist_track VALUES(3,3342);
INSERT INTO playlist_track VALUES(3,3346);
INSERT INTO playlist_track VALUES(3,3343);
INSERT INTO playlist_track VALUES(3,3347);
INSERT INTO playlist_track VALUES(3,3344);
INSERT INTO playlist_track VALUES(3,3348);
INSERT INTO playlist_track VALUES(3,3360);
INSERT INTO playlist_track VALUES(3,3361);
INSERT INTO playlist_track VALUES(3,3362);
INSERT INTO playlist_track VALUES(3,3363);
INSERT INTO playlist_track VALUES(3,3364);
INSERT INTO playlist_track VALUES(3,3172);
INSERT INTO playlist_track VALUES(3,3173);
INSERT INTO playlist_track VALUES(3,3174);
INSERT INTO playlist_track VALUES(3,3175);
INSERT INTO playlist_track VALUES(3,3176);
INSERT INTO playlist_track VALUES(3,3177);
INSERT INTO playlist_track VALUES(3,3178);
INSERT INTO playlist_track VALUES(3,3179);
INSERT INTO playlist_track VALUES(3,3180);
INSERT INTO playlist_track VALUES(3,3181);
INSERT INTO playlist_track VALUES(3,3182);
INSERT INTO playlist_track VALUES(3,3183);
INSERT INTO playlist_track VALUES(3,3184);
INSERT INTO playlist_track VALUES(3,3185);
INSERT INTO playlist_track VALUES(3,3186);
INSERT INTO playlist_track VALUES(3,3187);
INSERT INTO playlist_track VALUES(3,3188);
INSERT INTO playlist_track VALUES(3,3189);
INSERT INTO playlist_track VALUES(3,3190);
INSERT INTO playlist_track VALUES(3,3191);
INSERT INTO playlist_track VALUES(3,3192);
INSERT INTO playlist_track VALUES(3,3193);
INSERT INTO playlist_track VALUES(3,3194);
INSERT INTO playlist_track VALUES(3,3195);
INSERT INTO playlist_track VALUES(3,3196);
INSERT INTO playlist_track VALUES(3,3197);
INSERT INTO playlist_track VALUES(3,3198);
INSERT INTO playlist_track VALUES(3,3199);
INSERT INTO playlist_track VALUES(3,3200);
INSERT INTO playlist_track VALUES(3,3201);
INSERT INTO playlist_track VALUES(3,3202);
INSERT INTO playlist_track VALUES(3,3203);
INSERT INTO playlist_track VALUES(3,3204);
INSERT INTO playlist_track VALUES(3,3205);
INSERT INTO playlist_track VALUES(3,3206);
INSERT INTO playlist_track VALUES(3,3428);
INSERT INTO playlist_track VALUES(3,3207);
INSERT INTO playlist_track VALUES(3,3208);
INSERT INTO playlist_track VALUES(3,3209);
INSERT INTO playlist_track VALUES(3,3210);
INSERT INTO playlist_track VALUES(3,3211);
INSERT INTO playlist_track VALUES(3,3212);
INSERT INTO playlist_track VALUES(3,3429);
INSERT INTO playlist_track VALUES(3,3213);
INSERT INTO playlist_track VALUES(3,3214);
INSERT INTO playlist_track VALUES(3,3215);
INSERT INTO playlist_track VALUES(3,3216);
INSERT INTO playlist_track VALUES(3,3217);
INSERT INTO playlist_track VALUES(3,3218);
INSERT INTO playlist_track VALUES(3,3219);
INSERT INTO playlist_track VALUES(3,3220);
INSERT INTO playlist_track VALUES(3,3221);
INSERT INTO playlist_track VALUES(3,3222);
INSERT INTO playlist_track VALUES(5,51);
INSERT INTO playlist_track VALUES(5,52);
INSERT INTO playlist_track VALUES(5,53);
INSERT INTO playlist_track VALUES(5,54);
INSERT INTO playlist_track VALUES(5,55);
INSERT INTO playlist_track VALUES(5,56);
INSERT INTO playlist_track VALUES(5,57);
INSERT INTO playlist_track VALUES(5,58);
INSERT INTO playlist_track VALUES(5,59);
INSERT INTO playlist_track VALUES(5,60);
INSERT INTO playlist_track VALUES(5,61);
INSERT INTO playlist_track VALUES(5,62);
INSERT INTO playlist_track VALUES(5,798);
INSERT INTO playlist_track VALUES(5,799);
INSERT INTO playlist_track VALUES(5,800);
INSERT INTO playlist_track VALUES(5,801);
INSERT INTO playlist_track VALUES(5,802);
INSERT INTO playlist_track VALUES(5,803);
INSERT INTO playlist_track VALUES(5,804);
INSERT INTO playlist_track VALUES(5,805);
INSERT INTO playlist_track VALUES(5,806);
INSERT INTO playlist_track VALUES(5,3225);
INSERT INTO playlist_track VALUES(5,1325);
INSERT INTO playlist_track VALUES(5,1326);
INSERT INTO playlist_track VALUES(5,1327);
INSERT INTO playlist_track VALUES(5,1328);
INSERT INTO playlist_track VALUES(5,1329);
INSERT INTO playlist_track VALUES(5,1330);
INSERT INTO playlist_track VALUES(5,1331);
INSERT INTO playlist_track VALUES(5,1332);
INSERT INTO playlist_track VALUES(5,1333);
INSERT INTO playlist_track VALUES(5,1334);
INSERT INTO playlist_track VALUES(5,1557);
INSERT INTO playlist_track VALUES(5,2506);
INSERT INTO playlist_track VALUES(5,2591);
INSERT INTO playlist_track VALUES(5,2592);
INSERT INTO playlist_track VALUES(5,2593);
INSERT INTO playlist_track VALUES(5,2594);
INSERT INTO playlist_track VALUES(5,2595);
INSERT INTO playlist_track VALUES(5,2596);
INSERT INTO playlist_track VALUES(5,2597);
INSERT INTO playlist_track VALUES(5,2598);
INSERT INTO playlist_track VALUES(5,2599);
INSERT INTO playlist_track VALUES(5,2600);
INSERT INTO playlist_track VALUES(5,2601);
INSERT INTO playlist_track VALUES(5,2602);
INSERT INTO playlist_track VALUES(5,2603);
INSERT INTO playlist_track VALUES(5,2604);
INSERT INTO playlist_track VALUES(5,2605);
INSERT INTO playlist_track VALUES(5,2606);
INSERT INTO playlist_track VALUES(5,2607);
INSERT INTO playlist_track VALUES(5,2608);
INSERT INTO playlist_track VALUES(5,2627);
INSERT INTO playlist_track VALUES(5,2631);
INSERT INTO playlist_track VALUES(5,2638);
INSERT INTO playlist_track VALUES(5,1158);
INSERT INTO playlist_track VALUES(5,1159);
INSERT INTO playlist_track VALUES(5,1160);
INSERT INTO playlist_track VALUES(5,1161);
INSERT INTO playlist_track VALUES(5,1162);
INSERT INTO playlist_track VALUES(5,1163);
INSERT INTO playlist_track VALUES(5,1164);
INSERT INTO playlist_track VALUES(5,1165);
INSERT INTO playlist_track VALUES(5,1166);
INSERT INTO playlist_track VALUES(5,1167);
INSERT INTO playlist_track VALUES(5,1168);
INSERT INTO playlist_track VALUES(5,1169);
INSERT INTO playlist_track VALUES(5,1170);
INSERT INTO playlist_track VALUES(5,1171);
INSERT INTO playlist_track VALUES(5,1172);
INSERT INTO playlist_track VALUES(5,1173);
INSERT INTO playlist_track VALUES(5,1174);
INSERT INTO playlist_track VALUES(5,1175);
INSERT INTO playlist_track VALUES(5,1176);
INSERT INTO playlist_track VALUES(5,1177);
INSERT INTO playlist_track VALUES(5,1178);
INSERT INTO playlist_track VALUES(5,1179);
INSERT INTO playlist_track VALUES(5,1180);
INSERT INTO playlist_track VALUES(5,1181);
INSERT INTO playlist_track VALUES(5,1182);
INSERT INTO playlist_track VALUES(5,1183);
INSERT INTO playlist_track VALUES(5,1184);
INSERT INTO playlist_track VALUES(5,1185);
INSERT INTO playlist_track VALUES(5,1186);
INSERT INTO playlist_track VALUES(5,1187);
INSERT INTO playlist_track VALUES(5,1414);
INSERT INTO playlist_track VALUES(5,1415);
INSERT INTO playlist_track VALUES(5,1416);
INSERT INTO playlist_track VALUES(5,1417);
INSERT INTO playlist_track VALUES(5,1418);
INSERT INTO playlist_track VALUES(5,1419);
INSERT INTO playlist_track VALUES(5,1420);
INSERT INTO playlist_track VALUES(5,1421);
INSERT INTO playlist_track VALUES(5,1422);
INSERT INTO playlist_track VALUES(5,1423);
INSERT INTO playlist_track VALUES(5,1424);
INSERT INTO playlist_track VALUES(5,1425);
INSERT INTO playlist_track VALUES(5,1426);
INSERT INTO playlist_track VALUES(5,1427);
INSERT INTO playlist_track VALUES(5,1428);
INSERT INTO playlist_track VALUES(5,1429);
INSERT INTO playlist_track VALUES(5,1430);
INSERT INTO playlist_track VALUES(5,1431);
INSERT INTO playlist_track VALUES(5,1432);
INSERT INTO playlist_track VALUES(5,1433);
INSERT INTO playlist_track VALUES(5,1801);
INSERT INTO playlist_track VALUES(5,1802);
INSERT INTO playlist_track VALUES(5,1803);
INSERT INTO playlist_track VALUES(5,1804);
INSERT INTO playlist_track VALUES(5,1805);
INSERT INTO playlist_track VALUES(5,1806);
INSERT INTO playlist_track VALUES(5,1807);
INSERT INTO playlist_track VALUES(5,1808);
INSERT INTO playlist_track VALUES(5,1809);
INSERT INTO playlist_track VALUES(5,1810);
INSERT INTO playlist_track VALUES(5,1811);
INSERT INTO playlist_track VALUES(5,1812);
INSERT INTO playlist_track VALUES(5,2003);
INSERT INTO playlist_track VALUES(5,2004);
INSERT INTO playlist_track VALUES(5,2005);
INSERT INTO playlist_track VALUES(5,2006);
INSERT INTO playlist_track VALUES(5,2007);
INSERT INTO playlist_track VALUES(5,2008);
INSERT INTO playlist_track VALUES(5,2009);
INSERT INTO playlist_track VALUES(5,2010);
INSERT INTO playlist_track VALUES(5,2011);
INSERT INTO playlist_track VALUES(5,2012);
INSERT INTO playlist_track VALUES(5,2013);
INSERT INTO playlist_track VALUES(5,2014);
INSERT INTO playlist_track VALUES(5,2193);
INSERT INTO playlist_track VALUES(5,2194);
INSERT INTO playlist_track VALUES(5,2195);
INSERT INTO playlist_track VALUES(5,2196);
INSERT INTO playlist_track VALUES(5,2197);
INSERT INTO playlist_track VALUES(5,2198);
INSERT INTO playlist_track VALUES(5,2199);
INSERT INTO playlist_track VALUES(5,2200);
INSERT INTO playlist_track VALUES(5,2201);
INSERT INTO playlist_track VALUES(5,2202);
INSERT INTO playlist_track VALUES(5,2203);
INSERT INTO playlist_track VALUES(5,424);
INSERT INTO playlist_track VALUES(5,428);
INSERT INTO playlist_track VALUES(5,430);
INSERT INTO playlist_track VALUES(5,434);
INSERT INTO playlist_track VALUES(5,2310);
INSERT INTO playlist_track VALUES(5,2311);
INSERT INTO playlist_track VALUES(5,2312);
INSERT INTO playlist_track VALUES(5,2313);
INSERT INTO playlist_track VALUES(5,2314);
INSERT INTO playlist_track VALUES(5,2315);
INSERT INTO playlist_track VALUES(5,2316);
INSERT INTO playlist_track VALUES(5,2317);
INSERT INTO playlist_track VALUES(5,2282);
INSERT INTO playlist_track VALUES(5,2283);
INSERT INTO playlist_track VALUES(5,2284);
INSERT INTO playlist_track VALUES(5,2358);
INSERT INTO playlist_track VALUES(5,2359);
INSERT INTO playlist_track VALUES(5,2360);
INSERT INTO playlist_track VALUES(5,2361);
INSERT INTO playlist_track VALUES(5,2362);
INSERT INTO playlist_track VALUES(5,2363);
INSERT INTO playlist_track VALUES(5,2364);
INSERT INTO playlist_track VALUES(5,2365);
INSERT INTO playlist_track VALUES(5,2366);
INSERT INTO playlist_track VALUES(5,2367);
INSERT INTO playlist_track VALUES(5,2368);
INSERT INTO playlist_track VALUES(5,2369);
INSERT INTO playlist_track VALUES(5,2370);
INSERT INTO playlist_track VALUES(5,2371);
INSERT INTO playlist_track VALUES(5,2372);
INSERT INTO playlist_track VALUES(5,2373);
INSERT INTO playlist_track VALUES(5,2374);
INSERT INTO playlist_track VALUES(5,2420);
INSERT INTO playlist_track VALUES(5,2421);
INSERT INTO playlist_track VALUES(5,2422);
INSERT INTO playlist_track VALUES(5,2423);
INSERT INTO playlist_track VALUES(5,2424);
INSERT INTO playlist_track VALUES(5,2425);
INSERT INTO playlist_track VALUES(5,2426);
INSERT INTO playlist_track VALUES(5,2427);
INSERT INTO playlist_track VALUES(5,2488);
INSERT INTO playlist_track VALUES(5,2489);
INSERT INTO playlist_track VALUES(5,2511);
INSERT INTO playlist_track VALUES(5,2512);
INSERT INTO playlist_track VALUES(5,2513);
INSERT INTO playlist_track VALUES(5,2711);
INSERT INTO playlist_track VALUES(5,2715);
INSERT INTO playlist_track VALUES(5,3365);
INSERT INTO playlist_track VALUES(5,3366);
INSERT INTO playlist_track VALUES(5,3367);
INSERT INTO playlist_track VALUES(5,3368);
INSERT INTO playlist_track VALUES(5,3369);
INSERT INTO playlist_track VALUES(5,3370);
INSERT INTO playlist_track VALUES(5,3371);
INSERT INTO playlist_track VALUES(5,3372);
INSERT INTO playlist_track VALUES(5,3373);
INSERT INTO playlist_track VALUES(5,3374);
INSERT INTO playlist_track VALUES(5,2926);
INSERT INTO playlist_track VALUES(5,2927);
INSERT INTO playlist_track VALUES(5,2928);
INSERT INTO playlist_track VALUES(5,2929);
INSERT INTO playlist_track VALUES(5,2930);
INSERT INTO playlist_track VALUES(5,2931);
INSERT INTO playlist_track VALUES(5,2932);
INSERT INTO playlist_track VALUES(5,2933);
INSERT INTO playlist_track VALUES(5,2934);
INSERT INTO playlist_track VALUES(5,2935);
INSERT INTO playlist_track VALUES(5,2936);
INSERT INTO playlist_track VALUES(5,2937);
INSERT INTO playlist_track VALUES(5,3075);
INSERT INTO playlist_track VALUES(5,3076);
INSERT INTO playlist_track VALUES(5,166);
INSERT INTO playlist_track VALUES(5,167);
INSERT INTO playlist_track VALUES(5,168);
INSERT INTO playlist_track VALUES(5,169);
INSERT INTO playlist_track VALUES(5,170);
INSERT INTO playlist_track VALUES(5,171);
INSERT INTO playlist_track VALUES(5,172);
INSERT INTO playlist_track VALUES(5,173);
INSERT INTO playlist_track VALUES(5,174);
INSERT INTO playlist_track VALUES(5,175);
INSERT INTO playlist_track VALUES(5,176);
INSERT INTO playlist_track VALUES(5,177);
INSERT INTO playlist_track VALUES(5,178);
INSERT INTO playlist_track VALUES(5,179);
INSERT INTO playlist_track VALUES(5,180);
INSERT INTO playlist_track VALUES(5,181);
INSERT INTO playlist_track VALUES(5,182);
INSERT INTO playlist_track VALUES(5,3426);
INSERT INTO playlist_track VALUES(5,2625);
INSERT INTO playlist_track VALUES(5,816);
INSERT INTO playlist_track VALUES(5,817);
INSERT INTO playlist_track VALUES(5,818);
INSERT INTO playlist_track VALUES(5,819);
INSERT INTO playlist_track VALUES(5,820);
INSERT INTO playlist_track VALUES(5,821);
INSERT INTO playlist_track VALUES(5,822);
INSERT INTO playlist_track VALUES(5,823);
INSERT INTO playlist_track VALUES(5,824);
INSERT INTO playlist_track VALUES(5,825);
INSERT INTO playlist_track VALUES(5,768);
INSERT INTO playlist_track VALUES(5,769);
INSERT INTO playlist_track VALUES(5,770);
INSERT INTO playlist_track VALUES(5,771);
INSERT INTO playlist_track VALUES(5,772);
INSERT INTO playlist_track VALUES(5,773);
INSERT INTO playlist_track VALUES(5,774);
INSERT INTO playlist_track VALUES(5,775);
INSERT INTO playlist_track VALUES(5,776);
INSERT INTO playlist_track VALUES(5,777);
INSERT INTO playlist_track VALUES(5,778);
INSERT INTO playlist_track VALUES(5,909);
INSERT INTO playlist_track VALUES(5,910);
INSERT INTO playlist_track VALUES(5,911);
INSERT INTO playlist_track VALUES(5,912);
INSERT INTO playlist_track VALUES(5,913);
INSERT INTO playlist_track VALUES(5,914);
INSERT INTO playlist_track VALUES(5,915);
INSERT INTO playlist_track VALUES(5,916);
INSERT INTO playlist_track VALUES(5,917);
INSERT INTO playlist_track VALUES(5,918);
INSERT INTO playlist_track VALUES(5,919);
INSERT INTO playlist_track VALUES(5,920);
INSERT INTO playlist_track VALUES(5,921);
INSERT INTO playlist_track VALUES(5,922);
INSERT INTO playlist_track VALUES(5,935);
INSERT INTO playlist_track VALUES(5,936);
INSERT INTO playlist_track VALUES(5,937);
INSERT INTO playlist_track VALUES(5,938);
INSERT INTO playlist_track VALUES(5,939);
INSERT INTO playlist_track VALUES(5,940);
INSERT INTO playlist_track VALUES(5,941);
INSERT INTO playlist_track VALUES(5,942);
INSERT INTO playlist_track VALUES(5,943);
INSERT INTO playlist_track VALUES(5,944);
INSERT INTO playlist_track VALUES(5,945);
INSERT INTO playlist_track VALUES(5,946);
INSERT INTO playlist_track VALUES(5,947);
INSERT INTO playlist_track VALUES(5,948);
INSERT INTO playlist_track VALUES(5,3301);
INSERT INTO playlist_track VALUES(5,3300);
INSERT INTO playlist_track VALUES(5,3302);
INSERT INTO playlist_track VALUES(5,3303);
INSERT INTO playlist_track VALUES(5,3304);
INSERT INTO playlist_track VALUES(5,3305);
INSERT INTO playlist_track VALUES(5,3306);
INSERT INTO playlist_track VALUES(5,3307);
INSERT INTO playlist_track VALUES(5,3308);
INSERT INTO playlist_track VALUES(5,3309);
INSERT INTO playlist_track VALUES(5,3310);
INSERT INTO playlist_track VALUES(5,3311);
INSERT INTO playlist_track VALUES(5,3312);
INSERT INTO playlist_track VALUES(5,3313);
INSERT INTO playlist_track VALUES(5,3314);
INSERT INTO playlist_track VALUES(5,3315);
INSERT INTO playlist_track VALUES(5,3316);
INSERT INTO playlist_track VALUES(5,3317);
INSERT INTO playlist_track VALUES(5,3318);
INSERT INTO playlist_track VALUES(5,1256);
INSERT INTO playlist_track VALUES(5,1257);
INSERT INTO playlist_track VALUES(5,1258);
INSERT INTO playlist_track VALUES(5,1259);
INSERT INTO playlist_track VALUES(5,1260);
INSERT INTO playlist_track VALUES(5,1261);
INSERT INTO playlist_track VALUES(5,1262);
INSERT INTO playlist_track VALUES(5,1263);
INSERT INTO playlist_track VALUES(5,1264);
INSERT INTO playlist_track VALUES(5,1265);
INSERT INTO playlist_track VALUES(5,1266);
INSERT INTO playlist_track VALUES(5,1267);
INSERT INTO playlist_track VALUES(5,2490);
INSERT INTO playlist_track VALUES(5,2542);
INSERT INTO playlist_track VALUES(5,2543);
INSERT INTO playlist_track VALUES(5,2544);
INSERT INTO playlist_track VALUES(5,2545);
INSERT INTO playlist_track VALUES(5,2546);
INSERT INTO playlist_track VALUES(5,2547);
INSERT INTO playlist_track VALUES(5,2548);
INSERT INTO playlist_track VALUES(5,2549);
INSERT INTO playlist_track VALUES(5,2550);
INSERT INTO playlist_track VALUES(5,2551);
INSERT INTO playlist_track VALUES(5,2552);
INSERT INTO playlist_track VALUES(5,2553);
INSERT INTO playlist_track VALUES(5,3411);
INSERT INTO playlist_track VALUES(5,3403);
INSERT INTO playlist_track VALUES(5,3423);
INSERT INTO playlist_track VALUES(5,1212);
INSERT INTO playlist_track VALUES(5,1213);
INSERT INTO playlist_track VALUES(5,1214);
INSERT INTO playlist_track VALUES(5,1215);
INSERT INTO playlist_track VALUES(5,1216);
INSERT INTO playlist_track VALUES(5,1217);
INSERT INTO playlist_track VALUES(5,1218);
INSERT INTO playlist_track VALUES(5,1219);
INSERT INTO playlist_track VALUES(5,1220);
INSERT INTO playlist_track VALUES(5,1221);
INSERT INTO playlist_track VALUES(5,1222);
INSERT INTO playlist_track VALUES(5,1223);
INSERT INTO playlist_track VALUES(5,1224);
INSERT INTO playlist_track VALUES(5,1225);
INSERT INTO playlist_track VALUES(5,1226);
INSERT INTO playlist_track VALUES(5,1227);
INSERT INTO playlist_track VALUES(5,1228);
INSERT INTO playlist_track VALUES(5,1229);
INSERT INTO playlist_track VALUES(5,1230);
INSERT INTO playlist_track VALUES(5,1231);
INSERT INTO playlist_track VALUES(5,1232);
INSERT INTO playlist_track VALUES(5,1233);
INSERT INTO playlist_track VALUES(5,1234);
INSERT INTO playlist_track VALUES(5,1434);
INSERT INTO playlist_track VALUES(5,1435);
INSERT INTO playlist_track VALUES(5,1436);
INSERT INTO playlist_track VALUES(5,1437);
INSERT INTO playlist_track VALUES(5,1438);
INSERT INTO playlist_track VALUES(5,1439);
INSERT INTO playlist_track VALUES(5,1440);
INSERT INTO playlist_track VALUES(5,1441);
INSERT INTO playlist_track VALUES(5,1442);
INSERT INTO playlist_track VALUES(5,1443);
INSERT INTO playlist_track VALUES(5,2204);
INSERT INTO playlist_track VALUES(5,2205);
INSERT INTO playlist_track VALUES(5,2206);
INSERT INTO playlist_track VALUES(5,2207);
INSERT INTO playlist_track VALUES(5,2208);
INSERT INTO playlist_track VALUES(5,2209);
INSERT INTO playlist_track VALUES(5,2210);
INSERT INTO playlist_track VALUES(5,2211);
INSERT INTO playlist_track VALUES(5,2212);
INSERT INTO playlist_track VALUES(5,2213);
INSERT INTO playlist_track VALUES(5,2214);
INSERT INTO playlist_track VALUES(5,2215);
INSERT INTO playlist_track VALUES(5,3404);
INSERT INTO playlist_track VALUES(5,2491);
INSERT INTO playlist_track VALUES(5,2492);
INSERT INTO playlist_track VALUES(5,2493);
INSERT INTO playlist_track VALUES(5,3028);
INSERT INTO playlist_track VALUES(5,3029);
INSERT INTO playlist_track VALUES(5,3030);
INSERT INTO playlist_track VALUES(5,3031);
INSERT INTO playlist_track VALUES(5,3032);
INSERT INTO playlist_track VALUES(5,3033);
INSERT INTO playlist_track VALUES(5,3034);
INSERT INTO playlist_track VALUES(5,3035);
INSERT INTO playlist_track VALUES(5,3036);
INSERT INTO playlist_track VALUES(5,3037);
INSERT INTO playlist_track VALUES(5,23);
INSERT INTO playlist_track VALUES(5,24);
INSERT INTO playlist_track VALUES(5,25);
INSERT INTO playlist_track VALUES(5,26);
INSERT INTO playlist_track VALUES(5,27);
INSERT INTO playlist_track VALUES(5,28);
INSERT INTO playlist_track VALUES(5,29);
INSERT INTO playlist_track VALUES(5,30);
INSERT INTO playlist_track VALUES(5,31);
INSERT INTO playlist_track VALUES(5,32);
INSERT INTO playlist_track VALUES(5,33);
INSERT INTO playlist_track VALUES(5,34);
INSERT INTO playlist_track VALUES(5,35);
INSERT INTO playlist_track VALUES(5,36);
INSERT INTO playlist_track VALUES(5,37);
INSERT INTO playlist_track VALUES(5,111);
INSERT INTO playlist_track VALUES(5,112);
INSERT INTO playlist_track VALUES(5,113);
INSERT INTO playlist_track VALUES(5,114);
INSERT INTO playlist_track VALUES(5,115);
INSERT INTO playlist_track VALUES(5,116);
INSERT INTO playlist_track VALUES(5,117);
INSERT INTO playlist_track VALUES(5,118);
INSERT INTO playlist_track VALUES(5,119);
INSERT INTO playlist_track VALUES(5,120);
INSERT INTO playlist_track VALUES(5,121);
INSERT INTO playlist_track VALUES(5,122);
INSERT INTO playlist_track VALUES(5,515);
INSERT INTO playlist_track VALUES(5,516);
INSERT INTO playlist_track VALUES(5,517);
INSERT INTO playlist_track VALUES(5,518);
INSERT INTO playlist_track VALUES(5,519);
INSERT INTO playlist_track VALUES(5,520);
INSERT INTO playlist_track VALUES(5,521);
INSERT INTO playlist_track VALUES(5,522);
INSERT INTO playlist_track VALUES(5,523);
INSERT INTO playlist_track VALUES(5,524);
INSERT INTO playlist_track VALUES(5,525);
INSERT INTO playlist_track VALUES(5,526);
INSERT INTO playlist_track VALUES(5,527);
INSERT INTO playlist_track VALUES(5,528);
INSERT INTO playlist_track VALUES(5,269);
INSERT INTO playlist_track VALUES(5,270);
INSERT INTO playlist_track VALUES(5,271);
INSERT INTO playlist_track VALUES(5,272);
INSERT INTO playlist_track VALUES(5,273);
INSERT INTO playlist_track VALUES(5,274);
INSERT INTO playlist_track VALUES(5,275);
INSERT INTO playlist_track VALUES(5,276);
INSERT INTO playlist_track VALUES(5,277);
INSERT INTO playlist_track VALUES(5,278);
INSERT INTO playlist_track VALUES(5,279);
INSERT INTO playlist_track VALUES(5,280);
INSERT INTO playlist_track VALUES(5,281);
INSERT INTO playlist_track VALUES(5,891);
INSERT INTO playlist_track VALUES(5,892);
INSERT INTO playlist_track VALUES(5,893);
INSERT INTO playlist_track VALUES(5,894);
INSERT INTO playlist_track VALUES(5,895);
INSERT INTO playlist_track VALUES(5,896);
INSERT INTO playlist_track VALUES(5,897);
INSERT INTO playlist_track VALUES(5,898);
INSERT INTO playlist_track VALUES(5,899);
INSERT INTO playlist_track VALUES(5,900);
INSERT INTO playlist_track VALUES(5,901);
INSERT INTO playlist_track VALUES(5,902);
INSERT INTO playlist_track VALUES(5,903);
INSERT INTO playlist_track VALUES(5,904);
INSERT INTO playlist_track VALUES(5,905);
INSERT INTO playlist_track VALUES(5,906);
INSERT INTO playlist_track VALUES(5,907);
INSERT INTO playlist_track VALUES(5,908);
INSERT INTO playlist_track VALUES(5,1105);
INSERT INTO playlist_track VALUES(5,1106);
INSERT INTO playlist_track VALUES(5,1107);
INSERT INTO playlist_track VALUES(5,1108);
INSERT INTO playlist_track VALUES(5,1109);
INSERT INTO playlist_track VALUES(5,1110);
INSERT INTO playlist_track VALUES(5,1111);
INSERT INTO playlist_track VALUES(5,1112);
INSERT INTO playlist_track VALUES(5,1113);
INSERT INTO playlist_track VALUES(5,1114);
INSERT INTO playlist_track VALUES(5,1115);
INSERT INTO playlist_track VALUES(5,1116);
INSERT INTO playlist_track VALUES(5,1117);
INSERT INTO playlist_track VALUES(5,1118);
INSERT INTO playlist_track VALUES(5,1119);
INSERT INTO playlist_track VALUES(5,1120);
INSERT INTO playlist_track VALUES(5,470);
INSERT INTO playlist_track VALUES(5,471);
INSERT INTO playlist_track VALUES(5,472);
INSERT INTO playlist_track VALUES(5,473);
INSERT INTO playlist_track VALUES(5,474);
INSERT INTO playlist_track VALUES(5,3424);
INSERT INTO playlist_track VALUES(5,2690);
INSERT INTO playlist_track VALUES(5,2691);
INSERT INTO playlist_track VALUES(5,2692);
INSERT INTO playlist_track VALUES(5,2693);
INSERT INTO playlist_track VALUES(5,2694);
INSERT INTO playlist_track VALUES(5,2695);
INSERT INTO playlist_track VALUES(5,2696);
INSERT INTO playlist_track VALUES(5,2697);
INSERT INTO playlist_track VALUES(5,2698);
INSERT INTO playlist_track VALUES(5,2699);
INSERT INTO playlist_track VALUES(5,2700);
INSERT INTO playlist_track VALUES(5,2701);
INSERT INTO playlist_track VALUES(5,2702);
INSERT INTO playlist_track VALUES(5,2703);
INSERT INTO playlist_track VALUES(5,2704);
INSERT INTO playlist_track VALUES(5,2494);
INSERT INTO playlist_track VALUES(5,2514);
INSERT INTO playlist_track VALUES(5,2515);
INSERT INTO playlist_track VALUES(5,2516);
INSERT INTO playlist_track VALUES(5,2517);
INSERT INTO playlist_track VALUES(5,3132);
INSERT INTO playlist_track VALUES(5,3133);
INSERT INTO playlist_track VALUES(5,3134);
INSERT INTO playlist_track VALUES(5,3135);
INSERT INTO playlist_track VALUES(5,3136);
INSERT INTO playlist_track VALUES(5,3137);
INSERT INTO playlist_track VALUES(5,3138);
INSERT INTO playlist_track VALUES(5,3139);
INSERT INTO playlist_track VALUES(5,3140);
INSERT INTO playlist_track VALUES(5,3141);
INSERT INTO playlist_track VALUES(5,3142);
INSERT INTO playlist_track VALUES(5,3143);
INSERT INTO playlist_track VALUES(5,3144);
INSERT INTO playlist_track VALUES(5,3145);
INSERT INTO playlist_track VALUES(5,3408);
INSERT INTO playlist_track VALUES(5,3);
INSERT INTO playlist_track VALUES(5,4);
INSERT INTO playlist_track VALUES(5,5);
INSERT INTO playlist_track VALUES(5,38);
INSERT INTO playlist_track VALUES(5,39);
INSERT INTO playlist_track VALUES(5,40);
INSERT INTO playlist_track VALUES(5,41);
INSERT INTO playlist_track VALUES(5,42);
INSERT INTO playlist_track VALUES(5,43);
INSERT INTO playlist_track VALUES(5,44);
INSERT INTO playlist_track VALUES(5,45);
INSERT INTO playlist_track VALUES(5,46);
INSERT INTO playlist_track VALUES(5,47);
INSERT INTO playlist_track VALUES(5,48);
INSERT INTO playlist_track VALUES(5,49);
INSERT INTO playlist_track VALUES(5,50);
INSERT INTO playlist_track VALUES(5,826);
INSERT INTO playlist_track VALUES(5,827);
INSERT INTO playlist_track VALUES(5,828);
INSERT INTO playlist_track VALUES(5,829);
INSERT INTO playlist_track VALUES(5,830);
INSERT INTO playlist_track VALUES(5,831);
INSERT INTO playlist_track VALUES(5,832);
INSERT INTO playlist_track VALUES(5,833);
INSERT INTO playlist_track VALUES(5,834);
INSERT INTO playlist_track VALUES(5,835);
INSERT INTO playlist_track VALUES(5,836);
INSERT INTO playlist_track VALUES(5,837);
INSERT INTO playlist_track VALUES(5,838);
INSERT INTO playlist_track VALUES(5,839);
INSERT INTO playlist_track VALUES(5,840);
INSERT INTO playlist_track VALUES(5,841);
INSERT INTO playlist_track VALUES(5,949);
INSERT INTO playlist_track VALUES(5,950);
INSERT INTO playlist_track VALUES(5,951);
INSERT INTO playlist_track VALUES(5,952);
INSERT INTO playlist_track VALUES(5,953);
INSERT INTO playlist_track VALUES(5,954);
INSERT INTO playlist_track VALUES(5,955);
INSERT INTO playlist_track VALUES(5,956);
INSERT INTO playlist_track VALUES(5,957);
INSERT INTO playlist_track VALUES(5,958);
INSERT INTO playlist_track VALUES(5,959);
INSERT INTO playlist_track VALUES(5,960);
INSERT INTO playlist_track VALUES(5,961);
INSERT INTO playlist_track VALUES(5,962);
INSERT INTO playlist_track VALUES(5,963);
INSERT INTO playlist_track VALUES(5,475);
INSERT INTO playlist_track VALUES(5,476);
INSERT INTO playlist_track VALUES(5,477);
INSERT INTO playlist_track VALUES(5,478);
INSERT INTO playlist_track VALUES(5,479);
INSERT INTO playlist_track VALUES(5,480);
INSERT INTO playlist_track VALUES(5,3354);
INSERT INTO playlist_track VALUES(5,3351);
INSERT INTO playlist_track VALUES(5,1395);
INSERT INTO playlist_track VALUES(5,1396);
INSERT INTO playlist_track VALUES(5,1397);
INSERT INTO playlist_track VALUES(5,1398);
INSERT INTO playlist_track VALUES(5,1399);
INSERT INTO playlist_track VALUES(5,1400);
INSERT INTO playlist_track VALUES(5,1401);
INSERT INTO playlist_track VALUES(5,1402);
INSERT INTO playlist_track VALUES(5,1403);
INSERT INTO playlist_track VALUES(5,1404);
INSERT INTO playlist_track VALUES(5,1405);
INSERT INTO playlist_track VALUES(5,1455);
INSERT INTO playlist_track VALUES(5,1456);
INSERT INTO playlist_track VALUES(5,1457);
INSERT INTO playlist_track VALUES(5,1458);
INSERT INTO playlist_track VALUES(5,1459);
INSERT INTO playlist_track VALUES(5,1460);
INSERT INTO playlist_track VALUES(5,1461);
INSERT INTO playlist_track VALUES(5,1462);
INSERT INTO playlist_track VALUES(5,1463);
INSERT INTO playlist_track VALUES(5,1464);
INSERT INTO playlist_track VALUES(5,1465);
INSERT INTO playlist_track VALUES(5,1520);
INSERT INTO playlist_track VALUES(5,1521);
INSERT INTO playlist_track VALUES(5,1522);
INSERT INTO playlist_track VALUES(5,1523);
INSERT INTO playlist_track VALUES(5,1524);
INSERT INTO playlist_track VALUES(5,1525);
INSERT INTO playlist_track VALUES(5,1526);
INSERT INTO playlist_track VALUES(5,1527);
INSERT INTO playlist_track VALUES(5,1528);
INSERT INTO playlist_track VALUES(5,1529);
INSERT INTO playlist_track VALUES(5,1530);
INSERT INTO playlist_track VALUES(5,1531);
INSERT INTO playlist_track VALUES(5,3276);
INSERT INTO playlist_track VALUES(5,3277);
INSERT INTO playlist_track VALUES(5,3278);
INSERT INTO playlist_track VALUES(5,3279);
INSERT INTO playlist_track VALUES(5,3280);
INSERT INTO playlist_track VALUES(5,3281);
INSERT INTO playlist_track VALUES(5,3282);
INSERT INTO playlist_track VALUES(5,3283);
INSERT INTO playlist_track VALUES(5,3284);
INSERT INTO playlist_track VALUES(5,3285);
INSERT INTO playlist_track VALUES(5,3286);
INSERT INTO playlist_track VALUES(5,3287);
INSERT INTO playlist_track VALUES(5,2125);
INSERT INTO playlist_track VALUES(5,2126);
INSERT INTO playlist_track VALUES(5,2127);
INSERT INTO playlist_track VALUES(5,2128);
INSERT INTO playlist_track VALUES(5,2129);
INSERT INTO playlist_track VALUES(5,2130);
INSERT INTO playlist_track VALUES(5,2131);
INSERT INTO playlist_track VALUES(5,2132);
INSERT INTO playlist_track VALUES(5,2133);
INSERT INTO playlist_track VALUES(5,2134);
INSERT INTO playlist_track VALUES(5,2135);
INSERT INTO playlist_track VALUES(5,2136);
INSERT INTO playlist_track VALUES(5,2137);
INSERT INTO playlist_track VALUES(5,2138);
INSERT INTO playlist_track VALUES(5,3410);
INSERT INTO playlist_track VALUES(5,2476);
INSERT INTO playlist_track VALUES(5,2484);
INSERT INTO playlist_track VALUES(5,2495);
INSERT INTO playlist_track VALUES(5,2496);
INSERT INTO playlist_track VALUES(5,2497);
INSERT INTO playlist_track VALUES(5,2498);
INSERT INTO playlist_track VALUES(5,2709);
INSERT INTO playlist_track VALUES(5,2710);
INSERT INTO playlist_track VALUES(5,2712);
INSERT INTO playlist_track VALUES(5,3038);
INSERT INTO playlist_track VALUES(5,3039);
INSERT INTO playlist_track VALUES(5,3040);
INSERT INTO playlist_track VALUES(5,3041);
INSERT INTO playlist_track VALUES(5,3042);
INSERT INTO playlist_track VALUES(5,3043);
INSERT INTO playlist_track VALUES(5,3044);
INSERT INTO playlist_track VALUES(5,3045);
INSERT INTO playlist_track VALUES(5,3046);
INSERT INTO playlist_track VALUES(5,3047);
INSERT INTO playlist_track VALUES(5,3048);
INSERT INTO playlist_track VALUES(5,3049);
INSERT INTO playlist_track VALUES(5,3050);
INSERT INTO playlist_track VALUES(5,3051);
INSERT INTO playlist_track VALUES(5,3077);
INSERT INTO playlist_track VALUES(5,77);
INSERT INTO playlist_track VALUES(5,78);
INSERT INTO playlist_track VALUES(5,79);
INSERT INTO playlist_track VALUES(5,80);
INSERT INTO playlist_track VALUES(5,81);
INSERT INTO playlist_track VALUES(5,82);
INSERT INTO playlist_track VALUES(5,83);
INSERT INTO playlist_track VALUES(5,84);
INSERT INTO playlist_track VALUES(5,3421);
INSERT INTO playlist_track VALUES(5,246);
INSERT INTO playlist_track VALUES(5,247);
INSERT INTO playlist_track VALUES(5,248);
INSERT INTO playlist_track VALUES(5,249);
INSERT INTO playlist_track VALUES(5,250);
INSERT INTO playlist_track VALUES(5,251);
INSERT INTO playlist_track VALUES(5,252);
INSERT INTO playlist_track VALUES(5,253);
INSERT INTO playlist_track VALUES(5,254);
INSERT INTO playlist_track VALUES(5,255);
INSERT INTO playlist_track VALUES(5,256);
INSERT INTO playlist_track VALUES(5,257);
INSERT INTO playlist_track VALUES(5,258);
INSERT INTO playlist_track VALUES(5,259);
INSERT INTO playlist_track VALUES(5,260);
INSERT INTO playlist_track VALUES(5,261);
INSERT INTO playlist_track VALUES(5,262);
INSERT INTO playlist_track VALUES(5,263);
INSERT INTO playlist_track VALUES(5,264);
INSERT INTO playlist_track VALUES(5,265);
INSERT INTO playlist_track VALUES(5,266);
INSERT INTO playlist_track VALUES(5,267);
INSERT INTO playlist_track VALUES(5,268);
INSERT INTO playlist_track VALUES(5,786);
INSERT INTO playlist_track VALUES(5,787);
INSERT INTO playlist_track VALUES(5,788);
INSERT INTO playlist_track VALUES(5,789);
INSERT INTO playlist_track VALUES(5,790);
INSERT INTO playlist_track VALUES(5,791);
INSERT INTO playlist_track VALUES(5,792);
INSERT INTO playlist_track VALUES(5,793);
INSERT INTO playlist_track VALUES(5,794);
INSERT INTO playlist_track VALUES(5,795);
INSERT INTO playlist_track VALUES(5,796);
INSERT INTO playlist_track VALUES(5,797);
INSERT INTO playlist_track VALUES(5,1562);
INSERT INTO playlist_track VALUES(5,1563);
INSERT INTO playlist_track VALUES(5,1564);
INSERT INTO playlist_track VALUES(5,1565);
INSERT INTO playlist_track VALUES(5,1566);
INSERT INTO playlist_track VALUES(5,1567);
INSERT INTO playlist_track VALUES(5,1568);
INSERT INTO playlist_track VALUES(5,1569);
INSERT INTO playlist_track VALUES(5,1570);
INSERT INTO playlist_track VALUES(5,1571);
INSERT INTO playlist_track VALUES(5,1572);
INSERT INTO playlist_track VALUES(5,1573);
INSERT INTO playlist_track VALUES(5,1574);
INSERT INTO playlist_track VALUES(5,1575);
INSERT INTO playlist_track VALUES(5,1576);
INSERT INTO playlist_track VALUES(5,1839);
INSERT INTO playlist_track VALUES(5,1840);
INSERT INTO playlist_track VALUES(5,1841);
INSERT INTO playlist_track VALUES(5,1842);
INSERT INTO playlist_track VALUES(5,1843);
INSERT INTO playlist_track VALUES(5,1844);
INSERT INTO playlist_track VALUES(5,1845);
INSERT INTO playlist_track VALUES(5,1846);
INSERT INTO playlist_track VALUES(5,1847);
INSERT INTO playlist_track VALUES(5,1848);
INSERT INTO playlist_track VALUES(5,1849);
INSERT INTO playlist_track VALUES(5,1850);
INSERT INTO playlist_track VALUES(5,1851);
INSERT INTO playlist_track VALUES(5,1852);
INSERT INTO playlist_track VALUES(5,1986);
INSERT INTO playlist_track VALUES(5,1987);
INSERT INTO playlist_track VALUES(5,1988);
INSERT INTO playlist_track VALUES(5,1989);
INSERT INTO playlist_track VALUES(5,1990);
INSERT INTO playlist_track VALUES(5,1991);
INSERT INTO playlist_track VALUES(5,1992);
INSERT INTO playlist_track VALUES(5,1993);
INSERT INTO playlist_track VALUES(5,1994);
INSERT INTO playlist_track VALUES(5,1995);
INSERT INTO playlist_track VALUES(5,1996);
INSERT INTO playlist_track VALUES(5,1997);
INSERT INTO playlist_track VALUES(5,1998);
INSERT INTO playlist_track VALUES(5,1999);
INSERT INTO playlist_track VALUES(5,2000);
INSERT INTO playlist_track VALUES(5,2001);
INSERT INTO playlist_track VALUES(5,2002);
INSERT INTO playlist_track VALUES(5,3415);
INSERT INTO playlist_track VALUES(5,2650);
INSERT INTO playlist_track VALUES(5,2651);
INSERT INTO playlist_track VALUES(5,2652);
INSERT INTO playlist_track VALUES(5,2653);
INSERT INTO playlist_track VALUES(5,2654);
INSERT INTO playlist_track VALUES(5,2655);
INSERT INTO playlist_track VALUES(5,2656);
INSERT INTO playlist_track VALUES(5,2657);
INSERT INTO playlist_track VALUES(5,2658);
INSERT INTO playlist_track VALUES(5,2659);
INSERT INTO playlist_track VALUES(5,2660);
INSERT INTO playlist_track VALUES(5,2661);
INSERT INTO playlist_track VALUES(5,2662);
INSERT INTO playlist_track VALUES(5,2663);
INSERT INTO playlist_track VALUES(5,2296);
INSERT INTO playlist_track VALUES(5,2297);
INSERT INTO playlist_track VALUES(5,2298);
INSERT INTO playlist_track VALUES(5,2299);
INSERT INTO playlist_track VALUES(5,2300);
INSERT INTO playlist_track VALUES(5,2301);
INSERT INTO playlist_track VALUES(5,2302);
INSERT INTO playlist_track VALUES(5,2303);
INSERT INTO playlist_track VALUES(5,2304);
INSERT INTO playlist_track VALUES(5,2305);
INSERT INTO playlist_track VALUES(5,2306);
INSERT INTO playlist_track VALUES(5,2307);
INSERT INTO playlist_track VALUES(5,2308);
INSERT INTO playlist_track VALUES(5,2309);
INSERT INTO playlist_track VALUES(5,2334);
INSERT INTO playlist_track VALUES(5,2335);
INSERT INTO playlist_track VALUES(5,2336);
INSERT INTO playlist_track VALUES(5,2337);
INSERT INTO playlist_track VALUES(5,2338);
INSERT INTO playlist_track VALUES(5,2339);
INSERT INTO playlist_track VALUES(5,2340);
INSERT INTO playlist_track VALUES(5,2341);
INSERT INTO playlist_track VALUES(5,2342);
INSERT INTO playlist_track VALUES(5,2343);
INSERT INTO playlist_track VALUES(5,2434);
INSERT INTO playlist_track VALUES(5,2435);
INSERT INTO playlist_track VALUES(5,2436);
INSERT INTO playlist_track VALUES(5,2437);
INSERT INTO playlist_track VALUES(5,2438);
INSERT INTO playlist_track VALUES(5,2439);
INSERT INTO playlist_track VALUES(5,2440);
INSERT INTO playlist_track VALUES(5,2441);
INSERT INTO playlist_track VALUES(5,2442);
INSERT INTO playlist_track VALUES(5,2443);
INSERT INTO playlist_track VALUES(5,2444);
INSERT INTO playlist_track VALUES(5,2445);
INSERT INTO playlist_track VALUES(5,2446);
INSERT INTO playlist_track VALUES(5,2447);
INSERT INTO playlist_track VALUES(5,2448);
INSERT INTO playlist_track VALUES(5,2461);
INSERT INTO playlist_track VALUES(5,2462);
INSERT INTO playlist_track VALUES(5,2463);
INSERT INTO playlist_track VALUES(5,2464);
INSERT INTO playlist_track VALUES(5,2465);
INSERT INTO playlist_track VALUES(5,2466);
INSERT INTO playlist_track VALUES(5,2467);
INSERT INTO playlist_track VALUES(5,2468);
INSERT INTO playlist_track VALUES(5,2469);
INSERT INTO playlist_track VALUES(5,2470);
INSERT INTO playlist_track VALUES(5,2471);
INSERT INTO playlist_track VALUES(5,2478);
INSERT INTO playlist_track VALUES(5,2518);
INSERT INTO playlist_track VALUES(5,2519);
INSERT INTO playlist_track VALUES(5,2520);
INSERT INTO playlist_track VALUES(5,2521);
INSERT INTO playlist_track VALUES(5,2522);
INSERT INTO playlist_track VALUES(5,456);
INSERT INTO playlist_track VALUES(5,457);
INSERT INTO playlist_track VALUES(5,458);
INSERT INTO playlist_track VALUES(5,459);
INSERT INTO playlist_track VALUES(5,460);
INSERT INTO playlist_track VALUES(5,461);
INSERT INTO playlist_track VALUES(5,462);
INSERT INTO playlist_track VALUES(5,463);
INSERT INTO playlist_track VALUES(5,464);
INSERT INTO playlist_track VALUES(5,465);
INSERT INTO playlist_track VALUES(5,466);
INSERT INTO playlist_track VALUES(5,467);
INSERT INTO playlist_track VALUES(5,3078);
INSERT INTO playlist_track VALUES(5,3079);
INSERT INTO playlist_track VALUES(5,3080);
INSERT INTO playlist_track VALUES(5,3416);
INSERT INTO playlist_track VALUES(5,923);
INSERT INTO playlist_track VALUES(5,924);
INSERT INTO playlist_track VALUES(5,925);
INSERT INTO playlist_track VALUES(5,926);
INSERT INTO playlist_track VALUES(5,927);
INSERT INTO playlist_track VALUES(5,928);
INSERT INTO playlist_track VALUES(5,929);
INSERT INTO playlist_track VALUES(5,930);
INSERT INTO playlist_track VALUES(5,931);
INSERT INTO playlist_track VALUES(5,932);
INSERT INTO playlist_track VALUES(5,933);
INSERT INTO playlist_track VALUES(5,934);
INSERT INTO playlist_track VALUES(5,1020);
INSERT INTO playlist_track VALUES(5,1021);
INSERT INTO playlist_track VALUES(5,1022);
INSERT INTO playlist_track VALUES(5,1023);
INSERT INTO playlist_track VALUES(5,1024);
INSERT INTO playlist_track VALUES(5,1025);
INSERT INTO playlist_track VALUES(5,1026);
INSERT INTO playlist_track VALUES(5,1027);
INSERT INTO playlist_track VALUES(5,1028);
INSERT INTO playlist_track VALUES(5,1029);
INSERT INTO playlist_track VALUES(5,1030);
INSERT INTO playlist_track VALUES(5,1031);
INSERT INTO playlist_track VALUES(5,1032);
INSERT INTO playlist_track VALUES(5,481);
INSERT INTO playlist_track VALUES(5,482);
INSERT INTO playlist_track VALUES(5,483);
INSERT INTO playlist_track VALUES(5,484);
INSERT INTO playlist_track VALUES(5,1188);
INSERT INTO playlist_track VALUES(5,1189);
INSERT INTO playlist_track VALUES(5,1190);
INSERT INTO playlist_track VALUES(5,1191);
INSERT INTO playlist_track VALUES(5,1192);
INSERT INTO playlist_track VALUES(5,1193);
INSERT INTO playlist_track VALUES(5,1194);
INSERT INTO playlist_track VALUES(5,1195);
INSERT INTO playlist_track VALUES(5,1196);
INSERT INTO playlist_track VALUES(5,1197);
INSERT INTO playlist_track VALUES(5,1198);
INSERT INTO playlist_track VALUES(5,1199);
INSERT INTO playlist_track VALUES(5,1200);
INSERT INTO playlist_track VALUES(5,436);
INSERT INTO playlist_track VALUES(5,437);
INSERT INTO playlist_track VALUES(5,438);
INSERT INTO playlist_track VALUES(5,439);
INSERT INTO playlist_track VALUES(5,440);
INSERT INTO playlist_track VALUES(5,441);
INSERT INTO playlist_track VALUES(5,442);
INSERT INTO playlist_track VALUES(5,443);
INSERT INTO playlist_track VALUES(5,444);
INSERT INTO playlist_track VALUES(5,445);
INSERT INTO playlist_track VALUES(5,446);
INSERT INTO playlist_track VALUES(5,447);
INSERT INTO playlist_track VALUES(5,448);
INSERT INTO playlist_track VALUES(5,449);
INSERT INTO playlist_track VALUES(5,450);
INSERT INTO playlist_track VALUES(5,451);
INSERT INTO playlist_track VALUES(5,453);
INSERT INTO playlist_track VALUES(5,454);
INSERT INTO playlist_track VALUES(5,455);
INSERT INTO playlist_track VALUES(5,337);
INSERT INTO playlist_track VALUES(5,338);
INSERT INTO playlist_track VALUES(5,339);
INSERT INTO playlist_track VALUES(5,340);
INSERT INTO playlist_track VALUES(5,341);
INSERT INTO playlist_track VALUES(5,342);
INSERT INTO playlist_track VALUES(5,343);
INSERT INTO playlist_track VALUES(5,344);
INSERT INTO playlist_track VALUES(5,345);
INSERT INTO playlist_track VALUES(5,346);
INSERT INTO playlist_track VALUES(5,347);
INSERT INTO playlist_track VALUES(5,348);
INSERT INTO playlist_track VALUES(5,349);
INSERT INTO playlist_track VALUES(5,350);
INSERT INTO playlist_track VALUES(5,1577);
INSERT INTO playlist_track VALUES(5,1578);
INSERT INTO playlist_track VALUES(5,1579);
INSERT INTO playlist_track VALUES(5,1580);
INSERT INTO playlist_track VALUES(5,1581);
INSERT INTO playlist_track VALUES(5,1582);
INSERT INTO playlist_track VALUES(5,1583);
INSERT INTO playlist_track VALUES(5,1584);
INSERT INTO playlist_track VALUES(5,1585);
INSERT INTO playlist_track VALUES(5,1586);
INSERT INTO playlist_track VALUES(5,1861);
INSERT INTO playlist_track VALUES(5,1862);
INSERT INTO playlist_track VALUES(5,1863);
INSERT INTO playlist_track VALUES(5,1864);
INSERT INTO playlist_track VALUES(5,1865);
INSERT INTO playlist_track VALUES(5,1866);
INSERT INTO playlist_track VALUES(5,1867);
INSERT INTO playlist_track VALUES(5,1868);
INSERT INTO playlist_track VALUES(5,1869);
INSERT INTO playlist_track VALUES(5,1870);
INSERT INTO playlist_track VALUES(5,1871);
INSERT INTO playlist_track VALUES(5,1872);
INSERT INTO playlist_track VALUES(5,1873);
INSERT INTO playlist_track VALUES(5,3359);
INSERT INTO playlist_track VALUES(5,2406);
INSERT INTO playlist_track VALUES(5,2407);
INSERT INTO playlist_track VALUES(5,2408);
INSERT INTO playlist_track VALUES(5,2409);
INSERT INTO playlist_track VALUES(5,2410);
INSERT INTO playlist_track VALUES(5,2411);
INSERT INTO playlist_track VALUES(5,2412);
INSERT INTO playlist_track VALUES(5,2413);
INSERT INTO playlist_track VALUES(5,2414);
INSERT INTO playlist_track VALUES(5,2415);
INSERT INTO playlist_track VALUES(5,2416);
INSERT INTO playlist_track VALUES(5,2417);
INSERT INTO playlist_track VALUES(5,2418);
INSERT INTO playlist_track VALUES(5,2419);
INSERT INTO playlist_track VALUES(5,2499);
INSERT INTO playlist_track VALUES(5,2706);
INSERT INTO playlist_track VALUES(5,2708);
INSERT INTO playlist_track VALUES(5,2713);
INSERT INTO playlist_track VALUES(5,2716);
INSERT INTO playlist_track VALUES(5,2720);
INSERT INTO playlist_track VALUES(5,2721);
INSERT INTO playlist_track VALUES(5,2722);
INSERT INTO playlist_track VALUES(5,2723);
INSERT INTO playlist_track VALUES(5,2724);
INSERT INTO playlist_track VALUES(5,2725);
INSERT INTO playlist_track VALUES(5,2726);
INSERT INTO playlist_track VALUES(5,2727);
INSERT INTO playlist_track VALUES(5,2728);
INSERT INTO playlist_track VALUES(5,2729);
INSERT INTO playlist_track VALUES(5,2730);
INSERT INTO playlist_track VALUES(5,2565);
INSERT INTO playlist_track VALUES(5,2566);
INSERT INTO playlist_track VALUES(5,2567);
INSERT INTO playlist_track VALUES(5,2568);
INSERT INTO playlist_track VALUES(5,2569);
INSERT INTO playlist_track VALUES(5,2570);
INSERT INTO playlist_track VALUES(5,2571);
INSERT INTO playlist_track VALUES(5,2781);
INSERT INTO playlist_track VALUES(5,2782);
INSERT INTO playlist_track VALUES(5,2783);
INSERT INTO playlist_track VALUES(5,2784);
INSERT INTO playlist_track VALUES(5,2785);
INSERT INTO playlist_track VALUES(5,2786);
INSERT INTO playlist_track VALUES(5,2787);
INSERT INTO playlist_track VALUES(5,2788);
INSERT INTO playlist_track VALUES(5,2789);
INSERT INTO playlist_track VALUES(5,2790);
INSERT INTO playlist_track VALUES(5,2791);
INSERT INTO playlist_track VALUES(5,2792);
INSERT INTO playlist_track VALUES(5,2793);
INSERT INTO playlist_track VALUES(5,2794);
INSERT INTO playlist_track VALUES(5,2795);
INSERT INTO playlist_track VALUES(5,2796);
INSERT INTO playlist_track VALUES(5,2797);
INSERT INTO playlist_track VALUES(5,2798);
INSERT INTO playlist_track VALUES(5,2799);
INSERT INTO playlist_track VALUES(5,2800);
INSERT INTO playlist_track VALUES(5,2801);
INSERT INTO playlist_track VALUES(5,2802);
INSERT INTO playlist_track VALUES(5,2975);
INSERT INTO playlist_track VALUES(5,2976);
INSERT INTO playlist_track VALUES(5,2977);
INSERT INTO playlist_track VALUES(5,2978);
INSERT INTO playlist_track VALUES(5,2979);
INSERT INTO playlist_track VALUES(5,2980);
INSERT INTO playlist_track VALUES(5,2981);
INSERT INTO playlist_track VALUES(5,2982);
INSERT INTO playlist_track VALUES(5,2983);
INSERT INTO playlist_track VALUES(5,2984);
INSERT INTO playlist_track VALUES(5,2985);
INSERT INTO playlist_track VALUES(5,2986);
INSERT INTO playlist_track VALUES(5,183);
INSERT INTO playlist_track VALUES(5,184);
INSERT INTO playlist_track VALUES(5,185);
INSERT INTO playlist_track VALUES(5,186);
INSERT INTO playlist_track VALUES(5,187);
INSERT INTO playlist_track VALUES(5,188);
INSERT INTO playlist_track VALUES(5,189);
INSERT INTO playlist_track VALUES(5,190);
INSERT INTO playlist_track VALUES(5,191);
INSERT INTO playlist_track VALUES(5,192);
INSERT INTO playlist_track VALUES(5,193);
INSERT INTO playlist_track VALUES(5,205);
INSERT INTO playlist_track VALUES(5,206);
INSERT INTO playlist_track VALUES(5,207);
INSERT INTO playlist_track VALUES(5,208);
INSERT INTO playlist_track VALUES(5,209);
INSERT INTO playlist_track VALUES(5,210);
INSERT INTO playlist_track VALUES(5,211);
INSERT INTO playlist_track VALUES(5,212);
INSERT INTO playlist_track VALUES(5,213);
INSERT INTO playlist_track VALUES(5,214);
INSERT INTO playlist_track VALUES(5,215);
INSERT INTO playlist_track VALUES(5,216);
INSERT INTO playlist_track VALUES(5,217);
INSERT INTO playlist_track VALUES(5,218);
INSERT INTO playlist_track VALUES(5,219);
INSERT INTO playlist_track VALUES(5,220);
INSERT INTO playlist_track VALUES(5,221);
INSERT INTO playlist_track VALUES(5,222);
INSERT INTO playlist_track VALUES(5,3417);
INSERT INTO playlist_track VALUES(5,583);
INSERT INTO playlist_track VALUES(5,584);
INSERT INTO playlist_track VALUES(5,585);
INSERT INTO playlist_track VALUES(5,586);
INSERT INTO playlist_track VALUES(5,587);
INSERT INTO playlist_track VALUES(5,588);
INSERT INTO playlist_track VALUES(5,589);
INSERT INTO playlist_track VALUES(5,590);
INSERT INTO playlist_track VALUES(5,591);
INSERT INTO playlist_track VALUES(5,592);
INSERT INTO playlist_track VALUES(5,593);
INSERT INTO playlist_track VALUES(5,594);
INSERT INTO playlist_track VALUES(5,595);
INSERT INTO playlist_track VALUES(5,596);
INSERT INTO playlist_track VALUES(5,976);
INSERT INTO playlist_track VALUES(5,977);
INSERT INTO playlist_track VALUES(5,978);
INSERT INTO playlist_track VALUES(5,979);
INSERT INTO playlist_track VALUES(5,984);
INSERT INTO playlist_track VALUES(5,1087);
INSERT INTO playlist_track VALUES(5,1088);
INSERT INTO playlist_track VALUES(5,1089);
INSERT INTO playlist_track VALUES(5,1090);
INSERT INTO playlist_track VALUES(5,1091);
INSERT INTO playlist_track VALUES(5,1092);
INSERT INTO playlist_track VALUES(5,1093);
INSERT INTO playlist_track VALUES(5,1094);
INSERT INTO playlist_track VALUES(5,1095);
INSERT INTO playlist_track VALUES(5,1096);
INSERT INTO playlist_track VALUES(5,1097);
INSERT INTO playlist_track VALUES(5,1098);
INSERT INTO playlist_track VALUES(5,1099);
INSERT INTO playlist_track VALUES(5,1100);
INSERT INTO playlist_track VALUES(5,1101);
INSERT INTO playlist_track VALUES(5,1305);
INSERT INTO playlist_track VALUES(5,1306);
INSERT INTO playlist_track VALUES(5,1307);
INSERT INTO playlist_track VALUES(5,1308);
INSERT INTO playlist_track VALUES(5,1309);
INSERT INTO playlist_track VALUES(5,1310);
INSERT INTO playlist_track VALUES(5,1311);
INSERT INTO playlist_track VALUES(5,1312);
INSERT INTO playlist_track VALUES(5,1313);
INSERT INTO playlist_track VALUES(5,1314);
INSERT INTO playlist_track VALUES(5,1315);
INSERT INTO playlist_track VALUES(5,1316);
INSERT INTO playlist_track VALUES(5,1317);
INSERT INTO playlist_track VALUES(5,1318);
INSERT INTO playlist_track VALUES(5,1319);
INSERT INTO playlist_track VALUES(5,1320);
INSERT INTO playlist_track VALUES(5,1321);
INSERT INTO playlist_track VALUES(5,1322);
INSERT INTO playlist_track VALUES(5,1323);
INSERT INTO playlist_track VALUES(5,1324);
INSERT INTO playlist_track VALUES(5,1406);
INSERT INTO playlist_track VALUES(5,1407);
INSERT INTO playlist_track VALUES(5,1408);
INSERT INTO playlist_track VALUES(5,1409);
INSERT INTO playlist_track VALUES(5,1410);
INSERT INTO playlist_track VALUES(5,1411);
INSERT INTO playlist_track VALUES(5,1412);
INSERT INTO playlist_track VALUES(5,1413);
INSERT INTO playlist_track VALUES(5,1686);
INSERT INTO playlist_track VALUES(5,1687);
INSERT INTO playlist_track VALUES(5,1688);
INSERT INTO playlist_track VALUES(5,1689);
INSERT INTO playlist_track VALUES(5,1690);
INSERT INTO playlist_track VALUES(5,1691);
INSERT INTO playlist_track VALUES(5,1692);
INSERT INTO playlist_track VALUES(5,1693);
INSERT INTO playlist_track VALUES(5,1694);
INSERT INTO playlist_track VALUES(5,1695);
INSERT INTO playlist_track VALUES(5,1696);
INSERT INTO playlist_track VALUES(5,1697);
INSERT INTO playlist_track VALUES(5,1698);
INSERT INTO playlist_track VALUES(5,1699);
INSERT INTO playlist_track VALUES(5,1700);
INSERT INTO playlist_track VALUES(5,1701);
INSERT INTO playlist_track VALUES(5,408);
INSERT INTO playlist_track VALUES(5,409);
INSERT INTO playlist_track VALUES(5,410);
INSERT INTO playlist_track VALUES(5,411);
INSERT INTO playlist_track VALUES(5,412);
INSERT INTO playlist_track VALUES(5,413);
INSERT INTO playlist_track VALUES(5,414);
INSERT INTO playlist_track VALUES(5,415);
INSERT INTO playlist_track VALUES(5,416);
INSERT INTO playlist_track VALUES(5,417);
INSERT INTO playlist_track VALUES(5,418);
INSERT INTO playlist_track VALUES(5,1813);
INSERT INTO playlist_track VALUES(5,1814);
INSERT INTO playlist_track VALUES(5,1815);
INSERT INTO playlist_track VALUES(5,1816);
INSERT INTO playlist_track VALUES(5,1817);
INSERT INTO playlist_track VALUES(5,1818);
INSERT INTO playlist_track VALUES(5,1819);
INSERT INTO playlist_track VALUES(5,1820);
INSERT INTO playlist_track VALUES(5,1821);
INSERT INTO playlist_track VALUES(5,1822);
INSERT INTO playlist_track VALUES(5,1823);
INSERT INTO playlist_track VALUES(5,1824);
INSERT INTO playlist_track VALUES(5,1825);
INSERT INTO playlist_track VALUES(5,1826);
INSERT INTO playlist_track VALUES(5,1827);
INSERT INTO playlist_track VALUES(5,1828);
INSERT INTO playlist_track VALUES(5,1969);
INSERT INTO playlist_track VALUES(5,1970);
INSERT INTO playlist_track VALUES(5,1971);
INSERT INTO playlist_track VALUES(5,1972);
INSERT INTO playlist_track VALUES(5,1973);
INSERT INTO playlist_track VALUES(5,1974);
INSERT INTO playlist_track VALUES(5,1975);
INSERT INTO playlist_track VALUES(5,1976);
INSERT INTO playlist_track VALUES(5,1977);
INSERT INTO playlist_track VALUES(5,1978);
INSERT INTO playlist_track VALUES(5,1979);
INSERT INTO playlist_track VALUES(5,1980);
INSERT INTO playlist_track VALUES(5,1981);
INSERT INTO playlist_track VALUES(5,1982);
INSERT INTO playlist_track VALUES(5,1983);
INSERT INTO playlist_track VALUES(5,1984);
INSERT INTO playlist_track VALUES(5,1985);
INSERT INTO playlist_track VALUES(5,2113);
INSERT INTO playlist_track VALUES(5,2114);
INSERT INTO playlist_track VALUES(5,2115);
INSERT INTO playlist_track VALUES(5,2116);
INSERT INTO playlist_track VALUES(5,2117);
INSERT INTO playlist_track VALUES(5,2118);
INSERT INTO playlist_track VALUES(5,2119);
INSERT INTO playlist_track VALUES(5,2120);
INSERT INTO playlist_track VALUES(5,2121);
INSERT INTO playlist_track VALUES(5,2122);
INSERT INTO playlist_track VALUES(5,2123);
INSERT INTO playlist_track VALUES(5,2124);
INSERT INTO playlist_track VALUES(5,2149);
INSERT INTO playlist_track VALUES(5,2150);
INSERT INTO playlist_track VALUES(5,2151);
INSERT INTO playlist_track VALUES(5,2152);
INSERT INTO playlist_track VALUES(5,2153);
INSERT INTO playlist_track VALUES(5,2154);
INSERT INTO playlist_track VALUES(5,2155);
INSERT INTO playlist_track VALUES(5,2156);
INSERT INTO playlist_track VALUES(5,2157);
INSERT INTO playlist_track VALUES(5,2158);
INSERT INTO playlist_track VALUES(5,2159);
INSERT INTO playlist_track VALUES(5,2160);
INSERT INTO playlist_track VALUES(5,2161);
INSERT INTO playlist_track VALUES(5,2162);
INSERT INTO playlist_track VALUES(5,2163);
INSERT INTO playlist_track VALUES(5,2164);
INSERT INTO playlist_track VALUES(5,2676);
INSERT INTO playlist_track VALUES(5,2677);
INSERT INTO playlist_track VALUES(5,2678);
INSERT INTO playlist_track VALUES(5,2679);
INSERT INTO playlist_track VALUES(5,2680);
INSERT INTO playlist_track VALUES(5,2681);
INSERT INTO playlist_track VALUES(5,2682);
INSERT INTO playlist_track VALUES(5,2683);
INSERT INTO playlist_track VALUES(5,2684);
INSERT INTO playlist_track VALUES(5,2685);
INSERT INTO playlist_track VALUES(5,2686);
INSERT INTO playlist_track VALUES(5,2687);
INSERT INTO playlist_track VALUES(5,2688);
INSERT INTO playlist_track VALUES(5,2689);
INSERT INTO playlist_track VALUES(5,3418);
INSERT INTO playlist_track VALUES(5,2500);
INSERT INTO playlist_track VALUES(5,2501);
INSERT INTO playlist_track VALUES(5,2803);
INSERT INTO playlist_track VALUES(5,2804);
INSERT INTO playlist_track VALUES(5,2805);
INSERT INTO playlist_track VALUES(5,2806);
INSERT INTO playlist_track VALUES(5,2807);
INSERT INTO playlist_track VALUES(5,2808);
INSERT INTO playlist_track VALUES(5,2809);
INSERT INTO playlist_track VALUES(5,2810);
INSERT INTO playlist_track VALUES(5,2811);
INSERT INTO playlist_track VALUES(5,2812);
INSERT INTO playlist_track VALUES(5,2813);
INSERT INTO playlist_track VALUES(5,2814);
INSERT INTO playlist_track VALUES(5,2815);
INSERT INTO playlist_track VALUES(5,2816);
INSERT INTO playlist_track VALUES(5,2817);
INSERT INTO playlist_track VALUES(5,2818);
INSERT INTO playlist_track VALUES(5,2949);
INSERT INTO playlist_track VALUES(5,2950);
INSERT INTO playlist_track VALUES(5,2951);
INSERT INTO playlist_track VALUES(5,2952);
INSERT INTO playlist_track VALUES(5,2953);
INSERT INTO playlist_track VALUES(5,2954);
INSERT INTO playlist_track VALUES(5,2955);
INSERT INTO playlist_track VALUES(5,2956);
INSERT INTO playlist_track VALUES(5,2957);
INSERT INTO playlist_track VALUES(5,2958);
INSERT INTO playlist_track VALUES(5,2959);
INSERT INTO playlist_track VALUES(5,2960);
INSERT INTO playlist_track VALUES(5,2961);
INSERT INTO playlist_track VALUES(5,2962);
INSERT INTO playlist_track VALUES(5,2963);
INSERT INTO playlist_track VALUES(5,3004);
INSERT INTO playlist_track VALUES(5,3005);
INSERT INTO playlist_track VALUES(5,3006);
INSERT INTO playlist_track VALUES(5,3007);
INSERT INTO playlist_track VALUES(5,3008);
INSERT INTO playlist_track VALUES(5,3009);
INSERT INTO playlist_track VALUES(5,3010);
INSERT INTO playlist_track VALUES(5,3011);
INSERT INTO playlist_track VALUES(5,3012);
INSERT INTO playlist_track VALUES(5,3013);
INSERT INTO playlist_track VALUES(5,3014);
INSERT INTO playlist_track VALUES(5,3015);
INSERT INTO playlist_track VALUES(5,3016);
INSERT INTO playlist_track VALUES(5,3017);
INSERT INTO playlist_track VALUES(5,3092);
INSERT INTO playlist_track VALUES(5,3093);
INSERT INTO playlist_track VALUES(5,3094);
INSERT INTO playlist_track VALUES(5,3095);
INSERT INTO playlist_track VALUES(5,3096);
INSERT INTO playlist_track VALUES(5,3097);
INSERT INTO playlist_track VALUES(5,3098);
INSERT INTO playlist_track VALUES(5,3099);
INSERT INTO playlist_track VALUES(5,3100);
INSERT INTO playlist_track VALUES(5,3101);
INSERT INTO playlist_track VALUES(5,3102);
INSERT INTO playlist_track VALUES(5,3103);
INSERT INTO playlist_track VALUES(5,3409);
INSERT INTO playlist_track VALUES(5,299);
INSERT INTO playlist_track VALUES(5,300);
INSERT INTO playlist_track VALUES(5,301);
INSERT INTO playlist_track VALUES(5,302);
INSERT INTO playlist_track VALUES(5,303);
INSERT INTO playlist_track VALUES(5,304);
INSERT INTO playlist_track VALUES(5,305);
INSERT INTO playlist_track VALUES(5,306);
INSERT INTO playlist_track VALUES(5,307);
INSERT INTO playlist_track VALUES(5,308);
INSERT INTO playlist_track VALUES(5,309);
INSERT INTO playlist_track VALUES(5,310);
INSERT INTO playlist_track VALUES(5,311);
INSERT INTO playlist_track VALUES(5,312);
INSERT INTO playlist_track VALUES(5,851);
INSERT INTO playlist_track VALUES(5,852);
INSERT INTO playlist_track VALUES(5,853);
INSERT INTO playlist_track VALUES(5,854);
INSERT INTO playlist_track VALUES(5,855);
INSERT INTO playlist_track VALUES(5,856);
INSERT INTO playlist_track VALUES(5,857);
INSERT INTO playlist_track VALUES(5,858);
INSERT INTO playlist_track VALUES(5,859);
INSERT INTO playlist_track VALUES(5,860);
INSERT INTO playlist_track VALUES(5,861);
INSERT INTO playlist_track VALUES(5,862);
INSERT INTO playlist_track VALUES(5,863);
INSERT INTO playlist_track VALUES(5,864);
INSERT INTO playlist_track VALUES(5,865);
INSERT INTO playlist_track VALUES(5,866);
INSERT INTO playlist_track VALUES(5,867);
INSERT INTO playlist_track VALUES(5,868);
INSERT INTO playlist_track VALUES(5,869);
INSERT INTO playlist_track VALUES(5,870);
INSERT INTO playlist_track VALUES(5,871);
INSERT INTO playlist_track VALUES(5,872);
INSERT INTO playlist_track VALUES(5,873);
INSERT INTO playlist_track VALUES(5,874);
INSERT INTO playlist_track VALUES(5,875);
INSERT INTO playlist_track VALUES(5,876);
INSERT INTO playlist_track VALUES(5,1057);
INSERT INTO playlist_track VALUES(5,1058);
INSERT INTO playlist_track VALUES(5,1059);
INSERT INTO playlist_track VALUES(5,1060);
INSERT INTO playlist_track VALUES(5,1061);
INSERT INTO playlist_track VALUES(5,1062);
INSERT INTO playlist_track VALUES(5,1063);
INSERT INTO playlist_track VALUES(5,1064);
INSERT INTO playlist_track VALUES(5,1065);
INSERT INTO playlist_track VALUES(5,1066);
INSERT INTO playlist_track VALUES(5,1067);
INSERT INTO playlist_track VALUES(5,1068);
INSERT INTO playlist_track VALUES(5,1069);
INSERT INTO playlist_track VALUES(5,1070);
INSERT INTO playlist_track VALUES(5,1071);
INSERT INTO playlist_track VALUES(5,1072);
INSERT INTO playlist_track VALUES(5,501);
INSERT INTO playlist_track VALUES(5,502);
INSERT INTO playlist_track VALUES(5,503);
INSERT INTO playlist_track VALUES(5,504);
INSERT INTO playlist_track VALUES(5,505);
INSERT INTO playlist_track VALUES(5,506);
INSERT INTO playlist_track VALUES(5,507);
INSERT INTO playlist_track VALUES(5,508);
INSERT INTO playlist_track VALUES(5,509);
INSERT INTO playlist_track VALUES(5,510);
INSERT INTO playlist_track VALUES(5,511);
INSERT INTO playlist_track VALUES(5,512);
INSERT INTO playlist_track VALUES(5,513);
INSERT INTO playlist_track VALUES(5,514);
INSERT INTO playlist_track VALUES(5,1444);
INSERT INTO playlist_track VALUES(5,1445);
INSERT INTO playlist_track VALUES(5,1446);
INSERT INTO playlist_track VALUES(5,1447);
INSERT INTO playlist_track VALUES(5,1448);
INSERT INTO playlist_track VALUES(5,1449);
INSERT INTO playlist_track VALUES(5,1450);
INSERT INTO playlist_track VALUES(5,1451);
INSERT INTO playlist_track VALUES(5,1452);
INSERT INTO playlist_track VALUES(5,1453);
INSERT INTO playlist_track VALUES(5,1454);
INSERT INTO playlist_track VALUES(5,1496);
INSERT INTO playlist_track VALUES(5,1497);
INSERT INTO playlist_track VALUES(5,1498);
INSERT INTO playlist_track VALUES(5,1499);
INSERT INTO playlist_track VALUES(5,1500);
INSERT INTO playlist_track VALUES(5,1501);
INSERT INTO playlist_track VALUES(5,1502);
INSERT INTO playlist_track VALUES(5,1503);
INSERT INTO playlist_track VALUES(5,1504);
INSERT INTO playlist_track VALUES(5,1505);
INSERT INTO playlist_track VALUES(5,1671);
INSERT INTO playlist_track VALUES(5,1672);
INSERT INTO playlist_track VALUES(5,1673);
INSERT INTO playlist_track VALUES(5,1674);
INSERT INTO playlist_track VALUES(5,1675);
INSERT INTO playlist_track VALUES(5,1676);
INSERT INTO playlist_track VALUES(5,1677);
INSERT INTO playlist_track VALUES(5,1678);
INSERT INTO playlist_track VALUES(5,1679);
INSERT INTO playlist_track VALUES(5,1680);
INSERT INTO playlist_track VALUES(5,1681);
INSERT INTO playlist_track VALUES(5,1682);
INSERT INTO playlist_track VALUES(5,1683);
INSERT INTO playlist_track VALUES(5,1684);
INSERT INTO playlist_track VALUES(5,1685);
INSERT INTO playlist_track VALUES(5,2044);
INSERT INTO playlist_track VALUES(5,2045);
INSERT INTO playlist_track VALUES(5,2046);
INSERT INTO playlist_track VALUES(5,2047);
INSERT INTO playlist_track VALUES(5,2048);
INSERT INTO playlist_track VALUES(5,2049);
INSERT INTO playlist_track VALUES(5,2050);
INSERT INTO playlist_track VALUES(5,2051);
INSERT INTO playlist_track VALUES(5,2052);
INSERT INTO playlist_track VALUES(5,2053);
INSERT INTO playlist_track VALUES(5,2054);
INSERT INTO playlist_track VALUES(5,2055);
INSERT INTO playlist_track VALUES(5,2056);
INSERT INTO playlist_track VALUES(5,2057);
INSERT INTO playlist_track VALUES(5,2058);
INSERT INTO playlist_track VALUES(5,2059);
INSERT INTO playlist_track VALUES(5,2060);
INSERT INTO playlist_track VALUES(5,2061);
INSERT INTO playlist_track VALUES(5,2062);
INSERT INTO playlist_track VALUES(5,2063);
INSERT INTO playlist_track VALUES(5,2064);
INSERT INTO playlist_track VALUES(5,2238);
INSERT INTO playlist_track VALUES(5,2239);
INSERT INTO playlist_track VALUES(5,2240);
INSERT INTO playlist_track VALUES(5,2241);
INSERT INTO playlist_track VALUES(5,2242);
INSERT INTO playlist_track VALUES(5,2243);
INSERT INTO playlist_track VALUES(5,2244);
INSERT INTO playlist_track VALUES(5,2245);
INSERT INTO playlist_track VALUES(5,2246);
INSERT INTO playlist_track VALUES(5,2247);
INSERT INTO playlist_track VALUES(5,2248);
INSERT INTO playlist_track VALUES(5,2249);
INSERT INTO playlist_track VALUES(5,2250);
INSERT INTO playlist_track VALUES(5,2251);
INSERT INTO playlist_track VALUES(5,2252);
INSERT INTO playlist_track VALUES(5,2253);
INSERT INTO playlist_track VALUES(5,2391);
INSERT INTO playlist_track VALUES(5,2392);
INSERT INTO playlist_track VALUES(5,2393);
INSERT INTO playlist_track VALUES(5,2394);
INSERT INTO playlist_track VALUES(5,2395);
INSERT INTO playlist_track VALUES(5,2396);
INSERT INTO playlist_track VALUES(5,2397);
INSERT INTO playlist_track VALUES(5,2398);
INSERT INTO playlist_track VALUES(5,2399);
INSERT INTO playlist_track VALUES(5,2400);
INSERT INTO playlist_track VALUES(5,2401);
INSERT INTO playlist_track VALUES(5,2402);
INSERT INTO playlist_track VALUES(5,2403);
INSERT INTO playlist_track VALUES(5,2404);
INSERT INTO playlist_track VALUES(5,2405);
INSERT INTO playlist_track VALUES(5,570);
INSERT INTO playlist_track VALUES(5,573);
INSERT INTO playlist_track VALUES(5,577);
INSERT INTO playlist_track VALUES(5,580);
INSERT INTO playlist_track VALUES(5,581);
INSERT INTO playlist_track VALUES(5,571);
INSERT INTO playlist_track VALUES(5,579);
INSERT INTO playlist_track VALUES(5,582);
INSERT INTO playlist_track VALUES(5,572);
INSERT INTO playlist_track VALUES(5,575);
INSERT INTO playlist_track VALUES(5,578);
INSERT INTO playlist_track VALUES(5,574);
INSERT INTO playlist_track VALUES(5,576);
INSERT INTO playlist_track VALUES(5,2707);
INSERT INTO playlist_track VALUES(5,2714);
INSERT INTO playlist_track VALUES(5,2717);
INSERT INTO playlist_track VALUES(5,2718);
INSERT INTO playlist_track VALUES(5,3146);
INSERT INTO playlist_track VALUES(5,3147);
INSERT INTO playlist_track VALUES(5,3148);
INSERT INTO playlist_track VALUES(5,3149);
INSERT INTO playlist_track VALUES(5,3150);
INSERT INTO playlist_track VALUES(5,3151);
INSERT INTO playlist_track VALUES(5,3152);
INSERT INTO playlist_track VALUES(5,3153);
INSERT INTO playlist_track VALUES(5,3154);
INSERT INTO playlist_track VALUES(5,3155);
INSERT INTO playlist_track VALUES(5,3156);
INSERT INTO playlist_track VALUES(5,3157);
INSERT INTO playlist_track VALUES(5,3158);
INSERT INTO playlist_track VALUES(5,3159);
INSERT INTO playlist_track VALUES(5,3160);
INSERT INTO playlist_track VALUES(5,3161);
INSERT INTO playlist_track VALUES(5,3162);
INSERT INTO playlist_track VALUES(5,3163);
INSERT INTO playlist_track VALUES(5,3164);
INSERT INTO playlist_track VALUES(5,3438);
INSERT INTO playlist_track VALUES(5,3442);
INSERT INTO playlist_track VALUES(5,3436);
INSERT INTO playlist_track VALUES(5,3454);
INSERT INTO playlist_track VALUES(5,3432);
INSERT INTO playlist_track VALUES(5,3447);
INSERT INTO playlist_track VALUES(5,3434);
INSERT INTO playlist_track VALUES(5,3449);
INSERT INTO playlist_track VALUES(5,3445);
INSERT INTO playlist_track VALUES(5,3439);
INSERT INTO playlist_track VALUES(5,3435);
INSERT INTO playlist_track VALUES(5,3448);
INSERT INTO playlist_track VALUES(5,3437);
INSERT INTO playlist_track VALUES(5,3446);
INSERT INTO playlist_track VALUES(5,3444);
INSERT INTO playlist_track VALUES(5,3451);
INSERT INTO playlist_track VALUES(5,3430);
INSERT INTO playlist_track VALUES(5,3482);
INSERT INTO playlist_track VALUES(5,3485);
INSERT INTO playlist_track VALUES(5,3499);
INSERT INTO playlist_track VALUES(5,3490);
INSERT INTO playlist_track VALUES(5,3489);
INSERT INTO playlist_track VALUES(5,3492);
INSERT INTO playlist_track VALUES(5,3493);
INSERT INTO playlist_track VALUES(5,3498);
INSERT INTO playlist_track VALUES(5,3481);
INSERT INTO playlist_track VALUES(5,3503);
INSERT INTO playlist_track VALUES(8,3427);
INSERT INTO playlist_track VALUES(8,3357);
INSERT INTO playlist_track VALUES(8,1);
INSERT INTO playlist_track VALUES(8,6);
INSERT INTO playlist_track VALUES(8,7);
INSERT INTO playlist_track VALUES(8,8);
INSERT INTO playlist_track VALUES(8,9);
INSERT INTO playlist_track VALUES(8,10);
INSERT INTO playlist_track VALUES(8,11);
INSERT INTO playlist_track VALUES(8,12);
INSERT INTO playlist_track VALUES(8,13);
INSERT INTO playlist_track VALUES(8,14);
INSERT INTO playlist_track VALUES(8,15);
INSERT INTO playlist_track VALUES(8,16);
INSERT INTO playlist_track VALUES(8,17);
INSERT INTO playlist_track VALUES(8,18);
INSERT INTO playlist_track VALUES(8,19);
INSERT INTO playlist_track VALUES(8,20);
INSERT INTO playlist_track VALUES(8,21);
INSERT INTO playlist_track VALUES(8,22);
INSERT INTO playlist_track VALUES(8,3411);
INSERT INTO playlist_track VALUES(8,3412);
INSERT INTO playlist_track VALUES(8,3419);
INSERT INTO playlist_track VALUES(8,2);
INSERT INTO playlist_track VALUES(8,3);
INSERT INTO playlist_track VALUES(8,4);
INSERT INTO playlist_track VALUES(8,5);
INSERT INTO playlist_track VALUES(8,23);
INSERT INTO playlist_track VALUES(8,24);
INSERT INTO playlist_track VALUES(8,25);
INSERT INTO playlist_track VALUES(8,26);
INSERT INTO playlist_track VALUES(8,27);
INSERT INTO playlist_track VALUES(8,28);
INSERT INTO playlist_track VALUES(8,29);
INSERT INTO playlist_track VALUES(8,30);
INSERT INTO playlist_track VALUES(8,31);
INSERT INTO playlist_track VALUES(8,32);
INSERT INTO playlist_track VALUES(8,33);
INSERT INTO playlist_track VALUES(8,34);
INSERT INTO playlist_track VALUES(8,35);
INSERT INTO playlist_track VALUES(8,36);
INSERT INTO playlist_track VALUES(8,37);
INSERT INTO playlist_track VALUES(8,3256);
INSERT INTO playlist_track VALUES(8,3350);
INSERT INTO playlist_track VALUES(8,3349);
INSERT INTO playlist_track VALUES(8,38);
INSERT INTO playlist_track VALUES(8,39);
INSERT INTO playlist_track VALUES(8,40);
INSERT INTO playlist_track VALUES(8,41);
INSERT INTO playlist_track VALUES(8,42);
INSERT INTO playlist_track VALUES(8,43);
INSERT INTO playlist_track VALUES(8,44);
INSERT INTO playlist_track VALUES(8,45);
INSERT INTO playlist_track VALUES(8,46);
INSERT INTO playlist_track VALUES(8,47);
INSERT INTO playlist_track VALUES(8,48);
INSERT INTO playlist_track VALUES(8,49);
INSERT INTO playlist_track VALUES(8,50);
INSERT INTO playlist_track VALUES(8,3403);
INSERT INTO playlist_track VALUES(8,51);
INSERT INTO playlist_track VALUES(8,52);
INSERT INTO playlist_track VALUES(8,53);
INSERT INTO playlist_track VALUES(8,54);
INSERT INTO playlist_track VALUES(8,55);
INSERT INTO playlist_track VALUES(8,56);
INSERT INTO playlist_track VALUES(8,57);
INSERT INTO playlist_track VALUES(8,58);
INSERT INTO playlist_track VALUES(8,59);
INSERT INTO playlist_track VALUES(8,60);
INSERT INTO playlist_track VALUES(8,61);
INSERT INTO playlist_track VALUES(8,62);
INSERT INTO playlist_track VALUES(8,3406);
INSERT INTO playlist_track VALUES(8,379);
INSERT INTO playlist_track VALUES(8,391);
INSERT INTO playlist_track VALUES(8,63);
INSERT INTO playlist_track VALUES(8,64);
INSERT INTO playlist_track VALUES(8,65);
INSERT INTO playlist_track VALUES(8,66);
INSERT INTO playlist_track VALUES(8,67);
INSERT INTO playlist_track VALUES(8,68);
INSERT INTO playlist_track VALUES(8,69);
INSERT INTO playlist_track VALUES(8,70);
INSERT INTO playlist_track VALUES(8,71);
INSERT INTO playlist_track VALUES(8,72);
INSERT INTO playlist_track VALUES(8,73);
INSERT INTO playlist_track VALUES(8,74);
INSERT INTO playlist_track VALUES(8,75);
INSERT INTO playlist_track VALUES(8,76);
INSERT INTO playlist_track VALUES(8,77);
INSERT INTO playlist_track VALUES(8,78);
INSERT INTO playlist_track VALUES(8,79);
INSERT INTO playlist_track VALUES(8,80);
INSERT INTO playlist_track VALUES(8,81);
INSERT INTO playlist_track VALUES(8,82);
INSERT INTO playlist_track VALUES(8,83);
INSERT INTO playlist_track VALUES(8,84);
INSERT INTO playlist_track VALUES(8,85);
INSERT INTO playlist_track VALUES(8,86);
INSERT INTO playlist_track VALUES(8,87);
INSERT INTO playlist_track VALUES(8,88);
INSERT INTO playlist_track VALUES(8,89);
INSERT INTO playlist_track VALUES(8,90);
INSERT INTO playlist_track VALUES(8,91);
INSERT INTO playlist_track VALUES(8,92);
INSERT INTO playlist_track VALUES(8,93);
INSERT INTO playlist_track VALUES(8,94);
INSERT INTO playlist_track VALUES(8,95);
INSERT INTO playlist_track VALUES(8,96);
INSERT INTO playlist_track VALUES(8,97);
INSERT INTO playlist_track VALUES(8,98);
INSERT INTO playlist_track VALUES(8,99);
INSERT INTO playlist_track VALUES(8,100);
INSERT INTO playlist_track VALUES(8,101);
INSERT INTO playlist_track VALUES(8,102);
INSERT INTO playlist_track VALUES(8,103);
INSERT INTO playlist_track VALUES(8,104);
INSERT INTO playlist_track VALUES(8,105);
INSERT INTO playlist_track VALUES(8,106);
INSERT INTO playlist_track VALUES(8,107);
INSERT INTO playlist_track VALUES(8,108);
INSERT INTO playlist_track VALUES(8,109);
INSERT INTO playlist_track VALUES(8,110);
INSERT INTO playlist_track VALUES(8,3402);
INSERT INTO playlist_track VALUES(8,3389);
INSERT INTO playlist_track VALUES(8,3390);
INSERT INTO playlist_track VALUES(8,3391);
INSERT INTO playlist_track VALUES(8,3392);
INSERT INTO playlist_track VALUES(8,3393);
INSERT INTO playlist_track VALUES(8,3394);
INSERT INTO playlist_track VALUES(8,3395);
INSERT INTO playlist_track VALUES(8,3396);
INSERT INTO playlist_track VALUES(8,3397);
INSERT INTO playlist_track VALUES(8,3398);
INSERT INTO playlist_track VALUES(8,3399);
INSERT INTO playlist_track VALUES(8,3400);
INSERT INTO playlist_track VALUES(8,3401);
INSERT INTO playlist_track VALUES(8,3262);
INSERT INTO playlist_track VALUES(8,376);
INSERT INTO playlist_track VALUES(8,397);
INSERT INTO playlist_track VALUES(8,382);
INSERT INTO playlist_track VALUES(8,111);
INSERT INTO playlist_track VALUES(8,112);
INSERT INTO playlist_track VALUES(8,113);
INSERT INTO playlist_track VALUES(8,114);
INSERT INTO playlist_track VALUES(8,115);
INSERT INTO playlist_track VALUES(8,116);
INSERT INTO playlist_track VALUES(8,117);
INSERT INTO playlist_track VALUES(8,118);
INSERT INTO playlist_track VALUES(8,119);
INSERT INTO playlist_track VALUES(8,120);
INSERT INTO playlist_track VALUES(8,121);
INSERT INTO playlist_track VALUES(8,122);
INSERT INTO playlist_track VALUES(8,389);
INSERT INTO playlist_track VALUES(8,404);
INSERT INTO playlist_track VALUES(8,406);
INSERT INTO playlist_track VALUES(8,3421);
INSERT INTO playlist_track VALUES(8,380);
INSERT INTO playlist_track VALUES(8,394);
INSERT INTO playlist_track VALUES(8,3268);
INSERT INTO playlist_track VALUES(8,3413);
INSERT INTO playlist_track VALUES(8,3263);
INSERT INTO playlist_track VALUES(8,123);
INSERT INTO playlist_track VALUES(8,124);
INSERT INTO playlist_track VALUES(8,125);
INSERT INTO playlist_track VALUES(8,126);
INSERT INTO playlist_track VALUES(8,127);
INSERT INTO playlist_track VALUES(8,128);
INSERT INTO playlist_track VALUES(8,129);
INSERT INTO playlist_track VALUES(8,130);
INSERT INTO playlist_track VALUES(8,2572);
INSERT INTO playlist_track VALUES(8,2573);
INSERT INTO playlist_track VALUES(8,2574);
INSERT INTO playlist_track VALUES(8,2575);
INSERT INTO playlist_track VALUES(8,2576);
INSERT INTO playlist_track VALUES(8,2577);
INSERT INTO playlist_track VALUES(8,2578);
INSERT INTO playlist_track VALUES(8,2579);
INSERT INTO playlist_track VALUES(8,2580);
INSERT INTO playlist_track VALUES(8,2581);
INSERT INTO playlist_track VALUES(8,2582);
INSERT INTO playlist_track VALUES(8,2583);
INSERT INTO playlist_track VALUES(8,2584);
INSERT INTO playlist_track VALUES(8,2585);
INSERT INTO playlist_track VALUES(8,2586);
INSERT INTO playlist_track VALUES(8,2587);
INSERT INTO playlist_track VALUES(8,2588);
INSERT INTO playlist_track VALUES(8,2589);
INSERT INTO playlist_track VALUES(8,2590);
INSERT INTO playlist_track VALUES(8,3266);
INSERT INTO playlist_track VALUES(8,131);
INSERT INTO playlist_track VALUES(8,132);
INSERT INTO playlist_track VALUES(8,133);
INSERT INTO playlist_track VALUES(8,134);
INSERT INTO playlist_track VALUES(8,135);
INSERT INTO playlist_track VALUES(8,136);
INSERT INTO playlist_track VALUES(8,137);
INSERT INTO playlist_track VALUES(8,138);
INSERT INTO playlist_track VALUES(8,139);
INSERT INTO playlist_track VALUES(8,140);
INSERT INTO playlist_track VALUES(8,141);
INSERT INTO playlist_track VALUES(8,142);
INSERT INTO playlist_track VALUES(8,143);
INSERT INTO playlist_track VALUES(8,144);
INSERT INTO playlist_track VALUES(8,145);
INSERT INTO playlist_track VALUES(8,146);
INSERT INTO playlist_track VALUES(8,147);
INSERT INTO playlist_track VALUES(8,148);
INSERT INTO playlist_track VALUES(8,149);
INSERT INTO playlist_track VALUES(8,150);
INSERT INTO playlist_track VALUES(8,151);
INSERT INTO playlist_track VALUES(8,152);
INSERT INTO playlist_track VALUES(8,153);
INSERT INTO playlist_track VALUES(8,154);
INSERT INTO playlist_track VALUES(8,155);
INSERT INTO playlist_track VALUES(8,156);
INSERT INTO playlist_track VALUES(8,157);
INSERT INTO playlist_track VALUES(8,158);
INSERT INTO playlist_track VALUES(8,159);
INSERT INTO playlist_track VALUES(8,160);
INSERT INTO playlist_track VALUES(8,161);
INSERT INTO playlist_track VALUES(8,162);
INSERT INTO playlist_track VALUES(8,163);
INSERT INTO playlist_track VALUES(8,164);
INSERT INTO playlist_track VALUES(8,165);
INSERT INTO playlist_track VALUES(8,166);
INSERT INTO playlist_track VALUES(8,167);
INSERT INTO playlist_track VALUES(8,168);
INSERT INTO playlist_track VALUES(8,169);
INSERT INTO playlist_track VALUES(8,170);
INSERT INTO playlist_track VALUES(8,171);
INSERT INTO playlist_track VALUES(8,172);
INSERT INTO playlist_track VALUES(8,173);
INSERT INTO playlist_track VALUES(8,174);
INSERT INTO playlist_track VALUES(8,175);
INSERT INTO playlist_track VALUES(8,176);
INSERT INTO playlist_track VALUES(8,177);
INSERT INTO playlist_track VALUES(8,178);
INSERT INTO playlist_track VALUES(8,179);
INSERT INTO playlist_track VALUES(8,180);
INSERT INTO playlist_track VALUES(8,181);
INSERT INTO playlist_track VALUES(8,182);
INSERT INTO playlist_track VALUES(8,3426);
INSERT INTO playlist_track VALUES(8,3416);
INSERT INTO playlist_track VALUES(8,183);
INSERT INTO playlist_track VALUES(8,184);
INSERT INTO playlist_track VALUES(8,185);
INSERT INTO playlist_track VALUES(8,186);
INSERT INTO playlist_track VALUES(8,187);
INSERT INTO playlist_track VALUES(8,188);
INSERT INTO playlist_track VALUES(8,189);
INSERT INTO playlist_track VALUES(8,190);
INSERT INTO playlist_track VALUES(8,191);
INSERT INTO playlist_track VALUES(8,192);
INSERT INTO playlist_track VALUES(8,193);
INSERT INTO playlist_track VALUES(8,194);
INSERT INTO playlist_track VALUES(8,195);
INSERT INTO playlist_track VALUES(8,196);
INSERT INTO playlist_track VALUES(8,197);
INSERT INTO playlist_track VALUES(8,198);
INSERT INTO playlist_track VALUES(8,199);
INSERT INTO playlist_track VALUES(8,200);
INSERT INTO playlist_track VALUES(8,201);
INSERT INTO playlist_track VALUES(8,202);
INSERT INTO playlist_track VALUES(8,203);
INSERT INTO playlist_track VALUES(8,204);
INSERT INTO playlist_track VALUES(8,515);
INSERT INTO playlist_track VALUES(8,516);
INSERT INTO playlist_track VALUES(8,517);
INSERT INTO playlist_track VALUES(8,518);
INSERT INTO playlist_track VALUES(8,519);
INSERT INTO playlist_track VALUES(8,520);
INSERT INTO playlist_track VALUES(8,521);
INSERT INTO playlist_track VALUES(8,522);
INSERT INTO playlist_track VALUES(8,523);
INSERT INTO playlist_track VALUES(8,524);
INSERT INTO playlist_track VALUES(8,525);
INSERT INTO playlist_track VALUES(8,526);
INSERT INTO playlist_track VALUES(8,527);
INSERT INTO playlist_track VALUES(8,528);
INSERT INTO playlist_track VALUES(8,205);
INSERT INTO playlist_track VALUES(8,206);
INSERT INTO playlist_track VALUES(8,207);
INSERT INTO playlist_track VALUES(8,208);
INSERT INTO playlist_track VALUES(8,209);
INSERT INTO playlist_track VALUES(8,210);
INSERT INTO playlist_track VALUES(8,211);
INSERT INTO playlist_track VALUES(8,212);
INSERT INTO playlist_track VALUES(8,213);
INSERT INTO playlist_track VALUES(8,214);
INSERT INTO playlist_track VALUES(8,215);
INSERT INTO playlist_track VALUES(8,216);
INSERT INTO playlist_track VALUES(8,217);
INSERT INTO playlist_track VALUES(8,218);
INSERT INTO playlist_track VALUES(8,219);
INSERT INTO playlist_track VALUES(8,220);
INSERT INTO playlist_track VALUES(8,221);
INSERT INTO playlist_track VALUES(8,222);
INSERT INTO playlist_track VALUES(8,223);
INSERT INTO playlist_track VALUES(8,224);
INSERT INTO playlist_track VALUES(8,225);
INSERT INTO playlist_track VALUES(8,3336);
INSERT INTO playlist_track VALUES(8,715);
INSERT INTO playlist_track VALUES(8,716);
INSERT INTO playlist_track VALUES(8,717);
INSERT INTO playlist_track VALUES(8,718);
INSERT INTO playlist_track VALUES(8,719);
INSERT INTO playlist_track VALUES(8,720);
INSERT INTO playlist_track VALUES(8,721);
INSERT INTO playlist_track VALUES(8,722);
INSERT INTO playlist_track VALUES(8,723);
INSERT INTO playlist_track VALUES(8,724);
INSERT INTO playlist_track VALUES(8,725);
INSERT INTO playlist_track VALUES(8,726);
INSERT INTO playlist_track VALUES(8,727);
INSERT INTO playlist_track VALUES(8,728);
INSERT INTO playlist_track VALUES(8,729);
INSERT INTO playlist_track VALUES(8,730);
INSERT INTO playlist_track VALUES(8,731);
INSERT INTO playlist_track VALUES(8,732);
INSERT INTO playlist_track VALUES(8,733);
INSERT INTO playlist_track VALUES(8,734);
INSERT INTO playlist_track VALUES(8,735);
INSERT INTO playlist_track VALUES(8,736);
INSERT INTO playlist_track VALUES(8,737);
INSERT INTO playlist_track VALUES(8,738);
INSERT INTO playlist_track VALUES(8,739);
INSERT INTO playlist_track VALUES(8,740);
INSERT INTO playlist_track VALUES(8,741);
INSERT INTO playlist_track VALUES(8,742);
INSERT INTO playlist_track VALUES(8,743);
INSERT INTO playlist_track VALUES(8,744);
INSERT INTO playlist_track VALUES(8,3324);
INSERT INTO playlist_track VALUES(8,3417);
INSERT INTO playlist_track VALUES(8,226);
INSERT INTO playlist_track VALUES(8,227);
INSERT INTO playlist_track VALUES(8,228);
INSERT INTO playlist_track VALUES(8,229);
INSERT INTO playlist_track VALUES(8,230);
INSERT INTO playlist_track VALUES(8,231);
INSERT INTO playlist_track VALUES(8,232);
INSERT INTO playlist_track VALUES(8,233);
INSERT INTO playlist_track VALUES(8,234);
INSERT INTO playlist_track VALUES(8,235);
INSERT INTO playlist_track VALUES(8,236);
INSERT INTO playlist_track VALUES(8,237);
INSERT INTO playlist_track VALUES(8,238);
INSERT INTO playlist_track VALUES(8,239);
INSERT INTO playlist_track VALUES(8,240);
INSERT INTO playlist_track VALUES(8,241);
INSERT INTO playlist_track VALUES(8,242);
INSERT INTO playlist_track VALUES(8,243);
INSERT INTO playlist_track VALUES(8,244);
INSERT INTO playlist_track VALUES(8,245);
INSERT INTO playlist_track VALUES(8,246);
INSERT INTO playlist_track VALUES(8,247);
INSERT INTO playlist_track VALUES(8,248);
INSERT INTO playlist_track VALUES(8,249);
INSERT INTO playlist_track VALUES(8,250);
INSERT INTO playlist_track VALUES(8,251);
INSERT INTO playlist_track VALUES(8,252);
INSERT INTO playlist_track VALUES(8,253);
INSERT INTO playlist_track VALUES(8,254);
INSERT INTO playlist_track VALUES(8,255);
INSERT INTO playlist_track VALUES(8,256);
INSERT INTO playlist_track VALUES(8,257);
INSERT INTO playlist_track VALUES(8,258);
INSERT INTO playlist_track VALUES(8,259);
INSERT INTO playlist_track VALUES(8,260);
INSERT INTO playlist_track VALUES(8,261);
INSERT INTO playlist_track VALUES(8,262);
INSERT INTO playlist_track VALUES(8,263);
INSERT INTO playlist_track VALUES(8,264);
INSERT INTO playlist_track VALUES(8,265);
INSERT INTO playlist_track VALUES(8,266);
INSERT INTO playlist_track VALUES(8,267);
INSERT INTO playlist_track VALUES(8,268);
INSERT INTO playlist_track VALUES(8,269);
INSERT INTO playlist_track VALUES(8,270);
INSERT INTO playlist_track VALUES(8,271);
INSERT INTO playlist_track VALUES(8,272);
INSERT INTO playlist_track VALUES(8,273);
INSERT INTO playlist_track VALUES(8,274);
INSERT INTO playlist_track VALUES(8,275);
INSERT INTO playlist_track VALUES(8,276);
INSERT INTO playlist_track VALUES(8,277);
INSERT INTO playlist_track VALUES(8,278);
INSERT INTO playlist_track VALUES(8,279);
INSERT INTO playlist_track VALUES(8,280);
INSERT INTO playlist_track VALUES(8,281);
INSERT INTO playlist_track VALUES(8,3375);
INSERT INTO playlist_track VALUES(8,3376);
INSERT INTO playlist_track VALUES(8,3377);
INSERT INTO playlist_track VALUES(8,3378);
INSERT INTO playlist_track VALUES(8,3379);
INSERT INTO playlist_track VALUES(8,3380);
INSERT INTO playlist_track VALUES(8,3381);
INSERT INTO playlist_track VALUES(8,3382);
INSERT INTO playlist_track VALUES(8,3383);
INSERT INTO playlist_track VALUES(8,3384);
INSERT INTO playlist_track VALUES(8,3385);
INSERT INTO playlist_track VALUES(8,3386);
INSERT INTO playlist_track VALUES(8,3387);
INSERT INTO playlist_track VALUES(8,3388);
INSERT INTO playlist_track VALUES(8,3255);
INSERT INTO playlist_track VALUES(8,282);
INSERT INTO playlist_track VALUES(8,283);
INSERT INTO playlist_track VALUES(8,284);
INSERT INTO playlist_track VALUES(8,285);
INSERT INTO playlist_track VALUES(8,286);
INSERT INTO playlist_track VALUES(8,287);
INSERT INTO playlist_track VALUES(8,288);
INSERT INTO playlist_track VALUES(8,289);
INSERT INTO playlist_track VALUES(8,290);
INSERT INTO playlist_track VALUES(8,291);
INSERT INTO playlist_track VALUES(8,292);
INSERT INTO playlist_track VALUES(8,293);
INSERT INTO playlist_track VALUES(8,294);
INSERT INTO playlist_track VALUES(8,295);
INSERT INTO playlist_track VALUES(8,296);
INSERT INTO playlist_track VALUES(8,297);
INSERT INTO playlist_track VALUES(8,298);
INSERT INTO playlist_track VALUES(8,299);
INSERT INTO playlist_track VALUES(8,300);
INSERT INTO playlist_track VALUES(8,301);
INSERT INTO playlist_track VALUES(8,302);
INSERT INTO playlist_track VALUES(8,303);
INSERT INTO playlist_track VALUES(8,304);
INSERT INTO playlist_track VALUES(8,305);
INSERT INTO playlist_track VALUES(8,306);
INSERT INTO playlist_track VALUES(8,307);
INSERT INTO playlist_track VALUES(8,308);
INSERT INTO playlist_track VALUES(8,309);
INSERT INTO playlist_track VALUES(8,310);
INSERT INTO playlist_track VALUES(8,311);
INSERT INTO playlist_track VALUES(8,312);
INSERT INTO playlist_track VALUES(8,2591);
INSERT INTO playlist_track VALUES(8,2592);
INSERT INTO playlist_track VALUES(8,2593);
INSERT INTO playlist_track VALUES(8,2594);
INSERT INTO playlist_track VALUES(8,2595);
INSERT INTO playlist_track VALUES(8,2596);
INSERT INTO playlist_track VALUES(8,2597);
INSERT INTO playlist_track VALUES(8,2598);
INSERT INTO playlist_track VALUES(8,2599);
INSERT INTO playlist_track VALUES(8,2600);
INSERT INTO playlist_track VALUES(8,2601);
INSERT INTO playlist_track VALUES(8,2602);
INSERT INTO playlist_track VALUES(8,2603);
INSERT INTO playlist_track VALUES(8,2604);
INSERT INTO playlist_track VALUES(8,2605);
INSERT INTO playlist_track VALUES(8,2606);
INSERT INTO playlist_track VALUES(8,2607);
INSERT INTO playlist_track VALUES(8,2608);
INSERT INTO playlist_track VALUES(8,313);
INSERT INTO playlist_track VALUES(8,314);
INSERT INTO playlist_track VALUES(8,315);
INSERT INTO playlist_track VALUES(8,316);
INSERT INTO playlist_track VALUES(8,317);
INSERT INTO playlist_track VALUES(8,318);
INSERT INTO playlist_track VALUES(8,319);
INSERT INTO playlist_track VALUES(8,320);
INSERT INTO playlist_track VALUES(8,321);
INSERT INTO playlist_track VALUES(8,322);
INSERT INTO playlist_track VALUES(8,399);
INSERT INTO playlist_track VALUES(8,3259);
INSERT INTO playlist_track VALUES(8,675);
INSERT INTO playlist_track VALUES(8,676);
INSERT INTO playlist_track VALUES(8,677);
INSERT INTO playlist_track VALUES(8,678);
INSERT INTO playlist_track VALUES(8,679);
INSERT INTO playlist_track VALUES(8,680);
INSERT INTO playlist_track VALUES(8,681);
INSERT INTO playlist_track VALUES(8,682);
INSERT INTO playlist_track VALUES(8,683);
INSERT INTO playlist_track VALUES(8,684);
INSERT INTO playlist_track VALUES(8,685);
INSERT INTO playlist_track VALUES(8,686);
INSERT INTO playlist_track VALUES(8,687);
INSERT INTO playlist_track VALUES(8,688);
INSERT INTO playlist_track VALUES(8,689);
INSERT INTO playlist_track VALUES(8,690);
INSERT INTO playlist_track VALUES(8,691);
INSERT INTO playlist_track VALUES(8,692);
INSERT INTO playlist_track VALUES(8,693);
INSERT INTO playlist_track VALUES(8,694);
INSERT INTO playlist_track VALUES(8,695);
INSERT INTO playlist_track VALUES(8,696);
INSERT INTO playlist_track VALUES(8,697);
INSERT INTO playlist_track VALUES(8,698);
INSERT INTO playlist_track VALUES(8,699);
INSERT INTO playlist_track VALUES(8,700);
INSERT INTO playlist_track VALUES(8,701);
INSERT INTO playlist_track VALUES(8,702);
INSERT INTO playlist_track VALUES(8,703);
INSERT INTO playlist_track VALUES(8,704);
INSERT INTO playlist_track VALUES(8,705);
INSERT INTO playlist_track VALUES(8,706);
INSERT INTO playlist_track VALUES(8,707);
INSERT INTO playlist_track VALUES(8,708);
INSERT INTO playlist_track VALUES(8,709);
INSERT INTO playlist_track VALUES(8,710);
INSERT INTO playlist_track VALUES(8,711);
INSERT INTO playlist_track VALUES(8,712);
INSERT INTO playlist_track VALUES(8,713);
INSERT INTO playlist_track VALUES(8,714);
INSERT INTO playlist_track VALUES(8,2609);
INSERT INTO playlist_track VALUES(8,2610);
INSERT INTO playlist_track VALUES(8,2611);
INSERT INTO playlist_track VALUES(8,2612);
INSERT INTO playlist_track VALUES(8,2613);
INSERT INTO playlist_track VALUES(8,2614);
INSERT INTO playlist_track VALUES(8,2615);
INSERT INTO playlist_track VALUES(8,2616);
INSERT INTO playlist_track VALUES(8,2617);
INSERT INTO playlist_track VALUES(8,2618);
INSERT INTO playlist_track VALUES(8,2619);
INSERT INTO playlist_track VALUES(8,2620);
INSERT INTO playlist_track VALUES(8,2621);
INSERT INTO playlist_track VALUES(8,2622);
INSERT INTO playlist_track VALUES(8,2623);
INSERT INTO playlist_track VALUES(8,2624);
INSERT INTO playlist_track VALUES(8,2625);
INSERT INTO playlist_track VALUES(8,2626);
INSERT INTO playlist_track VALUES(8,2627);
INSERT INTO playlist_track VALUES(8,2628);
INSERT INTO playlist_track VALUES(8,2629);
INSERT INTO playlist_track VALUES(8,2630);
INSERT INTO playlist_track VALUES(8,2631);
INSERT INTO playlist_track VALUES(8,2632);
INSERT INTO playlist_track VALUES(8,2633);
INSERT INTO playlist_track VALUES(8,2634);
INSERT INTO playlist_track VALUES(8,2635);
INSERT INTO playlist_track VALUES(8,2636);
INSERT INTO playlist_track VALUES(8,2637);
INSERT INTO playlist_track VALUES(8,2638);
INSERT INTO playlist_track VALUES(8,489);
INSERT INTO playlist_track VALUES(8,490);
INSERT INTO playlist_track VALUES(8,491);
INSERT INTO playlist_track VALUES(8,492);
INSERT INTO playlist_track VALUES(8,493);
INSERT INTO playlist_track VALUES(8,494);
INSERT INTO playlist_track VALUES(8,495);
INSERT INTO playlist_track VALUES(8,496);
INSERT INTO playlist_track VALUES(8,497);
INSERT INTO playlist_track VALUES(8,498);
INSERT INTO playlist_track VALUES(8,499);
INSERT INTO playlist_track VALUES(8,500);
INSERT INTO playlist_track VALUES(8,816);
INSERT INTO playlist_track VALUES(8,817);
INSERT INTO playlist_track VALUES(8,818);
INSERT INTO playlist_track VALUES(8,819);
INSERT INTO playlist_track VALUES(8,820);
INSERT INTO playlist_track VALUES(8,821);
INSERT INTO playlist_track VALUES(8,822);
INSERT INTO playlist_track VALUES(8,823);
INSERT INTO playlist_track VALUES(8,824);
INSERT INTO playlist_track VALUES(8,825);
INSERT INTO playlist_track VALUES(8,745);
INSERT INTO playlist_track VALUES(8,746);
INSERT INTO playlist_track VALUES(8,747);
INSERT INTO playlist_track VALUES(8,748);
INSERT INTO playlist_track VALUES(8,749);
INSERT INTO playlist_track VALUES(8,750);
INSERT INTO playlist_track VALUES(8,751);
INSERT INTO playlist_track VALUES(8,752);
INSERT INTO playlist_track VALUES(8,753);
INSERT INTO playlist_track VALUES(8,754);
INSERT INTO playlist_track VALUES(8,755);
INSERT INTO playlist_track VALUES(8,756);
INSERT INTO playlist_track VALUES(8,757);
INSERT INTO playlist_track VALUES(8,758);
INSERT INTO playlist_track VALUES(8,759);
INSERT INTO playlist_track VALUES(8,760);
INSERT INTO playlist_track VALUES(8,620);
INSERT INTO playlist_track VALUES(8,621);
INSERT INTO playlist_track VALUES(8,622);
INSERT INTO playlist_track VALUES(8,623);
INSERT INTO playlist_track VALUES(8,761);
INSERT INTO playlist_track VALUES(8,762);
INSERT INTO playlist_track VALUES(8,763);
INSERT INTO playlist_track VALUES(8,764);
INSERT INTO playlist_track VALUES(8,765);
INSERT INTO playlist_track VALUES(8,766);
INSERT INTO playlist_track VALUES(8,767);
INSERT INTO playlist_track VALUES(8,768);
INSERT INTO playlist_track VALUES(8,769);
INSERT INTO playlist_track VALUES(8,770);
INSERT INTO playlist_track VALUES(8,771);
INSERT INTO playlist_track VALUES(8,772);
INSERT INTO playlist_track VALUES(8,773);
INSERT INTO playlist_track VALUES(8,774);
INSERT INTO playlist_track VALUES(8,775);
INSERT INTO playlist_track VALUES(8,776);
INSERT INTO playlist_track VALUES(8,777);
INSERT INTO playlist_track VALUES(8,778);
INSERT INTO playlist_track VALUES(8,779);
INSERT INTO playlist_track VALUES(8,780);
INSERT INTO playlist_track VALUES(8,781);
INSERT INTO playlist_track VALUES(8,782);
INSERT INTO playlist_track VALUES(8,783);
INSERT INTO playlist_track VALUES(8,784);
INSERT INTO playlist_track VALUES(8,785);
INSERT INTO playlist_track VALUES(8,543);
INSERT INTO playlist_track VALUES(8,544);
INSERT INTO playlist_track VALUES(8,545);
INSERT INTO playlist_track VALUES(8,546);
INSERT INTO playlist_track VALUES(8,547);
INSERT INTO playlist_track VALUES(8,548);
INSERT INTO playlist_track VALUES(8,549);
INSERT INTO playlist_track VALUES(8,786);
INSERT INTO playlist_track VALUES(8,787);
INSERT INTO playlist_track VALUES(8,788);
INSERT INTO playlist_track VALUES(8,789);
INSERT INTO playlist_track VALUES(8,790);
INSERT INTO playlist_track VALUES(8,791);
INSERT INTO playlist_track VALUES(8,792);
INSERT INTO playlist_track VALUES(8,793);
INSERT INTO playlist_track VALUES(8,794);
INSERT INTO playlist_track VALUES(8,795);
INSERT INTO playlist_track VALUES(8,796);
INSERT INTO playlist_track VALUES(8,797);
INSERT INTO playlist_track VALUES(8,798);
INSERT INTO playlist_track VALUES(8,799);
INSERT INTO playlist_track VALUES(8,800);
INSERT INTO playlist_track VALUES(8,801);
INSERT INTO playlist_track VALUES(8,802);
INSERT INTO playlist_track VALUES(8,803);
INSERT INTO playlist_track VALUES(8,804);
INSERT INTO playlist_track VALUES(8,805);
INSERT INTO playlist_track VALUES(8,806);
INSERT INTO playlist_track VALUES(8,807);
INSERT INTO playlist_track VALUES(8,808);
INSERT INTO playlist_track VALUES(8,809);
INSERT INTO playlist_track VALUES(8,810);
INSERT INTO playlist_track VALUES(8,811);
INSERT INTO playlist_track VALUES(8,812);
INSERT INTO playlist_track VALUES(8,813);
INSERT INTO playlist_track VALUES(8,814);
INSERT INTO playlist_track VALUES(8,815);
INSERT INTO playlist_track VALUES(8,826);
INSERT INTO playlist_track VALUES(8,827);
INSERT INTO playlist_track VALUES(8,828);
INSERT INTO playlist_track VALUES(8,829);
INSERT INTO playlist_track VALUES(8,830);
INSERT INTO playlist_track VALUES(8,831);
INSERT INTO playlist_track VALUES(8,832);
INSERT INTO playlist_track VALUES(8,833);
INSERT INTO playlist_track VALUES(8,834);
INSERT INTO playlist_track VALUES(8,835);
INSERT INTO playlist_track VALUES(8,836);
INSERT INTO playlist_track VALUES(8,837);
INSERT INTO playlist_track VALUES(8,838);
INSERT INTO playlist_track VALUES(8,839);
INSERT INTO playlist_track VALUES(8,840);
INSERT INTO playlist_track VALUES(8,841);
INSERT INTO playlist_track VALUES(8,842);
INSERT INTO playlist_track VALUES(8,843);
INSERT INTO playlist_track VALUES(8,844);
INSERT INTO playlist_track VALUES(8,845);
INSERT INTO playlist_track VALUES(8,846);
INSERT INTO playlist_track VALUES(8,847);
INSERT INTO playlist_track VALUES(8,848);
INSERT INTO playlist_track VALUES(8,849);
INSERT INTO playlist_track VALUES(8,850);
INSERT INTO playlist_track VALUES(8,3260);
INSERT INTO playlist_track VALUES(8,3331);
INSERT INTO playlist_track VALUES(8,851);
INSERT INTO playlist_track VALUES(8,852);
INSERT INTO playlist_track VALUES(8,853);
INSERT INTO playlist_track VALUES(8,854);
INSERT INTO playlist_track VALUES(8,855);
INSERT INTO playlist_track VALUES(8,856);
INSERT INTO playlist_track VALUES(8,857);
INSERT INTO playlist_track VALUES(8,858);
INSERT INTO playlist_track VALUES(8,859);
INSERT INTO playlist_track VALUES(8,860);
INSERT INTO playlist_track VALUES(8,861);
INSERT INTO playlist_track VALUES(8,862);
INSERT INTO playlist_track VALUES(8,863);
INSERT INTO playlist_track VALUES(8,864);
INSERT INTO playlist_track VALUES(8,865);
INSERT INTO playlist_track VALUES(8,866);
INSERT INTO playlist_track VALUES(8,867);
INSERT INTO playlist_track VALUES(8,868);
INSERT INTO playlist_track VALUES(8,869);
INSERT INTO playlist_track VALUES(8,870);
INSERT INTO playlist_track VALUES(8,871);
INSERT INTO playlist_track VALUES(8,872);
INSERT INTO playlist_track VALUES(8,873);
INSERT INTO playlist_track VALUES(8,874);
INSERT INTO playlist_track VALUES(8,875);
INSERT INTO playlist_track VALUES(8,876);
INSERT INTO playlist_track VALUES(8,2639);
INSERT INTO playlist_track VALUES(8,2640);
INSERT INTO playlist_track VALUES(8,2641);
INSERT INTO playlist_track VALUES(8,2642);
INSERT INTO playlist_track VALUES(8,2643);
INSERT INTO playlist_track VALUES(8,2644);
INSERT INTO playlist_track VALUES(8,2645);
INSERT INTO playlist_track VALUES(8,2646);
INSERT INTO playlist_track VALUES(8,2647);
INSERT INTO playlist_track VALUES(8,2648);
INSERT INTO playlist_track VALUES(8,2649);
INSERT INTO playlist_track VALUES(8,3225);
INSERT INTO playlist_track VALUES(8,583);
INSERT INTO playlist_track VALUES(8,584);
INSERT INTO playlist_track VALUES(8,585);
INSERT INTO playlist_track VALUES(8,586);
INSERT INTO playlist_track VALUES(8,587);
INSERT INTO playlist_track VALUES(8,588);
INSERT INTO playlist_track VALUES(8,589);
INSERT INTO playlist_track VALUES(8,590);
INSERT INTO playlist_track VALUES(8,591);
INSERT INTO playlist_track VALUES(8,592);
INSERT INTO playlist_track VALUES(8,593);
INSERT INTO playlist_track VALUES(8,594);
INSERT INTO playlist_track VALUES(8,595);
INSERT INTO playlist_track VALUES(8,596);
INSERT INTO playlist_track VALUES(8,388);
INSERT INTO playlist_track VALUES(8,402);
INSERT INTO playlist_track VALUES(8,407);
INSERT INTO playlist_track VALUES(8,396);
INSERT INTO playlist_track VALUES(8,877);
INSERT INTO playlist_track VALUES(8,878);
INSERT INTO playlist_track VALUES(8,879);
INSERT INTO playlist_track VALUES(8,880);
INSERT INTO playlist_track VALUES(8,881);
INSERT INTO playlist_track VALUES(8,882);
INSERT INTO playlist_track VALUES(8,883);
INSERT INTO playlist_track VALUES(8,884);
INSERT INTO playlist_track VALUES(8,885);
INSERT INTO playlist_track VALUES(8,886);
INSERT INTO playlist_track VALUES(8,887);
INSERT INTO playlist_track VALUES(8,888);
INSERT INTO playlist_track VALUES(8,889);
INSERT INTO playlist_track VALUES(8,890);
INSERT INTO playlist_track VALUES(8,3405);
INSERT INTO playlist_track VALUES(8,891);
INSERT INTO playlist_track VALUES(8,892);
INSERT INTO playlist_track VALUES(8,893);
INSERT INTO playlist_track VALUES(8,894);
INSERT INTO playlist_track VALUES(8,895);
INSERT INTO playlist_track VALUES(8,896);
INSERT INTO playlist_track VALUES(8,897);
INSERT INTO playlist_track VALUES(8,898);
INSERT INTO playlist_track VALUES(8,899);
INSERT INTO playlist_track VALUES(8,900);
INSERT INTO playlist_track VALUES(8,901);
INSERT INTO playlist_track VALUES(8,902);
INSERT INTO playlist_track VALUES(8,903);
INSERT INTO playlist_track VALUES(8,904);
INSERT INTO playlist_track VALUES(8,905);
INSERT INTO playlist_track VALUES(8,906);
INSERT INTO playlist_track VALUES(8,907);
INSERT INTO playlist_track VALUES(8,908);
INSERT INTO playlist_track VALUES(8,909);
INSERT INTO playlist_track VALUES(8,910);
INSERT INTO playlist_track VALUES(8,911);
INSERT INTO playlist_track VALUES(8,912);
INSERT INTO playlist_track VALUES(8,913);
INSERT INTO playlist_track VALUES(8,914);
INSERT INTO playlist_track VALUES(8,915);
INSERT INTO playlist_track VALUES(8,916);
INSERT INTO playlist_track VALUES(8,917);
INSERT INTO playlist_track VALUES(8,918);
INSERT INTO playlist_track VALUES(8,919);
INSERT INTO playlist_track VALUES(8,920);
INSERT INTO playlist_track VALUES(8,921);
INSERT INTO playlist_track VALUES(8,922);
INSERT INTO playlist_track VALUES(8,3423);
INSERT INTO playlist_track VALUES(8,923);
INSERT INTO playlist_track VALUES(8,924);
INSERT INTO playlist_track VALUES(8,925);
INSERT INTO playlist_track VALUES(8,926);
INSERT INTO playlist_track VALUES(8,927);
INSERT INTO playlist_track VALUES(8,928);
INSERT INTO playlist_track VALUES(8,929);
INSERT INTO playlist_track VALUES(8,930);
INSERT INTO playlist_track VALUES(8,931);
INSERT INTO playlist_track VALUES(8,932);
INSERT INTO playlist_track VALUES(8,933);
INSERT INTO playlist_track VALUES(8,934);
INSERT INTO playlist_track VALUES(8,935);
INSERT INTO playlist_track VALUES(8,936);
INSERT INTO playlist_track VALUES(8,937);
INSERT INTO playlist_track VALUES(8,938);
INSERT INTO playlist_track VALUES(8,939);
INSERT INTO playlist_track VALUES(8,940);
INSERT INTO playlist_track VALUES(8,941);
INSERT INTO playlist_track VALUES(8,942);
INSERT INTO playlist_track VALUES(8,943);
INSERT INTO playlist_track VALUES(8,944);
INSERT INTO playlist_track VALUES(8,945);
INSERT INTO playlist_track VALUES(8,946);
INSERT INTO playlist_track VALUES(8,947);
INSERT INTO playlist_track VALUES(8,948);
INSERT INTO playlist_track VALUES(8,949);
INSERT INTO playlist_track VALUES(8,950);
INSERT INTO playlist_track VALUES(8,951);
INSERT INTO playlist_track VALUES(8,952);
INSERT INTO playlist_track VALUES(8,953);
INSERT INTO playlist_track VALUES(8,954);
INSERT INTO playlist_track VALUES(8,955);
INSERT INTO playlist_track VALUES(8,956);
INSERT INTO playlist_track VALUES(8,957);
INSERT INTO playlist_track VALUES(8,958);
INSERT INTO playlist_track VALUES(8,959);
INSERT INTO playlist_track VALUES(8,960);
INSERT INTO playlist_track VALUES(8,961);
INSERT INTO playlist_track VALUES(8,962);
INSERT INTO playlist_track VALUES(8,963);
INSERT INTO playlist_track VALUES(8,964);
INSERT INTO playlist_track VALUES(8,965);
INSERT INTO playlist_track VALUES(8,966);
INSERT INTO playlist_track VALUES(8,967);
INSERT INTO playlist_track VALUES(8,968);
INSERT INTO playlist_track VALUES(8,969);
INSERT INTO playlist_track VALUES(8,970);
INSERT INTO playlist_track VALUES(8,971);
INSERT INTO playlist_track VALUES(8,972);
INSERT INTO playlist_track VALUES(8,973);
INSERT INTO playlist_track VALUES(8,974);
INSERT INTO playlist_track VALUES(8,975);
INSERT INTO playlist_track VALUES(8,976);
INSERT INTO playlist_track VALUES(8,977);
INSERT INTO playlist_track VALUES(8,978);
INSERT INTO playlist_track VALUES(8,979);
INSERT INTO playlist_track VALUES(8,980);
INSERT INTO playlist_track VALUES(8,981);
INSERT INTO playlist_track VALUES(8,982);
INSERT INTO playlist_track VALUES(8,983);
INSERT INTO playlist_track VALUES(8,984);
INSERT INTO playlist_track VALUES(8,985);
INSERT INTO playlist_track VALUES(8,986);
INSERT INTO playlist_track VALUES(8,987);
INSERT INTO playlist_track VALUES(8,988);
INSERT INTO playlist_track VALUES(8,390);
INSERT INTO playlist_track VALUES(8,3273);
INSERT INTO playlist_track VALUES(8,1020);
INSERT INTO playlist_track VALUES(8,1021);
INSERT INTO playlist_track VALUES(8,1022);
INSERT INTO playlist_track VALUES(8,1023);
INSERT INTO playlist_track VALUES(8,1024);
INSERT INTO playlist_track VALUES(8,1025);
INSERT INTO playlist_track VALUES(8,1026);
INSERT INTO playlist_track VALUES(8,1027);
INSERT INTO playlist_track VALUES(8,1028);
INSERT INTO playlist_track VALUES(8,1029);
INSERT INTO playlist_track VALUES(8,1030);
INSERT INTO playlist_track VALUES(8,1031);
INSERT INTO playlist_track VALUES(8,1032);
INSERT INTO playlist_track VALUES(8,989);
INSERT INTO playlist_track VALUES(8,990);
INSERT INTO playlist_track VALUES(8,991);
INSERT INTO playlist_track VALUES(8,992);
INSERT INTO playlist_track VALUES(8,993);
INSERT INTO playlist_track VALUES(8,994);
INSERT INTO playlist_track VALUES(8,995);
INSERT INTO playlist_track VALUES(8,996);
INSERT INTO playlist_track VALUES(8,997);
INSERT INTO playlist_track VALUES(8,998);
INSERT INTO playlist_track VALUES(8,999);
INSERT INTO playlist_track VALUES(8,1000);
INSERT INTO playlist_track VALUES(8,1001);
INSERT INTO playlist_track VALUES(8,1002);
INSERT INTO playlist_track VALUES(8,1003);
INSERT INTO playlist_track VALUES(8,1004);
INSERT INTO playlist_track VALUES(8,1005);
INSERT INTO playlist_track VALUES(8,1006);
INSERT INTO playlist_track VALUES(8,1007);
INSERT INTO playlist_track VALUES(8,1008);
INSERT INTO playlist_track VALUES(8,1009);
INSERT INTO playlist_track VALUES(8,1010);
INSERT INTO playlist_track VALUES(8,1011);
INSERT INTO playlist_track VALUES(8,1012);
INSERT INTO playlist_track VALUES(8,1013);
INSERT INTO playlist_track VALUES(8,1014);
INSERT INTO playlist_track VALUES(8,1015);
INSERT INTO playlist_track VALUES(8,1016);
INSERT INTO playlist_track VALUES(8,1017);
INSERT INTO playlist_track VALUES(8,1018);
INSERT INTO playlist_track VALUES(8,1019);
INSERT INTO playlist_track VALUES(8,1033);
INSERT INTO playlist_track VALUES(8,1034);
INSERT INTO playlist_track VALUES(8,1035);
INSERT INTO playlist_track VALUES(8,1036);
INSERT INTO playlist_track VALUES(8,1037);
INSERT INTO playlist_track VALUES(8,1038);
INSERT INTO playlist_track VALUES(8,1039);
INSERT INTO playlist_track VALUES(8,1040);
INSERT INTO playlist_track VALUES(8,1041);
INSERT INTO playlist_track VALUES(8,1042);
INSERT INTO playlist_track VALUES(8,1043);
INSERT INTO playlist_track VALUES(8,1044);
INSERT INTO playlist_track VALUES(8,1045);
INSERT INTO playlist_track VALUES(8,1046);
INSERT INTO playlist_track VALUES(8,1047);
INSERT INTO playlist_track VALUES(8,1048);
INSERT INTO playlist_track VALUES(8,1049);
INSERT INTO playlist_track VALUES(8,1050);
INSERT INTO playlist_track VALUES(8,1051);
INSERT INTO playlist_track VALUES(8,1052);
INSERT INTO playlist_track VALUES(8,1053);
INSERT INTO playlist_track VALUES(8,1054);
INSERT INTO playlist_track VALUES(8,1055);
INSERT INTO playlist_track VALUES(8,1056);
INSERT INTO playlist_track VALUES(8,351);
INSERT INTO playlist_track VALUES(8,352);
INSERT INTO playlist_track VALUES(8,353);
INSERT INTO playlist_track VALUES(8,354);
INSERT INTO playlist_track VALUES(8,355);
INSERT INTO playlist_track VALUES(8,356);
INSERT INTO playlist_track VALUES(8,357);
INSERT INTO playlist_track VALUES(8,358);
INSERT INTO playlist_track VALUES(8,359);
INSERT INTO playlist_track VALUES(8,3332);
INSERT INTO playlist_track VALUES(8,1057);
INSERT INTO playlist_track VALUES(8,1058);
INSERT INTO playlist_track VALUES(8,1059);
INSERT INTO playlist_track VALUES(8,1060);
INSERT INTO playlist_track VALUES(8,1061);
INSERT INTO playlist_track VALUES(8,1062);
INSERT INTO playlist_track VALUES(8,1063);
INSERT INTO playlist_track VALUES(8,1064);
INSERT INTO playlist_track VALUES(8,1065);
INSERT INTO playlist_track VALUES(8,1066);
INSERT INTO playlist_track VALUES(8,1067);
INSERT INTO playlist_track VALUES(8,1068);
INSERT INTO playlist_track VALUES(8,1069);
INSERT INTO playlist_track VALUES(8,1070);
INSERT INTO playlist_track VALUES(8,1071);
INSERT INTO playlist_track VALUES(8,1072);
INSERT INTO playlist_track VALUES(8,624);
INSERT INTO playlist_track VALUES(8,625);
INSERT INTO playlist_track VALUES(8,626);
INSERT INTO playlist_track VALUES(8,627);
INSERT INTO playlist_track VALUES(8,628);
INSERT INTO playlist_track VALUES(8,629);
INSERT INTO playlist_track VALUES(8,630);
INSERT INTO playlist_track VALUES(8,631);
INSERT INTO playlist_track VALUES(8,632);
INSERT INTO playlist_track VALUES(8,633);
INSERT INTO playlist_track VALUES(8,634);
INSERT INTO playlist_track VALUES(8,635);
INSERT INTO playlist_track VALUES(8,636);
INSERT INTO playlist_track VALUES(8,637);
INSERT INTO playlist_track VALUES(8,638);
INSERT INTO playlist_track VALUES(8,639);
INSERT INTO playlist_track VALUES(8,640);
INSERT INTO playlist_track VALUES(8,641);
INSERT INTO playlist_track VALUES(8,642);
INSERT INTO playlist_track VALUES(8,643);
INSERT INTO playlist_track VALUES(8,644);
INSERT INTO playlist_track VALUES(8,645);
INSERT INTO playlist_track VALUES(8,1073);
INSERT INTO playlist_track VALUES(8,1074);
INSERT INTO playlist_track VALUES(8,1075);
INSERT INTO playlist_track VALUES(8,1076);
INSERT INTO playlist_track VALUES(8,1077);
INSERT INTO playlist_track VALUES(8,1078);
INSERT INTO playlist_track VALUES(8,1079);
INSERT INTO playlist_track VALUES(8,1080);
INSERT INTO playlist_track VALUES(8,1081);
INSERT INTO playlist_track VALUES(8,1082);
INSERT INTO playlist_track VALUES(8,1083);
INSERT INTO playlist_track VALUES(8,1084);
INSERT INTO playlist_track VALUES(8,1085);
INSERT INTO playlist_track VALUES(8,1086);
INSERT INTO playlist_track VALUES(8,377);
INSERT INTO playlist_track VALUES(8,395);
INSERT INTO playlist_track VALUES(8,1102);
INSERT INTO playlist_track VALUES(8,1103);
INSERT INTO playlist_track VALUES(8,1104);
INSERT INTO playlist_track VALUES(8,1087);
INSERT INTO playlist_track VALUES(8,1088);
INSERT INTO playlist_track VALUES(8,1089);
INSERT INTO playlist_track VALUES(8,1090);
INSERT INTO playlist_track VALUES(8,1091);
INSERT INTO playlist_track VALUES(8,1092);
INSERT INTO playlist_track VALUES(8,1093);
INSERT INTO playlist_track VALUES(8,1094);
INSERT INTO playlist_track VALUES(8,1095);
INSERT INTO playlist_track VALUES(8,1096);
INSERT INTO playlist_track VALUES(8,1097);
INSERT INTO playlist_track VALUES(8,1098);
INSERT INTO playlist_track VALUES(8,1099);
INSERT INTO playlist_track VALUES(8,1100);
INSERT INTO playlist_track VALUES(8,1101);
INSERT INTO playlist_track VALUES(8,1105);
INSERT INTO playlist_track VALUES(8,1106);
INSERT INTO playlist_track VALUES(8,1107);
INSERT INTO playlist_track VALUES(8,1108);
INSERT INTO playlist_track VALUES(8,1109);
INSERT INTO playlist_track VALUES(8,1110);
INSERT INTO playlist_track VALUES(8,1111);
INSERT INTO playlist_track VALUES(8,1112);
INSERT INTO playlist_track VALUES(8,1113);
INSERT INTO playlist_track VALUES(8,1114);
INSERT INTO playlist_track VALUES(8,1115);
INSERT INTO playlist_track VALUES(8,1116);
INSERT INTO playlist_track VALUES(8,1117);
INSERT INTO playlist_track VALUES(8,1118);
INSERT INTO playlist_track VALUES(8,1119);
INSERT INTO playlist_track VALUES(8,1120);
INSERT INTO playlist_track VALUES(8,1121);
INSERT INTO playlist_track VALUES(8,1122);
INSERT INTO playlist_track VALUES(8,1123);
INSERT INTO playlist_track VALUES(8,1124);
INSERT INTO playlist_track VALUES(8,1125);
INSERT INTO playlist_track VALUES(8,1126);
INSERT INTO playlist_track VALUES(8,1127);
INSERT INTO playlist_track VALUES(8,1128);
INSERT INTO playlist_track VALUES(8,1129);
INSERT INTO playlist_track VALUES(8,1130);
INSERT INTO playlist_track VALUES(8,1131);
INSERT INTO playlist_track VALUES(8,1132);
INSERT INTO playlist_track VALUES(8,501);
INSERT INTO playlist_track VALUES(8,502);
INSERT INTO playlist_track VALUES(8,503);
INSERT INTO playlist_track VALUES(8,504);
INSERT INTO playlist_track VALUES(8,505);
INSERT INTO playlist_track VALUES(8,506);
INSERT INTO playlist_track VALUES(8,507);
INSERT INTO playlist_track VALUES(8,508);
INSERT INTO playlist_track VALUES(8,509);
INSERT INTO playlist_track VALUES(8,510);
INSERT INTO playlist_track VALUES(8,511);
INSERT INTO playlist_track VALUES(8,512);
INSERT INTO playlist_track VALUES(8,513);
INSERT INTO playlist_track VALUES(8,514);
INSERT INTO playlist_track VALUES(8,1133);
INSERT INTO playlist_track VALUES(8,1134);
INSERT INTO playlist_track VALUES(8,1135);
INSERT INTO playlist_track VALUES(8,1136);
INSERT INTO playlist_track VALUES(8,1137);
INSERT INTO playlist_track VALUES(8,1138);
INSERT INTO playlist_track VALUES(8,1139);
INSERT INTO playlist_track VALUES(8,1140);
INSERT INTO playlist_track VALUES(8,1141);
INSERT INTO playlist_track VALUES(8,1142);
INSERT INTO playlist_track VALUES(8,1143);
INSERT INTO playlist_track VALUES(8,1144);
INSERT INTO playlist_track VALUES(8,1145);
INSERT INTO playlist_track VALUES(8,3265);
INSERT INTO playlist_track VALUES(8,468);
INSERT INTO playlist_track VALUES(8,469);
INSERT INTO playlist_track VALUES(8,470);
INSERT INTO playlist_track VALUES(8,471);
INSERT INTO playlist_track VALUES(8,472);
INSERT INTO playlist_track VALUES(8,473);
INSERT INTO playlist_track VALUES(8,474);
INSERT INTO playlist_track VALUES(8,475);
INSERT INTO playlist_track VALUES(8,476);
INSERT INTO playlist_track VALUES(8,477);
INSERT INTO playlist_track VALUES(8,478);
INSERT INTO playlist_track VALUES(8,479);
INSERT INTO playlist_track VALUES(8,480);
INSERT INTO playlist_track VALUES(8,481);
INSERT INTO playlist_track VALUES(8,482);
INSERT INTO playlist_track VALUES(8,483);
INSERT INTO playlist_track VALUES(8,484);
INSERT INTO playlist_track VALUES(8,485);
INSERT INTO playlist_track VALUES(8,486);
INSERT INTO playlist_track VALUES(8,487);
INSERT INTO playlist_track VALUES(8,488);
INSERT INTO playlist_track VALUES(8,1146);
INSERT INTO playlist_track VALUES(8,1147);
INSERT INTO playlist_track VALUES(8,1148);
INSERT INTO playlist_track VALUES(8,1149);
INSERT INTO playlist_track VALUES(8,1150);
INSERT INTO playlist_track VALUES(8,1151);
INSERT INTO playlist_track VALUES(8,1152);
INSERT INTO playlist_track VALUES(8,1153);
INSERT INTO playlist_track VALUES(8,1154);
INSERT INTO playlist_track VALUES(8,1155);
INSERT INTO playlist_track VALUES(8,1156);
INSERT INTO playlist_track VALUES(8,1157);
INSERT INTO playlist_track VALUES(8,1158);
INSERT INTO playlist_track VALUES(8,1159);
INSERT INTO playlist_track VALUES(8,1160);
INSERT INTO playlist_track VALUES(8,1161);
INSERT INTO playlist_track VALUES(8,1162);
INSERT INTO playlist_track VALUES(8,1163);
INSERT INTO playlist_track VALUES(8,1164);
INSERT INTO playlist_track VALUES(8,1165);
INSERT INTO playlist_track VALUES(8,1166);
INSERT INTO playlist_track VALUES(8,1167);
INSERT INTO playlist_track VALUES(8,1168);
INSERT INTO playlist_track VALUES(8,1169);
INSERT INTO playlist_track VALUES(8,1170);
INSERT INTO playlist_track VALUES(8,1171);
INSERT INTO playlist_track VALUES(8,1172);
INSERT INTO playlist_track VALUES(8,1173);
INSERT INTO playlist_track VALUES(8,1174);
INSERT INTO playlist_track VALUES(8,1175);
INSERT INTO playlist_track VALUES(8,1176);
INSERT INTO playlist_track VALUES(8,1177);
INSERT INTO playlist_track VALUES(8,1178);
INSERT INTO playlist_track VALUES(8,1179);
INSERT INTO playlist_track VALUES(8,1180);
INSERT INTO playlist_track VALUES(8,1181);
INSERT INTO playlist_track VALUES(8,1182);
INSERT INTO playlist_track VALUES(8,1183);
INSERT INTO playlist_track VALUES(8,1184);
INSERT INTO playlist_track VALUES(8,1185);
INSERT INTO playlist_track VALUES(8,1186);
INSERT INTO playlist_track VALUES(8,1187);
INSERT INTO playlist_track VALUES(8,3322);
INSERT INTO playlist_track VALUES(8,3354);
INSERT INTO playlist_track VALUES(8,3351);
INSERT INTO playlist_track VALUES(8,3422);
INSERT INTO playlist_track VALUES(8,405);
INSERT INTO playlist_track VALUES(8,3407);
INSERT INTO playlist_track VALUES(8,3301);
INSERT INTO playlist_track VALUES(8,3300);
INSERT INTO playlist_track VALUES(8,3302);
INSERT INTO playlist_track VALUES(8,3303);
INSERT INTO playlist_track VALUES(8,3304);
INSERT INTO playlist_track VALUES(8,3305);
INSERT INTO playlist_track VALUES(8,3306);
INSERT INTO playlist_track VALUES(8,3307);
INSERT INTO playlist_track VALUES(8,3308);
INSERT INTO playlist_track VALUES(8,3309);
INSERT INTO playlist_track VALUES(8,3310);
INSERT INTO playlist_track VALUES(8,3311);
INSERT INTO playlist_track VALUES(8,3312);
INSERT INTO playlist_track VALUES(8,3313);
INSERT INTO playlist_track VALUES(8,3314);
INSERT INTO playlist_track VALUES(8,3315);
INSERT INTO playlist_track VALUES(8,3316);
INSERT INTO playlist_track VALUES(8,3317);
INSERT INTO playlist_track VALUES(8,3318);
INSERT INTO playlist_track VALUES(8,1188);
INSERT INTO playlist_track VALUES(8,1189);
INSERT INTO playlist_track VALUES(8,1190);
INSERT INTO playlist_track VALUES(8,1191);
INSERT INTO playlist_track VALUES(8,1192);
INSERT INTO playlist_track VALUES(8,1193);
INSERT INTO playlist_track VALUES(8,1194);
INSERT INTO playlist_track VALUES(8,1195);
INSERT INTO playlist_track VALUES(8,1196);
INSERT INTO playlist_track VALUES(8,1197);
INSERT INTO playlist_track VALUES(8,1198);
INSERT INTO playlist_track VALUES(8,1199);
INSERT INTO playlist_track VALUES(8,1200);
INSERT INTO playlist_track VALUES(8,3329);
INSERT INTO playlist_track VALUES(8,1235);
INSERT INTO playlist_track VALUES(8,1236);
INSERT INTO playlist_track VALUES(8,1237);
INSERT INTO playlist_track VALUES(8,1238);
INSERT INTO playlist_track VALUES(8,1239);
INSERT INTO playlist_track VALUES(8,1240);
INSERT INTO playlist_track VALUES(8,1241);
INSERT INTO playlist_track VALUES(8,1242);
INSERT INTO playlist_track VALUES(8,1243);
INSERT INTO playlist_track VALUES(8,1244);
INSERT INTO playlist_track VALUES(8,1245);
INSERT INTO playlist_track VALUES(8,1246);
INSERT INTO playlist_track VALUES(8,1247);
INSERT INTO playlist_track VALUES(8,1248);
INSERT INTO playlist_track VALUES(8,1249);
INSERT INTO playlist_track VALUES(8,1250);
INSERT INTO playlist_track VALUES(8,1251);
INSERT INTO playlist_track VALUES(8,1252);
INSERT INTO playlist_track VALUES(8,1253);
INSERT INTO playlist_track VALUES(8,1254);
INSERT INTO playlist_track VALUES(8,1255);
INSERT INTO playlist_track VALUES(8,1256);
INSERT INTO playlist_track VALUES(8,1257);
INSERT INTO playlist_track VALUES(8,1258);
INSERT INTO playlist_track VALUES(8,1259);
INSERT INTO playlist_track VALUES(8,1260);
INSERT INTO playlist_track VALUES(8,1261);
INSERT INTO playlist_track VALUES(8,1262);
INSERT INTO playlist_track VALUES(8,1263);
INSERT INTO playlist_track VALUES(8,1264);
INSERT INTO playlist_track VALUES(8,1265);
INSERT INTO playlist_track VALUES(8,1266);
INSERT INTO playlist_track VALUES(8,1267);
INSERT INTO playlist_track VALUES(8,1268);
INSERT INTO playlist_track VALUES(8,1269);
INSERT INTO playlist_track VALUES(8,1270);
INSERT INTO playlist_track VALUES(8,1271);
INSERT INTO playlist_track VALUES(8,1272);
INSERT INTO playlist_track VALUES(8,1273);
INSERT INTO playlist_track VALUES(8,1274);
INSERT INTO playlist_track VALUES(8,1275);
INSERT INTO playlist_track VALUES(8,1276);
INSERT INTO playlist_track VALUES(8,1277);
INSERT INTO playlist_track VALUES(8,1278);
INSERT INTO playlist_track VALUES(8,1279);
INSERT INTO playlist_track VALUES(8,1280);
INSERT INTO playlist_track VALUES(8,1281);
INSERT INTO playlist_track VALUES(8,1282);
INSERT INTO playlist_track VALUES(8,1283);
INSERT INTO playlist_track VALUES(8,1284);
INSERT INTO playlist_track VALUES(8,1285);
INSERT INTO playlist_track VALUES(8,1286);
INSERT INTO playlist_track VALUES(8,1287);
INSERT INTO playlist_track VALUES(8,1288);
INSERT INTO playlist_track VALUES(8,1289);
INSERT INTO playlist_track VALUES(8,1290);
INSERT INTO playlist_track VALUES(8,1291);
INSERT INTO playlist_track VALUES(8,1292);
INSERT INTO playlist_track VALUES(8,1293);
INSERT INTO playlist_track VALUES(8,1294);
INSERT INTO playlist_track VALUES(8,1295);
INSERT INTO playlist_track VALUES(8,1296);
INSERT INTO playlist_track VALUES(8,1297);
INSERT INTO playlist_track VALUES(8,1298);
INSERT INTO playlist_track VALUES(8,1299);
INSERT INTO playlist_track VALUES(8,1300);
INSERT INTO playlist_track VALUES(8,1301);
INSERT INTO playlist_track VALUES(8,1302);
INSERT INTO playlist_track VALUES(8,1303);
INSERT INTO playlist_track VALUES(8,1304);
INSERT INTO playlist_track VALUES(8,1305);
INSERT INTO playlist_track VALUES(8,1306);
INSERT INTO playlist_track VALUES(8,1307);
INSERT INTO playlist_track VALUES(8,1308);
INSERT INTO playlist_track VALUES(8,1309);
INSERT INTO playlist_track VALUES(8,1310);
INSERT INTO playlist_track VALUES(8,1311);
INSERT INTO playlist_track VALUES(8,1312);
INSERT INTO playlist_track VALUES(8,1313);
INSERT INTO playlist_track VALUES(8,1314);
INSERT INTO playlist_track VALUES(8,1315);
INSERT INTO playlist_track VALUES(8,1316);
INSERT INTO playlist_track VALUES(8,1317);
INSERT INTO playlist_track VALUES(8,1318);
INSERT INTO playlist_track VALUES(8,1319);
INSERT INTO playlist_track VALUES(8,1320);
INSERT INTO playlist_track VALUES(8,1321);
INSERT INTO playlist_track VALUES(8,1322);
INSERT INTO playlist_track VALUES(8,1323);
INSERT INTO playlist_track VALUES(8,1324);
INSERT INTO playlist_track VALUES(8,1201);
INSERT INTO playlist_track VALUES(8,1202);
INSERT INTO playlist_track VALUES(8,1203);
INSERT INTO playlist_track VALUES(8,1204);
INSERT INTO playlist_track VALUES(8,1205);
INSERT INTO playlist_track VALUES(8,1206);
INSERT INTO playlist_track VALUES(8,1207);
INSERT INTO playlist_track VALUES(8,1208);
INSERT INTO playlist_track VALUES(8,1209);
INSERT INTO playlist_track VALUES(8,1210);
INSERT INTO playlist_track VALUES(8,1211);
INSERT INTO playlist_track VALUES(8,1325);
INSERT INTO playlist_track VALUES(8,1326);
INSERT INTO playlist_track VALUES(8,1327);
INSERT INTO playlist_track VALUES(8,1328);
INSERT INTO playlist_track VALUES(8,1329);
INSERT INTO playlist_track VALUES(8,1330);
INSERT INTO playlist_track VALUES(8,1331);
INSERT INTO playlist_track VALUES(8,1332);
INSERT INTO playlist_track VALUES(8,1333);
INSERT INTO playlist_track VALUES(8,1334);
INSERT INTO playlist_track VALUES(8,1391);
INSERT INTO playlist_track VALUES(8,1393);
INSERT INTO playlist_track VALUES(8,1388);
INSERT INTO playlist_track VALUES(8,1394);
INSERT INTO playlist_track VALUES(8,1387);
INSERT INTO playlist_track VALUES(8,1392);
INSERT INTO playlist_track VALUES(8,1389);
INSERT INTO playlist_track VALUES(8,1390);
INSERT INTO playlist_track VALUES(8,1335);
INSERT INTO playlist_track VALUES(8,1336);
INSERT INTO playlist_track VALUES(8,1337);
INSERT INTO playlist_track VALUES(8,1338);
INSERT INTO playlist_track VALUES(8,1339);
INSERT INTO playlist_track VALUES(8,1340);
INSERT INTO playlist_track VALUES(8,1341);
INSERT INTO playlist_track VALUES(8,1342);
INSERT INTO playlist_track VALUES(8,1343);
INSERT INTO playlist_track VALUES(8,1344);
INSERT INTO playlist_track VALUES(8,1345);
INSERT INTO playlist_track VALUES(8,1346);
INSERT INTO playlist_track VALUES(8,1347);
INSERT INTO playlist_track VALUES(8,1348);
INSERT INTO playlist_track VALUES(8,1349);
INSERT INTO playlist_track VALUES(8,1350);
INSERT INTO playlist_track VALUES(8,1351);
INSERT INTO playlist_track VALUES(8,1212);
INSERT INTO playlist_track VALUES(8,1213);
INSERT INTO playlist_track VALUES(8,1214);
INSERT INTO playlist_track VALUES(8,1215);
INSERT INTO playlist_track VALUES(8,1216);
INSERT INTO playlist_track VALUES(8,1217);
INSERT INTO playlist_track VALUES(8,1218);
INSERT INTO playlist_track VALUES(8,1219);
INSERT INTO playlist_track VALUES(8,1220);
INSERT INTO playlist_track VALUES(8,1221);
INSERT INTO playlist_track VALUES(8,1222);
INSERT INTO playlist_track VALUES(8,1223);
INSERT INTO playlist_track VALUES(8,1224);
INSERT INTO playlist_track VALUES(8,1225);
INSERT INTO playlist_track VALUES(8,1226);
INSERT INTO playlist_track VALUES(8,1227);
INSERT INTO playlist_track VALUES(8,1228);
INSERT INTO playlist_track VALUES(8,1229);
INSERT INTO playlist_track VALUES(8,1230);
INSERT INTO playlist_track VALUES(8,1231);
INSERT INTO playlist_track VALUES(8,1232);
INSERT INTO playlist_track VALUES(8,1233);
INSERT INTO playlist_track VALUES(8,1234);
INSERT INTO playlist_track VALUES(8,1352);
INSERT INTO playlist_track VALUES(8,1353);
INSERT INTO playlist_track VALUES(8,1354);
INSERT INTO playlist_track VALUES(8,1355);
INSERT INTO playlist_track VALUES(8,1356);
INSERT INTO playlist_track VALUES(8,1357);
INSERT INTO playlist_track VALUES(8,1358);
INSERT INTO playlist_track VALUES(8,1359);
INSERT INTO playlist_track VALUES(8,1360);
INSERT INTO playlist_track VALUES(8,1361);
INSERT INTO playlist_track VALUES(8,1362);
INSERT INTO playlist_track VALUES(8,1363);
INSERT INTO playlist_track VALUES(8,1364);
INSERT INTO playlist_track VALUES(8,1365);
INSERT INTO playlist_track VALUES(8,1366);
INSERT INTO playlist_track VALUES(8,1367);
INSERT INTO playlist_track VALUES(8,1368);
INSERT INTO playlist_track VALUES(8,1369);
INSERT INTO playlist_track VALUES(8,1370);
INSERT INTO playlist_track VALUES(8,1371);
INSERT INTO playlist_track VALUES(8,1372);
INSERT INTO playlist_track VALUES(8,1373);
INSERT INTO playlist_track VALUES(8,1374);
INSERT INTO playlist_track VALUES(8,1375);
INSERT INTO playlist_track VALUES(8,1376);
INSERT INTO playlist_track VALUES(8,1377);
INSERT INTO playlist_track VALUES(8,1378);
INSERT INTO playlist_track VALUES(8,1379);
INSERT INTO playlist_track VALUES(8,1380);
INSERT INTO playlist_track VALUES(8,1381);
INSERT INTO playlist_track VALUES(8,1382);
INSERT INTO playlist_track VALUES(8,1386);
INSERT INTO playlist_track VALUES(8,1383);
INSERT INTO playlist_track VALUES(8,1385);
INSERT INTO playlist_track VALUES(8,1384);
INSERT INTO playlist_track VALUES(8,1406);
INSERT INTO playlist_track VALUES(8,1407);
INSERT INTO playlist_track VALUES(8,1408);
INSERT INTO playlist_track VALUES(8,1409);
INSERT INTO playlist_track VALUES(8,1410);
INSERT INTO playlist_track VALUES(8,1411);
INSERT INTO playlist_track VALUES(8,1412);
INSERT INTO playlist_track VALUES(8,1413);
INSERT INTO playlist_track VALUES(8,1395);
INSERT INTO playlist_track VALUES(8,1396);
INSERT INTO playlist_track VALUES(8,1397);
INSERT INTO playlist_track VALUES(8,1398);
INSERT INTO playlist_track VALUES(8,1399);
INSERT INTO playlist_track VALUES(8,1400);
INSERT INTO playlist_track VALUES(8,1401);
INSERT INTO playlist_track VALUES(8,1402);
INSERT INTO playlist_track VALUES(8,1403);
INSERT INTO playlist_track VALUES(8,1404);
INSERT INTO playlist_track VALUES(8,1405);
INSERT INTO playlist_track VALUES(8,3274);
INSERT INTO playlist_track VALUES(8,3267);
INSERT INTO playlist_track VALUES(8,3261);
INSERT INTO playlist_track VALUES(8,3272);
INSERT INTO playlist_track VALUES(8,1414);
INSERT INTO playlist_track VALUES(8,1415);
INSERT INTO playlist_track VALUES(8,1416);
INSERT INTO playlist_track VALUES(8,1417);
INSERT INTO playlist_track VALUES(8,1418);
INSERT INTO playlist_track VALUES(8,1419);
INSERT INTO playlist_track VALUES(8,1420);
INSERT INTO playlist_track VALUES(8,1421);
INSERT INTO playlist_track VALUES(8,1422);
INSERT INTO playlist_track VALUES(8,1423);
INSERT INTO playlist_track VALUES(8,1424);
INSERT INTO playlist_track VALUES(8,1425);
INSERT INTO playlist_track VALUES(8,1426);
INSERT INTO playlist_track VALUES(8,1427);
INSERT INTO playlist_track VALUES(8,1428);
INSERT INTO playlist_track VALUES(8,1429);
INSERT INTO playlist_track VALUES(8,1430);
INSERT INTO playlist_track VALUES(8,1431);
INSERT INTO playlist_track VALUES(8,1432);
INSERT INTO playlist_track VALUES(8,1433);
INSERT INTO playlist_track VALUES(8,1434);
INSERT INTO playlist_track VALUES(8,1435);
INSERT INTO playlist_track VALUES(8,1436);
INSERT INTO playlist_track VALUES(8,1437);
INSERT INTO playlist_track VALUES(8,1438);
INSERT INTO playlist_track VALUES(8,1439);
INSERT INTO playlist_track VALUES(8,1440);
INSERT INTO playlist_track VALUES(8,1441);
INSERT INTO playlist_track VALUES(8,1442);
INSERT INTO playlist_track VALUES(8,1443);
INSERT INTO playlist_track VALUES(8,1455);
INSERT INTO playlist_track VALUES(8,1456);
INSERT INTO playlist_track VALUES(8,1457);
INSERT INTO playlist_track VALUES(8,1458);
INSERT INTO playlist_track VALUES(8,1459);
INSERT INTO playlist_track VALUES(8,1460);
INSERT INTO playlist_track VALUES(8,1461);
INSERT INTO playlist_track VALUES(8,1462);
INSERT INTO playlist_track VALUES(8,1463);
INSERT INTO playlist_track VALUES(8,1464);
INSERT INTO playlist_track VALUES(8,1465);
INSERT INTO playlist_track VALUES(8,1444);
INSERT INTO playlist_track VALUES(8,1445);
INSERT INTO playlist_track VALUES(8,1446);
INSERT INTO playlist_track VALUES(8,1447);
INSERT INTO playlist_track VALUES(8,1448);
INSERT INTO playlist_track VALUES(8,1449);
INSERT INTO playlist_track VALUES(8,1450);
INSERT INTO playlist_track VALUES(8,1451);
INSERT INTO playlist_track VALUES(8,1452);
INSERT INTO playlist_track VALUES(8,1453);
INSERT INTO playlist_track VALUES(8,1454);
INSERT INTO playlist_track VALUES(8,1466);
INSERT INTO playlist_track VALUES(8,1467);
INSERT INTO playlist_track VALUES(8,1468);
INSERT INTO playlist_track VALUES(8,1469);
INSERT INTO playlist_track VALUES(8,1470);
INSERT INTO playlist_track VALUES(8,1471);
INSERT INTO playlist_track VALUES(8,1472);
INSERT INTO playlist_track VALUES(8,1473);
INSERT INTO playlist_track VALUES(8,1474);
INSERT INTO playlist_track VALUES(8,1475);
INSERT INTO playlist_track VALUES(8,1476);
INSERT INTO playlist_track VALUES(8,1477);
INSERT INTO playlist_track VALUES(8,1478);
INSERT INTO playlist_track VALUES(8,1479);
INSERT INTO playlist_track VALUES(8,1480);
INSERT INTO playlist_track VALUES(8,1481);
INSERT INTO playlist_track VALUES(8,1482);
INSERT INTO playlist_track VALUES(8,1483);
INSERT INTO playlist_track VALUES(8,1484);
INSERT INTO playlist_track VALUES(8,1485);
INSERT INTO playlist_track VALUES(8,1486);
INSERT INTO playlist_track VALUES(8,1487);
INSERT INTO playlist_track VALUES(8,1488);
INSERT INTO playlist_track VALUES(8,1489);
INSERT INTO playlist_track VALUES(8,1490);
INSERT INTO playlist_track VALUES(8,1491);
INSERT INTO playlist_track VALUES(8,1492);
INSERT INTO playlist_track VALUES(8,1493);
INSERT INTO playlist_track VALUES(8,1494);
INSERT INTO playlist_track VALUES(8,1495);
INSERT INTO playlist_track VALUES(8,378);
INSERT INTO playlist_track VALUES(8,392);
INSERT INTO playlist_track VALUES(8,1532);
INSERT INTO playlist_track VALUES(8,1533);
INSERT INTO playlist_track VALUES(8,1534);
INSERT INTO playlist_track VALUES(8,1535);
INSERT INTO playlist_track VALUES(8,1536);
INSERT INTO playlist_track VALUES(8,1537);
INSERT INTO playlist_track VALUES(8,1538);
INSERT INTO playlist_track VALUES(8,1539);
INSERT INTO playlist_track VALUES(8,1540);
INSERT INTO playlist_track VALUES(8,1541);
INSERT INTO playlist_track VALUES(8,1542);
INSERT INTO playlist_track VALUES(8,1543);
INSERT INTO playlist_track VALUES(8,1544);
INSERT INTO playlist_track VALUES(8,1545);
INSERT INTO playlist_track VALUES(8,1496);
INSERT INTO playlist_track VALUES(8,1497);
INSERT INTO playlist_track VALUES(8,1498);
INSERT INTO playlist_track VALUES(8,1499);
INSERT INTO playlist_track VALUES(8,1500);
INSERT INTO playlist_track VALUES(8,1501);
INSERT INTO playlist_track VALUES(8,1502);
INSERT INTO playlist_track VALUES(8,1503);
INSERT INTO playlist_track VALUES(8,1504);
INSERT INTO playlist_track VALUES(8,1505);
INSERT INTO playlist_track VALUES(8,403);
INSERT INTO playlist_track VALUES(8,1506);
INSERT INTO playlist_track VALUES(8,1507);
INSERT INTO playlist_track VALUES(8,1508);
INSERT INTO playlist_track VALUES(8,1509);
INSERT INTO playlist_track VALUES(8,1510);
INSERT INTO playlist_track VALUES(8,1511);
INSERT INTO playlist_track VALUES(8,1512);
INSERT INTO playlist_track VALUES(8,1513);
INSERT INTO playlist_track VALUES(8,1514);
INSERT INTO playlist_track VALUES(8,1515);
INSERT INTO playlist_track VALUES(8,1516);
INSERT INTO playlist_track VALUES(8,1517);
INSERT INTO playlist_track VALUES(8,1518);
INSERT INTO playlist_track VALUES(8,1519);
INSERT INTO playlist_track VALUES(8,381);
INSERT INTO playlist_track VALUES(8,1520);
INSERT INTO playlist_track VALUES(8,1521);
INSERT INTO playlist_track VALUES(8,1522);
INSERT INTO playlist_track VALUES(8,1523);
INSERT INTO playlist_track VALUES(8,1524);
INSERT INTO playlist_track VALUES(8,1525);
INSERT INTO playlist_track VALUES(8,1526);
INSERT INTO playlist_track VALUES(8,1527);
INSERT INTO playlist_track VALUES(8,1528);
INSERT INTO playlist_track VALUES(8,1529);
INSERT INTO playlist_track VALUES(8,1530);
INSERT INTO playlist_track VALUES(8,1531);
INSERT INTO playlist_track VALUES(8,1546);
INSERT INTO playlist_track VALUES(8,1547);
INSERT INTO playlist_track VALUES(8,1548);
INSERT INTO playlist_track VALUES(8,1549);
INSERT INTO playlist_track VALUES(8,1550);
INSERT INTO playlist_track VALUES(8,1551);
INSERT INTO playlist_track VALUES(8,1552);
INSERT INTO playlist_track VALUES(8,1553);
INSERT INTO playlist_track VALUES(8,1554);
INSERT INTO playlist_track VALUES(8,1555);
INSERT INTO playlist_track VALUES(8,1556);
INSERT INTO playlist_track VALUES(8,1557);
INSERT INTO playlist_track VALUES(8,1558);
INSERT INTO playlist_track VALUES(8,1559);
INSERT INTO playlist_track VALUES(8,1560);
INSERT INTO playlist_track VALUES(8,1561);
INSERT INTO playlist_track VALUES(8,3352);
INSERT INTO playlist_track VALUES(8,3358);
INSERT INTO playlist_track VALUES(8,400);
INSERT INTO playlist_track VALUES(8,436);
INSERT INTO playlist_track VALUES(8,437);
INSERT INTO playlist_track VALUES(8,438);
INSERT INTO playlist_track VALUES(8,439);
INSERT INTO playlist_track VALUES(8,440);
INSERT INTO playlist_track VALUES(8,441);
INSERT INTO playlist_track VALUES(8,442);
INSERT INTO playlist_track VALUES(8,443);
INSERT INTO playlist_track VALUES(8,444);
INSERT INTO playlist_track VALUES(8,445);
INSERT INTO playlist_track VALUES(8,446);
INSERT INTO playlist_track VALUES(8,447);
INSERT INTO playlist_track VALUES(8,448);
INSERT INTO playlist_track VALUES(8,449);
INSERT INTO playlist_track VALUES(8,450);
INSERT INTO playlist_track VALUES(8,451);
INSERT INTO playlist_track VALUES(8,452);
INSERT INTO playlist_track VALUES(8,453);
INSERT INTO playlist_track VALUES(8,454);
INSERT INTO playlist_track VALUES(8,455);
INSERT INTO playlist_track VALUES(8,1562);
INSERT INTO playlist_track VALUES(8,1563);
INSERT INTO playlist_track VALUES(8,1564);
INSERT INTO playlist_track VALUES(8,1565);
INSERT INTO playlist_track VALUES(8,1566);
INSERT INTO playlist_track VALUES(8,1567);
INSERT INTO playlist_track VALUES(8,1568);
INSERT INTO playlist_track VALUES(8,1569);
INSERT INTO playlist_track VALUES(8,1570);
INSERT INTO playlist_track VALUES(8,1571);
INSERT INTO playlist_track VALUES(8,1572);
INSERT INTO playlist_track VALUES(8,1573);
INSERT INTO playlist_track VALUES(8,1574);
INSERT INTO playlist_track VALUES(8,1575);
INSERT INTO playlist_track VALUES(8,1576);
INSERT INTO playlist_track VALUES(8,337);
INSERT INTO playlist_track VALUES(8,338);
INSERT INTO playlist_track VALUES(8,339);
INSERT INTO playlist_track VALUES(8,340);
INSERT INTO playlist_track VALUES(8,341);
INSERT INTO playlist_track VALUES(8,342);
INSERT INTO playlist_track VALUES(8,343);
INSERT INTO playlist_track VALUES(8,344);
INSERT INTO playlist_track VALUES(8,345);
INSERT INTO playlist_track VALUES(8,346);
INSERT INTO playlist_track VALUES(8,347);
INSERT INTO playlist_track VALUES(8,348);
INSERT INTO playlist_track VALUES(8,349);
INSERT INTO playlist_track VALUES(8,350);
INSERT INTO playlist_track VALUES(8,1577);
INSERT INTO playlist_track VALUES(8,1578);
INSERT INTO playlist_track VALUES(8,1579);
INSERT INTO playlist_track VALUES(8,1580);
INSERT INTO playlist_track VALUES(8,1581);
INSERT INTO playlist_track VALUES(8,1582);
INSERT INTO playlist_track VALUES(8,1583);
INSERT INTO playlist_track VALUES(8,1584);
INSERT INTO playlist_track VALUES(8,1585);
INSERT INTO playlist_track VALUES(8,1586);
INSERT INTO playlist_track VALUES(8,1587);
INSERT INTO playlist_track VALUES(8,1588);
INSERT INTO playlist_track VALUES(8,1589);
INSERT INTO playlist_track VALUES(8,1590);
INSERT INTO playlist_track VALUES(8,1591);
INSERT INTO playlist_track VALUES(8,1592);
INSERT INTO playlist_track VALUES(8,1593);
INSERT INTO playlist_track VALUES(8,1594);
INSERT INTO playlist_track VALUES(8,1595);
INSERT INTO playlist_track VALUES(8,1596);
INSERT INTO playlist_track VALUES(8,1597);
INSERT INTO playlist_track VALUES(8,1598);
INSERT INTO playlist_track VALUES(8,1599);
INSERT INTO playlist_track VALUES(8,1600);
INSERT INTO playlist_track VALUES(8,1601);
INSERT INTO playlist_track VALUES(8,1602);
INSERT INTO playlist_track VALUES(8,1603);
INSERT INTO playlist_track VALUES(8,1604);
INSERT INTO playlist_track VALUES(8,1605);
INSERT INTO playlist_track VALUES(8,1606);
INSERT INTO playlist_track VALUES(8,1607);
INSERT INTO playlist_track VALUES(8,1608);
INSERT INTO playlist_track VALUES(8,1609);
INSERT INTO playlist_track VALUES(8,1610);
INSERT INTO playlist_track VALUES(8,1611);
INSERT INTO playlist_track VALUES(8,1612);
INSERT INTO playlist_track VALUES(8,1613);
INSERT INTO playlist_track VALUES(8,1614);
INSERT INTO playlist_track VALUES(8,1615);
INSERT INTO playlist_track VALUES(8,1616);
INSERT INTO playlist_track VALUES(8,1617);
INSERT INTO playlist_track VALUES(8,1618);
INSERT INTO playlist_track VALUES(8,1619);
INSERT INTO playlist_track VALUES(8,1620);
INSERT INTO playlist_track VALUES(8,1621);
INSERT INTO playlist_track VALUES(8,1622);
INSERT INTO playlist_track VALUES(8,1623);
INSERT INTO playlist_track VALUES(8,1624);
INSERT INTO playlist_track VALUES(8,1625);
INSERT INTO playlist_track VALUES(8,1626);
INSERT INTO playlist_track VALUES(8,1627);
INSERT INTO playlist_track VALUES(8,1628);
INSERT INTO playlist_track VALUES(8,1629);
INSERT INTO playlist_track VALUES(8,1630);
INSERT INTO playlist_track VALUES(8,1631);
INSERT INTO playlist_track VALUES(8,1632);
INSERT INTO playlist_track VALUES(8,1633);
INSERT INTO playlist_track VALUES(8,1634);
INSERT INTO playlist_track VALUES(8,1635);
INSERT INTO playlist_track VALUES(8,1636);
INSERT INTO playlist_track VALUES(8,1637);
INSERT INTO playlist_track VALUES(8,1638);
INSERT INTO playlist_track VALUES(8,1639);
INSERT INTO playlist_track VALUES(8,1640);
INSERT INTO playlist_track VALUES(8,1641);
INSERT INTO playlist_track VALUES(8,1642);
INSERT INTO playlist_track VALUES(8,1643);
INSERT INTO playlist_track VALUES(8,1644);
INSERT INTO playlist_track VALUES(8,1645);
INSERT INTO playlist_track VALUES(8,550);
INSERT INTO playlist_track VALUES(8,551);
INSERT INTO playlist_track VALUES(8,552);
INSERT INTO playlist_track VALUES(8,553);
INSERT INTO playlist_track VALUES(8,554);
INSERT INTO playlist_track VALUES(8,555);
INSERT INTO playlist_track VALUES(8,1646);
INSERT INTO playlist_track VALUES(8,1647);
INSERT INTO playlist_track VALUES(8,1648);
INSERT INTO playlist_track VALUES(8,1649);
INSERT INTO playlist_track VALUES(8,1650);
INSERT INTO playlist_track VALUES(8,1651);
INSERT INTO playlist_track VALUES(8,1652);
INSERT INTO playlist_track VALUES(8,1653);
INSERT INTO playlist_track VALUES(8,1654);
INSERT INTO playlist_track VALUES(8,1655);
INSERT INTO playlist_track VALUES(8,1656);
INSERT INTO playlist_track VALUES(8,1657);
INSERT INTO playlist_track VALUES(8,1658);
INSERT INTO playlist_track VALUES(8,1659);
INSERT INTO playlist_track VALUES(8,1660);
INSERT INTO playlist_track VALUES(8,1661);
INSERT INTO playlist_track VALUES(8,1662);
INSERT INTO playlist_track VALUES(8,1663);
INSERT INTO playlist_track VALUES(8,1664);
INSERT INTO playlist_track VALUES(8,1665);
INSERT INTO playlist_track VALUES(8,1666);
INSERT INTO playlist_track VALUES(8,1667);
INSERT INTO playlist_track VALUES(8,1668);
INSERT INTO playlist_track VALUES(8,1669);
INSERT INTO playlist_track VALUES(8,1670);
INSERT INTO playlist_track VALUES(8,1686);
INSERT INTO playlist_track VALUES(8,1687);
INSERT INTO playlist_track VALUES(8,1688);
INSERT INTO playlist_track VALUES(8,1689);
INSERT INTO playlist_track VALUES(8,1690);
INSERT INTO playlist_track VALUES(8,1691);
INSERT INTO playlist_track VALUES(8,1692);
INSERT INTO playlist_track VALUES(8,1693);
INSERT INTO playlist_track VALUES(8,1694);
INSERT INTO playlist_track VALUES(8,1695);
INSERT INTO playlist_track VALUES(8,1696);
INSERT INTO playlist_track VALUES(8,1697);
INSERT INTO playlist_track VALUES(8,1698);
INSERT INTO playlist_track VALUES(8,1699);
INSERT INTO playlist_track VALUES(8,1700);
INSERT INTO playlist_track VALUES(8,1701);
INSERT INTO playlist_track VALUES(8,1671);
INSERT INTO playlist_track VALUES(8,1672);
INSERT INTO playlist_track VALUES(8,1673);
INSERT INTO playlist_track VALUES(8,1674);
INSERT INTO playlist_track VALUES(8,1675);
INSERT INTO playlist_track VALUES(8,1676);
INSERT INTO playlist_track VALUES(8,1677);
INSERT INTO playlist_track VALUES(8,1678);
INSERT INTO playlist_track VALUES(8,1679);
INSERT INTO playlist_track VALUES(8,1680);
INSERT INTO playlist_track VALUES(8,1681);
INSERT INTO playlist_track VALUES(8,1682);
INSERT INTO playlist_track VALUES(8,1683);
INSERT INTO playlist_track VALUES(8,1684);
INSERT INTO playlist_track VALUES(8,1685);
INSERT INTO playlist_track VALUES(8,1702);
INSERT INTO playlist_track VALUES(8,1703);
INSERT INTO playlist_track VALUES(8,1704);
INSERT INTO playlist_track VALUES(8,1705);
INSERT INTO playlist_track VALUES(8,1706);
INSERT INTO playlist_track VALUES(8,1707);
INSERT INTO playlist_track VALUES(8,1708);
INSERT INTO playlist_track VALUES(8,1709);
INSERT INTO playlist_track VALUES(8,1710);
INSERT INTO playlist_track VALUES(8,1711);
INSERT INTO playlist_track VALUES(8,1712);
INSERT INTO playlist_track VALUES(8,1713);
INSERT INTO playlist_track VALUES(8,1714);
INSERT INTO playlist_track VALUES(8,1715);
INSERT INTO playlist_track VALUES(8,1716);
INSERT INTO playlist_track VALUES(8,3257);
INSERT INTO playlist_track VALUES(8,3425);
INSERT INTO playlist_track VALUES(8,3420);
INSERT INTO playlist_track VALUES(8,3326);
INSERT INTO playlist_track VALUES(8,3258);
INSERT INTO playlist_track VALUES(8,3356);
INSERT INTO playlist_track VALUES(8,3424);
INSERT INTO playlist_track VALUES(8,384);
INSERT INTO playlist_track VALUES(8,1717);
INSERT INTO playlist_track VALUES(8,1720);
INSERT INTO playlist_track VALUES(8,1722);
INSERT INTO playlist_track VALUES(8,1723);
INSERT INTO playlist_track VALUES(8,1726);
INSERT INTO playlist_track VALUES(8,1727);
INSERT INTO playlist_track VALUES(8,1730);
INSERT INTO playlist_track VALUES(8,1731);
INSERT INTO playlist_track VALUES(8,1733);
INSERT INTO playlist_track VALUES(8,1736);
INSERT INTO playlist_track VALUES(8,1737);
INSERT INTO playlist_track VALUES(8,1740);
INSERT INTO playlist_track VALUES(8,1742);
INSERT INTO playlist_track VALUES(8,1743);
INSERT INTO playlist_track VALUES(8,1718);
INSERT INTO playlist_track VALUES(8,1719);
INSERT INTO playlist_track VALUES(8,1721);
INSERT INTO playlist_track VALUES(8,1724);
INSERT INTO playlist_track VALUES(8,1725);
INSERT INTO playlist_track VALUES(8,1728);
INSERT INTO playlist_track VALUES(8,1729);
INSERT INTO playlist_track VALUES(8,1732);
INSERT INTO playlist_track VALUES(8,1734);
INSERT INTO playlist_track VALUES(8,1735);
INSERT INTO playlist_track VALUES(8,1738);
INSERT INTO playlist_track VALUES(8,1739);
INSERT INTO playlist_track VALUES(8,1741);
INSERT INTO playlist_track VALUES(8,1744);
INSERT INTO playlist_track VALUES(8,374);
INSERT INTO playlist_track VALUES(8,1745);
INSERT INTO playlist_track VALUES(8,1746);
INSERT INTO playlist_track VALUES(8,1747);
INSERT INTO playlist_track VALUES(8,1748);
INSERT INTO playlist_track VALUES(8,1749);
INSERT INTO playlist_track VALUES(8,1750);
INSERT INTO playlist_track VALUES(8,1751);
INSERT INTO playlist_track VALUES(8,1752);
INSERT INTO playlist_track VALUES(8,1753);
INSERT INTO playlist_track VALUES(8,1754);
INSERT INTO playlist_track VALUES(8,1755);
INSERT INTO playlist_track VALUES(8,1762);
INSERT INTO playlist_track VALUES(8,1763);
INSERT INTO playlist_track VALUES(8,1756);
INSERT INTO playlist_track VALUES(8,1764);
INSERT INTO playlist_track VALUES(8,1757);
INSERT INTO playlist_track VALUES(8,1758);
INSERT INTO playlist_track VALUES(8,1765);
INSERT INTO playlist_track VALUES(8,1766);
INSERT INTO playlist_track VALUES(8,1759);
INSERT INTO playlist_track VALUES(8,1760);
INSERT INTO playlist_track VALUES(8,1767);
INSERT INTO playlist_track VALUES(8,1761);
INSERT INTO playlist_track VALUES(8,1768);
INSERT INTO playlist_track VALUES(8,1769);
INSERT INTO playlist_track VALUES(8,1770);
INSERT INTO playlist_track VALUES(8,1771);
INSERT INTO playlist_track VALUES(8,1772);
INSERT INTO playlist_track VALUES(8,1773);
INSERT INTO playlist_track VALUES(8,1774);
INSERT INTO playlist_track VALUES(8,1775);
INSERT INTO playlist_track VALUES(8,1776);
INSERT INTO playlist_track VALUES(8,1777);
INSERT INTO playlist_track VALUES(8,1778);
INSERT INTO playlist_track VALUES(8,1779);
INSERT INTO playlist_track VALUES(8,1780);
INSERT INTO playlist_track VALUES(8,1781);
INSERT INTO playlist_track VALUES(8,1782);
INSERT INTO playlist_track VALUES(8,1783);
INSERT INTO playlist_track VALUES(8,1784);
INSERT INTO playlist_track VALUES(8,1785);
INSERT INTO playlist_track VALUES(8,1786);
INSERT INTO playlist_track VALUES(8,1787);
INSERT INTO playlist_track VALUES(8,1788);
INSERT INTO playlist_track VALUES(8,1789);
INSERT INTO playlist_track VALUES(8,1790);
INSERT INTO playlist_track VALUES(8,3270);
INSERT INTO playlist_track VALUES(8,1791);
INSERT INTO playlist_track VALUES(8,1792);
INSERT INTO playlist_track VALUES(8,1793);
INSERT INTO playlist_track VALUES(8,1794);
INSERT INTO playlist_track VALUES(8,1795);
INSERT INTO playlist_track VALUES(8,1796);
INSERT INTO playlist_track VALUES(8,1797);
INSERT INTO playlist_track VALUES(8,1798);
INSERT INTO playlist_track VALUES(8,1799);
INSERT INTO playlist_track VALUES(8,1800);
INSERT INTO playlist_track VALUES(8,1893);
INSERT INTO playlist_track VALUES(8,1894);
INSERT INTO playlist_track VALUES(8,1895);
INSERT INTO playlist_track VALUES(8,1896);
INSERT INTO playlist_track VALUES(8,1897);
INSERT INTO playlist_track VALUES(8,1898);
INSERT INTO playlist_track VALUES(8,1899);
INSERT INTO playlist_track VALUES(8,1900);
INSERT INTO playlist_track VALUES(8,1901);
INSERT INTO playlist_track VALUES(8,1801);
INSERT INTO playlist_track VALUES(8,1802);
INSERT INTO playlist_track VALUES(8,1803);
INSERT INTO playlist_track VALUES(8,1804);
INSERT INTO playlist_track VALUES(8,1805);
INSERT INTO playlist_track VALUES(8,1806);
INSERT INTO playlist_track VALUES(8,1807);
INSERT INTO playlist_track VALUES(8,1808);
INSERT INTO playlist_track VALUES(8,1809);
INSERT INTO playlist_track VALUES(8,1810);
INSERT INTO playlist_track VALUES(8,1811);
INSERT INTO playlist_track VALUES(8,1812);
INSERT INTO playlist_track VALUES(8,408);
INSERT INTO playlist_track VALUES(8,409);
INSERT INTO playlist_track VALUES(8,410);
INSERT INTO playlist_track VALUES(8,411);
INSERT INTO playlist_track VALUES(8,412);
INSERT INTO playlist_track VALUES(8,413);
INSERT INTO playlist_track VALUES(8,414);
INSERT INTO playlist_track VALUES(8,415);
INSERT INTO playlist_track VALUES(8,416);
INSERT INTO playlist_track VALUES(8,417);
INSERT INTO playlist_track VALUES(8,418);
INSERT INTO playlist_track VALUES(8,1813);
INSERT INTO playlist_track VALUES(8,1814);
INSERT INTO playlist_track VALUES(8,1815);
INSERT INTO playlist_track VALUES(8,1816);
INSERT INTO playlist_track VALUES(8,1817);
INSERT INTO playlist_track VALUES(8,1818);
INSERT INTO playlist_track VALUES(8,1819);
INSERT INTO playlist_track VALUES(8,1820);
INSERT INTO playlist_track VALUES(8,1821);
INSERT INTO playlist_track VALUES(8,1822);
INSERT INTO playlist_track VALUES(8,1823);
INSERT INTO playlist_track VALUES(8,1824);
INSERT INTO playlist_track VALUES(8,1825);
INSERT INTO playlist_track VALUES(8,1826);
INSERT INTO playlist_track VALUES(8,1827);
INSERT INTO playlist_track VALUES(8,1828);
INSERT INTO playlist_track VALUES(8,1829);
INSERT INTO playlist_track VALUES(8,1830);
INSERT INTO playlist_track VALUES(8,1831);
INSERT INTO playlist_track VALUES(8,1832);
INSERT INTO playlist_track VALUES(8,1833);
INSERT INTO playlist_track VALUES(8,1834);
INSERT INTO playlist_track VALUES(8,1835);
INSERT INTO playlist_track VALUES(8,1836);
INSERT INTO playlist_track VALUES(8,1837);
INSERT INTO playlist_track VALUES(8,1838);
INSERT INTO playlist_track VALUES(8,1839);
INSERT INTO playlist_track VALUES(8,1840);
INSERT INTO playlist_track VALUES(8,1841);
INSERT INTO playlist_track VALUES(8,1842);
INSERT INTO playlist_track VALUES(8,1843);
INSERT INTO playlist_track VALUES(8,1844);
INSERT INTO playlist_track VALUES(8,1845);
INSERT INTO playlist_track VALUES(8,1846);
INSERT INTO playlist_track VALUES(8,1847);
INSERT INTO playlist_track VALUES(8,1848);
INSERT INTO playlist_track VALUES(8,1849);
INSERT INTO playlist_track VALUES(8,1850);
INSERT INTO playlist_track VALUES(8,1851);
INSERT INTO playlist_track VALUES(8,1852);
INSERT INTO playlist_track VALUES(8,1853);
INSERT INTO playlist_track VALUES(8,1854);
INSERT INTO playlist_track VALUES(8,1855);
INSERT INTO playlist_track VALUES(8,1856);
INSERT INTO playlist_track VALUES(8,1857);
INSERT INTO playlist_track VALUES(8,1858);
INSERT INTO playlist_track VALUES(8,1859);
INSERT INTO playlist_track VALUES(8,1860);
INSERT INTO playlist_track VALUES(8,1861);
INSERT INTO playlist_track VALUES(8,1862);
INSERT INTO playlist_track VALUES(8,1863);
INSERT INTO playlist_track VALUES(8,1864);
INSERT INTO playlist_track VALUES(8,1865);
INSERT INTO playlist_track VALUES(8,1866);
INSERT INTO playlist_track VALUES(8,1867);
INSERT INTO playlist_track VALUES(8,1868);
INSERT INTO playlist_track VALUES(8,1869);
INSERT INTO playlist_track VALUES(8,1870);
INSERT INTO playlist_track VALUES(8,1871);
INSERT INTO playlist_track VALUES(8,1872);
INSERT INTO playlist_track VALUES(8,1873);
INSERT INTO playlist_track VALUES(8,1874);
INSERT INTO playlist_track VALUES(8,1875);
INSERT INTO playlist_track VALUES(8,1876);
INSERT INTO playlist_track VALUES(8,1877);
INSERT INTO playlist_track VALUES(8,1878);
INSERT INTO playlist_track VALUES(8,1879);
INSERT INTO playlist_track VALUES(8,1880);
INSERT INTO playlist_track VALUES(8,1881);
INSERT INTO playlist_track VALUES(8,1882);
INSERT INTO playlist_track VALUES(8,1883);
INSERT INTO playlist_track VALUES(8,1884);
INSERT INTO playlist_track VALUES(8,1885);
INSERT INTO playlist_track VALUES(8,1886);
INSERT INTO playlist_track VALUES(8,1887);
INSERT INTO playlist_track VALUES(8,1888);
INSERT INTO playlist_track VALUES(8,1889);
INSERT INTO playlist_track VALUES(8,1890);
INSERT INTO playlist_track VALUES(8,1891);
INSERT INTO playlist_track VALUES(8,1892);
INSERT INTO playlist_track VALUES(8,597);
INSERT INTO playlist_track VALUES(8,598);
INSERT INTO playlist_track VALUES(8,599);
INSERT INTO playlist_track VALUES(8,600);
INSERT INTO playlist_track VALUES(8,601);
INSERT INTO playlist_track VALUES(8,602);
INSERT INTO playlist_track VALUES(8,603);
INSERT INTO playlist_track VALUES(8,604);
INSERT INTO playlist_track VALUES(8,605);
INSERT INTO playlist_track VALUES(8,606);
INSERT INTO playlist_track VALUES(8,607);
INSERT INTO playlist_track VALUES(8,608);
INSERT INTO playlist_track VALUES(8,609);
INSERT INTO playlist_track VALUES(8,610);
INSERT INTO playlist_track VALUES(8,611);
INSERT INTO playlist_track VALUES(8,612);
INSERT INTO playlist_track VALUES(8,613);
INSERT INTO playlist_track VALUES(8,614);
INSERT INTO playlist_track VALUES(8,615);
INSERT INTO playlist_track VALUES(8,616);
INSERT INTO playlist_track VALUES(8,617);
INSERT INTO playlist_track VALUES(8,618);
INSERT INTO playlist_track VALUES(8,619);
INSERT INTO playlist_track VALUES(8,1902);
INSERT INTO playlist_track VALUES(8,1903);
INSERT INTO playlist_track VALUES(8,1904);
INSERT INTO playlist_track VALUES(8,1905);
INSERT INTO playlist_track VALUES(8,1906);
INSERT INTO playlist_track VALUES(8,1907);
INSERT INTO playlist_track VALUES(8,1908);
INSERT INTO playlist_track VALUES(8,1909);
INSERT INTO playlist_track VALUES(8,1910);
INSERT INTO playlist_track VALUES(8,1911);
INSERT INTO playlist_track VALUES(8,1912);
INSERT INTO playlist_track VALUES(8,1913);
INSERT INTO playlist_track VALUES(8,1914);
INSERT INTO playlist_track VALUES(8,1915);
INSERT INTO playlist_track VALUES(8,398);
INSERT INTO playlist_track VALUES(8,1916);
INSERT INTO playlist_track VALUES(8,1917);
INSERT INTO playlist_track VALUES(8,1918);
INSERT INTO playlist_track VALUES(8,1919);
INSERT INTO playlist_track VALUES(8,1920);
INSERT INTO playlist_track VALUES(8,1921);
INSERT INTO playlist_track VALUES(8,1922);
INSERT INTO playlist_track VALUES(8,1923);
INSERT INTO playlist_track VALUES(8,1924);
INSERT INTO playlist_track VALUES(8,1925);
INSERT INTO playlist_track VALUES(8,1926);
INSERT INTO playlist_track VALUES(8,1927);
INSERT INTO playlist_track VALUES(8,1928);
INSERT INTO playlist_track VALUES(8,1929);
INSERT INTO playlist_track VALUES(8,1930);
INSERT INTO playlist_track VALUES(8,1931);
INSERT INTO playlist_track VALUES(8,1932);
INSERT INTO playlist_track VALUES(8,1933);
INSERT INTO playlist_track VALUES(8,1934);
INSERT INTO playlist_track VALUES(8,1935);
INSERT INTO playlist_track VALUES(8,1936);
INSERT INTO playlist_track VALUES(8,1937);
INSERT INTO playlist_track VALUES(8,1938);
INSERT INTO playlist_track VALUES(8,1939);
INSERT INTO playlist_track VALUES(8,1940);
INSERT INTO playlist_track VALUES(8,1941);
INSERT INTO playlist_track VALUES(8,375);
INSERT INTO playlist_track VALUES(8,1957);
INSERT INTO playlist_track VALUES(8,1958);
INSERT INTO playlist_track VALUES(8,1959);
INSERT INTO playlist_track VALUES(8,1960);
INSERT INTO playlist_track VALUES(8,1961);
INSERT INTO playlist_track VALUES(8,1962);
INSERT INTO playlist_track VALUES(8,1963);
INSERT INTO playlist_track VALUES(8,1964);
INSERT INTO playlist_track VALUES(8,1965);
INSERT INTO playlist_track VALUES(8,1966);
INSERT INTO playlist_track VALUES(8,1967);
INSERT INTO playlist_track VALUES(8,1968);
INSERT INTO playlist_track VALUES(8,1969);
INSERT INTO playlist_track VALUES(8,1970);
INSERT INTO playlist_track VALUES(8,1971);
INSERT INTO playlist_track VALUES(8,1972);
INSERT INTO playlist_track VALUES(8,1973);
INSERT INTO playlist_track VALUES(8,1974);
INSERT INTO playlist_track VALUES(8,1975);
INSERT INTO playlist_track VALUES(8,1976);
INSERT INTO playlist_track VALUES(8,1977);
INSERT INTO playlist_track VALUES(8,1978);
INSERT INTO playlist_track VALUES(8,1979);
INSERT INTO playlist_track VALUES(8,1980);
INSERT INTO playlist_track VALUES(8,1981);
INSERT INTO playlist_track VALUES(8,1982);
INSERT INTO playlist_track VALUES(8,1983);
INSERT INTO playlist_track VALUES(8,1984);
INSERT INTO playlist_track VALUES(8,1985);
INSERT INTO playlist_track VALUES(8,1942);
INSERT INTO playlist_track VALUES(8,1943);
INSERT INTO playlist_track VALUES(8,1944);
INSERT INTO playlist_track VALUES(8,1945);
INSERT INTO playlist_track VALUES(8,1946);
INSERT INTO playlist_track VALUES(8,1947);
INSERT INTO playlist_track VALUES(8,1948);
INSERT INTO playlist_track VALUES(8,1949);
INSERT INTO playlist_track VALUES(8,1950);
INSERT INTO playlist_track VALUES(8,1951);
INSERT INTO playlist_track VALUES(8,1952);
INSERT INTO playlist_track VALUES(8,1953);
INSERT INTO playlist_track VALUES(8,1954);
INSERT INTO playlist_track VALUES(8,1955);
INSERT INTO playlist_track VALUES(8,1956);
INSERT INTO playlist_track VALUES(8,3327);
INSERT INTO playlist_track VALUES(8,3330);
INSERT INTO playlist_track VALUES(8,385);
INSERT INTO playlist_track VALUES(8,3321);
INSERT INTO playlist_track VALUES(8,383);
INSERT INTO playlist_track VALUES(8,3359);
INSERT INTO playlist_track VALUES(8,1986);
INSERT INTO playlist_track VALUES(8,1987);
INSERT INTO playlist_track VALUES(8,1988);
INSERT INTO playlist_track VALUES(8,1989);
INSERT INTO playlist_track VALUES(8,1990);
INSERT INTO playlist_track VALUES(8,1991);
INSERT INTO playlist_track VALUES(8,1992);
INSERT INTO playlist_track VALUES(8,1993);
INSERT INTO playlist_track VALUES(8,1994);
INSERT INTO playlist_track VALUES(8,1995);
INSERT INTO playlist_track VALUES(8,1996);
INSERT INTO playlist_track VALUES(8,1997);
INSERT INTO playlist_track VALUES(8,1998);
INSERT INTO playlist_track VALUES(8,1999);
INSERT INTO playlist_track VALUES(8,2000);
INSERT INTO playlist_track VALUES(8,2001);
INSERT INTO playlist_track VALUES(8,2002);
INSERT INTO playlist_track VALUES(8,2003);
INSERT INTO playlist_track VALUES(8,2004);
INSERT INTO playlist_track VALUES(8,2005);
INSERT INTO playlist_track VALUES(8,2006);
INSERT INTO playlist_track VALUES(8,2007);
INSERT INTO playlist_track VALUES(8,2008);
INSERT INTO playlist_track VALUES(8,2009);
INSERT INTO playlist_track VALUES(8,2010);
INSERT INTO playlist_track VALUES(8,2011);
INSERT INTO playlist_track VALUES(8,2012);
INSERT INTO playlist_track VALUES(8,2013);
INSERT INTO playlist_track VALUES(8,2014);
INSERT INTO playlist_track VALUES(8,387);
INSERT INTO playlist_track VALUES(8,3319);
INSERT INTO playlist_track VALUES(8,2015);
INSERT INTO playlist_track VALUES(8,2016);
INSERT INTO playlist_track VALUES(8,2017);
INSERT INTO playlist_track VALUES(8,2018);
INSERT INTO playlist_track VALUES(8,2019);
INSERT INTO playlist_track VALUES(8,2020);
INSERT INTO playlist_track VALUES(8,2021);
INSERT INTO playlist_track VALUES(8,2022);
INSERT INTO playlist_track VALUES(8,2023);
INSERT INTO playlist_track VALUES(8,2024);
INSERT INTO playlist_track VALUES(8,2025);
INSERT INTO playlist_track VALUES(8,2026);
INSERT INTO playlist_track VALUES(8,2027);
INSERT INTO playlist_track VALUES(8,2028);
INSERT INTO playlist_track VALUES(8,2029);
INSERT INTO playlist_track VALUES(8,2030);
INSERT INTO playlist_track VALUES(8,2031);
INSERT INTO playlist_track VALUES(8,2032);
INSERT INTO playlist_track VALUES(8,2033);
INSERT INTO playlist_track VALUES(8,2034);
INSERT INTO playlist_track VALUES(8,2035);
INSERT INTO playlist_track VALUES(8,2036);
INSERT INTO playlist_track VALUES(8,2037);
INSERT INTO playlist_track VALUES(8,2038);
INSERT INTO playlist_track VALUES(8,2039);
INSERT INTO playlist_track VALUES(8,2040);
INSERT INTO playlist_track VALUES(8,2041);
INSERT INTO playlist_track VALUES(8,2042);
INSERT INTO playlist_track VALUES(8,2043);
INSERT INTO playlist_track VALUES(8,3415);
INSERT INTO playlist_track VALUES(8,393);
INSERT INTO playlist_track VALUES(8,529);
INSERT INTO playlist_track VALUES(8,530);
INSERT INTO playlist_track VALUES(8,531);
INSERT INTO playlist_track VALUES(8,532);
INSERT INTO playlist_track VALUES(8,533);
INSERT INTO playlist_track VALUES(8,534);
INSERT INTO playlist_track VALUES(8,535);
INSERT INTO playlist_track VALUES(8,536);
INSERT INTO playlist_track VALUES(8,537);
INSERT INTO playlist_track VALUES(8,538);
INSERT INTO playlist_track VALUES(8,539);
INSERT INTO playlist_track VALUES(8,540);
INSERT INTO playlist_track VALUES(8,541);
INSERT INTO playlist_track VALUES(8,542);
INSERT INTO playlist_track VALUES(8,2044);
INSERT INTO playlist_track VALUES(8,2045);
INSERT INTO playlist_track VALUES(8,2046);
INSERT INTO playlist_track VALUES(8,2047);
INSERT INTO playlist_track VALUES(8,2048);
INSERT INTO playlist_track VALUES(8,2049);
INSERT INTO playlist_track VALUES(8,2050);
INSERT INTO playlist_track VALUES(8,2051);
INSERT INTO playlist_track VALUES(8,2052);
INSERT INTO playlist_track VALUES(8,2053);
INSERT INTO playlist_track VALUES(8,2054);
INSERT INTO playlist_track VALUES(8,2055);
INSERT INTO playlist_track VALUES(8,2056);
INSERT INTO playlist_track VALUES(8,2057);
INSERT INTO playlist_track VALUES(8,2058);
INSERT INTO playlist_track VALUES(8,2059);
INSERT INTO playlist_track VALUES(8,2060);
INSERT INTO playlist_track VALUES(8,2061);
INSERT INTO playlist_track VALUES(8,2062);
INSERT INTO playlist_track VALUES(8,2063);
INSERT INTO playlist_track VALUES(8,2064);
INSERT INTO playlist_track VALUES(8,2065);
INSERT INTO playlist_track VALUES(8,2066);
INSERT INTO playlist_track VALUES(8,2067);
INSERT INTO playlist_track VALUES(8,2068);
INSERT INTO playlist_track VALUES(8,2069);
INSERT INTO playlist_track VALUES(8,2070);
INSERT INTO playlist_track VALUES(8,2071);
INSERT INTO playlist_track VALUES(8,2072);
INSERT INTO playlist_track VALUES(8,2073);
INSERT INTO playlist_track VALUES(8,2074);
INSERT INTO playlist_track VALUES(8,2075);
INSERT INTO playlist_track VALUES(8,2076);
INSERT INTO playlist_track VALUES(8,2077);
INSERT INTO playlist_track VALUES(8,2078);
INSERT INTO playlist_track VALUES(8,2079);
INSERT INTO playlist_track VALUES(8,2080);
INSERT INTO playlist_track VALUES(8,2081);
INSERT INTO playlist_track VALUES(8,2082);
INSERT INTO playlist_track VALUES(8,2083);
INSERT INTO playlist_track VALUES(8,2084);
INSERT INTO playlist_track VALUES(8,2085);
INSERT INTO playlist_track VALUES(8,2086);
INSERT INTO playlist_track VALUES(8,2087);
INSERT INTO playlist_track VALUES(8,2088);
INSERT INTO playlist_track VALUES(8,2089);
INSERT INTO playlist_track VALUES(8,2090);
INSERT INTO playlist_track VALUES(8,2091);
INSERT INTO playlist_track VALUES(8,2092);
INSERT INTO playlist_track VALUES(8,3328);
INSERT INTO playlist_track VALUES(8,2093);
INSERT INTO playlist_track VALUES(8,2094);
INSERT INTO playlist_track VALUES(8,2095);
INSERT INTO playlist_track VALUES(8,2096);
INSERT INTO playlist_track VALUES(8,2097);
INSERT INTO playlist_track VALUES(8,2098);
INSERT INTO playlist_track VALUES(8,3276);
INSERT INTO playlist_track VALUES(8,3277);
INSERT INTO playlist_track VALUES(8,3278);
INSERT INTO playlist_track VALUES(8,3279);
INSERT INTO playlist_track VALUES(8,3280);
INSERT INTO playlist_track VALUES(8,3281);
INSERT INTO playlist_track VALUES(8,3282);
INSERT INTO playlist_track VALUES(8,3283);
INSERT INTO playlist_track VALUES(8,3284);
INSERT INTO playlist_track VALUES(8,3285);
INSERT INTO playlist_track VALUES(8,3286);
INSERT INTO playlist_track VALUES(8,3287);
INSERT INTO playlist_track VALUES(8,2099);
INSERT INTO playlist_track VALUES(8,2100);
INSERT INTO playlist_track VALUES(8,2101);
INSERT INTO playlist_track VALUES(8,2102);
INSERT INTO playlist_track VALUES(8,2103);
INSERT INTO playlist_track VALUES(8,2104);
INSERT INTO playlist_track VALUES(8,2105);
INSERT INTO playlist_track VALUES(8,2106);
INSERT INTO playlist_track VALUES(8,2107);
INSERT INTO playlist_track VALUES(8,2108);
INSERT INTO playlist_track VALUES(8,2109);
INSERT INTO playlist_track VALUES(8,2110);
INSERT INTO playlist_track VALUES(8,2111);
INSERT INTO playlist_track VALUES(8,2112);
INSERT INTO playlist_track VALUES(8,2113);
INSERT INTO playlist_track VALUES(8,2114);
INSERT INTO playlist_track VALUES(8,2115);
INSERT INTO playlist_track VALUES(8,2116);
INSERT INTO playlist_track VALUES(8,2117);
INSERT INTO playlist_track VALUES(8,2118);
INSERT INTO playlist_track VALUES(8,2119);
INSERT INTO playlist_track VALUES(8,2120);
INSERT INTO playlist_track VALUES(8,2121);
INSERT INTO playlist_track VALUES(8,2122);
INSERT INTO playlist_track VALUES(8,2123);
INSERT INTO playlist_track VALUES(8,2124);
INSERT INTO playlist_track VALUES(8,2125);
INSERT INTO playlist_track VALUES(8,2126);
INSERT INTO playlist_track VALUES(8,2127);
INSERT INTO playlist_track VALUES(8,2128);
INSERT INTO playlist_track VALUES(8,2129);
INSERT INTO playlist_track VALUES(8,2130);
INSERT INTO playlist_track VALUES(8,2131);
INSERT INTO playlist_track VALUES(8,2132);
INSERT INTO playlist_track VALUES(8,2133);
INSERT INTO playlist_track VALUES(8,2134);
INSERT INTO playlist_track VALUES(8,2135);
INSERT INTO playlist_track VALUES(8,2136);
INSERT INTO playlist_track VALUES(8,2137);
INSERT INTO playlist_track VALUES(8,2138);
INSERT INTO playlist_track VALUES(8,2139);
INSERT INTO playlist_track VALUES(8,2140);
INSERT INTO playlist_track VALUES(8,2141);
INSERT INTO playlist_track VALUES(8,2142);
INSERT INTO playlist_track VALUES(8,2143);
INSERT INTO playlist_track VALUES(8,2144);
INSERT INTO playlist_track VALUES(8,2145);
INSERT INTO playlist_track VALUES(8,2146);
INSERT INTO playlist_track VALUES(8,2147);
INSERT INTO playlist_track VALUES(8,2148);
INSERT INTO playlist_track VALUES(8,2149);
INSERT INTO playlist_track VALUES(8,2150);
INSERT INTO playlist_track VALUES(8,2151);
INSERT INTO playlist_track VALUES(8,2152);
INSERT INTO playlist_track VALUES(8,2153);
INSERT INTO playlist_track VALUES(8,2154);
INSERT INTO playlist_track VALUES(8,2155);
INSERT INTO playlist_track VALUES(8,2156);
INSERT INTO playlist_track VALUES(8,2157);
INSERT INTO playlist_track VALUES(8,2158);
INSERT INTO playlist_track VALUES(8,2159);
INSERT INTO playlist_track VALUES(8,2160);
INSERT INTO playlist_track VALUES(8,2161);
INSERT INTO playlist_track VALUES(8,2162);
INSERT INTO playlist_track VALUES(8,2163);
INSERT INTO playlist_track VALUES(8,2164);
INSERT INTO playlist_track VALUES(8,2165);
INSERT INTO playlist_track VALUES(8,2166);
INSERT INTO playlist_track VALUES(8,2167);
INSERT INTO playlist_track VALUES(8,2168);
INSERT INTO playlist_track VALUES(8,2169);
INSERT INTO playlist_track VALUES(8,2170);
INSERT INTO playlist_track VALUES(8,2171);
INSERT INTO playlist_track VALUES(8,2172);
INSERT INTO playlist_track VALUES(8,2173);
INSERT INTO playlist_track VALUES(8,2174);
INSERT INTO playlist_track VALUES(8,2175);
INSERT INTO playlist_track VALUES(8,2176);
INSERT INTO playlist_track VALUES(8,2177);
INSERT INTO playlist_track VALUES(8,2178);
INSERT INTO playlist_track VALUES(8,2179);
INSERT INTO playlist_track VALUES(8,2180);
INSERT INTO playlist_track VALUES(8,2181);
INSERT INTO playlist_track VALUES(8,2182);
INSERT INTO playlist_track VALUES(8,2183);
INSERT INTO playlist_track VALUES(8,2184);
INSERT INTO playlist_track VALUES(8,2185);
INSERT INTO playlist_track VALUES(8,2186);
INSERT INTO playlist_track VALUES(8,2187);
INSERT INTO playlist_track VALUES(8,2188);
INSERT INTO playlist_track VALUES(8,2189);
INSERT INTO playlist_track VALUES(8,2190);
INSERT INTO playlist_track VALUES(8,2191);
INSERT INTO playlist_track VALUES(8,2192);
INSERT INTO playlist_track VALUES(8,2193);
INSERT INTO playlist_track VALUES(8,2194);
INSERT INTO playlist_track VALUES(8,2195);
INSERT INTO playlist_track VALUES(8,2196);
INSERT INTO playlist_track VALUES(8,2197);
INSERT INTO playlist_track VALUES(8,2198);
INSERT INTO playlist_track VALUES(8,2199);
INSERT INTO playlist_track VALUES(8,2200);
INSERT INTO playlist_track VALUES(8,2201);
INSERT INTO playlist_track VALUES(8,2202);
INSERT INTO playlist_track VALUES(8,2203);
INSERT INTO playlist_track VALUES(8,2204);
INSERT INTO playlist_track VALUES(8,2205);
INSERT INTO playlist_track VALUES(8,2206);
INSERT INTO playlist_track VALUES(8,2207);
INSERT INTO playlist_track VALUES(8,2208);
INSERT INTO playlist_track VALUES(8,2209);
INSERT INTO playlist_track VALUES(8,2210);
INSERT INTO playlist_track VALUES(8,2211);
INSERT INTO playlist_track VALUES(8,2212);
INSERT INTO playlist_track VALUES(8,2213);
INSERT INTO playlist_track VALUES(8,2214);
INSERT INTO playlist_track VALUES(8,2215);
INSERT INTO playlist_track VALUES(8,386);
INSERT INTO playlist_track VALUES(8,3325);
INSERT INTO playlist_track VALUES(8,2216);
INSERT INTO playlist_track VALUES(8,2217);
INSERT INTO playlist_track VALUES(8,2218);
INSERT INTO playlist_track VALUES(8,2219);
INSERT INTO playlist_track VALUES(8,2220);
INSERT INTO playlist_track VALUES(8,2221);
INSERT INTO playlist_track VALUES(8,2222);
INSERT INTO playlist_track VALUES(8,2223);
INSERT INTO playlist_track VALUES(8,2224);
INSERT INTO playlist_track VALUES(8,2225);
INSERT INTO playlist_track VALUES(8,2226);
INSERT INTO playlist_track VALUES(8,2227);
INSERT INTO playlist_track VALUES(8,2228);
INSERT INTO playlist_track VALUES(8,2229);
INSERT INTO playlist_track VALUES(8,2230);
INSERT INTO playlist_track VALUES(8,2231);
INSERT INTO playlist_track VALUES(8,2232);
INSERT INTO playlist_track VALUES(8,2233);
INSERT INTO playlist_track VALUES(8,2234);
INSERT INTO playlist_track VALUES(8,2235);
INSERT INTO playlist_track VALUES(8,2236);
INSERT INTO playlist_track VALUES(8,2237);
INSERT INTO playlist_track VALUES(8,2238);
INSERT INTO playlist_track VALUES(8,2239);
INSERT INTO playlist_track VALUES(8,2240);
INSERT INTO playlist_track VALUES(8,2241);
INSERT INTO playlist_track VALUES(8,2242);
INSERT INTO playlist_track VALUES(8,2243);
INSERT INTO playlist_track VALUES(8,2244);
INSERT INTO playlist_track VALUES(8,2245);
INSERT INTO playlist_track VALUES(8,2246);
INSERT INTO playlist_track VALUES(8,2247);
INSERT INTO playlist_track VALUES(8,2248);
INSERT INTO playlist_track VALUES(8,2249);
INSERT INTO playlist_track VALUES(8,2250);
INSERT INTO playlist_track VALUES(8,2251);
INSERT INTO playlist_track VALUES(8,2252);
INSERT INTO playlist_track VALUES(8,2253);
INSERT INTO playlist_track VALUES(8,2650);
INSERT INTO playlist_track VALUES(8,2651);
INSERT INTO playlist_track VALUES(8,2652);
INSERT INTO playlist_track VALUES(8,2653);
INSERT INTO playlist_track VALUES(8,2654);
INSERT INTO playlist_track VALUES(8,2655);
INSERT INTO playlist_track VALUES(8,2656);
INSERT INTO playlist_track VALUES(8,2657);
INSERT INTO playlist_track VALUES(8,2658);
INSERT INTO playlist_track VALUES(8,2659);
INSERT INTO playlist_track VALUES(8,2660);
INSERT INTO playlist_track VALUES(8,2661);
INSERT INTO playlist_track VALUES(8,2662);
INSERT INTO playlist_track VALUES(8,2663);
INSERT INTO playlist_track VALUES(8,3353);
INSERT INTO playlist_track VALUES(8,3355);
INSERT INTO playlist_track VALUES(8,3271);
INSERT INTO playlist_track VALUES(8,2254);
INSERT INTO playlist_track VALUES(8,2255);
INSERT INTO playlist_track VALUES(8,2256);
INSERT INTO playlist_track VALUES(8,2257);
INSERT INTO playlist_track VALUES(8,2258);
INSERT INTO playlist_track VALUES(8,2259);
INSERT INTO playlist_track VALUES(8,2260);
INSERT INTO playlist_track VALUES(8,2261);
INSERT INTO playlist_track VALUES(8,2262);
INSERT INTO playlist_track VALUES(8,2263);
INSERT INTO playlist_track VALUES(8,2264);
INSERT INTO playlist_track VALUES(8,2265);
INSERT INTO playlist_track VALUES(8,2266);
INSERT INTO playlist_track VALUES(8,2267);
INSERT INTO playlist_track VALUES(8,2268);
INSERT INTO playlist_track VALUES(8,2269);
INSERT INTO playlist_track VALUES(8,2270);
INSERT INTO playlist_track VALUES(8,419);
INSERT INTO playlist_track VALUES(8,420);
INSERT INTO playlist_track VALUES(8,421);
INSERT INTO playlist_track VALUES(8,422);
INSERT INTO playlist_track VALUES(8,423);
INSERT INTO playlist_track VALUES(8,424);
INSERT INTO playlist_track VALUES(8,425);
INSERT INTO playlist_track VALUES(8,426);
INSERT INTO playlist_track VALUES(8,427);
INSERT INTO playlist_track VALUES(8,428);
INSERT INTO playlist_track VALUES(8,429);
INSERT INTO playlist_track VALUES(8,430);
INSERT INTO playlist_track VALUES(8,431);
INSERT INTO playlist_track VALUES(8,432);
INSERT INTO playlist_track VALUES(8,433);
INSERT INTO playlist_track VALUES(8,434);
INSERT INTO playlist_track VALUES(8,435);
INSERT INTO playlist_track VALUES(8,2271);
INSERT INTO playlist_track VALUES(8,2272);
INSERT INTO playlist_track VALUES(8,2273);
INSERT INTO playlist_track VALUES(8,2274);
INSERT INTO playlist_track VALUES(8,2275);
INSERT INTO playlist_track VALUES(8,2276);
INSERT INTO playlist_track VALUES(8,2277);
INSERT INTO playlist_track VALUES(8,2278);
INSERT INTO playlist_track VALUES(8,2279);
INSERT INTO playlist_track VALUES(8,2280);
INSERT INTO playlist_track VALUES(8,2281);
INSERT INTO playlist_track VALUES(8,2318);
INSERT INTO playlist_track VALUES(8,2319);
INSERT INTO playlist_track VALUES(8,2320);
INSERT INTO playlist_track VALUES(8,2321);
INSERT INTO playlist_track VALUES(8,2322);
INSERT INTO playlist_track VALUES(8,2323);
INSERT INTO playlist_track VALUES(8,2324);
INSERT INTO playlist_track VALUES(8,2325);
INSERT INTO playlist_track VALUES(8,2326);
INSERT INTO playlist_track VALUES(8,2327);
INSERT INTO playlist_track VALUES(8,2328);
INSERT INTO playlist_track VALUES(8,2329);
INSERT INTO playlist_track VALUES(8,2330);
INSERT INTO playlist_track VALUES(8,2331);
INSERT INTO playlist_track VALUES(8,2332);
INSERT INTO playlist_track VALUES(8,2333);
INSERT INTO playlist_track VALUES(8,2285);
INSERT INTO playlist_track VALUES(8,2286);
INSERT INTO playlist_track VALUES(8,2287);
INSERT INTO playlist_track VALUES(8,2288);
INSERT INTO playlist_track VALUES(8,2289);
INSERT INTO playlist_track VALUES(8,2290);
INSERT INTO playlist_track VALUES(8,2291);
INSERT INTO playlist_track VALUES(8,2292);
INSERT INTO playlist_track VALUES(8,2293);
INSERT INTO playlist_track VALUES(8,2294);
INSERT INTO playlist_track VALUES(8,2295);
INSERT INTO playlist_track VALUES(8,3254);
INSERT INTO playlist_track VALUES(8,2296);
INSERT INTO playlist_track VALUES(8,2297);
INSERT INTO playlist_track VALUES(8,2298);
INSERT INTO playlist_track VALUES(8,2299);
INSERT INTO playlist_track VALUES(8,2300);
INSERT INTO playlist_track VALUES(8,2301);
INSERT INTO playlist_track VALUES(8,2302);
INSERT INTO playlist_track VALUES(8,2303);
INSERT INTO playlist_track VALUES(8,2304);
INSERT INTO playlist_track VALUES(8,2305);
INSERT INTO playlist_track VALUES(8,2306);
INSERT INTO playlist_track VALUES(8,2307);
INSERT INTO playlist_track VALUES(8,2308);
INSERT INTO playlist_track VALUES(8,2309);
INSERT INTO playlist_track VALUES(8,2310);
INSERT INTO playlist_track VALUES(8,2311);
INSERT INTO playlist_track VALUES(8,2312);
INSERT INTO playlist_track VALUES(8,2313);
INSERT INTO playlist_track VALUES(8,2314);
INSERT INTO playlist_track VALUES(8,2315);
INSERT INTO playlist_track VALUES(8,2316);
INSERT INTO playlist_track VALUES(8,2317);
INSERT INTO playlist_track VALUES(8,2282);
INSERT INTO playlist_track VALUES(8,2283);
INSERT INTO playlist_track VALUES(8,2284);
INSERT INTO playlist_track VALUES(8,2334);
INSERT INTO playlist_track VALUES(8,2335);
INSERT INTO playlist_track VALUES(8,2336);
INSERT INTO playlist_track VALUES(8,2337);
INSERT INTO playlist_track VALUES(8,2338);
INSERT INTO playlist_track VALUES(8,2339);
INSERT INTO playlist_track VALUES(8,2340);
INSERT INTO playlist_track VALUES(8,2341);
INSERT INTO playlist_track VALUES(8,2342);
INSERT INTO playlist_track VALUES(8,2343);
INSERT INTO playlist_track VALUES(8,2344);
INSERT INTO playlist_track VALUES(8,2345);
INSERT INTO playlist_track VALUES(8,2346);
INSERT INTO playlist_track VALUES(8,2347);
INSERT INTO playlist_track VALUES(8,2348);
INSERT INTO playlist_track VALUES(8,2349);
INSERT INTO playlist_track VALUES(8,2350);
INSERT INTO playlist_track VALUES(8,2351);
INSERT INTO playlist_track VALUES(8,2352);
INSERT INTO playlist_track VALUES(8,2353);
INSERT INTO playlist_track VALUES(8,2354);
INSERT INTO playlist_track VALUES(8,2355);
INSERT INTO playlist_track VALUES(8,2356);
INSERT INTO playlist_track VALUES(8,2357);
INSERT INTO playlist_track VALUES(8,2358);
INSERT INTO playlist_track VALUES(8,2359);
INSERT INTO playlist_track VALUES(8,2360);
INSERT INTO playlist_track VALUES(8,2361);
INSERT INTO playlist_track VALUES(8,2362);
INSERT INTO playlist_track VALUES(8,2363);
INSERT INTO playlist_track VALUES(8,2364);
INSERT INTO playlist_track VALUES(8,2365);
INSERT INTO playlist_track VALUES(8,2366);
INSERT INTO playlist_track VALUES(8,2367);
INSERT INTO playlist_track VALUES(8,2368);
INSERT INTO playlist_track VALUES(8,2369);
INSERT INTO playlist_track VALUES(8,2370);
INSERT INTO playlist_track VALUES(8,2371);
INSERT INTO playlist_track VALUES(8,2372);
INSERT INTO playlist_track VALUES(8,2373);
INSERT INTO playlist_track VALUES(8,2374);
INSERT INTO playlist_track VALUES(8,2375);
INSERT INTO playlist_track VALUES(8,2376);
INSERT INTO playlist_track VALUES(8,2377);
INSERT INTO playlist_track VALUES(8,2378);
INSERT INTO playlist_track VALUES(8,2379);
INSERT INTO playlist_track VALUES(8,2380);
INSERT INTO playlist_track VALUES(8,2381);
INSERT INTO playlist_track VALUES(8,2382);
INSERT INTO playlist_track VALUES(8,2383);
INSERT INTO playlist_track VALUES(8,2384);
INSERT INTO playlist_track VALUES(8,2385);
INSERT INTO playlist_track VALUES(8,2386);
INSERT INTO playlist_track VALUES(8,2387);
INSERT INTO playlist_track VALUES(8,2388);
INSERT INTO playlist_track VALUES(8,2389);
INSERT INTO playlist_track VALUES(8,2390);
INSERT INTO playlist_track VALUES(8,2391);
INSERT INTO playlist_track VALUES(8,2392);
INSERT INTO playlist_track VALUES(8,2393);
INSERT INTO playlist_track VALUES(8,2394);
INSERT INTO playlist_track VALUES(8,2395);
INSERT INTO playlist_track VALUES(8,2396);
INSERT INTO playlist_track VALUES(8,2397);
INSERT INTO playlist_track VALUES(8,2398);
INSERT INTO playlist_track VALUES(8,2399);
INSERT INTO playlist_track VALUES(8,2400);
INSERT INTO playlist_track VALUES(8,2401);
INSERT INTO playlist_track VALUES(8,2402);
INSERT INTO playlist_track VALUES(8,2403);
INSERT INTO playlist_track VALUES(8,2404);
INSERT INTO playlist_track VALUES(8,2405);
INSERT INTO playlist_track VALUES(8,3275);
INSERT INTO playlist_track VALUES(8,3404);
INSERT INTO playlist_track VALUES(8,3323);
INSERT INTO playlist_track VALUES(8,2664);
INSERT INTO playlist_track VALUES(8,2665);
INSERT INTO playlist_track VALUES(8,2666);
INSERT INTO playlist_track VALUES(8,2667);
INSERT INTO playlist_track VALUES(8,2668);
INSERT INTO playlist_track VALUES(8,2669);
INSERT INTO playlist_track VALUES(8,2670);
INSERT INTO playlist_track VALUES(8,2671);
INSERT INTO playlist_track VALUES(8,2672);
INSERT INTO playlist_track VALUES(8,2673);
INSERT INTO playlist_track VALUES(8,2674);
INSERT INTO playlist_track VALUES(8,2675);
INSERT INTO playlist_track VALUES(8,2676);
INSERT INTO playlist_track VALUES(8,2677);
INSERT INTO playlist_track VALUES(8,2678);
INSERT INTO playlist_track VALUES(8,2679);
INSERT INTO playlist_track VALUES(8,2680);
INSERT INTO playlist_track VALUES(8,2681);
INSERT INTO playlist_track VALUES(8,2682);
INSERT INTO playlist_track VALUES(8,2683);
INSERT INTO playlist_track VALUES(8,2684);
INSERT INTO playlist_track VALUES(8,2685);
INSERT INTO playlist_track VALUES(8,2686);
INSERT INTO playlist_track VALUES(8,2687);
INSERT INTO playlist_track VALUES(8,2688);
INSERT INTO playlist_track VALUES(8,2689);
INSERT INTO playlist_track VALUES(8,2690);
INSERT INTO playlist_track VALUES(8,2691);
INSERT INTO playlist_track VALUES(8,2692);
INSERT INTO playlist_track VALUES(8,2693);
INSERT INTO playlist_track VALUES(8,2694);
INSERT INTO playlist_track VALUES(8,2695);
INSERT INTO playlist_track VALUES(8,2696);
INSERT INTO playlist_track VALUES(8,2697);
INSERT INTO playlist_track VALUES(8,2698);
INSERT INTO playlist_track VALUES(8,2699);
INSERT INTO playlist_track VALUES(8,2700);
INSERT INTO playlist_track VALUES(8,2701);
INSERT INTO playlist_track VALUES(8,2702);
INSERT INTO playlist_track VALUES(8,2703);
INSERT INTO playlist_track VALUES(8,2704);
INSERT INTO playlist_track VALUES(8,3414);
INSERT INTO playlist_track VALUES(8,2406);
INSERT INTO playlist_track VALUES(8,2407);
INSERT INTO playlist_track VALUES(8,2408);
INSERT INTO playlist_track VALUES(8,2409);
INSERT INTO playlist_track VALUES(8,2410);
INSERT INTO playlist_track VALUES(8,2411);
INSERT INTO playlist_track VALUES(8,2412);
INSERT INTO playlist_track VALUES(8,2413);
INSERT INTO playlist_track VALUES(8,2414);
INSERT INTO playlist_track VALUES(8,2415);
INSERT INTO playlist_track VALUES(8,2416);
INSERT INTO playlist_track VALUES(8,2417);
INSERT INTO playlist_track VALUES(8,2418);
INSERT INTO playlist_track VALUES(8,2419);
INSERT INTO playlist_track VALUES(8,3334);
INSERT INTO playlist_track VALUES(8,401);
INSERT INTO playlist_track VALUES(8,2420);
INSERT INTO playlist_track VALUES(8,2421);
INSERT INTO playlist_track VALUES(8,2422);
INSERT INTO playlist_track VALUES(8,2423);
INSERT INTO playlist_track VALUES(8,2424);
INSERT INTO playlist_track VALUES(8,2425);
INSERT INTO playlist_track VALUES(8,2426);
INSERT INTO playlist_track VALUES(8,2427);
INSERT INTO playlist_track VALUES(8,2428);
INSERT INTO playlist_track VALUES(8,2429);
INSERT INTO playlist_track VALUES(8,2430);
INSERT INTO playlist_track VALUES(8,2431);
INSERT INTO playlist_track VALUES(8,2432);
INSERT INTO playlist_track VALUES(8,2433);
INSERT INTO playlist_track VALUES(8,570);
INSERT INTO playlist_track VALUES(8,573);
INSERT INTO playlist_track VALUES(8,577);
INSERT INTO playlist_track VALUES(8,580);
INSERT INTO playlist_track VALUES(8,581);
INSERT INTO playlist_track VALUES(8,571);
INSERT INTO playlist_track VALUES(8,579);
INSERT INTO playlist_track VALUES(8,582);
INSERT INTO playlist_track VALUES(8,572);
INSERT INTO playlist_track VALUES(8,575);
INSERT INTO playlist_track VALUES(8,578);
INSERT INTO playlist_track VALUES(8,574);
INSERT INTO playlist_track VALUES(8,576);
INSERT INTO playlist_track VALUES(8,3410);
INSERT INTO playlist_track VALUES(8,3288);
INSERT INTO playlist_track VALUES(8,3289);
INSERT INTO playlist_track VALUES(8,3290);
INSERT INTO playlist_track VALUES(8,3291);
INSERT INTO playlist_track VALUES(8,3292);
INSERT INTO playlist_track VALUES(8,3293);
INSERT INTO playlist_track VALUES(8,3294);
INSERT INTO playlist_track VALUES(8,3295);
INSERT INTO playlist_track VALUES(8,3296);
INSERT INTO playlist_track VALUES(8,3297);
INSERT INTO playlist_track VALUES(8,3298);
INSERT INTO playlist_track VALUES(8,3299);
INSERT INTO playlist_track VALUES(8,3333);
INSERT INTO playlist_track VALUES(8,2434);
INSERT INTO playlist_track VALUES(8,2435);
INSERT INTO playlist_track VALUES(8,2436);
INSERT INTO playlist_track VALUES(8,2437);
INSERT INTO playlist_track VALUES(8,2438);
INSERT INTO playlist_track VALUES(8,2439);
INSERT INTO playlist_track VALUES(8,2440);
INSERT INTO playlist_track VALUES(8,2441);
INSERT INTO playlist_track VALUES(8,2442);
INSERT INTO playlist_track VALUES(8,2443);
INSERT INTO playlist_track VALUES(8,2444);
INSERT INTO playlist_track VALUES(8,2445);
INSERT INTO playlist_track VALUES(8,2446);
INSERT INTO playlist_track VALUES(8,2447);
INSERT INTO playlist_track VALUES(8,2448);
INSERT INTO playlist_track VALUES(8,3418);
INSERT INTO playlist_track VALUES(8,2449);
INSERT INTO playlist_track VALUES(8,2450);
INSERT INTO playlist_track VALUES(8,2451);
INSERT INTO playlist_track VALUES(8,2452);
INSERT INTO playlist_track VALUES(8,2453);
INSERT INTO playlist_track VALUES(8,2454);
INSERT INTO playlist_track VALUES(8,2455);
INSERT INTO playlist_track VALUES(8,2456);
INSERT INTO playlist_track VALUES(8,2457);
INSERT INTO playlist_track VALUES(8,2458);
INSERT INTO playlist_track VALUES(8,2459);
INSERT INTO playlist_track VALUES(8,2460);
INSERT INTO playlist_track VALUES(8,2461);
INSERT INTO playlist_track VALUES(8,2462);
INSERT INTO playlist_track VALUES(8,2463);
INSERT INTO playlist_track VALUES(8,2464);
INSERT INTO playlist_track VALUES(8,2465);
INSERT INTO playlist_track VALUES(8,2466);
INSERT INTO playlist_track VALUES(8,2467);
INSERT INTO playlist_track VALUES(8,2468);
INSERT INTO playlist_track VALUES(8,2469);
INSERT INTO playlist_track VALUES(8,2470);
INSERT INTO playlist_track VALUES(8,2471);
INSERT INTO playlist_track VALUES(8,2472);
INSERT INTO playlist_track VALUES(8,2473);
INSERT INTO playlist_track VALUES(8,2474);
INSERT INTO playlist_track VALUES(8,2475);
INSERT INTO playlist_track VALUES(8,2476);
INSERT INTO playlist_track VALUES(8,2477);
INSERT INTO playlist_track VALUES(8,2478);
INSERT INTO playlist_track VALUES(8,2479);
INSERT INTO playlist_track VALUES(8,2480);
INSERT INTO playlist_track VALUES(8,2481);
INSERT INTO playlist_track VALUES(8,2482);
INSERT INTO playlist_track VALUES(8,2483);
INSERT INTO playlist_track VALUES(8,2484);
INSERT INTO playlist_track VALUES(8,2485);
INSERT INTO playlist_track VALUES(8,2486);
INSERT INTO playlist_track VALUES(8,2487);
INSERT INTO playlist_track VALUES(8,2488);
INSERT INTO playlist_track VALUES(8,2489);
INSERT INTO playlist_track VALUES(8,2490);
INSERT INTO playlist_track VALUES(8,2491);
INSERT INTO playlist_track VALUES(8,2492);
INSERT INTO playlist_track VALUES(8,2493);
INSERT INTO playlist_track VALUES(8,2494);
INSERT INTO playlist_track VALUES(8,2495);
INSERT INTO playlist_track VALUES(8,2496);
INSERT INTO playlist_track VALUES(8,2497);
INSERT INTO playlist_track VALUES(8,2498);
INSERT INTO playlist_track VALUES(8,2499);
INSERT INTO playlist_track VALUES(8,2500);
INSERT INTO playlist_track VALUES(8,2501);
INSERT INTO playlist_track VALUES(8,2502);
INSERT INTO playlist_track VALUES(8,2503);
INSERT INTO playlist_track VALUES(8,2504);
INSERT INTO playlist_track VALUES(8,2505);
INSERT INTO playlist_track VALUES(8,3269);
INSERT INTO playlist_track VALUES(8,2506);
INSERT INTO playlist_track VALUES(8,2507);
INSERT INTO playlist_track VALUES(8,2508);
INSERT INTO playlist_track VALUES(8,2509);
INSERT INTO playlist_track VALUES(8,2510);
INSERT INTO playlist_track VALUES(8,2511);
INSERT INTO playlist_track VALUES(8,2512);
INSERT INTO playlist_track VALUES(8,2513);
INSERT INTO playlist_track VALUES(8,2514);
INSERT INTO playlist_track VALUES(8,2515);
INSERT INTO playlist_track VALUES(8,2516);
INSERT INTO playlist_track VALUES(8,2517);
INSERT INTO playlist_track VALUES(8,2518);
INSERT INTO playlist_track VALUES(8,2519);
INSERT INTO playlist_track VALUES(8,2520);
INSERT INTO playlist_track VALUES(8,2521);
INSERT INTO playlist_track VALUES(8,2522);
INSERT INTO playlist_track VALUES(8,456);
INSERT INTO playlist_track VALUES(8,457);
INSERT INTO playlist_track VALUES(8,458);
INSERT INTO playlist_track VALUES(8,459);
INSERT INTO playlist_track VALUES(8,460);
INSERT INTO playlist_track VALUES(8,461);
INSERT INTO playlist_track VALUES(8,462);
INSERT INTO playlist_track VALUES(8,463);
INSERT INTO playlist_track VALUES(8,464);
INSERT INTO playlist_track VALUES(8,465);
INSERT INTO playlist_track VALUES(8,466);
INSERT INTO playlist_track VALUES(8,467);
INSERT INTO playlist_track VALUES(8,2523);
INSERT INTO playlist_track VALUES(8,2524);
INSERT INTO playlist_track VALUES(8,2525);
INSERT INTO playlist_track VALUES(8,2526);
INSERT INTO playlist_track VALUES(8,2527);
INSERT INTO playlist_track VALUES(8,2528);
INSERT INTO playlist_track VALUES(8,2529);
INSERT INTO playlist_track VALUES(8,2530);
INSERT INTO playlist_track VALUES(8,2531);
INSERT INTO playlist_track VALUES(8,3335);
INSERT INTO playlist_track VALUES(8,2532);
INSERT INTO playlist_track VALUES(8,2533);
INSERT INTO playlist_track VALUES(8,2534);
INSERT INTO playlist_track VALUES(8,2535);
INSERT INTO playlist_track VALUES(8,2536);
INSERT INTO playlist_track VALUES(8,2537);
INSERT INTO playlist_track VALUES(8,2538);
INSERT INTO playlist_track VALUES(8,2539);
INSERT INTO playlist_track VALUES(8,2540);
INSERT INTO playlist_track VALUES(8,2541);
INSERT INTO playlist_track VALUES(8,2542);
INSERT INTO playlist_track VALUES(8,2543);
INSERT INTO playlist_track VALUES(8,2544);
INSERT INTO playlist_track VALUES(8,2545);
INSERT INTO playlist_track VALUES(8,2546);
INSERT INTO playlist_track VALUES(8,2547);
INSERT INTO playlist_track VALUES(8,2548);
INSERT INTO playlist_track VALUES(8,2549);
INSERT INTO playlist_track VALUES(8,2550);
INSERT INTO playlist_track VALUES(8,2551);
INSERT INTO playlist_track VALUES(8,2552);
INSERT INTO playlist_track VALUES(8,2553);
INSERT INTO playlist_track VALUES(8,2554);
INSERT INTO playlist_track VALUES(8,2555);
INSERT INTO playlist_track VALUES(8,2556);
INSERT INTO playlist_track VALUES(8,2557);
INSERT INTO playlist_track VALUES(8,2558);
INSERT INTO playlist_track VALUES(8,2559);
INSERT INTO playlist_track VALUES(8,2560);
INSERT INTO playlist_track VALUES(8,2561);
INSERT INTO playlist_track VALUES(8,2562);
INSERT INTO playlist_track VALUES(8,2563);
INSERT INTO playlist_track VALUES(8,2564);
INSERT INTO playlist_track VALUES(8,2705);
INSERT INTO playlist_track VALUES(8,2706);
INSERT INTO playlist_track VALUES(8,2707);
INSERT INTO playlist_track VALUES(8,2708);
INSERT INTO playlist_track VALUES(8,2709);
INSERT INTO playlist_track VALUES(8,2710);
INSERT INTO playlist_track VALUES(8,2711);
INSERT INTO playlist_track VALUES(8,2712);
INSERT INTO playlist_track VALUES(8,2713);
INSERT INTO playlist_track VALUES(8,2714);
INSERT INTO playlist_track VALUES(8,2715);
INSERT INTO playlist_track VALUES(8,2716);
INSERT INTO playlist_track VALUES(8,2717);
INSERT INTO playlist_track VALUES(8,2718);
INSERT INTO playlist_track VALUES(8,2719);
INSERT INTO playlist_track VALUES(8,2720);
INSERT INTO playlist_track VALUES(8,2721);
INSERT INTO playlist_track VALUES(8,2722);
INSERT INTO playlist_track VALUES(8,2723);
INSERT INTO playlist_track VALUES(8,2724);
INSERT INTO playlist_track VALUES(8,2725);
INSERT INTO playlist_track VALUES(8,2726);
INSERT INTO playlist_track VALUES(8,2727);
INSERT INTO playlist_track VALUES(8,2728);
INSERT INTO playlist_track VALUES(8,2729);
INSERT INTO playlist_track VALUES(8,2730);
INSERT INTO playlist_track VALUES(8,3365);
INSERT INTO playlist_track VALUES(8,3366);
INSERT INTO playlist_track VALUES(8,3367);
INSERT INTO playlist_track VALUES(8,3368);
INSERT INTO playlist_track VALUES(8,3369);
INSERT INTO playlist_track VALUES(8,3370);
INSERT INTO playlist_track VALUES(8,3371);
INSERT INTO playlist_track VALUES(8,3372);
INSERT INTO playlist_track VALUES(8,3373);
INSERT INTO playlist_track VALUES(8,3374);
INSERT INTO playlist_track VALUES(8,2565);
INSERT INTO playlist_track VALUES(8,2566);
INSERT INTO playlist_track VALUES(8,2567);
INSERT INTO playlist_track VALUES(8,2568);
INSERT INTO playlist_track VALUES(8,2569);
INSERT INTO playlist_track VALUES(8,2570);
INSERT INTO playlist_track VALUES(8,2571);
INSERT INTO playlist_track VALUES(8,2751);
INSERT INTO playlist_track VALUES(8,2752);
INSERT INTO playlist_track VALUES(8,2753);
INSERT INTO playlist_track VALUES(8,2754);
INSERT INTO playlist_track VALUES(8,2755);
INSERT INTO playlist_track VALUES(8,2756);
INSERT INTO playlist_track VALUES(8,2757);
INSERT INTO playlist_track VALUES(8,2758);
INSERT INTO playlist_track VALUES(8,2759);
INSERT INTO playlist_track VALUES(8,2760);
INSERT INTO playlist_track VALUES(8,2761);
INSERT INTO playlist_track VALUES(8,2762);
INSERT INTO playlist_track VALUES(8,2763);
INSERT INTO playlist_track VALUES(8,2764);
INSERT INTO playlist_track VALUES(8,2765);
INSERT INTO playlist_track VALUES(8,2766);
INSERT INTO playlist_track VALUES(8,2767);
INSERT INTO playlist_track VALUES(8,2768);
INSERT INTO playlist_track VALUES(8,2769);
INSERT INTO playlist_track VALUES(8,2770);
INSERT INTO playlist_track VALUES(8,2771);
INSERT INTO playlist_track VALUES(8,2772);
INSERT INTO playlist_track VALUES(8,2773);
INSERT INTO playlist_track VALUES(8,2774);
INSERT INTO playlist_track VALUES(8,2775);
INSERT INTO playlist_track VALUES(8,2776);
INSERT INTO playlist_track VALUES(8,2777);
INSERT INTO playlist_track VALUES(8,2778);
INSERT INTO playlist_track VALUES(8,2779);
INSERT INTO playlist_track VALUES(8,2780);
INSERT INTO playlist_track VALUES(8,2781);
INSERT INTO playlist_track VALUES(8,2782);
INSERT INTO playlist_track VALUES(8,2783);
INSERT INTO playlist_track VALUES(8,2784);
INSERT INTO playlist_track VALUES(8,2785);
INSERT INTO playlist_track VALUES(8,2786);
INSERT INTO playlist_track VALUES(8,2787);
INSERT INTO playlist_track VALUES(8,2788);
INSERT INTO playlist_track VALUES(8,2789);
INSERT INTO playlist_track VALUES(8,2790);
INSERT INTO playlist_track VALUES(8,2791);
INSERT INTO playlist_track VALUES(8,2792);
INSERT INTO playlist_track VALUES(8,2793);
INSERT INTO playlist_track VALUES(8,2794);
INSERT INTO playlist_track VALUES(8,2795);
INSERT INTO playlist_track VALUES(8,2796);
INSERT INTO playlist_track VALUES(8,2797);
INSERT INTO playlist_track VALUES(8,2798);
INSERT INTO playlist_track VALUES(8,2799);
INSERT INTO playlist_track VALUES(8,2800);
INSERT INTO playlist_track VALUES(8,2801);
INSERT INTO playlist_track VALUES(8,2802);
INSERT INTO playlist_track VALUES(8,2803);
INSERT INTO playlist_track VALUES(8,2804);
INSERT INTO playlist_track VALUES(8,2805);
INSERT INTO playlist_track VALUES(8,2806);
INSERT INTO playlist_track VALUES(8,2807);
INSERT INTO playlist_track VALUES(8,2808);
INSERT INTO playlist_track VALUES(8,2809);
INSERT INTO playlist_track VALUES(8,2810);
INSERT INTO playlist_track VALUES(8,2811);
INSERT INTO playlist_track VALUES(8,2812);
INSERT INTO playlist_track VALUES(8,2813);
INSERT INTO playlist_track VALUES(8,2814);
INSERT INTO playlist_track VALUES(8,2815);
INSERT INTO playlist_track VALUES(8,2816);
INSERT INTO playlist_track VALUES(8,2817);
INSERT INTO playlist_track VALUES(8,2818);
INSERT INTO playlist_track VALUES(8,646);
INSERT INTO playlist_track VALUES(8,647);
INSERT INTO playlist_track VALUES(8,648);
INSERT INTO playlist_track VALUES(8,649);
INSERT INTO playlist_track VALUES(8,651);
INSERT INTO playlist_track VALUES(8,653);
INSERT INTO playlist_track VALUES(8,655);
INSERT INTO playlist_track VALUES(8,658);
INSERT INTO playlist_track VALUES(8,2926);
INSERT INTO playlist_track VALUES(8,2927);
INSERT INTO playlist_track VALUES(8,2928);
INSERT INTO playlist_track VALUES(8,2929);
INSERT INTO playlist_track VALUES(8,2930);
INSERT INTO playlist_track VALUES(8,2931);
INSERT INTO playlist_track VALUES(8,2932);
INSERT INTO playlist_track VALUES(8,2933);
INSERT INTO playlist_track VALUES(8,2934);
INSERT INTO playlist_track VALUES(8,2935);
INSERT INTO playlist_track VALUES(8,2936);
INSERT INTO playlist_track VALUES(8,2937);
INSERT INTO playlist_track VALUES(8,2938);
INSERT INTO playlist_track VALUES(8,2939);
INSERT INTO playlist_track VALUES(8,2940);
INSERT INTO playlist_track VALUES(8,2941);
INSERT INTO playlist_track VALUES(8,2942);
INSERT INTO playlist_track VALUES(8,2943);
INSERT INTO playlist_track VALUES(8,2944);
INSERT INTO playlist_track VALUES(8,2945);
INSERT INTO playlist_track VALUES(8,2946);
INSERT INTO playlist_track VALUES(8,2947);
INSERT INTO playlist_track VALUES(8,2948);
INSERT INTO playlist_track VALUES(8,2949);
INSERT INTO playlist_track VALUES(8,2950);
INSERT INTO playlist_track VALUES(8,2951);
INSERT INTO playlist_track VALUES(8,2952);
INSERT INTO playlist_track VALUES(8,2953);
INSERT INTO playlist_track VALUES(8,2954);
INSERT INTO playlist_track VALUES(8,2955);
INSERT INTO playlist_track VALUES(8,2956);
INSERT INTO playlist_track VALUES(8,2957);
INSERT INTO playlist_track VALUES(8,2958);
INSERT INTO playlist_track VALUES(8,2959);
INSERT INTO playlist_track VALUES(8,2960);
INSERT INTO playlist_track VALUES(8,2961);
INSERT INTO playlist_track VALUES(8,2962);
INSERT INTO playlist_track VALUES(8,2963);
INSERT INTO playlist_track VALUES(8,3004);
INSERT INTO playlist_track VALUES(8,3005);
INSERT INTO playlist_track VALUES(8,3006);
INSERT INTO playlist_track VALUES(8,3007);
INSERT INTO playlist_track VALUES(8,3008);
INSERT INTO playlist_track VALUES(8,3009);
INSERT INTO playlist_track VALUES(8,3010);
INSERT INTO playlist_track VALUES(8,3011);
INSERT INTO playlist_track VALUES(8,3012);
INSERT INTO playlist_track VALUES(8,3013);
INSERT INTO playlist_track VALUES(8,3014);
INSERT INTO playlist_track VALUES(8,3015);
INSERT INTO playlist_track VALUES(8,3016);
INSERT INTO playlist_track VALUES(8,3017);
INSERT INTO playlist_track VALUES(8,2964);
INSERT INTO playlist_track VALUES(8,2965);
INSERT INTO playlist_track VALUES(8,2966);
INSERT INTO playlist_track VALUES(8,2967);
INSERT INTO playlist_track VALUES(8,2968);
INSERT INTO playlist_track VALUES(8,2969);
INSERT INTO playlist_track VALUES(8,2970);
INSERT INTO playlist_track VALUES(8,2971);
INSERT INTO playlist_track VALUES(8,2972);
INSERT INTO playlist_track VALUES(8,2973);
INSERT INTO playlist_track VALUES(8,2974);
INSERT INTO playlist_track VALUES(8,3253);
INSERT INTO playlist_track VALUES(8,2975);
INSERT INTO playlist_track VALUES(8,2976);
INSERT INTO playlist_track VALUES(8,2977);
INSERT INTO playlist_track VALUES(8,2978);
INSERT INTO playlist_track VALUES(8,2979);
INSERT INTO playlist_track VALUES(8,2980);
INSERT INTO playlist_track VALUES(8,2981);
INSERT INTO playlist_track VALUES(8,2982);
INSERT INTO playlist_track VALUES(8,2983);
INSERT INTO playlist_track VALUES(8,2984);
INSERT INTO playlist_track VALUES(8,2985);
INSERT INTO playlist_track VALUES(8,2986);
INSERT INTO playlist_track VALUES(8,2987);
INSERT INTO playlist_track VALUES(8,2988);
INSERT INTO playlist_track VALUES(8,2989);
INSERT INTO playlist_track VALUES(8,2990);
INSERT INTO playlist_track VALUES(8,2991);
INSERT INTO playlist_track VALUES(8,2992);
INSERT INTO playlist_track VALUES(8,2993);
INSERT INTO playlist_track VALUES(8,2994);
INSERT INTO playlist_track VALUES(8,2995);
INSERT INTO playlist_track VALUES(8,2996);
INSERT INTO playlist_track VALUES(8,2997);
INSERT INTO playlist_track VALUES(8,2998);
INSERT INTO playlist_track VALUES(8,2999);
INSERT INTO playlist_track VALUES(8,3000);
INSERT INTO playlist_track VALUES(8,3001);
INSERT INTO playlist_track VALUES(8,3002);
INSERT INTO playlist_track VALUES(8,3003);
INSERT INTO playlist_track VALUES(8,3018);
INSERT INTO playlist_track VALUES(8,3019);
INSERT INTO playlist_track VALUES(8,3020);
INSERT INTO playlist_track VALUES(8,3021);
INSERT INTO playlist_track VALUES(8,3022);
INSERT INTO playlist_track VALUES(8,3023);
INSERT INTO playlist_track VALUES(8,3024);
INSERT INTO playlist_track VALUES(8,3025);
INSERT INTO playlist_track VALUES(8,3026);
INSERT INTO playlist_track VALUES(8,3027);
INSERT INTO playlist_track VALUES(8,3028);
INSERT INTO playlist_track VALUES(8,3029);
INSERT INTO playlist_track VALUES(8,3030);
INSERT INTO playlist_track VALUES(8,3031);
INSERT INTO playlist_track VALUES(8,3032);
INSERT INTO playlist_track VALUES(8,3033);
INSERT INTO playlist_track VALUES(8,3034);
INSERT INTO playlist_track VALUES(8,3035);
INSERT INTO playlist_track VALUES(8,3036);
INSERT INTO playlist_track VALUES(8,3037);
INSERT INTO playlist_track VALUES(8,3038);
INSERT INTO playlist_track VALUES(8,3039);
INSERT INTO playlist_track VALUES(8,3040);
INSERT INTO playlist_track VALUES(8,3041);
INSERT INTO playlist_track VALUES(8,3042);
INSERT INTO playlist_track VALUES(8,3043);
INSERT INTO playlist_track VALUES(8,3044);
INSERT INTO playlist_track VALUES(8,3045);
INSERT INTO playlist_track VALUES(8,3046);
INSERT INTO playlist_track VALUES(8,3047);
INSERT INTO playlist_track VALUES(8,3048);
INSERT INTO playlist_track VALUES(8,3049);
INSERT INTO playlist_track VALUES(8,3050);
INSERT INTO playlist_track VALUES(8,3051);
INSERT INTO playlist_track VALUES(8,3064);
INSERT INTO playlist_track VALUES(8,3065);
INSERT INTO playlist_track VALUES(8,3066);
INSERT INTO playlist_track VALUES(8,3067);
INSERT INTO playlist_track VALUES(8,3068);
INSERT INTO playlist_track VALUES(8,3069);
INSERT INTO playlist_track VALUES(8,3070);
INSERT INTO playlist_track VALUES(8,3071);
INSERT INTO playlist_track VALUES(8,3072);
INSERT INTO playlist_track VALUES(8,3073);
INSERT INTO playlist_track VALUES(8,3074);
INSERT INTO playlist_track VALUES(8,3075);
INSERT INTO playlist_track VALUES(8,3076);
INSERT INTO playlist_track VALUES(8,3077);
INSERT INTO playlist_track VALUES(8,3078);
INSERT INTO playlist_track VALUES(8,3079);
INSERT INTO playlist_track VALUES(8,3080);
INSERT INTO playlist_track VALUES(8,3052);
INSERT INTO playlist_track VALUES(8,3053);
INSERT INTO playlist_track VALUES(8,3054);
INSERT INTO playlist_track VALUES(8,3055);
INSERT INTO playlist_track VALUES(8,3056);
INSERT INTO playlist_track VALUES(8,3057);
INSERT INTO playlist_track VALUES(8,3058);
INSERT INTO playlist_track VALUES(8,3059);
INSERT INTO playlist_track VALUES(8,3060);
INSERT INTO playlist_track VALUES(8,3061);
INSERT INTO playlist_track VALUES(8,3062);
INSERT INTO playlist_track VALUES(8,3063);
INSERT INTO playlist_track VALUES(8,3081);
INSERT INTO playlist_track VALUES(8,3082);
INSERT INTO playlist_track VALUES(8,3083);
INSERT INTO playlist_track VALUES(8,3084);
INSERT INTO playlist_track VALUES(8,3085);
INSERT INTO playlist_track VALUES(8,3086);
INSERT INTO playlist_track VALUES(8,3087);
INSERT INTO playlist_track VALUES(8,3088);
INSERT INTO playlist_track VALUES(8,3089);
INSERT INTO playlist_track VALUES(8,3090);
INSERT INTO playlist_track VALUES(8,3091);
INSERT INTO playlist_track VALUES(8,3092);
INSERT INTO playlist_track VALUES(8,3093);
INSERT INTO playlist_track VALUES(8,3094);
INSERT INTO playlist_track VALUES(8,3095);
INSERT INTO playlist_track VALUES(8,3096);
INSERT INTO playlist_track VALUES(8,3097);
INSERT INTO playlist_track VALUES(8,3098);
INSERT INTO playlist_track VALUES(8,3099);
INSERT INTO playlist_track VALUES(8,3100);
INSERT INTO playlist_track VALUES(8,3101);
INSERT INTO playlist_track VALUES(8,3102);
INSERT INTO playlist_track VALUES(8,3103);
INSERT INTO playlist_track VALUES(8,323);
INSERT INTO playlist_track VALUES(8,324);
INSERT INTO playlist_track VALUES(8,325);
INSERT INTO playlist_track VALUES(8,326);
INSERT INTO playlist_track VALUES(8,327);
INSERT INTO playlist_track VALUES(8,328);
INSERT INTO playlist_track VALUES(8,329);
INSERT INTO playlist_track VALUES(8,330);
INSERT INTO playlist_track VALUES(8,331);
INSERT INTO playlist_track VALUES(8,332);
INSERT INTO playlist_track VALUES(8,333);
INSERT INTO playlist_track VALUES(8,334);
INSERT INTO playlist_track VALUES(8,335);
INSERT INTO playlist_track VALUES(8,336);
INSERT INTO playlist_track VALUES(8,360);
INSERT INTO playlist_track VALUES(8,361);
INSERT INTO playlist_track VALUES(8,362);
INSERT INTO playlist_track VALUES(8,363);
INSERT INTO playlist_track VALUES(8,364);
INSERT INTO playlist_track VALUES(8,365);
INSERT INTO playlist_track VALUES(8,366);
INSERT INTO playlist_track VALUES(8,367);
INSERT INTO playlist_track VALUES(8,368);
INSERT INTO playlist_track VALUES(8,369);
INSERT INTO playlist_track VALUES(8,370);
INSERT INTO playlist_track VALUES(8,371);
INSERT INTO playlist_track VALUES(8,372);
INSERT INTO playlist_track VALUES(8,373);
INSERT INTO playlist_track VALUES(8,556);
INSERT INTO playlist_track VALUES(8,557);
INSERT INTO playlist_track VALUES(8,558);
INSERT INTO playlist_track VALUES(8,559);
INSERT INTO playlist_track VALUES(8,560);
INSERT INTO playlist_track VALUES(8,561);
INSERT INTO playlist_track VALUES(8,562);
INSERT INTO playlist_track VALUES(8,563);
INSERT INTO playlist_track VALUES(8,564);
INSERT INTO playlist_track VALUES(8,565);
INSERT INTO playlist_track VALUES(8,566);
INSERT INTO playlist_track VALUES(8,567);
INSERT INTO playlist_track VALUES(8,568);
INSERT INTO playlist_track VALUES(8,569);
INSERT INTO playlist_track VALUES(8,661);
INSERT INTO playlist_track VALUES(8,662);
INSERT INTO playlist_track VALUES(8,663);
INSERT INTO playlist_track VALUES(8,664);
INSERT INTO playlist_track VALUES(8,665);
INSERT INTO playlist_track VALUES(8,666);
INSERT INTO playlist_track VALUES(8,667);
INSERT INTO playlist_track VALUES(8,668);
INSERT INTO playlist_track VALUES(8,669);
INSERT INTO playlist_track VALUES(8,670);
INSERT INTO playlist_track VALUES(8,671);
INSERT INTO playlist_track VALUES(8,672);
INSERT INTO playlist_track VALUES(8,673);
INSERT INTO playlist_track VALUES(8,674);
INSERT INTO playlist_track VALUES(8,3104);
INSERT INTO playlist_track VALUES(8,3105);
INSERT INTO playlist_track VALUES(8,3106);
INSERT INTO playlist_track VALUES(8,3107);
INSERT INTO playlist_track VALUES(8,3108);
INSERT INTO playlist_track VALUES(8,3109);
INSERT INTO playlist_track VALUES(8,3110);
INSERT INTO playlist_track VALUES(8,3111);
INSERT INTO playlist_track VALUES(8,3112);
INSERT INTO playlist_track VALUES(8,3113);
INSERT INTO playlist_track VALUES(8,3114);
INSERT INTO playlist_track VALUES(8,3115);
INSERT INTO playlist_track VALUES(8,3116);
INSERT INTO playlist_track VALUES(8,3117);
INSERT INTO playlist_track VALUES(8,3118);
INSERT INTO playlist_track VALUES(8,3119);
INSERT INTO playlist_track VALUES(8,3120);
INSERT INTO playlist_track VALUES(8,3121);
INSERT INTO playlist_track VALUES(8,3122);
INSERT INTO playlist_track VALUES(8,3123);
INSERT INTO playlist_track VALUES(8,3124);
INSERT INTO playlist_track VALUES(8,3125);
INSERT INTO playlist_track VALUES(8,3126);
INSERT INTO playlist_track VALUES(8,3127);
INSERT INTO playlist_track VALUES(8,3128);
INSERT INTO playlist_track VALUES(8,3129);
INSERT INTO playlist_track VALUES(8,3130);
INSERT INTO playlist_track VALUES(8,3131);
INSERT INTO playlist_track VALUES(8,652);
INSERT INTO playlist_track VALUES(8,656);
INSERT INTO playlist_track VALUES(8,657);
INSERT INTO playlist_track VALUES(8,650);
INSERT INTO playlist_track VALUES(8,659);
INSERT INTO playlist_track VALUES(8,654);
INSERT INTO playlist_track VALUES(8,660);
INSERT INTO playlist_track VALUES(8,3132);
INSERT INTO playlist_track VALUES(8,3133);
INSERT INTO playlist_track VALUES(8,3134);
INSERT INTO playlist_track VALUES(8,3135);
INSERT INTO playlist_track VALUES(8,3136);
INSERT INTO playlist_track VALUES(8,3137);
INSERT INTO playlist_track VALUES(8,3138);
INSERT INTO playlist_track VALUES(8,3139);
INSERT INTO playlist_track VALUES(8,3140);
INSERT INTO playlist_track VALUES(8,3141);
INSERT INTO playlist_track VALUES(8,3142);
INSERT INTO playlist_track VALUES(8,3143);
INSERT INTO playlist_track VALUES(8,3144);
INSERT INTO playlist_track VALUES(8,3145);
INSERT INTO playlist_track VALUES(8,2731);
INSERT INTO playlist_track VALUES(8,2732);
INSERT INTO playlist_track VALUES(8,2733);
INSERT INTO playlist_track VALUES(8,2734);
INSERT INTO playlist_track VALUES(8,2735);
INSERT INTO playlist_track VALUES(8,2736);
INSERT INTO playlist_track VALUES(8,2737);
INSERT INTO playlist_track VALUES(8,2738);
INSERT INTO playlist_track VALUES(8,2739);
INSERT INTO playlist_track VALUES(8,2740);
INSERT INTO playlist_track VALUES(8,2741);
INSERT INTO playlist_track VALUES(8,2742);
INSERT INTO playlist_track VALUES(8,2743);
INSERT INTO playlist_track VALUES(8,2744);
INSERT INTO playlist_track VALUES(8,2745);
INSERT INTO playlist_track VALUES(8,2746);
INSERT INTO playlist_track VALUES(8,2747);
INSERT INTO playlist_track VALUES(8,2748);
INSERT INTO playlist_track VALUES(8,2749);
INSERT INTO playlist_track VALUES(8,2750);
INSERT INTO playlist_track VALUES(8,3408);
INSERT INTO playlist_track VALUES(8,3320);
INSERT INTO playlist_track VALUES(8,3409);
INSERT INTO playlist_track VALUES(8,3264);
INSERT INTO playlist_track VALUES(8,3146);
INSERT INTO playlist_track VALUES(8,3147);
INSERT INTO playlist_track VALUES(8,3148);
INSERT INTO playlist_track VALUES(8,3149);
INSERT INTO playlist_track VALUES(8,3150);
INSERT INTO playlist_track VALUES(8,3151);
INSERT INTO playlist_track VALUES(8,3152);
INSERT INTO playlist_track VALUES(8,3153);
INSERT INTO playlist_track VALUES(8,3154);
INSERT INTO playlist_track VALUES(8,3155);
INSERT INTO playlist_track VALUES(8,3156);
INSERT INTO playlist_track VALUES(8,3157);
INSERT INTO playlist_track VALUES(8,3158);
INSERT INTO playlist_track VALUES(8,3159);
INSERT INTO playlist_track VALUES(8,3160);
INSERT INTO playlist_track VALUES(8,3161);
INSERT INTO playlist_track VALUES(8,3162);
INSERT INTO playlist_track VALUES(8,3163);
INSERT INTO playlist_track VALUES(8,3164);
INSERT INTO playlist_track VALUES(8,3438);
INSERT INTO playlist_track VALUES(8,3442);
INSERT INTO playlist_track VALUES(8,3436);
INSERT INTO playlist_track VALUES(8,3450);
INSERT INTO playlist_track VALUES(8,3454);
INSERT INTO playlist_track VALUES(8,3432);
INSERT INTO playlist_track VALUES(8,3443);
INSERT INTO playlist_track VALUES(8,3447);
INSERT INTO playlist_track VALUES(8,3452);
INSERT INTO playlist_track VALUES(8,3441);
INSERT INTO playlist_track VALUES(8,3434);
INSERT INTO playlist_track VALUES(8,3449);
INSERT INTO playlist_track VALUES(8,3445);
INSERT INTO playlist_track VALUES(8,3440);
INSERT INTO playlist_track VALUES(8,3453);
INSERT INTO playlist_track VALUES(8,3439);
INSERT INTO playlist_track VALUES(8,3435);
INSERT INTO playlist_track VALUES(8,3448);
INSERT INTO playlist_track VALUES(8,3437);
INSERT INTO playlist_track VALUES(8,3446);
INSERT INTO playlist_track VALUES(8,3444);
INSERT INTO playlist_track VALUES(8,3433);
INSERT INTO playlist_track VALUES(8,3431);
INSERT INTO playlist_track VALUES(8,3451);
INSERT INTO playlist_track VALUES(8,3430);
INSERT INTO playlist_track VALUES(8,3455);
INSERT INTO playlist_track VALUES(8,3456);
INSERT INTO playlist_track VALUES(8,3457);
INSERT INTO playlist_track VALUES(8,3458);
INSERT INTO playlist_track VALUES(8,3459);
INSERT INTO playlist_track VALUES(8,3460);
INSERT INTO playlist_track VALUES(8,3461);
INSERT INTO playlist_track VALUES(8,3462);
INSERT INTO playlist_track VALUES(8,3463);
INSERT INTO playlist_track VALUES(8,3464);
INSERT INTO playlist_track VALUES(8,3465);
INSERT INTO playlist_track VALUES(8,3466);
INSERT INTO playlist_track VALUES(8,3467);
INSERT INTO playlist_track VALUES(8,3468);
INSERT INTO playlist_track VALUES(8,3469);
INSERT INTO playlist_track VALUES(8,3470);
INSERT INTO playlist_track VALUES(8,3471);
INSERT INTO playlist_track VALUES(8,3472);
INSERT INTO playlist_track VALUES(8,3473);
INSERT INTO playlist_track VALUES(8,3474);
INSERT INTO playlist_track VALUES(8,3475);
INSERT INTO playlist_track VALUES(8,3476);
INSERT INTO playlist_track VALUES(8,3477);
INSERT INTO playlist_track VALUES(8,3478);
INSERT INTO playlist_track VALUES(8,3482);
INSERT INTO playlist_track VALUES(8,3485);
INSERT INTO playlist_track VALUES(8,3491);
INSERT INTO playlist_track VALUES(8,3501);
INSERT INTO playlist_track VALUES(8,3487);
INSERT INTO playlist_track VALUES(8,3500);
INSERT INTO playlist_track VALUES(8,3488);
INSERT INTO playlist_track VALUES(8,3499);
INSERT INTO playlist_track VALUES(8,3497);
INSERT INTO playlist_track VALUES(8,3494);
INSERT INTO playlist_track VALUES(8,3495);
INSERT INTO playlist_track VALUES(8,3490);
INSERT INTO playlist_track VALUES(8,3489);
INSERT INTO playlist_track VALUES(8,3492);
INSERT INTO playlist_track VALUES(8,3483);
INSERT INTO playlist_track VALUES(8,3493);
INSERT INTO playlist_track VALUES(8,3498);
INSERT INTO playlist_track VALUES(8,3496);
INSERT INTO playlist_track VALUES(8,3502);
INSERT INTO playlist_track VALUES(8,3479);
INSERT INTO playlist_track VALUES(8,3481);
INSERT INTO playlist_track VALUES(8,3503);
INSERT INTO playlist_track VALUES(8,3486);
INSERT INTO playlist_track VALUES(8,3480);
INSERT INTO playlist_track VALUES(8,3484);
INSERT INTO playlist_track VALUES(9,3402);
INSERT INTO playlist_track VALUES(10,3250);
INSERT INTO playlist_track VALUES(10,2819);
INSERT INTO playlist_track VALUES(10,2820);
INSERT INTO playlist_track VALUES(10,2821);
INSERT INTO playlist_track VALUES(10,2822);
INSERT INTO playlist_track VALUES(10,2823);
INSERT INTO playlist_track VALUES(10,2824);
INSERT INTO playlist_track VALUES(10,2825);
INSERT INTO playlist_track VALUES(10,2826);
INSERT INTO playlist_track VALUES(10,2827);
INSERT INTO playlist_track VALUES(10,2828);
INSERT INTO playlist_track VALUES(10,2829);
INSERT INTO playlist_track VALUES(10,2830);
INSERT INTO playlist_track VALUES(10,2831);
INSERT INTO playlist_track VALUES(10,2832);
INSERT INTO playlist_track VALUES(10,2833);
INSERT INTO playlist_track VALUES(10,2834);
INSERT INTO playlist_track VALUES(10,2835);
INSERT INTO playlist_track VALUES(10,2836);
INSERT INTO playlist_track VALUES(10,2837);
INSERT INTO playlist_track VALUES(10,2838);
INSERT INTO playlist_track VALUES(10,3226);
INSERT INTO playlist_track VALUES(10,3227);
INSERT INTO playlist_track VALUES(10,3228);
INSERT INTO playlist_track VALUES(10,3229);
INSERT INTO playlist_track VALUES(10,3230);
INSERT INTO playlist_track VALUES(10,3231);
INSERT INTO playlist_track VALUES(10,3232);
INSERT INTO playlist_track VALUES(10,3233);
INSERT INTO playlist_track VALUES(10,3234);
INSERT INTO playlist_track VALUES(10,3235);
INSERT INTO playlist_track VALUES(10,3236);
INSERT INTO playlist_track VALUES(10,3237);
INSERT INTO playlist_track VALUES(10,3238);
INSERT INTO playlist_track VALUES(10,3239);
INSERT INTO playlist_track VALUES(10,3240);
INSERT INTO playlist_track VALUES(10,3241);
INSERT INTO playlist_track VALUES(10,3242);
INSERT INTO playlist_track VALUES(10,3243);
INSERT INTO playlist_track VALUES(10,3244);
INSERT INTO playlist_track VALUES(10,3245);
INSERT INTO playlist_track VALUES(10,3246);
INSERT INTO playlist_track VALUES(10,3247);
INSERT INTO playlist_track VALUES(10,3248);
INSERT INTO playlist_track VALUES(10,3249);
INSERT INTO playlist_track VALUES(10,2839);
INSERT INTO playlist_track VALUES(10,2840);
INSERT INTO playlist_track VALUES(10,2841);
INSERT INTO playlist_track VALUES(10,2842);
INSERT INTO playlist_track VALUES(10,2843);
INSERT INTO playlist_track VALUES(10,2844);
INSERT INTO playlist_track VALUES(10,2845);
INSERT INTO playlist_track VALUES(10,2846);
INSERT INTO playlist_track VALUES(10,2847);
INSERT INTO playlist_track VALUES(10,2848);
INSERT INTO playlist_track VALUES(10,2849);
INSERT INTO playlist_track VALUES(10,2850);
INSERT INTO playlist_track VALUES(10,2851);
INSERT INTO playlist_track VALUES(10,2852);
INSERT INTO playlist_track VALUES(10,2853);
INSERT INTO playlist_track VALUES(10,2854);
INSERT INTO playlist_track VALUES(10,2855);
INSERT INTO playlist_track VALUES(10,2856);
INSERT INTO playlist_track VALUES(10,3166);
INSERT INTO playlist_track VALUES(10,3167);
INSERT INTO playlist_track VALUES(10,3168);
INSERT INTO playlist_track VALUES(10,3171);
INSERT INTO playlist_track VALUES(10,3223);
INSERT INTO playlist_track VALUES(10,2858);
INSERT INTO playlist_track VALUES(10,2861);
INSERT INTO playlist_track VALUES(10,2865);
INSERT INTO playlist_track VALUES(10,2868);
INSERT INTO playlist_track VALUES(10,2871);
INSERT INTO playlist_track VALUES(10,2873);
INSERT INTO playlist_track VALUES(10,2877);
INSERT INTO playlist_track VALUES(10,2880);
INSERT INTO playlist_track VALUES(10,2883);
INSERT INTO playlist_track VALUES(10,2885);
INSERT INTO playlist_track VALUES(10,2888);
INSERT INTO playlist_track VALUES(10,2893);
INSERT INTO playlist_track VALUES(10,2894);
INSERT INTO playlist_track VALUES(10,2898);
INSERT INTO playlist_track VALUES(10,2901);
INSERT INTO playlist_track VALUES(10,2904);
INSERT INTO playlist_track VALUES(10,2906);
INSERT INTO playlist_track VALUES(10,2911);
INSERT INTO playlist_track VALUES(10,2913);
INSERT INTO playlist_track VALUES(10,2915);
INSERT INTO playlist_track VALUES(10,2917);
INSERT INTO playlist_track VALUES(10,2919);
INSERT INTO playlist_track VALUES(10,2921);
INSERT INTO playlist_track VALUES(10,2923);
INSERT INTO playlist_track VALUES(10,2925);
INSERT INTO playlist_track VALUES(10,2859);
INSERT INTO playlist_track VALUES(10,2860);
INSERT INTO playlist_track VALUES(10,2864);
INSERT INTO playlist_track VALUES(10,2867);
INSERT INTO playlist_track VALUES(10,2869);
INSERT INTO playlist_track VALUES(10,2872);
INSERT INTO playlist_track VALUES(10,2878);
INSERT INTO playlist_track VALUES(10,2879);
INSERT INTO playlist_track VALUES(10,2884);
INSERT INTO playlist_track VALUES(10,2887);
INSERT INTO playlist_track VALUES(10,2889);
INSERT INTO playlist_track VALUES(10,2892);
INSERT INTO playlist_track VALUES(10,2896);
INSERT INTO playlist_track VALUES(10,2897);
INSERT INTO playlist_track VALUES(10,2902);
INSERT INTO playlist_track VALUES(10,2905);
INSERT INTO playlist_track VALUES(10,2907);
INSERT INTO playlist_track VALUES(10,2910);
INSERT INTO playlist_track VALUES(10,2914);
INSERT INTO playlist_track VALUES(10,2916);
INSERT INTO playlist_track VALUES(10,2918);
INSERT INTO playlist_track VALUES(10,2920);
INSERT INTO playlist_track VALUES(10,2922);
INSERT INTO playlist_track VALUES(10,2924);
INSERT INTO playlist_track VALUES(10,2857);
INSERT INTO playlist_track VALUES(10,2862);
INSERT INTO playlist_track VALUES(10,2863);
INSERT INTO playlist_track VALUES(10,2866);
INSERT INTO playlist_track VALUES(10,2870);
INSERT INTO playlist_track VALUES(10,2874);
INSERT INTO playlist_track VALUES(10,2875);
INSERT INTO playlist_track VALUES(10,2876);
INSERT INTO playlist_track VALUES(10,2881);
INSERT INTO playlist_track VALUES(10,2882);
INSERT INTO playlist_track VALUES(10,2886);
INSERT INTO playlist_track VALUES(10,2890);
INSERT INTO playlist_track VALUES(10,2891);
INSERT INTO playlist_track VALUES(10,2895);
INSERT INTO playlist_track VALUES(10,2899);
INSERT INTO playlist_track VALUES(10,2900);
INSERT INTO playlist_track VALUES(10,2903);
INSERT INTO playlist_track VALUES(10,2908);
INSERT INTO playlist_track VALUES(10,2909);
INSERT INTO playlist_track VALUES(10,2912);
INSERT INTO playlist_track VALUES(10,3165);
INSERT INTO playlist_track VALUES(10,3169);
INSERT INTO playlist_track VALUES(10,3170);
INSERT INTO playlist_track VALUES(10,3252);
INSERT INTO playlist_track VALUES(10,3224);
INSERT INTO playlist_track VALUES(10,3251);
INSERT INTO playlist_track VALUES(10,3340);
INSERT INTO playlist_track VALUES(10,3339);
INSERT INTO playlist_track VALUES(10,3338);
INSERT INTO playlist_track VALUES(10,3337);
INSERT INTO playlist_track VALUES(10,3341);
INSERT INTO playlist_track VALUES(10,3345);
INSERT INTO playlist_track VALUES(10,3342);
INSERT INTO playlist_track VALUES(10,3346);
INSERT INTO playlist_track VALUES(10,3343);
INSERT INTO playlist_track VALUES(10,3347);
INSERT INTO playlist_track VALUES(10,3344);
INSERT INTO playlist_track VALUES(10,3348);
INSERT INTO playlist_track VALUES(10,3360);
INSERT INTO playlist_track VALUES(10,3361);
INSERT INTO playlist_track VALUES(10,3362);
INSERT INTO playlist_track VALUES(10,3363);
INSERT INTO playlist_track VALUES(10,3364);
INSERT INTO playlist_track VALUES(10,3172);
INSERT INTO playlist_track VALUES(10,3173);
INSERT INTO playlist_track VALUES(10,3174);
INSERT INTO playlist_track VALUES(10,3175);
INSERT INTO playlist_track VALUES(10,3176);
INSERT INTO playlist_track VALUES(10,3177);
INSERT INTO playlist_track VALUES(10,3178);
INSERT INTO playlist_track VALUES(10,3179);
INSERT INTO playlist_track VALUES(10,3180);
INSERT INTO playlist_track VALUES(10,3181);
INSERT INTO playlist_track VALUES(10,3182);
INSERT INTO playlist_track VALUES(10,3183);
INSERT INTO playlist_track VALUES(10,3184);
INSERT INTO playlist_track VALUES(10,3185);
INSERT INTO playlist_track VALUES(10,3186);
INSERT INTO playlist_track VALUES(10,3187);
INSERT INTO playlist_track VALUES(10,3188);
INSERT INTO playlist_track VALUES(10,3189);
INSERT INTO playlist_track VALUES(10,3190);
INSERT INTO playlist_track VALUES(10,3191);
INSERT INTO playlist_track VALUES(10,3192);
INSERT INTO playlist_track VALUES(10,3193);
INSERT INTO playlist_track VALUES(10,3194);
INSERT INTO playlist_track VALUES(10,3195);
INSERT INTO playlist_track VALUES(10,3196);
INSERT INTO playlist_track VALUES(10,3197);
INSERT INTO playlist_track VALUES(10,3198);
INSERT INTO playlist_track VALUES(10,3199);
INSERT INTO playlist_track VALUES(10,3200);
INSERT INTO playlist_track VALUES(10,3201);
INSERT INTO playlist_track VALUES(10,3202);
INSERT INTO playlist_track VALUES(10,3203);
INSERT INTO playlist_track VALUES(10,3204);
INSERT INTO playlist_track VALUES(10,3205);
INSERT INTO playlist_track VALUES(10,3206);
INSERT INTO playlist_track VALUES(10,3207);
INSERT INTO playlist_track VALUES(10,3208);
INSERT INTO playlist_track VALUES(10,3209);
INSERT INTO playlist_track VALUES(10,3210);
INSERT INTO playlist_track VALUES(10,3211);
INSERT INTO playlist_track VALUES(10,3212);
INSERT INTO playlist_track VALUES(10,3213);
INSERT INTO playlist_track VALUES(10,3214);
INSERT INTO playlist_track VALUES(10,3215);
INSERT INTO playlist_track VALUES(10,3216);
INSERT INTO playlist_track VALUES(10,3217);
INSERT INTO playlist_track VALUES(10,3218);
INSERT INTO playlist_track VALUES(10,3219);
INSERT INTO playlist_track VALUES(10,3220);
INSERT INTO playlist_track VALUES(10,3221);
INSERT INTO playlist_track VALUES(10,3222);
INSERT INTO playlist_track VALUES(10,3428);
INSERT INTO playlist_track VALUES(10,3429);
INSERT INTO playlist_track VALUES(11,391);
INSERT INTO playlist_track VALUES(11,516);
INSERT INTO playlist_track VALUES(11,523);
INSERT INTO playlist_track VALUES(11,219);
INSERT INTO playlist_track VALUES(11,220);
INSERT INTO playlist_track VALUES(11,215);
INSERT INTO playlist_track VALUES(11,730);
INSERT INTO playlist_track VALUES(11,738);
INSERT INTO playlist_track VALUES(11,228);
INSERT INTO playlist_track VALUES(11,230);
INSERT INTO playlist_track VALUES(11,236);
INSERT INTO playlist_track VALUES(11,852);
INSERT INTO playlist_track VALUES(11,858);
INSERT INTO playlist_track VALUES(11,864);
INSERT INTO playlist_track VALUES(11,867);
INSERT INTO playlist_track VALUES(11,874);
INSERT INTO playlist_track VALUES(11,877);
INSERT INTO playlist_track VALUES(11,885);
INSERT INTO playlist_track VALUES(11,888);
INSERT INTO playlist_track VALUES(11,1088);
INSERT INTO playlist_track VALUES(11,1093);
INSERT INTO playlist_track VALUES(11,1099);
INSERT INTO playlist_track VALUES(11,1105);
INSERT INTO playlist_track VALUES(11,501);
INSERT INTO playlist_track VALUES(11,504);
INSERT INTO playlist_track VALUES(11,1518);
INSERT INTO playlist_track VALUES(11,1519);
INSERT INTO playlist_track VALUES(11,1514);
INSERT INTO playlist_track VALUES(11,1916);
INSERT INTO playlist_track VALUES(11,1928);
INSERT INTO playlist_track VALUES(11,1921);
INSERT INTO playlist_track VALUES(11,2752);
INSERT INTO playlist_track VALUES(11,2753);
INSERT INTO playlist_track VALUES(11,2754);
INSERT INTO playlist_track VALUES(11,2758);
INSERT INTO playlist_track VALUES(11,2767);
INSERT INTO playlist_track VALUES(11,2768);
INSERT INTO playlist_track VALUES(11,2769);
INSERT INTO playlist_track VALUES(11,393);
INSERT INTO playlist_track VALUES(12,3479);
INSERT INTO playlist_track VALUES(12,3480);
INSERT INTO playlist_track VALUES(12,3481);
INSERT INTO playlist_track VALUES(12,3482);
INSERT INTO playlist_track VALUES(12,3483);
INSERT INTO playlist_track VALUES(12,3484);
INSERT INTO playlist_track VALUES(12,3485);
INSERT INTO playlist_track VALUES(12,3486);
INSERT INTO playlist_track VALUES(12,3487);
INSERT INTO playlist_track VALUES(12,3488);
INSERT INTO playlist_track VALUES(12,3489);
INSERT INTO playlist_track VALUES(12,3490);
INSERT INTO playlist_track VALUES(12,3491);
INSERT INTO playlist_track VALUES(12,3492);
INSERT INTO playlist_track VALUES(12,3493);
INSERT INTO playlist_track VALUES(12,3494);
INSERT INTO playlist_track VALUES(12,3495);
INSERT INTO playlist_track VALUES(12,3496);
INSERT INTO playlist_track VALUES(12,3497);
INSERT INTO playlist_track VALUES(12,3498);
INSERT INTO playlist_track VALUES(12,3499);
INSERT INTO playlist_track VALUES(12,3500);
INSERT INTO playlist_track VALUES(12,3501);
INSERT INTO playlist_track VALUES(12,3502);
INSERT INTO playlist_track VALUES(12,3503);
INSERT INTO playlist_track VALUES(12,3430);
INSERT INTO playlist_track VALUES(12,3431);
INSERT INTO playlist_track VALUES(12,3432);
INSERT INTO playlist_track VALUES(12,3433);
INSERT INTO playlist_track VALUES(12,3434);
INSERT INTO playlist_track VALUES(12,3435);
INSERT INTO playlist_track VALUES(12,3436);
INSERT INTO playlist_track VALUES(12,3437);
INSERT INTO playlist_track VALUES(12,3438);
INSERT INTO playlist_track VALUES(12,3439);
INSERT INTO playlist_track VALUES(12,3440);
INSERT INTO playlist_track VALUES(12,3441);
INSERT INTO playlist_track VALUES(12,3442);
INSERT INTO playlist_track VALUES(12,3443);
INSERT INTO playlist_track VALUES(12,3444);
INSERT INTO playlist_track VALUES(12,3445);
INSERT INTO playlist_track VALUES(12,3446);
INSERT INTO playlist_track VALUES(12,3447);
INSERT INTO playlist_track VALUES(12,3448);
INSERT INTO playlist_track VALUES(12,3449);
INSERT INTO playlist_track VALUES(12,3450);
INSERT INTO playlist_track VALUES(12,3451);
INSERT INTO playlist_track VALUES(12,3452);
INSERT INTO playlist_track VALUES(12,3453);
INSERT INTO playlist_track VALUES(12,3454);
INSERT INTO playlist_track VALUES(12,3403);
INSERT INTO playlist_track VALUES(12,3404);
INSERT INTO playlist_track VALUES(12,3405);
INSERT INTO playlist_track VALUES(12,3406);
INSERT INTO playlist_track VALUES(12,3407);
INSERT INTO playlist_track VALUES(12,3408);
INSERT INTO playlist_track VALUES(12,3409);
INSERT INTO playlist_track VALUES(12,3410);
INSERT INTO playlist_track VALUES(12,3411);
INSERT INTO playlist_track VALUES(12,3412);
INSERT INTO playlist_track VALUES(12,3413);
INSERT INTO playlist_track VALUES(12,3414);
INSERT INTO playlist_track VALUES(12,3415);
INSERT INTO playlist_track VALUES(12,3416);
INSERT INTO playlist_track VALUES(12,3417);
INSERT INTO playlist_track VALUES(12,3418);
INSERT INTO playlist_track VALUES(12,3419);
INSERT INTO playlist_track VALUES(12,3420);
INSERT INTO playlist_track VALUES(12,3421);
INSERT INTO playlist_track VALUES(12,3422);
INSERT INTO playlist_track VALUES(12,3423);
INSERT INTO playlist_track VALUES(12,3424);
INSERT INTO playlist_track VALUES(12,3425);
INSERT INTO playlist_track VALUES(12,3426);
INSERT INTO playlist_track VALUES(12,3427);
INSERT INTO playlist_track VALUES(13,3479);
INSERT INTO playlist_track VALUES(13,3480);
INSERT INTO playlist_track VALUES(13,3481);
INSERT INTO playlist_track VALUES(13,3482);
INSERT INTO playlist_track VALUES(13,3483);
INSERT INTO playlist_track VALUES(13,3484);
INSERT INTO playlist_track VALUES(13,3485);
INSERT INTO playlist_track VALUES(13,3486);
INSERT INTO playlist_track VALUES(13,3487);
INSERT INTO playlist_track VALUES(13,3488);
INSERT INTO playlist_track VALUES(13,3489);
INSERT INTO playlist_track VALUES(13,3490);
INSERT INTO playlist_track VALUES(13,3491);
INSERT INTO playlist_track VALUES(13,3492);
INSERT INTO playlist_track VALUES(13,3493);
INSERT INTO playlist_track VALUES(13,3494);
INSERT INTO playlist_track VALUES(13,3495);
INSERT INTO playlist_track VALUES(13,3496);
INSERT INTO playlist_track VALUES(13,3497);
INSERT INTO playlist_track VALUES(13,3498);
INSERT INTO playlist_track VALUES(13,3499);
INSERT INTO playlist_track VALUES(13,3500);
INSERT INTO playlist_track VALUES(13,3501);
INSERT INTO playlist_track VALUES(13,3502);
INSERT INTO playlist_track VALUES(13,3503);
INSERT INTO playlist_track VALUES(14,3430);
INSERT INTO playlist_track VALUES(14,3431);
INSERT INTO playlist_track VALUES(14,3432);
INSERT INTO playlist_track VALUES(14,3433);
INSERT INTO playlist_track VALUES(14,3434);
INSERT INTO playlist_track VALUES(14,3435);
INSERT INTO playlist_track VALUES(14,3436);
INSERT INTO playlist_track VALUES(14,3437);
INSERT INTO playlist_track VALUES(14,3438);
INSERT INTO playlist_track VALUES(14,3439);
INSERT INTO playlist_track VALUES(14,3440);
INSERT INTO playlist_track VALUES(14,3441);
INSERT INTO playlist_track VALUES(14,3442);
INSERT INTO playlist_track VALUES(14,3443);
INSERT INTO playlist_track VALUES(14,3444);
INSERT INTO playlist_track VALUES(14,3445);
INSERT INTO playlist_track VALUES(14,3446);
INSERT INTO playlist_track VALUES(14,3447);
INSERT INTO playlist_track VALUES(14,3448);
INSERT INTO playlist_track VALUES(14,3449);
INSERT INTO playlist_track VALUES(14,3450);
INSERT INTO playlist_track VALUES(14,3451);
INSERT INTO playlist_track VALUES(14,3452);
INSERT INTO playlist_track VALUES(14,3453);
INSERT INTO playlist_track VALUES(14,3454);
INSERT INTO playlist_track VALUES(15,3403);
INSERT INTO playlist_track VALUES(15,3404);
INSERT INTO playlist_track VALUES(15,3405);
INSERT INTO playlist_track VALUES(15,3406);
INSERT INTO playlist_track VALUES(15,3407);
INSERT INTO playlist_track VALUES(15,3408);
INSERT INTO playlist_track VALUES(15,3409);
INSERT INTO playlist_track VALUES(15,3410);
INSERT INTO playlist_track VALUES(15,3411);
INSERT INTO playlist_track VALUES(15,3412);
INSERT INTO playlist_track VALUES(15,3413);
INSERT INTO playlist_track VALUES(15,3414);
INSERT INTO playlist_track VALUES(15,3415);
INSERT INTO playlist_track VALUES(15,3416);
INSERT INTO playlist_track VALUES(15,3417);
INSERT INTO playlist_track VALUES(15,3418);
INSERT INTO playlist_track VALUES(15,3419);
INSERT INTO playlist_track VALUES(15,3420);
INSERT INTO playlist_track VALUES(15,3421);
INSERT INTO playlist_track VALUES(15,3422);
INSERT INTO playlist_track VALUES(15,3423);
INSERT INTO playlist_track VALUES(15,3424);
INSERT INTO playlist_track VALUES(15,3425);
INSERT INTO playlist_track VALUES(15,3426);
INSERT INTO playlist_track VALUES(15,3427);
INSERT INTO playlist_track VALUES(16,3367);
INSERT INTO playlist_track VALUES(16,52);
INSERT INTO playlist_track VALUES(16,2194);
INSERT INTO playlist_track VALUES(16,2195);
INSERT INTO playlist_track VALUES(16,2198);
INSERT INTO playlist_track VALUES(16,2206);
INSERT INTO playlist_track VALUES(16,2512);
INSERT INTO playlist_track VALUES(16,2516);
INSERT INTO playlist_track VALUES(16,2550);
INSERT INTO playlist_track VALUES(16,2003);
INSERT INTO playlist_track VALUES(16,2004);
INSERT INTO playlist_track VALUES(16,2005);
INSERT INTO playlist_track VALUES(16,2007);
INSERT INTO playlist_track VALUES(16,2010);
INSERT INTO playlist_track VALUES(16,2013);
INSERT INTO playlist_track VALUES(17,1);
INSERT INTO playlist_track VALUES(17,2);
INSERT INTO playlist_track VALUES(17,3);
INSERT INTO playlist_track VALUES(17,4);
INSERT INTO playlist_track VALUES(17,5);
INSERT INTO playlist_track VALUES(17,152);
INSERT INTO playlist_track VALUES(17,160);
INSERT INTO playlist_track VALUES(17,1278);
INSERT INTO playlist_track VALUES(17,1283);
INSERT INTO playlist_track VALUES(17,1392);
INSERT INTO playlist_track VALUES(17,1335);
INSERT INTO playlist_track VALUES(17,1345);
INSERT INTO playlist_track VALUES(17,1380);
INSERT INTO playlist_track VALUES(17,1801);
INSERT INTO playlist_track VALUES(17,1830);
INSERT INTO playlist_track VALUES(17,1837);
INSERT INTO playlist_track VALUES(17,1854);
INSERT INTO playlist_track VALUES(17,1876);
INSERT INTO playlist_track VALUES(17,1880);
INSERT INTO playlist_track VALUES(17,1984);
INSERT INTO playlist_track VALUES(17,1942);
INSERT INTO playlist_track VALUES(17,1945);
INSERT INTO playlist_track VALUES(17,2094);
INSERT INTO playlist_track VALUES(17,2095);
INSERT INTO playlist_track VALUES(17,2096);
INSERT INTO playlist_track VALUES(17,3290);
INSERT INTO playlist_track VALUES(18,597);
CREATE TABLE IF NOT EXISTS "tracks"
(
    [TrackId] INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    [Name] NVARCHAR(200)  NOT NULL,
    [AlbumId] INTEGER,
    [MediaTypeId] INTEGER  NOT NULL,
    [GenreId] INTEGER,
    [Composer] NVARCHAR(220),
    [Milliseconds] INTEGER  NOT NULL,
    [Bytes] INTEGER,
    [UnitPrice] NUMERIC(10,2)  NOT NULL,
    FOREIGN KEY ([AlbumId]) REFERENCES "albums" ([AlbumId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION,
    FOREIGN KEY ([GenreId]) REFERENCES "genres" ([GenreId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION,
    FOREIGN KEY ([MediaTypeId]) REFERENCES "media_types" ([MediaTypeId]) 
		ON DELETE NO ACTION ON UPDATE NO ACTION
);
INSERT INTO tracks VALUES(1,'For Those About To Rock (We Salute You)',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',343719,11170334,0.98999999999999999111);
INSERT INTO tracks VALUES(2,'Balls to the Wall',2,2,1,NULL,342562,5510424,0.98999999999999999111);
INSERT INTO tracks VALUES(3,'Fast As a Shark',3,2,1,'F. Baltes, S. Kaufman, U. Dirkscneider & W. Hoffman',230619,3990994,0.98999999999999999111);
INSERT INTO tracks VALUES(4,'Restless and Wild',3,2,1,'F. Baltes, R.A. Smith-Diesel, S. Kaufman, U. Dirkscneider & W. Hoffman',252051,4331779,0.98999999999999999111);
INSERT INTO tracks VALUES(5,'Princess of the Dawn',3,2,1,'Deaffy & R.A. Smith-Diesel',375418,6290521,0.98999999999999999111);
INSERT INTO tracks VALUES(6,'Put The Finger On You',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',205662,6713451,0.98999999999999999111);
INSERT INTO tracks VALUES(7,'Let''s Get It Up',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',233926,7636561,0.98999999999999999111);
INSERT INTO tracks VALUES(8,'Inject The Venom',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',210834,6852860,0.98999999999999999111);
INSERT INTO tracks VALUES(9,'Snowballed',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',203102,6599424,0.98999999999999999111);
INSERT INTO tracks VALUES(10,'Evil Walks',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',263497,8611245,0.98999999999999999111);
INSERT INTO tracks VALUES(11,'C.O.D.',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',199836,6566314,0.98999999999999999111);
INSERT INTO tracks VALUES(12,'Breaking The Rules',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',263288,8596840,0.98999999999999999111);
INSERT INTO tracks VALUES(13,'Night Of The Long Knives',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',205688,6706347,0.98999999999999999111);
INSERT INTO tracks VALUES(14,'Spellbound',1,1,1,'Angus Young, Malcolm Young, Brian Johnson',270863,8817038,0.98999999999999999111);
INSERT INTO tracks VALUES(15,'Go Down',4,1,1,'AC/DC',331180,10847611,0.98999999999999999111);
INSERT INTO tracks VALUES(16,'Dog Eat Dog',4,1,1,'AC/DC',215196,7032162,0.98999999999999999111);
INSERT INTO tracks VALUES(17,'Let There Be Rock',4,1,1,'AC/DC',366654,12021261,0.98999999999999999111);
INSERT INTO tracks VALUES(18,'Bad Boy Boogie',4,1,1,'AC/DC',267728,8776140,0.98999999999999999111);
INSERT INTO tracks VALUES(19,'Problem Child',4,1,1,'AC/DC',325041,10617116,0.98999999999999999111);
INSERT INTO tracks VALUES(20,'Overdose',4,1,1,'AC/DC',369319,12066294,0.98999999999999999111);
INSERT INTO tracks VALUES(21,'Hell Ain''t A Bad Place To Be',4,1,1,'AC/DC',254380,8331286,0.98999999999999999111);
INSERT INTO tracks VALUES(22,'Whole Lotta Rosie',4,1,1,'AC/DC',323761,10547154,0.98999999999999999111);
INSERT INTO tracks VALUES(23,'Walk On Water',5,1,1,'Steven Tyler, Joe Perry, Jack Blades, Tommy Shaw',295680,9719579,0.98999999999999999111);
INSERT INTO tracks VALUES(24,'Love In An Elevator',5,1,1,'Steven Tyler, Joe Perry',321828,10552051,0.98999999999999999111);
INSERT INTO tracks VALUES(25,'Rag Doll',5,1,1,'Steven Tyler, Joe Perry, Jim Vallance, Holly Knight',264698,8675345,0.98999999999999999111);
INSERT INTO tracks VALUES(26,'What It Takes',5,1,1,'Steven Tyler, Joe Perry, Desmond Child',310622,10144730,0.98999999999999999111);
INSERT INTO tracks VALUES(27,'Dude (Looks Like A Lady)',5,1,1,'Steven Tyler, Joe Perry, Desmond Child',264855,8679940,0.98999999999999999111);
INSERT INTO tracks VALUES(28,'Janie''s Got A Gun',5,1,1,'Steven Tyler, Tom Hamilton',330736,10869391,0.98999999999999999111);
INSERT INTO tracks VALUES(29,'Cryin''',5,1,1,'Steven Tyler, Joe Perry, Taylor Rhodes',309263,10056995,0.98999999999999999111);
INSERT INTO tracks VALUES(30,'Amazing',5,1,1,'Steven Tyler, Richie Supa',356519,11616195,0.98999999999999999111);
INSERT INTO tracks VALUES(31,'Blind Man',5,1,1,'Steven Tyler, Joe Perry, Taylor Rhodes',240718,7877453,0.98999999999999999111);
INSERT INTO tracks VALUES(32,'Deuces Are Wild',5,1,1,'Steven Tyler, Jim Vallance',215875,7074167,0.98999999999999999111);
INSERT INTO tracks VALUES(33,'The Other Side',5,1,1,'Steven Tyler, Jim Vallance',244375,7983270,0.98999999999999999111);
INSERT INTO tracks VALUES(34,'Crazy',5,1,1,'Steven Tyler, Joe Perry, Desmond Child',316656,10402398,0.98999999999999999111);
INSERT INTO tracks VALUES(35,'Eat The Rich',5,1,1,'Steven Tyler, Joe Perry, Jim Vallance',251036,8262039,0.98999999999999999111);
INSERT INTO tracks VALUES(36,'Angel',5,1,1,'Steven Tyler, Desmond Child',307617,9989331,0.98999999999999999111);
INSERT INTO tracks VALUES(37,'Livin'' On The Edge',5,1,1,'Steven Tyler, Joe Perry, Mark Hudson',381231,12374569,0.98999999999999999111);
INSERT INTO tracks VALUES(38,'All I Really Want',6,1,1,'Alanis Morissette & Glenn Ballard',284891,9375567,0.98999999999999999111);
INSERT INTO tracks VALUES(39,'You Oughta Know',6,1,1,'Alanis Morissette & Glenn Ballard',249234,8196916,0.98999999999999999111);
INSERT INTO tracks VALUES(40,'Perfect',6,1,1,'Alanis Morissette & Glenn Ballard',188133,6145404,0.98999999999999999111);
INSERT INTO tracks VALUES(41,'Hand In My Pocket',6,1,1,'Alanis Morissette & Glenn Ballard',221570,7224246,0.98999999999999999111);
INSERT INTO tracks VALUES(42,'Right Through You',6,1,1,'Alanis Morissette & Glenn Ballard',176117,5793082,0.98999999999999999111);
INSERT INTO tracks VALUES(43,'Forgiven',6,1,1,'Alanis Morissette & Glenn Ballard',300355,9753256,0.98999999999999999111);
INSERT INTO tracks VALUES(44,'You Learn',6,1,1,'Alanis Morissette & Glenn Ballard',239699,7824837,0.98999999999999999111);
INSERT INTO tracks VALUES(45,'Head Over Feet',6,1,1,'Alanis Morissette & Glenn Ballard',267493,8758008,0.98999999999999999111);
INSERT INTO tracks VALUES(46,'Mary Jane',6,1,1,'Alanis Morissette & Glenn Ballard',280607,9163588,0.98999999999999999111);
INSERT INTO tracks VALUES(47,'Ironic',6,1,1,'Alanis Morissette & Glenn Ballard',229825,7598866,0.98999999999999999111);
INSERT INTO tracks VALUES(48,'Not The Doctor',6,1,1,'Alanis Morissette & Glenn Ballard',227631,7604601,0.98999999999999999111);
INSERT INTO tracks VALUES(49,'Wake Up',6,1,1,'Alanis Morissette & Glenn Ballard',293485,9703359,0.98999999999999999111);
INSERT INTO tracks VALUES(50,'You Oughta Know (Alternate)',6,1,1,'Alanis Morissette & Glenn Ballard',491885,16008629,0.98999999999999999111);
INSERT INTO tracks VALUES(51,'We Die Young',7,1,1,'Jerry Cantrell',152084,4925362,0.98999999999999999111);
INSERT INTO tracks VALUES(52,'Man In The Box',7,1,1,'Jerry Cantrell, Layne Staley',286641,9310272,0.98999999999999999111);
INSERT INTO tracks VALUES(53,'Sea Of Sorrow',7,1,1,'Jerry Cantrell',349831,11316328,0.98999999999999999111);
INSERT INTO tracks VALUES(54,'Bleed The Freak',7,1,1,'Jerry Cantrell',241946,7847716,0.98999999999999999111);
INSERT INTO tracks VALUES(55,'I Can''t Remember',7,1,1,'Jerry Cantrell, Layne Staley',222955,7302550,0.98999999999999999111);
INSERT INTO tracks VALUES(56,'Love, Hate, Love',7,1,1,'Jerry Cantrell, Layne Staley',387134,12575396,0.98999999999999999111);
INSERT INTO tracks VALUES(57,'It Ain''t Like That',7,1,1,'Jerry Cantrell, Michael Starr, Sean Kinney',277577,8993793,0.98999999999999999111);
INSERT INTO tracks VALUES(58,'Sunshine',7,1,1,'Jerry Cantrell',284969,9216057,0.98999999999999999111);
INSERT INTO tracks VALUES(59,'Put You Down',7,1,1,'Jerry Cantrell',196231,6420530,0.98999999999999999111);
INSERT INTO tracks VALUES(60,'Confusion',7,1,1,'Jerry Cantrell, Michael Starr, Layne Staley',344163,11183647,0.98999999999999999111);
INSERT INTO tracks VALUES(61,'I Know Somethin (Bout You)',7,1,1,'Jerry Cantrell',261955,8497788,0.98999999999999999111);
INSERT INTO tracks VALUES(62,'Real Thing',7,1,1,'Jerry Cantrell, Layne Staley',243879,7937731,0.98999999999999999111);
INSERT INTO tracks VALUES(63,'Desafinado',8,1,2,NULL,185338,5990473,0.98999999999999999111);
INSERT INTO tracks VALUES(64,'Garota De Ipanema',8,1,2,NULL,285048,9348428,0.98999999999999999111);
INSERT INTO tracks VALUES(65,'Samba De Uma Nota Só (One Note Samba)',8,1,2,NULL,137273,4535401,0.98999999999999999111);
INSERT INTO tracks VALUES(66,'Por Causa De Você',8,1,2,NULL,169900,5536496,0.98999999999999999111);
INSERT INTO tracks VALUES(67,'Ligia',8,1,2,NULL,251977,8226934,0.98999999999999999111);
INSERT INTO tracks VALUES(68,'Fotografia',8,1,2,NULL,129227,4198774,0.98999999999999999111);
INSERT INTO tracks VALUES(69,'Dindi (Dindi)',8,1,2,NULL,253178,8149148,0.98999999999999999111);
INSERT INTO tracks VALUES(70,'Se Todos Fossem Iguais A Você (Instrumental)',8,1,2,NULL,134948,4393377,0.98999999999999999111);
INSERT INTO tracks VALUES(71,'Falando De Amor',8,1,2,NULL,219663,7121735,0.98999999999999999111);
INSERT INTO tracks VALUES(72,'Angela',8,1,2,NULL,169508,5574957,0.98999999999999999111);
INSERT INTO tracks VALUES(73,'Corcovado (Quiet Nights Of Quiet Stars)',8,1,2,NULL,205662,6687994,0.98999999999999999111);
INSERT INTO tracks VALUES(74,'Outra Vez',8,1,2,NULL,126511,4110053,0.98999999999999999111);
INSERT INTO tracks VALUES(75,'O Boto (Bôto)',8,1,2,NULL,366837,12089673,0.98999999999999999111);
INSERT INTO tracks VALUES(76,'Canta, Canta Mais',8,1,2,NULL,271856,8719426,0.98999999999999999111);
INSERT INTO tracks VALUES(77,'Enter Sandman',9,1,3,'Apocalyptica',221701,7286305,0.98999999999999999111);
INSERT INTO tracks VALUES(78,'Master Of Puppets',9,1,3,'Apocalyptica',436453,14375310,0.98999999999999999111);
INSERT INTO tracks VALUES(79,'Harvester Of Sorrow',9,1,3,'Apocalyptica',374543,12372536,0.98999999999999999111);
INSERT INTO tracks VALUES(80,'The Unforgiven',9,1,3,'Apocalyptica',322925,10422447,0.98999999999999999111);
INSERT INTO tracks VALUES(81,'Sad But True',9,1,3,'Apocalyptica',288208,9405526,0.98999999999999999111);
INSERT INTO tracks VALUES(82,'Creeping Death',9,1,3,'Apocalyptica',308035,10110980,0.98999999999999999111);
INSERT INTO tracks VALUES(83,'Wherever I May Roam',9,1,3,'Apocalyptica',369345,12033110,0.98999999999999999111);
INSERT INTO tracks VALUES(84,'Welcome Home (Sanitarium)',9,1,3,'Apocalyptica',350197,11406431,0.98999999999999999111);
INSERT INTO tracks VALUES(85,'Cochise',10,1,1,'Audioslave/Chris Cornell',222380,5339931,0.98999999999999999111);
INSERT INTO tracks VALUES(86,'Show Me How to Live',10,1,1,'Audioslave/Chris Cornell',277890,6672176,0.98999999999999999111);
INSERT INTO tracks VALUES(87,'Gasoline',10,1,1,'Audioslave/Chris Cornell',279457,6709793,0.98999999999999999111);
INSERT INTO tracks VALUES(88,'What You Are',10,1,1,'Audioslave/Chris Cornell',249391,5988186,0.98999999999999999111);
INSERT INTO tracks VALUES(89,'Like a Stone',10,1,1,'Audioslave/Chris Cornell',294034,7059624,0.98999999999999999111);
INSERT INTO tracks VALUES(90,'Set It Off',10,1,1,'Audioslave/Chris Cornell',263262,6321091,0.98999999999999999111);
INSERT INTO tracks VALUES(91,'Shadow on the Sun',10,1,1,'Audioslave/Chris Cornell',343457,8245793,0.98999999999999999111);
INSERT INTO tracks VALUES(92,'I am the Highway',10,1,1,'Audioslave/Chris Cornell',334942,8041411,0.98999999999999999111);
INSERT INTO tracks VALUES(93,'Exploder',10,1,1,'Audioslave/Chris Cornell',206053,4948095,0.98999999999999999111);
INSERT INTO tracks VALUES(94,'Hypnotize',10,1,1,'Audioslave/Chris Cornell',206628,4961887,0.98999999999999999111);
INSERT INTO tracks VALUES(95,'Bring''em Back Alive',10,1,1,'Audioslave/Chris Cornell',329534,7911634,0.98999999999999999111);
INSERT INTO tracks VALUES(96,'Light My Way',10,1,1,'Audioslave/Chris Cornell',303595,7289084,0.98999999999999999111);
INSERT INTO tracks VALUES(97,'Getaway Car',10,1,1,'Audioslave/Chris Cornell',299598,7193162,0.98999999999999999111);
INSERT INTO tracks VALUES(98,'The Last Remaining Light',10,1,1,'Audioslave/Chris Cornell',317492,7622615,0.98999999999999999111);
INSERT INTO tracks VALUES(99,'Your Time Has Come',11,1,4,'Cornell, Commerford, Morello, Wilk',255529,8273592,0.98999999999999999111);
INSERT INTO tracks VALUES(100,'Out Of Exile',11,1,4,'Cornell, Commerford, Morello, Wilk',291291,9506571,0.98999999999999999111);
INSERT INTO tracks VALUES(101,'Be Yourself',11,1,4,'Cornell, Commerford, Morello, Wilk',279484,9106160,0.98999999999999999111);
INSERT INTO tracks VALUES(102,'Doesn''t Remind Me',11,1,4,'Cornell, Commerford, Morello, Wilk',255869,8357387,0.98999999999999999111);
INSERT INTO tracks VALUES(103,'Drown Me Slowly',11,1,4,'Cornell, Commerford, Morello, Wilk',233691,7609178,0.98999999999999999111);
INSERT INTO tracks VALUES(104,'Heaven''s Dead',11,1,4,'Cornell, Commerford, Morello, Wilk',276688,9006158,0.98999999999999999111);
INSERT INTO tracks VALUES(105,'The Worm',11,1,4,'Cornell, Commerford, Morello, Wilk',237714,7710800,0.98999999999999999111);
INSERT INTO tracks VALUES(106,'Man Or Animal',11,1,4,'Cornell, Commerford, Morello, Wilk',233195,7542942,0.98999999999999999111);
INSERT INTO tracks VALUES(107,'Yesterday To Tomorrow',11,1,4,'Cornell, Commerford, Morello, Wilk',273763,8944205,0.98999999999999999111);
INSERT INTO tracks VALUES(108,'Dandelion',11,1,4,'Cornell, Commerford, Morello, Wilk',278125,9003592,0.98999999999999999111);
INSERT INTO tracks VALUES(109,'#1 Zero',11,1,4,'Cornell, Commerford, Morello, Wilk',299102,9731988,0.98999999999999999111);
INSERT INTO tracks VALUES(110,'The Curse',11,1,4,'Cornell, Commerford, Morello, Wilk',309786,10029406,0.98999999999999999111);
INSERT INTO tracks VALUES(111,'Money',12,1,5,'Berry Gordy, Jr./Janie Bradford',147591,2365897,0.98999999999999999111);
INSERT INTO tracks VALUES(112,'Long Tall Sally',12,1,5,'Enotris Johnson/Little Richard/Robert "Bumps" Blackwell',106396,1707084,0.98999999999999999111);
INSERT INTO tracks VALUES(113,'Bad Boy',12,1,5,'Larry Williams',116088,1862126,0.98999999999999999111);
INSERT INTO tracks VALUES(114,'Twist And Shout',12,1,5,'Bert Russell/Phil Medley',161123,2582553,0.98999999999999999111);
INSERT INTO tracks VALUES(115,'Please Mr. Postman',12,1,5,'Brian Holland/Freddie Gorman/Georgia Dobbins/Robert Bateman/William Garrett',137639,2206986,0.98999999999999999111);
INSERT INTO tracks VALUES(116,'C''Mon Everybody',12,1,5,'Eddie Cochran/Jerry Capehart',140199,2247846,0.98999999999999999111);
INSERT INTO tracks VALUES(117,'Rock ''N'' Roll Music',12,1,5,'Chuck Berry',141923,2276788,0.98999999999999999111);
INSERT INTO tracks VALUES(118,'Slow Down',12,1,5,'Larry Williams',163265,2616981,0.98999999999999999111);
INSERT INTO tracks VALUES(119,'Roadrunner',12,1,5,'Bo Diddley',143595,2301989,0.98999999999999999111);
INSERT INTO tracks VALUES(120,'Carol',12,1,5,'Chuck Berry',143830,2306019,0.98999999999999999111);
INSERT INTO tracks VALUES(121,'Good Golly Miss Molly',12,1,5,'Little Richard',106266,1704918,0.98999999999999999111);
INSERT INTO tracks VALUES(122,'20 Flight Rock',12,1,5,'Ned Fairchild',107807,1299960,0.98999999999999999111);
INSERT INTO tracks VALUES(123,'Quadrant',13,1,2,'Billy Cobham',261851,8538199,0.98999999999999999111);
INSERT INTO tracks VALUES(124,'Snoopy''s search-Red baron',13,1,2,'Billy Cobham',456071,15075616,0.98999999999999999111);
INSERT INTO tracks VALUES(125,'Spanish moss-"A sound portrait"-Spanish moss',13,1,2,'Billy Cobham',248084,8217867,0.98999999999999999111);
INSERT INTO tracks VALUES(126,'Moon germs',13,1,2,'Billy Cobham',294060,9714812,0.98999999999999999111);
INSERT INTO tracks VALUES(127,'Stratus',13,1,2,'Billy Cobham',582086,19115680,0.98999999999999999111);
INSERT INTO tracks VALUES(128,'The pleasant pheasant',13,1,2,'Billy Cobham',318066,10630578,0.98999999999999999111);
INSERT INTO tracks VALUES(129,'Solo-Panhandler',13,1,2,'Billy Cobham',246151,8230661,0.98999999999999999111);
INSERT INTO tracks VALUES(130,'Do what cha wanna',13,1,2,'George Duke',274155,9018565,0.98999999999999999111);
INSERT INTO tracks VALUES(131,'Intro/ Low Down',14,1,3,NULL,323683,10642901,0.98999999999999999111);
INSERT INTO tracks VALUES(132,'13 Years Of Grief',14,1,3,NULL,246987,8137421,0.98999999999999999111);
INSERT INTO tracks VALUES(133,'Stronger Than Death',14,1,3,NULL,300747,9869647,0.98999999999999999111);
INSERT INTO tracks VALUES(134,'All For You',14,1,3,NULL,235833,7726948,0.98999999999999999111);
INSERT INTO tracks VALUES(135,'Super Terrorizer',14,1,3,NULL,319373,10513905,0.98999999999999999111);
INSERT INTO tracks VALUES(136,'Phoney Smile Fake Hellos',14,1,3,NULL,273606,9011701,0.98999999999999999111);
INSERT INTO tracks VALUES(137,'Lost My Better Half',14,1,3,NULL,284081,9355309,0.98999999999999999111);
INSERT INTO tracks VALUES(138,'Bored To Tears',14,1,3,NULL,247327,8130090,0.98999999999999999111);
INSERT INTO tracks VALUES(139,'A.N.D.R.O.T.A.Z.',14,1,3,NULL,266266,8574746,0.98999999999999999111);
INSERT INTO tracks VALUES(140,'Born To Booze',14,1,3,NULL,282122,9257358,0.98999999999999999111);
INSERT INTO tracks VALUES(141,'World Of Trouble',14,1,3,NULL,359157,11820932,0.98999999999999999111);
INSERT INTO tracks VALUES(142,'No More Tears',14,1,3,NULL,555075,18041629,0.98999999999999999111);
INSERT INTO tracks VALUES(143,'The Begining... At Last',14,1,3,NULL,365662,11965109,0.98999999999999999111);
INSERT INTO tracks VALUES(144,'Heart Of Gold',15,1,3,NULL,194873,6417460,0.98999999999999999111);
INSERT INTO tracks VALUES(145,'Snowblind',15,1,3,NULL,420022,13842549,0.98999999999999999111);
INSERT INTO tracks VALUES(146,'Like A Bird',15,1,3,NULL,276532,9115657,0.98999999999999999111);
INSERT INTO tracks VALUES(147,'Blood In The Wall',15,1,3,NULL,284368,9359475,0.98999999999999999111);
INSERT INTO tracks VALUES(148,'The Beginning...At Last',15,1,3,NULL,271960,8975814,0.98999999999999999111);
INSERT INTO tracks VALUES(149,'Black Sabbath',16,1,3,NULL,382066,12440200,0.98999999999999999111);
INSERT INTO tracks VALUES(150,'The Wizard',16,1,3,NULL,264829,8646737,0.98999999999999999111);
INSERT INTO tracks VALUES(151,'Behind The Wall Of Sleep',16,1,3,NULL,217573,7169049,0.98999999999999999111);
INSERT INTO tracks VALUES(152,'N.I.B.',16,1,3,NULL,368770,12029390,0.98999999999999999111);
INSERT INTO tracks VALUES(153,'Evil Woman',16,1,3,NULL,204930,6655170,0.98999999999999999111);
INSERT INTO tracks VALUES(154,'Sleeping Village',16,1,3,NULL,644571,21128525,0.98999999999999999111);
INSERT INTO tracks VALUES(155,'Warning',16,1,3,NULL,212062,6893363,0.98999999999999999111);
INSERT INTO tracks VALUES(156,'Wheels Of Confusion / The Straightener',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',494524,16065830,0.98999999999999999111);
INSERT INTO tracks VALUES(157,'Tomorrow''s Dream',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',192496,6252071,0.98999999999999999111);
INSERT INTO tracks VALUES(158,'Changes',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',286275,9175517,0.98999999999999999111);
INSERT INTO tracks VALUES(159,'FX',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',103157,3331776,0.98999999999999999111);
INSERT INTO tracks VALUES(160,'Supernaut',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',285779,9245971,0.98999999999999999111);
INSERT INTO tracks VALUES(161,'Snowblind',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',331676,10813386,0.98999999999999999111);
INSERT INTO tracks VALUES(162,'Cornucopia',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',234814,7653880,0.98999999999999999111);
INSERT INTO tracks VALUES(163,'Laguna Sunrise',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',173087,5671374,0.98999999999999999111);
INSERT INTO tracks VALUES(164,'St. Vitus Dance',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',149655,4884969,0.98999999999999999111);
INSERT INTO tracks VALUES(165,'Under The Sun/Every Day Comes and Goes',17,1,3,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',350458,11360486,0.98999999999999999111);
INSERT INTO tracks VALUES(166,'Smoked Pork',18,1,4,NULL,47333,1549074,0.98999999999999999111);
INSERT INTO tracks VALUES(167,'Body Count''s In The House',18,1,4,NULL,204251,6715413,0.98999999999999999111);
INSERT INTO tracks VALUES(168,'Now Sports',18,1,4,NULL,4884,161266,0.98999999999999999111);
INSERT INTO tracks VALUES(169,'Body Count',18,1,4,NULL,317936,10489139,0.98999999999999999111);
INSERT INTO tracks VALUES(170,'A Statistic',18,1,4,NULL,6373,211997,0.98999999999999999111);
INSERT INTO tracks VALUES(171,'Bowels Of The Devil',18,1,4,NULL,223216,7324125,0.98999999999999999111);
INSERT INTO tracks VALUES(172,'The Real Problem',18,1,4,NULL,11650,387360,0.98999999999999999111);
INSERT INTO tracks VALUES(173,'KKK Bitch',18,1,4,NULL,173008,5709631,0.98999999999999999111);
INSERT INTO tracks VALUES(174,'D Note',18,1,4,NULL,95738,3067064,0.98999999999999999111);
INSERT INTO tracks VALUES(175,'Voodoo',18,1,4,NULL,300721,9875962,0.98999999999999999111);
INSERT INTO tracks VALUES(176,'The Winner Loses',18,1,4,NULL,392254,12843821,0.98999999999999999111);
INSERT INTO tracks VALUES(177,'There Goes The Neighborhood',18,1,4,NULL,350171,11443471,0.98999999999999999111);
INSERT INTO tracks VALUES(178,'Oprah',18,1,4,NULL,6635,224313,0.98999999999999999111);
INSERT INTO tracks VALUES(179,'Evil Dick',18,1,4,NULL,239020,7828873,0.98999999999999999111);
INSERT INTO tracks VALUES(180,'Body Count Anthem',18,1,4,NULL,166426,5463690,0.98999999999999999111);
INSERT INTO tracks VALUES(181,'Momma''s Gotta Die Tonight',18,1,4,NULL,371539,12122946,0.98999999999999999111);
INSERT INTO tracks VALUES(182,'Freedom Of Speech',18,1,4,NULL,281234,9337917,0.98999999999999999111);
INSERT INTO tracks VALUES(183,'King In Crimson',19,1,3,'Roy Z',283167,9218499,0.98999999999999999111);
INSERT INTO tracks VALUES(184,'Chemical Wedding',19,1,3,'Roy Z',246177,8022764,0.98999999999999999111);
INSERT INTO tracks VALUES(185,'The Tower',19,1,3,'Roy Z',285257,9435693,0.98999999999999999111);
INSERT INTO tracks VALUES(186,'Killing Floor',19,1,3,'Adrian Smith',269557,8854240,0.98999999999999999111);
INSERT INTO tracks VALUES(187,'Book Of Thel',19,1,3,'Eddie Casillas/Roy Z',494393,16034404,0.98999999999999999111);
INSERT INTO tracks VALUES(188,'Gates Of Urizen',19,1,3,'Roy Z',265351,8627004,0.98999999999999999111);
INSERT INTO tracks VALUES(189,'Jerusalem',19,1,3,'Roy Z',402390,13194463,0.98999999999999999111);
INSERT INTO tracks VALUES(190,'Trupets Of Jericho',19,1,3,'Roy Z',359131,11820908,0.98999999999999999111);
INSERT INTO tracks VALUES(191,'Machine Men',19,1,3,'Adrian Smith',341655,11138147,0.98999999999999999111);
INSERT INTO tracks VALUES(192,'The Alchemist',19,1,3,'Roy Z',509413,16545657,0.98999999999999999111);
INSERT INTO tracks VALUES(193,'Realword',19,1,3,'Roy Z',237531,7802095,0.98999999999999999111);
INSERT INTO tracks VALUES(194,'First Time I Met The Blues',20,1,6,'Eurreal Montgomery',140434,4604995,0.98999999999999999111);
INSERT INTO tracks VALUES(195,'Let Me Love You Baby',20,1,6,'Willie Dixon',175386,5716994,0.98999999999999999111);
INSERT INTO tracks VALUES(196,'Stone Crazy',20,1,6,'Buddy Guy',433397,14184984,0.98999999999999999111);
INSERT INTO tracks VALUES(197,'Pretty Baby',20,1,6,'Willie Dixon',237662,7848282,0.98999999999999999111);
INSERT INTO tracks VALUES(198,'When My Left Eye Jumps',20,1,6,'Al Perkins/Willie Dixon',235311,7685363,0.98999999999999999111);
INSERT INTO tracks VALUES(199,'Leave My Girl Alone',20,1,6,'Buddy Guy',204721,6859518,0.98999999999999999111);
INSERT INTO tracks VALUES(200,'She Suits Me To A Tee',20,1,6,'Buddy Guy',136803,4456321,0.98999999999999999111);
INSERT INTO tracks VALUES(201,'Keep It To Myself (Aka Keep It To Yourself)',20,1,6,'Sonny Boy Williamson [I]',166060,5487056,0.98999999999999999111);
INSERT INTO tracks VALUES(202,'My Time After Awhile',20,1,6,'Robert Geddins/Ron Badger/Sheldon Feinberg',182491,6022698,0.98999999999999999111);
INSERT INTO tracks VALUES(203,'Too Many Ways (Alternate)',20,1,6,'Willie Dixon',135053,4459946,0.98999999999999999111);
INSERT INTO tracks VALUES(204,'Talkin'' ''Bout Women Obviously',20,1,6,'Amos Blakemore/Buddy Guy',589531,19161377,0.98999999999999999111);
INSERT INTO tracks VALUES(205,'Jorge Da Capadócia',21,1,7,'Jorge Ben',177397,5842196,0.98999999999999999111);
INSERT INTO tracks VALUES(206,'Prenda Minha',21,1,7,'Tradicional',99369,3225364,0.98999999999999999111);
INSERT INTO tracks VALUES(207,'Meditação',21,1,7,'Tom Jobim - Newton Mendoça',148793,4865597,0.98999999999999999111);
INSERT INTO tracks VALUES(208,'Terra',21,1,7,'Caetano Veloso',482429,15889054,0.98999999999999999111);
INSERT INTO tracks VALUES(209,'Eclipse Oculto',21,1,7,'Caetano Veloso',221936,7382703,0.98999999999999999111);
INSERT INTO tracks VALUES(210,'Texto "Verdade Tropical"',21,1,7,'Caetano Veloso',84088,2752161,0.98999999999999999111);
INSERT INTO tracks VALUES(211,'Bem Devagar',21,1,7,'Gilberto Gil',133172,4333651,0.98999999999999999111);
INSERT INTO tracks VALUES(212,'Drão',21,1,7,'Gilberto Gil',156264,5065932,0.98999999999999999111);
INSERT INTO tracks VALUES(213,'Saudosismo',21,1,7,'Caetano Veloso',144326,4726981,0.98999999999999999111);
INSERT INTO tracks VALUES(214,'Carolina',21,1,7,'Chico Buarque',181812,5924159,0.98999999999999999111);
INSERT INTO tracks VALUES(215,'Sozinho',21,1,7,'Peninha',190589,6253200,0.98999999999999999111);
INSERT INTO tracks VALUES(216,'Esse Cara',21,1,7,'Caetano Veloso',223111,7217126,0.98999999999999999111);
INSERT INTO tracks VALUES(217,'Mel',21,1,7,'Caetano Veloso - Waly Salomão',294765,9854062,0.98999999999999999111);
INSERT INTO tracks VALUES(218,'Linha Do Equador',21,1,7,'Caetano Veloso - Djavan',299337,10003747,0.98999999999999999111);
INSERT INTO tracks VALUES(219,'Odara',21,1,7,'Caetano Veloso',141270,4704104,0.98999999999999999111);
INSERT INTO tracks VALUES(220,'A Luz De Tieta',21,1,7,'Caetano Veloso',251742,8507446,0.98999999999999999111);
INSERT INTO tracks VALUES(221,'Atrás Da Verd-E-Rosa Só Não Vai Quem Já Morreu',21,1,7,'David Corrêa - Paulinho Carvalho - Carlos Sena - Bira do Ponto',307252,10364247,0.98999999999999999111);
INSERT INTO tracks VALUES(222,'Vida Boa',21,1,7,'Fausto Nilo - Armandinho',281730,9411272,0.98999999999999999111);
INSERT INTO tracks VALUES(223,'Sozinho (Hitmakers Classic Mix)',22,1,7,NULL,436636,14462072,0.98999999999999999111);
INSERT INTO tracks VALUES(224,'Sozinho (Hitmakers Classic Radio Edit)',22,1,7,NULL,195004,6455134,0.98999999999999999111);
INSERT INTO tracks VALUES(225,'Sozinho (Caêdrum ''n'' Bass)',22,1,7,NULL,328071,10975007,0.98999999999999999111);
INSERT INTO tracks VALUES(226,'Carolina',23,1,7,NULL,163056,5375395,0.98999999999999999111);
INSERT INTO tracks VALUES(227,'Essa Moça Ta Diferente',23,1,7,NULL,167235,5568574,0.98999999999999999111);
INSERT INTO tracks VALUES(228,'Vai Passar',23,1,7,NULL,369763,12359161,0.98999999999999999111);
INSERT INTO tracks VALUES(229,'Samba De Orly',23,1,7,NULL,162429,5431854,0.98999999999999999111);
INSERT INTO tracks VALUES(230,'Bye, Bye Brasil',23,1,7,NULL,283402,9499590,0.98999999999999999111);
INSERT INTO tracks VALUES(231,'Atras Da Porta',23,1,7,NULL,189675,6132843,0.98999999999999999111);
INSERT INTO tracks VALUES(232,'Tatuagem',23,1,7,NULL,172120,5645703,0.98999999999999999111);
INSERT INTO tracks VALUES(233,'O Que Será (À Flor Da Terra)',23,1,7,NULL,167288,5574848,0.98999999999999999111);
INSERT INTO tracks VALUES(234,'Morena De Angola',23,1,7,NULL,186801,6373932,0.98999999999999999111);
INSERT INTO tracks VALUES(235,'Apesar De Você',23,1,7,NULL,234501,7886937,0.98999999999999999111);
INSERT INTO tracks VALUES(236,'A Banda',23,1,7,NULL,132493,4349539,0.98999999999999999111);
INSERT INTO tracks VALUES(237,'Minha Historia',23,1,7,NULL,182256,6029673,0.98999999999999999111);
INSERT INTO tracks VALUES(238,'Com Açúcar E Com Afeto',23,1,7,NULL,175386,5846442,0.98999999999999999111);
INSERT INTO tracks VALUES(239,'Brejo Da Cruz',23,1,7,NULL,214099,7270749,0.98999999999999999111);
INSERT INTO tracks VALUES(240,'Meu Caro Amigo',23,1,7,NULL,260257,8778172,0.98999999999999999111);
INSERT INTO tracks VALUES(241,'Geni E O Zepelim',23,1,7,NULL,317570,10342226,0.98999999999999999111);
INSERT INTO tracks VALUES(242,'Trocando Em Miúdos',23,1,7,NULL,169717,5461468,0.98999999999999999111);
INSERT INTO tracks VALUES(243,'Vai Trabalhar Vagabundo',23,1,7,NULL,139154,4693941,0.98999999999999999111);
INSERT INTO tracks VALUES(244,'Gota D''água',23,1,7,NULL,153208,5074189,0.98999999999999999111);
INSERT INTO tracks VALUES(245,'Construção / Deus Lhe Pague',23,1,7,NULL,383059,12675305,0.98999999999999999111);
INSERT INTO tracks VALUES(246,'Mateus Enter',24,1,7,'Chico Science',33149,1103013,0.98999999999999999111);
INSERT INTO tracks VALUES(247,'O Cidadão Do Mundo',24,1,7,'Chico Science',200933,6724966,0.98999999999999999111);
INSERT INTO tracks VALUES(248,'Etnia',24,1,7,'Chico Science',152555,5061413,0.98999999999999999111);
INSERT INTO tracks VALUES(249,'Quilombo Groove [Instrumental]',24,1,7,'Chico Science',151823,5042447,0.98999999999999999111);
INSERT INTO tracks VALUES(250,'Macô',24,1,7,'Chico Science',249600,8253934,0.98999999999999999111);
INSERT INTO tracks VALUES(251,'Um Passeio No Mundo Livre',24,1,7,'Chico Science',240091,7984291,0.98999999999999999111);
INSERT INTO tracks VALUES(252,'Samba Do Lado',24,1,7,'Chico Science',227317,7541688,0.98999999999999999111);
INSERT INTO tracks VALUES(253,'Maracatu Atômico',24,1,7,'Chico Science',284264,9670057,0.98999999999999999111);
INSERT INTO tracks VALUES(254,'O Encontro De Isaac Asimov Com Santos Dumont No Céu',24,1,7,'Chico Science',99108,3240816,0.98999999999999999111);
INSERT INTO tracks VALUES(255,'Corpo De Lama',24,1,7,'Chico Science',232672,7714954,0.98999999999999999111);
INSERT INTO tracks VALUES(256,'Sobremesa',24,1,7,'Chico Science',240091,7960868,0.98999999999999999111);
INSERT INTO tracks VALUES(257,'Manguetown',24,1,7,'Chico Science',194560,6475159,0.98999999999999999111);
INSERT INTO tracks VALUES(258,'Um Satélite Na Cabeça',24,1,7,'Chico Science',126615,4272821,0.98999999999999999111);
INSERT INTO tracks VALUES(259,'Baião Ambiental [Instrumental]',24,1,7,'Chico Science',152659,5198539,0.98999999999999999111);
INSERT INTO tracks VALUES(260,'Sangue De Bairro',24,1,7,'Chico Science',132231,4415557,0.98999999999999999111);
INSERT INTO tracks VALUES(261,'Enquanto O Mundo Explode',24,1,7,'Chico Science',88764,2968650,0.98999999999999999111);
INSERT INTO tracks VALUES(262,'Interlude Zumbi',24,1,7,'Chico Science',71627,2408550,0.98999999999999999111);
INSERT INTO tracks VALUES(263,'Criança De Domingo',24,1,7,'Chico Science',208222,6984813,0.98999999999999999111);
INSERT INTO tracks VALUES(264,'Amor De Muito',24,1,7,'Chico Science',175333,5881293,0.98999999999999999111);
INSERT INTO tracks VALUES(265,'Samidarish [Instrumental]',24,1,7,'Chico Science',272431,8911641,0.98999999999999999111);
INSERT INTO tracks VALUES(266,'Maracatu Atômico [Atomic Version]',24,1,7,'Chico Science',273084,9019677,0.98999999999999999111);
INSERT INTO tracks VALUES(267,'Maracatu Atômico [Ragga Mix]',24,1,7,'Chico Science',210155,6986421,0.98999999999999999111);
INSERT INTO tracks VALUES(268,'Maracatu Atômico [Trip Hop]',24,1,7,'Chico Science',221492,7380787,0.98999999999999999111);
INSERT INTO tracks VALUES(269,'Banditismo Por Uma Questa',25,1,7,NULL,307095,10251097,0.98999999999999999111);
INSERT INTO tracks VALUES(270,'Banditismo Por Uma Questa',25,1,7,NULL,243644,8147224,0.98999999999999999111);
INSERT INTO tracks VALUES(271,'Rios Pontes & Overdrives',25,1,7,NULL,286720,9659152,0.98999999999999999111);
INSERT INTO tracks VALUES(272,'Cidade',25,1,7,NULL,216346,7241817,0.98999999999999999111);
INSERT INTO tracks VALUES(273,'Praiera',25,1,7,NULL,183640,6172781,0.98999999999999999111);
INSERT INTO tracks VALUES(274,'Samba Makossa',25,1,7,NULL,271856,9095410,0.98999999999999999111);
INSERT INTO tracks VALUES(275,'Da Lama Ao Caos',25,1,7,NULL,251559,8378065,0.98999999999999999111);
INSERT INTO tracks VALUES(276,'Maracatu De Tiro Certeiro',25,1,7,NULL,88868,2901397,0.98999999999999999111);
INSERT INTO tracks VALUES(277,'Salustiano Song',25,1,7,NULL,215405,7183969,0.98999999999999999111);
INSERT INTO tracks VALUES(278,'Antene Se',25,1,7,NULL,248372,8253618,0.98999999999999999111);
INSERT INTO tracks VALUES(279,'Risoflora',25,1,7,NULL,105586,3536938,0.98999999999999999111);
INSERT INTO tracks VALUES(280,'Lixo Do Mangue',25,1,7,NULL,193253,6534200,0.98999999999999999111);
INSERT INTO tracks VALUES(281,'Computadores Fazem Arte',25,1,7,NULL,404323,13702771,0.98999999999999999111);
INSERT INTO tracks VALUES(282,'Girassol',26,1,8,'Bino Farias/Da Gama/Lazão/Pedro Luis/Toni Garrido',249808,8327676,0.98999999999999999111);
INSERT INTO tracks VALUES(283,'A Sombra Da Maldade',26,1,8,'Da Gama/Toni Garrido',230922,7697230,0.98999999999999999111);
INSERT INTO tracks VALUES(284,'Johnny B. Goode',26,1,8,'Chuck Berry',254615,8505985,0.98999999999999999111);
INSERT INTO tracks VALUES(285,'Soldado Da Paz',26,1,8,'Herbert Vianna',194220,6455080,0.98999999999999999111);
INSERT INTO tracks VALUES(286,'Firmamento',26,1,8,'Bino Farias/Da Gama/Henry Lawes/Lazão/Toni Garrido/Winston Foser-Vers',222145,7402658,0.98999999999999999111);
INSERT INTO tracks VALUES(287,'Extra',26,1,8,'Gilberto Gil',304352,10078050,0.98999999999999999111);
INSERT INTO tracks VALUES(288,'O Erê',26,1,8,'Bernardo Vilhena/Bino Farias/Da Gama/Lazão/Toni Garrido',236382,7866924,0.98999999999999999111);
INSERT INTO tracks VALUES(289,'Podes Crer',26,1,8,'Bino Farias/Da Gama/Lazão/Toni Garrido',232280,7747747,0.98999999999999999111);
INSERT INTO tracks VALUES(290,'A Estrada',26,1,8,'Bino Farias/Da Gama/Lazão/Toni Garrido',248842,8275673,0.98999999999999999111);
INSERT INTO tracks VALUES(291,'Berlim',26,1,8,'Da Gama/Toni Garrido',207542,6920424,0.98999999999999999111);
INSERT INTO tracks VALUES(292,'Já Foi',26,1,8,'Bino Farias/Da Gama/Lazão/Toni Garrido',221544,7388466,0.98999999999999999111);
INSERT INTO tracks VALUES(293,'Onde Você Mora?',26,1,8,'Marisa Monte/Nando Reis',256026,8502588,0.98999999999999999111);
INSERT INTO tracks VALUES(294,'Pensamento',26,1,8,'Bino Farias/Da Gamma/Lazão/Rás Bernard',173008,5748424,0.98999999999999999111);
INSERT INTO tracks VALUES(295,'Conciliação',26,1,8,'Da Gama/Lazão/Rás Bernardo',257619,8552474,0.98999999999999999111);
INSERT INTO tracks VALUES(296,'Realidade Virtual',26,1,8,'Bino Farias/Da Gama/Lazão/Toni Garrido',195239,6503533,0.98999999999999999111);
INSERT INTO tracks VALUES(297,'Mensagem',26,1,8,'Bino Farias/Da Gama/Lazão/Rás Bernardo',225332,7488852,0.98999999999999999111);
INSERT INTO tracks VALUES(298,'A Cor Do Sol',26,1,8,'Bernardo Vilhena/Da Gama/Lazão',231392,7663348,0.98999999999999999111);
INSERT INTO tracks VALUES(299,'Onde Você Mora?',27,1,8,'Marisa Monte/Nando Reis',298396,10056970,0.98999999999999999111);
INSERT INTO tracks VALUES(300,'O Erê',27,1,8,'Bernardo Vilhena/Bino/Da Gama/Lazao/Toni Garrido',206942,6950332,0.98999999999999999111);
INSERT INTO tracks VALUES(301,'A Sombra Da Maldade',27,1,8,'Da Gama/Toni Garrido',285231,9544383,0.98999999999999999111);
INSERT INTO tracks VALUES(302,'A Estrada',27,1,8,'Da Gama/Lazao/Toni Garrido',282174,9344477,0.98999999999999999111);
INSERT INTO tracks VALUES(303,'Falar A Verdade',27,1,8,'Bino/Da Gama/Ras Bernardo',244950,8189093,0.98999999999999999111);
INSERT INTO tracks VALUES(304,'Firmamento',27,1,8,'Harry Lawes/Winston Foster-Vers',225488,7507866,0.98999999999999999111);
INSERT INTO tracks VALUES(305,'Pensamento',27,1,8,'Bino/Da Gama/Ras Bernardo',192391,6399761,0.98999999999999999111);
INSERT INTO tracks VALUES(306,'Realidade Virtual',27,1,8,'Bino/Da Gamma/Lazao/Toni Garrido',240300,8069934,0.98999999999999999111);
INSERT INTO tracks VALUES(307,'Doutor',27,1,8,'Bino/Da Gama/Toni Garrido',178155,5950952,0.98999999999999999111);
INSERT INTO tracks VALUES(308,'Na Frente Da TV',27,1,8,'Bino/Da Gama/Lazao/Ras Bernardo',289750,9633659,0.98999999999999999111);
INSERT INTO tracks VALUES(309,'Downtown',27,1,8,'Cidade Negra',239725,8024386,0.98999999999999999111);
INSERT INTO tracks VALUES(310,'Sábado A Noite',27,1,8,'Lulu Santos',267363,8895073,0.98999999999999999111);
INSERT INTO tracks VALUES(311,'A Cor Do Sol',27,1,8,'Bernardo Vilhena/Da Gama/Lazao',273031,9142937,0.98999999999999999111);
INSERT INTO tracks VALUES(312,'Eu Também Quero Beijar',27,1,8,'Fausto Nilo/Moraes Moreira/Pepeu Gomes',211147,7029400,0.98999999999999999111);
INSERT INTO tracks VALUES(313,'Noite Do Prazer',28,1,7,NULL,311353,10309980,0.98999999999999999111);
INSERT INTO tracks VALUES(314,'À Francesa',28,1,7,NULL,244532,8150846,0.98999999999999999111);
INSERT INTO tracks VALUES(315,'Cada Um Cada Um (A Namoradeira)',28,1,7,NULL,253492,8441034,0.98999999999999999111);
INSERT INTO tracks VALUES(316,'Linha Do Equador',28,1,7,NULL,244715,8123466,0.98999999999999999111);
INSERT INTO tracks VALUES(317,'Amor Demais',28,1,7,NULL,254040,8420093,0.98999999999999999111);
INSERT INTO tracks VALUES(318,'Férias',28,1,7,NULL,264202,8731945,0.98999999999999999111);
INSERT INTO tracks VALUES(319,'Gostava Tanto De Você',28,1,7,NULL,230452,7685326,0.98999999999999999111);
INSERT INTO tracks VALUES(320,'Flor Do Futuro',28,1,7,NULL,275748,9205941,0.98999999999999999111);
INSERT INTO tracks VALUES(321,'Felicidade Urgente',28,1,7,NULL,266605,8873358,0.98999999999999999111);
INSERT INTO tracks VALUES(322,'Livre Pra Viver',28,1,7,NULL,214595,7111596,0.98999999999999999111);
INSERT INTO tracks VALUES(323,'Dig-Dig, Lambe-Lambe (Ao Vivo)',29,1,9,'Cassiano Costa/Cintia Maviane/J.F./Lucas Costa',205479,6892516,0.98999999999999999111);
INSERT INTO tracks VALUES(324,'Pererê',29,1,9,'Augusto Conceição/Chiclete Com Banana',198661,6643207,0.98999999999999999111);
INSERT INTO tracks VALUES(325,'TriboTchan',29,1,9,'Cal Adan/Paulo Levi',194194,6507950,0.98999999999999999111);
INSERT INTO tracks VALUES(326,'Tapa Aqui, Descobre Ali',29,1,9,'Paulo Levi/W. Rangel',188630,6327391,0.98999999999999999111);
INSERT INTO tracks VALUES(327,'Daniela',29,1,9,'Jorge Cardoso/Pierre Onasis',230791,7748006,0.98999999999999999111);
INSERT INTO tracks VALUES(328,'Bate Lata',29,1,9,'Fábio Nolasco/Gal Sales/Ivan Brasil',206733,7034985,0.98999999999999999111);
INSERT INTO tracks VALUES(329,'Garotas do Brasil',29,1,9,'Garay, Ricardo Engels/Luca Predabom/Ludwig, Carlos Henrique/Maurício Vieira',210155,6973625,0.98999999999999999111);
INSERT INTO tracks VALUES(330,'Levada do Amor (Ailoviu)',29,1,9,'Luiz Wanderley/Paulo Levi',190093,6457752,0.98999999999999999111);
INSERT INTO tracks VALUES(331,'Lavadeira',29,1,9,'Do Vale, Valverde/Gal Oliveira/Luciano Pinto',214256,7254147,0.98999999999999999111);
INSERT INTO tracks VALUES(332,'Reboladeira',29,1,9,'Cal Adan/Ferrugem/Julinho Carioca/Tríona Ní Dhomhnaill',210599,7027525,0.98999999999999999111);
INSERT INTO tracks VALUES(333,'É que Nessa Encarnação Eu Nasci Manga',29,1,9,'Lucina/Luli',196519,6568081,0.98999999999999999111);
INSERT INTO tracks VALUES(334,'Reggae Tchan',29,1,9,'Cal Adan/Del Rey, Tension/Edu Casanova',206654,6931328,0.98999999999999999111);
INSERT INTO tracks VALUES(335,'My Love',29,1,9,'Jauperi/Zeu Góes',203493,6772813,0.98999999999999999111);
INSERT INTO tracks VALUES(336,'Latinha de Cerveja',29,1,9,'Adriano Bernandes/Edmar Neves',166687,5532564,0.98999999999999999111);
INSERT INTO tracks VALUES(337,'You Shook Me',30,1,1,'J B Lenoir/Willie Dixon',315951,10249958,0.98999999999999999111);
INSERT INTO tracks VALUES(338,'I Can''t Quit You Baby',30,1,1,'Willie Dixon',263836,8581414,0.98999999999999999111);
INSERT INTO tracks VALUES(339,'Communication Breakdown',30,1,1,'Jimmy Page/John Bonham/John Paul Jones',192653,6287257,0.98999999999999999111);
INSERT INTO tracks VALUES(340,'Dazed and Confused',30,1,1,'Jimmy Page',401920,13035765,0.98999999999999999111);
INSERT INTO tracks VALUES(341,'The Girl I Love She Got Long Black Wavy Hair',30,1,1,'Jimmy Page/John Bonham/John Estes/John Paul Jones/Robert Plant',183327,5995686,0.98999999999999999111);
INSERT INTO tracks VALUES(342,'What is and Should Never Be',30,1,1,'Jimmy Page/Robert Plant',260675,8497116,0.98999999999999999111);
INSERT INTO tracks VALUES(343,'Communication Breakdown(2)',30,1,1,'Jimmy Page/John Bonham/John Paul Jones',161149,5261022,0.98999999999999999111);
INSERT INTO tracks VALUES(344,'Travelling Riverside Blues',30,1,1,'Jimmy Page/Robert Johnson/Robert Plant',312032,10232581,0.98999999999999999111);
INSERT INTO tracks VALUES(345,'Whole Lotta Love',30,1,1,'Jimmy Page/John Bonham/John Paul Jones/Robert Plant/Willie Dixon',373394,12258175,0.98999999999999999111);
INSERT INTO tracks VALUES(346,'Somethin'' Else',30,1,1,'Bob Cochran/Sharon Sheeley',127869,4165650,0.98999999999999999111);
INSERT INTO tracks VALUES(347,'Communication Breakdown(3)',30,1,1,'Jimmy Page/John Bonham/John Paul Jones',185260,6041133,0.98999999999999999111);
INSERT INTO tracks VALUES(348,'I Can''t Quit You Baby(2)',30,1,1,'Willie Dixon',380551,12377615,0.98999999999999999111);
INSERT INTO tracks VALUES(349,'You Shook Me(2)',30,1,1,'J B Lenoir/Willie Dixon',619467,20138673,0.98999999999999999111);
INSERT INTO tracks VALUES(350,'How Many More Times',30,1,1,'Chester Burnett/Jimmy Page/John Bonham/John Paul Jones/Robert Plant',711836,23092953,0.98999999999999999111);
INSERT INTO tracks VALUES(351,'Debra Kadabra',31,1,1,'Frank Zappa',234553,7649679,0.98999999999999999111);
INSERT INTO tracks VALUES(352,'Carolina Hard-Core Ecstasy',31,1,1,'Frank Zappa',359680,11731061,0.98999999999999999111);
INSERT INTO tracks VALUES(353,'Sam With The Showing Scalp Flat Top',31,1,1,'Don Van Vliet',171284,5572993,0.98999999999999999111);
INSERT INTO tracks VALUES(354,'Poofter''s Froth Wyoming Plans Ahead',31,1,1,'Frank Zappa',183902,6007019,0.98999999999999999111);
INSERT INTO tracks VALUES(355,'200 Years Old',31,1,1,'Frank Zappa',272561,8912465,0.98999999999999999111);
INSERT INTO tracks VALUES(356,'Cucamonga',31,1,1,'Frank Zappa',144483,4728586,0.98999999999999999111);
INSERT INTO tracks VALUES(357,'Advance Romance',31,1,1,'Frank Zappa',677694,22080051,0.98999999999999999111);
INSERT INTO tracks VALUES(358,'Man With The Woman Head',31,1,1,'Don Van Vliet',88894,2922044,0.98999999999999999111);
INSERT INTO tracks VALUES(359,'Muffin Man',31,1,1,'Frank Zappa',332878,10891682,0.98999999999999999111);
INSERT INTO tracks VALUES(360,'Vai-Vai 2001',32,1,10,NULL,276349,9402241,0.98999999999999999111);
INSERT INTO tracks VALUES(361,'X-9 2001',32,1,10,NULL,273920,9310370,0.98999999999999999111);
INSERT INTO tracks VALUES(362,'Gavioes 2001',32,1,10,NULL,282723,9616640,0.98999999999999999111);
INSERT INTO tracks VALUES(363,'Nene 2001',32,1,10,NULL,284969,9694508,0.98999999999999999111);
INSERT INTO tracks VALUES(364,'Rosas De Ouro 2001',32,1,10,NULL,284342,9721084,0.98999999999999999111);
INSERT INTO tracks VALUES(365,'Mocidade Alegre 2001',32,1,10,NULL,282488,9599937,0.98999999999999999111);
INSERT INTO tracks VALUES(366,'Camisa Verde 2001',32,1,10,NULL,283454,9633755,0.98999999999999999111);
INSERT INTO tracks VALUES(367,'Leandro De Itaquera 2001',32,1,10,NULL,274808,9451845,0.98999999999999999111);
INSERT INTO tracks VALUES(368,'Tucuruvi 2001',32,1,10,NULL,287921,9883335,0.98999999999999999111);
INSERT INTO tracks VALUES(369,'Aguia De Ouro 2001',32,1,10,NULL,284160,9698729,0.98999999999999999111);
INSERT INTO tracks VALUES(370,'Ipiranga 2001',32,1,10,NULL,248293,8522591,0.98999999999999999111);
INSERT INTO tracks VALUES(371,'Morro Da Casa Verde 2001',32,1,10,NULL,284708,9718778,0.98999999999999999111);
INSERT INTO tracks VALUES(372,'Perola Negra 2001',32,1,10,NULL,281626,9619196,0.98999999999999999111);
INSERT INTO tracks VALUES(373,'Sao Lucas 2001',32,1,10,NULL,296254,10020122,0.98999999999999999111);
INSERT INTO tracks VALUES(374,'Guanabara',33,1,7,'Marcos Valle',247614,8499591,0.98999999999999999111);
INSERT INTO tracks VALUES(375,'Mas Que Nada',33,1,7,'Jorge Ben',248398,8255254,0.98999999999999999111);
INSERT INTO tracks VALUES(376,'Vôo Sobre o Horizonte',33,1,7,'J.r.Bertami/Parana',225097,7528825,0.98999999999999999111);
INSERT INTO tracks VALUES(377,'A Paz',33,1,7,'Donato/Gilberto Gil',263183,8619173,0.98999999999999999111);
INSERT INTO tracks VALUES(378,'Wave (Vou te Contar)',33,1,7,'Antonio Carlos Jobim',271647,9057557,0.98999999999999999111);
INSERT INTO tracks VALUES(379,'Água de Beber',33,1,7,'Antonio Carlos Jobim/Vinicius de Moraes',146677,4866476,0.98999999999999999111);
INSERT INTO tracks VALUES(380,'Samba da Bençaco',33,1,7,'Baden Powell/Vinicius de Moraes',282200,9440676,0.98999999999999999111);
INSERT INTO tracks VALUES(381,'Pode Parar',33,1,7,'Jorge Vercilo/Jota Maranhao',179408,6046678,0.98999999999999999111);
INSERT INTO tracks VALUES(382,'Menino do Rio',33,1,7,'Caetano Veloso',262713,8737489,0.98999999999999999111);
INSERT INTO tracks VALUES(383,'Ando Meio Desligado',33,1,7,'Caetano Veloso',195813,6547648,0.98999999999999999111);
INSERT INTO tracks VALUES(384,'Mistério da Raça',33,1,7,'Luiz Melodia/Ricardo Augusto',184320,6191752,0.98999999999999999111);
INSERT INTO tracks VALUES(385,'All Star',33,1,7,'Nando Reis',176326,5891697,0.98999999999999999111);
INSERT INTO tracks VALUES(386,'Menina Bonita',33,1,7,'Alexandre Brazil/Pedro Luis/Rodrigo Cabelo',237087,7938246,0.98999999999999999111);
INSERT INTO tracks VALUES(387,'Pescador de Ilusões',33,1,7,'Macelo Yuka/O Rappa',245524,8267067,0.98999999999999999111);
INSERT INTO tracks VALUES(388,'À Vontade (Live Mix)',33,1,7,'Bombom/Ed Motta',180636,5972430,0.98999999999999999111);
INSERT INTO tracks VALUES(389,'Maria Fumaça',33,1,7,'Luiz Carlos/Oberdan',141008,4743149,0.98999999999999999111);
INSERT INTO tracks VALUES(390,'Sambassim (dj patife remix)',33,1,7,'Alba Carvalho/Fernando Porto',213655,7243166,0.98999999999999999111);
INSERT INTO tracks VALUES(391,'Garota De Ipanema',34,1,7,'Vários',279536,9141343,0.98999999999999999111);
INSERT INTO tracks VALUES(392,'Tim Tim Por Tim Tim',34,1,7,'Vários',213237,7143328,0.98999999999999999111);
INSERT INTO tracks VALUES(393,'Tarde Em Itapoã',34,1,7,'Vários',313704,10344491,0.98999999999999999111);
INSERT INTO tracks VALUES(394,'Tanto Tempo',34,1,7,'Vários',170292,5572240,0.98999999999999999111);
INSERT INTO tracks VALUES(395,'Eu Vim Da Bahia - Live',34,1,7,'Vários',157988,5115428,0.98999999999999999111);
INSERT INTO tracks VALUES(396,'Alô Alô Marciano',34,1,7,'Vários',238106,8013065,0.98999999999999999111);
INSERT INTO tracks VALUES(397,'Linha Do Horizonte',34,1,7,'Vários',279484,9275929,0.98999999999999999111);
INSERT INTO tracks VALUES(398,'Only A Dream In Rio',34,1,7,'Vários',371356,12192989,0.98999999999999999111);
INSERT INTO tracks VALUES(399,'Abrir A Porta',34,1,7,'Vários',271960,8991141,0.98999999999999999111);
INSERT INTO tracks VALUES(400,'Alice',34,1,7,'Vários',165982,5594341,0.98999999999999999111);
INSERT INTO tracks VALUES(401,'Momentos Que Marcam',34,1,7,'Vários',280137,9313740,0.98999999999999999111);
INSERT INTO tracks VALUES(402,'Um Jantar Pra Dois',34,1,7,'Vários',237714,7819755,0.98999999999999999111);
INSERT INTO tracks VALUES(403,'Bumbo Da Mangueira',34,1,7,'Vários',270158,9073350,0.98999999999999999111);
INSERT INTO tracks VALUES(404,'Mr Funk Samba',34,1,7,'Vários',213890,7102545,0.98999999999999999111);
INSERT INTO tracks VALUES(405,'Santo Antonio',34,1,7,'Vários',162716,5492069,0.98999999999999999111);
INSERT INTO tracks VALUES(406,'Por Você',34,1,7,'Vários',205557,6792493,0.98999999999999999111);
INSERT INTO tracks VALUES(407,'Só Tinha De Ser Com Você',34,1,7,'Vários',389642,13085596,0.98999999999999999111);
INSERT INTO tracks VALUES(408,'Free Speech For The Dumb',35,1,3,'Molaney/Morris/Roberts/Wainwright',155428,5076048,0.98999999999999999111);
INSERT INTO tracks VALUES(409,'It''s Electric',35,1,3,'Harris/Tatler',213995,6978601,0.98999999999999999111);
INSERT INTO tracks VALUES(410,'Sabbra Cadabra',35,1,3,'Black Sabbath',380342,12418147,0.98999999999999999111);
INSERT INTO tracks VALUES(411,'Turn The Page',35,1,3,'Seger',366524,11946327,0.98999999999999999111);
INSERT INTO tracks VALUES(412,'Die Die My Darling',35,1,3,'Danzig',149315,4867667,0.98999999999999999111);
INSERT INTO tracks VALUES(413,'Loverman',35,1,3,'Cave',472764,15446975,0.98999999999999999111);
INSERT INTO tracks VALUES(414,'Mercyful Fate',35,1,3,'Diamond/Shermann',671712,21942829,0.98999999999999999111);
INSERT INTO tracks VALUES(415,'Astronomy',35,1,3,'A.Bouchard/J.Bouchard/S.Pearlman',397531,13065612,0.98999999999999999111);
INSERT INTO tracks VALUES(416,'Whiskey In The Jar',35,1,3,'Traditional',305005,9943129,0.98999999999999999111);
INSERT INTO tracks VALUES(417,'Tuesday''s Gone',35,1,3,'Collins/Van Zandt',545750,17900787,0.98999999999999999111);
INSERT INTO tracks VALUES(418,'The More I See',35,1,3,'Molaney/Morris/Roberts/Wainwright',287973,9378873,0.98999999999999999111);
INSERT INTO tracks VALUES(419,'A Kind Of Magic',36,1,1,'Roger Taylor',262608,8689618,0.98999999999999999111);
INSERT INTO tracks VALUES(420,'Under Pressure',36,1,1,'Queen & David Bowie',236617,7739042,0.98999999999999999111);
INSERT INTO tracks VALUES(421,'Radio GA GA',36,1,1,'Roger Taylor',343745,11358573,0.98999999999999999111);
INSERT INTO tracks VALUES(422,'I Want It All',36,1,1,'Queen',241684,7876564,0.98999999999999999111);
INSERT INTO tracks VALUES(423,'I Want To Break Free',36,1,1,'John Deacon',259108,8552861,0.98999999999999999111);
INSERT INTO tracks VALUES(424,'Innuendo',36,1,1,'Queen',387761,12664591,0.98999999999999999111);
INSERT INTO tracks VALUES(425,'It''s A Hard Life',36,1,1,'Freddie Mercury',249417,8112242,0.98999999999999999111);
INSERT INTO tracks VALUES(426,'Breakthru',36,1,1,'Queen',249234,8150479,0.98999999999999999111);
INSERT INTO tracks VALUES(427,'Who Wants To Live Forever',36,1,1,'Brian May',297691,9577577,0.98999999999999999111);
INSERT INTO tracks VALUES(428,'Headlong',36,1,1,'Queen',273057,8921404,0.98999999999999999111);
INSERT INTO tracks VALUES(429,'The Miracle',36,1,1,'Queen',294974,9671923,0.98999999999999999111);
INSERT INTO tracks VALUES(430,'I''m Going Slightly Mad',36,1,1,'Queen',248032,8192339,0.98999999999999999111);
INSERT INTO tracks VALUES(431,'The Invisible Man',36,1,1,'Queen',238994,7920353,0.98999999999999999111);
INSERT INTO tracks VALUES(432,'Hammer To Fall',36,1,1,'Brian May',220316,7255404,0.98999999999999999111);
INSERT INTO tracks VALUES(433,'Friends Will Be Friends',36,1,1,'Freddie Mercury & John Deacon',248920,8114582,0.98999999999999999111);
INSERT INTO tracks VALUES(434,'The Show Must Go On',36,1,1,'Queen',263784,8526760,0.98999999999999999111);
INSERT INTO tracks VALUES(435,'One Vision',36,1,1,'Queen',242599,7936928,0.98999999999999999111);
INSERT INTO tracks VALUES(436,'Detroit Rock City',37,1,1,'Paul Stanley, B. Ezrin',218880,7146372,0.98999999999999999111);
INSERT INTO tracks VALUES(437,'Black Diamond',37,1,1,'Paul Stanley',314148,10266007,0.98999999999999999111);
INSERT INTO tracks VALUES(438,'Hard Luck Woman',37,1,1,'Paul Stanley',216032,7109267,0.98999999999999999111);
INSERT INTO tracks VALUES(439,'Sure Know Something',37,1,1,'Paul Stanley, Vincent Poncia',242468,7939886,0.98999999999999999111);
INSERT INTO tracks VALUES(440,'Love Gun',37,1,1,'Paul Stanley',196257,6424915,0.98999999999999999111);
INSERT INTO tracks VALUES(441,'Deuce',37,1,1,'Gene Simmons',185077,6097210,0.98999999999999999111);
INSERT INTO tracks VALUES(442,'Goin'' Blind',37,1,1,'Gene Simmons, S. Coronel',216215,7045314,0.98999999999999999111);
INSERT INTO tracks VALUES(443,'Shock Me',37,1,1,'Ace Frehley',227291,7529336,0.98999999999999999111);
INSERT INTO tracks VALUES(444,'Do You Love Me',37,1,1,'Paul Stanley, B. Ezrin, K. Fowley',214987,6976194,0.98999999999999999111);
INSERT INTO tracks VALUES(445,'She',37,1,1,'Gene Simmons, S. Coronel',248346,8229734,0.98999999999999999111);
INSERT INTO tracks VALUES(446,'I Was Made For Loving You',37,1,1,'Paul Stanley, Vincent Poncia, Desmond Child',271360,9018078,0.98999999999999999111);
INSERT INTO tracks VALUES(447,'Shout It Out Loud',37,1,1,'Paul Stanley, Gene Simmons, B. Ezrin',219742,7194424,0.98999999999999999111);
INSERT INTO tracks VALUES(448,'God Of Thunder',37,1,1,'Paul Stanley',255791,8309077,0.98999999999999999111);
INSERT INTO tracks VALUES(449,'Calling Dr. Love',37,1,1,'Gene Simmons',225332,7395034,0.98999999999999999111);
INSERT INTO tracks VALUES(450,'Beth',37,1,1,'S. Penridge, Bob Ezrin, Peter Criss',166974,5360574,0.98999999999999999111);
INSERT INTO tracks VALUES(451,'Strutter',37,1,1,'Paul Stanley, Gene Simmons',192496,6317021,0.98999999999999999111);
INSERT INTO tracks VALUES(452,'Rock And Roll All Nite',37,1,1,'Paul Stanley, Gene Simmons',173609,5735902,0.98999999999999999111);
INSERT INTO tracks VALUES(453,'Cold Gin',37,1,1,'Ace Frehley',262243,8609783,0.98999999999999999111);
INSERT INTO tracks VALUES(454,'Plaster Caster',37,1,1,'Gene Simmons',207333,6801116,0.98999999999999999111);
INSERT INTO tracks VALUES(455,'God Gave Rock ''n'' Roll To You',37,1,1,'Paul Stanley, Gene Simmons, Rus Ballard, Bob Ezrin',320444,10441590,0.98999999999999999111);
INSERT INTO tracks VALUES(456,'Heart of the Night',38,1,2,NULL,273737,9098263,0.98999999999999999111);
INSERT INTO tracks VALUES(457,'De La Luz',38,1,2,NULL,315219,10518284,0.98999999999999999111);
INSERT INTO tracks VALUES(458,'Westwood Moon',38,1,2,NULL,295627,9765802,0.98999999999999999111);
INSERT INTO tracks VALUES(459,'Midnight',38,1,2,NULL,266866,8851060,0.98999999999999999111);
INSERT INTO tracks VALUES(460,'Playtime',38,1,2,NULL,273580,9070880,0.98999999999999999111);
INSERT INTO tracks VALUES(461,'Surrender',38,1,2,NULL,287634,9422926,0.98999999999999999111);
INSERT INTO tracks VALUES(462,'Valentino''s',38,1,2,NULL,296124,9848545,0.98999999999999999111);
INSERT INTO tracks VALUES(463,'Believe',38,1,2,NULL,310778,10317185,0.98999999999999999111);
INSERT INTO tracks VALUES(464,'As We Sleep',38,1,2,NULL,316865,10429398,0.98999999999999999111);
INSERT INTO tracks VALUES(465,'When Evening Falls',38,1,2,NULL,298135,9863942,0.98999999999999999111);
INSERT INTO tracks VALUES(466,'J Squared',38,1,2,NULL,288757,9480777,0.98999999999999999111);
INSERT INTO tracks VALUES(467,'Best Thing',38,1,2,NULL,274259,9069394,0.98999999999999999111);
INSERT INTO tracks VALUES(468,'Maria',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',167262,5484747,0.98999999999999999111);
INSERT INTO tracks VALUES(469,'Poprocks And Coke',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',158354,5243078,0.98999999999999999111);
INSERT INTO tracks VALUES(470,'Longview',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',234083,7714939,0.98999999999999999111);
INSERT INTO tracks VALUES(471,'Welcome To Paradise',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',224208,7406008,0.98999999999999999111);
INSERT INTO tracks VALUES(472,'Basket Case',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',181629,5951736,0.98999999999999999111);
INSERT INTO tracks VALUES(473,'When I Come Around',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',178364,5839426,0.98999999999999999111);
INSERT INTO tracks VALUES(474,'She',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',134164,4425128,0.98999999999999999111);
INSERT INTO tracks VALUES(475,'J.A.R. (Jason Andrew Relva)',39,1,4,'Mike Dirnt -Words Green Day -Music',170997,5645755,0.98999999999999999111);
INSERT INTO tracks VALUES(476,'Geek Stink Breath',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',135888,4408983,0.98999999999999999111);
INSERT INTO tracks VALUES(477,'Brain Stew',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',193149,6305550,0.98999999999999999111);
INSERT INTO tracks VALUES(478,'Jaded',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',90331,2950224,0.98999999999999999111);
INSERT INTO tracks VALUES(479,'Walking Contradiction',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',151170,4932366,0.98999999999999999111);
INSERT INTO tracks VALUES(480,'Stuck With Me',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',135523,4431357,0.98999999999999999111);
INSERT INTO tracks VALUES(481,'Hitchin'' A Ride',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',171546,5616891,0.98999999999999999111);
INSERT INTO tracks VALUES(482,'Good Riddance (Time Of Your Life)',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',153600,5075241,0.98999999999999999111);
INSERT INTO tracks VALUES(483,'Redundant',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',198164,6481753,0.98999999999999999111);
INSERT INTO tracks VALUES(484,'Nice Guys Finish Last',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',170187,5604618,0.98999999999999999111);
INSERT INTO tracks VALUES(485,'Minority',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',168803,5535061,0.98999999999999999111);
INSERT INTO tracks VALUES(486,'Warning',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',221910,7343176,0.98999999999999999111);
INSERT INTO tracks VALUES(487,'Waiting',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',192757,6316430,0.98999999999999999111);
INSERT INTO tracks VALUES(488,'Macy''s Day Parade',39,1,4,'Billie Joe Armstrong -Words Green Day -Music',213420,7075573,0.98999999999999999111);
INSERT INTO tracks VALUES(489,'Into The Light',40,1,1,'David Coverdale',76303,2452653,0.98999999999999999111);
INSERT INTO tracks VALUES(490,'River Song',40,1,1,'David Coverdale',439510,14359478,0.98999999999999999111);
INSERT INTO tracks VALUES(491,'She Give Me ...',40,1,1,'David Coverdale',252551,8385478,0.98999999999999999111);
INSERT INTO tracks VALUES(492,'Don''t You Cry',40,1,1,'David Coverdale',347036,11269612,0.98999999999999999111);
INSERT INTO tracks VALUES(493,'Love Is Blind',40,1,1,'David Coverdale/Earl Slick',344999,11409720,0.98999999999999999111);
INSERT INTO tracks VALUES(494,'Slave',40,1,1,'David Coverdale/Earl Slick',291892,9425200,0.98999999999999999111);
INSERT INTO tracks VALUES(495,'Cry For Love',40,1,1,'Bossi/David Coverdale/Earl Slick',293015,9567075,0.98999999999999999111);
INSERT INTO tracks VALUES(496,'Living On Love',40,1,1,'Bossi/David Coverdale/Earl Slick',391549,12785876,0.98999999999999999111);
INSERT INTO tracks VALUES(497,'Midnight Blue',40,1,1,'David Coverdale/Earl Slick',298631,9750990,0.98999999999999999111);
INSERT INTO tracks VALUES(498,'Too Many Tears',40,1,1,'Adrian Vanderberg/David Coverdale',359497,11810238,0.98999999999999999111);
INSERT INTO tracks VALUES(499,'Don''t Lie To Me',40,1,1,'David Coverdale/Earl Slick',283585,9288007,0.98999999999999999111);
INSERT INTO tracks VALUES(500,'Wherever You May Go',40,1,1,'David Coverdale',239699,7803074,0.98999999999999999111);
INSERT INTO tracks VALUES(501,'Grito De Alerta',41,1,7,'Gonzaga Jr.',202213,6539422,0.98999999999999999111);
INSERT INTO tracks VALUES(502,'Não Dá Mais Pra Segurar (Explode Coração)',41,1,7,NULL,219768,7083012,0.98999999999999999111);
INSERT INTO tracks VALUES(503,'Começaria Tudo Outra Vez',41,1,7,NULL,196545,6473395,0.98999999999999999111);
INSERT INTO tracks VALUES(504,'O Que É O Que É ?',41,1,7,NULL,259291,8650647,0.98999999999999999111);
INSERT INTO tracks VALUES(505,'Sangrando',41,1,7,'Gonzaga Jr/Gonzaguinha',169717,5494406,0.98999999999999999111);
INSERT INTO tracks VALUES(506,'Diga Lá, Coração',41,1,7,NULL,255921,8280636,0.98999999999999999111);
INSERT INTO tracks VALUES(507,'Lindo Lago Do Amor',41,1,7,'Gonzaga Jr.',249678,8353191,0.98999999999999999111);
INSERT INTO tracks VALUES(508,'Eu Apenas Queria Que Voçê Soubesse',41,1,7,NULL,155637,5130056,0.98999999999999999111);
INSERT INTO tracks VALUES(509,'Com A Perna No Mundo',41,1,7,'Gonzaga Jr.',227448,7747108,0.98999999999999999111);
INSERT INTO tracks VALUES(510,'E Vamos À Luta',41,1,7,NULL,222406,7585112,0.98999999999999999111);
INSERT INTO tracks VALUES(511,'Um Homem Também Chora (Guerreiro Menino)',41,1,7,NULL,207229,6854219,0.98999999999999999111);
INSERT INTO tracks VALUES(512,'Comportamento Geral',41,1,7,'Gonzaga Jr',181577,5997444,0.98999999999999999111);
INSERT INTO tracks VALUES(513,'Ponto De Interrogação',41,1,7,NULL,180950,5946265,0.98999999999999999111);
INSERT INTO tracks VALUES(514,'Espere Por Mim, Morena',41,1,7,'Gonzaguinha',207072,6796523,0.98999999999999999111);
INSERT INTO tracks VALUES(515,'Meia-Lua Inteira',23,1,7,NULL,222093,7466288,0.98999999999999999111);
INSERT INTO tracks VALUES(516,'Voce e Linda',23,1,7,NULL,242938,8050268,0.98999999999999999111);
INSERT INTO tracks VALUES(517,'Um Indio',23,1,7,NULL,195944,6453213,0.98999999999999999111);
INSERT INTO tracks VALUES(518,'Podres Poderes',23,1,7,NULL,259761,8622495,0.98999999999999999111);
INSERT INTO tracks VALUES(519,'Voce Nao Entende Nada - Cotidiano',23,1,7,NULL,421982,13885612,0.98999999999999999111);
INSERT INTO tracks VALUES(520,'O Estrangeiro',23,1,7,NULL,374700,12472890,0.98999999999999999111);
INSERT INTO tracks VALUES(521,'Menino Do Rio',23,1,7,NULL,147670,4862277,0.98999999999999999111);
INSERT INTO tracks VALUES(522,'Qualquer Coisa',23,1,7,NULL,193410,6372433,0.98999999999999999111);
INSERT INTO tracks VALUES(523,'Sampa',23,1,7,NULL,185051,6151831,0.98999999999999999111);
INSERT INTO tracks VALUES(524,'Queixa',23,1,7,NULL,299676,9953962,0.98999999999999999111);
INSERT INTO tracks VALUES(525,'O Leaozinho',23,1,7,NULL,184398,6098150,0.98999999999999999111);
INSERT INTO tracks VALUES(526,'Fora Da Ordem',23,1,7,NULL,354011,11746781,0.98999999999999999111);
INSERT INTO tracks VALUES(527,'Terra',23,1,7,NULL,401319,13224055,0.98999999999999999111);
INSERT INTO tracks VALUES(528,'Alegria, Alegria',23,1,7,NULL,169221,5497025,0.98999999999999999111);
INSERT INTO tracks VALUES(529,'Balada Do Louco',42,1,4,'Arnaldo Baptista - Rita Lee',241057,7852328,0.98999999999999999111);
INSERT INTO tracks VALUES(530,'Ando Meio Desligado',42,1,4,'Arnaldo Baptista - Rita Lee -  Sérgio Dias',287817,9484504,0.98999999999999999111);
INSERT INTO tracks VALUES(531,'Top Top',42,1,4,'Os Mutantes - Arnolpho Lima Filho',146938,4875374,0.98999999999999999111);
INSERT INTO tracks VALUES(532,'Baby',42,1,4,'Caetano Veloso',177188,5798202,0.98999999999999999111);
INSERT INTO tracks VALUES(533,'A E O Z',42,1,4,'Mutantes',518556,16873005,0.98999999999999999111);
INSERT INTO tracks VALUES(534,'Panis Et Circenses',42,1,4,'Caetano Veloso - Gilberto Gil',125152,4069688,0.98999999999999999111);
INSERT INTO tracks VALUES(535,'Chão De Estrelas',42,1,4,'Orestes Barbosa-Sílvio Caldas',284813,9433620,0.98999999999999999111);
INSERT INTO tracks VALUES(536,'Vida De Cachorro',42,1,4,'Rita Lee - Arnaldo Baptista - Sérgio Baptista',195186,6411149,0.98999999999999999111);
INSERT INTO tracks VALUES(537,'Bat Macumba',42,1,4,'Gilberto Gil - Caetano Veloso',187794,6295223,0.98999999999999999111);
INSERT INTO tracks VALUES(538,'Desculpe Babe',42,1,4,'Arnaldo Baptista - Rita Lee',170422,5637959,0.98999999999999999111);
INSERT INTO tracks VALUES(539,'Rita Lee',42,1,4,'Arnaldo Baptista/Rita Lee/Sérgio Dias',189257,6270503,0.98999999999999999111);
INSERT INTO tracks VALUES(540,'Posso Perder Minha Mulher, Minha Mãe, Desde Que Eu Tenha O Rock And Roll',42,1,4,'Arnaldo Baptista - Rita Lee - Arnolpho Lima Filho',222955,7346254,0.98999999999999999111);
INSERT INTO tracks VALUES(541,'Banho De Lua',42,1,4,'B. de Filippi - F. Migliaci - Versão: Fred Jorge',221831,7232123,0.98999999999999999111);
INSERT INTO tracks VALUES(542,'Meu Refrigerador Não Funciona',42,1,4,'Arnaldo Baptista - Rita Lee - Sérgio Dias',382981,12495906,0.98999999999999999111);
INSERT INTO tracks VALUES(543,'Burn',43,1,1,'Coverdale/Lord/Paice',453955,14775708,0.98999999999999999111);
INSERT INTO tracks VALUES(544,'Stormbringer',43,1,1,'Coverdale',277133,9050022,0.98999999999999999111);
INSERT INTO tracks VALUES(545,'Gypsy',43,1,1,'Coverdale/Hughes/Lord/Paice',339173,11046952,0.98999999999999999111);
INSERT INTO tracks VALUES(546,'Lady Double Dealer',43,1,1,'Coverdale',233586,7608759,0.98999999999999999111);
INSERT INTO tracks VALUES(547,'Mistreated',43,1,1,'Coverdale',758648,24596235,0.98999999999999999111);
INSERT INTO tracks VALUES(548,'Smoke On The Water',43,1,1,'Gillan/Glover/Lord/Paice',618031,20103125,0.98999999999999999111);
INSERT INTO tracks VALUES(549,'You Fool No One',43,1,1,'Coverdale/Lord/Paice',804101,26369966,0.98999999999999999111);
INSERT INTO tracks VALUES(550,'Custard Pie',44,1,1,'Jimmy Page/Robert Plant',253962,8348257,0.98999999999999999111);
INSERT INTO tracks VALUES(551,'The Rover',44,1,1,'Jimmy Page/Robert Plant',337084,11011286,0.98999999999999999111);
INSERT INTO tracks VALUES(552,'In My Time Of Dying',44,1,1,'John Bonham/John Paul Jones',666017,21676727,0.98999999999999999111);
INSERT INTO tracks VALUES(553,'Houses Of The Holy',44,1,1,'Jimmy Page/Robert Plant',242494,7972503,0.98999999999999999111);
INSERT INTO tracks VALUES(554,'Trampled Under Foot',44,1,1,'John Paul Jones',336692,11154468,0.98999999999999999111);
INSERT INTO tracks VALUES(555,'Kashmir',44,1,1,'John Bonham',508604,16686580,0.98999999999999999111);
INSERT INTO tracks VALUES(556,'Imperatriz',45,1,7,'Guga/Marquinho Lessa/Tuninho Professor',339173,11348710,0.98999999999999999111);
INSERT INTO tracks VALUES(557,'Beija-Flor',45,1,7,'Caruso/Cleber/Deo/Osmar',327000,10991159,0.98999999999999999111);
INSERT INTO tracks VALUES(558,'Viradouro',45,1,7,'Dadinho/Gilbreto Gomes/Gustavo/P.C. Portugal/R. Mocoto',344320,11484362,0.98999999999999999111);
INSERT INTO tracks VALUES(559,'Mocidade',45,1,7,'Domenil/J. Brito/Joaozinho/Rap, Marcelo Do',261720,8817757,0.98999999999999999111);
INSERT INTO tracks VALUES(560,'Unidos Da Tijuca',45,1,7,'Douglas/Neves, Vicente Das/Silva, Gilmar L./Toninho Gentil/Wantuir',338834,11440689,0.98999999999999999111);
INSERT INTO tracks VALUES(561,'Salgueiro',45,1,7,'Augusto/Craig Negoescu/Rocco Filho/Saara, Ze Carlos Da',305920,10294741,0.98999999999999999111);
INSERT INTO tracks VALUES(562,'Mangueira',45,1,7,'Bizuca/Clóvis Pê/Gilson Bernini/Marelo D''Aguia',298318,9999506,0.98999999999999999111);
INSERT INTO tracks VALUES(563,'União Da Ilha',45,1,7,'Dito/Djalma Falcao/Ilha, Almir Da/Márcio André',330945,11100945,0.98999999999999999111);
INSERT INTO tracks VALUES(564,'Grande Rio',45,1,7,'Carlos Santos/Ciro/Claudio Russo/Zé Luiz',307252,10251428,0.98999999999999999111);
INSERT INTO tracks VALUES(565,'Portela',45,1,7,'Flavio Bororo/Paulo Apparicio/Wagner Alves/Zeca Sereno',319608,10712216,0.98999999999999999111);
INSERT INTO tracks VALUES(566,'Caprichosos',45,1,7,'Gule/Jorge 101/Lequinho/Luiz Piao',351320,11870956,0.98999999999999999111);
INSERT INTO tracks VALUES(567,'Tradição',45,1,7,'Adalto Magalha/Lourenco',269165,9114880,0.98999999999999999111);
INSERT INTO tracks VALUES(568,'Império Serrano',45,1,7,'Arlindo Cruz/Carlos Sena/Elmo Caetano/Mauricao',334942,11161196,0.98999999999999999111);
INSERT INTO tracks VALUES(569,'Tuiuti',45,1,7,'Claudio Martins/David Lima/Kleber Rodrigues/Livre, Cesare Som',259657,8749492,0.98999999999999999111);
INSERT INTO tracks VALUES(570,'(Da Le) Yaleo',46,1,1,'Santana',353488,11769507,0.98999999999999999111);
INSERT INTO tracks VALUES(571,'Love Of My Life',46,1,1,'Carlos Santana & Dave Matthews',347820,11634337,0.98999999999999999111);
INSERT INTO tracks VALUES(572,'Put Your Lights On',46,1,1,'E. Shrody',285178,9394769,0.98999999999999999111);
INSERT INTO tracks VALUES(573,'Africa Bamba',46,1,1,'I. Toure, S. Tidiane Toure, Carlos Santana & K. Perazzo',282827,9492487,0.98999999999999999111);
INSERT INTO tracks VALUES(574,'Smooth',46,1,1,'M. Itaal Shur & Rob Thomas',298161,9867455,0.98999999999999999111);
INSERT INTO tracks VALUES(575,'Do You Like The Way',46,1,1,'L. Hill',354899,11741062,0.98999999999999999111);
INSERT INTO tracks VALUES(576,'Maria Maria',46,1,1,'W. Jean, J. Duplessis, Carlos Santana, K. Perazzo & R. Rekow',262635,8664601,0.98999999999999999111);
INSERT INTO tracks VALUES(577,'Migra',46,1,1,'R. Taha, Carlos Santana & T. Lindsay',329064,10963305,0.98999999999999999111);
INSERT INTO tracks VALUES(578,'Corazon Espinado',46,1,1,'F. Olivera',276114,9206802,0.98999999999999999111);
INSERT INTO tracks VALUES(579,'Wishing It Was',46,1,1,'Eale-Eye Cherry, M. Simpson, J. King & M. Nishita',292832,9771348,0.98999999999999999111);
INSERT INTO tracks VALUES(580,'El Farol',46,1,1,'Carlos Santana & KC Porter',291160,9599353,0.98999999999999999111);
INSERT INTO tracks VALUES(581,'Primavera',46,1,1,'KC Porter & JB Eckl',378618,12504234,0.98999999999999999111);
INSERT INTO tracks VALUES(582,'The Calling',46,1,1,'Carlos Santana & C. Thompson',747755,24703884,0.98999999999999999111);
INSERT INTO tracks VALUES(583,'Solução',47,1,7,NULL,247431,8100449,0.98999999999999999111);
INSERT INTO tracks VALUES(584,'Manuel',47,1,7,NULL,230269,7677671,0.98999999999999999111);
INSERT INTO tracks VALUES(585,'Entre E Ouça',47,1,7,NULL,286302,9391004,0.98999999999999999111);
INSERT INTO tracks VALUES(586,'Um Contrato Com Deus',47,1,7,NULL,202501,6636465,0.98999999999999999111);
INSERT INTO tracks VALUES(587,'Um Jantar Pra Dois',47,1,7,NULL,244009,8021589,0.98999999999999999111);
INSERT INTO tracks VALUES(588,'Vamos Dançar',47,1,7,NULL,226194,7617432,0.98999999999999999111);
INSERT INTO tracks VALUES(589,'Um Love',47,1,7,NULL,181603,6095524,0.98999999999999999111);
INSERT INTO tracks VALUES(590,'Seis Da Tarde',47,1,7,NULL,238445,7935898,0.98999999999999999111);
INSERT INTO tracks VALUES(591,'Baixo Rio',47,1,7,NULL,198008,6521676,0.98999999999999999111);
INSERT INTO tracks VALUES(592,'Sombras Do Meu Destino',47,1,7,NULL,280685,9161539,0.98999999999999999111);
INSERT INTO tracks VALUES(593,'Do You Have Other Loves?',47,1,7,NULL,295235,9604273,0.98999999999999999111);
INSERT INTO tracks VALUES(594,'Agora Que O Dia Acordou',47,1,7,NULL,323213,10572752,0.98999999999999999111);
INSERT INTO tracks VALUES(595,'Já!!!',47,1,7,NULL,217782,7103608,0.98999999999999999111);
INSERT INTO tracks VALUES(596,'A Rua',47,1,7,NULL,238027,7930264,0.98999999999999999111);
INSERT INTO tracks VALUES(597,'Now''s The Time',48,1,2,'Miles Davis',197459,6358868,0.98999999999999999111);
INSERT INTO tracks VALUES(598,'Jeru',48,1,2,'Miles Davis',193410,6222536,0.98999999999999999111);
INSERT INTO tracks VALUES(599,'Compulsion',48,1,2,'Miles Davis',345025,11254474,0.98999999999999999111);
INSERT INTO tracks VALUES(600,'Tempus Fugit',48,1,2,'Miles Davis',231784,7548434,0.98999999999999999111);
INSERT INTO tracks VALUES(601,'Walkin''',48,1,2,'Miles Davis',807392,26411634,0.98999999999999999111);
INSERT INTO tracks VALUES(602,'''Round Midnight',48,1,2,'Miles Davis',357459,11590284,0.98999999999999999111);
INSERT INTO tracks VALUES(603,'Bye Bye Blackbird',48,1,2,'Miles Davis',476003,15549224,0.98999999999999999111);
INSERT INTO tracks VALUES(604,'New Rhumba',48,1,2,'Miles Davis',277968,9018024,0.98999999999999999111);
INSERT INTO tracks VALUES(605,'Generique',48,1,2,'Miles Davis',168777,5437017,0.98999999999999999111);
INSERT INTO tracks VALUES(606,'Summertime',48,1,2,'Miles Davis',200437,6461370,0.98999999999999999111);
INSERT INTO tracks VALUES(607,'So What',48,1,2,'Miles Davis',564009,18360449,0.98999999999999999111);
INSERT INTO tracks VALUES(608,'The Pan Piper',48,1,2,'Miles Davis',233769,7593713,0.98999999999999999111);
INSERT INTO tracks VALUES(609,'Someday My Prince Will Come',48,1,2,'Miles Davis',544078,17890773,0.98999999999999999111);
INSERT INTO tracks VALUES(610,'My Funny Valentine (Live)',49,1,2,'Miles Davis',907520,29416781,0.98999999999999999111);
INSERT INTO tracks VALUES(611,'E.S.P.',49,1,2,'Miles Davis',330684,11079866,0.98999999999999999111);
INSERT INTO tracks VALUES(612,'Nefertiti',49,1,2,'Miles Davis',473495,15478450,0.98999999999999999111);
INSERT INTO tracks VALUES(613,'Petits Machins (Little Stuff)',49,1,2,'Miles Davis',487392,16131272,0.98999999999999999111);
INSERT INTO tracks VALUES(614,'Miles Runs The Voodoo Down',49,1,2,'Miles Davis',843964,27967919,0.98999999999999999111);
INSERT INTO tracks VALUES(615,'Little Church (Live)',49,1,2,'Miles Davis',196101,6273225,0.98999999999999999111);
INSERT INTO tracks VALUES(616,'Black Satin',49,1,2,'Miles Davis',316682,10529483,0.98999999999999999111);
INSERT INTO tracks VALUES(617,'Jean Pierre (Live)',49,1,2,'Miles Davis',243461,7955114,0.98999999999999999111);
INSERT INTO tracks VALUES(618,'Time After Time',49,1,2,'Miles Davis',220734,7292197,0.98999999999999999111);
INSERT INTO tracks VALUES(619,'Portia',49,1,2,'Miles Davis',378775,12520126,0.98999999999999999111);
INSERT INTO tracks VALUES(620,'Space Truckin''',50,1,1,'Blackmore/Gillan/Glover/Lord/Paice',1196094,39267613,0.98999999999999999111);
INSERT INTO tracks VALUES(621,'Going Down / Highway Star',50,1,1,'Gillan/Glover/Lord/Nix - Blackmore/Paice',913658,29846063,0.98999999999999999111);
INSERT INTO tracks VALUES(622,'Mistreated (Alternate Version)',50,1,1,'Blackmore/Coverdale',854700,27775442,0.98999999999999999111);
INSERT INTO tracks VALUES(623,'You Fool No One (Alternate Version)',50,1,1,'Blackmore/Coverdale/Lord/Paice',763924,24887209,0.98999999999999999111);
INSERT INTO tracks VALUES(624,'Jeepers Creepers',51,1,2,NULL,185965,5991903,0.98999999999999999111);
INSERT INTO tracks VALUES(625,'Blue Rythm Fantasy',51,1,2,NULL,348212,11204006,0.98999999999999999111);
INSERT INTO tracks VALUES(626,'Drum Boogie',51,1,2,NULL,191555,6185636,0.98999999999999999111);
INSERT INTO tracks VALUES(627,'Let Me Off Uptown',51,1,2,NULL,187637,6034685,0.98999999999999999111);
INSERT INTO tracks VALUES(628,'Leave Us Leap',51,1,2,NULL,182726,5898810,0.98999999999999999111);
INSERT INTO tracks VALUES(629,'Opus No.1',51,1,2,NULL,179800,5846041,0.98999999999999999111);
INSERT INTO tracks VALUES(630,'Boogie Blues',51,1,2,NULL,204199,6603153,0.98999999999999999111);
INSERT INTO tracks VALUES(631,'How High The Moon',51,1,2,NULL,201430,6529487,0.98999999999999999111);
INSERT INTO tracks VALUES(632,'Disc Jockey Jump',51,1,2,NULL,193149,6260820,0.98999999999999999111);
INSERT INTO tracks VALUES(633,'Up An'' Atom',51,1,2,NULL,179565,5822645,0.98999999999999999111);
INSERT INTO tracks VALUES(634,'Bop Boogie',51,1,2,NULL,189596,6093124,0.98999999999999999111);
INSERT INTO tracks VALUES(635,'Lemon Drop',51,1,2,NULL,194089,6287531,0.98999999999999999111);
INSERT INTO tracks VALUES(636,'Coronation Drop',51,1,2,NULL,176222,5899898,0.98999999999999999111);
INSERT INTO tracks VALUES(637,'Overtime',51,1,2,NULL,163030,5432236,0.98999999999999999111);
INSERT INTO tracks VALUES(638,'Imagination',51,1,2,NULL,289306,9444385,0.98999999999999999111);
INSERT INTO tracks VALUES(639,'Don''t Take Your Love From Me',51,1,2,NULL,282331,9244238,0.98999999999999999111);
INSERT INTO tracks VALUES(640,'Midget',51,1,2,NULL,217025,7257663,0.98999999999999999111);
INSERT INTO tracks VALUES(641,'I''m Coming Virginia',51,1,2,NULL,280163,9209827,0.98999999999999999111);
INSERT INTO tracks VALUES(642,'Payin'' Them Dues Blues',51,1,2,NULL,198556,6536918,0.98999999999999999111);
INSERT INTO tracks VALUES(643,'Jungle Drums',51,1,2,NULL,199627,6546063,0.98999999999999999111);
INSERT INTO tracks VALUES(644,'Showcase',51,1,2,NULL,201560,6697510,0.98999999999999999111);
INSERT INTO tracks VALUES(645,'Swedish Schnapps',51,1,2,NULL,191268,6359750,0.98999999999999999111);
INSERT INTO tracks VALUES(646,'Samba Da Bênção',52,1,11,NULL,409965,13490008,0.98999999999999999111);
INSERT INTO tracks VALUES(647,'Pot-Pourri N.º 4',52,1,11,NULL,392437,13125975,0.98999999999999999111);
INSERT INTO tracks VALUES(648,'Onde Anda Você',52,1,11,NULL,168437,5550356,0.98999999999999999111);
INSERT INTO tracks VALUES(649,'Samba Da Volta',52,1,11,NULL,170631,5676090,0.98999999999999999111);
INSERT INTO tracks VALUES(650,'Canto De Ossanha',52,1,11,NULL,204956,6771624,0.98999999999999999111);
INSERT INTO tracks VALUES(651,'Pot-Pourri N.º 5',52,1,11,NULL,219898,7117769,0.98999999999999999111);
INSERT INTO tracks VALUES(652,'Formosa',52,1,11,NULL,137482,4560873,0.98999999999999999111);
INSERT INTO tracks VALUES(653,'Como É Duro Trabalhar',52,1,11,NULL,226168,7541177,0.98999999999999999111);
INSERT INTO tracks VALUES(654,'Minha Namorada',52,1,11,NULL,244297,7927967,0.98999999999999999111);
INSERT INTO tracks VALUES(655,'Por Que Será',52,1,11,NULL,162142,5371483,0.98999999999999999111);
INSERT INTO tracks VALUES(656,'Berimbau',52,1,11,NULL,190667,6335548,0.98999999999999999111);
INSERT INTO tracks VALUES(657,'Deixa',52,1,11,NULL,179826,5932799,0.98999999999999999111);
INSERT INTO tracks VALUES(658,'Pot-Pourri N.º 2',52,1,11,NULL,211748,6878359,0.98999999999999999111);
INSERT INTO tracks VALUES(659,'Samba Em Prelúdio',52,1,11,NULL,212636,6923473,0.98999999999999999111);
INSERT INTO tracks VALUES(660,'Carta Ao Tom 74',52,1,11,NULL,162560,5382354,0.98999999999999999111);
INSERT INTO tracks VALUES(661,'Linha de Passe (João Bosco)',53,1,7,NULL,230948,7902328,0.98999999999999999111);
INSERT INTO tracks VALUES(662,'Pela Luz dos Olhos Teus (Miúcha e Tom Jobim)',53,1,7,NULL,163970,5399626,0.98999999999999999111);
INSERT INTO tracks VALUES(663,'Chão de Giz (Elba Ramalho)',53,1,7,NULL,274834,9016916,0.98999999999999999111);
INSERT INTO tracks VALUES(664,'Marina (Dorival Caymmi)',53,1,7,NULL,172643,5523628,0.98999999999999999111);
INSERT INTO tracks VALUES(665,'Aquarela (Toquinho)',53,1,7,NULL,259944,8480140,0.98999999999999999111);
INSERT INTO tracks VALUES(666,'Coração do Agreste (Fafá de Belém)',53,1,7,NULL,258194,8380320,0.98999999999999999111);
INSERT INTO tracks VALUES(667,'Dona (Roupa Nova)',53,1,7,NULL,243356,7991295,0.98999999999999999111);
INSERT INTO tracks VALUES(668,'Começaria Tudo Outra Vez (Maria Creuza)',53,1,7,NULL,206994,6851151,0.98999999999999999111);
INSERT INTO tracks VALUES(669,'Caçador de Mim (Sá & Guarabyra)',53,1,7,NULL,238341,7751360,0.98999999999999999111);
INSERT INTO tracks VALUES(670,'Romaria (Renato Teixeira)',53,1,7,NULL,244793,8033885,0.98999999999999999111);
INSERT INTO tracks VALUES(671,'As Rosas Não Falam (Beth Carvalho)',53,1,7,NULL,116767,3836641,0.98999999999999999111);
INSERT INTO tracks VALUES(672,'Wave (Os Cariocas)',53,1,7,NULL,130063,4298006,0.98999999999999999111);
INSERT INTO tracks VALUES(673,'Garota de Ipanema (Dick Farney)',53,1,7,NULL,174367,5767474,0.98999999999999999111);
INSERT INTO tracks VALUES(674,'Preciso Apender a Viver Só (Maysa)',53,1,7,NULL,143464,4642359,0.98999999999999999111);
INSERT INTO tracks VALUES(675,'Susie Q',54,1,1,'Hawkins-Lewis-Broadwater',275565,9043825,0.98999999999999999111);
INSERT INTO tracks VALUES(676,'I Put A Spell On You',54,1,1,'Jay Hawkins',272091,8943000,0.98999999999999999111);
INSERT INTO tracks VALUES(677,'Proud Mary',54,1,1,'J. C. Fogerty',189022,6229590,0.98999999999999999111);
INSERT INTO tracks VALUES(678,'Bad Moon Rising',54,1,1,'J. C. Fogerty',140146,4609835,0.98999999999999999111);
INSERT INTO tracks VALUES(679,'Lodi',54,1,1,'J. C. Fogerty',191451,6260214,0.98999999999999999111);
INSERT INTO tracks VALUES(680,'Green River',54,1,1,'J. C. Fogerty',154279,5105874,0.98999999999999999111);
INSERT INTO tracks VALUES(681,'Commotion',54,1,1,'J. C. Fogerty',162899,5354252,0.98999999999999999111);
INSERT INTO tracks VALUES(682,'Down On The Corner',54,1,1,'J. C. Fogerty',164858,5521804,0.98999999999999999111);
INSERT INTO tracks VALUES(683,'Fortunate Son',54,1,1,'J. C. Fogerty',140329,4617559,0.98999999999999999111);
INSERT INTO tracks VALUES(684,'Travelin'' Band',54,1,1,'J. C. Fogerty',129358,4270414,0.98999999999999999111);
INSERT INTO tracks VALUES(685,'Who''ll Stop The Rain',54,1,1,'J. C. Fogerty',149394,4899579,0.98999999999999999111);
INSERT INTO tracks VALUES(686,'Up Around The Bend',54,1,1,'J. C. Fogerty',162429,5368701,0.98999999999999999111);
INSERT INTO tracks VALUES(687,'Run Through The Jungle',54,1,1,'J. C. Fogerty',186044,6156567,0.98999999999999999111);
INSERT INTO tracks VALUES(688,'Lookin'' Out My Back Door',54,1,1,'J. C. Fogerty',152946,5034670,0.98999999999999999111);
INSERT INTO tracks VALUES(689,'Long As I Can See The Light',54,1,1,'J. C. Fogerty',213237,6924024,0.98999999999999999111);
INSERT INTO tracks VALUES(690,'I Heard It Through The Grapevine',54,1,1,'Whitfield-Strong',664894,21947845,0.98999999999999999111);
INSERT INTO tracks VALUES(691,'Have You Ever Seen The Rain?',54,1,1,'J. C. Fogerty',160052,5263675,0.98999999999999999111);
INSERT INTO tracks VALUES(692,'Hey Tonight',54,1,1,'J. C. Fogerty',162847,5343807,0.98999999999999999111);
INSERT INTO tracks VALUES(693,'Sweet Hitch-Hiker',54,1,1,'J. C. Fogerty',175490,5716603,0.98999999999999999111);
INSERT INTO tracks VALUES(694,'Someday Never Comes',54,1,1,'J. C. Fogerty',239360,7945235,0.98999999999999999111);
INSERT INTO tracks VALUES(695,'Walking On The Water',55,1,1,'J.C. Fogerty',281286,9302129,0.98999999999999999111);
INSERT INTO tracks VALUES(696,'Suzie-Q, Pt. 2',55,1,1,'J.C. Fogerty',244114,7986637,0.98999999999999999111);
INSERT INTO tracks VALUES(697,'Born On The Bayou',55,1,1,'J.C. Fogerty',316630,10361866,0.98999999999999999111);
INSERT INTO tracks VALUES(698,'Good Golly Miss Molly',55,1,1,'J.C. Fogerty',163604,5348175,0.98999999999999999111);
INSERT INTO tracks VALUES(699,'Tombstone Shadow',55,1,1,'J.C. Fogerty',218880,7209080,0.98999999999999999111);
INSERT INTO tracks VALUES(700,'Wrote A Song For Everyone',55,1,1,'J.C. Fogerty',296385,9675875,0.98999999999999999111);
INSERT INTO tracks VALUES(701,'Night Time Is The Right Time',55,1,1,'J.C. Fogerty',190119,6211173,0.98999999999999999111);
INSERT INTO tracks VALUES(702,'Cotton Fields',55,1,1,'J.C. Fogerty',178181,5919224,0.98999999999999999111);
INSERT INTO tracks VALUES(703,'It Came Out Of The Sky',55,1,1,'J.C. Fogerty',176718,5807474,0.98999999999999999111);
INSERT INTO tracks VALUES(704,'Don''t Look Now',55,1,1,'J.C. Fogerty',131918,4366455,0.98999999999999999111);
INSERT INTO tracks VALUES(705,'The Midnight Special',55,1,1,'J.C. Fogerty',253596,8297482,0.98999999999999999111);
INSERT INTO tracks VALUES(706,'Before You Accuse Me',55,1,1,'J.C. Fogerty',207804,6815126,0.98999999999999999111);
INSERT INTO tracks VALUES(707,'My Baby Left Me',55,1,1,'J.C. Fogerty',140460,4633440,0.98999999999999999111);
INSERT INTO tracks VALUES(708,'Pagan Baby',55,1,1,'J.C. Fogerty',385619,12713813,0.98999999999999999111);
INSERT INTO tracks VALUES(709,'(Wish I Could) Hideaway',55,1,1,'J.C. Fogerty',228466,7432978,0.98999999999999999111);
INSERT INTO tracks VALUES(710,'It''s Just A Thought',55,1,1,'J.C. Fogerty',237374,7778319,0.98999999999999999111);
INSERT INTO tracks VALUES(711,'Molina',55,1,1,'J.C. Fogerty',163239,5390811,0.98999999999999999111);
INSERT INTO tracks VALUES(712,'Born To Move',55,1,1,'J.C. Fogerty',342804,11260814,0.98999999999999999111);
INSERT INTO tracks VALUES(713,'Lookin'' For A Reason',55,1,1,'J.C. Fogerty',209789,6933135,0.98999999999999999111);
INSERT INTO tracks VALUES(714,'Hello Mary Lou',55,1,1,'J.C. Fogerty',132832,4476563,0.98999999999999999111);
INSERT INTO tracks VALUES(715,'Gatas Extraordinárias',56,1,7,NULL,212506,7095702,0.98999999999999999111);
INSERT INTO tracks VALUES(716,'Brasil',56,1,7,NULL,243696,7911683,0.98999999999999999111);
INSERT INTO tracks VALUES(717,'Eu Sou Neguinha (Ao Vivo)',56,1,7,NULL,251768,8376000,0.98999999999999999111);
INSERT INTO tracks VALUES(718,'Geração Coca-Cola (Ao Vivo)',56,1,7,NULL,228153,7573301,0.98999999999999999111);
INSERT INTO tracks VALUES(719,'Lanterna Dos Afogados',56,1,7,NULL,204538,6714582,0.98999999999999999111);
INSERT INTO tracks VALUES(720,'Coroné Antonio Bento',56,1,7,NULL,200437,6713066,0.98999999999999999111);
INSERT INTO tracks VALUES(721,'Você Passa, Eu Acho Graça (Ao Vivo)',56,1,7,NULL,206733,6943576,0.98999999999999999111);
INSERT INTO tracks VALUES(722,'Meu Mundo Fica Completo (Com Você)',56,1,7,NULL,247771,8322240,0.98999999999999999111);
INSERT INTO tracks VALUES(723,'1° De Julho',56,1,7,NULL,270262,9017535,0.98999999999999999111);
INSERT INTO tracks VALUES(724,'Música Urbana 2',56,1,7,NULL,194899,6383472,0.98999999999999999111);
INSERT INTO tracks VALUES(725,'Vida Bandida (Ao Vivo)',56,1,7,NULL,192626,6360785,0.98999999999999999111);
INSERT INTO tracks VALUES(726,'Palavras Ao Vento',56,1,7,NULL,212453,7048676,0.98999999999999999111);
INSERT INTO tracks VALUES(727,'Não Sei O Que Eu Quero Da Vida',56,1,7,NULL,151849,5024963,0.98999999999999999111);
INSERT INTO tracks VALUES(728,'Woman Is The Nigger Of The World (Ao Vivo)',56,1,7,NULL,298919,9724145,0.98999999999999999111);
INSERT INTO tracks VALUES(729,'Juventude Transviada (Ao Vivo)',56,1,7,NULL,278622,9183808,0.98999999999999999111);
INSERT INTO tracks VALUES(730,'Malandragem',57,1,7,NULL,247588,8165048,0.98999999999999999111);
INSERT INTO tracks VALUES(731,'O Segundo Sol',57,1,7,NULL,252133,8335629,0.98999999999999999111);
INSERT INTO tracks VALUES(732,'Smells Like Teen Spirit (Ao Vivo)',57,1,7,NULL,316865,10384506,0.98999999999999999111);
INSERT INTO tracks VALUES(733,'E.C.T.',57,1,7,NULL,227500,7571834,0.98999999999999999111);
INSERT INTO tracks VALUES(734,'Todo Amor Que Houver Nesta Vida',57,1,7,NULL,227160,7420347,0.98999999999999999111);
INSERT INTO tracks VALUES(735,'Metrô. Linha 743',57,1,7,NULL,174654,5837495,0.98999999999999999111);
INSERT INTO tracks VALUES(736,'Nós (Ao Vivo)',57,1,7,NULL,193828,6498661,0.98999999999999999111);
INSERT INTO tracks VALUES(737,'Na Cadência Do Samba',57,1,7,NULL,196075,6483952,0.98999999999999999111);
INSERT INTO tracks VALUES(738,'Admirável Gado Novo',57,1,7,NULL,274390,9144031,0.98999999999999999111);
INSERT INTO tracks VALUES(739,'Eleanor Rigby',57,1,7,NULL,189466,6303205,0.98999999999999999111);
INSERT INTO tracks VALUES(740,'Socorro',57,1,7,NULL,258586,8549393,0.98999999999999999111);
INSERT INTO tracks VALUES(741,'Blues Da Piedade',57,1,7,NULL,257123,8472964,0.98999999999999999111);
INSERT INTO tracks VALUES(742,'Rubens',57,1,7,NULL,211853,7026317,0.98999999999999999111);
INSERT INTO tracks VALUES(743,'Não Deixe O Samba Morrer - Cassia Eller e Alcione',57,1,7,NULL,268173,8936345,0.98999999999999999111);
INSERT INTO tracks VALUES(744,'Mis Penas Lloraba Yo (Ao Vivo) Soy Gitano (Tangos)',57,1,7,NULL,188473,6195854,0.98999999999999999111);
INSERT INTO tracks VALUES(745,'Comin'' Home',58,1,1,'Bolin/Coverdale/Paice',235781,7644604,0.98999999999999999111);
INSERT INTO tracks VALUES(746,'Lady Luck',58,1,1,'Cook/Coverdale',168202,5501379,0.98999999999999999111);
INSERT INTO tracks VALUES(747,'Gettin'' Tighter',58,1,1,'Bolin/Hughes',218044,7176909,0.98999999999999999111);
INSERT INTO tracks VALUES(748,'Dealer',58,1,1,'Bolin/Coverdale',230922,7591066,0.98999999999999999111);
INSERT INTO tracks VALUES(749,'I Need Love',58,1,1,'Bolin/Coverdale',263836,8701064,0.98999999999999999111);
INSERT INTO tracks VALUES(750,'Drifter',58,1,1,'Bolin/Coverdale',242834,8001505,0.98999999999999999111);
INSERT INTO tracks VALUES(751,'Love Child',58,1,1,'Bolin/Coverdale',188160,6173806,0.98999999999999999111);
INSERT INTO tracks VALUES(752,'This Time Around / Owed to ''G'' [Instrumental]',58,1,1,'Bolin/Hughes/Lord',370102,11995679,0.98999999999999999111);
INSERT INTO tracks VALUES(753,'You Keep On Moving',58,1,1,'Coverdale/Hughes',319111,10447868,0.98999999999999999111);
INSERT INTO tracks VALUES(754,'Speed King',59,1,1,'Blackmore, Gillan, Glover, Lord, Paice',264385,8587578,0.98999999999999999111);
INSERT INTO tracks VALUES(755,'Bloodsucker',59,1,1,'Blackmore, Gillan, Glover, Lord, Paice',256261,8344405,0.98999999999999999111);
INSERT INTO tracks VALUES(756,'Child In Time',59,1,1,'Blackmore, Gillan, Glover, Lord, Paice',620460,20230089,0.98999999999999999111);
INSERT INTO tracks VALUES(757,'Flight Of The Rat',59,1,1,'Blackmore, Gillan, Glover, Lord, Paice',478302,15563967,0.98999999999999999111);
INSERT INTO tracks VALUES(758,'Into The Fire',59,1,1,'Blackmore, Gillan, Glover, Lord, Paice',210259,6849310,0.98999999999999999111);
INSERT INTO tracks VALUES(759,'Living Wreck',59,1,1,'Blackmore, Gillan, Glover, Lord, Paice',274886,8993056,0.98999999999999999111);
INSERT INTO tracks VALUES(760,'Hard Lovin'' Man',59,1,1,'Blackmore, Gillan, Glover, Lord, Paice',431203,13931179,0.98999999999999999111);
INSERT INTO tracks VALUES(761,'Fireball',60,1,1,'Ritchie Blackmore, Ian Gillan, Roger Glover, Jon Lord, Ian Paice',204721,6714807,0.98999999999999999111);
INSERT INTO tracks VALUES(762,'No No No',60,1,1,'Ritchie Blackmore, Ian Gillan, Roger Glover, Jon Lord, Ian Paice',414902,13646606,0.98999999999999999111);
INSERT INTO tracks VALUES(763,'Strange Kind Of Woman',60,1,1,'Ritchie Blackmore, Ian Gillan, Roger Glover, Jon Lord, Ian Paice',247092,8072036,0.98999999999999999111);
INSERT INTO tracks VALUES(764,'Anyone''s Daughter',60,1,1,'Ritchie Blackmore, Ian Gillan, Roger Glover, Jon Lord, Ian Paice',284682,9354480,0.98999999999999999111);
INSERT INTO tracks VALUES(765,'The Mule',60,1,1,'Ritchie Blackmore, Ian Gillan, Roger Glover, Jon Lord, Ian Paice',322063,10638390,0.98999999999999999111);
INSERT INTO tracks VALUES(766,'Fools',60,1,1,'Ritchie Blackmore, Ian Gillan, Roger Glover, Jon Lord, Ian Paice',500427,16279366,0.98999999999999999111);
INSERT INTO tracks VALUES(767,'No One Came',60,1,1,'Ritchie Blackmore, Ian Gillan, Roger Glover, Jon Lord, Ian Paice',385880,12643813,0.98999999999999999111);
INSERT INTO tracks VALUES(768,'Knocking At Your Back Door',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover',424829,13779332,0.98999999999999999111);
INSERT INTO tracks VALUES(769,'Bad Attitude',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Jon Lord',307905,10035180,0.98999999999999999111);
INSERT INTO tracks VALUES(770,'Child In Time (Son Of Aleric - Instrumental)',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Jon Lord, Ian Paice',602880,19712753,0.98999999999999999111);
INSERT INTO tracks VALUES(771,'Nobody''s Home',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Jon Lord, Ian Paice',243017,7929493,0.98999999999999999111);
INSERT INTO tracks VALUES(772,'Black Night',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Jon Lord, Ian Paice',368770,12058906,0.98999999999999999111);
INSERT INTO tracks VALUES(773,'Perfect Strangers',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover',321149,10445353,0.98999999999999999111);
INSERT INTO tracks VALUES(774,'The Unwritten Law',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Ian Paice',295053,9740361,0.98999999999999999111);
INSERT INTO tracks VALUES(775,'Call Of The Wild',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Jon Lord',293851,9575295,0.98999999999999999111);
INSERT INTO tracks VALUES(776,'Hush',61,1,1,'South',213054,6944928,0.98999999999999999111);
INSERT INTO tracks VALUES(777,'Smoke On The Water',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Jon Lord, Ian Paice',464378,15180849,0.98999999999999999111);
INSERT INTO tracks VALUES(778,'Space Trucking',61,1,1,'Richie Blackmore, Ian Gillian, Roger Glover, Jon Lord, Ian Paice',341185,11122183,0.98999999999999999111);
INSERT INTO tracks VALUES(779,'Highway Star',62,1,1,'Ian Gillan/Ian Paice/Jon Lord/Ritchie Blckmore/Roger Glover',368770,12012452,0.98999999999999999111);
INSERT INTO tracks VALUES(780,'Maybe I''m A Leo',62,1,1,'Ian Gillan/Ian Paice/Jon Lord/Ritchie Blckmore/Roger Glover',290455,9502646,0.98999999999999999111);
INSERT INTO tracks VALUES(781,'Pictures Of Home',62,1,1,'Ian Gillan/Ian Paice/Jon Lord/Ritchie Blckmore/Roger Glover',303777,9903835,0.98999999999999999111);
INSERT INTO tracks VALUES(782,'Never Before',62,1,1,'Ian Gillan/Ian Paice/Jon Lord/Ritchie Blckmore/Roger Glover',239830,7832790,0.98999999999999999111);
INSERT INTO tracks VALUES(783,'Smoke On The Water',62,1,1,'Ian Gillan/Ian Paice/Jon Lord/Ritchie Blckmore/Roger Glover',340871,11246496,0.98999999999999999111);
INSERT INTO tracks VALUES(784,'Lazy',62,1,1,'Ian Gillan/Ian Paice/Jon Lord/Ritchie Blckmore/Roger Glover',442096,14397671,0.98999999999999999111);
INSERT INTO tracks VALUES(785,'Space Truckin''',62,1,1,'Ian Gillan/Ian Paice/Jon Lord/Ritchie Blckmore/Roger Glover',272796,8981030,0.98999999999999999111);
INSERT INTO tracks VALUES(786,'Vavoom : Ted The Mechanic',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',257384,8510755,0.98999999999999999111);
INSERT INTO tracks VALUES(787,'Loosen My Strings',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',359680,11702232,0.98999999999999999111);
INSERT INTO tracks VALUES(788,'Soon Forgotten',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',287791,9401383,0.98999999999999999111);
INSERT INTO tracks VALUES(789,'Sometimes I Feel Like Screaming',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',451840,14789410,0.98999999999999999111);
INSERT INTO tracks VALUES(790,'Cascades : I''m Not Your Lover',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',283689,9209693,0.98999999999999999111);
INSERT INTO tracks VALUES(791,'The Aviator',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',320992,10532053,0.98999999999999999111);
INSERT INTO tracks VALUES(792,'Rosa''s Cantina',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',312372,10323804,0.98999999999999999111);
INSERT INTO tracks VALUES(793,'A Castle Full Of Rascals',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',311693,10159566,0.98999999999999999111);
INSERT INTO tracks VALUES(794,'A Touch Away',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',276323,9098561,0.98999999999999999111);
INSERT INTO tracks VALUES(795,'Hey Cisco',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',354089,11600029,0.98999999999999999111);
INSERT INTO tracks VALUES(796,'Somebody Stole My Guitar',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',249443,8180421,0.98999999999999999111);
INSERT INTO tracks VALUES(797,'The Purpendicular Waltz',63,1,1,'Ian Gillan, Roger Glover, Jon Lord, Steve Morse, Ian Paice',283924,9299131,0.98999999999999999111);
INSERT INTO tracks VALUES(798,'King Of Dreams',64,1,1,'Blackmore, Glover, Turner',328385,10733847,0.98999999999999999111);
INSERT INTO tracks VALUES(799,'The Cut Runs Deep',64,1,1,'Blackmore, Glover, Turner, Lord, Paice',342752,11191650,0.98999999999999999111);
INSERT INTO tracks VALUES(800,'Fire In The Basement',64,1,1,'Blackmore, Glover, Turner, Lord, Paice',283977,9267550,0.98999999999999999111);
INSERT INTO tracks VALUES(801,'Truth Hurts',64,1,1,'Blackmore, Glover, Turner',314827,10224612,0.98999999999999999111);
INSERT INTO tracks VALUES(802,'Breakfast In Bed',64,1,1,'Blackmore, Glover, Turner',317126,10323804,0.98999999999999999111);
INSERT INTO tracks VALUES(803,'Love Conquers All',64,1,1,'Blackmore, Glover, Turner',227186,7328516,0.98999999999999999111);
INSERT INTO tracks VALUES(804,'Fortuneteller',64,1,1,'Blackmore, Glover, Turner, Lord, Paice',349335,11369671,0.98999999999999999111);
INSERT INTO tracks VALUES(805,'Too Much Is Not Enough',64,1,1,'Turner, Held, Greenwood',257724,8382800,0.98999999999999999111);
INSERT INTO tracks VALUES(806,'Wicked Ways',64,1,1,'Blackmore, Glover, Turner, Lord, Paice',393691,12826582,0.98999999999999999111);
INSERT INTO tracks VALUES(807,'Stormbringer',65,1,1,'D.Coverdale/R.Blackmore/Ritchie Blackmore',246413,8044864,0.98999999999999999111);
INSERT INTO tracks VALUES(808,'Love Don''t Mean a Thing',65,1,1,'D.Coverdale/G.Hughes/Glenn Hughes/I.Paice/Ian Paice/J.Lord/John Lord/R.Blackmore/Ritchie Blackmore',263862,8675026,0.98999999999999999111);
INSERT INTO tracks VALUES(809,'Holy Man',65,1,1,'D.Coverdale/G.Hughes/Glenn Hughes/J.Lord/John Lord',270236,8818093,0.98999999999999999111);
INSERT INTO tracks VALUES(810,'Hold On',65,1,1,'D.Coverdal/G.Hughes/Glenn Hughes/I.Paice/Ian Paice/J.Lord/John Lord',306860,10022428,0.98999999999999999111);
INSERT INTO tracks VALUES(811,'Lady Double Dealer',65,1,1,'D.Coverdale/R.Blackmore/Ritchie Blackmore',201482,6554330,0.98999999999999999111);
INSERT INTO tracks VALUES(812,'You Can''t Do it Right (With the One You Love)',65,1,1,'D.Coverdale/G.Hughes/Glenn Hughes/R.Blackmore/Ritchie Blackmore',203755,6709579,0.98999999999999999111);
INSERT INTO tracks VALUES(813,'High Ball Shooter',65,1,1,'D.Coverdale/G.Hughes/Glenn Hughes/I.Paice/Ian Paice/J.Lord/John Lord/R.Blackmore/Ritchie Blackmore',267833,8772471,0.98999999999999999111);
INSERT INTO tracks VALUES(814,'The Gypsy',65,1,1,'D.Coverdale/G.Hughes/Glenn Hughes/I.Paice/Ian Paice/J.Lord/John Lord/R.Blackmore/Ritchie Blackmore',242886,7946614,0.98999999999999999111);
INSERT INTO tracks VALUES(815,'Soldier Of Fortune',65,1,1,'D.Coverdale/R.Blackmore/Ritchie Blackmore',193750,6315321,0.98999999999999999111);
INSERT INTO tracks VALUES(816,'The Battle Rages On',66,1,1,'ian paice/jon lord',356963,11626228,0.98999999999999999111);
INSERT INTO tracks VALUES(817,'Lick It Up',66,1,1,'roger glover',240274,7792604,0.98999999999999999111);
INSERT INTO tracks VALUES(818,'Anya',66,1,1,'jon lord/roger glover',392437,12754921,0.98999999999999999111);
INSERT INTO tracks VALUES(819,'Talk About Love',66,1,1,'roger glover',247823,8072171,0.98999999999999999111);
INSERT INTO tracks VALUES(820,'Time To Kill',66,1,1,'roger glover',351033,11354742,0.98999999999999999111);
INSERT INTO tracks VALUES(821,'Ramshackle Man',66,1,1,'roger glover',334445,10874679,0.98999999999999999111);
INSERT INTO tracks VALUES(822,'A Twist In The Tail',66,1,1,'roger glover',257462,8413103,0.98999999999999999111);
INSERT INTO tracks VALUES(823,'Nasty Piece Of Work',66,1,1,'jon lord/roger glover',276662,9076997,0.98999999999999999111);
INSERT INTO tracks VALUES(824,'Solitaire',66,1,1,'roger glover',282226,9157021,0.98999999999999999111);
INSERT INTO tracks VALUES(825,'One Man''s Meat',66,1,1,'roger glover',278804,9068960,0.98999999999999999111);
INSERT INTO tracks VALUES(826,'Pour Some Sugar On Me',67,1,1,NULL,292519,9518842,0.98999999999999999111);
INSERT INTO tracks VALUES(827,'Photograph',67,1,1,NULL,248633,8108507,0.98999999999999999111);
INSERT INTO tracks VALUES(828,'Love Bites',67,1,1,NULL,346853,11305791,0.98999999999999999111);
INSERT INTO tracks VALUES(829,'Let''s Get Rocked',67,1,1,NULL,296019,9724150,0.98999999999999999111);
INSERT INTO tracks VALUES(830,'Two Steps Behind [Acoustic Version]',67,1,1,NULL,259787,8523388,0.98999999999999999111);
INSERT INTO tracks VALUES(831,'Animal',67,1,1,NULL,244741,7985133,0.98999999999999999111);
INSERT INTO tracks VALUES(832,'Heaven Is',67,1,1,NULL,214021,6988128,0.98999999999999999111);
INSERT INTO tracks VALUES(833,'Rocket',67,1,1,NULL,247248,8092463,0.98999999999999999111);
INSERT INTO tracks VALUES(834,'When Love & Hate Collide',67,1,1,NULL,257280,8364633,0.98999999999999999111);
INSERT INTO tracks VALUES(835,'Action',67,1,1,NULL,220604,7130830,0.98999999999999999111);
INSERT INTO tracks VALUES(836,'Make Love Like A Man',67,1,1,NULL,255660,8309725,0.98999999999999999111);
INSERT INTO tracks VALUES(837,'Armageddon It',67,1,1,NULL,322455,10522352,0.98999999999999999111);
INSERT INTO tracks VALUES(838,'Have You Ever Needed Someone So Bad',67,1,1,NULL,319320,10400020,0.98999999999999999111);
INSERT INTO tracks VALUES(839,'Rock Of Ages',67,1,1,NULL,248424,8150318,0.98999999999999999111);
INSERT INTO tracks VALUES(840,'Hysteria',67,1,1,NULL,355056,11622738,0.98999999999999999111);
INSERT INTO tracks VALUES(841,'Bringin'' On The Heartbreak',67,1,1,NULL,272457,8853324,0.98999999999999999111);
INSERT INTO tracks VALUES(842,'Roll Call',68,1,2,'Jim Beard',321358,10653494,0.98999999999999999111);
INSERT INTO tracks VALUES(843,'Otay',68,1,2,'John Scofield, Robert Aries, Milton Chambers and Gary Grainger',423653,14176083,0.98999999999999999111);
INSERT INTO tracks VALUES(844,'Groovus Interruptus',68,1,2,'Jim Beard',319373,10602166,0.98999999999999999111);
INSERT INTO tracks VALUES(845,'Paris On Mine',68,1,2,'Jon Herington',368875,12059507,0.98999999999999999111);
INSERT INTO tracks VALUES(846,'In Time',68,1,2,'Sylvester Stewart',368953,12287103,0.98999999999999999111);
INSERT INTO tracks VALUES(847,'Plan B',68,1,2,'Dean Brown, Dennis Chambers & Jim Beard',272039,9032315,0.98999999999999999111);
INSERT INTO tracks VALUES(848,'Outbreak',68,1,2,'Jim Beard & Jon Herington',659226,21685807,0.98999999999999999111);
INSERT INTO tracks VALUES(849,'Baltimore, DC',68,1,2,'John Scofield',346932,11394473,0.98999999999999999111);
INSERT INTO tracks VALUES(850,'Talkin Loud and Saying Nothin',68,1,2,'James Brown & Bobby Byrd',360411,11994859,0.98999999999999999111);
INSERT INTO tracks VALUES(851,'Pétala',69,1,7,NULL,270080,8856165,0.98999999999999999111);
INSERT INTO tracks VALUES(852,'Meu Bem-Querer',69,1,7,NULL,255608,8330047,0.98999999999999999111);
INSERT INTO tracks VALUES(853,'Cigano',69,1,7,NULL,304692,10037362,0.98999999999999999111);
INSERT INTO tracks VALUES(854,'Boa Noite',69,1,7,NULL,338755,11283582,0.98999999999999999111);
INSERT INTO tracks VALUES(855,'Fato Consumado',69,1,7,NULL,211565,7018586,0.98999999999999999111);
INSERT INTO tracks VALUES(856,'Faltando Um Pedaço',69,1,7,NULL,267728,8788760,0.98999999999999999111);
INSERT INTO tracks VALUES(857,'Álibi',69,1,7,NULL,213237,6928434,0.98999999999999999111);
INSERT INTO tracks VALUES(858,'Esquinas',69,1,7,NULL,280999,9096726,0.98999999999999999111);
INSERT INTO tracks VALUES(859,'Se...',69,1,7,NULL,286432,9413777,0.98999999999999999111);
INSERT INTO tracks VALUES(860,'Eu Te Devoro',69,1,7,NULL,311614,10312775,0.98999999999999999111);
INSERT INTO tracks VALUES(861,'Lilás',69,1,7,NULL,274181,9049542,0.98999999999999999111);
INSERT INTO tracks VALUES(862,'Acelerou',69,1,7,NULL,284081,9396942,0.98999999999999999111);
INSERT INTO tracks VALUES(863,'Um Amor Puro',69,1,7,NULL,327784,10687311,0.98999999999999999111);
INSERT INTO tracks VALUES(864,'Samurai',70,1,7,'Djavan',330997,10872787,0.98999999999999999111);
INSERT INTO tracks VALUES(865,'Nem Um Dia',70,1,7,'Djavan',337423,11181446,0.98999999999999999111);
INSERT INTO tracks VALUES(866,'Oceano',70,1,7,'Djavan',217338,7026441,0.98999999999999999111);
INSERT INTO tracks VALUES(867,'Açai',70,1,7,'Djavan',270968,8893682,0.98999999999999999111);
INSERT INTO tracks VALUES(868,'Serrado',70,1,7,'Djavan',295314,9842240,0.98999999999999999111);
INSERT INTO tracks VALUES(869,'Flor De Lis',70,1,7,'Djavan',236355,7801108,0.98999999999999999111);
INSERT INTO tracks VALUES(870,'Amar É Tudo',70,1,7,'Djavan',211617,7073899,0.98999999999999999111);
INSERT INTO tracks VALUES(871,'Azul',70,1,7,'Djavan',253962,8381029,0.98999999999999999111);
INSERT INTO tracks VALUES(872,'Seduzir',70,1,7,'Djavan',277524,9163253,0.98999999999999999111);
INSERT INTO tracks VALUES(873,'A Carta',70,1,7,'Djavan - Gabriel, O Pensador',347297,11493463,0.98999999999999999111);
INSERT INTO tracks VALUES(874,'Sina',70,1,7,'Djavan',268173,8906539,0.98999999999999999111);
INSERT INTO tracks VALUES(875,'Acelerou',70,1,7,'Djavan',284133,9391439,0.98999999999999999111);
INSERT INTO tracks VALUES(876,'Um Amor Puro',70,1,7,'Djavan',327105,10664698,0.98999999999999999111);
INSERT INTO tracks VALUES(877,'O Bêbado e a Equilibrista',71,1,7,NULL,223059,7306143,0.98999999999999999111);
INSERT INTO tracks VALUES(878,'O Mestre-Sala dos Mares',71,1,7,NULL,186226,6180414,0.98999999999999999111);
INSERT INTO tracks VALUES(879,'Atrás da Porta',71,1,7,NULL,166608,5432518,0.98999999999999999111);
INSERT INTO tracks VALUES(880,'Dois Pra Lá, Dois Pra Cá',71,1,7,NULL,263026,8684639,0.98999999999999999111);
INSERT INTO tracks VALUES(881,'Casa no Campo',71,1,7,NULL,170788,5531841,0.98999999999999999111);
INSERT INTO tracks VALUES(882,'Romaria',71,1,7,NULL,242834,7968525,0.98999999999999999111);
INSERT INTO tracks VALUES(883,'Alô, Alô, Marciano',71,1,7,NULL,241397,8137254,0.98999999999999999111);
INSERT INTO tracks VALUES(884,'Me Deixas Louca',71,1,7,NULL,214831,6888030,0.98999999999999999111);
INSERT INTO tracks VALUES(885,'Fascinação',71,1,7,NULL,180793,5793959,0.98999999999999999111);
INSERT INTO tracks VALUES(886,'Saudosa Maloca',71,1,7,NULL,278125,9059416,0.98999999999999999111);
INSERT INTO tracks VALUES(887,'As Aparências Enganam',71,1,7,NULL,247379,8014346,0.98999999999999999111);
INSERT INTO tracks VALUES(888,'Madalena',71,1,7,NULL,157387,5243721,0.98999999999999999111);
INSERT INTO tracks VALUES(889,'Maria Rosa',71,1,7,NULL,232803,7592504,0.98999999999999999111);
INSERT INTO tracks VALUES(890,'Aprendendo A Jogar',71,1,7,NULL,290664,9391041,0.98999999999999999111);
INSERT INTO tracks VALUES(891,'Layla',72,1,6,'Clapton/Gordon',430733,14115792,0.98999999999999999111);
INSERT INTO tracks VALUES(892,'Badge',72,1,6,'Clapton/Harrison',163552,5322942,0.98999999999999999111);
INSERT INTO tracks VALUES(893,'I Feel Free',72,1,6,'Bruce/Clapton',174576,5725684,0.98999999999999999111);
INSERT INTO tracks VALUES(894,'Sunshine Of Your Love',72,1,6,'Bruce/Clapton',252891,8225889,0.98999999999999999111);
INSERT INTO tracks VALUES(895,'Crossroads',72,1,6,'Clapton/Robert Johnson Arr: Eric Clapton',253335,8273540,0.98999999999999999111);
INSERT INTO tracks VALUES(896,'Strange Brew',72,1,6,'Clapton/Collins/Pappalardi',167810,5489787,0.98999999999999999111);
INSERT INTO tracks VALUES(897,'White Room',72,1,6,'Bruce/Clapton',301583,9872606,0.98999999999999999111);
INSERT INTO tracks VALUES(898,'Bell Bottom Blues',72,1,6,'Clapton',304744,9946681,0.98999999999999999111);
INSERT INTO tracks VALUES(899,'Cocaine',72,1,6,'Cale/Clapton',215928,7138399,0.98999999999999999111);
INSERT INTO tracks VALUES(900,'I Shot The Sheriff',72,1,6,'Marley',263862,8738973,0.98999999999999999111);
INSERT INTO tracks VALUES(901,'After Midnight',72,1,6,'Clapton/J. J. Cale',191320,6460941,0.98999999999999999111);
INSERT INTO tracks VALUES(902,'Swing Low Sweet Chariot',72,1,6,'Clapton/Trad. Arr. Clapton',208143,6896288,0.98999999999999999111);
INSERT INTO tracks VALUES(903,'Lay Down Sally',72,1,6,'Clapton/Levy',231732,7774207,0.98999999999999999111);
INSERT INTO tracks VALUES(904,'Knockin On Heavens Door',72,1,6,'Clapton/Dylan',264411,8758819,0.98999999999999999111);
INSERT INTO tracks VALUES(905,'Wonderful Tonight',72,1,6,'Clapton',221387,7326923,0.98999999999999999111);
INSERT INTO tracks VALUES(906,'Let It Grow',72,1,6,'Clapton',297064,9742568,0.98999999999999999111);
INSERT INTO tracks VALUES(907,'Promises',72,1,6,'Clapton/F.eldman/Linn',180401,6006154,0.98999999999999999111);
INSERT INTO tracks VALUES(908,'I Can''t Stand It',72,1,6,'Clapton',249730,8271980,0.98999999999999999111);
INSERT INTO tracks VALUES(909,'Signe',73,1,6,'Eric Clapton',193515,6475042,0.98999999999999999111);
INSERT INTO tracks VALUES(910,'Before You Accuse Me',73,1,6,'Eugene McDaniel',224339,7456807,0.98999999999999999111);
INSERT INTO tracks VALUES(911,'Hey Hey',73,1,6,'Big Bill Broonzy',196466,6543487,0.98999999999999999111);
INSERT INTO tracks VALUES(912,'Tears In Heaven',73,1,6,'Eric Clapton, Will Jennings',274729,9032835,0.98999999999999999111);
INSERT INTO tracks VALUES(913,'Lonely Stranger',73,1,6,'Eric Clapton',328724,10894406,0.98999999999999999111);
INSERT INTO tracks VALUES(914,'Nobody Knows You When You''re Down & Out',73,1,6,'Jimmy Cox',231836,7669922,0.98999999999999999111);
INSERT INTO tracks VALUES(915,'Layla',73,1,6,'Eric Clapton, Jim Gordon',285387,9490542,0.98999999999999999111);
INSERT INTO tracks VALUES(916,'Running On Faith',73,1,6,'Jerry Lynn Williams',378984,12536275,0.98999999999999999111);
INSERT INTO tracks VALUES(917,'Walkin'' Blues',73,1,6,'Robert Johnson',226429,7435192,0.98999999999999999111);
INSERT INTO tracks VALUES(918,'Alberta',73,1,6,'Traditional',222406,7412975,0.98999999999999999111);
INSERT INTO tracks VALUES(919,'San Francisco Bay Blues',73,1,6,'Jesse Fuller',203363,6724021,0.98999999999999999111);
INSERT INTO tracks VALUES(920,'Malted Milk',73,1,6,'Robert Johnson',216528,7096781,0.98999999999999999111);
INSERT INTO tracks VALUES(921,'Old Love',73,1,6,'Eric Clapton, Robert Cray',472920,15780747,0.98999999999999999111);
INSERT INTO tracks VALUES(922,'Rollin'' And Tumblin''',73,1,6,'McKinley Morgenfield (Muddy Waters)',251768,8407355,0.98999999999999999111);
INSERT INTO tracks VALUES(923,'Collision',74,1,4,'Jon Hudson/Mike Patton',204303,6656596,0.98999999999999999111);
INSERT INTO tracks VALUES(924,'Stripsearch',74,1,4,'Jon Hudson/Mike Bordin/Mike Patton',270106,8861119,0.98999999999999999111);
INSERT INTO tracks VALUES(925,'Last Cup Of Sorrow',74,1,4,'Bill Gould/Mike Patton',251663,8221247,0.98999999999999999111);
INSERT INTO tracks VALUES(926,'Naked In Front Of The Computer',74,1,4,'Mike Patton',128757,4225077,0.98999999999999999111);
INSERT INTO tracks VALUES(927,'Helpless',74,1,4,'Bill Gould/Mike Bordin/Mike Patton',326217,10753135,0.98999999999999999111);
INSERT INTO tracks VALUES(928,'Mouth To Mouth',74,1,4,'Bill Gould/Jon Hudson/Mike Bordin/Mike Patton',228493,7505887,0.98999999999999999111);
INSERT INTO tracks VALUES(929,'Ashes To Ashes',74,1,4,'Bill Gould/Jon Hudson/Mike Bordin/Mike Patton/Roddy Bottum',217391,7093746,0.98999999999999999111);
INSERT INTO tracks VALUES(930,'She Loves Me Not',74,1,4,'Bill Gould/Mike Bordin/Mike Patton',209867,6887544,0.98999999999999999111);
INSERT INTO tracks VALUES(931,'Got That Feeling',74,1,4,'Mike Patton',140852,4643227,0.98999999999999999111);
INSERT INTO tracks VALUES(932,'Paths Of Glory',74,1,4,'Bill Gould/Jon Hudson/Mike Bordin/Mike Patton/Roddy Bottum',257253,8436300,0.98999999999999999111);
INSERT INTO tracks VALUES(933,'Home Sick Home',74,1,4,'Mike Patton',119040,3898976,0.98999999999999999111);
INSERT INTO tracks VALUES(934,'Pristina',74,1,4,'Bill Gould/Mike Patton',232698,7497361,0.98999999999999999111);
INSERT INTO tracks VALUES(935,'Land Of Sunshine',75,1,4,NULL,223921,7353567,0.98999999999999999111);
INSERT INTO tracks VALUES(936,'Caffeine',75,1,4,NULL,267937,8747367,0.98999999999999999111);
INSERT INTO tracks VALUES(937,'Midlife Crisis',75,1,4,NULL,263235,8628841,0.98999999999999999111);
INSERT INTO tracks VALUES(938,'RV',75,1,4,NULL,223242,7288162,0.98999999999999999111);
INSERT INTO tracks VALUES(939,'Smaller And Smaller',75,1,4,NULL,310831,10180103,0.98999999999999999111);
INSERT INTO tracks VALUES(940,'Everything''s Ruined',75,1,4,NULL,273658,9010917,0.98999999999999999111);
INSERT INTO tracks VALUES(941,'Malpractice',75,1,4,NULL,241371,7900683,0.98999999999999999111);
INSERT INTO tracks VALUES(942,'Kindergarten',75,1,4,NULL,270680,8853647,0.98999999999999999111);
INSERT INTO tracks VALUES(943,'Be Aggressive',75,1,4,NULL,222432,7298027,0.98999999999999999111);
INSERT INTO tracks VALUES(944,'A Small Victory',75,1,4,NULL,297168,9733572,0.98999999999999999111);
INSERT INTO tracks VALUES(945,'Crack Hitler',75,1,4,NULL,279144,9162435,0.98999999999999999111);
INSERT INTO tracks VALUES(946,'Jizzlobber',75,1,4,NULL,398341,12926140,0.98999999999999999111);
INSERT INTO tracks VALUES(947,'Midnight Cowboy',75,1,4,NULL,251924,8242626,0.98999999999999999111);
INSERT INTO tracks VALUES(948,'Easy',75,1,4,NULL,185835,6073008,0.98999999999999999111);
INSERT INTO tracks VALUES(949,'Get Out',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',137482,4524972,0.98999999999999999111);
INSERT INTO tracks VALUES(950,'Ricochet',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',269400,8808812,0.98999999999999999111);
INSERT INTO tracks VALUES(951,'Evidence',76,1,1,'Mike Bordin, Billy Gould, Mike Patton, Trey Spruance',293590,9626136,0.98999999999999999111);
INSERT INTO tracks VALUES(952,'The Gentle Art Of Making Enemies',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',209319,6908609,0.98999999999999999111);
INSERT INTO tracks VALUES(953,'Star A.D.',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',203807,6747658,0.98999999999999999111);
INSERT INTO tracks VALUES(954,'Cuckoo For Caca',76,1,1,'Mike Bordin, Billy Gould, Mike Patton, Trey Spruance',222902,7388369,0.98999999999999999111);
INSERT INTO tracks VALUES(955,'Caralho Voador',76,1,1,'Mike Bordin, Billy Gould, Mike Patton, Trey Spruance',242102,8029054,0.98999999999999999111);
INSERT INTO tracks VALUES(956,'Ugly In The Morning',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',186435,6224997,0.98999999999999999111);
INSERT INTO tracks VALUES(957,'Digging The Grave',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',185129,6109259,0.98999999999999999111);
INSERT INTO tracks VALUES(958,'Take This Bottle',76,1,1,'Mike Bordin, Billy Gould, Mike Patton, Trey Spruance',298997,9779971,0.98999999999999999111);
INSERT INTO tracks VALUES(959,'King For A Day',76,1,1,'Mike Bordin, Billy Gould, Mike Patton, Trey Spruance',395859,13163733,0.98999999999999999111);
INSERT INTO tracks VALUES(960,'What A Day',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',158275,5203430,0.98999999999999999111);
INSERT INTO tracks VALUES(961,'The Last To Know',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',267833,8736776,0.98999999999999999111);
INSERT INTO tracks VALUES(962,'Just A Man',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',336666,11031254,0.98999999999999999111);
INSERT INTO tracks VALUES(963,'Absolute Zero',76,1,1,'Mike Bordin, Billy Gould, Mike Patton',181995,5929427,0.98999999999999999111);
INSERT INTO tracks VALUES(964,'From Out Of Nowhere',77,1,4,'Faith No More',202527,6587802,0.98999999999999999111);
INSERT INTO tracks VALUES(965,'Epic',77,1,4,'Faith No More',294008,9631296,0.98999999999999999111);
INSERT INTO tracks VALUES(966,'Falling To Pieces',77,1,4,'Faith No More',316055,10333123,0.98999999999999999111);
INSERT INTO tracks VALUES(967,'Surprise! You''re Dead!',77,1,4,'Faith No More',147226,4823036,0.98999999999999999111);
INSERT INTO tracks VALUES(968,'Zombie Eaters',77,1,4,'Faith No More',360881,11835367,0.98999999999999999111);
INSERT INTO tracks VALUES(969,'The Real Thing',77,1,4,'Faith No More',493635,16233080,0.98999999999999999111);
INSERT INTO tracks VALUES(970,'Underwater Love',77,1,4,'Faith No More',231993,7634387,0.98999999999999999111);
INSERT INTO tracks VALUES(971,'The Morning After',77,1,4,'Faith No More',223764,7355898,0.98999999999999999111);
INSERT INTO tracks VALUES(972,'Woodpecker From Mars',77,1,4,'Faith No More',340532,11174250,0.98999999999999999111);
INSERT INTO tracks VALUES(973,'War Pigs',77,1,4,'Tony Iommi, Bill Ward, Geezer Butler, Ozzy Osbourne',464770,15267802,0.98999999999999999111);
INSERT INTO tracks VALUES(974,'Edge Of The World',77,1,4,'Faith No More',250357,8235607,0.98999999999999999111);
INSERT INTO tracks VALUES(975,'Deixa Entrar',78,1,7,NULL,33619,1095012,0.98999999999999999111);
INSERT INTO tracks VALUES(976,'Falamansa Song',78,1,7,NULL,237165,7921313,0.98999999999999999111);
INSERT INTO tracks VALUES(977,'Xote Dos Milagres',78,1,7,NULL,269557,8897778,0.98999999999999999111);
INSERT INTO tracks VALUES(978,'Rindo À Toa',78,1,7,NULL,222066,7365321,0.98999999999999999111);
INSERT INTO tracks VALUES(979,'Confidência',78,1,7,NULL,222197,7460829,0.98999999999999999111);
INSERT INTO tracks VALUES(980,'Forró De Tóquio',78,1,7,NULL,169273,5588756,0.98999999999999999111);
INSERT INTO tracks VALUES(981,'Zeca Violeiro',78,1,7,NULL,143673,4781949,0.98999999999999999111);
INSERT INTO tracks VALUES(982,'Avisa',78,1,7,NULL,355030,11844320,0.98999999999999999111);
INSERT INTO tracks VALUES(983,'Principiando/Decolagem',78,1,7,NULL,116767,3923789,0.98999999999999999111);
INSERT INTO tracks VALUES(984,'Asas',78,1,7,NULL,231915,7711669,0.98999999999999999111);
INSERT INTO tracks VALUES(985,'Medo De Escuro',78,1,7,NULL,213760,7056323,0.98999999999999999111);
INSERT INTO tracks VALUES(986,'Oração',78,1,7,NULL,271072,9003882,0.98999999999999999111);
INSERT INTO tracks VALUES(987,'Minha Gata',78,1,7,NULL,181838,6039502,0.98999999999999999111);
INSERT INTO tracks VALUES(988,'Desaforo',78,1,7,NULL,174524,5853561,0.98999999999999999111);
INSERT INTO tracks VALUES(989,'In Your Honor',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',230191,7468463,0.98999999999999999111);
INSERT INTO tracks VALUES(990,'No Way Back',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',196675,6421400,0.98999999999999999111);
INSERT INTO tracks VALUES(991,'Best Of You',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',255712,8363467,0.98999999999999999111);
INSERT INTO tracks VALUES(992,'DOA',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',252186,8232342,0.98999999999999999111);
INSERT INTO tracks VALUES(993,'Hell',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',117080,3819255,0.98999999999999999111);
INSERT INTO tracks VALUES(994,'The Last Song',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',199523,6496742,0.98999999999999999111);
INSERT INTO tracks VALUES(995,'Free Me',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',278700,9109340,0.98999999999999999111);
INSERT INTO tracks VALUES(996,'Resolve',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',288731,9416186,0.98999999999999999111);
INSERT INTO tracks VALUES(997,'The Deepest Blues Are Black',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',238419,7735473,0.98999999999999999111);
INSERT INTO tracks VALUES(998,'End Over End',79,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett',352078,11395296,0.98999999999999999111);
INSERT INTO tracks VALUES(999,'Still',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',313182,10323157,0.98999999999999999111);
INSERT INTO tracks VALUES(1000,'What If I Do?',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',302994,9929799,0.98999999999999999111);
INSERT INTO tracks VALUES(1001,'Miracle',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',209684,6877994,0.98999999999999999111);
INSERT INTO tracks VALUES(1002,'Another Round',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',265848,8752670,0.98999999999999999111);
INSERT INTO tracks VALUES(1003,'Friend Of A Friend',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',193280,6355088,0.98999999999999999111);
INSERT INTO tracks VALUES(1004,'Over And Out',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',316264,10428382,0.98999999999999999111);
INSERT INTO tracks VALUES(1005,'On The Mend',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',271908,9071997,0.98999999999999999111);
INSERT INTO tracks VALUES(1006,'Virginia Moon',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',229198,7494639,0.98999999999999999111);
INSERT INTO tracks VALUES(1007,'Cold Day In The Sun',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',200724,6596617,0.98999999999999999111);
INSERT INTO tracks VALUES(1008,'Razor',80,1,1,'Dave Grohl, Taylor Hawkins, Nate Mendel, Chris Shiflett/FOO FIGHTERS',293276,9721373,0.98999999999999999111);
INSERT INTO tracks VALUES(1009,'All My Life',81,1,4,'Foo Fighters',263653,8665545,0.98999999999999999111);
INSERT INTO tracks VALUES(1010,'Low',81,1,4,'Foo Fighters',268120,8847196,0.98999999999999999111);
INSERT INTO tracks VALUES(1011,'Have It All',81,1,4,'Foo Fighters',298057,9729292,0.98999999999999999111);
INSERT INTO tracks VALUES(1012,'Times Like These',81,1,4,'Foo Fighters',266370,8624691,0.98999999999999999111);
INSERT INTO tracks VALUES(1013,'Disenchanted Lullaby',81,1,4,'Foo Fighters',273528,8919111,0.98999999999999999111);
INSERT INTO tracks VALUES(1014,'Tired Of You',81,1,4,'Foo Fighters',311353,10094743,0.98999999999999999111);
INSERT INTO tracks VALUES(1015,'Halo',81,1,4,'Foo Fighters',306442,10026371,0.98999999999999999111);
INSERT INTO tracks VALUES(1016,'Lonely As You',81,1,4,'Foo Fighters',277185,9022628,0.98999999999999999111);
INSERT INTO tracks VALUES(1017,'Overdrive',81,1,4,'Foo Fighters',270550,8793187,0.98999999999999999111);
INSERT INTO tracks VALUES(1018,'Burn Away',81,1,4,'Foo Fighters',298396,9678073,0.98999999999999999111);
INSERT INTO tracks VALUES(1019,'Come Back',81,1,4,'Foo Fighters',469968,15371980,0.98999999999999999111);
INSERT INTO tracks VALUES(1020,'Doll',82,1,1,'Dave, Taylor, Nate, Chris',83487,2702572,0.98999999999999999111);
INSERT INTO tracks VALUES(1021,'Monkey Wrench',82,1,1,'Dave, Taylor, Nate, Chris',231523,7527531,0.98999999999999999111);
INSERT INTO tracks VALUES(1022,'Hey, Johnny Park!',82,1,1,'Dave, Taylor, Nate, Chris',248528,8079480,0.98999999999999999111);
INSERT INTO tracks VALUES(1023,'My Poor Brain',82,1,1,'Dave, Taylor, Nate, Chris',213446,6973746,0.98999999999999999111);
INSERT INTO tracks VALUES(1024,'Wind Up',82,1,1,'Dave, Taylor, Nate, Chris',152163,4950667,0.98999999999999999111);
INSERT INTO tracks VALUES(1025,'Up In Arms',82,1,1,'Dave, Taylor, Nate, Chris',135732,4406227,0.98999999999999999111);
INSERT INTO tracks VALUES(1026,'My Hero',82,1,1,'Dave, Taylor, Nate, Chris',260101,8472365,0.98999999999999999111);
INSERT INTO tracks VALUES(1027,'See You',82,1,1,'Dave, Taylor, Nate, Chris',146782,4888173,0.98999999999999999111);
INSERT INTO tracks VALUES(1028,'Enough Space',82,1,1,'Dave Grohl',157387,5169280,0.98999999999999999111);
INSERT INTO tracks VALUES(1029,'February Stars',82,1,1,'Dave, Taylor, Nate, Chris',289306,9344875,0.98999999999999999111);
INSERT INTO tracks VALUES(1030,'Everlong',82,1,1,'Dave Grohl',250749,8270816,0.98999999999999999111);
INSERT INTO tracks VALUES(1031,'Walking After You',82,1,1,'Dave Grohl',303856,9898992,0.98999999999999999111);
INSERT INTO tracks VALUES(1032,'New Way Home',82,1,1,'Dave, Taylor, Nate, Chris',342230,11205664,0.98999999999999999111);
INSERT INTO tracks VALUES(1033,'My Way',83,1,12,'claude françois/gilles thibault/jacques revaux/paul anka',275879,8928684,0.98999999999999999111);
INSERT INTO tracks VALUES(1034,'Strangers In The Night',83,1,12,'berthold kaempfert/charles singleton/eddie snyder',155794,5055295,0.98999999999999999111);
INSERT INTO tracks VALUES(1035,'New York, New York',83,1,12,'fred ebb/john kander',206001,6707993,0.98999999999999999111);
INSERT INTO tracks VALUES(1036,'I Get A Kick Out Of You',83,1,12,'cole porter',194429,6332441,0.98999999999999999111);
INSERT INTO tracks VALUES(1037,'Something Stupid',83,1,12,'carson c. parks',158615,5210643,0.98999999999999999111);
INSERT INTO tracks VALUES(1038,'Moon River',83,1,12,'henry mancini/johnny mercer',198922,6395808,0.98999999999999999111);
INSERT INTO tracks VALUES(1039,'What Now My Love',83,1,12,'carl sigman/gilbert becaud/pierre leroyer',149995,4913383,0.98999999999999999111);
INSERT INTO tracks VALUES(1040,'Summer Love',83,1,12,'hans bradtke/heinz meier/johnny mercer',174994,5693242,0.98999999999999999111);
INSERT INTO tracks VALUES(1041,'For Once In My Life',83,1,12,'orlando murden/ronald miller',171154,5557537,0.98999999999999999111);
INSERT INTO tracks VALUES(1042,'Love And Marriage',83,1,12,'jimmy van heusen/sammy cahn',89730,2930596,0.98999999999999999111);
INSERT INTO tracks VALUES(1043,'They Can''t Take That Away From Me',83,1,12,'george gershwin/ira gershwin',161227,5240043,0.98999999999999999111);
INSERT INTO tracks VALUES(1044,'My Kind Of Town',83,1,12,'jimmy van heusen/sammy cahn',188499,6119915,0.98999999999999999111);
INSERT INTO tracks VALUES(1045,'Fly Me To The Moon',83,1,12,'bart howard',149263,4856954,0.98999999999999999111);
INSERT INTO tracks VALUES(1046,'I''ve Got You Under My Skin',83,1,12,'cole porter',210808,6883787,0.98999999999999999111);
INSERT INTO tracks VALUES(1047,'The Best Is Yet To Come',83,1,12,'carolyn leigh/cy coleman',173583,5633730,0.98999999999999999111);
INSERT INTO tracks VALUES(1048,'It Was A Very Good Year',83,1,12,'ervin drake',266605,8554066,0.98999999999999999111);
INSERT INTO tracks VALUES(1049,'Come Fly With Me',83,1,12,'jimmy van heusen/sammy cahn',190458,6231029,0.98999999999999999111);
INSERT INTO tracks VALUES(1050,'That''s Life',83,1,12,'dean kay thompson/kelly gordon',187010,6095727,0.98999999999999999111);
INSERT INTO tracks VALUES(1051,'The Girl From Ipanema',83,1,12,'antonio carlos jobim/norman gimbel/vinicius de moraes',193750,6410674,0.98999999999999999111);
INSERT INTO tracks VALUES(1052,'The Lady Is A Tramp',83,1,12,'lorenz hart/richard rodgers',184111,5987372,0.98999999999999999111);
INSERT INTO tracks VALUES(1053,'Bad, Bad Leroy Brown',83,1,12,'jim croce',169900,5548581,0.98999999999999999111);
INSERT INTO tracks VALUES(1054,'Mack The Knife',83,1,12,'bert brecht/kurt weill/marc blitzstein',292075,9541052,0.98999999999999999111);
INSERT INTO tracks VALUES(1055,'Loves Been Good To Me',83,1,12,'rod mckuen',203964,6645365,0.98999999999999999111);
INSERT INTO tracks VALUES(1056,'L.A. Is My Lady',83,1,12,'alan bergman/marilyn bergman/peggy lipton jones/quincy jones',193175,6378511,0.98999999999999999111);
INSERT INTO tracks VALUES(1057,'Entrando Na Sua (Intro)',84,1,7,NULL,179252,5840027,0.98999999999999999111);
INSERT INTO tracks VALUES(1058,'Nervosa',84,1,7,NULL,229537,7680421,0.98999999999999999111);
INSERT INTO tracks VALUES(1059,'Funk De Bamba (Com Fernanda Abreu)',84,1,7,NULL,237191,7866165,0.98999999999999999111);
INSERT INTO tracks VALUES(1060,'Call Me At Cleo´s',84,1,7,NULL,236617,7920510,0.98999999999999999111);
INSERT INTO tracks VALUES(1061,'Olhos Coloridos (Com Sandra De Sá)',84,1,7,NULL,321332,10567404,0.98999999999999999111);
INSERT INTO tracks VALUES(1062,'Zambação',84,1,7,NULL,301113,10030604,0.98999999999999999111);
INSERT INTO tracks VALUES(1063,'Funk Hum',84,1,7,NULL,244453,8084475,0.98999999999999999111);
INSERT INTO tracks VALUES(1064,'Forty Days (Com DJ Hum)',84,1,7,NULL,221727,7347172,0.98999999999999999111);
INSERT INTO tracks VALUES(1065,'Balada Da Paula',84,1,7,'Emerson Villani',322821,10603717,0.98999999999999999111);
INSERT INTO tracks VALUES(1066,'Dujji',84,1,7,NULL,324597,10833935,0.98999999999999999111);
INSERT INTO tracks VALUES(1067,'Meu Guarda-Chuva',84,1,7,NULL,248528,8216625,0.98999999999999999111);
INSERT INTO tracks VALUES(1068,'Motéis',84,1,7,NULL,213498,7041077,0.98999999999999999111);
INSERT INTO tracks VALUES(1069,'Whistle Stop',84,1,7,NULL,526132,17533664,0.98999999999999999111);
INSERT INTO tracks VALUES(1070,'16 Toneladas',84,1,7,NULL,191634,6390885,0.98999999999999999111);
INSERT INTO tracks VALUES(1071,'Divirta-Se (Saindo Da Sua)',84,1,7,NULL,74919,2439206,0.98999999999999999111);
INSERT INTO tracks VALUES(1072,'Forty Days Instrumental',84,1,7,NULL,292493,9584317,0.98999999999999999111);
INSERT INTO tracks VALUES(1073,'Óia Eu Aqui De Novo',85,1,10,NULL,219454,7469735,0.98999999999999999111);
INSERT INTO tracks VALUES(1074,'Baião Da Penha',85,1,10,NULL,247928,8393047,0.98999999999999999111);
INSERT INTO tracks VALUES(1075,'Esperando Na Janela',85,1,10,'Manuca/Raimundinho DoAcordion/Targino Godim',261041,8660617,0.98999999999999999111);
INSERT INTO tracks VALUES(1076,'Juazeiro',85,1,10,'Humberto Teixeira/Luiz Gonzaga',222275,7349779,0.98999999999999999111);
INSERT INTO tracks VALUES(1077,'Último Pau-De-Arara',85,1,10,'Corumbá/José Gumarães/Venancio',200437,6638563,0.98999999999999999111);
INSERT INTO tracks VALUES(1078,'Asa Branca',85,1,10,'Humberto Teixeira/Luiz Gonzaga',217051,7387183,0.98999999999999999111);
INSERT INTO tracks VALUES(1079,'Qui Nem Jiló',85,1,10,'Humberto Teixeira/Luiz Gonzaga',204695,6937472,0.98999999999999999111);
INSERT INTO tracks VALUES(1080,'Assum Preto',85,1,10,'Humberto Teixeira/Luiz Gonzaga',199653,6625000,0.98999999999999999111);
INSERT INTO tracks VALUES(1081,'Pau-De-Arara',85,1,10,'Guio De Morais E Seus "Parentes"/Luiz Gonzaga',191660,6340649,0.98999999999999999111);
INSERT INTO tracks VALUES(1082,'A Volta Da Asa Branca',85,1,10,'Luiz Gonzaga/Zé Dantas',271020,9098093,0.98999999999999999111);
INSERT INTO tracks VALUES(1083,'O Amor Daqui De Casa',85,1,10,'Gilberto Gil',148636,4888292,0.98999999999999999111);
INSERT INTO tracks VALUES(1084,'As Pegadas Do Amor',85,1,10,'Gilberto Gil',209136,6899062,0.98999999999999999111);
INSERT INTO tracks VALUES(1085,'Lamento Sertanejo',85,1,10,'Dominguinhos/Gilberto Gil',260963,8518290,0.98999999999999999111);
INSERT INTO tracks VALUES(1086,'Casinha Feliz',85,1,10,'Gilberto Gil',32287,1039615,0.98999999999999999111);
INSERT INTO tracks VALUES(1087,'Introdução (Live)',86,1,7,NULL,154096,5227579,0.98999999999999999111);
INSERT INTO tracks VALUES(1088,'Palco (Live)',86,1,7,NULL,238315,8026622,0.98999999999999999111);
INSERT INTO tracks VALUES(1089,'Is This Love (Live)',86,1,7,NULL,295262,9819759,0.98999999999999999111);
INSERT INTO tracks VALUES(1090,'Stir It Up (Live)',86,1,7,NULL,282409,9594738,0.98999999999999999111);
INSERT INTO tracks VALUES(1091,'Refavela (Live)',86,1,7,NULL,236695,7985305,0.98999999999999999111);
INSERT INTO tracks VALUES(1092,'Vendedor De Caranguejo (Live)',86,1,7,NULL,248842,8358128,0.98999999999999999111);
INSERT INTO tracks VALUES(1093,'Quanta (Live)',86,1,7,NULL,357485,11774865,0.98999999999999999111);
INSERT INTO tracks VALUES(1094,'Estrela (Live)',86,1,7,NULL,285309,9436411,0.98999999999999999111);
INSERT INTO tracks VALUES(1095,'Pela Internet (Live)',86,1,7,NULL,263471,8804401,0.98999999999999999111);
INSERT INTO tracks VALUES(1096,'Cérebro Eletrônico (Live)',86,1,7,NULL,231627,7805352,0.98999999999999999111);
INSERT INTO tracks VALUES(1097,'Opachorô (Live)',86,1,7,NULL,259526,8596384,0.98999999999999999111);
INSERT INTO tracks VALUES(1098,'Copacabana (Live)',86,1,7,NULL,289671,9673672,0.98999999999999999111);
INSERT INTO tracks VALUES(1099,'A Novidade (Live)',86,1,7,NULL,316969,10508000,0.98999999999999999111);
INSERT INTO tracks VALUES(1100,'Ghandi (Live)',86,1,7,NULL,222458,7481950,0.98999999999999999111);
INSERT INTO tracks VALUES(1101,'De Ouro E Marfim (Live)',86,1,7,NULL,234971,7838453,0.98999999999999999111);
INSERT INTO tracks VALUES(1102,'Doce De Carnaval (Candy All)',87,1,2,NULL,356101,11998470,0.98999999999999999111);
INSERT INTO tracks VALUES(1103,'Lamento De Carnaval',87,1,2,NULL,294530,9819276,0.98999999999999999111);
INSERT INTO tracks VALUES(1104,'Pretinha',87,1,2,NULL,265273,8914579,0.98999999999999999111);
INSERT INTO tracks VALUES(1105,'A Novidade',73,1,7,'Gilberto Gil',324780,10765600,0.98999999999999999111);
INSERT INTO tracks VALUES(1106,'Tenho Sede',73,1,7,'Gilberto Gil',261616,8708114,0.98999999999999999111);
INSERT INTO tracks VALUES(1107,'Refazenda',73,1,7,'Gilberto Gil',218305,7237784,0.98999999999999999111);
INSERT INTO tracks VALUES(1108,'Realce',73,1,7,'Gilberto Gil',264489,8847612,0.98999999999999999111);
INSERT INTO tracks VALUES(1109,'Esotérico',73,1,7,'Gilberto Gil',317779,10530533,0.98999999999999999111);
INSERT INTO tracks VALUES(1110,'Drão',73,1,7,'Gilberto Gil',301453,9931950,0.98999999999999999111);
INSERT INTO tracks VALUES(1111,'A Paz',73,1,7,'Gilberto Gil',293093,9593064,0.98999999999999999111);
INSERT INTO tracks VALUES(1112,'Beira Mar',73,1,7,'Gilberto Gil',295444,9597994,0.98999999999999999111);
INSERT INTO tracks VALUES(1113,'Sampa',73,1,7,'Gilberto Gil',225697,7469905,0.98999999999999999111);
INSERT INTO tracks VALUES(1114,'Parabolicamará',73,1,7,'Gilberto Gil',284943,9543435,0.98999999999999999111);
INSERT INTO tracks VALUES(1115,'Tempo Rei',73,1,7,'Gilberto Gil',302733,10019269,0.98999999999999999111);
INSERT INTO tracks VALUES(1116,'Expresso 2222',73,1,7,'Gilberto Gil',284760,9690577,0.98999999999999999111);
INSERT INTO tracks VALUES(1117,'Aquele Abraço',73,1,7,'Gilberto Gil',263993,8805003,0.98999999999999999111);
INSERT INTO tracks VALUES(1118,'Palco',73,1,7,'Gilberto Gil',270550,9049901,0.98999999999999999111);
INSERT INTO tracks VALUES(1119,'Toda Menina Baiana',73,1,7,'Gilberto Gil',278177,9351000,0.98999999999999999111);
INSERT INTO tracks VALUES(1120,'Sítio Do Pica-Pau Amarelo',73,1,7,'Gilberto Gil',218070,7217955,0.98999999999999999111);
INSERT INTO tracks VALUES(1121,'Straight Out Of Line',88,1,3,'Sully Erna',259213,8511877,0.98999999999999999111);
INSERT INTO tracks VALUES(1122,'Faceless',88,1,3,'Sully Erna',216006,6992417,0.98999999999999999111);
INSERT INTO tracks VALUES(1123,'Changes',88,1,3,'Sully Erna; Tony Rombola',260022,8455835,0.98999999999999999111);
INSERT INTO tracks VALUES(1124,'Make Me Believe',88,1,3,'Sully Erna',248607,8075050,0.98999999999999999111);
INSERT INTO tracks VALUES(1125,'I Stand Alone',88,1,3,'Sully Erna',246125,8017041,0.98999999999999999111);
INSERT INTO tracks VALUES(1126,'Re-Align',88,1,3,'Sully Erna',260884,8513891,0.98999999999999999111);
INSERT INTO tracks VALUES(1127,'I Fucking Hate You',88,1,3,'Sully Erna',247170,8059642,0.98999999999999999111);
INSERT INTO tracks VALUES(1128,'Releasing The Demons',88,1,3,'Sully Erna',252760,8276372,0.98999999999999999111);
INSERT INTO tracks VALUES(1129,'Dead And Broken',88,1,3,'Sully Erna',251454,8206611,0.98999999999999999111);
INSERT INTO tracks VALUES(1130,'I Am',88,1,3,'Sully Erna',239516,7803270,0.98999999999999999111);
INSERT INTO tracks VALUES(1131,'The Awakening',88,1,3,'Sully Erna',89547,3035251,0.98999999999999999111);
INSERT INTO tracks VALUES(1132,'Serenity',88,1,3,'Sully Erna; Tony Rombola',274834,9172976,0.98999999999999999111);
INSERT INTO tracks VALUES(1133,'American Idiot',89,1,4,'Billie Joe Armstrong, Mike Dirnt, Tré Cool',174419,5705793,0.98999999999999999111);
INSERT INTO tracks VALUES(1134,'Jesus Of Suburbia / City Of The Damned / I Don''t Care / Dearly Beloved / Tales Of Another Broken Home',89,1,4,'Billie Joe Armstrong/Green Day',548336,17875209,0.98999999999999999111);
INSERT INTO tracks VALUES(1135,'Holiday',89,1,4,'Billie Joe Armstrong, Mike Dirnt, Tré Cool',232724,7599602,0.98999999999999999111);
INSERT INTO tracks VALUES(1136,'Boulevard Of Broken Dreams',89,1,4,'Mike Dint, Billie Joe, Tré Cool',260858,8485122,0.98999999999999999111);
INSERT INTO tracks VALUES(1137,'Are We The Waiting',89,1,4,'Green Day',163004,5328329,0.98999999999999999111);
INSERT INTO tracks VALUES(1138,'St. Jimmy',89,1,4,'Green Day',175307,5716589,0.98999999999999999111);
INSERT INTO tracks VALUES(1139,'Give Me Novacaine',89,1,4,'Green Day',205871,6752485,0.98999999999999999111);
INSERT INTO tracks VALUES(1140,'She''s A Rebel',89,1,4,'Green Day',120528,3901226,0.98999999999999999111);
INSERT INTO tracks VALUES(1141,'Extraordinary Girl',89,1,4,'Green Day',214021,6975177,0.98999999999999999111);
INSERT INTO tracks VALUES(1142,'Letterbomb',89,1,4,'Green Day',246151,7980902,0.98999999999999999111);
INSERT INTO tracks VALUES(1143,'Wake Me Up When September Ends',89,1,4,'Mike Dint, Billie Joe, Tré Cool',285753,9325597,0.98999999999999999111);
INSERT INTO tracks VALUES(1144,'Homecoming / The Death Of St. Jimmy / East 12th St. / Nobody Likes You / Rock And Roll Girlfriend / We''re Coming Home Again',89,1,4,'Mike Dirnt/Tré Cool',558602,18139840,0.98999999999999999111);
INSERT INTO tracks VALUES(1145,'Whatsername',89,1,4,'Green Day',252316,8244843,0.98999999999999999111);
INSERT INTO tracks VALUES(1146,'Welcome to the Jungle',90,2,1,NULL,273552,4538451,0.98999999999999999111);
INSERT INTO tracks VALUES(1147,'It''s So Easy',90,2,1,NULL,202824,3394019,0.98999999999999999111);
INSERT INTO tracks VALUES(1148,'Nightrain',90,2,1,NULL,268537,4457283,0.98999999999999999111);
INSERT INTO tracks VALUES(1149,'Out Ta Get Me',90,2,1,NULL,263893,4382147,0.98999999999999999111);
INSERT INTO tracks VALUES(1150,'Mr. Brownstone',90,2,1,NULL,228924,3816323,0.98999999999999999111);
INSERT INTO tracks VALUES(1151,'Paradise City',90,2,1,NULL,406347,6687123,0.98999999999999999111);
INSERT INTO tracks VALUES(1152,'My Michelle',90,2,1,NULL,219961,3671299,0.98999999999999999111);
INSERT INTO tracks VALUES(1153,'Think About You',90,2,1,NULL,231640,3860275,0.98999999999999999111);
INSERT INTO tracks VALUES(1154,'Sweet Child O'' Mine',90,2,1,NULL,356424,5879347,0.98999999999999999111);
INSERT INTO tracks VALUES(1155,'You''re Crazy',90,2,1,NULL,197135,3301971,0.98999999999999999111);
INSERT INTO tracks VALUES(1156,'Anything Goes',90,2,1,NULL,206400,3451891,0.98999999999999999111);
INSERT INTO tracks VALUES(1157,'Rocket Queen',90,2,1,NULL,375349,6185539,0.98999999999999999111);
INSERT INTO tracks VALUES(1158,'Right Next Door to Hell',91,2,1,NULL,182321,3175950,0.98999999999999999111);
INSERT INTO tracks VALUES(1159,'Dust N'' Bones',91,2,1,NULL,298374,5053742,0.98999999999999999111);
INSERT INTO tracks VALUES(1160,'Live and Let Die',91,2,1,NULL,184016,3203390,0.98999999999999999111);
INSERT INTO tracks VALUES(1161,'Don''t Cry (Original)',91,2,1,NULL,284744,4833259,0.98999999999999999111);
INSERT INTO tracks VALUES(1162,'Perfect Crime',91,2,1,NULL,143637,2550030,0.98999999999999999111);
INSERT INTO tracks VALUES(1163,'You Ain''t the First',91,2,1,NULL,156268,2754414,0.98999999999999999111);
INSERT INTO tracks VALUES(1164,'Bad Obsession',91,2,1,NULL,328282,5537678,0.98999999999999999111);
INSERT INTO tracks VALUES(1165,'Back off Bitch',91,2,1,NULL,303436,5135662,0.98999999999999999111);
INSERT INTO tracks VALUES(1166,'Double Talkin'' Jive',91,2,1,NULL,203637,3520862,0.98999999999999999111);
INSERT INTO tracks VALUES(1167,'November Rain',91,2,1,NULL,537540,8923566,0.98999999999999999111);
INSERT INTO tracks VALUES(1168,'The Garden',91,2,1,NULL,322175,5438862,0.98999999999999999111);
INSERT INTO tracks VALUES(1169,'Garden of Eden',91,2,1,NULL,161539,2839694,0.98999999999999999111);
INSERT INTO tracks VALUES(1170,'Don''t Damn Me',91,2,1,NULL,318901,5385886,0.98999999999999999111);
INSERT INTO tracks VALUES(1171,'Bad Apples',91,2,1,NULL,268351,4567966,0.98999999999999999111);
INSERT INTO tracks VALUES(1172,'Dead Horse',91,2,1,NULL,257600,4394014,0.98999999999999999111);
INSERT INTO tracks VALUES(1173,'Coma',91,2,1,NULL,616511,10201342,0.98999999999999999111);
INSERT INTO tracks VALUES(1174,'Civil War',92,1,3,'Duff McKagan/Slash/W. Axl Rose',461165,15046579,0.98999999999999999111);
INSERT INTO tracks VALUES(1175,'14 Years',92,1,3,'Izzy Stradlin''/W. Axl Rose',261355,8543664,0.98999999999999999111);
INSERT INTO tracks VALUES(1176,'Yesterdays',92,1,3,'Billy/Del James/W. Axl Rose/West Arkeen',196205,6398489,0.98999999999999999111);
INSERT INTO tracks VALUES(1177,'Knockin'' On Heaven''s Door',92,1,3,'Bob Dylan',336457,10986716,0.98999999999999999111);
INSERT INTO tracks VALUES(1178,'Get In The Ring',92,1,3,'Duff McKagan/Slash/W. Axl Rose',341054,11134105,0.98999999999999999111);
INSERT INTO tracks VALUES(1179,'Shotgun Blues',92,1,3,'W. Axl Rose',203206,6623916,0.98999999999999999111);
INSERT INTO tracks VALUES(1180,'Breakdown',92,1,3,'W. Axl Rose',424960,13978284,0.98999999999999999111);
INSERT INTO tracks VALUES(1181,'Pretty Tied Up',92,1,3,'Izzy Stradlin''',287477,9408754,0.98999999999999999111);
INSERT INTO tracks VALUES(1182,'Locomotive',92,1,3,'Slash/W. Axl Rose',522396,17236842,0.98999999999999999111);
INSERT INTO tracks VALUES(1183,'So Fine',92,1,3,'Duff McKagan',246491,8039484,0.98999999999999999111);
INSERT INTO tracks VALUES(1184,'Estranged',92,1,3,'W. Axl Rose',563800,18343787,0.98999999999999999111);
INSERT INTO tracks VALUES(1185,'You Could Be Mine',92,1,3,'Izzy Stradlin''/W. Axl Rose',343875,11207355,0.98999999999999999111);
INSERT INTO tracks VALUES(1186,'Don''t Cry',92,1,3,'Izzy Stradlin''/W. Axl Rose',284238,9222458,0.98999999999999999111);
INSERT INTO tracks VALUES(1187,'My World',92,1,3,'W. Axl Rose',84532,2764045,0.98999999999999999111);
INSERT INTO tracks VALUES(1188,'Colibri',93,1,2,'Richard Bull',361012,12055329,0.98999999999999999111);
INSERT INTO tracks VALUES(1189,'Love Is The Colour',93,1,2,'R. Carless',251585,8419165,0.98999999999999999111);
INSERT INTO tracks VALUES(1190,'Magnetic Ocean',93,1,2,'Patrick Claher/Richard Bull',321123,10720741,0.98999999999999999111);
INSERT INTO tracks VALUES(1191,'Deep Waters',93,1,2,'Richard Bull',396460,13075359,0.98999999999999999111);
INSERT INTO tracks VALUES(1192,'L''Arc En Ciel De Miles',93,1,2,'Kevin Robinson/Richard Bull',242390,8053997,0.98999999999999999111);
INSERT INTO tracks VALUES(1193,'Gypsy',93,1,2,'Kevin Robinson',330997,11083374,0.98999999999999999111);
INSERT INTO tracks VALUES(1194,'Journey Into Sunlight',93,1,2,'Jean Paul Maunick',249756,8241177,0.98999999999999999111);
INSERT INTO tracks VALUES(1195,'Sunchild',93,1,2,'Graham Harvey',259970,8593143,0.98999999999999999111);
INSERT INTO tracks VALUES(1196,'Millenium',93,1,2,'Maxton Gig Beesley Jnr.',379167,12511939,0.98999999999999999111);
INSERT INTO tracks VALUES(1197,'Thinking ''Bout Tomorrow',93,1,2,'Fayyaz Virgi/Richard Bull',355395,11865384,0.98999999999999999111);
INSERT INTO tracks VALUES(1198,'Jacob''s Ladder',93,1,2,'Julian Crampton',367647,12201595,0.98999999999999999111);
INSERT INTO tracks VALUES(1199,'She Wears Black',93,1,2,'G Harvey/R Hope-Taylor',528666,17617944,0.98999999999999999111);
INSERT INTO tracks VALUES(1200,'Dark Side Of The Cog',93,1,2,'Jean Paul Maunick',377155,12491122,0.98999999999999999111);
INSERT INTO tracks VALUES(1201,'Different World',94,2,1,NULL,258692,4383764,0.98999999999999999111);
INSERT INTO tracks VALUES(1202,'These Colours Don''t Run',94,2,1,NULL,412152,6883500,0.98999999999999999111);
INSERT INTO tracks VALUES(1203,'Brighter Than a Thousand Suns',94,2,1,NULL,526255,8721490,0.98999999999999999111);
INSERT INTO tracks VALUES(1204,'The Pilgrim',94,2,1,NULL,307593,5172144,0.98999999999999999111);
INSERT INTO tracks VALUES(1205,'The Longest Day',94,2,1,NULL,467810,7785748,0.98999999999999999111);
INSERT INTO tracks VALUES(1206,'Out of the Shadows',94,2,1,NULL,336896,5647303,0.98999999999999999111);
INSERT INTO tracks VALUES(1207,'The Reincarnation of Benjamin Breeg',94,2,1,NULL,442106,7367736,0.98999999999999999111);
INSERT INTO tracks VALUES(1208,'For the Greater Good of God',94,2,1,NULL,564893,9367328,0.98999999999999999111);
INSERT INTO tracks VALUES(1209,'Lord of Light',94,2,1,NULL,444614,7393698,0.98999999999999999111);
INSERT INTO tracks VALUES(1210,'The Legacy',94,2,1,NULL,562966,9314287,0.98999999999999999111);
INSERT INTO tracks VALUES(1211,'Hallowed Be Thy Name (Live) [Non Album Bonus Track]',94,2,1,NULL,431262,7205816,0.98999999999999999111);
INSERT INTO tracks VALUES(1212,'The Number Of The Beast',95,1,3,'Steve Harris',294635,4718897,0.98999999999999999111);
INSERT INTO tracks VALUES(1213,'The Trooper',95,1,3,'Steve Harris',235311,3766272,0.98999999999999999111);
INSERT INTO tracks VALUES(1214,'Prowler',95,1,3,'Steve Harris',255634,4091904,0.98999999999999999111);
INSERT INTO tracks VALUES(1215,'Transylvania',95,1,3,'Steve Harris',265874,4255744,0.98999999999999999111);
INSERT INTO tracks VALUES(1216,'Remember Tomorrow',95,1,3,'Paul Di''Anno/Steve Harris',352731,5648438,0.98999999999999999111);
INSERT INTO tracks VALUES(1217,'Where Eagles Dare',95,1,3,'Steve Harris',289358,4630528,0.98999999999999999111);
INSERT INTO tracks VALUES(1218,'Sanctuary',95,1,3,'David Murray/Paul Di''Anno/Steve Harris',293250,4694016,0.98999999999999999111);
INSERT INTO tracks VALUES(1219,'Running Free',95,1,3,'Paul Di''Anno/Steve Harris',228937,3663872,0.98999999999999999111);
INSERT INTO tracks VALUES(1220,'Run To The Hilss',95,1,3,'Steve Harris',237557,3803136,0.98999999999999999111);
INSERT INTO tracks VALUES(1221,'2 Minutes To Midnight',95,1,3,'Adrian Smith/Bruce Dickinson',337423,5400576,0.98999999999999999111);
INSERT INTO tracks VALUES(1222,'Iron Maiden',95,1,3,'Steve Harris',324623,5195776,0.98999999999999999111);
INSERT INTO tracks VALUES(1223,'Hallowed Be Thy Name',95,1,3,'Steve Harris',471849,7550976,0.98999999999999999111);
INSERT INTO tracks VALUES(1224,'Be Quick Or Be Dead',96,1,3,'Bruce Dickinson/Janick Gers',196911,3151872,0.98999999999999999111);
INSERT INTO tracks VALUES(1225,'From Here To Eternity',96,1,3,'Steve Harris',259866,4159488,0.98999999999999999111);
INSERT INTO tracks VALUES(1226,'Can I Play With Madness',96,1,3,'Adrian Smith/Bruce Dickinson/Steve Harris',282488,4521984,0.98999999999999999111);
INSERT INTO tracks VALUES(1227,'Wasting Love',96,1,3,'Bruce Dickinson/Janick Gers',347846,5566464,0.98999999999999999111);
INSERT INTO tracks VALUES(1228,'Tailgunner',96,1,3,'Bruce Dickinson/Steve Harris',249469,3993600,0.98999999999999999111);
INSERT INTO tracks VALUES(1229,'The Evil That Men Do',96,1,3,'Adrian Smith/Bruce Dickinson/Steve Harris',325929,5216256,0.98999999999999999111);
INSERT INTO tracks VALUES(1230,'Afraid To Shoot Strangers',96,1,3,'Steve Harris',407980,6529024,0.98999999999999999111);
INSERT INTO tracks VALUES(1231,'Bring Your Daughter... To The Slaughter',96,1,3,'Bruce Dickinson',317727,5085184,0.98999999999999999111);
INSERT INTO tracks VALUES(1232,'Heaven Can Wait',96,1,3,'Steve Harris',448574,7178240,0.98999999999999999111);
INSERT INTO tracks VALUES(1233,'The Clairvoyant',96,1,3,'Steve Harris',269871,4319232,0.98999999999999999111);
INSERT INTO tracks VALUES(1234,'Fear Of The Dark',96,1,3,'Steve Harris',431333,6906078,0.98999999999999999111);
INSERT INTO tracks VALUES(1235,'The Wicker Man',97,1,1,'Adrian Smith/Bruce Dickinson/Steve Harris',275539,11022464,0.98999999999999999111);
INSERT INTO tracks VALUES(1236,'Ghost Of The Navigator',97,1,1,'Bruce Dickinson/Janick Gers/Steve Harris',410070,16404608,0.98999999999999999111);
INSERT INTO tracks VALUES(1237,'Brave New World',97,1,1,'Bruce Dickinson/David Murray/Steve Harris',378984,15161472,0.98999999999999999111);
INSERT INTO tracks VALUES(1238,'Blood Brothers',97,1,1,'Steve Harris',434442,17379456,0.98999999999999999111);
INSERT INTO tracks VALUES(1239,'The Mercenary',97,1,1,'Janick Gers/Steve Harris',282488,11300992,0.98999999999999999111);
INSERT INTO tracks VALUES(1240,'Dream Of Mirrors',97,1,1,'Janick Gers/Steve Harris',561162,22448256,0.98999999999999999111);
INSERT INTO tracks VALUES(1241,'The Fallen Angel',97,1,1,'Adrian Smith/Steve Harris',240718,9629824,0.98999999999999999111);
INSERT INTO tracks VALUES(1242,'The Nomad',97,1,1,'David Murray/Steve Harris',546115,21846144,0.98999999999999999111);
INSERT INTO tracks VALUES(1243,'Out Of The Silent Planet',97,1,1,'Bruce Dickinson/Janick Gers/Steve Harris',385541,15423616,0.98999999999999999111);
INSERT INTO tracks VALUES(1244,'The Thin Line Between Love & Hate',97,1,1,'David Murray/Steve Harris',506801,20273280,0.98999999999999999111);
INSERT INTO tracks VALUES(1245,'Wildest Dreams',98,1,13,'Adrian Smith/Steve Harris',232777,9312384,0.98999999999999999111);
INSERT INTO tracks VALUES(1246,'Rainmaker',98,1,13,'Bruce Dickinson/David Murray/Steve Harris',228623,9146496,0.98999999999999999111);
INSERT INTO tracks VALUES(1247,'No More Lies',98,1,13,'Steve Harris',441782,17672320,0.98999999999999999111);
INSERT INTO tracks VALUES(1248,'Montsegur',98,1,13,'Bruce Dickinson/Janick Gers/Steve Harris',350484,14020736,0.98999999999999999111);
INSERT INTO tracks VALUES(1249,'Dance Of Death',98,1,13,'Janick Gers/Steve Harris',516649,20670727,0.98999999999999999111);
INSERT INTO tracks VALUES(1250,'Gates Of Tomorrow',98,1,13,'Bruce Dickinson/Janick Gers/Steve Harris',312032,12482688,0.98999999999999999111);
INSERT INTO tracks VALUES(1251,'New Frontier',98,1,13,'Adrian Smith/Bruce Dickinson/Nicko McBrain',304509,12181632,0.98999999999999999111);
INSERT INTO tracks VALUES(1252,'Paschendale',98,1,13,'Adrian Smith/Steve Harris',508107,20326528,0.98999999999999999111);
INSERT INTO tracks VALUES(1253,'Face In The Sand',98,1,13,'Adrian Smith/Bruce Dickinson/Steve Harris',391105,15648948,0.98999999999999999111);
INSERT INTO tracks VALUES(1254,'Age Of Innocence',98,1,13,'David Murray/Steve Harris',370468,14823478,0.98999999999999999111);
INSERT INTO tracks VALUES(1255,'Journeyman',98,1,13,'Bruce Dickinson/David Murray/Steve Harris',427023,17082496,0.98999999999999999111);
INSERT INTO tracks VALUES(1256,'Be Quick Or Be Dead',99,1,1,'Bruce Dickinson/Janick Gers',204512,8181888,0.98999999999999999111);
INSERT INTO tracks VALUES(1257,'From Here To Eternity',99,1,1,'Steve Harris',218357,8739038,0.98999999999999999111);
INSERT INTO tracks VALUES(1258,'Afraid To Shoot Strangers',99,1,1,'Steve Harris',416496,16664589,0.98999999999999999111);
INSERT INTO tracks VALUES(1259,'Fear Is The Key',99,1,1,'Bruce Dickinson/Janick Gers',335307,13414528,0.98999999999999999111);
INSERT INTO tracks VALUES(1260,'Childhood''s End',99,1,1,'Steve Harris',280607,11225216,0.98999999999999999111);
INSERT INTO tracks VALUES(1261,'Wasting Love',99,1,1,'Bruce Dickinson/Janick Gers',350981,14041216,0.98999999999999999111);
INSERT INTO tracks VALUES(1262,'The Fugitive',99,1,1,'Steve Harris',294112,11765888,0.98999999999999999111);
INSERT INTO tracks VALUES(1263,'Chains Of Misery',99,1,1,'Bruce Dickinson/David Murray',217443,8700032,0.98999999999999999111);
INSERT INTO tracks VALUES(1264,'The Apparition',99,1,1,'Janick Gers/Steve Harris',234605,9386112,0.98999999999999999111);
INSERT INTO tracks VALUES(1265,'Judas Be My Guide',99,1,1,'Bruce Dickinson/David Murray',188786,7553152,0.98999999999999999111);
INSERT INTO tracks VALUES(1266,'Weekend Warrior',99,1,1,'Janick Gers/Steve Harris',339748,13594678,0.98999999999999999111);
INSERT INTO tracks VALUES(1267,'Fear Of The Dark',99,1,1,'Steve Harris',436976,17483789,0.98999999999999999111);
INSERT INTO tracks VALUES(1268,'01 - Prowler',100,1,6,'Steve Harris',236173,5668992,0.98999999999999999111);
INSERT INTO tracks VALUES(1269,'02 - Sanctuary',100,1,6,'David Murray/Paul Di''Anno/Steve Harris',196284,4712576,0.98999999999999999111);
INSERT INTO tracks VALUES(1270,'03 - Remember Tomorrow',100,1,6,'Harris/Paul Di´Anno',328620,7889024,0.98999999999999999111);
INSERT INTO tracks VALUES(1271,'04 - Running Free',100,1,6,'Harris/Paul Di´Anno',197276,4739122,0.98999999999999999111);
INSERT INTO tracks VALUES(1272,'05 - Phantom of the Opera',100,1,6,'Steve Harris',428016,10276872,0.98999999999999999111);
INSERT INTO tracks VALUES(1273,'06 - Transylvania',100,1,6,'Steve Harris',259343,6226048,0.98999999999999999111);
INSERT INTO tracks VALUES(1274,'07 - Strange World',100,1,6,'Steve Harris',332460,7981184,0.98999999999999999111);
INSERT INTO tracks VALUES(1275,'08 - Charlotte the Harlot',100,1,6,'Murray  Dave',252708,6066304,0.98999999999999999111);
INSERT INTO tracks VALUES(1276,'09 - Iron Maiden',100,1,6,'Steve Harris',216058,5189891,0.98999999999999999111);
INSERT INTO tracks VALUES(1277,'The Ides Of March',101,1,13,'Steve Harris',105926,2543744,0.98999999999999999111);
INSERT INTO tracks VALUES(1278,'Wrathchild',101,1,13,'Steve Harris',174471,4188288,0.98999999999999999111);
INSERT INTO tracks VALUES(1279,'Murders In The Rue Morgue',101,1,13,'Steve Harris',258377,6205786,0.98999999999999999111);
INSERT INTO tracks VALUES(1280,'Another Life',101,1,13,'Steve Harris',203049,4874368,0.98999999999999999111);
INSERT INTO tracks VALUES(1281,'Genghis Khan',101,1,13,'Steve Harris',187141,4493440,0.98999999999999999111);
INSERT INTO tracks VALUES(1282,'Innocent Exile',101,1,13,'Di´Anno/Harris',232515,5584861,0.98999999999999999111);
INSERT INTO tracks VALUES(1283,'Killers',101,1,13,'Steve Harris',300956,7227440,0.98999999999999999111);
INSERT INTO tracks VALUES(1284,'Prodigal Son',101,1,13,'Steve Harris',372349,8937600,0.98999999999999999111);
INSERT INTO tracks VALUES(1285,'Purgatory',101,1,13,'Steve Harris',200150,4804736,0.98999999999999999111);
INSERT INTO tracks VALUES(1286,'Drifter',101,1,13,'Steve Harris',288757,6934660,0.98999999999999999111);
INSERT INTO tracks VALUES(1287,'Intro- Churchill S Speech',102,1,13,NULL,48013,1154488,0.98999999999999999111);
INSERT INTO tracks VALUES(1288,'Aces High',102,1,13,NULL,276375,6635187,0.98999999999999999111);
INSERT INTO tracks VALUES(1289,'2 Minutes To Midnight',102,1,3,'Smith/Dickinson',366550,8799380,0.98999999999999999111);
INSERT INTO tracks VALUES(1290,'The Trooper',102,1,3,'Harris',268878,6455255,0.98999999999999999111);
INSERT INTO tracks VALUES(1291,'Revelations',102,1,3,'Dickinson',371826,8926021,0.98999999999999999111);
INSERT INTO tracks VALUES(1292,'Flight Of Icarus',102,1,3,'Smith/Dickinson',229982,5521744,0.98999999999999999111);
INSERT INTO tracks VALUES(1293,'Rime Of The Ancient Mariner',102,1,3,NULL,789472,18949518,0.98999999999999999111);
INSERT INTO tracks VALUES(1294,'Powerslave',102,1,3,NULL,454974,10921567,0.98999999999999999111);
INSERT INTO tracks VALUES(1295,'The Number Of The Beast',102,1,3,'Harris',275121,6605094,0.98999999999999999111);
INSERT INTO tracks VALUES(1296,'Hallowed Be Thy Name',102,1,3,'Harris',451422,10836304,0.98999999999999999111);
INSERT INTO tracks VALUES(1297,'Iron Maiden',102,1,3,'Harris',261955,6289117,0.98999999999999999111);
INSERT INTO tracks VALUES(1298,'Run To The Hills',102,1,3,'Harris',231627,5561241,0.98999999999999999111);
INSERT INTO tracks VALUES(1299,'Running Free',102,1,3,'Harris/Di Anno',204617,4912986,0.98999999999999999111);
INSERT INTO tracks VALUES(1300,'Wrathchild',102,1,13,'Steve Harris',183666,4410181,0.98999999999999999111);
INSERT INTO tracks VALUES(1301,'Acacia Avenue',102,1,13,NULL,379872,9119118,0.98999999999999999111);
INSERT INTO tracks VALUES(1302,'Children Of The Damned',102,1,13,'Steve Harris',278177,6678446,0.98999999999999999111);
INSERT INTO tracks VALUES(1303,'Die With Your Boots On',102,1,13,'Adrian Smith/Bruce Dickinson/Steve Harris',314174,7542367,0.98999999999999999111);
INSERT INTO tracks VALUES(1304,'Phantom Of The Opera',102,1,13,'Steve Harris',441155,10589917,0.98999999999999999111);
INSERT INTO tracks VALUES(1305,'Be Quick Or Be Dead',103,1,1,NULL,233142,5599853,0.98999999999999999111);
INSERT INTO tracks VALUES(1306,'The Number Of The Beast',103,1,1,NULL,294008,7060625,0.98999999999999999111);
INSERT INTO tracks VALUES(1307,'Wrathchild',103,1,1,NULL,174106,4182963,0.98999999999999999111);
INSERT INTO tracks VALUES(1308,'From Here To Eternity',103,1,1,NULL,284447,6831163,0.98999999999999999111);
INSERT INTO tracks VALUES(1309,'Can I Play With Madness',103,1,1,NULL,213106,5118995,0.98999999999999999111);
INSERT INTO tracks VALUES(1310,'Wasting Love',103,1,1,NULL,336953,8091301,0.98999999999999999111);
INSERT INTO tracks VALUES(1311,'Tailgunner',103,1,1,NULL,247640,5947795,0.98999999999999999111);
INSERT INTO tracks VALUES(1312,'The Evil That Men Do',103,1,1,NULL,478145,11479913,0.98999999999999999111);
INSERT INTO tracks VALUES(1313,'Afraid To Shoot Strangers',103,1,1,NULL,412525,9905048,0.98999999999999999111);
INSERT INTO tracks VALUES(1314,'Fear Of The Dark',103,1,1,NULL,431542,10361452,0.98999999999999999111);
INSERT INTO tracks VALUES(1315,'Bring Your Daughter... To The Slaughter...',104,1,1,NULL,376711,9045532,0.98999999999999999111);
INSERT INTO tracks VALUES(1316,'The Clairvoyant',104,1,1,NULL,262426,6302648,0.98999999999999999111);
INSERT INTO tracks VALUES(1317,'Heaven Can Wait',104,1,1,NULL,440555,10577743,0.98999999999999999111);
INSERT INTO tracks VALUES(1318,'Run To The Hills',104,1,1,NULL,235859,5665052,0.98999999999999999111);
INSERT INTO tracks VALUES(1319,'2 Minutes To Midnight',104,1,1,'Adrian Smith/Bruce Dickinson',338233,8122030,0.98999999999999999111);
INSERT INTO tracks VALUES(1320,'Iron Maiden',104,1,1,NULL,494602,11874875,0.98999999999999999111);
INSERT INTO tracks VALUES(1321,'Hallowed Be Thy Name',104,1,1,NULL,447791,10751410,0.98999999999999999111);
INSERT INTO tracks VALUES(1322,'The Trooper',104,1,1,NULL,232672,5588560,0.98999999999999999111);
INSERT INTO tracks VALUES(1323,'Sanctuary',104,1,1,NULL,318511,7648679,0.98999999999999999111);
INSERT INTO tracks VALUES(1324,'Running Free',104,1,1,NULL,474017,11380851,0.98999999999999999111);
INSERT INTO tracks VALUES(1325,'Tailgunner',105,1,3,'Bruce Dickinson/Steve Harris',255582,4089856,0.98999999999999999111);
INSERT INTO tracks VALUES(1326,'Holy Smoke',105,1,3,'Bruce Dickinson/Steve Harris',229459,3672064,0.98999999999999999111);
INSERT INTO tracks VALUES(1327,'No Prayer For The Dying',105,1,3,'Steve Harris',263941,4225024,0.98999999999999999111);
INSERT INTO tracks VALUES(1328,'Public Enema Number One',105,1,3,'Bruce Dickinson/David Murray',254197,4071587,0.98999999999999999111);
INSERT INTO tracks VALUES(1329,'Fates Warning',105,1,3,'David Murray/Steve Harris',250853,4018088,0.98999999999999999111);
INSERT INTO tracks VALUES(1330,'The Assassin',105,1,3,'Steve Harris',258768,4141056,0.98999999999999999111);
INSERT INTO tracks VALUES(1331,'Run Silent Run Deep',105,1,3,'Bruce Dickinson/Steve Harris',275408,4407296,0.98999999999999999111);
INSERT INTO tracks VALUES(1332,'Hooks In You',105,1,3,'Adrian Smith/Bruce Dickinson',247510,3960832,0.98999999999999999111);
INSERT INTO tracks VALUES(1333,'Bring Your Daughter... ...To The Slaughter',105,1,3,'Bruce Dickinson',284238,4548608,0.98999999999999999111);
INSERT INTO tracks VALUES(1334,'Mother Russia',105,1,3,'Steve Harris',332617,5322752,0.98999999999999999111);
INSERT INTO tracks VALUES(1335,'Where Eagles Dare',106,1,3,'Steve Harris',369554,5914624,0.98999999999999999111);
INSERT INTO tracks VALUES(1336,'Revelations',106,1,3,'Bruce Dickinson',408607,6539264,0.98999999999999999111);
INSERT INTO tracks VALUES(1337,'Flight Of The Icarus',106,1,3,'Adrian Smith/Bruce Dickinson',230269,3686400,0.98999999999999999111);
INSERT INTO tracks VALUES(1338,'Die With Your Boots On',106,1,3,'Adrian Smith/Bruce Dickinson/Steve Harris',325694,5212160,0.98999999999999999111);
INSERT INTO tracks VALUES(1339,'The Trooper',106,1,3,'Steve Harris',251454,4024320,0.98999999999999999111);
INSERT INTO tracks VALUES(1340,'Still Life',106,1,3,'David Murray/Steve Harris',294347,4710400,0.98999999999999999111);
INSERT INTO tracks VALUES(1341,'Quest For Fire',106,1,3,'Steve Harris',221309,3543040,0.98999999999999999111);
INSERT INTO tracks VALUES(1342,'Sun And Steel',106,1,3,'Adrian Smith/Bruce Dickinson',206367,3306324,0.98999999999999999111);
INSERT INTO tracks VALUES(1343,'To Tame A Land',106,1,3,'Steve Harris',445283,7129264,0.98999999999999999111);
INSERT INTO tracks VALUES(1344,'Aces High',107,1,3,'Harris',269531,6472088,0.98999999999999999111);
INSERT INTO tracks VALUES(1345,'2 Minutes To Midnight',107,1,3,'Smith/Dickinson',359810,8638809,0.98999999999999999111);
INSERT INTO tracks VALUES(1346,'Losfer Words',107,1,3,'Steve Harris',252891,6074756,0.98999999999999999111);
INSERT INTO tracks VALUES(1347,'Flash of The Blade',107,1,3,'Dickinson',242729,5828861,0.98999999999999999111);
INSERT INTO tracks VALUES(1348,'Duelists',107,1,3,'Steve Harris',366471,8800686,0.98999999999999999111);
INSERT INTO tracks VALUES(1349,'Back in the Village',107,1,3,'Dickinson/Smith',320548,7696518,0.98999999999999999111);
INSERT INTO tracks VALUES(1350,'Powerslave',107,1,3,'Dickinson',407823,9791106,0.98999999999999999111);
INSERT INTO tracks VALUES(1351,'Rime of the Ancient Mariner',107,1,3,'Harris',816509,19599577,0.98999999999999999111);
INSERT INTO tracks VALUES(1352,'Intro',108,1,3,NULL,115931,4638848,0.98999999999999999111);
INSERT INTO tracks VALUES(1353,'The Wicker Man',108,1,3,'Adrian Smith/Bruce Dickinson/Steve Harris',281782,11272320,0.98999999999999999111);
INSERT INTO tracks VALUES(1354,'Ghost Of The Navigator',108,1,3,'Bruce Dickinson/Janick Gers/Steve Harris',408607,16345216,0.98999999999999999111);
INSERT INTO tracks VALUES(1355,'Brave New World',108,1,3,'Bruce Dickinson/David Murray/Steve Harris',366785,14676148,0.98999999999999999111);
INSERT INTO tracks VALUES(1356,'Wrathchild',108,1,3,'Steve Harris',185808,7434368,0.98999999999999999111);
INSERT INTO tracks VALUES(1357,'2 Minutes To Midnight',108,1,3,'Adrian Smith/Bruce Dickinson',386821,15474816,0.98999999999999999111);
INSERT INTO tracks VALUES(1358,'Blood Brothers',108,1,3,'Steve Harris',435513,17422464,0.98999999999999999111);
INSERT INTO tracks VALUES(1359,'Sign Of The Cross',108,1,3,'Steve Harris',649116,25966720,0.98999999999999999111);
INSERT INTO tracks VALUES(1360,'The Mercenary',108,1,3,'Janick Gers/Steve Harris',282697,11309184,0.98999999999999999111);
INSERT INTO tracks VALUES(1361,'The Trooper',108,1,3,'Steve Harris',273528,10942592,0.98999999999999999111);
INSERT INTO tracks VALUES(1362,'Dream Of Mirrors',109,1,1,'Janick Gers/Steve Harris',578324,23134336,0.98999999999999999111);
INSERT INTO tracks VALUES(1363,'The Clansman',109,1,1,'Steve Harris',559203,22370432,0.98999999999999999111);
INSERT INTO tracks VALUES(1364,'The Evil That Men Do',109,1,3,'Adrian Smith/Bruce Dickinson/Steve Harris',280737,11231360,0.98999999999999999111);
INSERT INTO tracks VALUES(1365,'Fear Of The Dark',109,1,1,'Steve Harris',460695,18430080,0.98999999999999999111);
INSERT INTO tracks VALUES(1366,'Iron Maiden',109,1,1,'Steve Harris',351869,14076032,0.98999999999999999111);
INSERT INTO tracks VALUES(1367,'The Number Of The Beast',109,1,1,'Steve Harris',300434,12022107,0.98999999999999999111);
INSERT INTO tracks VALUES(1368,'Hallowed Be Thy Name',109,1,1,'Steve Harris',443977,17760384,0.98999999999999999111);
INSERT INTO tracks VALUES(1369,'Sanctuary',109,1,1,'David Murray/Paul Di''Anno/Steve Harris',317335,12695680,0.98999999999999999111);
INSERT INTO tracks VALUES(1370,'Run To The Hills',109,1,1,'Steve Harris',292179,11688064,0.98999999999999999111);
INSERT INTO tracks VALUES(1371,'Moonchild',110,1,3,'Adrian Smith; Bruce Dickinson',340767,8179151,0.98999999999999999111);
INSERT INTO tracks VALUES(1372,'Infinite Dreams',110,1,3,'Steve Harris',369005,8858669,0.98999999999999999111);
INSERT INTO tracks VALUES(1373,'Can I Play With Madness',110,1,3,'Adrian Smith; Bruce Dickinson; Steve Harris',211043,5067867,0.98999999999999999111);
INSERT INTO tracks VALUES(1374,'The Evil That Men Do',110,1,3,'Adrian Smith; Bruce Dickinson; Steve Harris',273998,6578930,0.98999999999999999111);
INSERT INTO tracks VALUES(1375,'Seventh Son of a Seventh Son',110,1,3,'Steve Harris',593580,14249000,0.98999999999999999111);
INSERT INTO tracks VALUES(1376,'The Prophecy',110,1,3,'Dave Murray; Steve Harris',305475,7334450,0.98999999999999999111);
INSERT INTO tracks VALUES(1377,'The Clairvoyant',110,1,3,'Adrian Smith; Bruce Dickinson; Steve Harris',267023,6411510,0.98999999999999999111);
INSERT INTO tracks VALUES(1378,'Only the Good Die Young',110,1,3,'Bruce Dickinson; Harris',280894,6744431,0.98999999999999999111);
INSERT INTO tracks VALUES(1379,'Caught Somewhere in Time',111,1,3,'Steve Harris',445779,10701149,0.98999999999999999111);
INSERT INTO tracks VALUES(1380,'Wasted Years',111,1,3,'Adrian Smith',307565,7384358,0.98999999999999999111);
INSERT INTO tracks VALUES(1381,'Sea of Madness',111,1,3,'Adrian Smith',341995,8210695,0.98999999999999999111);
INSERT INTO tracks VALUES(1382,'Heaven Can Wait',111,1,3,'Steve Harris',441417,10596431,0.98999999999999999111);
INSERT INTO tracks VALUES(1383,'Stranger in a Strange Land',111,1,3,'Adrian Smith',344502,8270899,0.98999999999999999111);
INSERT INTO tracks VALUES(1384,'Alexander the Great',111,1,3,'Steve Harris',515631,12377742,0.98999999999999999111);
INSERT INTO tracks VALUES(1385,'De Ja Vu',111,1,3,'David Murray/Steve Harris',296176,7113035,0.98999999999999999111);
INSERT INTO tracks VALUES(1386,'The Loneliness of the Long Dis',111,1,3,'Steve Harris',391314,9393598,0.98999999999999999111);
INSERT INTO tracks VALUES(1387,'22 Acacia Avenue',112,1,3,'Adrian Smith/Steve Harris',395572,5542516,0.98999999999999999111);
INSERT INTO tracks VALUES(1388,'Children of the Damned',112,1,3,'Steve Harris',274364,3845631,0.98999999999999999111);
INSERT INTO tracks VALUES(1389,'Gangland',112,1,3,'Adrian Smith/Clive Burr/Steve Harris',228440,3202866,0.98999999999999999111);
INSERT INTO tracks VALUES(1390,'Hallowed Be Thy Name',112,1,3,'Steve Harris',428669,6006107,0.98999999999999999111);
INSERT INTO tracks VALUES(1391,'Invaders',112,1,3,'Steve Harris',203180,2849181,0.98999999999999999111);
INSERT INTO tracks VALUES(1392,'Run to the Hills',112,1,3,'Steve Harris',228884,3209124,0.98999999999999999111);
INSERT INTO tracks VALUES(1393,'The Number Of The Beast',112,1,1,'Steve Harris',293407,11737216,0.98999999999999999111);
INSERT INTO tracks VALUES(1394,'The Prisoner',112,1,3,'Adrian Smith/Steve Harris',361299,5062906,0.98999999999999999111);
INSERT INTO tracks VALUES(1395,'Sign Of The Cross',113,1,1,'Steve Harris',678008,27121792,0.98999999999999999111);
INSERT INTO tracks VALUES(1396,'Lord Of The Flies',113,1,1,'Janick Gers/Steve Harris',303699,12148864,0.98999999999999999111);
INSERT INTO tracks VALUES(1397,'Man On The Edge',113,1,1,'Blaze Bayley/Janick Gers',253413,10137728,0.98999999999999999111);
INSERT INTO tracks VALUES(1398,'Fortunes Of War',113,1,1,'Steve Harris',443977,17760384,0.98999999999999999111);
INSERT INTO tracks VALUES(1399,'Look For The Truth',113,1,1,'Blaze Bayley/Janick Gers/Steve Harris',310230,12411008,0.98999999999999999111);
INSERT INTO tracks VALUES(1400,'The Aftermath',113,1,1,'Blaze Bayley/Janick Gers/Steve Harris',380786,15233152,0.98999999999999999111);
INSERT INTO tracks VALUES(1401,'Judgement Of Heaven',113,1,1,'Steve Harris',312476,12501120,0.98999999999999999111);
INSERT INTO tracks VALUES(1402,'Blood On The World''s Hands',113,1,1,'Steve Harris',357799,14313600,0.98999999999999999111);
INSERT INTO tracks VALUES(1403,'The Edge Of Darkness',113,1,1,'Blaze Bayley/Janick Gers/Steve Harris',399333,15974528,0.98999999999999999111);
INSERT INTO tracks VALUES(1404,'2 A.M.',113,1,1,'Blaze Bayley/Janick Gers/Steve Harris',337658,13511087,0.98999999999999999111);
INSERT INTO tracks VALUES(1405,'The Unbeliever',113,1,1,'Janick Gers/Steve Harris',490422,19617920,0.98999999999999999111);
INSERT INTO tracks VALUES(1406,'Futureal',114,1,1,'Blaze Bayley/Steve Harris',175777,7032960,0.98999999999999999111);
INSERT INTO tracks VALUES(1407,'The Angel And The Gambler',114,1,1,'Steve Harris',592744,23711872,0.98999999999999999111);
INSERT INTO tracks VALUES(1408,'Lightning Strikes Twice',114,1,1,'David Murray/Steve Harris',290377,11616384,0.98999999999999999111);
INSERT INTO tracks VALUES(1409,'The Clansman',114,1,1,'Steve Harris',539689,21592327,0.98999999999999999111);
INSERT INTO tracks VALUES(1410,'When Two Worlds Collide',114,1,1,'Blaze Bayley/David Murray/Steve Harris',377312,15093888,0.98999999999999999111);
INSERT INTO tracks VALUES(1411,'The Educated Fool',114,1,1,'Steve Harris',404767,16191616,0.98999999999999999111);
INSERT INTO tracks VALUES(1412,'Don''t Look To The Eyes Of A Stranger',114,1,1,'Steve Harris',483657,19347584,0.98999999999999999111);
INSERT INTO tracks VALUES(1413,'Como Estais Amigos',114,1,1,'Blaze Bayley/Janick Gers',330292,13213824,0.98999999999999999111);
INSERT INTO tracks VALUES(1414,'Please Please Please',115,1,14,'James Brown/Johnny Terry',165067,5394585,0.98999999999999999111);
INSERT INTO tracks VALUES(1415,'Think',115,1,14,'Lowman Pauling',166739,5513208,0.98999999999999999111);
INSERT INTO tracks VALUES(1416,'Night Train',115,1,14,'Jimmy Forrest/Lewis C. Simpkins/Oscar Washington',212401,7027377,0.98999999999999999111);
INSERT INTO tracks VALUES(1417,'Out Of Sight',115,1,14,'Ted Wright',143725,4711323,0.98999999999999999111);
INSERT INTO tracks VALUES(1418,'Papa''s Got A Brand New Bag Pt.1',115,1,14,'James Brown',127399,4174420,0.98999999999999999111);
INSERT INTO tracks VALUES(1419,'I Got You (I Feel Good)',115,1,14,'James Brown',167392,5468472,0.98999999999999999111);
INSERT INTO tracks VALUES(1420,'It''s A Man''s Man''s Man''s World',115,1,14,'Betty Newsome/James Brown',168228,5541611,0.98999999999999999111);
INSERT INTO tracks VALUES(1421,'Cold Sweat',115,1,14,'Alfred Ellis/James Brown',172408,5643213,0.98999999999999999111);
INSERT INTO tracks VALUES(1422,'Say It Loud, I''m Black And I''m Proud Pt.1',115,1,14,'Alfred Ellis/James Brown',167392,5478117,0.98999999999999999111);
INSERT INTO tracks VALUES(1423,'Get Up (I Feel Like Being A) Sex Machine',115,1,14,'Bobby Byrd/James Brown/Ron Lenhoff',316551,10498031,0.98999999999999999111);
INSERT INTO tracks VALUES(1424,'Hey America',115,1,14,'Addie William Jones/Nat Jones',218226,7187857,0.98999999999999999111);
INSERT INTO tracks VALUES(1425,'Make It Funky Pt.1',115,1,14,'Charles Bobbitt/James Brown',196231,6507782,0.98999999999999999111);
INSERT INTO tracks VALUES(1426,'I''m A Greedy Man Pt.1',115,1,14,'Charles Bobbitt/James Brown',217730,7251211,0.98999999999999999111);
INSERT INTO tracks VALUES(1427,'Get On The Good Foot',115,1,14,'Fred Wesley/James Brown/Joseph Mims',215902,7182736,0.98999999999999999111);
INSERT INTO tracks VALUES(1428,'Get Up Offa That Thing',115,1,14,'Deanna Brown/Deidra Jenkins/Yamma Brown',250723,8355989,0.98999999999999999111);
INSERT INTO tracks VALUES(1429,'It''s Too Funky In Here',115,1,14,'Brad Shapiro/George Jackson/Robert Miller/Walter Shaw',239072,7973979,0.98999999999999999111);
INSERT INTO tracks VALUES(1430,'Living In America',115,1,14,'Charlie Midnight/Dan Hartman',282880,9432346,0.98999999999999999111);
INSERT INTO tracks VALUES(1431,'I''m Real',115,1,14,'Full Force/James Brown',334236,11183457,0.98999999999999999111);
INSERT INTO tracks VALUES(1432,'Hot Pants Pt.1',115,1,14,'Fred Wesley/James Brown',188212,6295110,0.98999999999999999111);
INSERT INTO tracks VALUES(1433,'Soul Power (Live)',115,1,14,'James Brown',260728,8593206,0.98999999999999999111);
INSERT INTO tracks VALUES(1434,'When You Gonna Learn (Digeridoo)',116,1,1,'Jay Kay/Kay, Jay',230635,7655482,0.98999999999999999111);
INSERT INTO tracks VALUES(1435,'Too Young To Die',116,1,1,'Smith, Toby',365818,12391660,0.98999999999999999111);
INSERT INTO tracks VALUES(1436,'Hooked Up',116,1,1,'Smith, Toby',275879,9301687,0.98999999999999999111);
INSERT INTO tracks VALUES(1437,'If I Like It, I Do It',116,1,1,'Gelder, Nick van',293093,9848207,0.98999999999999999111);
INSERT INTO tracks VALUES(1438,'Music Of The Wind',116,1,1,'Smith, Toby',383033,12870239,0.98999999999999999111);
INSERT INTO tracks VALUES(1439,'Emergency On Planet Earth',116,1,1,'Smith, Toby',245263,8117218,0.98999999999999999111);
INSERT INTO tracks VALUES(1440,'Whatever It Is, I Just Can''t Stop',116,1,1,'Jay Kay/Kay, Jay',247222,8249453,0.98999999999999999111);
INSERT INTO tracks VALUES(1441,'Blow Your Mind',116,1,1,'Smith, Toby',512339,17089176,0.98999999999999999111);
INSERT INTO tracks VALUES(1442,'Revolution 1993',116,1,1,'Smith, Toby',616829,20816872,0.98999999999999999111);
INSERT INTO tracks VALUES(1443,'Didgin'' Out',116,1,1,'Buchanan, Wallis',157100,5263555,0.98999999999999999111);
INSERT INTO tracks VALUES(1444,'Canned Heat',117,1,14,'Jay Kay',331964,11042037,0.98999999999999999111);
INSERT INTO tracks VALUES(1445,'Planet Home',117,1,14,'Jay Kay/Toby Smith',284447,9566237,0.98999999999999999111);
INSERT INTO tracks VALUES(1446,'Black Capricorn Day',117,1,14,'Jay Kay',341629,11477231,0.98999999999999999111);
INSERT INTO tracks VALUES(1447,'Soul Education',117,1,14,'Jay Kay/Toby Smith',255477,8575435,0.98999999999999999111);
INSERT INTO tracks VALUES(1448,'Failling',117,1,14,'Jay Kay/Toby Smith',225227,7503999,0.98999999999999999111);
INSERT INTO tracks VALUES(1449,'Destitute Illusions',117,1,14,'Derrick McKenzie/Jay Kay/Toby Smith',340218,11452651,0.98999999999999999111);
INSERT INTO tracks VALUES(1450,'Supersonic',117,1,14,'Jay Kay',315872,10699265,0.98999999999999999111);
INSERT INTO tracks VALUES(1451,'Butterfly',117,1,14,'Jay Kay/Toby Smith',268852,8947356,0.98999999999999999111);
INSERT INTO tracks VALUES(1452,'Were Do We Go From Here',117,1,14,'Jay Kay',313626,10504158,0.98999999999999999111);
INSERT INTO tracks VALUES(1453,'King For A Day',117,1,14,'Jay Kay/Toby Smith',221544,7335693,0.98999999999999999111);
INSERT INTO tracks VALUES(1454,'Deeper Underground',117,1,14,'Toby Smith',281808,9351277,0.98999999999999999111);
INSERT INTO tracks VALUES(1455,'Just Another Story',118,1,15,'Toby Smith',529684,17582818,0.98999999999999999111);
INSERT INTO tracks VALUES(1456,'Stillness In Time',118,1,15,'Toby Smith',257097,8644290,0.98999999999999999111);
INSERT INTO tracks VALUES(1457,'Half The Man',118,1,15,'Toby Smith',289854,9577679,0.98999999999999999111);
INSERT INTO tracks VALUES(1458,'Light Years',118,1,15,'Toby Smith',354560,11796244,0.98999999999999999111);
INSERT INTO tracks VALUES(1459,'Manifest Destiny',118,1,15,'Toby Smith',382197,12676962,0.98999999999999999111);
INSERT INTO tracks VALUES(1460,'The Kids',118,1,15,'Toby Smith',309995,10334529,0.98999999999999999111);
INSERT INTO tracks VALUES(1461,'Mr. Moon',118,1,15,'Stuard Zender/Toby Smith',329534,11043559,0.98999999999999999111);
INSERT INTO tracks VALUES(1462,'Scam',118,1,15,'Stuart Zender',422321,14019705,0.98999999999999999111);
INSERT INTO tracks VALUES(1463,'Journey To Arnhemland',118,1,15,'Toby Smith/Wallis Buchanan',322455,10843832,0.98999999999999999111);
INSERT INTO tracks VALUES(1464,'Morning Glory',118,1,15,'J. Kay/Jay Kay',384130,12777210,0.98999999999999999111);
INSERT INTO tracks VALUES(1465,'Space Cowboy',118,1,15,'J. Kay/Jay Kay',385697,12906520,0.98999999999999999111);
INSERT INTO tracks VALUES(1466,'Last Chance',119,1,4,'C. Cester/C. Muncey',112352,3683130,0.98999999999999999111);
INSERT INTO tracks VALUES(1467,'Are You Gonna Be My Girl',119,1,4,'C. Muncey/N. Cester',213890,6992324,0.98999999999999999111);
INSERT INTO tracks VALUES(1468,'Rollover D.J.',119,1,4,'C. Cester/N. Cester',196702,6406517,0.98999999999999999111);
INSERT INTO tracks VALUES(1469,'Look What You''ve Done',119,1,4,'N. Cester',230974,7517083,0.98999999999999999111);
INSERT INTO tracks VALUES(1470,'Get What You Need',119,1,4,'C. Cester/C. Muncey/N. Cester',247719,8043765,0.98999999999999999111);
INSERT INTO tracks VALUES(1471,'Move On',119,1,4,'C. Cester/N. Cester',260623,8519353,0.98999999999999999111);
INSERT INTO tracks VALUES(1472,'Radio Song',119,1,4,'C. Cester/C. Muncey/N. Cester',272117,8871509,0.98999999999999999111);
INSERT INTO tracks VALUES(1473,'Get Me Outta Here',119,1,4,'C. Cester/N. Cester',176274,5729098,0.98999999999999999111);
INSERT INTO tracks VALUES(1474,'Cold Hard Bitch',119,1,4,'C. Cester/C. Muncey/N. Cester',243278,7929610,0.98999999999999999111);
INSERT INTO tracks VALUES(1475,'Come Around Again',119,1,4,'C. Muncey/N. Cester',270497,8872405,0.98999999999999999111);
INSERT INTO tracks VALUES(1476,'Take It Or Leave It',119,1,4,'C. Muncey/N. Cester',142889,4643370,0.98999999999999999111);
INSERT INTO tracks VALUES(1477,'Lazy Gun',119,1,4,'C. Cester/N. Cester',282174,9186285,0.98999999999999999111);
INSERT INTO tracks VALUES(1478,'Timothy',119,1,4,'C. Cester',270341,8856507,0.98999999999999999111);
INSERT INTO tracks VALUES(1479,'Foxy Lady',120,1,1,'Jimi Hendrix',199340,6480896,0.98999999999999999111);
INSERT INTO tracks VALUES(1480,'Manic Depression',120,1,1,'Jimi Hendrix',222302,7289272,0.98999999999999999111);
INSERT INTO tracks VALUES(1481,'Red House',120,1,1,'Jimi Hendrix',224130,7285851,0.98999999999999999111);
INSERT INTO tracks VALUES(1482,'Can You See Me',120,1,1,'Jimi Hendrix',153077,4987068,0.98999999999999999111);
INSERT INTO tracks VALUES(1483,'Love Or Confusion',120,1,1,'Jimi Hendrix',193123,6329408,0.98999999999999999111);
INSERT INTO tracks VALUES(1484,'I Don''t Live Today',120,1,1,'Jimi Hendrix',235311,7661214,0.98999999999999999111);
INSERT INTO tracks VALUES(1485,'May This Be Love',120,1,1,'Jimi Hendrix',191216,6240028,0.98999999999999999111);
INSERT INTO tracks VALUES(1486,'Fire',120,1,1,'Jimi Hendrix',164989,5383075,0.98999999999999999111);
INSERT INTO tracks VALUES(1487,'Third Stone From The Sun',120,1,1,'Jimi Hendrix',404453,13186975,0.98999999999999999111);
INSERT INTO tracks VALUES(1488,'Remember',120,1,1,'Jimi Hendrix',168150,5509613,0.98999999999999999111);
INSERT INTO tracks VALUES(1489,'Are You Experienced?',120,1,1,'Jimi Hendrix',254537,8292497,0.98999999999999999111);
INSERT INTO tracks VALUES(1490,'Hey Joe',120,1,1,'Billy Roberts',210259,6870054,0.98999999999999999111);
INSERT INTO tracks VALUES(1491,'Stone Free',120,1,1,'Jimi Hendrix',216293,7002331,0.98999999999999999111);
INSERT INTO tracks VALUES(1492,'Purple Haze',120,1,1,'Jimi Hendrix',171572,5597056,0.98999999999999999111);
INSERT INTO tracks VALUES(1493,'51st Anniversary',120,1,1,'Jimi Hendrix',196388,6398044,0.98999999999999999111);
INSERT INTO tracks VALUES(1494,'The Wind Cries Mary',120,1,1,'Jimi Hendrix',200463,6540638,0.98999999999999999111);
INSERT INTO tracks VALUES(1495,'Highway Chile',120,1,1,'Jimi Hendrix',212453,6887949,0.98999999999999999111);
INSERT INTO tracks VALUES(1496,'Surfing with the Alien',121,2,1,NULL,263707,4418504,0.98999999999999999111);
INSERT INTO tracks VALUES(1497,'Ice 9',121,2,1,NULL,239721,4036215,0.98999999999999999111);
INSERT INTO tracks VALUES(1498,'Crushing Day',121,2,1,NULL,314768,5232158,0.98999999999999999111);
INSERT INTO tracks VALUES(1499,'Always With Me, Always With You',121,2,1,NULL,202035,3435777,0.98999999999999999111);
INSERT INTO tracks VALUES(1500,'Satch Boogie',121,2,1,NULL,193560,3300654,0.98999999999999999111);
INSERT INTO tracks VALUES(1501,'Hill of the Skull',121,2,1,'J. Satriani',108435,1944738,0.98999999999999999111);
INSERT INTO tracks VALUES(1502,'Circles',121,2,1,NULL,209071,3548553,0.98999999999999999111);
INSERT INTO tracks VALUES(1503,'Lords of Karma',121,2,1,'J. Satriani',288227,4809279,0.98999999999999999111);
INSERT INTO tracks VALUES(1504,'Midnight',121,2,1,'J. Satriani',102630,1851753,0.98999999999999999111);
INSERT INTO tracks VALUES(1505,'Echo',121,2,1,'J. Satriani',337570,5595557,0.98999999999999999111);
INSERT INTO tracks VALUES(1506,'Engenho De Dentro',122,1,7,NULL,310073,10211473,0.98999999999999999111);
INSERT INTO tracks VALUES(1507,'Alcohol',122,1,7,NULL,355239,12010478,0.98999999999999999111);
INSERT INTO tracks VALUES(1508,'Mama Africa',122,1,7,NULL,283062,9488316,0.98999999999999999111);
INSERT INTO tracks VALUES(1509,'Salve Simpatia',122,1,7,NULL,343484,11314756,0.98999999999999999111);
INSERT INTO tracks VALUES(1510,'W/Brasil (Chama O Síndico)',122,1,7,NULL,317100,10599953,0.98999999999999999111);
INSERT INTO tracks VALUES(1511,'País Tropical',122,1,7,NULL,452519,14946972,0.98999999999999999111);
INSERT INTO tracks VALUES(1512,'Os Alquimistas Estão Chegando',122,1,7,NULL,367281,12304520,0.98999999999999999111);
INSERT INTO tracks VALUES(1513,'Charles Anjo 45',122,1,7,NULL,389276,13022833,0.98999999999999999111);
INSERT INTO tracks VALUES(1514,'Selassiê',122,1,7,NULL,326321,10724982,0.98999999999999999111);
INSERT INTO tracks VALUES(1515,'Menina Sarará',122,1,7,NULL,191477,6393818,0.98999999999999999111);
INSERT INTO tracks VALUES(1516,'Que Maravilha',122,1,7,NULL,338076,10996656,0.98999999999999999111);
INSERT INTO tracks VALUES(1517,'Santa Clara Clareou',122,1,7,NULL,380081,12524725,0.98999999999999999111);
INSERT INTO tracks VALUES(1518,'Filho Maravilha',122,1,7,NULL,227526,7498259,0.98999999999999999111);
INSERT INTO tracks VALUES(1519,'Taj Mahal',122,1,7,NULL,289750,9502898,0.98999999999999999111);
INSERT INTO tracks VALUES(1520,'Rapidamente',123,1,7,NULL,252238,8470107,0.98999999999999999111);
INSERT INTO tracks VALUES(1521,'As Dores do Mundo',123,1,7,'Hyldon',255477,8537092,0.98999999999999999111);
INSERT INTO tracks VALUES(1522,'Vou Pra Ai',123,1,7,NULL,300878,10053718,0.98999999999999999111);
INSERT INTO tracks VALUES(1523,'My Brother',123,1,7,NULL,253231,8431821,0.98999999999999999111);
INSERT INTO tracks VALUES(1524,'Há Quanto Tempo',123,1,7,NULL,270027,9004470,0.98999999999999999111);
INSERT INTO tracks VALUES(1525,'Vício',123,1,7,NULL,269897,8887216,0.98999999999999999111);
INSERT INTO tracks VALUES(1526,'Encontrar Alguém',123,1,7,'Marco Tulio Lara/Rogerio Flausino',224078,7437935,0.98999999999999999111);
INSERT INTO tracks VALUES(1527,'Dance Enquanto é Tempo',123,1,7,NULL,229093,7583799,0.98999999999999999111);
INSERT INTO tracks VALUES(1528,'A Tarde',123,1,7,NULL,266919,8836127,0.98999999999999999111);
INSERT INTO tracks VALUES(1529,'Always Be All Right',123,1,7,NULL,128078,4299676,0.98999999999999999111);
INSERT INTO tracks VALUES(1530,'Sem Sentido',123,1,7,NULL,250462,8292108,0.98999999999999999111);
INSERT INTO tracks VALUES(1531,'Onibusfobia',123,1,7,NULL,315977,10474904,0.98999999999999999111);
INSERT INTO tracks VALUES(1532,'Pura Elegancia',124,1,16,'João Suplicy',284107,9632269,0.98999999999999999111);
INSERT INTO tracks VALUES(1533,'Choramingando',124,1,16,'João Suplicy',190484,6400532,0.98999999999999999111);
INSERT INTO tracks VALUES(1534,'Por Merecer',124,1,16,'João Suplicy',230582,7764601,0.98999999999999999111);
INSERT INTO tracks VALUES(1535,'No Futuro',124,1,16,'João Suplicy',182308,6056200,0.98999999999999999111);
INSERT INTO tracks VALUES(1536,'Voce Inteira',124,1,16,'João Suplicy',241084,8077282,0.98999999999999999111);
INSERT INTO tracks VALUES(1537,'Cuando A Noite Vai Chegando',124,1,16,'João Suplicy',270628,9081874,0.98999999999999999111);
INSERT INTO tracks VALUES(1538,'Naquele Dia',124,1,16,'João Suplicy',251768,8452654,0.98999999999999999111);
INSERT INTO tracks VALUES(1539,'Equinocio',124,1,16,'João Suplicy',269008,8871455,0.98999999999999999111);
INSERT INTO tracks VALUES(1540,'Papelão',124,1,16,'João Suplicy',213263,7257390,0.98999999999999999111);
INSERT INTO tracks VALUES(1541,'Cuando Eu For Pro Ceu',124,1,16,'João Suplicy',118804,3948371,0.98999999999999999111);
INSERT INTO tracks VALUES(1542,'Do Nosso Amor',124,1,16,'João Suplicy',203415,6774566,0.98999999999999999111);
INSERT INTO tracks VALUES(1543,'Borogodo',124,1,16,'João Suplicy',208457,7104588,0.98999999999999999111);
INSERT INTO tracks VALUES(1544,'Cafezinho',124,1,16,'João Suplicy',180924,6031174,0.98999999999999999111);
INSERT INTO tracks VALUES(1545,'Enquanto O Dia Não Vem',124,1,16,'João Suplicy',220891,7248336,0.98999999999999999111);
INSERT INTO tracks VALUES(1546,'The Green Manalishi',125,1,3,NULL,205792,6720789,0.98999999999999999111);
INSERT INTO tracks VALUES(1547,'Living After Midnight',125,1,3,NULL,213289,7056785,0.98999999999999999111);
INSERT INTO tracks VALUES(1548,'Breaking The Law (Live)',125,1,3,NULL,144195,4728246,0.98999999999999999111);
INSERT INTO tracks VALUES(1549,'Hot Rockin''',125,1,3,NULL,197328,6509179,0.98999999999999999111);
INSERT INTO tracks VALUES(1550,'Heading Out To The Highway (Live)',125,1,3,NULL,276427,9006022,0.98999999999999999111);
INSERT INTO tracks VALUES(1551,'The Hellion',125,1,3,NULL,41900,1351993,0.98999999999999999111);
INSERT INTO tracks VALUES(1552,'Electric Eye',125,1,3,NULL,222197,7231368,0.98999999999999999111);
INSERT INTO tracks VALUES(1553,'You''ve Got Another Thing Comin''',125,1,3,NULL,305162,9962558,0.98999999999999999111);
INSERT INTO tracks VALUES(1554,'Turbo Lover',125,1,3,NULL,335542,11068866,0.98999999999999999111);
INSERT INTO tracks VALUES(1555,'Freewheel Burning',125,1,3,NULL,265952,8713599,0.98999999999999999111);
INSERT INTO tracks VALUES(1556,'Some Heads Are Gonna Roll',125,1,3,NULL,249939,8198617,0.98999999999999999111);
INSERT INTO tracks VALUES(1557,'Metal Meltdown',125,1,3,NULL,290664,9390646,0.98999999999999999111);
INSERT INTO tracks VALUES(1558,'Ram It Down',125,1,3,NULL,292179,9554023,0.98999999999999999111);
INSERT INTO tracks VALUES(1559,'Diamonds And Rust (Live)',125,1,3,NULL,219350,7163147,0.98999999999999999111);
INSERT INTO tracks VALUES(1560,'Victim Of Change (Live)',125,1,3,NULL,430942,14067512,0.98999999999999999111);
INSERT INTO tracks VALUES(1561,'Tyrant (Live)',125,1,3,NULL,282253,9190536,0.98999999999999999111);
INSERT INTO tracks VALUES(1562,'Comin'' Home',126,1,1,'Paul Stanley, Ace Frehley',172068,5661120,0.98999999999999999111);
INSERT INTO tracks VALUES(1563,'Plaster Caster',126,1,1,'Gene Simmons',198060,6528719,0.98999999999999999111);
INSERT INTO tracks VALUES(1564,'Goin'' Blind',126,1,1,'Gene Simmons, Stephen Coronel',217652,7167523,0.98999999999999999111);
INSERT INTO tracks VALUES(1565,'Do You Love Me',126,1,1,'Paul Stanley, Bob Ezrin, Kim Fowley',193619,6343111,0.98999999999999999111);
INSERT INTO tracks VALUES(1566,'Domino',126,1,1,'Gene Simmons',226377,7488191,0.98999999999999999111);
INSERT INTO tracks VALUES(1567,'Sure Know Something',126,1,1,'Paul Stanley, Vincent Poncia',254354,8375190,0.98999999999999999111);
INSERT INTO tracks VALUES(1568,'A World Without Heroes',126,1,1,'Paul Stanley, Gene Simmons, Bob Ezrin, Lewis Reed',177815,5832524,0.98999999999999999111);
INSERT INTO tracks VALUES(1569,'Rock Bottom',126,1,1,'Paul Stanley, Ace Frehley',200594,6560818,0.98999999999999999111);
INSERT INTO tracks VALUES(1570,'See You Tonight',126,1,1,'Gene Simmons',146494,4817521,0.98999999999999999111);
INSERT INTO tracks VALUES(1571,'I Still Love You',126,1,1,'Paul Stanley',369815,12086145,0.98999999999999999111);
INSERT INTO tracks VALUES(1572,'Every Time I Look At You',126,1,1,'Paul Stanley, Vincent Cusano',283898,9290948,0.98999999999999999111);
INSERT INTO tracks VALUES(1573,'2,000 Man',126,1,1,'Mick Jagger, Keith Richard',312450,10292829,0.98999999999999999111);
INSERT INTO tracks VALUES(1574,'Beth',126,1,1,'Peter Criss, Stan Penridge, Bob Ezrin',170187,5577807,0.98999999999999999111);
INSERT INTO tracks VALUES(1575,'Nothin'' To Lose',126,1,1,'Gene Simmons',222354,7351460,0.98999999999999999111);
INSERT INTO tracks VALUES(1576,'Rock And Roll All Nite',126,1,1,'Paul Stanley, Gene Simmons',259631,8549296,0.98999999999999999111);
INSERT INTO tracks VALUES(1577,'Immigrant Song',127,1,1,'Robert Plant',201247,6457766,0.98999999999999999111);
INSERT INTO tracks VALUES(1578,'Heartbreaker',127,1,1,'John Bonham/John Paul Jones/Robert Plant',316081,10179657,0.98999999999999999111);
INSERT INTO tracks VALUES(1579,'Since I''ve Been Loving You',127,1,1,'John Paul Jones/Robert Plant',416365,13471959,0.98999999999999999111);
INSERT INTO tracks VALUES(1580,'Black Dog',127,1,1,'John Paul Jones/Robert Plant',317622,10267572,0.98999999999999999111);
INSERT INTO tracks VALUES(1581,'Dazed And Confused',127,1,1,'Jimmy Page/Led Zeppelin',1116734,36052247,0.98999999999999999111);
INSERT INTO tracks VALUES(1582,'Stairway To Heaven',127,1,1,'Robert Plant',529658,17050485,0.98999999999999999111);
INSERT INTO tracks VALUES(1583,'Going To California',127,1,1,'Robert Plant',234605,7646749,0.98999999999999999111);
INSERT INTO tracks VALUES(1584,'That''s The Way',127,1,1,'Robert Plant',343431,11248455,0.98999999999999999111);
INSERT INTO tracks VALUES(1585,'Whole Lotta Love (Medley)',127,1,1,'Arthur Crudup/Bernard Besman/Bukka White/Doc Pomus/John Bonham/John Lee Hooker/John Paul Jones/Mort Shuman/Robert Plant/Willie Dixon',825103,26742545,0.98999999999999999111);
INSERT INTO tracks VALUES(1586,'Thank You',127,1,1,'Robert Plant',398262,12831826,0.98999999999999999111);
INSERT INTO tracks VALUES(1587,'We''re Gonna Groove',128,1,1,'Ben E.King/James Bethea',157570,5180975,0.98999999999999999111);
INSERT INTO tracks VALUES(1588,'Poor Tom',128,1,1,'Jimmy Page/Robert Plant',182491,6016220,0.98999999999999999111);
INSERT INTO tracks VALUES(1589,'I Can''t Quit You Baby',128,1,1,'Willie Dixon',258168,8437098,0.98999999999999999111);
INSERT INTO tracks VALUES(1590,'Walter''s Walk',128,1,1,'Jimmy Page, Robert Plant',270785,8712499,0.98999999999999999111);
INSERT INTO tracks VALUES(1591,'Ozone Baby',128,1,1,'Jimmy Page, Robert Plant',215954,7079588,0.98999999999999999111);
INSERT INTO tracks VALUES(1592,'Darlene',128,1,1,'Jimmy Page, Robert Plant, John Bonham, John Paul Jones',307226,10078197,0.98999999999999999111);
INSERT INTO tracks VALUES(1593,'Bonzo''s Montreux',128,1,1,'John Bonham',258925,8557447,0.98999999999999999111);
INSERT INTO tracks VALUES(1594,'Wearing And Tearing',128,1,1,'Jimmy Page, Robert Plant',330004,10701590,0.98999999999999999111);
INSERT INTO tracks VALUES(1595,'The Song Remains The Same',129,1,1,'Jimmy Page/Jimmy Page & Robert Plant/Robert Plant',330004,10708950,0.98999999999999999111);
INSERT INTO tracks VALUES(1596,'The Rain Song',129,1,1,'Jimmy Page/Jimmy Page & Robert Plant/Robert Plant',459180,15029875,0.98999999999999999111);
INSERT INTO tracks VALUES(1597,'Over The Hills And Far Away',129,1,1,'Jimmy Page/Jimmy Page & Robert Plant/Robert Plant',290089,9552829,0.98999999999999999111);
INSERT INTO tracks VALUES(1598,'The Crunge',129,1,1,'John Bonham/John Paul Jones',197407,6460212,0.98999999999999999111);
INSERT INTO tracks VALUES(1599,'Dancing Days',129,1,1,'Jimmy Page/Jimmy Page & Robert Plant/Robert Plant',223216,7250104,0.98999999999999999111);
INSERT INTO tracks VALUES(1600,'D''Yer Mak''er',129,1,1,'John Bonham/John Paul Jones',262948,8645935,0.98999999999999999111);
INSERT INTO tracks VALUES(1601,'No Quarter',129,1,1,'John Paul Jones',420493,13656517,0.98999999999999999111);
INSERT INTO tracks VALUES(1602,'The Ocean',129,1,1,'John Bonham/John Paul Jones',271098,8846469,0.98999999999999999111);
INSERT INTO tracks VALUES(1603,'In The Evening',130,1,1,'Jimmy Page, Robert Plant & John Paul Jones',410566,13399734,0.98999999999999999111);
INSERT INTO tracks VALUES(1604,'South Bound Saurez',130,1,1,'John Paul Jones & Robert Plant',254406,8420427,0.98999999999999999111);
INSERT INTO tracks VALUES(1605,'Fool In The Rain',130,1,1,'Jimmy Page, Robert Plant & John Paul Jones',372950,12371433,0.98999999999999999111);
INSERT INTO tracks VALUES(1606,'Hot Dog',130,1,1,'Jimmy Page & Robert Plant',197198,6536167,0.98999999999999999111);
INSERT INTO tracks VALUES(1607,'Carouselambra',130,1,1,'John Paul Jones, Jimmy Page & Robert Plant',634435,20858315,0.98999999999999999111);
INSERT INTO tracks VALUES(1608,'All My Love',130,1,1,'Robert Plant & John Paul Jones',356284,11684862,0.98999999999999999111);
INSERT INTO tracks VALUES(1609,'I''m Gonna Crawl',130,1,1,'Jimmy Page, Robert Plant & John Paul Jones',329639,10737665,0.98999999999999999111);
INSERT INTO tracks VALUES(1610,'Black Dog',131,1,1,'Jimmy Page, Robert Plant, John Paul Jones',296672,9660588,0.98999999999999999111);
INSERT INTO tracks VALUES(1611,'Rock & Roll',131,1,1,'Jimmy Page, Robert Plant, John Paul Jones, John Bonham',220917,7142127,0.98999999999999999111);
INSERT INTO tracks VALUES(1612,'The Battle Of Evermore',131,1,1,'Jimmy Page, Robert Plant',351555,11525689,0.98999999999999999111);
INSERT INTO tracks VALUES(1613,'Stairway To Heaven',131,1,1,'Jimmy Page, Robert Plant',481619,15706767,0.98999999999999999111);
INSERT INTO tracks VALUES(1614,'Misty Mountain Hop',131,1,1,'Jimmy Page, Robert Plant, John Paul Jones',278857,9092799,0.98999999999999999111);
INSERT INTO tracks VALUES(1615,'Four Sticks',131,1,1,'Jimmy Page, Robert Plant',284447,9481301,0.98999999999999999111);
INSERT INTO tracks VALUES(1616,'Going To California',131,1,1,'Jimmy Page, Robert Plant',215693,7068737,0.98999999999999999111);
INSERT INTO tracks VALUES(1617,'When The Levee Breaks',131,1,1,'Jimmy Page, Robert Plant, John Paul Jones, John Bonham, Memphis Minnie',427702,13912107,0.98999999999999999111);
INSERT INTO tracks VALUES(1618,'Good Times Bad Times',132,1,1,'Jimmy Page/John Bonham/John Paul Jones',166164,5464077,0.98999999999999999111);
INSERT INTO tracks VALUES(1619,'Babe I''m Gonna Leave You',132,1,1,'Jimmy Page/Robert Plant',401475,13189312,0.98999999999999999111);
INSERT INTO tracks VALUES(1620,'You Shook Me',132,1,1,'J. B. Lenoir/Willie Dixon',388179,12643067,0.98999999999999999111);
INSERT INTO tracks VALUES(1621,'Dazed and Confused',132,1,1,'Jimmy Page',386063,12610326,0.98999999999999999111);
INSERT INTO tracks VALUES(1622,'Your Time Is Gonna Come',132,1,1,'Jimmy Page/John Paul Jones',274860,9011653,0.98999999999999999111);
INSERT INTO tracks VALUES(1623,'Black Mountain Side',132,1,1,'Jimmy Page',132702,4440602,0.98999999999999999111);
INSERT INTO tracks VALUES(1624,'Communication Breakdown',132,1,1,'Jimmy Page/John Bonham/John Paul Jones',150230,4899554,0.98999999999999999111);
INSERT INTO tracks VALUES(1625,'I Can''t Quit You Baby',132,1,1,'Willie Dixon',282671,9252733,0.98999999999999999111);
INSERT INTO tracks VALUES(1626,'How Many More Times',132,1,1,'Jimmy Page/John Bonham/John Paul Jones',508055,16541364,0.98999999999999999111);
INSERT INTO tracks VALUES(1627,'Whole Lotta Love',133,1,1,'Jimmy Page, Robert Plant, John Paul Jones, John Bonham',334471,11026243,0.98999999999999999111);
INSERT INTO tracks VALUES(1628,'What Is And What Should Never Be',133,1,1,'Jimmy Page, Robert Plant',287973,9369385,0.98999999999999999111);
INSERT INTO tracks VALUES(1629,'The Lemon Song',133,1,1,'Jimmy Page, Robert Plant, John Paul Jones, John Bonham',379141,12463496,0.98999999999999999111);
INSERT INTO tracks VALUES(1630,'Thank You',133,1,1,'Jimmy Page, Robert Plant',287791,9337392,0.98999999999999999111);
INSERT INTO tracks VALUES(1631,'Heartbreaker',133,1,1,'Jimmy Page, Robert Plant, John Paul Jones, John Bonham',253988,8387560,0.98999999999999999111);
INSERT INTO tracks VALUES(1632,'Living Loving Maid (She''s Just A Woman)',133,1,1,'Jimmy Page, Robert Plant',159216,5219819,0.98999999999999999111);
INSERT INTO tracks VALUES(1633,'Ramble On',133,1,1,'Jimmy Page, Robert Plant',275591,9199710,0.98999999999999999111);
INSERT INTO tracks VALUES(1634,'Moby Dick',133,1,1,'John Bonham, John Paul Jones, Jimmy Page',260728,8664210,0.98999999999999999111);
INSERT INTO tracks VALUES(1635,'Bring It On Home',133,1,1,'Jimmy Page, Robert Plant',259970,8494731,0.98999999999999999111);
INSERT INTO tracks VALUES(1636,'Immigrant Song',134,1,1,'Jimmy Page, Robert Plant',144875,4786461,0.98999999999999999111);
INSERT INTO tracks VALUES(1637,'Friends',134,1,1,'Jimmy Page, Robert Plant',233560,7694220,0.98999999999999999111);
INSERT INTO tracks VALUES(1638,'Celebration Day',134,1,1,'Jimmy Page, Robert Plant, John Paul Jones',209528,6871078,0.98999999999999999111);
INSERT INTO tracks VALUES(1639,'Since I''ve Been Loving You',134,1,1,'Jimmy Page, Robert Plant, John Paul Jones',444055,14482460,0.98999999999999999111);
INSERT INTO tracks VALUES(1640,'Out On The Tiles',134,1,1,'Jimmy Page, Robert Plant, John Bonham',246047,8060350,0.98999999999999999111);
INSERT INTO tracks VALUES(1641,'Gallows Pole',134,1,1,'Traditional',296228,9757151,0.98999999999999999111);
INSERT INTO tracks VALUES(1642,'Tangerine',134,1,1,'Jimmy Page',189675,6200893,0.98999999999999999111);
INSERT INTO tracks VALUES(1643,'That''s The Way',134,1,1,'Jimmy Page, Robert Plant',337345,11202499,0.98999999999999999111);
INSERT INTO tracks VALUES(1644,'Bron-Y-Aur Stomp',134,1,1,'Jimmy Page, Robert Plant, John Paul Jones',259500,8674508,0.98999999999999999111);
INSERT INTO tracks VALUES(1645,'Hats Off To (Roy) Harper',134,1,1,'Traditional',219376,7236640,0.98999999999999999111);
INSERT INTO tracks VALUES(1646,'In The Light',135,1,1,'John Paul Jones/Robert Plant',526785,17033046,0.98999999999999999111);
INSERT INTO tracks VALUES(1647,'Bron-Yr-Aur',135,1,1,'Jimmy Page',126641,4150746,0.98999999999999999111);
INSERT INTO tracks VALUES(1648,'Down By The Seaside',135,1,1,'Robert Plant',316186,10371282,0.98999999999999999111);
INSERT INTO tracks VALUES(1649,'Ten Years Gone',135,1,1,'Robert Plant',393116,12756366,0.98999999999999999111);
INSERT INTO tracks VALUES(1650,'Night Flight',135,1,1,'John Paul Jones/Robert Plant',217547,7160647,0.98999999999999999111);
INSERT INTO tracks VALUES(1651,'The Wanton Song',135,1,1,'Robert Plant',249887,8180988,0.98999999999999999111);
INSERT INTO tracks VALUES(1652,'Boogie With Stu',135,1,1,'Ian Stewart/John Bonham/John Paul Jones/Mrs. Valens/Robert Plant',233273,7657086,0.98999999999999999111);
INSERT INTO tracks VALUES(1653,'Black Country Woman',135,1,1,'Robert Plant',273084,8951732,0.98999999999999999111);
INSERT INTO tracks VALUES(1654,'Sick Again',135,1,1,'Robert Plant',283036,9279263,0.98999999999999999111);
INSERT INTO tracks VALUES(1655,'Achilles Last Stand',136,1,1,'Jimmy Page/Robert Plant',625502,20593955,0.98999999999999999111);
INSERT INTO tracks VALUES(1656,'For Your Life',136,1,1,'Jimmy Page/Robert Plant',384391,12633382,0.98999999999999999111);
INSERT INTO tracks VALUES(1657,'Royal Orleans',136,1,1,'John Bonham/John Paul Jones',179591,5930027,0.98999999999999999111);
INSERT INTO tracks VALUES(1658,'Nobody''s Fault But Mine',136,1,1,'Jimmy Page/Robert Plant',376215,12237859,0.98999999999999999111);
INSERT INTO tracks VALUES(1659,'Candy Store Rock',136,1,1,'Jimmy Page/Robert Plant',252055,8397423,0.98999999999999999111);
INSERT INTO tracks VALUES(1660,'Hots On For Nowhere',136,1,1,'Jimmy Page/Robert Plant',284107,9342342,0.98999999999999999111);
INSERT INTO tracks VALUES(1661,'Tea For One',136,1,1,'Jimmy Page/Robert Plant',566752,18475264,0.98999999999999999111);
INSERT INTO tracks VALUES(1662,'Rock & Roll',137,1,1,'John Bonham/John Paul Jones/Robert Plant',242442,7897065,0.98999999999999999111);
INSERT INTO tracks VALUES(1663,'Celebration Day',137,1,1,'John Paul Jones/Robert Plant',230034,7478487,0.98999999999999999111);
INSERT INTO tracks VALUES(1664,'The Song Remains The Same',137,1,1,'Robert Plant',353358,11465033,0.98999999999999999111);
INSERT INTO tracks VALUES(1665,'Rain Song',137,1,1,'Robert Plant',505808,16273705,0.98999999999999999111);
INSERT INTO tracks VALUES(1666,'Dazed And Confused',137,1,1,'Jimmy Page',1612329,52490554,0.98999999999999999111);
INSERT INTO tracks VALUES(1667,'No Quarter',138,1,1,'John Paul Jones/Robert Plant',749897,24399285,0.98999999999999999111);
INSERT INTO tracks VALUES(1668,'Stairway To Heaven',138,1,1,'Robert Plant',657293,21354766,0.98999999999999999111);
INSERT INTO tracks VALUES(1669,'Moby Dick',138,1,1,'John Bonham/John Paul Jones',766354,25345841,0.98999999999999999111);
INSERT INTO tracks VALUES(1670,'Whole Lotta Love',138,1,1,'John Bonham/John Paul Jones/Robert Plant/Willie Dixon',863895,28191437,0.98999999999999999111);
INSERT INTO tracks VALUES(1671,'Natália',139,1,7,'Renato Russo',235728,7640230,0.98999999999999999111);
INSERT INTO tracks VALUES(1672,'L''Avventura',139,1,7,'Renato Russo',278256,9165769,0.98999999999999999111);
INSERT INTO tracks VALUES(1673,'Música De Trabalho',139,1,7,'Renato Russo',260231,8590671,0.98999999999999999111);
INSERT INTO tracks VALUES(1674,'Longe Do Meu Lado',139,1,7,'Renato Russo - Marcelo Bonfá',266161,8655249,0.98999999999999999111);
INSERT INTO tracks VALUES(1675,'A Via Láctea',139,1,7,'Renato Russo',280084,9234879,0.98999999999999999111);
INSERT INTO tracks VALUES(1676,'Música Ambiente',139,1,7,'Renato Russo',247614,8234388,0.98999999999999999111);
INSERT INTO tracks VALUES(1677,'Aloha',139,1,7,'Renato Russo',325955,10793301,0.98999999999999999111);
INSERT INTO tracks VALUES(1678,'Soul Parsifal',139,1,7,'Renato Russo - Marisa Monte',295053,9853589,0.98999999999999999111);
INSERT INTO tracks VALUES(1679,'Dezesseis',139,1,7,'Renato Russo',323918,10573515,0.98999999999999999111);
INSERT INTO tracks VALUES(1680,'Mil Pedaços',139,1,7,'Renato Russo',203337,6643291,0.98999999999999999111);
INSERT INTO tracks VALUES(1681,'Leila',139,1,7,'Renato Russo',323056,10608239,0.98999999999999999111);
INSERT INTO tracks VALUES(1682,'1º De Julho',139,1,7,'Renato Russo',290298,9619257,0.98999999999999999111);
INSERT INTO tracks VALUES(1683,'Esperando Por Mim',139,1,7,'Renato Russo',261668,8844133,0.98999999999999999111);
INSERT INTO tracks VALUES(1684,'Quando Você Voltar',139,1,7,'Renato Russo',173897,5781046,0.98999999999999999111);
INSERT INTO tracks VALUES(1685,'O Livro Dos Dias',139,1,7,'Renato Russo',257253,8570929,0.98999999999999999111);
INSERT INTO tracks VALUES(1686,'Será',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',148401,4826528,0.98999999999999999111);
INSERT INTO tracks VALUES(1687,'Ainda É Cedo',140,1,7,'Dado Villa-Lobos/Ico Ouro-Preto/Marcelo Bonfá',236826,7796400,0.98999999999999999111);
INSERT INTO tracks VALUES(1688,'Geração Coca-Cola',140,1,7,'Renato Russo',141453,4625731,0.98999999999999999111);
INSERT INTO tracks VALUES(1689,'Eduardo E Mônica',140,1,7,'Renato Russo',271229,9026691,0.98999999999999999111);
INSERT INTO tracks VALUES(1690,'Tempo Perdido',140,1,7,'Renato Russo',302158,9963914,0.98999999999999999111);
INSERT INTO tracks VALUES(1691,'Indios',140,1,7,'Renato Russo',258168,8610226,0.98999999999999999111);
INSERT INTO tracks VALUES(1692,'Que País É Este',140,1,7,'Renato Russo',177606,5822124,0.98999999999999999111);
INSERT INTO tracks VALUES(1693,'Faroeste Caboclo',140,1,7,'Renato Russo',543007,18092739,0.98999999999999999111);
INSERT INTO tracks VALUES(1694,'Há Tempos',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',197146,6432922,0.98999999999999999111);
INSERT INTO tracks VALUES(1695,'Pais E Filhos',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',308401,10130685,0.98999999999999999111);
INSERT INTO tracks VALUES(1696,'Meninos E Meninas',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',203781,6667802,0.98999999999999999111);
INSERT INTO tracks VALUES(1697,'Vento No Litoral',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',366445,12063806,0.98999999999999999111);
INSERT INTO tracks VALUES(1698,'Perfeição',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',276558,9258489,0.98999999999999999111);
INSERT INTO tracks VALUES(1699,'Giz',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',202213,6677671,0.98999999999999999111);
INSERT INTO tracks VALUES(1700,'Dezesseis',140,1,7,'Dado Villa-Lobos/Marcelo Bonfá',321724,10501773,0.98999999999999999111);
INSERT INTO tracks VALUES(1701,'Antes Das Seis',140,1,7,'Dado Villa-Lobos',189231,6296531,0.98999999999999999111);
INSERT INTO tracks VALUES(1702,'Are You Gonna Go My Way',141,1,1,'Craig Ross/Lenny Kravitz',211591,6905135,0.98999999999999999111);
INSERT INTO tracks VALUES(1703,'Fly Away',141,1,1,'Lenny Kravitz',221962,7322085,0.98999999999999999111);
INSERT INTO tracks VALUES(1704,'Rock And Roll Is Dead',141,1,1,'Lenny Kravitz',204199,6680312,0.98999999999999999111);
INSERT INTO tracks VALUES(1705,'Again',141,1,1,'Lenny Kravitz',228989,7490476,0.98999999999999999111);
INSERT INTO tracks VALUES(1706,'It Ain''t Over ''Til It''s Over',141,1,1,'Lenny Kravitz',242703,8078936,0.98999999999999999111);
INSERT INTO tracks VALUES(1707,'Can''t Get You Off My Mind',141,1,1,'Lenny Kravitz',273815,8937150,0.98999999999999999111);
INSERT INTO tracks VALUES(1708,'Mr. Cab Driver',141,1,1,'Lenny Kravitz',230321,7668084,0.98999999999999999111);
INSERT INTO tracks VALUES(1709,'American Woman',141,1,1,'B. Cummings/G. Peterson/M.J. Kale/R. Bachman',261773,8538023,0.98999999999999999111);
INSERT INTO tracks VALUES(1710,'Stand By My Woman',141,1,1,'Henry Kirssch/Lenny Kravitz/S. Pasch A. Krizan',259683,8447611,0.98999999999999999111);
INSERT INTO tracks VALUES(1711,'Always On The Run',141,1,1,'Lenny Kravitz/Slash',232515,7593397,0.98999999999999999111);
INSERT INTO tracks VALUES(1712,'Heaven Help',141,1,1,'Gerry DeVeaux/Terry Britten',190354,6222092,0.98999999999999999111);
INSERT INTO tracks VALUES(1713,'I Belong To You',141,1,1,'Lenny Kravitz',257123,8477980,0.98999999999999999111);
INSERT INTO tracks VALUES(1714,'Believe',141,1,1,'Henry Hirsch/Lenny Kravitz',295131,9661978,0.98999999999999999111);
INSERT INTO tracks VALUES(1715,'Let Love Rule',141,1,1,'Lenny Kravitz',342648,11298085,0.98999999999999999111);
INSERT INTO tracks VALUES(1716,'Black Velveteen',141,1,1,'Lenny Kravitz',290899,9531301,0.98999999999999999111);
INSERT INTO tracks VALUES(1717,'Assim Caminha A Humanidade',142,1,7,NULL,210755,6993763,0.98999999999999999111);
INSERT INTO tracks VALUES(1718,'Honolulu',143,1,7,NULL,261433,8558481,0.98999999999999999111);
INSERT INTO tracks VALUES(1719,'Dancin´Days',143,1,7,NULL,237400,7875347,0.98999999999999999111);
INSERT INTO tracks VALUES(1720,'Um Pro Outro',142,1,7,NULL,236382,7825215,0.98999999999999999111);
INSERT INTO tracks VALUES(1721,'Aviso Aos Navegantes',143,1,7,NULL,242808,8058651,0.98999999999999999111);
INSERT INTO tracks VALUES(1722,'Casa',142,1,7,NULL,307591,10107269,0.98999999999999999111);
INSERT INTO tracks VALUES(1723,'Condição',142,1,7,NULL,263549,8778465,0.98999999999999999111);
INSERT INTO tracks VALUES(1724,'Hyperconectividade',143,1,7,NULL,180636,5948039,0.98999999999999999111);
INSERT INTO tracks VALUES(1725,'O Descobridor Dos Sete Mares',143,1,7,NULL,225854,7475780,0.98999999999999999111);
INSERT INTO tracks VALUES(1726,'Satisfação',142,1,7,NULL,208065,6901681,0.98999999999999999111);
INSERT INTO tracks VALUES(1727,'Brumário',142,1,7,NULL,216241,7243499,0.98999999999999999111);
INSERT INTO tracks VALUES(1728,'Um Certo Alguém',143,1,7,NULL,194063,6430939,0.98999999999999999111);
INSERT INTO tracks VALUES(1729,'Fullgás',143,1,7,NULL,346070,11505484,0.98999999999999999111);
INSERT INTO tracks VALUES(1730,'Sábado À Noite',142,1,7,NULL,193854,6435114,0.98999999999999999111);
INSERT INTO tracks VALUES(1731,'A Cura',142,1,7,NULL,280920,9260588,0.98999999999999999111);
INSERT INTO tracks VALUES(1732,'Aquilo',143,1,7,NULL,246073,8167819,0.98999999999999999111);
INSERT INTO tracks VALUES(1733,'Atrás Do Trio Elétrico',142,1,7,NULL,149080,4917615,0.98999999999999999111);
INSERT INTO tracks VALUES(1734,'Senta A Pua',143,1,7,NULL,217547,7205844,0.98999999999999999111);
INSERT INTO tracks VALUES(1735,'Ro-Que-Se-Da-Ne',143,1,7,NULL,146703,4805897,0.98999999999999999111);
INSERT INTO tracks VALUES(1736,'Tudo Bem',142,1,7,NULL,196101,6419139,0.98999999999999999111);
INSERT INTO tracks VALUES(1737,'Toda Forma De Amor',142,1,7,NULL,227813,7496584,0.98999999999999999111);
INSERT INTO tracks VALUES(1738,'Tudo Igual',143,1,7,NULL,276035,9201645,0.98999999999999999111);
INSERT INTO tracks VALUES(1739,'Fogo De Palha',143,1,7,NULL,246804,8133732,0.98999999999999999111);
INSERT INTO tracks VALUES(1740,'Sereia',142,1,7,NULL,278047,9121087,0.98999999999999999111);
INSERT INTO tracks VALUES(1741,'Assaltaram A Gramática',143,1,7,NULL,261041,8698959,0.98999999999999999111);
INSERT INTO tracks VALUES(1742,'Se Você Pensa',142,1,7,NULL,195996,6552490,0.98999999999999999111);
INSERT INTO tracks VALUES(1743,'Lá Vem O Sol (Here Comes The Sun)',142,1,7,NULL,189492,6229645,0.98999999999999999111);
INSERT INTO tracks VALUES(1744,'O Último Romântico (Ao Vivo)',143,1,7,NULL,231993,7692697,0.98999999999999999111);
INSERT INTO tracks VALUES(1745,'Pseudo Silk Kimono',144,1,1,'Kelly, Mosley, Rothery, Trewaves',134739,4334038,0.98999999999999999111);
INSERT INTO tracks VALUES(1746,'Kayleigh',144,1,1,'Kelly, Mosley, Rothery, Trewaves',234605,7716005,0.98999999999999999111);
INSERT INTO tracks VALUES(1747,'Lavender',144,1,1,'Kelly, Mosley, Rothery, Trewaves',153417,4999814,0.98999999999999999111);
INSERT INTO tracks VALUES(1748,'Bitter Suite: Brief Encounter / Lost Weekend / Blue Angel',144,1,1,'Kelly, Mosley, Rothery, Trewaves',356493,11791068,0.98999999999999999111);
INSERT INTO tracks VALUES(1749,'Heart Of Lothian: Wide Boy / Curtain Call',144,1,1,'Kelly, Mosley, Rothery, Trewaves',366053,11893723,0.98999999999999999111);
INSERT INTO tracks VALUES(1750,'Waterhole (Expresso Bongo)',144,1,1,'Kelly, Mosley, Rothery, Trewaves',133093,4378835,0.98999999999999999111);
INSERT INTO tracks VALUES(1751,'Lords Of The Backstage',144,1,1,'Kelly, Mosley, Rothery, Trewaves',112875,3741319,0.98999999999999999111);
INSERT INTO tracks VALUES(1752,'Blind Curve: Vocal Under A Bloodlight / Passing Strangers / Mylo / Perimeter Walk / Threshold',144,1,1,'Kelly, Mosley, Rothery, Trewaves',569704,18578995,0.98999999999999999111);
INSERT INTO tracks VALUES(1753,'Childhoods End?',144,1,1,'Kelly, Mosley, Rothery, Trewaves',272796,9015366,0.98999999999999999111);
INSERT INTO tracks VALUES(1754,'White Feather',144,1,1,'Kelly, Mosley, Rothery, Trewaves',143595,4711776,0.98999999999999999111);
INSERT INTO tracks VALUES(1755,'Arrepio',145,1,7,'Carlinhos Brown',136254,4511390,0.98999999999999999111);
INSERT INTO tracks VALUES(1756,'Magamalabares',145,1,7,'Carlinhos Brown',215875,7183757,0.98999999999999999111);
INSERT INTO tracks VALUES(1757,'Chuva No Brejo',145,1,7,'Morais',145606,4857761,0.98999999999999999111);
INSERT INTO tracks VALUES(1758,'Cérebro Eletrônico',145,1,7,'Gilberto Gil',172800,5760864,0.98999999999999999111);
INSERT INTO tracks VALUES(1759,'Tempos Modernos',145,1,7,'Lulu Santos',183066,6066234,0.98999999999999999111);
INSERT INTO tracks VALUES(1760,'Maraçá',145,1,7,'Carlinhos Brown',230008,7621482,0.98999999999999999111);
INSERT INTO tracks VALUES(1761,'Blanco',145,1,7,'Marisa Monte/poema de Octavio Paz/versão: Haroldo de Campos',45191,1454532,0.98999999999999999111);
INSERT INTO tracks VALUES(1762,'Panis Et Circenses',145,1,7,'Caetano Veloso e Gilberto Gil',192339,6318373,0.98999999999999999111);
INSERT INTO tracks VALUES(1763,'De Noite Na Cama',145,1,7,'Caetano Veloso e Gilberto Gil',209005,7012658,0.98999999999999999111);
INSERT INTO tracks VALUES(1764,'Beija Eu',145,1,7,'Caetano Veloso e Gilberto Gil',197276,6512544,0.98999999999999999111);
INSERT INTO tracks VALUES(1765,'Give Me Love',145,1,7,'Caetano Veloso e Gilberto Gil',249808,8196331,0.98999999999999999111);
INSERT INTO tracks VALUES(1766,'Ainda Lembro',145,1,7,'Caetano Veloso e Gilberto Gil',218801,7211247,0.98999999999999999111);
INSERT INTO tracks VALUES(1767,'A Menina Dança',145,1,7,'Caetano Veloso e Gilberto Gil',129410,4326918,0.98999999999999999111);
INSERT INTO tracks VALUES(1768,'Dança Da Solidão',145,1,7,'Caetano Veloso e Gilberto Gil',203520,6699368,0.98999999999999999111);
INSERT INTO tracks VALUES(1769,'Ao Meu Redor',145,1,7,'Caetano Veloso e Gilberto Gil',275591,9158834,0.98999999999999999111);
INSERT INTO tracks VALUES(1770,'Bem Leve',145,1,7,'Caetano Veloso e Gilberto Gil',159190,5246835,0.98999999999999999111);
INSERT INTO tracks VALUES(1771,'Segue O Seco',145,1,7,'Caetano Veloso e Gilberto Gil',178207,5922018,0.98999999999999999111);
INSERT INTO tracks VALUES(1772,'O Xote Das Meninas',145,1,7,'Caetano Veloso e Gilberto Gil',291866,9553228,0.98999999999999999111);
INSERT INTO tracks VALUES(1773,'Wherever I Lay My Hat',146,1,14,NULL,136986,4477321,0.98999999999999999111);
INSERT INTO tracks VALUES(1774,'Get My Hands On Some Lovin''',146,1,14,NULL,149054,4860380,0.98999999999999999111);
INSERT INTO tracks VALUES(1775,'No Good Without You',146,1,14,'William "Mickey" Stevenson',161410,5259218,0.98999999999999999111);
INSERT INTO tracks VALUES(1776,'You''ve Been A Long Time Coming',146,1,14,'Brian Holland/Eddie Holland/Lamont Dozier',137221,4437949,0.98999999999999999111);
INSERT INTO tracks VALUES(1777,'When I Had Your Love',146,1,14,'Robert Rogers/Warren "Pete" Moore/William "Mickey" Stevenson',152424,4972815,0.98999999999999999111);
INSERT INTO tracks VALUES(1778,'You''re What''s Happening (In The World Today)',146,1,14,'Allen Story/George Gordy/Robert Gordy',142027,4631104,0.98999999999999999111);
INSERT INTO tracks VALUES(1779,'Loving You Is Sweeter Than Ever',146,1,14,'Ivy Hunter/Stevie Wonder',166295,5377546,0.98999999999999999111);
INSERT INTO tracks VALUES(1780,'It''s A Bitter Pill To Swallow',146,1,14,'Smokey Robinson/Warren "Pete" Moore',194821,6477882,0.98999999999999999111);
INSERT INTO tracks VALUES(1781,'Seek And You Shall Find',146,1,14,'Ivy Hunter/William "Mickey" Stevenson',223451,7306719,0.98999999999999999111);
INSERT INTO tracks VALUES(1782,'Gonna Keep On Tryin'' Till I Win Your Love',146,1,14,'Barrett Strong/Norman Whitfield',176404,5789945,0.98999999999999999111);
INSERT INTO tracks VALUES(1783,'Gonna Give Her All The Love I''ve Got',146,1,14,'Barrett Strong/Norman Whitfield',210886,6893603,0.98999999999999999111);
INSERT INTO tracks VALUES(1784,'I Wish It Would Rain',146,1,14,'Barrett Strong/Norman Whitfield/Roger Penzabene',172486,5647327,0.98999999999999999111);
INSERT INTO tracks VALUES(1785,'Abraham, Martin And John',146,1,14,'Dick Holler',273057,8888206,0.98999999999999999111);
INSERT INTO tracks VALUES(1786,'Save The Children',146,1,14,'Al Cleveland/Marvin Gaye/Renaldo Benson',194821,6342021,0.98999999999999999111);
INSERT INTO tracks VALUES(1787,'You Sure Love To Ball',146,1,14,'Marvin Gaye',218540,7217872,0.98999999999999999111);
INSERT INTO tracks VALUES(1788,'Ego Tripping Out',146,1,14,'Marvin Gaye',314514,10383887,0.98999999999999999111);
INSERT INTO tracks VALUES(1789,'Praise',146,1,14,'Marvin Gaye',235833,7839179,0.98999999999999999111);
INSERT INTO tracks VALUES(1790,'Heavy Love Affair',146,1,14,'Marvin Gaye',227892,7522232,0.98999999999999999111);
INSERT INTO tracks VALUES(1791,'Down Under',147,1,1,NULL,222171,7366142,0.98999999999999999111);
INSERT INTO tracks VALUES(1792,'Overkill',147,1,1,NULL,225410,7408652,0.98999999999999999111);
INSERT INTO tracks VALUES(1793,'Be Good Johnny',147,1,1,NULL,216320,7139814,0.98999999999999999111);
INSERT INTO tracks VALUES(1794,'Everything I Need',147,1,1,NULL,216476,7107625,0.98999999999999999111);
INSERT INTO tracks VALUES(1795,'Down by the Sea',147,1,1,NULL,408163,13314900,0.98999999999999999111);
INSERT INTO tracks VALUES(1796,'Who Can It Be Now?',147,1,1,NULL,202396,6682850,0.98999999999999999111);
INSERT INTO tracks VALUES(1797,'It''s a Mistake',147,1,1,NULL,273371,8979965,0.98999999999999999111);
INSERT INTO tracks VALUES(1798,'Dr. Heckyll & Mr. Jive',147,1,1,NULL,278465,9110403,0.98999999999999999111);
INSERT INTO tracks VALUES(1799,'Shakes and Ladders',147,1,1,NULL,198008,6560753,0.98999999999999999111);
INSERT INTO tracks VALUES(1800,'No Sign of Yesterday',147,1,1,NULL,362004,11829011,0.98999999999999999111);
INSERT INTO tracks VALUES(1801,'Enter Sandman',148,1,3,'James Hetfield, Lars Ulrich and Kirk Hammett',332251,10852002,0.98999999999999999111);
INSERT INTO tracks VALUES(1802,'Sad But True',148,1,3,'Ulrich',324754,10541258,0.98999999999999999111);
INSERT INTO tracks VALUES(1803,'Holier Than Thou',148,1,3,'Ulrich',227892,7462011,0.98999999999999999111);
INSERT INTO tracks VALUES(1804,'The Unforgiven',148,1,3,'James Hetfield, Lars Ulrich and Kirk Hammett',387082,12646886,0.98999999999999999111);
INSERT INTO tracks VALUES(1805,'Wherever I May Roam',148,1,3,'Ulrich',404323,13161169,0.98999999999999999111);
INSERT INTO tracks VALUES(1806,'Don''t Tread On Me',148,1,3,'Ulrich',240483,7827907,0.98999999999999999111);
INSERT INTO tracks VALUES(1807,'Through The Never',148,1,3,'James Hetfield, Lars Ulrich and Kirk Hammett',244375,8024047,0.98999999999999999111);
INSERT INTO tracks VALUES(1808,'Nothing Else Matters',148,1,3,'Ulrich',388832,12606241,0.98999999999999999111);
INSERT INTO tracks VALUES(1809,'Of Wolf And Man',148,1,3,'James Hetfield, Lars Ulrich and Kirk Hammett',256835,8339785,0.98999999999999999111);
INSERT INTO tracks VALUES(1810,'The God That Failed',148,1,3,'Ulrich',308610,10055959,0.98999999999999999111);
INSERT INTO tracks VALUES(1811,'My Friend Of Misery',148,1,3,'James Hetfield, Lars Ulrich and Jason Newsted',409547,13293515,0.98999999999999999111);
INSERT INTO tracks VALUES(1812,'The Struggle Within',148,1,3,'Ulrich',234240,7654052,0.98999999999999999111);
INSERT INTO tracks VALUES(1813,'Helpless',149,1,3,'Harris/Tatler',398315,12977902,0.98999999999999999111);
INSERT INTO tracks VALUES(1814,'The Small Hours',149,1,3,'Holocaust',403435,13215133,0.98999999999999999111);
INSERT INTO tracks VALUES(1815,'The Wait',149,1,3,'Killing Joke',295418,9688418,0.98999999999999999111);
INSERT INTO tracks VALUES(1816,'Crash Course In Brain Surgery',149,1,3,'Bourge/Phillips/Shelley',190406,6233729,0.98999999999999999111);
INSERT INTO tracks VALUES(1817,'Last Caress/Green Hell',149,1,3,'Danzig',209972,6854313,0.98999999999999999111);
INSERT INTO tracks VALUES(1818,'Am I Evil?',149,1,3,'Harris/Tatler',470256,15387219,0.98999999999999999111);
INSERT INTO tracks VALUES(1819,'Blitzkrieg',149,1,3,'Jones/Sirotto/Smith',216685,7090018,0.98999999999999999111);
INSERT INTO tracks VALUES(1820,'Breadfan',149,1,3,'Bourge/Phillips/Shelley',341551,11100130,0.98999999999999999111);
INSERT INTO tracks VALUES(1821,'The Prince',149,1,3,'Harris/Tatler',265769,8624492,0.98999999999999999111);
INSERT INTO tracks VALUES(1822,'Stone Cold Crazy',149,1,3,'Deacon/May/Mercury/Taylor',137717,4514830,0.98999999999999999111);
INSERT INTO tracks VALUES(1823,'So What',149,1,3,'Culmer/Exalt',189152,6162894,0.98999999999999999111);
INSERT INTO tracks VALUES(1824,'Killing Time',149,1,3,'Sweet Savage',183693,6021197,0.98999999999999999111);
INSERT INTO tracks VALUES(1825,'Overkill',149,1,3,'Clarke/Kilmister/Tayler',245133,7971330,0.98999999999999999111);
INSERT INTO tracks VALUES(1826,'Damage Case',149,1,3,'Clarke/Farren/Kilmister/Tayler',220212,7212997,0.98999999999999999111);
INSERT INTO tracks VALUES(1827,'Stone Dead Forever',149,1,3,'Clarke/Kilmister/Tayler',292127,9556060,0.98999999999999999111);
INSERT INTO tracks VALUES(1828,'Too Late Too Late',149,1,3,'Clarke/Kilmister/Tayler',192052,6276291,0.98999999999999999111);
INSERT INTO tracks VALUES(1829,'Hit The Lights',150,1,3,'James Hetfield, Lars Ulrich',257541,8357088,0.98999999999999999111);
INSERT INTO tracks VALUES(1830,'The Four Horsemen',150,1,3,'James Hetfield, Lars Ulrich, Dave Mustaine',433188,14178138,0.98999999999999999111);
INSERT INTO tracks VALUES(1831,'Motorbreath',150,1,3,'James Hetfield',188395,6153933,0.98999999999999999111);
INSERT INTO tracks VALUES(1832,'Jump In The Fire',150,1,3,'James Hetfield, Lars Ulrich, Dave Mustaine',281573,9135755,0.98999999999999999111);
INSERT INTO tracks VALUES(1833,'(Anesthesia) Pulling Teeth',150,1,3,'Cliff Burton',254955,8234710,0.98999999999999999111);
INSERT INTO tracks VALUES(1834,'Whiplash',150,1,3,'James Hetfield, Lars Ulrich',249208,8102839,0.98999999999999999111);
INSERT INTO tracks VALUES(1835,'Phantom Lord',150,1,3,'James Hetfield, Lars Ulrich, Dave Mustaine',302053,9817143,0.98999999999999999111);
INSERT INTO tracks VALUES(1836,'No Remorse',150,1,3,'James Hetfield, Lars Ulrich',386795,12672166,0.98999999999999999111);
INSERT INTO tracks VALUES(1837,'Seek & Destroy',150,1,3,'James Hetfield, Lars Ulrich',415817,13452301,0.98999999999999999111);
INSERT INTO tracks VALUES(1838,'Metal Militia',150,1,3,'James Hetfield, Lars Ulrich, Dave Mustaine',311327,10141785,0.98999999999999999111);
INSERT INTO tracks VALUES(1839,'Ain''t My Bitch',151,1,3,'James Hetfield, Lars Ulrich',304457,9931015,0.98999999999999999111);
INSERT INTO tracks VALUES(1840,'2 X 4',151,1,3,'James Hetfield, Lars Ulrich, Kirk Hammett',328254,10732251,0.98999999999999999111);
INSERT INTO tracks VALUES(1841,'The House Jack Built',151,1,3,'James Hetfield, Lars Ulrich, Kirk Hammett',398942,13005152,0.98999999999999999111);
INSERT INTO tracks VALUES(1842,'Until It Sleeps',151,1,3,'James Hetfield, Lars Ulrich',269740,8837394,0.98999999999999999111);
INSERT INTO tracks VALUES(1843,'King Nothing',151,1,3,'James Hetfield, Lars Ulrich, Kirk Hammett',328097,10681477,0.98999999999999999111);
INSERT INTO tracks VALUES(1844,'Hero Of The Day',151,1,3,'James Hetfield, Lars Ulrich, Kirk Hammett',261982,8540298,0.98999999999999999111);
INSERT INTO tracks VALUES(1845,'Bleeding Me',151,1,3,'James Hetfield, Lars Ulrich, Kirk Hammett',497998,16249420,0.98999999999999999111);
INSERT INTO tracks VALUES(1846,'Cure',151,1,3,'James Hetfield, Lars Ulrich',294347,9648615,0.98999999999999999111);
INSERT INTO tracks VALUES(1847,'Poor Twisted Me',151,1,3,'James Hetfield, Lars Ulrich',240065,7854349,0.98999999999999999111);
INSERT INTO tracks VALUES(1848,'Wasted My Hate',151,1,3,'James Hetfield, Lars Ulrich, Kirk Hammett',237296,7762300,0.98999999999999999111);
INSERT INTO tracks VALUES(1849,'Mama Said',151,1,3,'James Hetfield, Lars Ulrich',319764,10508310,0.98999999999999999111);
INSERT INTO tracks VALUES(1850,'Thorn Within',151,1,3,'James Hetfield, Lars Ulrich, Kirk Hammett',351738,11486686,0.98999999999999999111);
INSERT INTO tracks VALUES(1851,'Ronnie',151,1,3,'James Hetfield, Lars Ulrich',317204,10390947,0.98999999999999999111);
INSERT INTO tracks VALUES(1852,'The Outlaw Torn',151,1,3,'James Hetfield, Lars Ulrich',588721,19286261,0.98999999999999999111);
INSERT INTO tracks VALUES(1853,'Battery',152,1,3,'J.Hetfield/L.Ulrich',312424,10229577,0.98999999999999999111);
INSERT INTO tracks VALUES(1854,'Master Of Puppets',152,1,3,'K.Hammett',515239,16893720,0.98999999999999999111);
INSERT INTO tracks VALUES(1855,'The Thing That Should Not Be',152,1,3,'K.Hammett',396199,12952368,0.98999999999999999111);
INSERT INTO tracks VALUES(1856,'Welcome Home (Sanitarium)',152,1,3,'K.Hammett',387186,12679965,0.98999999999999999111);
INSERT INTO tracks VALUES(1857,'Disposable Heroes',152,1,3,'J.Hetfield/L.Ulrich',496718,16135560,0.98999999999999999111);
INSERT INTO tracks VALUES(1858,'Leper Messiah',152,1,3,'C.Burton',347428,11310434,0.98999999999999999111);
INSERT INTO tracks VALUES(1859,'Orion',152,1,3,'K.Hammett',500062,16378477,0.98999999999999999111);
INSERT INTO tracks VALUES(1860,'Damage Inc.',152,1,3,'K.Hammett',330919,10725029,0.98999999999999999111);
INSERT INTO tracks VALUES(1861,'Fuel',153,1,3,'Hetfield, Ulrich, Hammett',269557,8876811,0.98999999999999999111);
INSERT INTO tracks VALUES(1862,'The Memory Remains',153,1,3,'Hetfield, Ulrich',279353,9110730,0.98999999999999999111);
INSERT INTO tracks VALUES(1863,'Devil''s Dance',153,1,3,'Hetfield, Ulrich',318955,10414832,0.98999999999999999111);
INSERT INTO tracks VALUES(1864,'The Unforgiven II',153,1,3,'Hetfield, Ulrich, Hammett',395520,12886474,0.98999999999999999111);
INSERT INTO tracks VALUES(1865,'Better Than You',153,1,3,'Hetfield, Ulrich',322899,10549070,0.98999999999999999111);
INSERT INTO tracks VALUES(1866,'Slither',153,1,3,'Hetfield, Ulrich, Hammett',313103,10199789,0.98999999999999999111);
INSERT INTO tracks VALUES(1867,'Carpe Diem Baby',153,1,3,'Hetfield, Ulrich, Hammett',372480,12170693,0.98999999999999999111);
INSERT INTO tracks VALUES(1868,'Bad Seed',153,1,3,'Hetfield, Ulrich, Hammett',245394,8019586,0.98999999999999999111);
INSERT INTO tracks VALUES(1869,'Where The Wild Things Are',153,1,3,'Hetfield, Ulrich, Newsted',414380,13571280,0.98999999999999999111);
INSERT INTO tracks VALUES(1870,'Prince Charming',153,1,3,'Hetfield, Ulrich',365061,12009412,0.98999999999999999111);
INSERT INTO tracks VALUES(1871,'Low Man''s Lyric',153,1,3,'Hetfield, Ulrich',457639,14855583,0.98999999999999999111);
INSERT INTO tracks VALUES(1872,'Attitude',153,1,3,'Hetfield, Ulrich',315898,10335734,0.98999999999999999111);
INSERT INTO tracks VALUES(1873,'Fixxxer',153,1,3,'Hetfield, Ulrich, Hammett',496065,16190041,0.98999999999999999111);
INSERT INTO tracks VALUES(1874,'Fight Fire With Fire',154,1,3,'Metallica',285753,9420856,0.98999999999999999111);
INSERT INTO tracks VALUES(1875,'Ride The Lightning',154,1,3,'Metallica',397740,13055884,0.98999999999999999111);
INSERT INTO tracks VALUES(1876,'For Whom The Bell Tolls',154,1,3,'Metallica',311719,10159725,0.98999999999999999111);
INSERT INTO tracks VALUES(1877,'Fade To Black',154,1,3,'Metallica',414824,13531954,0.98999999999999999111);
INSERT INTO tracks VALUES(1878,'Trapped Under Ice',154,1,3,'Metallica',244532,7975942,0.98999999999999999111);
INSERT INTO tracks VALUES(1879,'Escape',154,1,3,'Metallica',264359,8652332,0.98999999999999999111);
INSERT INTO tracks VALUES(1880,'Creeping Death',154,1,3,'Metallica',396878,12955593,0.98999999999999999111);
INSERT INTO tracks VALUES(1881,'The Call Of Ktulu',154,1,3,'Metallica',534883,17486240,0.98999999999999999111);
INSERT INTO tracks VALUES(1882,'Frantic',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',350458,11510849,0.98999999999999999111);
INSERT INTO tracks VALUES(1883,'St. Anger',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',441234,14363779,0.98999999999999999111);
INSERT INTO tracks VALUES(1884,'Some Kind Of Monster',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',505626,16557497,0.98999999999999999111);
INSERT INTO tracks VALUES(1885,'Dirty Window',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',324989,10670604,0.98999999999999999111);
INSERT INTO tracks VALUES(1886,'Invisible Kid',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',510197,16591800,0.98999999999999999111);
INSERT INTO tracks VALUES(1887,'My World',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',345626,11253756,0.98999999999999999111);
INSERT INTO tracks VALUES(1888,'Shoot Me Again',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',430210,14093551,0.98999999999999999111);
INSERT INTO tracks VALUES(1889,'Sweet Amber',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',327235,10616595,0.98999999999999999111);
INSERT INTO tracks VALUES(1890,'The Unnamed Feeling',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',429479,14014582,0.98999999999999999111);
INSERT INTO tracks VALUES(1891,'Purify',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',314017,10232537,0.98999999999999999111);
INSERT INTO tracks VALUES(1892,'All Within My Hands',155,1,3,'Bob Rock/James Hetfield/Kirk Hammett/Lars Ulrich',527986,17162741,0.98999999999999999111);
INSERT INTO tracks VALUES(1893,'Blackened',156,1,3,'James Hetfield, Lars Ulrich & Jason Newsted',403382,13254874,0.98999999999999999111);
INSERT INTO tracks VALUES(1894,'...And Justice For All',156,1,3,'James Hetfield, Lars Ulrich & Kirk Hammett',585769,19262088,0.98999999999999999111);
INSERT INTO tracks VALUES(1895,'Eye Of The Beholder',156,1,3,'James Hetfield, Lars Ulrich & Kirk Hammett',385828,12747894,0.98999999999999999111);
INSERT INTO tracks VALUES(1896,'One',156,1,3,'James Hetfield & Lars Ulrich',446484,14695721,0.98999999999999999111);
INSERT INTO tracks VALUES(1897,'The Shortest Straw',156,1,3,'James Hetfield and Lars Ulrich',395389,13013990,0.98999999999999999111);
INSERT INTO tracks VALUES(1898,'Harvester Of Sorrow',156,1,3,'James Hetfield and Lars Ulrich',345547,11377339,0.98999999999999999111);
INSERT INTO tracks VALUES(1899,'The Frayed Ends Of Sanity',156,1,3,'James Hetfield, Lars Ulrich and Kirk Hammett',464039,15198986,0.98999999999999999111);
INSERT INTO tracks VALUES(1900,'To Live Is To Die',156,1,3,'James Hetfield, Lars Ulrich and Cliff Burton',588564,19243795,0.98999999999999999111);
INSERT INTO tracks VALUES(1901,'Dyers Eve',156,1,3,'James Hetfield, Lars Ulrich and Kirk Hammett',313991,10302828,0.98999999999999999111);
INSERT INTO tracks VALUES(1902,'Springsville',157,1,2,'J. Carisi',207725,6776219,0.98999999999999999111);
INSERT INTO tracks VALUES(1903,'The Maids Of Cadiz',157,1,2,'L. Delibes',233534,7505275,0.98999999999999999111);
INSERT INTO tracks VALUES(1904,'The Duke',157,1,2,'Dave Brubeck',214961,6977626,0.98999999999999999111);
INSERT INTO tracks VALUES(1905,'My Ship',157,1,2,'Ira Gershwin, Kurt Weill',268016,8581144,0.98999999999999999111);
INSERT INTO tracks VALUES(1906,'Miles Ahead',157,1,2,'Miles Davis, Gil Evans',209893,6807707,0.98999999999999999111);
INSERT INTO tracks VALUES(1907,'Blues For Pablo',157,1,2,'Gil Evans',318328,10218398,0.98999999999999999111);
INSERT INTO tracks VALUES(1908,'New Rhumba',157,1,2,'A. Jamal',276871,8980400,0.98999999999999999111);
INSERT INTO tracks VALUES(1909,'The Meaning Of The Blues',157,1,2,'R. Troup, L. Worth',168594,5395412,0.98999999999999999111);
INSERT INTO tracks VALUES(1910,'Lament',157,1,2,'J.J. Johnson',134191,4293394,0.98999999999999999111);
INSERT INTO tracks VALUES(1911,'I Don''t Wanna Be Kissed (By Anyone But You)',157,1,2,'H. Spina, J. Elliott',191320,6219487,0.98999999999999999111);
INSERT INTO tracks VALUES(1912,'Springsville (Alternate Take)',157,1,2,'J. Carisi',196388,6382079,0.98999999999999999111);
INSERT INTO tracks VALUES(1913,'Blues For Pablo (Alternate Take)',157,1,2,'Gil Evans',212558,6900619,0.98999999999999999111);
INSERT INTO tracks VALUES(1914,'The Meaning Of The Blues/Lament (Alternate Take)',157,1,2,'J.J. Johnson/R. Troup, L. Worth',309786,9912387,0.98999999999999999111);
INSERT INTO tracks VALUES(1915,'I Don''t Wanna Be Kissed (By Anyone But You) (Alternate Take)',157,1,2,'H. Spina, J. Elliott',192078,6254796,0.98999999999999999111);
INSERT INTO tracks VALUES(1916,'Coração De Estudante',158,1,7,'Wagner Tiso, Milton Nascimento',238550,7797308,0.98999999999999999111);
INSERT INTO tracks VALUES(1917,'A Noite Do Meu Bem',158,1,7,'Dolores Duran',220081,7125225,0.98999999999999999111);
INSERT INTO tracks VALUES(1918,'Paisagem Na Janela',158,1,7,'Lô Borges, Fernando Brant',197694,6523547,0.98999999999999999111);
INSERT INTO tracks VALUES(1919,'Cuitelinho',158,1,7,'Folclore',209397,6803970,0.98999999999999999111);
INSERT INTO tracks VALUES(1920,'Caxangá',158,1,7,'Milton Nascimento, Fernando Brant',245551,8144179,0.98999999999999999111);
INSERT INTO tracks VALUES(1921,'Nos Bailes Da Vida',158,1,7,'Milton Nascimento, Fernando Brant',275748,9126170,0.98999999999999999111);
INSERT INTO tracks VALUES(1922,'Menestrel Das Alagoas',158,1,7,'Milton Nascimento, Fernando Brant',199758,6542289,0.98999999999999999111);
INSERT INTO tracks VALUES(1923,'Brasil',158,1,7,'Milton Nascimento, Fernando Brant',155428,5252560,0.98999999999999999111);
INSERT INTO tracks VALUES(1924,'Canção Do Novo Mundo',158,1,7,'Beto Guedes, Ronaldo Bastos',215353,7032626,0.98999999999999999111);
INSERT INTO tracks VALUES(1925,'Um Gosto De Sol',158,1,7,'Milton Nascimento, Ronaldo Bastos',307200,9893875,0.98999999999999999111);
INSERT INTO tracks VALUES(1926,'Solar',158,1,7,'Milton Nascimento, Fernando Brant',156212,5098288,0.98999999999999999111);
INSERT INTO tracks VALUES(1927,'Para Lennon E McCartney',158,1,7,'Lô Borges, Márcio Borges, Fernando Brant',321828,10626920,0.98999999999999999111);
INSERT INTO tracks VALUES(1928,'Maria, Maria',158,1,7,'Milton Nascimento, Fernando Brant',72463,2371543,0.98999999999999999111);
INSERT INTO tracks VALUES(1929,'Minas',159,1,7,'Milton Nascimento, Caetano Veloso',152293,4921056,0.98999999999999999111);
INSERT INTO tracks VALUES(1930,'Fé Cega, Faca Amolada',159,1,7,'Milton Nascimento, Ronaldo Bastos',278099,9258649,0.98999999999999999111);
INSERT INTO tracks VALUES(1931,'Beijo Partido',159,1,7,'Toninho Horta',229564,7506969,0.98999999999999999111);
INSERT INTO tracks VALUES(1932,'Saudade Dos Aviões Da Panair (Conversando No Bar)',159,1,7,'Milton Nascimento, Fernando Brant',268721,8805088,0.98999999999999999111);
INSERT INTO tracks VALUES(1933,'Gran Circo',159,1,7,'Milton Nascimento, Márcio Borges',251297,8237026,0.98999999999999999111);
INSERT INTO tracks VALUES(1934,'Ponta de Areia',159,1,7,'Milton Nascimento, Fernando Brant',272796,8874285,0.98999999999999999111);
INSERT INTO tracks VALUES(1935,'Trastevere',159,1,7,'Milton Nascimento, Ronaldo Bastos',265665,8708399,0.98999999999999999111);
INSERT INTO tracks VALUES(1936,'Idolatrada',159,1,7,'Milton Nascimento, Fernando Brant',286249,9426153,0.98999999999999999111);
INSERT INTO tracks VALUES(1937,'Leila (Venha Ser Feliz)',159,1,7,'Milton Nascimento',209737,6898507,0.98999999999999999111);
INSERT INTO tracks VALUES(1938,'Paula E Bebeto',159,1,7,'Milton Nascimento, Caetano Veloso',135732,4583956,0.98999999999999999111);
INSERT INTO tracks VALUES(1939,'Simples',159,1,7,'Nelson Angelo',133093,4326333,0.98999999999999999111);
INSERT INTO tracks VALUES(1940,'Norwegian Wood',159,1,7,'John Lennon, Paul McCartney',413910,13520382,0.98999999999999999111);
INSERT INTO tracks VALUES(1941,'Caso Você Queira Saber',159,1,7,'Beto Guedes, Márcio Borges',205688,6787901,0.98999999999999999111);
INSERT INTO tracks VALUES(1942,'Ace Of Spades',160,1,3,'Clarke/Kilmister/Taylor',169926,5523552,0.98999999999999999111);
INSERT INTO tracks VALUES(1943,'Love Me Like A Reptile',160,1,3,'Clarke/Kilmister/Taylor',203546,6616389,0.98999999999999999111);
INSERT INTO tracks VALUES(1944,'Shoot You In The Back',160,1,3,'Clarke/Kilmister/Taylor',160026,5175327,0.98999999999999999111);
INSERT INTO tracks VALUES(1945,'Live To Win',160,1,3,'Clarke/Kilmister/Taylor',217626,7102182,0.98999999999999999111);
INSERT INTO tracks VALUES(1946,'Fast And Loose',160,1,3,'Clarke/Kilmister/Taylor',203337,6643350,0.98999999999999999111);
INSERT INTO tracks VALUES(1947,'(We Are) The Road Crew',160,1,3,'Clarke/Kilmister/Taylor',192600,6283035,0.98999999999999999111);
INSERT INTO tracks VALUES(1948,'Fire Fire',160,1,3,'Clarke/Kilmister/Taylor',164675,5416114,0.98999999999999999111);
INSERT INTO tracks VALUES(1949,'Jailbait',160,1,3,'Clarke/Kilmister/Taylor',213916,6983609,0.98999999999999999111);
INSERT INTO tracks VALUES(1950,'Dance',160,1,3,'Clarke/Kilmister/Taylor',158432,5155099,0.98999999999999999111);
INSERT INTO tracks VALUES(1951,'Bite The Bullet',160,1,3,'Clarke/Kilmister/Taylor',98115,3195536,0.98999999999999999111);
INSERT INTO tracks VALUES(1952,'The Chase Is Better Than The Catch',160,1,3,'Clarke/Kilmister/Taylor',258403,8393310,0.98999999999999999111);
INSERT INTO tracks VALUES(1953,'The Hammer',160,1,3,'Clarke/Kilmister/Taylor',168071,5543267,0.98999999999999999111);
INSERT INTO tracks VALUES(1954,'Dirty Love',160,1,3,'Clarke/Kilmister/Taylor',176457,5805241,0.98999999999999999111);
INSERT INTO tracks VALUES(1955,'Please Don''t Touch',160,1,3,'Heath/Robinson',169926,5557002,0.98999999999999999111);
INSERT INTO tracks VALUES(1956,'Emergency',160,1,3,'Dufort/Johnson/McAuliffe/Williams',180427,5828728,0.98999999999999999111);
INSERT INTO tracks VALUES(1957,'Kir Royal',161,1,16,'Mônica Marianno',234788,7706552,0.98999999999999999111);
INSERT INTO tracks VALUES(1958,'O Que Vai Em Meu Coração',161,1,16,'Mônica Marianno',255373,8366846,0.98999999999999999111);
INSERT INTO tracks VALUES(1959,'Aos Leões',161,1,16,'Mônica Marianno',234684,7790574,0.98999999999999999111);
INSERT INTO tracks VALUES(1960,'Dois Índios',161,1,16,'Mônica Marianno',219271,7213072,0.98999999999999999111);
INSERT INTO tracks VALUES(1961,'Noite Negra',161,1,16,'Mônica Marianno',206811,6819584,0.98999999999999999111);
INSERT INTO tracks VALUES(1962,'Beijo do Olhar',161,1,16,'Mônica Marianno',252682,8369029,0.98999999999999999111);
INSERT INTO tracks VALUES(1963,'É Fogo',161,1,16,'Mônica Marianno',194873,6501520,0.98999999999999999111);
INSERT INTO tracks VALUES(1964,'Já Foi',161,1,16,'Mônica Marianno',245681,8094872,0.98999999999999999111);
INSERT INTO tracks VALUES(1965,'Só Se For Pelo Cabelo',161,1,16,'Mônica Marianno',238288,8006345,0.98999999999999999111);
INSERT INTO tracks VALUES(1966,'No Clima',161,1,16,'Mônica Marianno',249495,8362040,0.98999999999999999111);
INSERT INTO tracks VALUES(1967,'A Moça e a Chuva',161,1,16,'Mônica Marianno',274625,8929357,0.98999999999999999111);
INSERT INTO tracks VALUES(1968,'Demorou!',161,1,16,'Mônica Marianno',39131,1287083,0.98999999999999999111);
INSERT INTO tracks VALUES(1969,'Bitter Pill',162,1,3,'Mick Mars/Nikki Sixx/Tommy Lee/Vince Neil',266814,8666786,0.98999999999999999111);
INSERT INTO tracks VALUES(1970,'Enslaved',162,1,3,'Mick Mars/Nikki Sixx/Tommy Lee',269844,8789966,0.98999999999999999111);
INSERT INTO tracks VALUES(1971,'Girls, Girls, Girls',162,1,3,'Mick Mars/Nikki Sixx/Tommy Lee',270288,8874814,0.98999999999999999111);
INSERT INTO tracks VALUES(1972,'Kickstart My Heart',162,1,3,'Nikki Sixx',283559,9237736,0.98999999999999999111);
INSERT INTO tracks VALUES(1973,'Wild Side',162,1,3,'Nikki Sixx/Tommy Lee/Vince Neil',276767,9116997,0.98999999999999999111);
INSERT INTO tracks VALUES(1974,'Glitter',162,1,3,'Bryan Adams/Nikki Sixx/Scott Humphrey',340114,11184094,0.98999999999999999111);
INSERT INTO tracks VALUES(1975,'Dr. Feelgood',162,1,3,'Mick Mars/Nikki Sixx',282618,9281875,0.98999999999999999111);
INSERT INTO tracks VALUES(1976,'Same Ol'' Situation',162,1,3,'Mick Mars/Nikki Sixx/Tommy Lee/Vince Neil',254511,8283958,0.98999999999999999111);
INSERT INTO tracks VALUES(1977,'Home Sweet Home',162,1,3,'Nikki Sixx/Tommy Lee/Vince Neil',236904,7697538,0.98999999999999999111);
INSERT INTO tracks VALUES(1978,'Afraid',162,1,3,'Nikki Sixx',248006,8077464,0.98999999999999999111);
INSERT INTO tracks VALUES(1979,'Don''t Go Away Mad (Just Go Away)',162,1,3,'Mick Mars/Nikki Sixx',279980,9188156,0.98999999999999999111);
INSERT INTO tracks VALUES(1980,'Without You',162,1,3,'Mick Mars/Nikki Sixx',268956,8738371,0.98999999999999999111);
INSERT INTO tracks VALUES(1981,'Smokin'' in The Boys Room',162,1,3,'Cub Coda/Michael Lutz',206837,6735408,0.98999999999999999111);
INSERT INTO tracks VALUES(1982,'Primal Scream',162,1,3,'Mick Mars/Nikki Sixx/Tommy Lee/Vince Neil',286197,9421164,0.98999999999999999111);
INSERT INTO tracks VALUES(1983,'Too Fast For Love',162,1,3,'Nikki Sixx',200829,6580542,0.98999999999999999111);
INSERT INTO tracks VALUES(1984,'Looks That Kill',162,1,3,'Nikki Sixx',240979,7831122,0.98999999999999999111);
INSERT INTO tracks VALUES(1985,'Shout At The Devil',162,1,3,'Nikki Sixx',221962,7281974,0.98999999999999999111);
INSERT INTO tracks VALUES(1986,'Intro',163,1,1,'Kurt Cobain',52218,1688527,0.98999999999999999111);
INSERT INTO tracks VALUES(1987,'School',163,1,1,'Kurt Cobain',160235,5234885,0.98999999999999999111);
INSERT INTO tracks VALUES(1988,'Drain You',163,1,1,'Kurt Cobain',215196,7013175,0.98999999999999999111);
INSERT INTO tracks VALUES(1989,'Aneurysm',163,1,1,'Nirvana',271516,8862545,0.98999999999999999111);
INSERT INTO tracks VALUES(1990,'Smells Like Teen Spirit',163,1,1,'Nirvana',287190,9425215,0.98999999999999999111);
INSERT INTO tracks VALUES(1991,'Been A Son',163,1,1,'Kurt Cobain',127555,4170369,0.98999999999999999111);
INSERT INTO tracks VALUES(1992,'Lithium',163,1,1,'Kurt Cobain',250017,8148800,0.98999999999999999111);
INSERT INTO tracks VALUES(1993,'Sliver',163,1,1,'Kurt Cobain',116218,3784567,0.98999999999999999111);
INSERT INTO tracks VALUES(1994,'Spank Thru',163,1,1,'Kurt Cobain',190354,6186487,0.98999999999999999111);
INSERT INTO tracks VALUES(1995,'Scentless Apprentice',163,1,1,'Nirvana',211200,6898177,0.98999999999999999111);
INSERT INTO tracks VALUES(1996,'Heart-Shaped Box',163,1,1,'Kurt Cobain',281887,9210982,0.98999999999999999111);
INSERT INTO tracks VALUES(1997,'Milk It',163,1,1,'Kurt Cobain',225724,7406945,0.98999999999999999111);
INSERT INTO tracks VALUES(1998,'Negative Creep',163,1,1,'Kurt Cobain',163761,5354854,0.98999999999999999111);
INSERT INTO tracks VALUES(1999,'Polly',163,1,1,'Kurt Cobain',149995,4885331,0.98999999999999999111);
INSERT INTO tracks VALUES(2000,'Breed',163,1,1,'Kurt Cobain',208378,6759080,0.98999999999999999111);
INSERT INTO tracks VALUES(2001,'Tourette''s',163,1,1,'Kurt Cobain',115591,3753246,0.98999999999999999111);
INSERT INTO tracks VALUES(2002,'Blew',163,1,1,'Kurt Cobain',216346,7096936,0.98999999999999999111);
INSERT INTO tracks VALUES(2003,'Smells Like Teen Spirit',164,1,1,'Kurt Cobain',301296,9823847,0.98999999999999999111);
INSERT INTO tracks VALUES(2004,'In Bloom',164,1,1,'Kurt Cobain',254928,8327077,0.98999999999999999111);
INSERT INTO tracks VALUES(2005,'Come As You Are',164,1,1,'Kurt Cobain',219219,7123357,0.98999999999999999111);
INSERT INTO tracks VALUES(2006,'Breed',164,1,1,'Kurt Cobain',183928,5984812,0.98999999999999999111);
INSERT INTO tracks VALUES(2007,'Lithium',164,1,1,'Kurt Cobain',256992,8404745,0.98999999999999999111);
INSERT INTO tracks VALUES(2008,'Polly',164,1,1,'Kurt Cobain',177031,5788407,0.98999999999999999111);
INSERT INTO tracks VALUES(2009,'Territorial Pissings',164,1,1,'Kurt Cobain',143281,4613880,0.98999999999999999111);
INSERT INTO tracks VALUES(2010,'Drain You',164,1,1,'Kurt Cobain',223973,7273440,0.98999999999999999111);
INSERT INTO tracks VALUES(2011,'Lounge Act',164,1,1,'Kurt Cobain',156786,5093635,0.98999999999999999111);
INSERT INTO tracks VALUES(2012,'Stay Away',164,1,1,'Kurt Cobain',212636,6956404,0.98999999999999999111);
INSERT INTO tracks VALUES(2013,'On A Plain',164,1,1,'Kurt Cobain',196440,6390635,0.98999999999999999111);
INSERT INTO tracks VALUES(2014,'Something In The Way',164,1,1,'Kurt Cobain',230556,7472168,0.98999999999999999111);
INSERT INTO tracks VALUES(2015,'Time',165,1,1,NULL,96888,3124455,0.98999999999999999111);
INSERT INTO tracks VALUES(2016,'P.S.Apareça',165,1,1,NULL,209188,6842244,0.98999999999999999111);
INSERT INTO tracks VALUES(2017,'Sangue Latino',165,1,1,NULL,223033,7354184,0.98999999999999999111);
INSERT INTO tracks VALUES(2018,'Folhas Secas',165,1,1,NULL,161253,5284522,0.98999999999999999111);
INSERT INTO tracks VALUES(2019,'Poeira',165,1,1,NULL,267075,8784141,0.98999999999999999111);
INSERT INTO tracks VALUES(2020,'Mágica',165,1,1,NULL,233743,7627348,0.98999999999999999111);
INSERT INTO tracks VALUES(2021,'Quem Mata A Mulher Mata O Melhor',165,1,1,NULL,262791,8640121,0.98999999999999999111);
INSERT INTO tracks VALUES(2022,'Mundaréu',165,1,1,NULL,217521,7158975,0.98999999999999999111);
INSERT INTO tracks VALUES(2023,'O Braço Da Minha Guitarra',165,1,1,NULL,258351,8469531,0.98999999999999999111);
INSERT INTO tracks VALUES(2024,'Deus',165,1,1,NULL,284160,9188110,0.98999999999999999111);
INSERT INTO tracks VALUES(2025,'Mãe Terra',165,1,1,NULL,306625,9949269,0.98999999999999999111);
INSERT INTO tracks VALUES(2026,'Às Vezes',165,1,1,NULL,330292,10706614,0.98999999999999999111);
INSERT INTO tracks VALUES(2027,'Menino De Rua',165,1,1,NULL,329795,10784595,0.98999999999999999111);
INSERT INTO tracks VALUES(2028,'Prazer E Fé',165,1,1,NULL,214831,7031383,0.98999999999999999111);
INSERT INTO tracks VALUES(2029,'Elza',165,1,1,NULL,199105,6517629,0.98999999999999999111);
INSERT INTO tracks VALUES(2030,'Requebra',166,1,7,NULL,240744,8010811,0.98999999999999999111);
INSERT INTO tracks VALUES(2031,'Nossa Gente (Avisa Là)',166,1,7,NULL,188212,6233201,0.98999999999999999111);
INSERT INTO tracks VALUES(2032,'Olodum - Alegria Geral',166,1,7,NULL,233404,7754245,0.98999999999999999111);
INSERT INTO tracks VALUES(2033,'Madagáscar Olodum',166,1,7,NULL,252264,8270584,0.98999999999999999111);
INSERT INTO tracks VALUES(2034,'Faraó Divindade Do Egito',166,1,7,NULL,228571,7523278,0.98999999999999999111);
INSERT INTO tracks VALUES(2035,'Todo Amor (Asas Da Liberdade)',166,1,7,NULL,245133,8121434,0.98999999999999999111);
INSERT INTO tracks VALUES(2036,'Denúncia',166,1,7,NULL,159555,5327433,0.98999999999999999111);
INSERT INTO tracks VALUES(2037,'Olodum, A Banda Do Pelô',166,1,7,NULL,146599,4900121,0.98999999999999999111);
INSERT INTO tracks VALUES(2038,'Cartao Postal',166,1,7,NULL,211565,7082301,0.98999999999999999111);
INSERT INTO tracks VALUES(2039,'Jeito Faceiro',166,1,7,NULL,217286,7233608,0.98999999999999999111);
INSERT INTO tracks VALUES(2040,'Revolta Olodum',166,1,7,NULL,230191,7557065,0.98999999999999999111);
INSERT INTO tracks VALUES(2041,'Reggae Odoyá',166,1,7,NULL,224470,7499807,0.98999999999999999111);
INSERT INTO tracks VALUES(2042,'Protesto Do Olodum (Ao Vivo)',166,1,7,NULL,206001,6766104,0.98999999999999999111);
INSERT INTO tracks VALUES(2043,'Olodum - Smile (Instrumental)',166,1,7,NULL,235833,7871409,0.98999999999999999111);
INSERT INTO tracks VALUES(2044,'Vulcão Dub - Fui Eu',167,1,7,'Bi Ribeira/Herbert Vianna/João Barone',287059,9495202,0.98999999999999999111);
INSERT INTO tracks VALUES(2045,'O Trem Da Juventude',167,1,7,'Herbert Vianna',225880,7507655,0.98999999999999999111);
INSERT INTO tracks VALUES(2046,'Manguetown',167,1,7,'Chico Science/Dengue/Lúcio Maia',162925,5382018,0.98999999999999999111);
INSERT INTO tracks VALUES(2047,'Um Amor, Um Lugar',167,1,7,'Herbert Vianna',184555,6090334,0.98999999999999999111);
INSERT INTO tracks VALUES(2048,'Bora-Bora',167,1,7,'Herbert Vianna',182987,6036046,0.98999999999999999111);
INSERT INTO tracks VALUES(2049,'Vai Valer',167,1,7,'Herbert Vianna',206524,6899778,0.98999999999999999111);
INSERT INTO tracks VALUES(2050,'I Feel Good (I Got You) - Sossego',167,1,7,'James Brown/Tim Maia',244976,8091302,0.98999999999999999111);
INSERT INTO tracks VALUES(2051,'Uns Dias',167,1,7,'Herbert Vianna',240796,7931552,0.98999999999999999111);
INSERT INTO tracks VALUES(2052,'Sincero Breu',167,1,7,'C. A./C.A./Celso Alvim/Herbert Vianna/Mário Moura/Pedro Luís/Sidon Silva',208013,6921669,0.98999999999999999111);
INSERT INTO tracks VALUES(2053,'Meu Erro',167,1,7,'Herbert Vianna',188577,6192791,0.98999999999999999111);
INSERT INTO tracks VALUES(2054,'Selvagem',167,1,7,'Bi Ribeiro/Herbert Vianna/João Barone',148558,4942831,0.98999999999999999111);
INSERT INTO tracks VALUES(2055,'Brasília 5:31',167,1,7,'Herbert Vianna',178337,5857116,0.98999999999999999111);
INSERT INTO tracks VALUES(2056,'Tendo A Lua',167,1,7,'Herbert Vianna/Tet Tillett',198922,6568180,0.98999999999999999111);
INSERT INTO tracks VALUES(2057,'Que País É Este',167,1,7,'Renato Russo',216685,7137865,0.98999999999999999111);
INSERT INTO tracks VALUES(2058,'Navegar Impreciso',167,1,7,'Herbert Vianna',262870,8761283,0.98999999999999999111);
INSERT INTO tracks VALUES(2059,'Feira Moderna',167,1,7,'Beto Guedes/Fernando Brant/L Borges',182517,6001793,0.98999999999999999111);
INSERT INTO tracks VALUES(2060,'Tequila - Lourinha Bombril (Parate Y Mira)',167,1,7,'Bahiano/Chuck Rio/Diego Blanco/Herbert Vianna',255738,8514961,0.98999999999999999111);
INSERT INTO tracks VALUES(2061,'Vamo Batê Lata',167,1,7,'Herbert Vianna',228754,7585707,0.98999999999999999111);
INSERT INTO tracks VALUES(2062,'Life During Wartime',167,1,7,'Chris Frantz/David Byrne/Jerry Harrison/Tina Weymouth',259186,8543439,0.98999999999999999111);
INSERT INTO tracks VALUES(2063,'Nebulosa Do Amor',167,1,7,'Herbert Vianna',203415,6732496,0.98999999999999999111);
INSERT INTO tracks VALUES(2064,'Caleidoscópio',167,1,7,'Herbert Vianna',256522,8484597,0.98999999999999999111);
INSERT INTO tracks VALUES(2065,'Trac Trac',168,1,7,'Fito Paez/Herbert Vianna',231653,7638256,0.98999999999999999111);
INSERT INTO tracks VALUES(2066,'Tendo A Lua',168,1,7,'Herbert Vianna/Tetê Tillet',219585,7342776,0.98999999999999999111);
INSERT INTO tracks VALUES(2067,'Mensagen De Amor (2000)',168,1,7,'Herbert Vianna',183588,6061324,0.98999999999999999111);
INSERT INTO tracks VALUES(2068,'Lourinha Bombril',168,1,7,'Bahiano/Diego Blanco/Herbert Vianna',159895,5301882,0.98999999999999999111);
INSERT INTO tracks VALUES(2069,'La Bella Luna',168,1,7,'Herbert Vianna',192653,6428598,0.98999999999999999111);
INSERT INTO tracks VALUES(2070,'Busca Vida',168,1,7,'Herbert Vianna',176431,5798663,0.98999999999999999111);
INSERT INTO tracks VALUES(2071,'Uma Brasileira',168,1,7,'Carlinhos Brown/Herbert Vianna',217573,7280574,0.98999999999999999111);
INSERT INTO tracks VALUES(2072,'Luis Inacio (300 Picaretas)',168,1,7,'Herbert Vianna',198191,6576790,0.98999999999999999111);
INSERT INTO tracks VALUES(2073,'Saber Amar',168,1,7,'Herbert Vianna',202788,6723733,0.98999999999999999111);
INSERT INTO tracks VALUES(2074,'Ela Disse Adeus',168,1,7,'Herbert Vianna',226298,7608999,0.98999999999999999111);
INSERT INTO tracks VALUES(2075,'O Amor Nao Sabe Esperar',168,1,7,'Herbert Vianna',241084,8042534,0.98999999999999999111);
INSERT INTO tracks VALUES(2076,'Aonde Quer Que Eu Va',168,1,7,'Herbert Vianna/Paulo Sérgio Valle',258089,8470121,0.98999999999999999111);
INSERT INTO tracks VALUES(2077,'Caleidoscópio',169,1,7,NULL,211330,7000017,0.98999999999999999111);
INSERT INTO tracks VALUES(2078,'Óculos',169,1,7,NULL,219271,7262419,0.98999999999999999111);
INSERT INTO tracks VALUES(2079,'Cinema Mudo',169,1,7,NULL,227918,7612168,0.98999999999999999111);
INSERT INTO tracks VALUES(2080,'Alagados',169,1,7,NULL,302393,10255463,0.98999999999999999111);
INSERT INTO tracks VALUES(2081,'Lanterna Dos Afogados',169,1,7,NULL,190197,6264318,0.98999999999999999111);
INSERT INTO tracks VALUES(2082,'Melô Do Marinheiro',169,1,7,NULL,208352,6905668,0.98999999999999999111);
INSERT INTO tracks VALUES(2083,'Vital E Sua Moto',169,1,7,NULL,210207,6902878,0.98999999999999999111);
INSERT INTO tracks VALUES(2084,'O Beco',169,1,7,NULL,189178,6293184,0.98999999999999999111);
INSERT INTO tracks VALUES(2085,'Meu Erro',169,1,7,NULL,208431,6893533,0.98999999999999999111);
INSERT INTO tracks VALUES(2086,'Perplexo',169,1,7,NULL,161175,5355013,0.98999999999999999111);
INSERT INTO tracks VALUES(2087,'Me Liga',169,1,7,NULL,229590,7565912,0.98999999999999999111);
INSERT INTO tracks VALUES(2088,'Quase Um Segundo',169,1,7,NULL,275644,8971355,0.98999999999999999111);
INSERT INTO tracks VALUES(2089,'Selvagem',169,1,7,NULL,245890,8141084,0.98999999999999999111);
INSERT INTO tracks VALUES(2090,'Romance Ideal',169,1,7,NULL,250070,8260477,0.98999999999999999111);
INSERT INTO tracks VALUES(2091,'Será Que Vai Chover?',169,1,7,NULL,337057,11133830,0.98999999999999999111);
INSERT INTO tracks VALUES(2092,'SKA',169,1,7,NULL,148871,4943540,0.98999999999999999111);
INSERT INTO tracks VALUES(2093,'Bark at the Moon',170,2,1,'O. Osbourne',257252,4601224,0.98999999999999999111);
INSERT INTO tracks VALUES(2094,'I Don''t Know',171,2,1,'B. Daisley, O. Osbourne & R. Rhoads',312980,5525339,0.98999999999999999111);
INSERT INTO tracks VALUES(2095,'Crazy Train',171,2,1,'B. Daisley, O. Osbourne & R. Rhoads',295960,5255083,0.98999999999999999111);
INSERT INTO tracks VALUES(2096,'Flying High Again',172,2,1,'L. Kerslake, O. Osbourne, R. Daisley & R. Rhoads',290851,5179599,0.98999999999999999111);
INSERT INTO tracks VALUES(2097,'Mama, I''m Coming Home',173,2,1,'L. Kilmister, O. Osbourne & Z. Wylde',251586,4302390,0.98999999999999999111);
INSERT INTO tracks VALUES(2098,'No More Tears',173,2,1,'J. Purdell, M. Inez, O. Osbourne, R. Castillo & Z. Wylde',444358,7362964,0.98999999999999999111);
INSERT INTO tracks VALUES(2099,'I Don''t Know',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',283088,9207869,0.98999999999999999111);
INSERT INTO tracks VALUES(2100,'Crazy Train',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',322716,10517408,0.98999999999999999111);
INSERT INTO tracks VALUES(2101,'Believer',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',308897,10003794,0.98999999999999999111);
INSERT INTO tracks VALUES(2102,'Mr. Crowley',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',344241,11184130,0.98999999999999999111);
INSERT INTO tracks VALUES(2103,'Flying High Again',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads, L. Kerslake',261224,8481822,0.98999999999999999111);
INSERT INTO tracks VALUES(2104,'Relvelation (Mother Earth)',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',349440,11367866,0.98999999999999999111);
INSERT INTO tracks VALUES(2105,'Steal Away (The Night)',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',485720,15945806,0.98999999999999999111);
INSERT INTO tracks VALUES(2106,'Suicide Solution (With Guitar Solo)',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',467069,15119938,0.98999999999999999111);
INSERT INTO tracks VALUES(2107,'Iron Man',174,1,3,'A. F. Iommi, W. Ward, T. Butler, J. Osbourne',172120,5609799,0.98999999999999999111);
INSERT INTO tracks VALUES(2108,'Children Of The Grave',174,1,3,'A. F. Iommi, W. Ward, T. Butler, J. Osbourne',357067,11626740,0.98999999999999999111);
INSERT INTO tracks VALUES(2109,'Paranoid',174,1,3,'A. F. Iommi, W. Ward, T. Butler, J. Osbourne',176352,5729813,0.98999999999999999111);
INSERT INTO tracks VALUES(2110,'Goodbye To Romance',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',334393,10841337,0.98999999999999999111);
INSERT INTO tracks VALUES(2111,'No Bone Movies',174,1,3,'O. Osbourne, R. Daisley, R. Rhoads',249208,8095199,0.98999999999999999111);
INSERT INTO tracks VALUES(2112,'Dee',174,1,3,'R. Rhoads',261302,8555963,0.98999999999999999111);
INSERT INTO tracks VALUES(2113,'Shining In The Light',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',240796,7951688,0.98999999999999999111);
INSERT INTO tracks VALUES(2114,'When The World Was Young',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',373394,12198930,0.98999999999999999111);
INSERT INTO tracks VALUES(2115,'Upon A Golden Horse',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',232359,7594829,0.98999999999999999111);
INSERT INTO tracks VALUES(2116,'Blue Train',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',405028,13170391,0.98999999999999999111);
INSERT INTO tracks VALUES(2117,'Please Read The Letter',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',262112,8603372,0.98999999999999999111);
INSERT INTO tracks VALUES(2118,'Most High',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',336535,10999203,0.98999999999999999111);
INSERT INTO tracks VALUES(2119,'Heart In Your Hand',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',230896,7598019,0.98999999999999999111);
INSERT INTO tracks VALUES(2120,'Walking Into Clarksdale',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',318511,10396315,0.98999999999999999111);
INSERT INTO tracks VALUES(2121,'Burning Up',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',321619,10525136,0.98999999999999999111);
INSERT INTO tracks VALUES(2122,'When I Was A Child',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',345626,11249456,0.98999999999999999111);
INSERT INTO tracks VALUES(2123,'House Of Love',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',335699,10990880,0.98999999999999999111);
INSERT INTO tracks VALUES(2124,'Sons Of Freedom',175,1,1,'Jimmy Page, Robert Plant, Charlie Jones, Michael Lee',246465,8087944,0.98999999999999999111);
INSERT INTO tracks VALUES(2125,'United Colours',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',330266,10939131,0.98999999999999999111);
INSERT INTO tracks VALUES(2126,'Slug',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',281469,9295950,0.98999999999999999111);
INSERT INTO tracks VALUES(2127,'Your Blue Room',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',328228,10867860,0.98999999999999999111);
INSERT INTO tracks VALUES(2128,'Always Forever Now',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',383764,12727928,0.98999999999999999111);
INSERT INTO tracks VALUES(2129,'A Different Kind Of Blue',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',120816,3884133,0.98999999999999999111);
INSERT INTO tracks VALUES(2130,'Beach Sequence',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',212297,6928259,0.98999999999999999111);
INSERT INTO tracks VALUES(2131,'Miss Sarajevo',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',340767,11064884,0.98999999999999999111);
INSERT INTO tracks VALUES(2132,'Ito Okashi',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',205087,6572813,0.98999999999999999111);
INSERT INTO tracks VALUES(2133,'One Minute Warning',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',279693,9335453,0.98999999999999999111);
INSERT INTO tracks VALUES(2134,'Corpse (These Chains Are Way Too Long)',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',214909,6920451,0.98999999999999999111);
INSERT INTO tracks VALUES(2135,'Elvis Ate America',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',180166,5851053,0.98999999999999999111);
INSERT INTO tracks VALUES(2136,'Plot 180',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',221596,7253729,0.98999999999999999111);
INSERT INTO tracks VALUES(2137,'Theme From The Swan',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',203911,6638076,0.98999999999999999111);
INSERT INTO tracks VALUES(2138,'Theme From Let''s Go Native',176,1,10,'Brian Eno, Bono, Adam Clayton, The Edge & Larry Mullen Jnr.',186723,6179777,0.98999999999999999111);
INSERT INTO tracks VALUES(2139,'Wrathchild',177,1,1,'Steve Harris',170396,5499390,0.98999999999999999111);
INSERT INTO tracks VALUES(2140,'Killers',177,1,1,'Paul Di''Anno/Steve Harris',309995,10009697,0.98999999999999999111);
INSERT INTO tracks VALUES(2141,'Prowler',177,1,1,'Steve Harris',240274,7782963,0.98999999999999999111);
INSERT INTO tracks VALUES(2142,'Murders In The Rue Morgue',177,1,1,'Steve Harris',258638,8360999,0.98999999999999999111);
INSERT INTO tracks VALUES(2143,'Women In Uniform',177,1,1,'Greg Macainsh',189936,6139651,0.98999999999999999111);
INSERT INTO tracks VALUES(2144,'Remember Tomorrow',177,1,1,'Paul Di''Anno/Steve Harris',326426,10577976,0.98999999999999999111);
INSERT INTO tracks VALUES(2145,'Sanctuary',177,1,1,'David Murray/Paul Di''Anno/Steve Harris',198844,6423543,0.98999999999999999111);
INSERT INTO tracks VALUES(2146,'Running Free',177,1,1,'Paul Di''Anno/Steve Harris',199706,6483496,0.98999999999999999111);
INSERT INTO tracks VALUES(2147,'Phantom Of The Opera',177,1,1,'Steve Harris',418168,13585530,0.98999999999999999111);
INSERT INTO tracks VALUES(2148,'Iron Maiden',177,1,1,'Steve Harris',235232,7600077,0.98999999999999999111);
INSERT INTO tracks VALUES(2149,'Corduroy',178,1,1,'Pearl Jam & Eddie Vedder',305293,9991106,0.98999999999999999111);
INSERT INTO tracks VALUES(2150,'Given To Fly',178,1,1,'Eddie Vedder & Mike McCready',233613,7678347,0.98999999999999999111);
INSERT INTO tracks VALUES(2151,'Hail, Hail',178,1,1,'Stone Gossard & Eddie Vedder & Jeff Ament & Mike McCready',223764,7364206,0.98999999999999999111);
INSERT INTO tracks VALUES(2152,'Daughter',178,1,1,'Dave Abbruzzese & Jeff Ament & Stone Gossard & Mike McCready & Eddie Vedder',407484,13420697,0.98999999999999999111);
INSERT INTO tracks VALUES(2153,'Elderly Woman Behind The Counter In A Small Town',178,1,1,'Dave Abbruzzese & Jeff Ament & Stone Gossard & Mike McCready & Eddie Vedder',229328,7509304,0.98999999999999999111);
INSERT INTO tracks VALUES(2154,'Untitled',178,1,1,'Pearl Jam',122801,3957141,0.98999999999999999111);
INSERT INTO tracks VALUES(2155,'MFC',178,1,1,'Eddie Vedder',148192,4817665,0.98999999999999999111);
INSERT INTO tracks VALUES(2156,'Go',178,1,1,'Dave Abbruzzese & Jeff Ament & Stone Gossard & Mike McCready & Eddie Vedder',161541,5290810,0.98999999999999999111);
INSERT INTO tracks VALUES(2157,'Red Mosquito',178,1,1,'Jeff Ament & Stone Gossard & Jack Irons & Mike McCready & Eddie Vedder',242991,7944923,0.98999999999999999111);
INSERT INTO tracks VALUES(2158,'Even Flow',178,1,1,'Stone Gossard & Eddie Vedder',317100,10394239,0.98999999999999999111);
INSERT INTO tracks VALUES(2159,'Off He Goes',178,1,1,'Eddie Vedder',343222,11245109,0.98999999999999999111);
INSERT INTO tracks VALUES(2160,'Nothingman',178,1,1,'Jeff Ament & Eddie Vedder',278595,9107017,0.98999999999999999111);
INSERT INTO tracks VALUES(2161,'Do The Evolution',178,1,1,'Eddie Vedder & Stone Gossard',225462,7377286,0.98999999999999999111);
INSERT INTO tracks VALUES(2162,'Better Man',178,1,1,'Eddie Vedder',246204,8019563,0.98999999999999999111);
INSERT INTO tracks VALUES(2163,'Black',178,1,1,'Stone Gossard & Eddie Vedder',415712,13580009,0.98999999999999999111);
INSERT INTO tracks VALUES(2164,'F*Ckin'' Up',178,1,1,'Neil Young',377652,12360893,0.98999999999999999111);
INSERT INTO tracks VALUES(2165,'Life Wasted',179,1,4,'Stone Gossard',234344,7610169,0.98999999999999999111);
INSERT INTO tracks VALUES(2166,'World Wide Suicide',179,1,4,'Eddie Vedder',209188,6885908,0.98999999999999999111);
INSERT INTO tracks VALUES(2167,'Comatose',179,1,4,'Mike McCready & Stone Gossard',139990,4574516,0.98999999999999999111);
INSERT INTO tracks VALUES(2168,'Severed Hand',179,1,4,'Eddie Vedder',270341,8817438,0.98999999999999999111);
INSERT INTO tracks VALUES(2169,'Marker In The Sand',179,1,4,'Mike McCready',263235,8656578,0.98999999999999999111);
INSERT INTO tracks VALUES(2170,'Parachutes',179,1,4,'Stone Gossard',216555,7074973,0.98999999999999999111);
INSERT INTO tracks VALUES(2171,'Unemployable',179,1,4,'Matt Cameron & Mike McCready',184398,6066542,0.98999999999999999111);
INSERT INTO tracks VALUES(2172,'Big Wave',179,1,4,'Jeff Ament',178573,5858788,0.98999999999999999111);
INSERT INTO tracks VALUES(2173,'Gone',179,1,4,'Eddie Vedder',249547,8158204,0.98999999999999999111);
INSERT INTO tracks VALUES(2174,'Wasted Reprise',179,1,4,'Stone Gossard',53733,1731020,0.98999999999999999111);
INSERT INTO tracks VALUES(2175,'Army Reserve',179,1,4,'Jeff Ament',225567,7393771,0.98999999999999999111);
INSERT INTO tracks VALUES(2176,'Come Back',179,1,4,'Eddie Vedder & Mike McCready',329743,10768701,0.98999999999999999111);
INSERT INTO tracks VALUES(2177,'Inside Job',179,1,4,'Eddie Vedder & Mike McCready',428643,14006924,0.98999999999999999111);
INSERT INTO tracks VALUES(2178,'Can''t Keep',180,1,1,'Eddie Vedder',219428,7215713,0.98999999999999999111);
INSERT INTO tracks VALUES(2179,'Save You',180,1,1,'Eddie Vedder/Jeff Ament/Matt Cameron/Mike McCready/Stone Gossard',230112,7609110,0.98999999999999999111);
INSERT INTO tracks VALUES(2180,'Love Boat Captain',180,1,1,'Eddie Vedder',276453,9016789,0.98999999999999999111);
INSERT INTO tracks VALUES(2181,'Cropduster',180,1,1,'Matt Cameron',231888,7588928,0.98999999999999999111);
INSERT INTO tracks VALUES(2182,'Ghost',180,1,1,'Jeff Ament',195108,6383772,0.98999999999999999111);
INSERT INTO tracks VALUES(2183,'I Am Mine',180,1,1,'Eddie Vedder',215719,7086901,0.98999999999999999111);
INSERT INTO tracks VALUES(2184,'Thumbing My Way',180,1,1,'Eddie Vedder',250226,8201437,0.98999999999999999111);
INSERT INTO tracks VALUES(2185,'You Are',180,1,1,'Matt Cameron',270863,8938409,0.98999999999999999111);
INSERT INTO tracks VALUES(2186,'Get Right',180,1,1,'Matt Cameron',158589,5223345,0.98999999999999999111);
INSERT INTO tracks VALUES(2187,'Green Disease',180,1,1,'Eddie Vedder',161253,5375818,0.98999999999999999111);
INSERT INTO tracks VALUES(2188,'Help Help',180,1,1,'Jeff Ament',215092,7033002,0.98999999999999999111);
INSERT INTO tracks VALUES(2189,'Bushleager',180,1,1,'Stone Gossard',237479,7849757,0.98999999999999999111);
INSERT INTO tracks VALUES(2190,'1/2 Full',180,1,1,'Jeff Ament',251010,8197219,0.98999999999999999111);
INSERT INTO tracks VALUES(2191,'Arc',180,1,1,'Pearl Jam',65593,2099421,0.98999999999999999111);
INSERT INTO tracks VALUES(2192,'All or None',180,1,1,'Stone Gossard',277655,9104728,0.98999999999999999111);
INSERT INTO tracks VALUES(2193,'Once',181,1,1,'Stone Gossard',231758,7561555,0.98999999999999999111);
INSERT INTO tracks VALUES(2194,'Evenflow',181,1,1,'Stone Gossard',293720,9622017,0.98999999999999999111);
INSERT INTO tracks VALUES(2195,'Alive',181,1,1,'Stone Gossard',341080,11176623,0.98999999999999999111);
INSERT INTO tracks VALUES(2196,'Why Go',181,1,1,'Jeff Ament',200254,6539287,0.98999999999999999111);
INSERT INTO tracks VALUES(2197,'Black',181,1,1,'Dave Krusen/Stone Gossard',343823,11213314,0.98999999999999999111);
INSERT INTO tracks VALUES(2198,'Jeremy',181,1,1,'Jeff Ament',318981,10447222,0.98999999999999999111);
INSERT INTO tracks VALUES(2199,'Oceans',181,1,1,'Jeff Ament/Stone Gossard',162194,5282368,0.98999999999999999111);
INSERT INTO tracks VALUES(2200,'Porch',181,1,1,'Eddie Vedder',210520,6877475,0.98999999999999999111);
INSERT INTO tracks VALUES(2201,'Garden',181,1,1,'Jeff Ament/Stone Gossard',299154,9740738,0.98999999999999999111);
INSERT INTO tracks VALUES(2202,'Deep',181,1,1,'Jeff Ament/Stone Gossard',258324,8432497,0.98999999999999999111);
INSERT INTO tracks VALUES(2203,'Release',181,1,1,'Jeff Ament/Mike McCready/Stone Gossard',546063,17802673,0.98999999999999999111);
INSERT INTO tracks VALUES(2204,'Go',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',193123,6351920,0.98999999999999999111);
INSERT INTO tracks VALUES(2205,'Animal',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',169325,5503459,0.98999999999999999111);
INSERT INTO tracks VALUES(2206,'Daughter',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',235598,7824586,0.98999999999999999111);
INSERT INTO tracks VALUES(2207,'Glorified G',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',206968,6772116,0.98999999999999999111);
INSERT INTO tracks VALUES(2208,'Dissident',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',215510,7034500,0.98999999999999999111);
INSERT INTO tracks VALUES(2209,'W.M.A.',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',359262,12037261,0.98999999999999999111);
INSERT INTO tracks VALUES(2210,'Blood',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',170631,5551478,0.98999999999999999111);
INSERT INTO tracks VALUES(2211,'Rearviewmirror',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',284186,9321053,0.98999999999999999111);
INSERT INTO tracks VALUES(2212,'Rats',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',255425,8341934,0.98999999999999999111);
INSERT INTO tracks VALUES(2213,'Elderly Woman Behind The Counter In A Small Town',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',196336,6499398,0.98999999999999999111);
INSERT INTO tracks VALUES(2214,'Leash',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',189257,6191560,0.98999999999999999111);
INSERT INTO tracks VALUES(2215,'Indifference',182,1,1,'Dave Abbruzzese/Eddie Vedder/Jeff Ament/Mike McCready/Stone Gossard',302053,9756133,0.98999999999999999111);
INSERT INTO tracks VALUES(2216,'Johnny B. Goode',141,1,8,NULL,243200,8092024,0.98999999999999999111);
INSERT INTO tracks VALUES(2217,'Don''t Look Back',141,1,8,NULL,221100,7344023,0.98999999999999999111);
INSERT INTO tracks VALUES(2218,'Jah Seh No',141,1,8,NULL,276871,9134476,0.98999999999999999111);
INSERT INTO tracks VALUES(2219,'I''m The Toughest',141,1,8,NULL,230191,7657594,0.98999999999999999111);
INSERT INTO tracks VALUES(2220,'Nothing But Love',141,1,8,NULL,221570,7335228,0.98999999999999999111);
INSERT INTO tracks VALUES(2221,'Buk-In-Hamm Palace',141,1,8,NULL,265665,8964369,0.98999999999999999111);
INSERT INTO tracks VALUES(2222,'Bush Doctor',141,1,8,NULL,239751,7942299,0.98999999999999999111);
INSERT INTO tracks VALUES(2223,'Wanted Dread And Alive',141,1,8,NULL,260310,8670933,0.98999999999999999111);
INSERT INTO tracks VALUES(2224,'Mystic Man',141,1,8,NULL,353671,11812170,0.98999999999999999111);
INSERT INTO tracks VALUES(2225,'Coming In Hot',141,1,8,NULL,213054,7109414,0.98999999999999999111);
INSERT INTO tracks VALUES(2226,'Pick Myself Up',141,1,8,NULL,234684,7788255,0.98999999999999999111);
INSERT INTO tracks VALUES(2227,'Crystal Ball',141,1,8,NULL,309733,10319296,0.98999999999999999111);
INSERT INTO tracks VALUES(2228,'Equal Rights Downpresser Man',141,1,8,NULL,366733,12086524,0.98999999999999999111);
INSERT INTO tracks VALUES(2229,'Speak To Me/Breathe',183,1,1,'Mason/Waters, Gilmour, Wright',234213,7631305,0.98999999999999999111);
INSERT INTO tracks VALUES(2230,'On The Run',183,1,1,'Gilmour, Waters',214595,7206300,0.98999999999999999111);
INSERT INTO tracks VALUES(2231,'Time',183,1,1,'Mason, Waters, Wright, Gilmour',425195,13955426,0.98999999999999999111);
INSERT INTO tracks VALUES(2232,'The Great Gig In The Sky',183,1,1,'Wright, Waters',284055,9147563,0.98999999999999999111);
INSERT INTO tracks VALUES(2233,'Money',183,1,1,'Waters',391888,12930070,0.98999999999999999111);
INSERT INTO tracks VALUES(2234,'Us And Them',183,1,1,'Waters, Wright',461035,15000299,0.98999999999999999111);
INSERT INTO tracks VALUES(2235,'Any Colour You Like',183,1,1,'Gilmour, Mason, Wright, Waters',205740,6707989,0.98999999999999999111);
INSERT INTO tracks VALUES(2236,'Brain Damage',183,1,1,'Waters',230556,7497655,0.98999999999999999111);
INSERT INTO tracks VALUES(2237,'Eclipse',183,1,1,'Waters',125361,4065299,0.98999999999999999111);
INSERT INTO tracks VALUES(2238,'ZeroVinteUm',184,1,17,NULL,315637,10426550,0.98999999999999999111);
INSERT INTO tracks VALUES(2239,'Queimando Tudo',184,1,17,NULL,172591,5723677,0.98999999999999999111);
INSERT INTO tracks VALUES(2240,'Hip Hop Rio',184,1,17,NULL,151536,4991935,0.98999999999999999111);
INSERT INTO tracks VALUES(2241,'Bossa',184,1,17,NULL,29048,967098,0.98999999999999999111);
INSERT INTO tracks VALUES(2242,'100% HardCore',184,1,17,NULL,165146,5407744,0.98999999999999999111);
INSERT INTO tracks VALUES(2243,'Biruta',184,1,17,NULL,213263,7108200,0.98999999999999999111);
INSERT INTO tracks VALUES(2244,'Mão Na Cabeça',184,1,17,NULL,202631,6642753,0.98999999999999999111);
INSERT INTO tracks VALUES(2245,'O Bicho Tá Pregando',184,1,17,NULL,171964,5683369,0.98999999999999999111);
INSERT INTO tracks VALUES(2246,'Adoled (Ocean)',184,1,17,NULL,185103,6009946,0.98999999999999999111);
INSERT INTO tracks VALUES(2247,'Seus Amigos',184,1,17,NULL,100858,3304738,0.98999999999999999111);
INSERT INTO tracks VALUES(2248,'Paga Pau',184,1,17,NULL,197485,6529041,0.98999999999999999111);
INSERT INTO tracks VALUES(2249,'Rappers Reais',184,1,17,NULL,202004,6684160,0.98999999999999999111);
INSERT INTO tracks VALUES(2250,'Nega Do Cabelo Duro',184,1,17,NULL,121808,4116536,0.98999999999999999111);
INSERT INTO tracks VALUES(2251,'Hemp Family',184,1,17,NULL,205923,6806900,0.98999999999999999111);
INSERT INTO tracks VALUES(2252,'Quem Me Cobrou?',184,1,17,NULL,121704,3947664,0.98999999999999999111);
INSERT INTO tracks VALUES(2253,'Se Liga',184,1,17,NULL,410409,13559173,0.98999999999999999111);
INSERT INTO tracks VALUES(2254,'Bohemian Rhapsody',185,1,1,'Mercury, Freddie',358948,11619868,0.98999999999999999111);
INSERT INTO tracks VALUES(2255,'Another One Bites The Dust',185,1,1,'Deacon, John',216946,7172355,0.98999999999999999111);
INSERT INTO tracks VALUES(2256,'Killer Queen',185,1,1,'Mercury, Freddie',182099,5967749,0.98999999999999999111);
INSERT INTO tracks VALUES(2257,'Fat Bottomed Girls',185,1,1,'May, Brian',204695,6630041,0.98999999999999999111);
INSERT INTO tracks VALUES(2258,'Bicycle Race',185,1,1,'Mercury, Freddie',183823,6012409,0.98999999999999999111);
INSERT INTO tracks VALUES(2259,'You''re My Best Friend',185,1,1,'Deacon, John',172225,5602173,0.98999999999999999111);
INSERT INTO tracks VALUES(2260,'Don''t Stop Me Now',185,1,1,'Mercury, Freddie',211826,6896666,0.98999999999999999111);
INSERT INTO tracks VALUES(2261,'Save Me',185,1,1,'May, Brian',228832,7444624,0.98999999999999999111);
INSERT INTO tracks VALUES(2262,'Crazy Little Thing Called Love',185,1,1,'Mercury, Freddie',164231,5435501,0.98999999999999999111);
INSERT INTO tracks VALUES(2263,'Somebody To Love',185,1,1,'Mercury, Freddie',297351,9650520,0.98999999999999999111);
INSERT INTO tracks VALUES(2264,'Now I''m Here',185,1,1,'May, Brian',255346,8328312,0.98999999999999999111);
INSERT INTO tracks VALUES(2265,'Good Old-Fashioned Lover Boy',185,1,1,'Mercury, Freddie',175960,5747506,0.98999999999999999111);
INSERT INTO tracks VALUES(2266,'Play The Game',185,1,1,'Mercury, Freddie',213368,6915832,0.98999999999999999111);
INSERT INTO tracks VALUES(2267,'Flash',185,1,1,'May, Brian',168489,5464986,0.98999999999999999111);
INSERT INTO tracks VALUES(2268,'Seven Seas Of Rhye',185,1,1,'Mercury, Freddie',170553,5539957,0.98999999999999999111);
INSERT INTO tracks VALUES(2269,'We Will Rock You',185,1,1,'Deacon, John/May, Brian',122880,4026955,0.98999999999999999111);
INSERT INTO tracks VALUES(2270,'We Are The Champions',185,1,1,'Mercury, Freddie',180950,5880231,0.98999999999999999111);
INSERT INTO tracks VALUES(2271,'We Will Rock You',186,1,1,'May',122671,4026815,0.98999999999999999111);
INSERT INTO tracks VALUES(2272,'We Are The Champions',186,1,1,'Mercury',182883,5939794,0.98999999999999999111);
INSERT INTO tracks VALUES(2273,'Sheer Heart Attack',186,1,1,'Taylor',207386,6642685,0.98999999999999999111);
INSERT INTO tracks VALUES(2274,'All Dead, All Dead',186,1,1,'May',190119,6144878,0.98999999999999999111);
INSERT INTO tracks VALUES(2275,'Spread Your Wings',186,1,1,'Deacon',275356,8936992,0.98999999999999999111);
INSERT INTO tracks VALUES(2276,'Fight From The Inside',186,1,1,'Taylor',184737,6078001,0.98999999999999999111);
INSERT INTO tracks VALUES(2277,'Get Down, Make Love',186,1,1,'Mercury',231235,7509333,0.98999999999999999111);
INSERT INTO tracks VALUES(2278,'Sleep On The Sidewalk',186,1,1,'May',187428,6099840,0.98999999999999999111);
INSERT INTO tracks VALUES(2279,'Who Needs You',186,1,1,'Deacon',186958,6292969,0.98999999999999999111);
INSERT INTO tracks VALUES(2280,'It''s Late',186,1,1,'May',386194,12519388,0.98999999999999999111);
INSERT INTO tracks VALUES(2281,'My Melancholy Blues',186,1,1,'Mercury',206471,6691838,0.98999999999999999111);
INSERT INTO tracks VALUES(2282,'Shiny Happy People',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',226298,7475323,0.98999999999999999111);
INSERT INTO tracks VALUES(2283,'Me In Honey',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',246674,8194751,0.98999999999999999111);
INSERT INTO tracks VALUES(2284,'Radio Song',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',255477,8421172,0.98999999999999999111);
INSERT INTO tracks VALUES(2285,'Pop Song 89',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',185730,6132218,0.98999999999999999111);
INSERT INTO tracks VALUES(2286,'Get Up',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',160235,5264376,0.98999999999999999111);
INSERT INTO tracks VALUES(2287,'You Are The Everything',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',226298,7373181,0.98999999999999999111);
INSERT INTO tracks VALUES(2288,'Stand',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',192862,6349090,0.98999999999999999111);
INSERT INTO tracks VALUES(2289,'World Leader Pretend',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',259761,8537282,0.98999999999999999111);
INSERT INTO tracks VALUES(2290,'The Wrong Child',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',216633,7065060,0.98999999999999999111);
INSERT INTO tracks VALUES(2291,'Orange Crush',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',231706,7742894,0.98999999999999999111);
INSERT INTO tracks VALUES(2292,'Turn You Inside-Out',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',257358,8395671,0.98999999999999999111);
INSERT INTO tracks VALUES(2293,'Hairshirt',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',235911,7753807,0.98999999999999999111);
INSERT INTO tracks VALUES(2294,'I Remember California',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',304013,9950311,0.98999999999999999111);
INSERT INTO tracks VALUES(2295,'Untitled',188,1,4,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',191503,6332426,0.98999999999999999111);
INSERT INTO tracks VALUES(2296,'How The West Was Won And Where It Got Us',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',271151,8994291,0.98999999999999999111);
INSERT INTO tracks VALUES(2297,'The Wake-Up Bomb',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',308532,10077337,0.98999999999999999111);
INSERT INTO tracks VALUES(2298,'New Test Leper',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',326791,10866447,0.98999999999999999111);
INSERT INTO tracks VALUES(2299,'Undertow',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',309498,10131005,0.98999999999999999111);
INSERT INTO tracks VALUES(2300,'E-Bow The Letter',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',324963,10714576,0.98999999999999999111);
INSERT INTO tracks VALUES(2301,'Leave',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',437968,14433365,0.98999999999999999111);
INSERT INTO tracks VALUES(2302,'Departure',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',209423,6818425,0.98999999999999999111);
INSERT INTO tracks VALUES(2303,'Bittersweet Me',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',245812,8114718,0.98999999999999999111);
INSERT INTO tracks VALUES(2304,'Be Mine',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',333087,10790541,0.98999999999999999111);
INSERT INTO tracks VALUES(2305,'Binky The Doormat',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',301688,9950320,0.98999999999999999111);
INSERT INTO tracks VALUES(2306,'Zither',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',154148,5032962,0.98999999999999999111);
INSERT INTO tracks VALUES(2307,'So Fast, So Numb',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',252682,8341223,0.98999999999999999111);
INSERT INTO tracks VALUES(2308,'Low Desert',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',212062,6989288,0.98999999999999999111);
INSERT INTO tracks VALUES(2309,'Electrolite',189,1,1,'Bill Berry-Peter Buck-Mike Mills-Michael Stipe',245315,8051199,0.98999999999999999111);
INSERT INTO tracks VALUES(2310,'Losing My Religion',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',269035,8885672,0.98999999999999999111);
INSERT INTO tracks VALUES(2311,'Low',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',296777,9633860,0.98999999999999999111);
INSERT INTO tracks VALUES(2312,'Near Wild Heaven',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',199862,6610009,0.98999999999999999111);
INSERT INTO tracks VALUES(2313,'Endgame',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',230687,7664479,0.98999999999999999111);
INSERT INTO tracks VALUES(2314,'Belong',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',247013,8219375,0.98999999999999999111);
INSERT INTO tracks VALUES(2315,'Half A World Away',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',208431,6837283,0.98999999999999999111);
INSERT INTO tracks VALUES(2316,'Texarkana',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',220081,7260681,0.98999999999999999111);
INSERT INTO tracks VALUES(2317,'Country Feedback',187,1,4,'Bill Berry/Michael Stipe/Mike Mills/Peter Buck',249782,8178943,0.98999999999999999111);
INSERT INTO tracks VALUES(2318,'Carnival Of Sorts',190,1,4,'R.E.M.',233482,7669658,0.98999999999999999111);
INSERT INTO tracks VALUES(2319,'Radio Free Aurope',190,1,4,'R.E.M.',245315,8163490,0.98999999999999999111);
INSERT INTO tracks VALUES(2320,'Perfect Circle',190,1,4,'R.E.M.',208509,6898067,0.98999999999999999111);
INSERT INTO tracks VALUES(2321,'Talk About The Passion',190,1,4,'R.E.M.',203206,6725435,0.98999999999999999111);
INSERT INTO tracks VALUES(2322,'So Central Rain',190,1,4,'R.E.M.',194768,6414550,0.98999999999999999111);
INSERT INTO tracks VALUES(2323,'Don''t Go Back To Rockville',190,1,4,'R.E.M.',272352,9010715,0.98999999999999999111);
INSERT INTO tracks VALUES(2324,'Pretty Persuasion',190,1,4,'R.E.M.',229929,7577754,0.98999999999999999111);
INSERT INTO tracks VALUES(2325,'Green Grow The Rushes',190,1,4,'R.E.M.',225671,7422425,0.98999999999999999111);
INSERT INTO tracks VALUES(2326,'Can''t Get There From Here',190,1,4,'R.E.M.',220630,7285936,0.98999999999999999111);
INSERT INTO tracks VALUES(2327,'Driver 8',190,1,4,'R.E.M.',204747,6779076,0.98999999999999999111);
INSERT INTO tracks VALUES(2328,'Fall On Me',190,1,4,'R.E.M.',172016,5676811,0.98999999999999999111);
INSERT INTO tracks VALUES(2329,'I Believe',190,1,4,'R.E.M.',227709,7542929,0.98999999999999999111);
INSERT INTO tracks VALUES(2330,'Cuyahoga',190,1,4,'R.E.M.',260623,8591057,0.98999999999999999111);
INSERT INTO tracks VALUES(2331,'The One I Love',190,1,4,'R.E.M.',197355,6495125,0.98999999999999999111);
INSERT INTO tracks VALUES(2332,'The Finest Worksong',190,1,4,'R.E.M.',229276,7574856,0.98999999999999999111);
INSERT INTO tracks VALUES(2333,'It''s The End Of The World As We Know It (And I Feel Fine)',190,1,4,'R.E.M.',244819,7998987,0.98999999999999999111);
INSERT INTO tracks VALUES(2334,'Infeliz Natal',191,1,4,'Rodolfo',138266,4503299,0.98999999999999999111);
INSERT INTO tracks VALUES(2335,'A Sua',191,1,4,'Rodolfo',142132,4622064,0.98999999999999999111);
INSERT INTO tracks VALUES(2336,'Papeau Nuky Doe',191,1,4,'Rodolfo',121652,3995022,0.98999999999999999111);
INSERT INTO tracks VALUES(2337,'Merry Christmas',191,1,4,'Rodolfo',126040,4166652,0.98999999999999999111);
INSERT INTO tracks VALUES(2338,'Bodies',191,1,4,'Rodolfo',180035,5873778,0.98999999999999999111);
INSERT INTO tracks VALUES(2339,'Puteiro Em João Pessoa',191,1,4,'Rodolfo',195578,6395490,0.98999999999999999111);
INSERT INTO tracks VALUES(2340,'Esporrei Na Manivela',191,1,4,'Rodolfo',293276,9618499,0.98999999999999999111);
INSERT INTO tracks VALUES(2341,'Bê-a-Bá',191,1,4,'Rodolfo',249051,8130636,0.98999999999999999111);
INSERT INTO tracks VALUES(2342,'Cajueiro',191,1,4,'Rodolfo',158589,5164837,0.98999999999999999111);
INSERT INTO tracks VALUES(2343,'Palhas Do Coqueiro',191,1,4,'Rodolfo',133851,4396466,0.98999999999999999111);
INSERT INTO tracks VALUES(2344,'Maluco Beleza',192,1,1,NULL,203206,6628067,0.98999999999999999111);
INSERT INTO tracks VALUES(2345,'O Dia Em Que A Terra Parou',192,1,1,NULL,261720,8586678,0.98999999999999999111);
INSERT INTO tracks VALUES(2346,'No Fundo Do Quintal Da Escola',192,1,1,NULL,177606,5836953,0.98999999999999999111);
INSERT INTO tracks VALUES(2347,'O Segredo Do Universo',192,1,1,NULL,192679,6315187,0.98999999999999999111);
INSERT INTO tracks VALUES(2348,'As Profecias',192,1,1,NULL,232515,7657732,0.98999999999999999111);
INSERT INTO tracks VALUES(2349,'Mata Virgem',192,1,1,NULL,142602,4690029,0.98999999999999999111);
INSERT INTO tracks VALUES(2350,'Sapato 36',192,1,1,NULL,196702,6507301,0.98999999999999999111);
INSERT INTO tracks VALUES(2351,'Todo Mundo Explica',192,1,1,NULL,134896,4449772,0.98999999999999999111);
INSERT INTO tracks VALUES(2352,'Que Luz É Essa',192,1,1,NULL,165067,5620058,0.98999999999999999111);
INSERT INTO tracks VALUES(2353,'Diamante De Mendigo',192,1,1,NULL,206053,6775101,0.98999999999999999111);
INSERT INTO tracks VALUES(2354,'Negócio É',192,1,1,NULL,175464,5826775,0.98999999999999999111);
INSERT INTO tracks VALUES(2355,'Muita Estrela, Pouca Constelação',192,1,1,NULL,268068,8781021,0.98999999999999999111);
INSERT INTO tracks VALUES(2356,'Século XXI',192,1,1,NULL,244897,8040563,0.98999999999999999111);
INSERT INTO tracks VALUES(2357,'Rock Das Aranhas (Ao Vivo) (Live)',192,1,1,NULL,231836,7591945,0.98999999999999999111);
INSERT INTO tracks VALUES(2358,'The Power Of Equality',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',243591,8148266,0.98999999999999999111);
INSERT INTO tracks VALUES(2359,'If You Have To Ask',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',216790,7199175,0.98999999999999999111);
INSERT INTO tracks VALUES(2360,'Breaking The Girl',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',295497,9805526,0.98999999999999999111);
INSERT INTO tracks VALUES(2361,'Funky Monks',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',323395,10708168,0.98999999999999999111);
INSERT INTO tracks VALUES(2362,'Suck My Kiss',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',217234,7129137,0.98999999999999999111);
INSERT INTO tracks VALUES(2363,'I Could Have Lied',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',244506,8088244,0.98999999999999999111);
INSERT INTO tracks VALUES(2364,'Mellowship Slinky In B Major',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',240091,7971384,0.98999999999999999111);
INSERT INTO tracks VALUES(2365,'The Righteous & The Wicked',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',248084,8134096,0.98999999999999999111);
INSERT INTO tracks VALUES(2366,'Give It Away',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',283010,9308997,0.98999999999999999111);
INSERT INTO tracks VALUES(2367,'Blood Sugar Sex Magik',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',271229,8940573,0.98999999999999999111);
INSERT INTO tracks VALUES(2368,'Under The Bridge',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',264359,8682716,0.98999999999999999111);
INSERT INTO tracks VALUES(2369,'Naked In The Rain',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',265717,8724674,0.98999999999999999111);
INSERT INTO tracks VALUES(2370,'Apache Rose Peacock',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',282226,9312588,0.98999999999999999111);
INSERT INTO tracks VALUES(2371,'The Greeting Song',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',193593,6346507,0.98999999999999999111);
INSERT INTO tracks VALUES(2372,'My Lovely Man',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',279118,9220114,0.98999999999999999111);
INSERT INTO tracks VALUES(2373,'Sir Psycho Sexy',193,1,4,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',496692,16354362,0.98999999999999999111);
INSERT INTO tracks VALUES(2374,'They''re Red Hot',193,1,4,'Robert Johnson',71941,2382220,0.98999999999999999111);
INSERT INTO tracks VALUES(2375,'By The Way',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',218017,7197430,0.98999999999999999111);
INSERT INTO tracks VALUES(2376,'Universally Speaking',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',259213,8501904,0.98999999999999999111);
INSERT INTO tracks VALUES(2377,'This Is The Place',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',257906,8469765,0.98999999999999999111);
INSERT INTO tracks VALUES(2378,'Dosed',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',312058,10235611,0.98999999999999999111);
INSERT INTO tracks VALUES(2379,'Don''t Forget Me',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',277995,9107071,0.98999999999999999111);
INSERT INTO tracks VALUES(2380,'The Zephyr Song',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',232960,7690312,0.98999999999999999111);
INSERT INTO tracks VALUES(2381,'Can''t Stop',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',269400,8872479,0.98999999999999999111);
INSERT INTO tracks VALUES(2382,'I Could Die For You',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',193906,6333311,0.98999999999999999111);
INSERT INTO tracks VALUES(2383,'Midnight',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',295810,9702450,0.98999999999999999111);
INSERT INTO tracks VALUES(2384,'Throw Away Your Television',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',224574,7483526,0.98999999999999999111);
INSERT INTO tracks VALUES(2385,'Cabron',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',218592,7458864,0.98999999999999999111);
INSERT INTO tracks VALUES(2386,'Tear',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',317413,10395500,0.98999999999999999111);
INSERT INTO tracks VALUES(2387,'On Mercury',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',208509,6834762,0.98999999999999999111);
INSERT INTO tracks VALUES(2388,'Minor Thing',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',217835,7148115,0.98999999999999999111);
INSERT INTO tracks VALUES(2389,'Warm Tape',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',256653,8358200,0.98999999999999999111);
INSERT INTO tracks VALUES(2390,'Venice Queen',194,1,1,'Anthony Kiedis, Flea, John Frusciante, and Chad Smith',369110,12280381,0.98999999999999999111);
INSERT INTO tracks VALUES(2391,'Around The World',195,1,1,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',238837,7859167,0.98999999999999999111);
INSERT INTO tracks VALUES(2392,'Parallel Universe',195,1,1,'Red Hot Chili Peppers',270654,8958519,0.98999999999999999111);
INSERT INTO tracks VALUES(2393,'Scar Tissue',195,1,1,'Red Hot Chili Peppers',217469,7153744,0.98999999999999999111);
INSERT INTO tracks VALUES(2394,'Otherside',195,1,1,'Red Hot Chili Peppers',255973,8357989,0.98999999999999999111);
INSERT INTO tracks VALUES(2395,'Get On Top',195,1,1,'Red Hot Chili Peppers',198164,6587883,0.98999999999999999111);
INSERT INTO tracks VALUES(2396,'Californication',195,1,1,'Red Hot Chili Peppers',321671,10568999,0.98999999999999999111);
INSERT INTO tracks VALUES(2397,'Easily',195,1,1,'Red Hot Chili Peppers',231418,7504534,0.98999999999999999111);
INSERT INTO tracks VALUES(2398,'Porcelain',195,1,1,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',163787,5278793,0.98999999999999999111);
INSERT INTO tracks VALUES(2399,'Emit Remmus',195,1,1,'Red Hot Chili Peppers',240300,7901717,0.98999999999999999111);
INSERT INTO tracks VALUES(2400,'I Like Dirt',195,1,1,'Red Hot Chili Peppers',157727,5225917,0.98999999999999999111);
INSERT INTO tracks VALUES(2401,'This Velvet Glove',195,1,1,'Red Hot Chili Peppers',225280,7480537,0.98999999999999999111);
INSERT INTO tracks VALUES(2402,'Savior',195,1,1,'Anthony Kiedis/Chad Smith/Flea/John Frusciante',292493,9551885,0.98999999999999999111);
INSERT INTO tracks VALUES(2403,'Purple Stain',195,1,1,'Red Hot Chili Peppers',253440,8359971,0.98999999999999999111);
INSERT INTO tracks VALUES(2404,'Right On Time',195,1,1,'Red Hot Chili Peppers',112613,3722219,0.98999999999999999111);
INSERT INTO tracks VALUES(2405,'Road Trippin''',195,1,1,'Red Hot Chili Peppers',205635,6685831,0.98999999999999999111);
INSERT INTO tracks VALUES(2406,'The Spirit Of Radio',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',299154,9862012,0.98999999999999999111);
INSERT INTO tracks VALUES(2407,'The Trees',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',285126,9345473,0.98999999999999999111);
INSERT INTO tracks VALUES(2408,'Something For Nothing',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',240770,7898395,0.98999999999999999111);
INSERT INTO tracks VALUES(2409,'Freewill',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',324362,10694110,0.98999999999999999111);
INSERT INTO tracks VALUES(2410,'Xanadu',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',667428,21753168,0.98999999999999999111);
INSERT INTO tracks VALUES(2411,'Bastille Day',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',280528,9264769,0.98999999999999999111);
INSERT INTO tracks VALUES(2412,'By-Tor And The Snow Dog',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',519888,17076397,0.98999999999999999111);
INSERT INTO tracks VALUES(2413,'Anthem',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',264515,8693343,0.98999999999999999111);
INSERT INTO tracks VALUES(2414,'Closer To The Heart',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',175412,5767005,0.98999999999999999111);
INSERT INTO tracks VALUES(2415,'2112 Overture',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',272718,8898066,0.98999999999999999111);
INSERT INTO tracks VALUES(2416,'The Temples Of Syrinx',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',133459,4360163,0.98999999999999999111);
INSERT INTO tracks VALUES(2417,'La Villa Strangiato',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',577488,19137855,0.98999999999999999111);
INSERT INTO tracks VALUES(2418,'Fly By Night',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',202318,6683061,0.98999999999999999111);
INSERT INTO tracks VALUES(2419,'Finding My Way',196,1,1,'Geddy Lee And Alex Lifeson/Geddy Lee And Neil Peart/Rush',305528,9985701,0.98999999999999999111);
INSERT INTO tracks VALUES(2420,'Jingo',197,1,1,'M.Babatunde Olantunji',592953,19736495,0.98999999999999999111);
INSERT INTO tracks VALUES(2421,'El Corazon Manda',197,1,1,'E.Weiss',713534,23519583,0.98999999999999999111);
INSERT INTO tracks VALUES(2422,'La Puesta Del Sol',197,1,1,'E.Weiss',628062,20614621,0.98999999999999999111);
INSERT INTO tracks VALUES(2423,'Persuasion',197,1,1,'Carlos Santana',318432,10354751,0.98999999999999999111);
INSERT INTO tracks VALUES(2424,'As The Years Go by',197,1,1,'Albert King',233064,7566829,0.98999999999999999111);
INSERT INTO tracks VALUES(2425,'Soul Sacrifice',197,1,1,'Carlos Santana',296437,9801120,0.98999999999999999111);
INSERT INTO tracks VALUES(2426,'Fried Neckbones And Home Fries',197,1,1,'W.Correa',638563,20939646,0.98999999999999999111);
INSERT INTO tracks VALUES(2427,'Santana Jam',197,1,1,'Carlos Santana',882834,29207100,0.98999999999999999111);
INSERT INTO tracks VALUES(2428,'Evil Ways',198,1,1,NULL,475402,15289235,0.98999999999999999111);
INSERT INTO tracks VALUES(2429,'We''ve Got To Get Together/Jingo',198,1,1,NULL,1070027,34618222,0.98999999999999999111);
INSERT INTO tracks VALUES(2430,'Rock Me',198,1,1,NULL,94720,3037596,0.98999999999999999111);
INSERT INTO tracks VALUES(2431,'Just Ain''t Good Enough',198,1,1,NULL,850259,27489067,0.98999999999999999111);
INSERT INTO tracks VALUES(2432,'Funky Piano',198,1,1,NULL,934791,30200730,0.98999999999999999111);
INSERT INTO tracks VALUES(2433,'The Way You Do To Mer',198,1,1,NULL,618344,20028702,0.98999999999999999111);
INSERT INTO tracks VALUES(2434,'Holding Back The Years',141,1,1,'Mick Hucknall and Neil Moss',270053,8833220,0.98999999999999999111);
INSERT INTO tracks VALUES(2435,'Money''s Too Tight To Mention',141,1,1,'John and William Valentine',268408,8861921,0.98999999999999999111);
INSERT INTO tracks VALUES(2436,'The Right Thing',141,1,1,'Mick Hucknall',262687,8624063,0.98999999999999999111);
INSERT INTO tracks VALUES(2437,'It''s Only Love',141,1,1,'Jimmy and Vella Cameron',232594,7659017,0.98999999999999999111);
INSERT INTO tracks VALUES(2438,'A New Flame',141,1,1,'Mick Hucknall',237662,7822875,0.98999999999999999111);
INSERT INTO tracks VALUES(2439,'You''ve Got It',141,1,1,'Mick Hucknall and Lamont Dozier',235232,7712845,0.98999999999999999111);
INSERT INTO tracks VALUES(2440,'If You Don''t Know Me By Now',141,1,1,'Kenny Gamble and Leon Huff',206524,6712634,0.98999999999999999111);
INSERT INTO tracks VALUES(2441,'Stars',141,1,1,'Mick Hucknall',248137,8194906,0.98999999999999999111);
INSERT INTO tracks VALUES(2442,'Something Got Me Started',141,1,1,'Mick Hucknall and Fritz McIntyre',239595,7997139,0.98999999999999999111);
INSERT INTO tracks VALUES(2443,'Thrill Me',141,1,1,'Mick Hucknall and Fritz McIntyre',303934,10034711,0.98999999999999999111);
INSERT INTO tracks VALUES(2444,'Your Mirror',141,1,1,'Mick Hucknall',240666,7893821,0.98999999999999999111);
INSERT INTO tracks VALUES(2445,'For Your Babies',141,1,1,'Mick Hucknall',256992,8408803,0.98999999999999999111);
INSERT INTO tracks VALUES(2446,'So Beautiful',141,1,1,'Mick Hucknall',298083,9837832,0.98999999999999999111);
INSERT INTO tracks VALUES(2447,'Angel',141,1,1,'Carolyn Franklin and Sonny Saunders',240561,7880256,0.98999999999999999111);
INSERT INTO tracks VALUES(2448,'Fairground',141,1,1,'Mick Hucknall',263888,8793094,0.98999999999999999111);
INSERT INTO tracks VALUES(2449,'Água E Fogo',199,1,1,'Chico Amaral/Edgard Scandurra/Samuel Rosa',278987,9272272,0.98999999999999999111);
INSERT INTO tracks VALUES(2450,'Três Lados',199,1,1,'Chico Amaral/Samuel Rosa',233665,7699609,0.98999999999999999111);
INSERT INTO tracks VALUES(2451,'Ela Desapareceu',199,1,1,'Chico Amaral/Samuel Rosa',250122,8289200,0.98999999999999999111);
INSERT INTO tracks VALUES(2452,'Balada Do Amor Inabalável',199,1,1,'Fausto Fawcett/Samuel Rosa',240613,8025816,0.98999999999999999111);
INSERT INTO tracks VALUES(2453,'Canção Noturna',199,1,1,'Chico Amaral/Lelo Zanettik',238628,7874774,0.98999999999999999111);
INSERT INTO tracks VALUES(2454,'Muçulmano',199,1,1,'Leão, Rodrigo F./Samuel Rosa',249600,8270613,0.98999999999999999111);
INSERT INTO tracks VALUES(2455,'Maquinarama',199,1,1,'Chico Amaral/Samuel Rosa',245629,8213710,0.98999999999999999111);
INSERT INTO tracks VALUES(2456,'Rebelião',199,1,1,'Chico Amaral/Samuel Rosa',298527,9817847,0.98999999999999999111);
INSERT INTO tracks VALUES(2457,'A Última Guerra',199,1,1,'Leão, Rodrigo F./Lô Borges/Samuel Rosa',314723,10480391,0.98999999999999999111);
INSERT INTO tracks VALUES(2458,'Fica',199,1,1,'Chico Amaral/Samuel Rosa',272169,8980972,0.98999999999999999111);
INSERT INTO tracks VALUES(2459,'Ali',199,1,1,'Nando Reis/Samuel Rosa',306390,10110351,0.98999999999999999111);
INSERT INTO tracks VALUES(2460,'Preto Damião',199,1,1,'Chico Amaral/Samuel Rosa',264568,8697658,0.98999999999999999111);
INSERT INTO tracks VALUES(2461,'É Uma Partida De Futebol',200,1,1,'Samuel Rosa',1071,38747,0.98999999999999999111);
INSERT INTO tracks VALUES(2462,'Eu Disse A Ela',200,1,1,'Samuel Rosa',254223,8479463,0.98999999999999999111);
INSERT INTO tracks VALUES(2463,'Zé Trindade',200,1,1,'Samuel Rosa',247954,8331310,0.98999999999999999111);
INSERT INTO tracks VALUES(2464,'Garota Nacional',200,1,1,'Samuel Rosa',317492,10511239,0.98999999999999999111);
INSERT INTO tracks VALUES(2465,'Tão Seu',200,1,1,'Samuel Rosa',243748,8133126,0.98999999999999999111);
INSERT INTO tracks VALUES(2466,'Sem Terra',200,1,1,'Samuel Rosa',279353,9196411,0.98999999999999999111);
INSERT INTO tracks VALUES(2467,'Os Exilados',200,1,1,'Samuel Rosa',245551,8222095,0.98999999999999999111);
INSERT INTO tracks VALUES(2468,'Um Dia Qualquer',200,1,1,'Samuel Rosa',292414,9805570,0.98999999999999999111);
INSERT INTO tracks VALUES(2469,'Los Pretos',200,1,1,'Samuel Rosa',239229,8025667,0.98999999999999999111);
INSERT INTO tracks VALUES(2470,'Sul Da América',200,1,1,'Samuel Rosa',254928,8484871,0.98999999999999999111);
INSERT INTO tracks VALUES(2471,'Poconé',200,1,1,'Samuel Rosa',318406,10771610,0.98999999999999999111);
INSERT INTO tracks VALUES(2472,'Lucky 13',201,1,4,'Billy Corgan',189387,6200617,0.98999999999999999111);
INSERT INTO tracks VALUES(2473,'Aeroplane Flies High',201,1,4,'Billy Corgan',473391,15408329,0.98999999999999999111);
INSERT INTO tracks VALUES(2474,'Because You Are',201,1,4,'Billy Corgan',226403,7405137,0.98999999999999999111);
INSERT INTO tracks VALUES(2475,'Slow Dawn',201,1,4,'Billy Corgan',192339,6269057,0.98999999999999999111);
INSERT INTO tracks VALUES(2476,'Believe',201,1,4,'James Iha',192940,6320652,0.98999999999999999111);
INSERT INTO tracks VALUES(2477,'My Mistake',201,1,4,'Billy Corgan',240901,7843477,0.98999999999999999111);
INSERT INTO tracks VALUES(2478,'Marquis In Spades',201,1,4,'Billy Corgan',192731,6304789,0.98999999999999999111);
INSERT INTO tracks VALUES(2479,'Here''s To The Atom Bomb',201,1,4,'Billy Corgan',266893,8763140,0.98999999999999999111);
INSERT INTO tracks VALUES(2480,'Sparrow',201,1,4,'Billy Corgan',176822,5696989,0.98999999999999999111);
INSERT INTO tracks VALUES(2481,'Waiting',201,1,4,'Billy Corgan',228336,7627641,0.98999999999999999111);
INSERT INTO tracks VALUES(2482,'Saturnine',201,1,4,'Billy Corgan',229877,7523502,0.98999999999999999111);
INSERT INTO tracks VALUES(2483,'Rock On',201,1,4,'David Cook',366471,12133825,0.98999999999999999111);
INSERT INTO tracks VALUES(2484,'Set The Ray To Jerry',201,1,4,'Billy Corgan',249364,8215184,0.98999999999999999111);
INSERT INTO tracks VALUES(2485,'Winterlong',201,1,4,'Billy Corgan',299389,9670616,0.98999999999999999111);
INSERT INTO tracks VALUES(2486,'Soot & Stars',201,1,4,'Billy Corgan',399986,12866557,0.98999999999999999111);
INSERT INTO tracks VALUES(2487,'Blissed & Gone',201,1,4,'Billy Corgan',286302,9305998,0.98999999999999999111);
INSERT INTO tracks VALUES(2488,'Siva',202,1,4,'Billy Corgan',261172,8576622,0.98999999999999999111);
INSERT INTO tracks VALUES(2489,'Rhinocerous',202,1,4,'Billy Corgan',353462,11526684,0.98999999999999999111);
INSERT INTO tracks VALUES(2490,'Drown',202,1,4,'Billy Corgan',270497,8883496,0.98999999999999999111);
INSERT INTO tracks VALUES(2491,'Cherub Rock',202,1,4,'Billy Corgan',299389,9786739,0.98999999999999999111);
INSERT INTO tracks VALUES(2492,'Today',202,1,4,'Billy Corgan',202213,6596933,0.98999999999999999111);
INSERT INTO tracks VALUES(2493,'Disarm',202,1,4,'Billy Corgan',198556,6508249,0.98999999999999999111);
INSERT INTO tracks VALUES(2494,'Landslide',202,1,4,'Stevie Nicks',190275,6187754,0.98999999999999999111);
INSERT INTO tracks VALUES(2495,'Bullet With Butterfly Wings',202,1,4,'Billy Corgan',257306,8431747,0.98999999999999999111);
INSERT INTO tracks VALUES(2496,'1979',202,1,4,'Billy Corgan',263653,8728470,0.98999999999999999111);
INSERT INTO tracks VALUES(2497,'Zero',202,1,4,'Billy Corgan',161123,5267176,0.98999999999999999111);
INSERT INTO tracks VALUES(2498,'Tonight, Tonight',202,1,4,'Billy Corgan',255686,8351543,0.98999999999999999111);
INSERT INTO tracks VALUES(2499,'Eye',202,1,4,'Billy Corgan',294530,9784201,0.98999999999999999111);
INSERT INTO tracks VALUES(2500,'Ava Adore',202,1,4,'Billy Corgan',261433,8590412,0.98999999999999999111);
INSERT INTO tracks VALUES(2501,'Perfect',202,1,4,'Billy Corgan',203023,6734636,0.98999999999999999111);
INSERT INTO tracks VALUES(2502,'The Everlasting Gaze',202,1,4,'Billy Corgan',242155,7844404,0.98999999999999999111);
INSERT INTO tracks VALUES(2503,'Stand Inside Your Love',202,1,4,'Billy Corgan',253753,8270113,0.98999999999999999111);
INSERT INTO tracks VALUES(2504,'Real Love',202,1,4,'Billy Corgan',250697,8025896,0.98999999999999999111);
INSERT INTO tracks VALUES(2505,'[Untitled]',202,1,4,'Billy Corgan',231784,7689713,0.98999999999999999111);
INSERT INTO tracks VALUES(2506,'Nothing To Say',203,1,1,'Chris Cornell/Kim Thayil',238027,7744833,0.98999999999999999111);
INSERT INTO tracks VALUES(2507,'Flower',203,1,1,'Chris Cornell/Kim Thayil',208822,6830732,0.98999999999999999111);
INSERT INTO tracks VALUES(2508,'Loud Love',203,1,1,'Chris Cornell',297456,9660953,0.98999999999999999111);
INSERT INTO tracks VALUES(2509,'Hands All Over',203,1,1,'Chris Cornell/Kim Thayil',362475,11893108,0.98999999999999999111);
INSERT INTO tracks VALUES(2510,'Get On The Snake',203,1,1,'Chris Cornell/Kim Thayil',225123,7313744,0.98999999999999999111);
INSERT INTO tracks VALUES(2511,'Jesus Christ Pose',203,1,1,'Ben Shepherd/Chris Cornell/Kim Thayil/Matt Cameron',352966,11739886,0.98999999999999999111);
INSERT INTO tracks VALUES(2512,'Outshined',203,1,1,'Chris Cornell',312476,10274629,0.98999999999999999111);
INSERT INTO tracks VALUES(2513,'Rusty Cage',203,1,1,'Chris Cornell',267728,8779485,0.98999999999999999111);
INSERT INTO tracks VALUES(2514,'Spoonman',203,1,1,'Chris Cornell',248476,8289906,0.98999999999999999111);
INSERT INTO tracks VALUES(2515,'The Day I Tried To Live',203,1,1,'Chris Cornell',321175,10507137,0.98999999999999999111);
INSERT INTO tracks VALUES(2516,'Black Hole Sun',203,1,1,'Soundgarden',320365,10425229,0.98999999999999999111);
INSERT INTO tracks VALUES(2517,'Fell On Black Days',203,1,1,'Chris Cornell',282331,9256082,0.98999999999999999111);
INSERT INTO tracks VALUES(2518,'Pretty Noose',203,1,1,'Chris Cornell',253570,8317931,0.98999999999999999111);
INSERT INTO tracks VALUES(2519,'Burden In My Hand',203,1,1,'Chris Cornell',292153,9659911,0.98999999999999999111);
INSERT INTO tracks VALUES(2520,'Blow Up The Outside World',203,1,1,'Chris Cornell',347898,11379527,0.98999999999999999111);
INSERT INTO tracks VALUES(2521,'Ty Cobb',203,1,1,'Ben Shepherd/Chris Cornell',188786,6233136,0.98999999999999999111);
INSERT INTO tracks VALUES(2522,'Bleed Together',203,1,1,'Chris Cornell',232202,7597074,0.98999999999999999111);
INSERT INTO tracks VALUES(2523,'Morning Dance',204,1,2,'Jay Beckenstein',238759,8101979,0.98999999999999999111);
INSERT INTO tracks VALUES(2524,'Jubilee',204,1,2,'Jeremy Wall',275147,9151846,0.98999999999999999111);
INSERT INTO tracks VALUES(2525,'Rasul',204,1,2,'Jeremy Wall',238315,7854737,0.98999999999999999111);
INSERT INTO tracks VALUES(2526,'Song For Lorraine',204,1,2,'Jay Beckenstein',240091,8101723,0.98999999999999999111);
INSERT INTO tracks VALUES(2527,'Starburst',204,1,2,'Jeremy Wall',291500,9768399,0.98999999999999999111);
INSERT INTO tracks VALUES(2528,'Heliopolis',204,1,2,'Jay Beckenstein',338729,11365655,0.98999999999999999111);
INSERT INTO tracks VALUES(2529,'It Doesn''t Matter',204,1,2,'Chet Catallo',270027,9034177,0.98999999999999999111);
INSERT INTO tracks VALUES(2530,'Little Linda',204,1,2,'Jeremy Wall',264019,8958743,0.98999999999999999111);
INSERT INTO tracks VALUES(2531,'End Of Romanticism',204,1,2,'Rick Strauss',320078,10553155,0.98999999999999999111);
INSERT INTO tracks VALUES(2532,'The House Is Rockin''',205,1,6,'Doyle Bramhall/Stevie Ray Vaughan',144352,4706253,0.98999999999999999111);
INSERT INTO tracks VALUES(2533,'Crossfire',205,1,6,'B. Carter/C. Layton/R. Ellsworth/R. Wynans/T. Shannon',251219,8238033,0.98999999999999999111);
INSERT INTO tracks VALUES(2534,'Tightrope',205,1,6,'Doyle Bramhall/Stevie Ray Vaughan',281155,9254906,0.98999999999999999111);
INSERT INTO tracks VALUES(2535,'Let Me Love You Baby',205,1,6,'Willie Dixon',164127,5378455,0.98999999999999999111);
INSERT INTO tracks VALUES(2536,'Leave My Girl Alone',205,1,6,'B. Guy',256365,8438021,0.98999999999999999111);
INSERT INTO tracks VALUES(2537,'Travis Walk',205,1,6,'Stevie Ray Vaughan',140826,4650979,0.98999999999999999111);
INSERT INTO tracks VALUES(2538,'Wall Of Denial',205,1,6,'Doyle Bramhall/Stevie Ray Vaughan',336927,11085915,0.98999999999999999111);
INSERT INTO tracks VALUES(2539,'Scratch-N-Sniff',205,1,6,'Doyle Bramhall/Stevie Ray Vaughan',163422,5353627,0.98999999999999999111);
INSERT INTO tracks VALUES(2540,'Love Me Darlin''',205,1,6,'C. Burnett',201586,6650869,0.98999999999999999111);
INSERT INTO tracks VALUES(2541,'Riviera Paradise',205,1,6,'Stevie Ray Vaughan',528692,17232776,0.98999999999999999111);
INSERT INTO tracks VALUES(2542,'Dead And Bloated',206,1,1,'R. DeLeo/Weiland',310386,10170433,0.98999999999999999111);
INSERT INTO tracks VALUES(2543,'Sex Type Thing',206,1,1,'D. DeLeo/Kretz/Weiland',218723,7102064,0.98999999999999999111);
INSERT INTO tracks VALUES(2544,'Wicked Garden',206,1,1,'D. DeLeo/R. DeLeo/Weiland',245368,7989505,0.98999999999999999111);
INSERT INTO tracks VALUES(2545,'No Memory',206,1,1,'Dean Deleo',80613,2660859,0.98999999999999999111);
INSERT INTO tracks VALUES(2546,'Sin',206,1,1,'R. DeLeo/Weiland',364800,12018823,0.98999999999999999111);
INSERT INTO tracks VALUES(2547,'Naked Sunday',206,1,1,'D. DeLeo/Kretz/R. DeLeo/Weiland',229720,7444201,0.98999999999999999111);
INSERT INTO tracks VALUES(2548,'Creep',206,1,1,'R. DeLeo/Weiland',333191,10894988,0.98999999999999999111);
INSERT INTO tracks VALUES(2549,'Piece Of Pie',206,1,1,'R. DeLeo/Weiland',324623,10605231,0.98999999999999999111);
INSERT INTO tracks VALUES(2550,'Plush',206,1,1,'R. DeLeo/Weiland',314017,10229848,0.98999999999999999111);
INSERT INTO tracks VALUES(2551,'Wet My Bed',206,1,1,'R. DeLeo/Weiland',96914,3198627,0.98999999999999999111);
INSERT INTO tracks VALUES(2552,'Crackerman',206,1,1,'Kretz/R. DeLeo/Weiland',194403,6317361,0.98999999999999999111);
INSERT INTO tracks VALUES(2553,'Where The River Goes',206,1,1,'D. DeLeo/Kretz/Weiland',505991,16468904,0.98999999999999999111);
INSERT INTO tracks VALUES(2554,'Soldier Side - Intro',207,1,3,'Dolmayan, John/Malakian, Daron/Odadjian, Shavo',63764,2056079,0.98999999999999999111);
INSERT INTO tracks VALUES(2555,'B.Y.O.B.',207,1,3,'Tankian, Serj',255555,8407935,0.98999999999999999111);
INSERT INTO tracks VALUES(2556,'Revenga',207,1,3,'Tankian, Serj',228127,7503805,0.98999999999999999111);
INSERT INTO tracks VALUES(2557,'Cigaro',207,1,3,'Tankian, Serj',131787,4321705,0.98999999999999999111);
INSERT INTO tracks VALUES(2558,'Radio/Video',207,1,3,'Dolmayan, John/Malakian, Daron/Odadjian, Shavo',249312,8224917,0.98999999999999999111);
INSERT INTO tracks VALUES(2559,'This Cocaine Makes Me Feel Like I''m On This Song',207,1,3,'Tankian, Serj',128339,4185193,0.98999999999999999111);
INSERT INTO tracks VALUES(2560,'Violent Pornography',207,1,3,'Dolmayan, John/Malakian, Daron/Odadjian, Shavo',211435,6985960,0.98999999999999999111);
INSERT INTO tracks VALUES(2561,'Question!',207,1,3,'Tankian, Serj',200698,6616398,0.98999999999999999111);
INSERT INTO tracks VALUES(2562,'Sad Statue',207,1,3,'Tankian, Serj',205897,6733449,0.98999999999999999111);
INSERT INTO tracks VALUES(2563,'Old School Hollywood',207,1,3,'Dolmayan, John/Malakian, Daron/Odadjian, Shavo',176953,5830258,0.98999999999999999111);
INSERT INTO tracks VALUES(2564,'Lost in Hollywood',207,1,3,'Tankian, Serj',320783,10535158,0.98999999999999999111);
INSERT INTO tracks VALUES(2565,'The Sun Road',208,1,1,'Terry Bozzio, Steve Stevens, Tony Levin',880640,29008407,0.98999999999999999111);
INSERT INTO tracks VALUES(2566,'Dark Corners',208,1,1,'Terry Bozzio, Steve Stevens, Tony Levin',513541,16839223,0.98999999999999999111);
INSERT INTO tracks VALUES(2567,'Duende',208,1,1,'Terry Bozzio, Steve Stevens, Tony Levin',447582,14956771,0.98999999999999999111);
INSERT INTO tracks VALUES(2568,'Black Light Syndrome',208,1,1,'Terry Bozzio, Steve Stevens, Tony Levin',526471,17300835,0.98999999999999999111);
INSERT INTO tracks VALUES(2569,'Falling in Circles',208,1,1,'Terry Bozzio, Steve Stevens, Tony Levin',549093,18263248,0.98999999999999999111);
INSERT INTO tracks VALUES(2570,'Book of Hours',208,1,1,'Terry Bozzio, Steve Stevens, Tony Levin',583366,19464726,0.98999999999999999111);
INSERT INTO tracks VALUES(2571,'Chaos-Control',208,1,1,'Terry Bozzio, Steve Stevens, Tony Levin',529841,17455568,0.98999999999999999111);
INSERT INTO tracks VALUES(2572,'Midnight From The Inside Out',209,1,6,'Chris Robinson/Rich Robinson',286981,9442157,0.98999999999999999111);
INSERT INTO tracks VALUES(2573,'Sting Me',209,1,6,'Chris Robinson/Rich Robinson',268094,8813561,0.98999999999999999111);
INSERT INTO tracks VALUES(2574,'Thick & Thin',209,1,6,'Chris Robinson/Rich Robinson',222720,7284377,0.98999999999999999111);
INSERT INTO tracks VALUES(2575,'Greasy Grass River',209,1,6,'Chris Robinson/Rich Robinson',218749,7157045,0.98999999999999999111);
INSERT INTO tracks VALUES(2576,'Sometimes Salvation',209,1,6,'Chris Robinson/Rich Robinson',389146,12749424,0.98999999999999999111);
INSERT INTO tracks VALUES(2577,'Cursed Diamonds',209,1,6,'Chris Robinson/Rich Robinson',368300,12047978,0.98999999999999999111);
INSERT INTO tracks VALUES(2578,'Miracle To Me',209,1,6,'Chris Robinson/Rich Robinson',372636,12222116,0.98999999999999999111);
INSERT INTO tracks VALUES(2579,'Wiser Time',209,1,6,'Chris Robinson/Rich Robinson',459990,15161907,0.98999999999999999111);
INSERT INTO tracks VALUES(2580,'Girl From A Pawnshop',209,1,6,'Chris Robinson/Rich Robinson',404688,13250848,0.98999999999999999111);
INSERT INTO tracks VALUES(2581,'Cosmic Fiend',209,1,6,'Chris Robinson/Rich Robinson',308401,10115556,0.98999999999999999111);
INSERT INTO tracks VALUES(2582,'Black Moon Creeping',210,1,6,'Chris Robinson/Rich Robinson',359314,11740886,0.98999999999999999111);
INSERT INTO tracks VALUES(2583,'High Head Blues',210,1,6,'Chris Robinson/Rich Robinson',371879,12227998,0.98999999999999999111);
INSERT INTO tracks VALUES(2584,'Title Song',210,1,6,'Chris Robinson/Rich Robinson',505521,16501316,0.98999999999999999111);
INSERT INTO tracks VALUES(2585,'She Talks To Angels',210,1,6,'Chris Robinson/Rich Robinson',361978,11837342,0.98999999999999999111);
INSERT INTO tracks VALUES(2586,'Twice As Hard',210,1,6,'Chris Robinson/Rich Robinson',275565,9008067,0.98999999999999999111);
INSERT INTO tracks VALUES(2587,'Lickin''',210,1,6,'Chris Robinson/Rich Robinson',314409,10331216,0.98999999999999999111);
INSERT INTO tracks VALUES(2588,'Soul Singing',210,1,6,'Chris Robinson/Rich Robinson',233639,7672489,0.98999999999999999111);
INSERT INTO tracks VALUES(2589,'Hard To Handle',210,1,6,'A.Isbell/A.Jones/O.Redding',206994,6786304,0.98999999999999999111);
INSERT INTO tracks VALUES(2590,'Remedy',210,1,6,'Chris Robinson/Rich Robinson',337084,11049098,0.98999999999999999111);
INSERT INTO tracks VALUES(2591,'White Riot',211,1,4,'Joe Strummer/Mick Jones',118726,3922819,0.98999999999999999111);
INSERT INTO tracks VALUES(2592,'Remote Control',211,1,4,'Joe Strummer/Mick Jones',180297,5949647,0.98999999999999999111);
INSERT INTO tracks VALUES(2593,'Complete Control',211,1,4,'Joe Strummer/Mick Jones',192653,6272081,0.98999999999999999111);
INSERT INTO tracks VALUES(2594,'Clash City Rockers',211,1,4,'Joe Strummer/Mick Jones',227500,7555054,0.98999999999999999111);
INSERT INTO tracks VALUES(2595,'(White Man) In Hammersmith Palais',211,1,4,'Joe Strummer/Mick Jones',240640,7883532,0.98999999999999999111);
INSERT INTO tracks VALUES(2596,'Tommy Gun',211,1,4,'Joe Strummer/Mick Jones',195526,6399872,0.98999999999999999111);
INSERT INTO tracks VALUES(2597,'English Civil War',211,1,4,'Mick Jones/Traditional arr. Joe Strummer',156708,5111226,0.98999999999999999111);
INSERT INTO tracks VALUES(2598,'I Fought The Law',211,1,4,'Sonny Curtis',159764,5245258,0.98999999999999999111);
INSERT INTO tracks VALUES(2599,'London Calling',211,1,4,'Joe Strummer/Mick Jones',199706,6569007,0.98999999999999999111);
INSERT INTO tracks VALUES(2600,'Train In Vain',211,1,4,'Joe Strummer/Mick Jones',189675,6329877,0.98999999999999999111);
INSERT INTO tracks VALUES(2601,'Bankrobber',211,1,4,'Joe Strummer/Mick Jones',272431,9067323,0.98999999999999999111);
INSERT INTO tracks VALUES(2602,'The Call Up',211,1,4,'The Clash',324336,10746937,0.98999999999999999111);
INSERT INTO tracks VALUES(2603,'Hitsville UK',211,1,4,'The Clash',261433,8606887,0.98999999999999999111);
INSERT INTO tracks VALUES(2604,'The Magnificent Seven',211,1,4,'The Clash',268486,8889821,0.98999999999999999111);
INSERT INTO tracks VALUES(2605,'This Is Radio Clash',211,1,4,'The Clash',249756,8366573,0.98999999999999999111);
INSERT INTO tracks VALUES(2606,'Know Your Rights',211,1,4,'The Clash',217678,7195726,0.98999999999999999111);
INSERT INTO tracks VALUES(2607,'Rock The Casbah',211,1,4,'The Clash',222145,7361500,0.98999999999999999111);
INSERT INTO tracks VALUES(2608,'Should I Stay Or Should I Go',211,1,4,'The Clash',187219,6188688,0.98999999999999999111);
INSERT INTO tracks VALUES(2609,'War (The Process)',212,1,1,'Billy Duffy/Ian Astbury',252630,8254842,0.98999999999999999111);
INSERT INTO tracks VALUES(2610,'The Saint',212,1,1,'Billy Duffy/Ian Astbury',216215,7061584,0.98999999999999999111);
INSERT INTO tracks VALUES(2611,'Rise',212,1,1,'Billy Duffy/Ian Astbury',219088,7106195,0.98999999999999999111);
INSERT INTO tracks VALUES(2612,'Take The Power',212,1,1,'Billy Duffy/Ian Astbury',235755,7650012,0.98999999999999999111);
INSERT INTO tracks VALUES(2613,'Breathe',212,1,1,'Billy Duffy/Ian Astbury/Marti Frederiksen/Mick Jones',299781,9742361,0.98999999999999999111);
INSERT INTO tracks VALUES(2614,'Nico',212,1,1,'Billy Duffy/Ian Astbury',289488,9412323,0.98999999999999999111);
INSERT INTO tracks VALUES(2615,'American Gothic',212,1,1,'Billy Duffy/Ian Astbury',236878,7739840,0.98999999999999999111);
INSERT INTO tracks VALUES(2616,'Ashes And Ghosts',212,1,1,'Billy Duffy/Bob Rock/Ian Astbury',300591,9787692,0.98999999999999999111);
INSERT INTO tracks VALUES(2617,'Shape The Sky',212,1,1,'Billy Duffy/Ian Astbury',209789,6885647,0.98999999999999999111);
INSERT INTO tracks VALUES(2618,'Speed Of Light',212,1,1,'Billy Duffy/Bob Rock/Ian Astbury',262817,8563352,0.98999999999999999111);
INSERT INTO tracks VALUES(2619,'True Believers',212,1,1,'Billy Duffy/Ian Astbury',308009,9981359,0.98999999999999999111);
INSERT INTO tracks VALUES(2620,'My Bridges Burn',212,1,1,'Billy Duffy/Ian Astbury',231862,7571370,0.98999999999999999111);
INSERT INTO tracks VALUES(2621,'She Sells Sanctuary',213,1,1,NULL,253727,8368634,0.98999999999999999111);
INSERT INTO tracks VALUES(2622,'Fire Woman',213,1,1,NULL,312790,10196995,0.98999999999999999111);
INSERT INTO tracks VALUES(2623,'Lil'' Evil',213,1,1,NULL,165825,5419655,0.98999999999999999111);
INSERT INTO tracks VALUES(2624,'Spirit Walker',213,1,1,NULL,230060,7555897,0.98999999999999999111);
INSERT INTO tracks VALUES(2625,'The Witch',213,1,1,NULL,258768,8725403,0.98999999999999999111);
INSERT INTO tracks VALUES(2626,'Revolution',213,1,1,NULL,256026,8371254,0.98999999999999999111);
INSERT INTO tracks VALUES(2627,'Wild Hearted Son',213,1,1,NULL,266893,8670550,0.98999999999999999111);
INSERT INTO tracks VALUES(2628,'Love Removal Machine',213,1,1,NULL,257619,8412167,0.98999999999999999111);
INSERT INTO tracks VALUES(2629,'Rain',213,1,1,NULL,236669,7788461,0.98999999999999999111);
INSERT INTO tracks VALUES(2630,'Edie (Ciao Baby)',213,1,1,NULL,241632,7846177,0.98999999999999999111);
INSERT INTO tracks VALUES(2631,'Heart Of Soul',213,1,1,NULL,274207,8967257,0.98999999999999999111);
INSERT INTO tracks VALUES(2632,'Love',213,1,1,NULL,326739,10729824,0.98999999999999999111);
INSERT INTO tracks VALUES(2633,'Wild Flower',213,1,1,NULL,215536,7084321,0.98999999999999999111);
INSERT INTO tracks VALUES(2634,'Go West',213,1,1,NULL,238158,7777749,0.98999999999999999111);
INSERT INTO tracks VALUES(2635,'Resurrection Joe',213,1,1,NULL,255451,8532840,0.98999999999999999111);
INSERT INTO tracks VALUES(2636,'Sun King',213,1,1,NULL,368431,12010865,0.98999999999999999111);
INSERT INTO tracks VALUES(2637,'Sweet Soul Sister',213,1,1,NULL,212009,6889883,0.98999999999999999111);
INSERT INTO tracks VALUES(2638,'Earth Mofo',213,1,1,NULL,282200,9204581,0.98999999999999999111);
INSERT INTO tracks VALUES(2639,'Break on Through',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',149342,4943144,0.98999999999999999111);
INSERT INTO tracks VALUES(2640,'Soul Kitchen',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',215066,7040865,0.98999999999999999111);
INSERT INTO tracks VALUES(2641,'The Crystal Ship',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',154853,5052658,0.98999999999999999111);
INSERT INTO tracks VALUES(2642,'Twentienth Century Fox',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',153913,5069211,0.98999999999999999111);
INSERT INTO tracks VALUES(2643,'Alabama Song',214,1,1,'Weill-Brecht',200097,6563411,0.98999999999999999111);
INSERT INTO tracks VALUES(2644,'Light My Fire',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',428329,13963351,0.98999999999999999111);
INSERT INTO tracks VALUES(2645,'Back Door Man',214,1,1,'Willie Dixon, C. Burnett',214360,7035636,0.98999999999999999111);
INSERT INTO tracks VALUES(2646,'I Looked At You',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',142080,4663988,0.98999999999999999111);
INSERT INTO tracks VALUES(2647,'End Of The Night',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',172695,5589732,0.98999999999999999111);
INSERT INTO tracks VALUES(2648,'Take It As It Comes',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',137168,4512656,0.98999999999999999111);
INSERT INTO tracks VALUES(2649,'The End',214,1,1,'Robby Krieger, Ray Manzarek, John Densmore, Jim Morrison',701831,22927336,0.98999999999999999111);
INSERT INTO tracks VALUES(2650,'Roxanne',215,1,1,'G M Sumner',192992,6330159,0.98999999999999999111);
INSERT INTO tracks VALUES(2651,'Can''t Stand Losing You',215,1,1,'G M Sumner',181159,5971983,0.98999999999999999111);
INSERT INTO tracks VALUES(2652,'Message in a Bottle',215,1,1,'G M Sumner',291474,9647829,0.98999999999999999111);
INSERT INTO tracks VALUES(2653,'Walking on the Moon',215,1,1,'G M Sumner',302080,10019861,0.98999999999999999111);
INSERT INTO tracks VALUES(2654,'Don''t Stand so Close to Me',215,1,1,'G M Sumner',241031,7956658,0.98999999999999999111);
INSERT INTO tracks VALUES(2655,'De Do Do Do, De Da Da Da',215,1,1,'G M Sumner',247196,8227075,0.98999999999999999111);
INSERT INTO tracks VALUES(2656,'Every Little Thing She Does is Magic',215,1,1,'G M Sumner',261120,8646853,0.98999999999999999111);
INSERT INTO tracks VALUES(2657,'Invisible Sun',215,1,1,'G M Sumner',225593,7304320,0.98999999999999999111);
INSERT INTO tracks VALUES(2658,'Spirit''s in the Material World',215,1,1,'G M Sumner',181133,5986622,0.98999999999999999111);
INSERT INTO tracks VALUES(2659,'Every Breath You Take',215,1,1,'G M Sumner',254615,8364520,0.98999999999999999111);
INSERT INTO tracks VALUES(2660,'King Of Pain',215,1,1,'G M Sumner',300512,9880303,0.98999999999999999111);
INSERT INTO tracks VALUES(2661,'Wrapped Around Your Finger',215,1,1,'G M Sumner',315454,10361490,0.98999999999999999111);
INSERT INTO tracks VALUES(2662,'Don''t Stand So Close to Me ''86',215,1,1,'G M Sumner',293590,9636683,0.98999999999999999111);
INSERT INTO tracks VALUES(2663,'Message in a Bottle (new classic rock mix)',215,1,1,'G M Sumner',290951,9640349,0.98999999999999999111);
INSERT INTO tracks VALUES(2664,'Time Is On My Side',216,1,1,'Jerry Ragavoy',179983,5855836,0.98999999999999999111);
INSERT INTO tracks VALUES(2665,'Heart Of Stone',216,1,1,'Jagger/Richards',164493,5329538,0.98999999999999999111);
INSERT INTO tracks VALUES(2666,'Play With Fire',216,1,1,'Nanker Phelge',132022,4265297,0.98999999999999999111);
INSERT INTO tracks VALUES(2667,'Satisfaction',216,1,1,'Jagger/Richards',226612,7398766,0.98999999999999999111);
INSERT INTO tracks VALUES(2668,'As Tears Go By',216,1,1,'Jagger/Richards/Oldham',164284,5357350,0.98999999999999999111);
INSERT INTO tracks VALUES(2669,'Get Off Of My Cloud',216,1,1,'Jagger/Richards',176013,5719514,0.98999999999999999111);
INSERT INTO tracks VALUES(2670,'Mother''s Little Helper',216,1,1,'Jagger/Richards',167549,5422434,0.98999999999999999111);
INSERT INTO tracks VALUES(2671,'19th Nervous Breakdown',216,1,1,'Jagger/Richards',237923,7742984,0.98999999999999999111);
INSERT INTO tracks VALUES(2672,'Paint It Black',216,1,1,'Jagger/Richards',226063,7442888,0.98999999999999999111);
INSERT INTO tracks VALUES(2673,'Under My Thumb',216,1,1,'Jagger/Richards',221387,7371799,0.98999999999999999111);
INSERT INTO tracks VALUES(2674,'Ruby Tuesday',216,1,1,'Jagger/Richards',197459,6433467,0.98999999999999999111);
INSERT INTO tracks VALUES(2675,'Let''s Spend The Night Together',216,1,1,'Jagger/Richards',217495,7137048,0.98999999999999999111);
INSERT INTO tracks VALUES(2676,'Intro',217,1,1,'Jagger/Richards',49737,1618591,0.98999999999999999111);
INSERT INTO tracks VALUES(2677,'You Got Me Rocking',217,1,1,'Jagger/Richards',205766,6734385,0.98999999999999999111);
INSERT INTO tracks VALUES(2678,'Gimmie Shelters',217,1,1,'Jagger/Richards',382119,12528764,0.98999999999999999111);
INSERT INTO tracks VALUES(2679,'Flip The Switch',217,1,1,'Jagger/Richards',252421,8336591,0.98999999999999999111);
INSERT INTO tracks VALUES(2680,'Memory Motel',217,1,1,'Jagger/Richards',365844,11982431,0.98999999999999999111);
INSERT INTO tracks VALUES(2681,'Corinna',217,1,1,'Jesse Ed Davis III/Taj Mahal',257488,8449471,0.98999999999999999111);
INSERT INTO tracks VALUES(2682,'Saint Of Me',217,1,1,'Jagger/Richards',325694,10725160,0.98999999999999999111);
INSERT INTO tracks VALUES(2683,'Wainting On A Friend',217,1,1,'Jagger/Richards',302497,9978046,0.98999999999999999111);
INSERT INTO tracks VALUES(2684,'Sister Morphine',217,1,1,'Faithfull/Jagger/Richards',376215,12345289,0.98999999999999999111);
INSERT INTO tracks VALUES(2685,'Live With Me',217,1,1,'Jagger/Richards',234893,7709006,0.98999999999999999111);
INSERT INTO tracks VALUES(2686,'Respectable',217,1,1,'Jagger/Richards',215693,7099669,0.98999999999999999111);
INSERT INTO tracks VALUES(2687,'Thief In The Night',217,1,1,'De Beauport/Jagger/Richards',337266,10952756,0.98999999999999999111);
INSERT INTO tracks VALUES(2688,'The Last Time',217,1,1,'Jagger/Richards',287294,9498758,0.98999999999999999111);
INSERT INTO tracks VALUES(2689,'Out Of Control',217,1,1,'Jagger/Richards',479242,15749289,0.98999999999999999111);
INSERT INTO tracks VALUES(2690,'Love Is Strong',218,1,1,'Jagger/Richards',230896,7639774,0.98999999999999999111);
INSERT INTO tracks VALUES(2691,'You Got Me Rocking',218,1,1,'Jagger/Richards',215928,7162159,0.98999999999999999111);
INSERT INTO tracks VALUES(2692,'Sparks Will Fly',218,1,1,'Jagger/Richards',196466,6492847,0.98999999999999999111);
INSERT INTO tracks VALUES(2693,'The Worst',218,1,1,'Jagger/Richards',144613,4750094,0.98999999999999999111);
INSERT INTO tracks VALUES(2694,'New Faces',218,1,1,'Jagger/Richards',172146,5689122,0.98999999999999999111);
INSERT INTO tracks VALUES(2695,'Moon Is Up',218,1,1,'Jagger/Richards',222119,7366316,0.98999999999999999111);
INSERT INTO tracks VALUES(2696,'Out Of Tears',218,1,1,'Jagger/Richards',327418,10677236,0.98999999999999999111);
INSERT INTO tracks VALUES(2697,'I Go Wild',218,1,1,'Jagger/Richards',264019,8630833,0.98999999999999999111);
INSERT INTO tracks VALUES(2698,'Brand New Car',218,1,1,'Jagger/Richards',256052,8459344,0.98999999999999999111);
INSERT INTO tracks VALUES(2699,'Sweethearts Together',218,1,1,'Jagger/Richards',285492,9550459,0.98999999999999999111);
INSERT INTO tracks VALUES(2700,'Suck On The Jugular',218,1,1,'Jagger/Richards',268225,8920566,0.98999999999999999111);
INSERT INTO tracks VALUES(2701,'Blinded By Rainbows',218,1,1,'Jagger/Richards',273946,8971343,0.98999999999999999111);
INSERT INTO tracks VALUES(2702,'Baby Break It Down',218,1,1,'Jagger/Richards',249417,8197309,0.98999999999999999111);
INSERT INTO tracks VALUES(2703,'Thru And Thru',218,1,1,'Jagger/Richards',375092,12175406,0.98999999999999999111);
INSERT INTO tracks VALUES(2704,'Mean Disposition',218,1,1,'Jagger/Richards',249155,8273602,0.98999999999999999111);
INSERT INTO tracks VALUES(2705,'Walking Wounded',219,1,4,'The Tea Party',277968,9184345,0.98999999999999999111);
INSERT INTO tracks VALUES(2706,'Temptation',219,1,4,'The Tea Party',205087,6711943,0.98999999999999999111);
INSERT INTO tracks VALUES(2707,'The Messenger',219,1,4,'Daniel Lanois',212062,6975437,0.98999999999999999111);
INSERT INTO tracks VALUES(2708,'Psychopomp',219,1,4,'The Tea Party',315559,10295199,0.98999999999999999111);
INSERT INTO tracks VALUES(2709,'Sister Awake',219,1,4,'The Tea Party',343875,11299407,0.98999999999999999111);
INSERT INTO tracks VALUES(2710,'The Bazaar',219,1,4,'The Tea Party',222458,7245691,0.98999999999999999111);
INSERT INTO tracks VALUES(2711,'Save Me (Remix)',219,1,4,'The Tea Party',396303,13053839,0.98999999999999999111);
INSERT INTO tracks VALUES(2712,'Fire In The Head',219,1,4,'The Tea Party',306337,10005675,0.98999999999999999111);
INSERT INTO tracks VALUES(2713,'Release',219,1,4,'The Tea Party',244114,8014606,0.98999999999999999111);
INSERT INTO tracks VALUES(2714,'Heaven Coming Down',219,1,4,'The Tea Party',241867,7846459,0.98999999999999999111);
INSERT INTO tracks VALUES(2715,'The River (Remix)',219,1,4,'The Tea Party',343170,11193268,0.98999999999999999111);
INSERT INTO tracks VALUES(2716,'Babylon',219,1,4,'The Tea Party',169795,5568808,0.98999999999999999111);
INSERT INTO tracks VALUES(2717,'Waiting On A Sign',219,1,4,'The Tea Party',261903,8558590,0.98999999999999999111);
INSERT INTO tracks VALUES(2718,'Life Line',219,1,4,'The Tea Party',277786,9082773,0.98999999999999999111);
INSERT INTO tracks VALUES(2719,'Paint It Black',219,1,4,'Keith Richards/Mick Jagger',214752,7101572,0.98999999999999999111);
INSERT INTO tracks VALUES(2720,'Temptation',220,1,4,'The Tea Party',205244,6719465,0.98999999999999999111);
INSERT INTO tracks VALUES(2721,'Army Ants',220,1,4,'The Tea Party',215405,7075838,0.98999999999999999111);
INSERT INTO tracks VALUES(2722,'Psychopomp',220,1,4,'The Tea Party',317231,10351778,0.98999999999999999111);
INSERT INTO tracks VALUES(2723,'Gyroscope',220,1,4,'The Tea Party',177711,5810323,0.98999999999999999111);
INSERT INTO tracks VALUES(2724,'Alarum',220,1,4,'The Tea Party',298187,9712545,0.98999999999999999111);
INSERT INTO tracks VALUES(2725,'Release',220,1,4,'The Tea Party',266292,8725824,0.98999999999999999111);
INSERT INTO tracks VALUES(2726,'Transmission',220,1,4,'The Tea Party',317257,10351152,0.98999999999999999111);
INSERT INTO tracks VALUES(2727,'Babylon',220,1,4,'The Tea Party',292466,9601786,0.98999999999999999111);
INSERT INTO tracks VALUES(2728,'Pulse',220,1,4,'The Tea Party',250253,8183872,0.98999999999999999111);
INSERT INTO tracks VALUES(2729,'Emerald',220,1,4,'The Tea Party',289750,9543789,0.98999999999999999111);
INSERT INTO tracks VALUES(2730,'Aftermath',220,1,4,'The Tea Party',343745,11085607,0.98999999999999999111);
INSERT INTO tracks VALUES(2731,'I Can''t Explain',221,1,1,'Pete Townshend',125152,4082896,0.98999999999999999111);
INSERT INTO tracks VALUES(2732,'Anyway, Anyhow, Anywhere',221,1,1,'Pete Townshend, Roger Daltrey',161253,5234173,0.98999999999999999111);
INSERT INTO tracks VALUES(2733,'My Generation',221,1,1,'John Entwistle/Pete Townshend',197825,6446634,0.98999999999999999111);
INSERT INTO tracks VALUES(2734,'Substitute',221,1,1,'Pete Townshend',228022,7409995,0.98999999999999999111);
INSERT INTO tracks VALUES(2735,'I''m A Boy',221,1,1,'Pete Townshend',157126,5120605,0.98999999999999999111);
INSERT INTO tracks VALUES(2736,'Boris The Spider',221,1,1,'John Entwistle',149472,4835202,0.98999999999999999111);
INSERT INTO tracks VALUES(2737,'Happy Jack',221,1,1,'Pete Townshend',132310,4353063,0.98999999999999999111);
INSERT INTO tracks VALUES(2738,'Pictures Of Lily',221,1,1,'Pete Townshend',164414,5329751,0.98999999999999999111);
INSERT INTO tracks VALUES(2739,'I Can See For Miles',221,1,1,'Pete Townshend',262791,8604989,0.98999999999999999111);
INSERT INTO tracks VALUES(2740,'Magic Bus',221,1,1,'Pete Townshend',197224,6452700,0.98999999999999999111);
INSERT INTO tracks VALUES(2741,'Pinball Wizard',221,1,1,'John Entwistle/Pete Townshend',181890,6055580,0.98999999999999999111);
INSERT INTO tracks VALUES(2742,'The Seeker',221,1,1,'Pete Townshend',204643,6736866,0.98999999999999999111);
INSERT INTO tracks VALUES(2743,'Baba O''Riley',221,1,1,'John Entwistle/Pete Townshend',309472,10141660,0.98999999999999999111);
INSERT INTO tracks VALUES(2744,'Won''t Get Fooled Again (Full Length Version)',221,1,1,'John Entwistle/Pete Townshend',513750,16855521,0.98999999999999999111);
INSERT INTO tracks VALUES(2745,'Let''s See Action',221,1,1,'Pete Townshend',243513,8078418,0.98999999999999999111);
INSERT INTO tracks VALUES(2746,'5.15',221,1,1,'Pete Townshend',289619,9458549,0.98999999999999999111);
INSERT INTO tracks VALUES(2747,'Join Together',221,1,1,'Pete Townshend',262556,8602485,0.98999999999999999111);
INSERT INTO tracks VALUES(2748,'Squeeze Box',221,1,1,'Pete Townshend',161280,5256508,0.98999999999999999111);
INSERT INTO tracks VALUES(2749,'Who Are You (Single Edit Version)',221,1,1,'John Entwistle/Pete Townshend',299232,9900469,0.98999999999999999111);
INSERT INTO tracks VALUES(2750,'You Better You Bet',221,1,1,'Pete Townshend',338520,11160877,0.98999999999999999111);
INSERT INTO tracks VALUES(2751,'Primavera',222,1,7,'Genival Cassiano/Silvio Rochael',126615,4152604,0.98999999999999999111);
INSERT INTO tracks VALUES(2752,'Chocolate',222,1,7,'Tim Maia',194690,6411587,0.98999999999999999111);
INSERT INTO tracks VALUES(2753,'Azul Da Cor Do Mar',222,1,7,'Tim Maia',197955,6475007,0.98999999999999999111);
INSERT INTO tracks VALUES(2754,'O Descobridor Dos Sete Mares',222,1,7,'Gilson Mendonça/Michel',262974,8749583,0.98999999999999999111);
INSERT INTO tracks VALUES(2755,'Até Que Enfim Encontrei Você',222,1,7,'Tim Maia',105064,3477751,0.98999999999999999111);
INSERT INTO tracks VALUES(2756,'Coroné Antonio Bento',222,1,7,'Do Vale, João/Luiz Wanderley',131317,4340326,0.98999999999999999111);
INSERT INTO tracks VALUES(2757,'New Love',222,1,7,'Tim Maia',237897,7786824,0.98999999999999999111);
INSERT INTO tracks VALUES(2758,'Não Vou Ficar',222,1,7,'Tim Maia',172068,5642919,0.98999999999999999111);
INSERT INTO tracks VALUES(2759,'Música No Ar',222,1,7,'Tim Maia',158511,5184891,0.98999999999999999111);
INSERT INTO tracks VALUES(2760,'Salve Nossa Senhora',222,1,7,'Carlos Imperial/Edardo Araújo',115461,3827629,0.98999999999999999111);
INSERT INTO tracks VALUES(2761,'Você Fugiu',222,1,7,'Genival Cassiano',238367,7971147,0.98999999999999999111);
INSERT INTO tracks VALUES(2762,'Cristina Nº 2',222,1,7,'Carlos Imperial/Tim Maia',90148,2978589,0.98999999999999999111);
INSERT INTO tracks VALUES(2763,'Compadre',222,1,7,'Tim Maia',171389,5631446,0.98999999999999999111);
INSERT INTO tracks VALUES(2764,'Over Again',222,1,7,'Tim Maia',200489,6612634,0.98999999999999999111);
INSERT INTO tracks VALUES(2765,'Réu Confesso',222,1,7,'Tim Maia',217391,7189874,0.98999999999999999111);
INSERT INTO tracks VALUES(2766,'O Que Me Importa',223,1,7,NULL,153155,4977852,0.98999999999999999111);
INSERT INTO tracks VALUES(2767,'Gostava Tanto De Você',223,1,7,NULL,253805,8380077,0.98999999999999999111);
INSERT INTO tracks VALUES(2768,'Você',223,1,7,NULL,242599,7911702,0.98999999999999999111);
INSERT INTO tracks VALUES(2769,'Não Quero Dinheiro',223,1,7,NULL,152607,5031797,0.98999999999999999111);
INSERT INTO tracks VALUES(2770,'Eu Amo Você',223,1,7,NULL,242782,7914628,0.98999999999999999111);
INSERT INTO tracks VALUES(2771,'A Festa Do Santo Reis',223,1,7,NULL,159791,5204995,0.98999999999999999111);
INSERT INTO tracks VALUES(2772,'I Don''t Know What To Do With Myself',223,1,7,NULL,221387,7251478,0.98999999999999999111);
INSERT INTO tracks VALUES(2773,'Padre Cícero',223,1,7,NULL,139598,4581685,0.98999999999999999111);
INSERT INTO tracks VALUES(2774,'Nosso Adeus',223,1,7,NULL,206471,6793270,0.98999999999999999111);
INSERT INTO tracks VALUES(2775,'Canário Do Reino',223,1,7,NULL,139337,4552858,0.98999999999999999111);
INSERT INTO tracks VALUES(2776,'Preciso Ser Amado',223,1,7,NULL,174001,5618895,0.98999999999999999111);
INSERT INTO tracks VALUES(2777,'Balanço',223,1,7,NULL,209737,6890327,0.98999999999999999111);
INSERT INTO tracks VALUES(2778,'Preciso Aprender A Ser Só',223,1,7,NULL,162220,5213894,0.98999999999999999111);
INSERT INTO tracks VALUES(2779,'Esta É A Canção',223,1,7,NULL,184450,6069933,0.98999999999999999111);
INSERT INTO tracks VALUES(2780,'Formigueiro',223,1,7,NULL,252943,8455132,0.98999999999999999111);
INSERT INTO tracks VALUES(2781,'Comida',224,1,4,'Titãs',322612,10786578,0.98999999999999999111);
INSERT INTO tracks VALUES(2782,'Go Back',224,1,4,'Titãs',230504,7668899,0.98999999999999999111);
INSERT INTO tracks VALUES(2783,'Prá Dizer Adeus',224,1,4,'Titãs',222484,7382048,0.98999999999999999111);
INSERT INTO tracks VALUES(2784,'Família',224,1,4,'Titãs',218331,7267458,0.98999999999999999111);
INSERT INTO tracks VALUES(2785,'Os Cegos Do Castelo',224,1,4,'Titãs',296829,9868187,0.98999999999999999111);
INSERT INTO tracks VALUES(2786,'O Pulso',224,1,4,'Titãs',199131,6566998,0.98999999999999999111);
INSERT INTO tracks VALUES(2787,'Marvin',224,1,4,'Titãs',264359,8741444,0.98999999999999999111);
INSERT INTO tracks VALUES(2788,'Nem 5 Minutos Guardados',224,1,4,'Titãs',245995,8143797,0.98999999999999999111);
INSERT INTO tracks VALUES(2789,'Flores',224,1,4,'Titãs',215510,7148017,0.98999999999999999111);
INSERT INTO tracks VALUES(2790,'Palavras',224,1,4,'Titãs',158458,5285715,0.98999999999999999111);
INSERT INTO tracks VALUES(2791,'Hereditário',224,1,4,'Titãs',151693,5020547,0.98999999999999999111);
INSERT INTO tracks VALUES(2792,'A Melhor Forma',224,1,4,'Titãs',191503,6349938,0.98999999999999999111);
INSERT INTO tracks VALUES(2793,'Cabeça Dinossauro',224,1,4,'Titãs',37120,1220930,0.98999999999999999111);
INSERT INTO tracks VALUES(2794,'32 Dentes',224,1,4,'Titãs',184946,6157904,0.98999999999999999111);
INSERT INTO tracks VALUES(2795,'Bichos Escrotos (Vinheta)',224,1,4,'Titãs',104986,3503755,0.98999999999999999111);
INSERT INTO tracks VALUES(2796,'Não Vou Lutar',224,1,4,'Titãs',189988,6308613,0.98999999999999999111);
INSERT INTO tracks VALUES(2797,'Homem Primata (Vinheta)',224,1,4,'Titãs',34168,1124909,0.98999999999999999111);
INSERT INTO tracks VALUES(2798,'Homem Primata',224,1,4,'Titãs',195500,6486470,0.98999999999999999111);
INSERT INTO tracks VALUES(2799,'Polícia (Vinheta)',224,1,4,'Titãs',56111,1824213,0.98999999999999999111);
INSERT INTO tracks VALUES(2800,'Querem Meu Sangue',224,1,4,'Titãs',212401,7069773,0.98999999999999999111);
INSERT INTO tracks VALUES(2801,'Diversão',224,1,4,'Titãs',285936,9531268,0.98999999999999999111);
INSERT INTO tracks VALUES(2802,'Televisão',224,1,4,'Titãs',293668,9776548,0.98999999999999999111);
INSERT INTO tracks VALUES(2803,'Sonifera Ilha',225,1,4,'Branco Mello/Carlos Barmack/Ciro Pessoa/Marcelo Fromer/Toni Belloto',170684,5678290,0.98999999999999999111);
INSERT INTO tracks VALUES(2804,'Lugar Nenhum',225,1,4,'Arnaldo Antunes/Charles Gavin/Marcelo Fromer/Sérgio Britto/Toni Bellotto',195840,6472780,0.98999999999999999111);
INSERT INTO tracks VALUES(2805,'Sua Impossivel Chance',225,1,4,'Nando Reis',246622,8073248,0.98999999999999999111);
INSERT INTO tracks VALUES(2806,'Desordem',225,1,4,'Charles Gavin/Marcelo Fromer/Sérgio Britto',213289,7067340,0.98999999999999999111);
INSERT INTO tracks VALUES(2807,'Não Vou Me Adaptar',225,1,4,'Arnaldo Antunes',221831,7304656,0.98999999999999999111);
INSERT INTO tracks VALUES(2808,'Domingo',225,1,4,'Sérgio Britto/Toni Bellotto',208613,6883180,0.98999999999999999111);
INSERT INTO tracks VALUES(2809,'Amanhã Não Se Sabe',225,1,4,'Sérgio Britto',189440,6243967,0.98999999999999999111);
INSERT INTO tracks VALUES(2810,'Caras Como Eu',225,1,4,'Toni Bellotto',183092,5999048,0.98999999999999999111);
INSERT INTO tracks VALUES(2811,'Senhora E Senhor',225,1,4,'Arnaldo Anutnes/Marcelo Fromer/Paulo Miklos',203702,6733733,0.98999999999999999111);
INSERT INTO tracks VALUES(2812,'Era Uma Vez',225,1,4,'Arnaldo Anutnes/Branco Mello/Marcelo Fromer/Sergio Brotto/Toni Bellotto',224261,7453156,0.98999999999999999111);
INSERT INTO tracks VALUES(2813,'Miséria',225,1,4,'Arnaldo Antunes/Britto, SergioMiklos, Paulo',262191,8727645,0.98999999999999999111);
INSERT INTO tracks VALUES(2814,'Insensível',225,1,4,'Sérgio Britto',207830,6893664,0.98999999999999999111);
INSERT INTO tracks VALUES(2815,'Eu E Ela',225,1,4,'Nando Reis',276035,9138846,0.98999999999999999111);
INSERT INTO tracks VALUES(2816,'Toda Cor',225,1,4,'Ciro Pressoa/Marcelo Fromer',209084,6939176,0.98999999999999999111);
INSERT INTO tracks VALUES(2817,'É Preciso Saber Viver',225,1,4,'Erasmo Carlos/Roberto Carlos',251115,8271418,0.98999999999999999111);
INSERT INTO tracks VALUES(2818,'Senhor Delegado/Eu Não Aguento',225,1,4,'Antonio Lopes',156656,5277983,0.98999999999999999111);
INSERT INTO tracks VALUES(2819,'Battlestar Galactica: The Story So Far',226,3,18,NULL,2622250,490750393,1.9899999999999999911);
INSERT INTO tracks VALUES(2820,'Occupation / Precipice',227,3,19,NULL,5286953,1054423946,1.9899999999999999911);
INSERT INTO tracks VALUES(2821,'Exodus, Pt. 1',227,3,19,NULL,2621708,475079441,1.9899999999999999911);
INSERT INTO tracks VALUES(2822,'Exodus, Pt. 2',227,3,19,NULL,2618000,466820021,1.9899999999999999911);
INSERT INTO tracks VALUES(2823,'Collaborators',227,3,19,NULL,2626626,483484911,1.9899999999999999911);
INSERT INTO tracks VALUES(2824,'Torn',227,3,19,NULL,2631291,495262585,1.9899999999999999911);
INSERT INTO tracks VALUES(2825,'A Measure of Salvation',227,3,18,NULL,2563938,489715554,1.9899999999999999911);
INSERT INTO tracks VALUES(2826,'Hero',227,3,18,NULL,2713755,506896959,1.9899999999999999911);
INSERT INTO tracks VALUES(2827,'Unfinished Business',227,3,18,NULL,2622038,528499160,1.9899999999999999911);
INSERT INTO tracks VALUES(2828,'The Passage',227,3,18,NULL,2623875,490375760,1.9899999999999999911);
INSERT INTO tracks VALUES(2829,'The Eye of Jupiter',227,3,18,NULL,2618750,517909587,1.9899999999999999911);
INSERT INTO tracks VALUES(2830,'Rapture',227,3,18,NULL,2624541,508406153,1.9899999999999999911);
INSERT INTO tracks VALUES(2831,'Taking a Break from All Your Worries',227,3,18,NULL,2624207,492700163,1.9899999999999999911);
INSERT INTO tracks VALUES(2832,'The Woman King',227,3,18,NULL,2626376,552893447,1.9899999999999999911);
INSERT INTO tracks VALUES(2833,'A Day In the Life',227,3,18,NULL,2620245,462818231,1.9899999999999999911);
INSERT INTO tracks VALUES(2834,'Dirty Hands',227,3,18,NULL,2627961,537648614,1.9899999999999999911);
INSERT INTO tracks VALUES(2835,'Maelstrom',227,3,18,NULL,2622372,514154275,1.9899999999999999911);
INSERT INTO tracks VALUES(2836,'The Son Also Rises',227,3,18,NULL,2621830,499258498,1.9899999999999999911);
INSERT INTO tracks VALUES(2837,'Crossroads, Pt. 1',227,3,20,NULL,2622622,486233524,1.9899999999999999911);
INSERT INTO tracks VALUES(2838,'Crossroads, Pt. 2',227,3,20,NULL,2869953,497335706,1.9899999999999999911);
INSERT INTO tracks VALUES(2839,'Genesis',228,3,19,NULL,2611986,515671080,1.9899999999999999911);
INSERT INTO tracks VALUES(2840,'Don''t Look Back',228,3,21,NULL,2571154,493628775,1.9899999999999999911);
INSERT INTO tracks VALUES(2841,'One Giant Leap',228,3,21,NULL,2607649,521616246,1.9899999999999999911);
INSERT INTO tracks VALUES(2842,'Collision',228,3,21,NULL,2605480,526182322,1.9899999999999999911);
INSERT INTO tracks VALUES(2843,'Hiros',228,3,21,NULL,2533575,488835454,1.9899999999999999911);
INSERT INTO tracks VALUES(2844,'Better Halves',228,3,21,NULL,2573031,549353481,1.9899999999999999911);
INSERT INTO tracks VALUES(2845,'Nothing to Hide',228,3,19,NULL,2605647,510058181,1.9899999999999999911);
INSERT INTO tracks VALUES(2846,'Seven Minutes to Midnight',228,3,21,NULL,2613988,515590682,1.9899999999999999911);
INSERT INTO tracks VALUES(2847,'Homecoming',228,3,21,NULL,2601351,516015339,1.9899999999999999911);
INSERT INTO tracks VALUES(2848,'Six Months Ago',228,3,19,NULL,2602852,505133869,1.9899999999999999911);
INSERT INTO tracks VALUES(2849,'Fallout',228,3,21,NULL,2594761,501145440,1.9899999999999999911);
INSERT INTO tracks VALUES(2850,'The Fix',228,3,21,NULL,2600266,507026323,1.9899999999999999911);
INSERT INTO tracks VALUES(2851,'Distractions',228,3,21,NULL,2590382,537111289,1.9899999999999999911);
INSERT INTO tracks VALUES(2852,'Run!',228,3,21,NULL,2602602,542936677,1.9899999999999999911);
INSERT INTO tracks VALUES(2853,'Unexpected',228,3,21,NULL,2598139,511777758,1.9899999999999999911);
INSERT INTO tracks VALUES(2854,'Company Man',228,3,21,NULL,2601226,493168135,1.9899999999999999911);
INSERT INTO tracks VALUES(2855,'Company Man',228,3,21,NULL,2601101,503786316,1.9899999999999999911);
INSERT INTO tracks VALUES(2856,'Parasite',228,3,21,NULL,2602727,487461520,1.9899999999999999911);
INSERT INTO tracks VALUES(2857,'A Tale of Two Cities',229,3,19,NULL,2636970,513691652,1.9899999999999999911);
INSERT INTO tracks VALUES(2858,'Lost (Pilot, Part 1) [Premiere]',230,3,19,NULL,2548875,217124866,1.9899999999999999911);
INSERT INTO tracks VALUES(2859,'Man of Science, Man of Faith (Premiere)',231,3,19,NULL,2612250,543342028,1.9899999999999999911);
INSERT INTO tracks VALUES(2860,'Adrift',231,3,19,NULL,2564958,502663995,1.9899999999999999911);
INSERT INTO tracks VALUES(2861,'Lost (Pilot, Part 2)',230,3,19,NULL,2436583,204995876,1.9899999999999999911);
INSERT INTO tracks VALUES(2862,'The Glass Ballerina',229,3,21,NULL,2637458,535729216,1.9899999999999999911);
INSERT INTO tracks VALUES(2863,'Further Instructions',229,3,19,NULL,2563980,502041019,1.9899999999999999911);
INSERT INTO tracks VALUES(2864,'Orientation',231,3,19,NULL,2609083,500600434,1.9899999999999999911);
INSERT INTO tracks VALUES(2865,'Tabula Rasa',230,3,19,NULL,2627105,210526410,1.9899999999999999911);
INSERT INTO tracks VALUES(2866,'Every Man for Himself',229,3,21,NULL,2637387,513803546,1.9899999999999999911);
INSERT INTO tracks VALUES(2867,'Everybody Hates Hugo',231,3,19,NULL,2609192,498163145,1.9899999999999999911);
INSERT INTO tracks VALUES(2868,'Walkabout',230,3,19,NULL,2587370,207748198,1.9899999999999999911);
INSERT INTO tracks VALUES(2869,'...And Found',231,3,19,NULL,2563833,500330548,1.9899999999999999911);
INSERT INTO tracks VALUES(2870,'The Cost of Living',229,3,19,NULL,2637500,505647192,1.9899999999999999911);
INSERT INTO tracks VALUES(2871,'White Rabbit',230,3,19,NULL,2571965,201654606,1.9899999999999999911);
INSERT INTO tracks VALUES(2872,'Abandoned',231,3,19,NULL,2587041,537348711,1.9899999999999999911);
INSERT INTO tracks VALUES(2873,'House of the Rising Sun',230,3,19,NULL,2590032,210379525,1.9899999999999999911);
INSERT INTO tracks VALUES(2874,'I Do',229,3,19,NULL,2627791,504676825,1.9899999999999999911);
INSERT INTO tracks VALUES(2875,'Not In Portland',229,3,21,NULL,2637303,499061234,1.9899999999999999911);
INSERT INTO tracks VALUES(2876,'Not In Portland',229,3,21,NULL,2637345,510546847,1.9899999999999999911);
INSERT INTO tracks VALUES(2877,'The Moth',230,3,19,NULL,2631327,228896396,1.9899999999999999911);
INSERT INTO tracks VALUES(2878,'The Other 48 Days',231,3,19,NULL,2610625,535256753,1.9899999999999999911);
INSERT INTO tracks VALUES(2879,'Collision',231,3,19,NULL,2564916,475656544,1.9899999999999999911);
INSERT INTO tracks VALUES(2880,'Confidence Man',230,3,19,NULL,2615244,223756475,1.9899999999999999911);
INSERT INTO tracks VALUES(2881,'Flashes Before Your Eyes',229,3,21,NULL,2636636,537760755,1.9899999999999999911);
INSERT INTO tracks VALUES(2882,'Lost Survival Guide',229,3,21,NULL,2632590,486675063,1.9899999999999999911);
INSERT INTO tracks VALUES(2883,'Solitary',230,3,19,NULL,2612894,207045178,1.9899999999999999911);
INSERT INTO tracks VALUES(2884,'What Kate Did',231,3,19,NULL,2610250,484583988,1.9899999999999999911);
INSERT INTO tracks VALUES(2885,'Raised By Another',230,3,19,NULL,2590459,223623810,1.9899999999999999911);
INSERT INTO tracks VALUES(2886,'Stranger In a Strange Land',229,3,21,NULL,2636428,505056021,1.9899999999999999911);
INSERT INTO tracks VALUES(2887,'The 23rd Psalm',231,3,19,NULL,2610416,487401604,1.9899999999999999911);
INSERT INTO tracks VALUES(2888,'All the Best Cowboys Have Daddy Issues',230,3,19,NULL,2555492,211743651,1.9899999999999999911);
INSERT INTO tracks VALUES(2889,'The Hunting Party',231,3,21,NULL,2611333,520350364,1.9899999999999999911);
INSERT INTO tracks VALUES(2890,'Tricia Tanaka Is Dead',229,3,21,NULL,2635010,548197162,1.9899999999999999911);
INSERT INTO tracks VALUES(2891,'Enter 77',229,3,21,NULL,2629796,517521422,1.9899999999999999911);
INSERT INTO tracks VALUES(2892,'Fire + Water',231,3,21,NULL,2600333,488458695,1.9899999999999999911);
INSERT INTO tracks VALUES(2893,'Whatever the Case May Be',230,3,19,NULL,2616410,183867185,1.9899999999999999911);
INSERT INTO tracks VALUES(2894,'Hearts and Minds',230,3,19,NULL,2619462,207607466,1.9899999999999999911);
INSERT INTO tracks VALUES(2895,'Par Avion',229,3,21,NULL,2629879,517079642,1.9899999999999999911);
INSERT INTO tracks VALUES(2896,'The Long Con',231,3,19,NULL,2679583,518376636,1.9899999999999999911);
INSERT INTO tracks VALUES(2897,'One of Them',231,3,21,NULL,2698791,542332389,1.9899999999999999911);
INSERT INTO tracks VALUES(2898,'Special',230,3,19,NULL,2618530,219961967,1.9899999999999999911);
INSERT INTO tracks VALUES(2899,'The Man from Tallahassee',229,3,21,NULL,2637637,550893556,1.9899999999999999911);
INSERT INTO tracks VALUES(2900,'Exposé',229,3,21,NULL,2593760,511338017,1.9899999999999999911);
INSERT INTO tracks VALUES(2901,'Homecoming',230,3,19,NULL,2515882,210675221,1.9899999999999999911);
INSERT INTO tracks VALUES(2902,'Maternity Leave',231,3,21,NULL,2780416,555244214,1.9899999999999999911);
INSERT INTO tracks VALUES(2903,'Left Behind',229,3,21,NULL,2635343,538491964,1.9899999999999999911);
INSERT INTO tracks VALUES(2904,'Outlaws',230,3,19,NULL,2619887,206500939,1.9899999999999999911);
INSERT INTO tracks VALUES(2905,'The Whole Truth',231,3,21,NULL,2610125,495487014,1.9899999999999999911);
INSERT INTO tracks VALUES(2906,'...In Translation',230,3,19,NULL,2604575,215441983,1.9899999999999999911);
INSERT INTO tracks VALUES(2907,'Lockdown',231,3,21,NULL,2610250,543886056,1.9899999999999999911);
INSERT INTO tracks VALUES(2908,'One of Us',229,3,21,NULL,2638096,502387276,1.9899999999999999911);
INSERT INTO tracks VALUES(2909,'Catch-22',229,3,21,NULL,2561394,489773399,1.9899999999999999911);
INSERT INTO tracks VALUES(2910,'Dave',231,3,19,NULL,2825166,574325829,1.9899999999999999911);
INSERT INTO tracks VALUES(2911,'Numbers',230,3,19,NULL,2609772,214709143,1.9899999999999999911);
INSERT INTO tracks VALUES(2912,'D.O.C.',229,3,21,NULL,2616032,518556641,1.9899999999999999911);
INSERT INTO tracks VALUES(2913,'Deus Ex Machina',230,3,19,NULL,2582009,214996732,1.9899999999999999911);
INSERT INTO tracks VALUES(2914,'S.O.S.',231,3,19,NULL,2639541,517979269,1.9899999999999999911);
INSERT INTO tracks VALUES(2915,'Do No Harm',230,3,19,NULL,2618487,212039309,1.9899999999999999911);
INSERT INTO tracks VALUES(2916,'Two for the Road',231,3,21,NULL,2610958,502404558,1.9899999999999999911);
INSERT INTO tracks VALUES(2917,'The Greater Good',230,3,19,NULL,2617784,214130273,1.9899999999999999911);
INSERT INTO tracks VALUES(2918,'"?"',231,3,19,NULL,2782333,528227089,1.9899999999999999911);
INSERT INTO tracks VALUES(2919,'Born to Run',230,3,19,NULL,2618619,213772057,1.9899999999999999911);
INSERT INTO tracks VALUES(2920,'Three Minutes',231,3,19,NULL,2763666,531556853,1.9899999999999999911);
INSERT INTO tracks VALUES(2921,'Exodus (Part 1)',230,3,19,NULL,2620747,213107744,1.9899999999999999911);
INSERT INTO tracks VALUES(2922,'Live Together, Die Alone, Pt. 1',231,3,21,NULL,2478041,457364940,1.9899999999999999911);
INSERT INTO tracks VALUES(2923,'Exodus (Part 2) [Season Finale]',230,3,19,NULL,2605557,208667059,1.9899999999999999911);
INSERT INTO tracks VALUES(2924,'Live Together, Die Alone, Pt. 2',231,3,19,NULL,2656531,503619265,1.9899999999999999911);
INSERT INTO tracks VALUES(2925,'Exodus (Part 3) [Season Finale]',230,3,19,NULL,2619869,197937785,1.9899999999999999911);
INSERT INTO tracks VALUES(2926,'Zoo Station',232,1,1,'U2',276349,9056902,0.98999999999999999111);
INSERT INTO tracks VALUES(2927,'Even Better Than The Real Thing',232,1,1,'U2',221361,7279392,0.98999999999999999111);
INSERT INTO tracks VALUES(2928,'One',232,1,1,'U2',276192,9158892,0.98999999999999999111);
INSERT INTO tracks VALUES(2929,'Until The End Of The World',232,1,1,'U2',278700,9132485,0.98999999999999999111);
INSERT INTO tracks VALUES(2930,'Who''s Gonna Ride Your Wild Horses',232,1,1,'U2',316551,10304369,0.98999999999999999111);
INSERT INTO tracks VALUES(2931,'So Cruel',232,1,1,'U2',349492,11527614,0.98999999999999999111);
INSERT INTO tracks VALUES(2932,'The Fly',232,1,1,'U2',268982,8825399,0.98999999999999999111);
INSERT INTO tracks VALUES(2933,'Mysterious Ways',232,1,1,'U2',243826,8062057,0.98999999999999999111);
INSERT INTO tracks VALUES(2934,'Tryin'' To Throw Your Arms Around The World',232,1,1,'U2',232463,7612124,0.98999999999999999111);
INSERT INTO tracks VALUES(2935,'Ultraviolet (Light My Way)',232,1,1,'U2',330788,10754631,0.98999999999999999111);
INSERT INTO tracks VALUES(2936,'Acrobat',232,1,1,'U2',270288,8824723,0.98999999999999999111);
INSERT INTO tracks VALUES(2937,'Love Is Blindness',232,1,1,'U2',263497,8531766,0.98999999999999999111);
INSERT INTO tracks VALUES(2938,'Beautiful Day',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',248163,8056723,0.98999999999999999111);
INSERT INTO tracks VALUES(2939,'Stuck In A Moment You Can''t Get Out Of',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',272378,8997366,0.98999999999999999111);
INSERT INTO tracks VALUES(2940,'Elevation',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',227552,7479414,0.98999999999999999111);
INSERT INTO tracks VALUES(2941,'Walk On',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',296280,9800861,0.98999999999999999111);
INSERT INTO tracks VALUES(2942,'Kite',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',266893,8765761,0.98999999999999999111);
INSERT INTO tracks VALUES(2943,'In A Little While',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',219271,7189647,0.98999999999999999111);
INSERT INTO tracks VALUES(2944,'Wild Honey',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',226768,7466069,0.98999999999999999111);
INSERT INTO tracks VALUES(2945,'Peace On Earth',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',288496,9476171,0.98999999999999999111);
INSERT INTO tracks VALUES(2946,'When I Look At The World',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',257776,8500491,0.98999999999999999111);
INSERT INTO tracks VALUES(2947,'New York',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',330370,10862323,0.98999999999999999111);
INSERT INTO tracks VALUES(2948,'Grace',233,1,1,'Adam Clayton, Bono, Larry Mullen, The Edge',330657,10877148,0.98999999999999999111);
INSERT INTO tracks VALUES(2949,'The Three Sunrises',234,1,1,'U2',234788,7717990,0.98999999999999999111);
INSERT INTO tracks VALUES(2950,'Spanish Eyes',234,1,1,'U2',196702,6392710,0.98999999999999999111);
INSERT INTO tracks VALUES(2951,'Sweetest Thing',234,1,1,'U2',185103,6154896,0.98999999999999999111);
INSERT INTO tracks VALUES(2952,'Love Comes Tumbling',234,1,1,'U2',282671,9328802,0.98999999999999999111);
INSERT INTO tracks VALUES(2953,'Bass Trap',234,1,1,'U2',213289,6834107,0.98999999999999999111);
INSERT INTO tracks VALUES(2954,'Dancing Barefoot',234,1,1,'Ivan Kral/Patti Smith',287895,9488294,0.98999999999999999111);
INSERT INTO tracks VALUES(2955,'Everlasting Love',234,1,1,'Buzz Cason/Mac Gayden',202631,6708932,0.98999999999999999111);
INSERT INTO tracks VALUES(2956,'Unchained Melody',234,1,1,'Alex North/Hy Zaret',294164,9597568,0.98999999999999999111);
INSERT INTO tracks VALUES(2957,'Walk To The Water',234,1,1,'U2',289253,9523336,0.98999999999999999111);
INSERT INTO tracks VALUES(2958,'Luminous Times (Hold On To Love)',234,1,1,'Brian Eno/U2',277760,9015513,0.98999999999999999111);
INSERT INTO tracks VALUES(2959,'Hallelujah Here She Comes',234,1,1,'U2',242364,8027028,0.98999999999999999111);
INSERT INTO tracks VALUES(2960,'Silver And Gold',234,1,1,'Bono',279875,9199746,0.98999999999999999111);
INSERT INTO tracks VALUES(2961,'Endless Deep',234,1,1,'U2',179879,5899070,0.98999999999999999111);
INSERT INTO tracks VALUES(2962,'A Room At The Heartbreak Hotel',234,1,1,'U2',274546,9015416,0.98999999999999999111);
INSERT INTO tracks VALUES(2963,'Trash, Trampoline And The Party Girl',234,1,1,'U2',153965,5083523,0.98999999999999999111);
INSERT INTO tracks VALUES(2964,'Vertigo',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',194612,6329502,0.98999999999999999111);
INSERT INTO tracks VALUES(2965,'Miracle Drug',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',239124,7760916,0.98999999999999999111);
INSERT INTO tracks VALUES(2966,'Sometimes You Can''t Make It On Your Own',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',308976,10112863,0.98999999999999999111);
INSERT INTO tracks VALUES(2967,'Love And Peace Or Else',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',290690,9476723,0.98999999999999999111);
INSERT INTO tracks VALUES(2968,'City Of Blinding Lights',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',347951,11432026,0.98999999999999999111);
INSERT INTO tracks VALUES(2969,'All Because Of You',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',219141,7198014,0.98999999999999999111);
INSERT INTO tracks VALUES(2970,'A Man And A Woman',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',270132,8938285,0.98999999999999999111);
INSERT INTO tracks VALUES(2971,'Crumbs From Your Table',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',303568,9892349,0.98999999999999999111);
INSERT INTO tracks VALUES(2972,'One Step Closer',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',231680,7512912,0.98999999999999999111);
INSERT INTO tracks VALUES(2973,'Original Of The Species',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',281443,9230041,0.98999999999999999111);
INSERT INTO tracks VALUES(2974,'Yahweh',235,1,1,'Adam Clayton, Bono, Larry Mullen & The Edge',262034,8636998,0.98999999999999999111);
INSERT INTO tracks VALUES(2975,'Discotheque',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',319582,10442206,0.98999999999999999111);
INSERT INTO tracks VALUES(2976,'Do You Feel Loved',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',307539,10122694,0.98999999999999999111);
INSERT INTO tracks VALUES(2977,'Mofo',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',349178,11583042,0.98999999999999999111);
INSERT INTO tracks VALUES(2978,'If God Will Send His Angels',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',322533,10563329,0.98999999999999999111);
INSERT INTO tracks VALUES(2979,'Staring At The Sun',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',276924,9082838,0.98999999999999999111);
INSERT INTO tracks VALUES(2980,'Last Night On Earth',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',285753,9401017,0.98999999999999999111);
INSERT INTO tracks VALUES(2981,'Gone',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',266866,8746301,0.98999999999999999111);
INSERT INTO tracks VALUES(2982,'Miami',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',293041,9741603,0.98999999999999999111);
INSERT INTO tracks VALUES(2983,'The Playboy Mansion',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',280555,9274144,0.98999999999999999111);
INSERT INTO tracks VALUES(2984,'If You Wear That Velvet Dress',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',315167,10227333,0.98999999999999999111);
INSERT INTO tracks VALUES(2985,'Please',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',302602,9909484,0.98999999999999999111);
INSERT INTO tracks VALUES(2986,'Wake Up Dead Man',236,1,1,'Bono, The Edge, Adam Clayton, and Larry Mullen',292832,9515903,0.98999999999999999111);
INSERT INTO tracks VALUES(2987,'Helter Skelter',237,1,1,'Lennon, John/McCartney, Paul',187350,6097636,0.98999999999999999111);
INSERT INTO tracks VALUES(2988,'Van Diemen''s Land',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',186044,5990280,0.98999999999999999111);
INSERT INTO tracks VALUES(2989,'Desire',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',179226,5874535,0.98999999999999999111);
INSERT INTO tracks VALUES(2990,'Hawkmoon 269',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',382458,12494987,0.98999999999999999111);
INSERT INTO tracks VALUES(2991,'All Along The Watchtower',237,1,1,'Dylan, Bob',264568,8623572,0.98999999999999999111);
INSERT INTO tracks VALUES(2992,'I Still Haven''t Found What I''m Looking for',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',353567,11542247,0.98999999999999999111);
INSERT INTO tracks VALUES(2993,'Freedom For My People',237,1,1,'Mabins, Macie/Magee, Sterling/Robinson, Bobby',38164,1249764,0.98999999999999999111);
INSERT INTO tracks VALUES(2994,'Silver And Gold',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',349831,11450194,0.98999999999999999111);
INSERT INTO tracks VALUES(2995,'Pride (In The Name Of Love)',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',267807,8806361,0.98999999999999999111);
INSERT INTO tracks VALUES(2996,'Angel Of Harlem',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',229276,7498022,0.98999999999999999111);
INSERT INTO tracks VALUES(2997,'Love Rescue Me',237,1,1,'Bono/Clayton, Adam/Dylan, Bob/Mullen Jr., Larry/The Edge',384522,12508716,0.98999999999999999111);
INSERT INTO tracks VALUES(2998,'When Love Comes To Town',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',255869,8340954,0.98999999999999999111);
INSERT INTO tracks VALUES(2999,'Heartland',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',303360,9867748,0.98999999999999999111);
INSERT INTO tracks VALUES(3000,'God Part II',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',195604,6497570,0.98999999999999999111);
INSERT INTO tracks VALUES(3001,'The Star Spangled Banner',237,1,1,'Hendrix, Jimi',43232,1385810,0.98999999999999999111);
INSERT INTO tracks VALUES(3002,'Bullet The Blue Sky',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',337005,10993607,0.98999999999999999111);
INSERT INTO tracks VALUES(3003,'All I Want Is You',237,1,1,'Bono/Clayton, Adam/Mullen Jr., Larry/The Edge',390243,12729820,0.98999999999999999111);
INSERT INTO tracks VALUES(3004,'Pride (In The Name Of Love)',238,1,1,'U2',230243,7549085,0.98999999999999999111);
INSERT INTO tracks VALUES(3005,'New Year''s Day',238,1,1,'U2',258925,8491818,0.98999999999999999111);
INSERT INTO tracks VALUES(3006,'With Or Without You',238,1,1,'U2',299023,9765188,0.98999999999999999111);
INSERT INTO tracks VALUES(3007,'I Still Haven''t Found What I''m Looking For',238,1,1,'U2',280764,9306737,0.98999999999999999111);
INSERT INTO tracks VALUES(3008,'Sunday Bloody Sunday',238,1,1,'U2',282174,9269668,0.98999999999999999111);
INSERT INTO tracks VALUES(3009,'Bad',238,1,1,'U2',351817,11628058,0.98999999999999999111);
INSERT INTO tracks VALUES(3010,'Where The Streets Have No Name',238,1,1,'U2',276218,9042305,0.98999999999999999111);
INSERT INTO tracks VALUES(3011,'I Will Follow',238,1,1,'U2',218253,7184825,0.98999999999999999111);
INSERT INTO tracks VALUES(3012,'The Unforgettable Fire',238,1,1,'U2',295183,9684664,0.98999999999999999111);
INSERT INTO tracks VALUES(3013,'Sweetest Thing',238,1,1,'U2 & Daragh O''Toole',183066,6071385,0.98999999999999999111);
INSERT INTO tracks VALUES(3014,'Desire',238,1,1,'U2',179853,5893206,0.98999999999999999111);
INSERT INTO tracks VALUES(3015,'When Love Comes To Town',238,1,1,'U2',258194,8479525,0.98999999999999999111);
INSERT INTO tracks VALUES(3016,'Angel Of Harlem',238,1,1,'U2',230217,7527339,0.98999999999999999111);
INSERT INTO tracks VALUES(3017,'All I Want Is You',238,1,1,'U2 & Van Dyke Parks',591986,19202252,0.98999999999999999111);
INSERT INTO tracks VALUES(3018,'Sunday Bloody Sunday',239,1,1,'U2',278204,9140849,0.98999999999999999111);
INSERT INTO tracks VALUES(3019,'Seconds',239,1,1,'U2',191582,6352121,0.98999999999999999111);
INSERT INTO tracks VALUES(3020,'New Year''s Day',239,1,1,'U2',336274,11054732,0.98999999999999999111);
INSERT INTO tracks VALUES(3021,'Like A Song...',239,1,1,'U2',287294,9365379,0.98999999999999999111);
INSERT INTO tracks VALUES(3022,'Drowning Man',239,1,1,'U2',254458,8457066,0.98999999999999999111);
INSERT INTO tracks VALUES(3023,'The Refugee',239,1,1,'U2',221283,7374043,0.98999999999999999111);
INSERT INTO tracks VALUES(3024,'Two Hearts Beat As One',239,1,1,'U2',243487,7998323,0.98999999999999999111);
INSERT INTO tracks VALUES(3025,'Red Light',239,1,1,'U2',225854,7453704,0.98999999999999999111);
INSERT INTO tracks VALUES(3026,'Surrender',239,1,1,'U2',333505,11221406,0.98999999999999999111);
INSERT INTO tracks VALUES(3027,'"40"',239,1,1,'U2',157962,5251767,0.98999999999999999111);
INSERT INTO tracks VALUES(3028,'Zooropa',240,1,1,'U2; Bono',392359,12807979,0.98999999999999999111);
INSERT INTO tracks VALUES(3029,'Babyface',240,1,1,'U2; Bono',241998,7942573,0.98999999999999999111);
INSERT INTO tracks VALUES(3030,'Numb',240,1,1,'U2; Edge, The',260284,8577861,0.98999999999999999111);
INSERT INTO tracks VALUES(3031,'Lemon',240,1,1,'U2; Bono',418324,13988878,0.98999999999999999111);
INSERT INTO tracks VALUES(3032,'Stay (Faraway, So Close!)',240,1,1,'U2; Bono',298475,9785480,0.98999999999999999111);
INSERT INTO tracks VALUES(3033,'Daddy''s Gonna Pay For Your Crashed Car',240,1,1,'U2; Bono',320287,10609581,0.98999999999999999111);
INSERT INTO tracks VALUES(3034,'Some Days Are Better Than Others',240,1,1,'U2; Bono',257436,8417690,0.98999999999999999111);
INSERT INTO tracks VALUES(3035,'The First Time',240,1,1,'U2; Bono',225697,7247651,0.98999999999999999111);
INSERT INTO tracks VALUES(3036,'Dirty Day',240,1,1,'U2; Bono & Edge, The',324440,10652877,0.98999999999999999111);
INSERT INTO tracks VALUES(3037,'The Wanderer',240,1,1,'U2; Bono',283951,9258717,0.98999999999999999111);
INSERT INTO tracks VALUES(3038,'Breakfast In Bed',241,1,8,NULL,196179,6513325,0.98999999999999999111);
INSERT INTO tracks VALUES(3039,'Where Did I Go Wrong',241,1,8,NULL,226742,7485054,0.98999999999999999111);
INSERT INTO tracks VALUES(3040,'I Would Do For You',241,1,8,NULL,334524,11193602,0.98999999999999999111);
INSERT INTO tracks VALUES(3041,'Homely Girl',241,1,8,NULL,203833,6790788,0.98999999999999999111);
INSERT INTO tracks VALUES(3042,'Here I Am (Come And Take Me)',241,1,8,NULL,242102,8106249,0.98999999999999999111);
INSERT INTO tracks VALUES(3043,'Kingston Town',241,1,8,NULL,226951,7638236,0.98999999999999999111);
INSERT INTO tracks VALUES(3044,'Wear You To The Ball',241,1,8,NULL,213342,7159527,0.98999999999999999111);
INSERT INTO tracks VALUES(3045,'(I Can''t Help) Falling In Love With You',241,1,8,NULL,207568,6905623,0.98999999999999999111);
INSERT INTO tracks VALUES(3046,'Higher Ground',241,1,8,NULL,260179,8665244,0.98999999999999999111);
INSERT INTO tracks VALUES(3047,'Bring Me Your Cup',241,1,8,NULL,341498,11346114,0.98999999999999999111);
INSERT INTO tracks VALUES(3048,'C''est La Vie',241,1,8,NULL,270053,9031661,0.98999999999999999111);
INSERT INTO tracks VALUES(3049,'Reggae Music',241,1,8,NULL,245106,8203931,0.98999999999999999111);
INSERT INTO tracks VALUES(3050,'Superstition',241,1,8,NULL,319582,10728099,0.98999999999999999111);
INSERT INTO tracks VALUES(3051,'Until My Dying Day',241,1,8,NULL,235807,7886195,0.98999999999999999111);
INSERT INTO tracks VALUES(3052,'Where Have All The Good Times Gone?',242,1,1,'Ray Davies',186723,6063937,0.98999999999999999111);
INSERT INTO tracks VALUES(3053,'Hang ''Em High',242,1,1,'Alex Van Halen/David Lee Roth/Edward Van Halen/Michael Anthony',210259,6872314,0.98999999999999999111);
INSERT INTO tracks VALUES(3054,'Cathedral',242,1,1,'Alex Van Halen/David Lee Roth/Edward Van Halen/Michael Anthony',82860,2650998,0.98999999999999999111);
INSERT INTO tracks VALUES(3055,'Secrets',242,1,1,'Alex Van Halen/David Lee Roth/Edward Van Halen/Michael Anthony',206968,6803255,0.98999999999999999111);
INSERT INTO tracks VALUES(3056,'Intruder',242,1,1,'Alex Van Halen/David Lee Roth/Edward Van Halen/Michael Anthony',100153,3282142,0.98999999999999999111);
INSERT INTO tracks VALUES(3057,'(Oh) Pretty Woman',242,1,1,'Bill Dees/Roy Orbison',174680,5665828,0.98999999999999999111);
INSERT INTO tracks VALUES(3058,'Dancing In The Street',242,1,1,'Ivy Jo Hunter/Marvin Gaye/William Stevenson',225985,7461499,0.98999999999999999111);
INSERT INTO tracks VALUES(3059,'Little Guitars (Intro)',242,1,1,'Alex Van Halen/David Lee Roth/Edward Van Halen/Michael Anthony',42240,1439530,0.98999999999999999111);
INSERT INTO tracks VALUES(3060,'Little Guitars',242,1,1,'Alex Van Halen/David Lee Roth/Edward Van Halen/Michael Anthony',228806,7453043,0.98999999999999999111);
INSERT INTO tracks VALUES(3061,'Big Bad Bill (Is Sweet William Now)',242,1,1,'Jack Yellen/Milton Ager',165146,5489609,0.98999999999999999111);
INSERT INTO tracks VALUES(3062,'The Full Bug',242,1,1,'Alex Van Halen/David Lee Roth/Edward Van Halen/Michael Anthony',201116,6551013,0.98999999999999999111);
INSERT INTO tracks VALUES(3063,'Happy Trails',242,1,1,'Dale Evans',65488,2111141,0.98999999999999999111);
INSERT INTO tracks VALUES(3064,'Eruption',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth, Michael Anthony',102164,3272891,0.98999999999999999111);
INSERT INTO tracks VALUES(3065,'Ain''t Talkin'' ''bout Love',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth, Michael Anthony',228336,7569506,0.98999999999999999111);
INSERT INTO tracks VALUES(3066,'Runnin'' With The Devil',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth, Michael Anthony',215902,7061901,0.98999999999999999111);
INSERT INTO tracks VALUES(3067,'Dance the Night Away',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth, Michael Anthony',185965,6087433,0.98999999999999999111);
INSERT INTO tracks VALUES(3068,'And the Cradle Will Rock...',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth, Michael Anthony',213968,7011402,0.98999999999999999111);
INSERT INTO tracks VALUES(3069,'Unchained',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth, Michael Anthony',208953,6777078,0.98999999999999999111);
INSERT INTO tracks VALUES(3070,'Jump',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth',241711,7911090,0.98999999999999999111);
INSERT INTO tracks VALUES(3071,'Panama',243,1,1,'Edward Van Halen, Alex Van Halen, David Lee Roth',211853,6921784,0.98999999999999999111);
INSERT INTO tracks VALUES(3072,'Why Can''t This Be Love',243,1,1,'Van Halen',227761,7457655,0.98999999999999999111);
INSERT INTO tracks VALUES(3073,'Dreams',243,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony,/Edward Van Halen, Alex Van Halen, Michael Anthony, Sammy Hagar',291813,9504119,0.98999999999999999111);
INSERT INTO tracks VALUES(3074,'When It''s Love',243,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony,/Edward Van Halen, Alex Van Halen, Michael Anthony, Sammy Hagar',338991,11049966,0.98999999999999999111);
INSERT INTO tracks VALUES(3075,'Poundcake',243,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony,/Edward Van Halen, Alex Van Halen, Michael Anthony, Sammy Hagar',321854,10366978,0.98999999999999999111);
INSERT INTO tracks VALUES(3076,'Right Now',243,1,1,'Van Halen',321828,10503352,0.98999999999999999111);
INSERT INTO tracks VALUES(3077,'Can''t Stop Loving You',243,1,1,'Van Halen',248502,8107896,0.98999999999999999111);
INSERT INTO tracks VALUES(3078,'Humans Being',243,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony,/Edward Van Halen, Alex Van Halen, Michael Anthony, Sammy Hagar',308950,10014683,0.98999999999999999111);
INSERT INTO tracks VALUES(3079,'Can''t Get This Stuff No More',243,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony,/Edward Van Halen, Alex Van Halen, Michael Anthony, David Lee Roth',315376,10355753,0.98999999999999999111);
INSERT INTO tracks VALUES(3080,'Me Wise Magic',243,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony,/Edward Van Halen, Alex Van Halen, Michael Anthony, David Lee Roth',366053,12013467,0.98999999999999999111);
INSERT INTO tracks VALUES(3081,'Runnin'' With The Devil',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',216032,7056863,0.98999999999999999111);
INSERT INTO tracks VALUES(3082,'Eruption',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',102556,3286026,0.98999999999999999111);
INSERT INTO tracks VALUES(3083,'You Really Got Me',244,1,1,'Ray Davies',158589,5194092,0.98999999999999999111);
INSERT INTO tracks VALUES(3084,'Ain''t Talkin'' ''Bout Love',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',230060,7617284,0.98999999999999999111);
INSERT INTO tracks VALUES(3085,'I''m The One',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',226507,7373922,0.98999999999999999111);
INSERT INTO tracks VALUES(3086,'Jamie''s Cryin''',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',210546,6946086,0.98999999999999999111);
INSERT INTO tracks VALUES(3087,'Atomic Punk',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',182073,5908861,0.98999999999999999111);
INSERT INTO tracks VALUES(3088,'Feel Your Love Tonight',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',222850,7293608,0.98999999999999999111);
INSERT INTO tracks VALUES(3089,'Little Dreamer',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',203258,6648122,0.98999999999999999111);
INSERT INTO tracks VALUES(3090,'Ice Cream Man',244,1,1,'John Brim',200306,6573145,0.98999999999999999111);
INSERT INTO tracks VALUES(3091,'On Fire',244,1,1,'Edward Van Halen, Alex Van Halen, Michael Anthony and David Lee Roth',180636,5879235,0.98999999999999999111);
INSERT INTO tracks VALUES(3092,'Neworld',245,1,1,'Van Halen',105639,3495897,0.98999999999999999111);
INSERT INTO tracks VALUES(3093,'Without You',245,1,1,'Van Halen',390295,12619558,0.98999999999999999111);
INSERT INTO tracks VALUES(3094,'One I Want',245,1,1,'Van Halen',330788,10743970,0.98999999999999999111);
INSERT INTO tracks VALUES(3095,'From Afar',245,1,1,'Van Halen',324414,10524554,0.98999999999999999111);
INSERT INTO tracks VALUES(3096,'Dirty Water Dog',245,1,1,'Van Halen',327392,10709202,0.98999999999999999111);
INSERT INTO tracks VALUES(3097,'Once',245,1,1,'Van Halen',462837,15378082,0.98999999999999999111);
INSERT INTO tracks VALUES(3098,'Fire in the Hole',245,1,1,'Van Halen',331728,10846768,0.98999999999999999111);
INSERT INTO tracks VALUES(3099,'Josephina',245,1,1,'Van Halen',342491,11161521,0.98999999999999999111);
INSERT INTO tracks VALUES(3100,'Year to the Day',245,1,1,'Van Halen',514612,16621333,0.98999999999999999111);
INSERT INTO tracks VALUES(3101,'Primary',245,1,1,'Van Halen',86987,2812555,0.98999999999999999111);
INSERT INTO tracks VALUES(3102,'Ballot or the Bullet',245,1,1,'Van Halen',342282,11212955,0.98999999999999999111);
INSERT INTO tracks VALUES(3103,'How Many Say I',245,1,1,'Van Halen',363937,11716855,0.98999999999999999111);
INSERT INTO tracks VALUES(3104,'Sucker Train Blues',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',267859,8738780,0.98999999999999999111);
INSERT INTO tracks VALUES(3105,'Do It For The Kids',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',235911,7693331,0.98999999999999999111);
INSERT INTO tracks VALUES(3106,'Big Machine',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',265613,8673442,0.98999999999999999111);
INSERT INTO tracks VALUES(3107,'Illegal I Song',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',257750,8483347,0.98999999999999999111);
INSERT INTO tracks VALUES(3108,'Spectacle',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',221701,7252876,0.98999999999999999111);
INSERT INTO tracks VALUES(3109,'Fall To Pieces',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',270889,8823096,0.98999999999999999111);
INSERT INTO tracks VALUES(3110,'Headspace',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',223033,7237986,0.98999999999999999111);
INSERT INTO tracks VALUES(3111,'Superhuman',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',255921,8365328,0.98999999999999999111);
INSERT INTO tracks VALUES(3112,'Set Me Free',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',247954,8053388,0.98999999999999999111);
INSERT INTO tracks VALUES(3113,'You Got No Right',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',335412,10991094,0.98999999999999999111);
INSERT INTO tracks VALUES(3114,'Slither',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',248398,8118785,0.98999999999999999111);
INSERT INTO tracks VALUES(3115,'Dirty Little Thing',246,1,1,'Dave Kushner, Duff, Keith Nelson, Matt Sorum, Scott Weiland & Slash',237844,7732982,0.98999999999999999111);
INSERT INTO tracks VALUES(3116,'Loving The Alien',246,1,1,'Dave Kushner, Duff, Matt Sorum, Scott Weiland & Slash',348786,11412762,0.98999999999999999111);
INSERT INTO tracks VALUES(3117,'Pela Luz Dos Olhos Teus',247,1,7,NULL,119196,3905715,0.98999999999999999111);
INSERT INTO tracks VALUES(3118,'A Bencao E Outros',247,1,7,NULL,421093,14234427,0.98999999999999999111);
INSERT INTO tracks VALUES(3119,'Tudo Na Mais Santa Paz',247,1,7,NULL,222406,7426757,0.98999999999999999111);
INSERT INTO tracks VALUES(3120,'O Velho E Aflor',247,1,7,NULL,275121,9126828,0.98999999999999999111);
INSERT INTO tracks VALUES(3121,'Cotidiano N 2',247,1,7,NULL,55902,1805797,0.98999999999999999111);
INSERT INTO tracks VALUES(3122,'Adeus',247,1,7,NULL,221884,7259351,0.98999999999999999111);
INSERT INTO tracks VALUES(3123,'Samba Pra Endrigo',247,1,7,NULL,259265,8823551,0.98999999999999999111);
INSERT INTO tracks VALUES(3124,'So Por Amor',247,1,7,NULL,236591,7745764,0.98999999999999999111);
INSERT INTO tracks VALUES(3125,'Meu Pranto Rolou',247,1,7,NULL,181760,6003345,0.98999999999999999111);
INSERT INTO tracks VALUES(3126,'Mulher Carioca',247,1,7,NULL,191686,6395048,0.98999999999999999111);
INSERT INTO tracks VALUES(3127,'Um Homem Chamado Alfredo',247,1,7,NULL,151640,4976227,0.98999999999999999111);
INSERT INTO tracks VALUES(3128,'Samba Do Jato',247,1,7,NULL,220813,7357840,0.98999999999999999111);
INSERT INTO tracks VALUES(3129,'Oi, La',247,1,7,NULL,167053,5562700,0.98999999999999999111);
INSERT INTO tracks VALUES(3130,'Vinicius, Poeta Do Encontro',247,1,7,NULL,336431,10858776,0.98999999999999999111);
INSERT INTO tracks VALUES(3131,'Soneto Da Separacao',247,1,7,NULL,193880,6277511,0.98999999999999999111);
INSERT INTO tracks VALUES(3132,'Still Of The Night',141,1,3,'Sykes',398210,13043817,0.98999999999999999111);
INSERT INTO tracks VALUES(3133,'Here I Go Again',141,1,3,'Marsden',233874,7652473,0.98999999999999999111);
INSERT INTO tracks VALUES(3134,'Is This Love',141,1,3,'Sykes',283924,9262360,0.98999999999999999111);
INSERT INTO tracks VALUES(3135,'Love Ain''t No Stranger',141,1,3,'Galley',259395,8490428,0.98999999999999999111);
INSERT INTO tracks VALUES(3136,'Looking For Love',141,1,3,'Sykes',391941,12769847,0.98999999999999999111);
INSERT INTO tracks VALUES(3137,'Now You''re Gone',141,1,3,'Vandenberg',251141,8162193,0.98999999999999999111);
INSERT INTO tracks VALUES(3138,'Slide It In',141,1,3,'Coverdale',202475,6615152,0.98999999999999999111);
INSERT INTO tracks VALUES(3139,'Slow An'' Easy',141,1,3,'Moody',367255,11961332,0.98999999999999999111);
INSERT INTO tracks VALUES(3140,'Judgement Day',141,1,3,'Vandenberg',317074,10326997,0.98999999999999999111);
INSERT INTO tracks VALUES(3141,'You''re Gonna Break My Hart Again',141,1,3,'Sykes',250853,8176847,0.98999999999999999111);
INSERT INTO tracks VALUES(3142,'The Deeper The Love',141,1,3,'Vandenberg',262791,8606504,0.98999999999999999111);
INSERT INTO tracks VALUES(3143,'Crying In The Rain',141,1,3,'Coverdale',337005,10931921,0.98999999999999999111);
INSERT INTO tracks VALUES(3144,'Fool For Your Loving',141,1,3,'Marsden/Moody',250801,8129820,0.98999999999999999111);
INSERT INTO tracks VALUES(3145,'Sweet Lady Luck',141,1,3,'Vandenberg',273737,8919163,0.98999999999999999111);
INSERT INTO tracks VALUES(3146,'Faixa Amarela',248,1,7,'Beto Gogo/Jessé Pai/Luiz Carlos/Zeca Pagodinho',240692,8082036,0.98999999999999999111);
INSERT INTO tracks VALUES(3147,'Posso Até Me Apaixonar',248,1,7,'Dudu Nobre',200698,6735526,0.98999999999999999111);
INSERT INTO tracks VALUES(3148,'Não Sou Mais Disso',248,1,7,'Jorge Aragão/Zeca Pagodinho',225985,7613817,0.98999999999999999111);
INSERT INTO tracks VALUES(3149,'Vivo Isolado Do Mundo',248,1,7,'Alcides Dias Lopes',180035,6073995,0.98999999999999999111);
INSERT INTO tracks VALUES(3150,'Coração Em Desalinho',248,1,7,'Mauro Diniz/Ratino Sigem',185208,6225948,0.98999999999999999111);
INSERT INTO tracks VALUES(3151,'Seu Balancê',248,1,7,'Paulinho Rezende/Toninho Geraes',219454,7311219,0.98999999999999999111);
INSERT INTO tracks VALUES(3152,'Vai Adiar',248,1,7,'Alcino Corrêa/Monarco',270393,9134882,0.98999999999999999111);
INSERT INTO tracks VALUES(3153,'Rugas',248,1,7,'Augusto Garcez/Nelson Cavaquinho',140930,4703182,0.98999999999999999111);
INSERT INTO tracks VALUES(3154,'Feirinha da Pavuna/Luz do Repente/Bagaço da Laranja',248,1,7,'Arlindo Cruz/Franco/Marquinhos PQD/Negro, Jovelina Pérolo/Zeca Pagodinho',107206,3593684,0.98999999999999999111);
INSERT INTO tracks VALUES(3155,'Sem Essa de Malandro Agulha',248,1,7,'Aldir Blanc/Jayme Vignoli',158484,5332668,0.98999999999999999111);
INSERT INTO tracks VALUES(3156,'Chico Não Vai na Corimba',248,1,7,'Dudu Nobre/Zeca Pagodinho',269374,9122188,0.98999999999999999111);
INSERT INTO tracks VALUES(3157,'Papel Principal',248,1,7,'Almir Guineto/Dedé Paraiso/Luverci Ernesto',217495,7325302,0.98999999999999999111);
INSERT INTO tracks VALUES(3158,'Saudade Louca',248,1,7,'Acyr Marques/Arlindo Cruz/Franco',243591,8136475,0.98999999999999999111);
INSERT INTO tracks VALUES(3159,'Camarão que Dorme e Onda Leva',248,1,7,'Acyi Marques/Arlindo Bruz/Braço, Beto Sem/Zeca Pagodinho',299102,10012231,0.98999999999999999111);
INSERT INTO tracks VALUES(3160,'Sapopemba e Maxambomba',248,1,7,'Nei Lopes/Wilson Moreira',245394,8268712,0.98999999999999999111);
INSERT INTO tracks VALUES(3161,'Minha Fé',248,1,7,'Murilão',206994,6981474,0.98999999999999999111);
INSERT INTO tracks VALUES(3162,'Lua de Ogum',248,1,7,'Ratinho/Zeca Pagodinho',168463,5719129,0.98999999999999999111);
INSERT INTO tracks VALUES(3163,'Samba pras moças',248,1,7,'Grazielle/Roque Ferreira',152816,5121366,0.98999999999999999111);
INSERT INTO tracks VALUES(3164,'Verdade',248,1,7,'Carlinhos Santana/Nelson Rufino',332826,11120708,0.98999999999999999111);
INSERT INTO tracks VALUES(3165,'The Brig',229,3,21,NULL,2617325,488919543,1.9899999999999999911);
INSERT INTO tracks VALUES(3166,'.07%',228,3,21,NULL,2585794,541715199,1.9899999999999999911);
INSERT INTO tracks VALUES(3167,'Five Years Gone',228,3,21,NULL,2587712,530551890,1.9899999999999999911);
INSERT INTO tracks VALUES(3168,'The Hard Part',228,3,21,NULL,2601017,475996611,1.9899999999999999911);
INSERT INTO tracks VALUES(3169,'The Man Behind the Curtain',229,3,21,NULL,2615990,493951081,1.9899999999999999911);
INSERT INTO tracks VALUES(3170,'Greatest Hits',229,3,21,NULL,2617117,522102916,1.9899999999999999911);
INSERT INTO tracks VALUES(3171,'Landslide',228,3,21,NULL,2600725,518677861,1.9899999999999999911);
INSERT INTO tracks VALUES(3172,'The Office: An American Workplace (Pilot)',249,3,19,NULL,1380833,290482361,1.9899999999999999911);
INSERT INTO tracks VALUES(3173,'Diversity Day',249,3,19,NULL,1306416,257879716,1.9899999999999999911);
INSERT INTO tracks VALUES(3174,'Health Care',249,3,19,NULL,1321791,260493577,1.9899999999999999911);
INSERT INTO tracks VALUES(3175,'The Alliance',249,3,19,NULL,1317125,266203162,1.9899999999999999911);
INSERT INTO tracks VALUES(3176,'Basketball',249,3,19,NULL,1323541,267464180,1.9899999999999999911);
INSERT INTO tracks VALUES(3177,'Hot Girl',249,3,19,NULL,1325458,267836576,1.9899999999999999911);
INSERT INTO tracks VALUES(3178,'The Dundies',250,3,19,NULL,1253541,246845576,1.9899999999999999911);
INSERT INTO tracks VALUES(3179,'Sexual Harassment',250,3,19,NULL,1294541,273069146,1.9899999999999999911);
INSERT INTO tracks VALUES(3180,'Office Olympics',250,3,19,NULL,1290458,256247623,1.9899999999999999911);
INSERT INTO tracks VALUES(3181,'The Fire',250,3,19,NULL,1288166,266856017,1.9899999999999999911);
INSERT INTO tracks VALUES(3182,'Halloween',250,3,19,NULL,1315333,249205209,1.9899999999999999911);
INSERT INTO tracks VALUES(3183,'The Fight',250,3,19,NULL,1320028,277149457,1.9899999999999999911);
INSERT INTO tracks VALUES(3184,'The Client',250,3,19,NULL,1299341,253836788,1.9899999999999999911);
INSERT INTO tracks VALUES(3185,'Performance Review',250,3,19,NULL,1292458,256143822,1.9899999999999999911);
INSERT INTO tracks VALUES(3186,'Email Surveillance',250,3,19,NULL,1328870,265101113,1.9899999999999999911);
INSERT INTO tracks VALUES(3187,'Christmas Party',250,3,19,NULL,1282115,260891300,1.9899999999999999911);
INSERT INTO tracks VALUES(3188,'Booze Cruise',250,3,19,NULL,1267958,252518021,1.9899999999999999911);
INSERT INTO tracks VALUES(3189,'The Injury',250,3,19,NULL,1275275,253912762,1.9899999999999999911);
INSERT INTO tracks VALUES(3190,'The Secret',250,3,19,NULL,1264875,253143200,1.9899999999999999911);
INSERT INTO tracks VALUES(3191,'The Carpet',250,3,19,NULL,1264375,256477011,1.9899999999999999911);
INSERT INTO tracks VALUES(3192,'Boys and Girls',250,3,19,NULL,1278333,255245729,1.9899999999999999911);
INSERT INTO tracks VALUES(3193,'Valentine''s Day',250,3,19,NULL,1270375,253552710,1.9899999999999999911);
INSERT INTO tracks VALUES(3194,'Dwight''s Speech',250,3,19,NULL,1278041,255001728,1.9899999999999999911);
INSERT INTO tracks VALUES(3195,'Take Your Daughter to Work Day',250,3,19,NULL,1268333,253451012,1.9899999999999999911);
INSERT INTO tracks VALUES(3196,'Michael''s Birthday',250,3,19,NULL,1237791,247238398,1.9899999999999999911);
INSERT INTO tracks VALUES(3197,'Drug Testing',250,3,19,NULL,1278625,244626927,1.9899999999999999911);
INSERT INTO tracks VALUES(3198,'Conflict Resolution',250,3,19,NULL,1274583,253808658,1.9899999999999999911);
INSERT INTO tracks VALUES(3199,'Casino Night - Season Finale',250,3,19,NULL,1712791,327642458,1.9899999999999999911);
INSERT INTO tracks VALUES(3200,'Gay Witch Hunt',251,3,19,NULL,1326534,276942637,1.9899999999999999911);
INSERT INTO tracks VALUES(3201,'The Convention',251,3,19,NULL,1297213,255117055,1.9899999999999999911);
INSERT INTO tracks VALUES(3202,'The Coup',251,3,19,NULL,1276526,267205501,1.9899999999999999911);
INSERT INTO tracks VALUES(3203,'Grief Counseling',251,3,19,NULL,1282615,256912833,1.9899999999999999911);
INSERT INTO tracks VALUES(3204,'The Initiation',251,3,19,NULL,1280113,251728257,1.9899999999999999911);
INSERT INTO tracks VALUES(3205,'Diwali',251,3,19,NULL,1279904,252726644,1.9899999999999999911);
INSERT INTO tracks VALUES(3206,'Branch Closing',251,3,19,NULL,1822781,358761786,1.9899999999999999911);
INSERT INTO tracks VALUES(3207,'The Merger',251,3,19,NULL,1801926,345960631,1.9899999999999999911);
INSERT INTO tracks VALUES(3208,'The Convict',251,3,22,NULL,1273064,248863427,1.9899999999999999911);
INSERT INTO tracks VALUES(3209,'A Benihana Christmas, Pts. 1 & 2',251,3,22,NULL,2519436,515301752,1.9899999999999999911);
INSERT INTO tracks VALUES(3210,'Back from Vacation',251,3,22,NULL,1271688,245378749,1.9899999999999999911);
INSERT INTO tracks VALUES(3211,'Traveling Salesmen',251,3,22,NULL,1289039,250822697,1.9899999999999999911);
INSERT INTO tracks VALUES(3212,'Producer''s Cut: The Return',251,3,22,NULL,1700241,337219980,1.9899999999999999911);
INSERT INTO tracks VALUES(3213,'Ben Franklin',251,3,22,NULL,1271938,264168080,1.9899999999999999911);
INSERT INTO tracks VALUES(3214,'Phyllis''s Wedding',251,3,22,NULL,1271521,258561054,1.9899999999999999911);
INSERT INTO tracks VALUES(3215,'Business School',251,3,22,NULL,1302093,254402605,1.9899999999999999911);
INSERT INTO tracks VALUES(3216,'Cocktails',251,3,22,NULL,1272522,259011909,1.9899999999999999911);
INSERT INTO tracks VALUES(3217,'The Negotiation',251,3,22,NULL,1767851,371663719,1.9899999999999999911);
INSERT INTO tracks VALUES(3218,'Safety Training',251,3,22,NULL,1271229,253054534,1.9899999999999999911);
INSERT INTO tracks VALUES(3219,'Product Recall',251,3,22,NULL,1268268,251208610,1.9899999999999999911);
INSERT INTO tracks VALUES(3220,'Women''s Appreciation',251,3,22,NULL,1732649,338778844,1.9899999999999999911);
INSERT INTO tracks VALUES(3221,'Beach Games',251,3,22,NULL,1676134,333671149,1.9899999999999999911);
INSERT INTO tracks VALUES(3222,'The Job',251,3,22,NULL,2541875,501060138,1.9899999999999999911);
INSERT INTO tracks VALUES(3223,'How to Stop an Exploding Man',228,3,21,NULL,2687103,487881159,1.9899999999999999911);
INSERT INTO tracks VALUES(3224,'Through a Looking Glass',229,3,21,NULL,5088838,1059546140,1.9899999999999999911);
INSERT INTO tracks VALUES(3225,'Your Time Is Gonna Come',252,2,1,'Page, Jones',310774,5126563,0.98999999999999999111);
INSERT INTO tracks VALUES(3226,'Battlestar Galactica, Pt. 1',253,3,20,NULL,2952702,541359437,1.9899999999999999911);
INSERT INTO tracks VALUES(3227,'Battlestar Galactica, Pt. 2',253,3,20,NULL,2956081,521387924,1.9899999999999999911);
INSERT INTO tracks VALUES(3228,'Battlestar Galactica, Pt. 3',253,3,20,NULL,2927802,554509033,1.9899999999999999911);
INSERT INTO tracks VALUES(3229,'Lost Planet of the Gods, Pt. 1',253,3,20,NULL,2922547,537812711,1.9899999999999999911);
INSERT INTO tracks VALUES(3230,'Lost Planet of the Gods, Pt. 2',253,3,20,NULL,2914664,534343985,1.9899999999999999911);
INSERT INTO tracks VALUES(3231,'The Lost Warrior',253,3,20,NULL,2920045,558872190,1.9899999999999999911);
INSERT INTO tracks VALUES(3232,'The Long Patrol',253,3,20,NULL,2925008,513122217,1.9899999999999999911);
INSERT INTO tracks VALUES(3233,'The Gun On Ice Planet Zero, Pt. 1',253,3,20,NULL,2907615,540980196,1.9899999999999999911);
INSERT INTO tracks VALUES(3234,'The Gun On Ice Planet Zero, Pt. 2',253,3,20,NULL,2924341,546542281,1.9899999999999999911);
INSERT INTO tracks VALUES(3235,'The Magnificent Warriors',253,3,20,NULL,2924716,570152232,1.9899999999999999911);
INSERT INTO tracks VALUES(3236,'The Young Lords',253,3,20,NULL,2863571,587051735,1.9899999999999999911);
INSERT INTO tracks VALUES(3237,'The Living Legend, Pt. 1',253,3,20,NULL,2924507,503641007,1.9899999999999999911);
INSERT INTO tracks VALUES(3238,'The Living Legend, Pt. 2',253,3,20,NULL,2923298,515632754,1.9899999999999999911);
INSERT INTO tracks VALUES(3239,'Fire In Space',253,3,20,NULL,2926593,536784757,1.9899999999999999911);
INSERT INTO tracks VALUES(3240,'War of the Gods, Pt. 1',253,3,20,NULL,2922630,505761343,1.9899999999999999911);
INSERT INTO tracks VALUES(3241,'War of the Gods, Pt. 2',253,3,20,NULL,2923381,487899692,1.9899999999999999911);
INSERT INTO tracks VALUES(3242,'The Man With Nine Lives',253,3,20,NULL,2956998,577829804,1.9899999999999999911);
INSERT INTO tracks VALUES(3243,'Murder On the Rising Star',253,3,20,NULL,2935894,551759986,1.9899999999999999911);
INSERT INTO tracks VALUES(3244,'Greetings from Earth, Pt. 1',253,3,20,NULL,2960293,536824558,1.9899999999999999911);
INSERT INTO tracks VALUES(3245,'Greetings from Earth, Pt. 2',253,3,20,NULL,2903778,527842860,1.9899999999999999911);
INSERT INTO tracks VALUES(3246,'Baltar''s Escape',253,3,20,NULL,2922088,525564224,1.9899999999999999911);
INSERT INTO tracks VALUES(3247,'Experiment In Terra',253,3,20,NULL,2923548,547982556,1.9899999999999999911);
INSERT INTO tracks VALUES(3248,'Take the Celestra',253,3,20,NULL,2927677,512381289,1.9899999999999999911);
INSERT INTO tracks VALUES(3249,'The Hand of God',253,3,20,NULL,2924007,536583079,1.9899999999999999911);
INSERT INTO tracks VALUES(3250,'Pilot',254,3,19,NULL,2484567,492670102,1.9899999999999999911);
INSERT INTO tracks VALUES(3251,'Through the Looking Glass, Pt. 2',229,3,21,NULL,2617117,550943353,1.9899999999999999911);
INSERT INTO tracks VALUES(3252,'Through the Looking Glass, Pt. 1',229,3,21,NULL,2610860,493211809,1.9899999999999999911);
INSERT INTO tracks VALUES(3253,'Instant Karma',255,2,9,NULL,193188,3150090,0.98999999999999999111);
INSERT INTO tracks VALUES(3254,'#9 Dream',255,2,9,NULL,278312,4506425,0.98999999999999999111);
INSERT INTO tracks VALUES(3255,'Mother',255,2,9,NULL,287740,4656660,0.98999999999999999111);
INSERT INTO tracks VALUES(3256,'Give Peace a Chance',255,2,9,NULL,274644,4448025,0.98999999999999999111);
INSERT INTO tracks VALUES(3257,'Cold Turkey',255,2,9,NULL,281424,4556003,0.98999999999999999111);
INSERT INTO tracks VALUES(3258,'Whatever Gets You Thru the Night',255,2,9,NULL,215084,3499018,0.98999999999999999111);
INSERT INTO tracks VALUES(3259,'I''m Losing You',255,2,9,NULL,240719,3907467,0.98999999999999999111);
INSERT INTO tracks VALUES(3260,'Gimme Some Truth',255,2,9,NULL,232778,3780807,0.98999999999999999111);
INSERT INTO tracks VALUES(3261,'Oh, My Love',255,2,9,NULL,159473,2612788,0.98999999999999999111);
INSERT INTO tracks VALUES(3262,'Imagine',255,2,9,NULL,192329,3136271,0.98999999999999999111);
INSERT INTO tracks VALUES(3263,'Nobody Told Me',255,2,9,NULL,210348,3423395,0.98999999999999999111);
INSERT INTO tracks VALUES(3264,'Jealous Guy',255,2,9,NULL,239094,3881620,0.98999999999999999111);
INSERT INTO tracks VALUES(3265,'Working Class Hero',255,2,9,NULL,265449,4301430,0.98999999999999999111);
INSERT INTO tracks VALUES(3266,'Power to the People',255,2,9,NULL,213018,3466029,0.98999999999999999111);
INSERT INTO tracks VALUES(3267,'Imagine',255,2,9,NULL,219078,3562542,0.98999999999999999111);
INSERT INTO tracks VALUES(3268,'Beautiful Boy',255,2,9,NULL,227995,3704642,0.98999999999999999111);
INSERT INTO tracks VALUES(3269,'Isolation',255,2,9,NULL,156059,2558399,0.98999999999999999111);
INSERT INTO tracks VALUES(3270,'Watching the Wheels',255,2,9,NULL,198645,3237063,0.98999999999999999111);
INSERT INTO tracks VALUES(3271,'Grow Old With Me',255,2,9,NULL,149093,2447453,0.98999999999999999111);
INSERT INTO tracks VALUES(3272,'Gimme Some Truth',255,2,9,NULL,187546,3060083,0.98999999999999999111);
INSERT INTO tracks VALUES(3273,'[Just Like] Starting Over',255,2,9,NULL,215549,3506308,0.98999999999999999111);
INSERT INTO tracks VALUES(3274,'God',255,2,9,NULL,260410,4221135,0.98999999999999999111);
INSERT INTO tracks VALUES(3275,'Real Love',255,2,9,NULL,236911,3846658,0.98999999999999999111);
INSERT INTO tracks VALUES(3276,'Sympton of the Universe',256,2,1,NULL,340890,5489313,0.98999999999999999111);
INSERT INTO tracks VALUES(3277,'Snowblind',256,2,1,NULL,295960,4773171,0.98999999999999999111);
INSERT INTO tracks VALUES(3278,'Black Sabbath',256,2,1,NULL,364180,5860455,0.98999999999999999111);
INSERT INTO tracks VALUES(3279,'Fairies Wear Boots',256,2,1,NULL,392764,6315916,0.98999999999999999111);
INSERT INTO tracks VALUES(3280,'War Pigs',256,2,1,NULL,515435,8270194,0.98999999999999999111);
INSERT INTO tracks VALUES(3281,'The Wizard',256,2,1,NULL,282678,4561796,0.98999999999999999111);
INSERT INTO tracks VALUES(3282,'N.I.B.',256,2,1,NULL,335248,5399456,0.98999999999999999111);
INSERT INTO tracks VALUES(3283,'Sweet Leaf',256,2,1,NULL,354706,5709700,0.98999999999999999111);
INSERT INTO tracks VALUES(3284,'Never Say Die',256,2,1,NULL,258343,4173799,0.98999999999999999111);
INSERT INTO tracks VALUES(3285,'Sabbath, Bloody Sabbath',256,2,1,NULL,333622,5373633,0.98999999999999999111);
INSERT INTO tracks VALUES(3286,'Iron Man/Children of the Grave',256,2,1,NULL,552308,8858616,0.98999999999999999111);
INSERT INTO tracks VALUES(3287,'Paranoid',256,2,1,NULL,189171,3071042,0.98999999999999999111);
INSERT INTO tracks VALUES(3288,'Rock You Like a Hurricane',257,2,1,NULL,255766,4300973,0.98999999999999999111);
INSERT INTO tracks VALUES(3289,'No One Like You',257,2,1,NULL,240325,4050259,0.98999999999999999111);
INSERT INTO tracks VALUES(3290,'The Zoo',257,2,1,NULL,332740,5550779,0.98999999999999999111);
INSERT INTO tracks VALUES(3291,'Loving You Sunday Morning',257,2,1,NULL,339125,5654493,0.98999999999999999111);
INSERT INTO tracks VALUES(3292,'Still Loving You',257,2,1,NULL,390674,6491444,0.98999999999999999111);
INSERT INTO tracks VALUES(3293,'Big City Nights',257,2,1,NULL,251865,4237651,0.98999999999999999111);
INSERT INTO tracks VALUES(3294,'Believe in Love',257,2,1,NULL,325774,5437651,0.98999999999999999111);
INSERT INTO tracks VALUES(3295,'Rhythm of Love',257,2,1,NULL,231246,3902834,0.98999999999999999111);
INSERT INTO tracks VALUES(3296,'I Can''t Explain',257,2,1,NULL,205332,3482099,0.98999999999999999111);
INSERT INTO tracks VALUES(3297,'Tease Me Please Me',257,2,1,NULL,287229,4811894,0.98999999999999999111);
INSERT INTO tracks VALUES(3298,'Wind of Change',257,2,1,NULL,315325,5268002,0.98999999999999999111);
INSERT INTO tracks VALUES(3299,'Send Me an Angel',257,2,1,NULL,273041,4581492,0.98999999999999999111);
INSERT INTO tracks VALUES(3300,'Jump Around',258,1,17,'E. Schrody/L. Muggerud',217835,8715653,0.98999999999999999111);
INSERT INTO tracks VALUES(3301,'Salutations',258,1,17,'E. Schrody/L. Dimant',69120,2767047,0.98999999999999999111);
INSERT INTO tracks VALUES(3302,'Put Your Head Out',258,1,17,'E. Schrody/L. Freese/L. Muggerud',182230,7291473,0.98999999999999999111);
INSERT INTO tracks VALUES(3303,'Top O'' The Morning To Ya',258,1,17,'E. Schrody/L. Dimant',216633,8667599,0.98999999999999999111);
INSERT INTO tracks VALUES(3304,'Commercial 1',258,1,17,'L. Muggerud',7941,319888,0.98999999999999999111);
INSERT INTO tracks VALUES(3305,'House And The Rising Sun',258,1,17,'E. Schrody/J. Vasquez/L. Dimant',219402,8778369,0.98999999999999999111);
INSERT INTO tracks VALUES(3306,'Shamrocks And Shenanigans',258,1,17,'E. Schrody/L. Dimant',218331,8735518,0.98999999999999999111);
INSERT INTO tracks VALUES(3307,'House Of Pain Anthem',258,1,17,'E. Schrody/L. Dimant',155611,6226713,0.98999999999999999111);
INSERT INTO tracks VALUES(3308,'Danny Boy, Danny Boy',258,1,17,'E. Schrody/L. Muggerud',114520,4583091,0.98999999999999999111);
INSERT INTO tracks VALUES(3309,'Guess Who''s Back',258,1,17,'E. Schrody/L. Muggerud',238393,9537994,0.98999999999999999111);
INSERT INTO tracks VALUES(3310,'Commercial 2',258,1,17,'L. Muggerud',21211,850698,0.98999999999999999111);
INSERT INTO tracks VALUES(3311,'Put On Your Shit Kickers',258,1,17,'E. Schrody/L. Muggerud',190432,7619569,0.98999999999999999111);
INSERT INTO tracks VALUES(3312,'Come And Get Some Of This',258,1,17,'E. Schrody/L. Muggerud/R. Medrano',170475,6821279,0.98999999999999999111);
INSERT INTO tracks VALUES(3313,'Life Goes On',258,1,17,'E. Schrody/R. Medrano',163030,6523458,0.98999999999999999111);
INSERT INTO tracks VALUES(3314,'One For The Road',258,1,17,'E. Schrody/L. Dimant/L. Muggerud',170213,6810820,0.98999999999999999111);
INSERT INTO tracks VALUES(3315,'Feel It',258,1,17,'E. Schrody/R. Medrano',239908,9598588,0.98999999999999999111);
INSERT INTO tracks VALUES(3316,'All My Love',258,1,17,'E. Schrody/L. Dimant',200620,8027065,0.98999999999999999111);
INSERT INTO tracks VALUES(3317,'Jump Around (Pete Rock Remix)',258,1,17,'E. Schrody/L. Muggerud',236120,9447101,0.98999999999999999111);
INSERT INTO tracks VALUES(3318,'Shamrocks And Shenanigans (Boom Shalock Lock Boom/Butch Vig Mix)',258,1,17,'E. Schrody/L. Dimant',237035,9483705,0.98999999999999999111);
INSERT INTO tracks VALUES(3319,'Instinto Colectivo',259,1,15,NULL,300564,12024875,0.98999999999999999111);
INSERT INTO tracks VALUES(3320,'Chapa o Coco',259,1,15,NULL,143830,5755478,0.98999999999999999111);
INSERT INTO tracks VALUES(3321,'Prostituta',259,1,15,NULL,359000,14362307,0.98999999999999999111);
INSERT INTO tracks VALUES(3322,'Eu So Queria Sumir',259,1,15,NULL,269740,10791921,0.98999999999999999111);
INSERT INTO tracks VALUES(3323,'Tres Reis',259,1,15,NULL,304143,12168015,0.98999999999999999111);
INSERT INTO tracks VALUES(3324,'Um Lugar ao Sol',259,1,15,NULL,212323,8495217,0.98999999999999999111);
INSERT INTO tracks VALUES(3325,'Batalha Naval',259,1,15,NULL,285727,11431382,0.98999999999999999111);
INSERT INTO tracks VALUES(3326,'Todo o Carnaval tem seu Fim',259,1,15,NULL,237426,9499371,0.98999999999999999111);
INSERT INTO tracks VALUES(3327,'O Misterio do Samba',259,1,15,NULL,226142,9047970,0.98999999999999999111);
INSERT INTO tracks VALUES(3328,'Armadura',259,1,15,NULL,232881,9317533,0.98999999999999999111);
INSERT INTO tracks VALUES(3329,'Na Ladeira',259,1,15,NULL,221570,8865099,0.98999999999999999111);
INSERT INTO tracks VALUES(3330,'Carimbo',259,1,15,NULL,328751,13152314,0.98999999999999999111);
INSERT INTO tracks VALUES(3331,'Catimbo',259,1,15,NULL,254484,10181692,0.98999999999999999111);
INSERT INTO tracks VALUES(3332,'Funk de Bamba',259,1,15,NULL,237322,9495184,0.98999999999999999111);
INSERT INTO tracks VALUES(3333,'Chega no Suingue',259,1,15,NULL,221805,8874509,0.98999999999999999111);
INSERT INTO tracks VALUES(3334,'Mun-Ra',259,1,15,NULL,274651,10988338,0.98999999999999999111);
INSERT INTO tracks VALUES(3335,'Freestyle Love',259,1,15,NULL,318484,12741680,0.98999999999999999111);
INSERT INTO tracks VALUES(3336,'War Pigs',260,4,23,NULL,234013,8052374,0.98999999999999999111);
INSERT INTO tracks VALUES(3337,'Past, Present, and Future',261,3,21,NULL,2492867,490796184,1.9899999999999999911);
INSERT INTO tracks VALUES(3338,'The Beginning of the End',261,3,21,NULL,2611903,526865050,1.9899999999999999911);
INSERT INTO tracks VALUES(3339,'LOST Season 4 Trailer',261,3,21,NULL,112712,20831818,1.9899999999999999911);
INSERT INTO tracks VALUES(3340,'LOST In 8:15',261,3,21,NULL,497163,98460675,1.9899999999999999911);
INSERT INTO tracks VALUES(3341,'Confirmed Dead',261,3,21,NULL,2611986,512168460,1.9899999999999999911);
INSERT INTO tracks VALUES(3342,'The Economist',261,3,21,NULL,2609025,516934914,1.9899999999999999911);
INSERT INTO tracks VALUES(3343,'Eggtown',261,3,19,NULL,2608817,501061240,1.9899999999999999911);
INSERT INTO tracks VALUES(3344,'The Constant',261,3,21,NULL,2611569,520209363,1.9899999999999999911);
INSERT INTO tracks VALUES(3345,'The Other Woman',261,3,21,NULL,2605021,513246663,1.9899999999999999911);
INSERT INTO tracks VALUES(3346,'Ji Yeon',261,3,19,NULL,2588797,506458858,1.9899999999999999911);
INSERT INTO tracks VALUES(3347,'Meet Kevin Johnson',261,3,19,NULL,2612028,504132981,1.9899999999999999911);
INSERT INTO tracks VALUES(3348,'The Shape of Things to Come',261,3,21,NULL,2591299,502284266,1.9899999999999999911);
INSERT INTO tracks VALUES(3349,'Amanda',262,5,2,'Luca Gusella',246503,4011615,0.98999999999999999111);
INSERT INTO tracks VALUES(3350,'Despertar',262,5,2,'Andrea Dulbecco',307385,4821485,0.98999999999999999111);
INSERT INTO tracks VALUES(3351,'Din Din Wo (Little Child)',263,5,16,'Habib Koité',285837,4615841,0.98999999999999999111);
INSERT INTO tracks VALUES(3352,'Distance',264,5,15,'Karsh Kale/Vishal Vaid',327122,5327463,0.98999999999999999111);
INSERT INTO tracks VALUES(3353,'I Guess You''re Right',265,5,1,'Darius "Take One" Minwalla/Jon Auer/Ken Stringfellow/Matt Harris',212044,3453849,0.98999999999999999111);
INSERT INTO tracks VALUES(3354,'I Ka Barra (Your Work)',263,5,16,'Habib Koité',300605,4855457,0.98999999999999999111);
INSERT INTO tracks VALUES(3355,'Love Comes',265,5,1,'Darius "Take One" Minwalla/Jon Auer/Ken Stringfellow/Matt Harris',199923,3240609,0.98999999999999999111);
INSERT INTO tracks VALUES(3356,'Muita Bobeira',266,5,7,'Luciana Souza',172710,2775071,0.98999999999999999111);
INSERT INTO tracks VALUES(3357,'OAM''s Blues',267,5,2,'Aaron Goldberg',266936,4292028,0.98999999999999999111);
INSERT INTO tracks VALUES(3358,'One Step Beyond',264,5,15,'Karsh Kale',366085,6034098,0.98999999999999999111);
INSERT INTO tracks VALUES(3359,'Symphony No. 3 in E-flat major, Op. 55, "Eroica" - Scherzo: Allegro Vivace',268,5,24,'Ludwig van Beethoven',356426,5817216,0.98999999999999999111);
INSERT INTO tracks VALUES(3360,'Something Nice Back Home',261,3,21,NULL,2612779,484711353,1.9899999999999999911);
INSERT INTO tracks VALUES(3361,'Cabin Fever',261,3,21,NULL,2612028,477733942,1.9899999999999999911);
INSERT INTO tracks VALUES(3362,'There''s No Place Like Home, Pt. 1',261,3,21,NULL,2609526,522919189,1.9899999999999999911);
INSERT INTO tracks VALUES(3363,'There''s No Place Like Home, Pt. 2',261,3,21,NULL,2497956,523748920,1.9899999999999999911);
INSERT INTO tracks VALUES(3364,'There''s No Place Like Home, Pt. 3',261,3,21,NULL,2582957,486161766,1.9899999999999999911);
INSERT INTO tracks VALUES(3365,'Say Hello 2 Heaven',269,2,23,NULL,384497,6477217,0.98999999999999999111);
INSERT INTO tracks VALUES(3366,'Reach Down',269,2,23,NULL,672773,11157785,0.98999999999999999111);
INSERT INTO tracks VALUES(3367,'Hunger Strike',269,2,23,NULL,246292,4233212,0.98999999999999999111);
INSERT INTO tracks VALUES(3368,'Pushin Forward Back',269,2,23,NULL,225278,3892066,0.98999999999999999111);
INSERT INTO tracks VALUES(3369,'Call Me a Dog',269,2,23,NULL,304458,5177612,0.98999999999999999111);
INSERT INTO tracks VALUES(3370,'Times of Trouble',269,2,23,NULL,342539,5795951,0.98999999999999999111);
INSERT INTO tracks VALUES(3371,'Wooden Jesus',269,2,23,NULL,250565,4302603,0.98999999999999999111);
INSERT INTO tracks VALUES(3372,'Your Savior',269,2,23,NULL,244226,4199626,0.98999999999999999111);
INSERT INTO tracks VALUES(3373,'Four Walled World',269,2,23,NULL,414474,6964048,0.98999999999999999111);
INSERT INTO tracks VALUES(3374,'All Night Thing',269,2,23,NULL,231803,3997982,0.98999999999999999111);
INSERT INTO tracks VALUES(3375,'No Such Thing',270,2,23,'Chris Cornell',224837,3691272,0.98999999999999999111);
INSERT INTO tracks VALUES(3376,'Poison Eye',270,2,23,'Chris Cornell',237120,3890037,0.98999999999999999111);
INSERT INTO tracks VALUES(3377,'Arms Around Your Love',270,2,23,'Chris Cornell',214016,3516224,0.98999999999999999111);
INSERT INTO tracks VALUES(3378,'Safe and Sound',270,2,23,'Chris Cornell',256764,4207769,0.98999999999999999111);
INSERT INTO tracks VALUES(3379,'She''ll Never Be Your Man',270,2,23,'Chris Cornell',204078,3355715,0.98999999999999999111);
INSERT INTO tracks VALUES(3380,'Ghosts',270,2,23,'Chris Cornell',231547,3799745,0.98999999999999999111);
INSERT INTO tracks VALUES(3381,'Killing Birds',270,2,23,'Chris Cornell',218498,3588776,0.98999999999999999111);
INSERT INTO tracks VALUES(3382,'Billie Jean',270,2,23,'Michael Jackson',281401,4606408,0.98999999999999999111);
INSERT INTO tracks VALUES(3383,'Scar On the Sky',270,2,23,'Chris Cornell',220193,3616618,0.98999999999999999111);
INSERT INTO tracks VALUES(3384,'Your Soul Today',270,2,23,'Chris Cornell',205959,3385722,0.98999999999999999111);
INSERT INTO tracks VALUES(3385,'Finally Forever',270,2,23,'Chris Cornell',217035,3565098,0.98999999999999999111);
INSERT INTO tracks VALUES(3386,'Silence the Voices',270,2,23,'Chris Cornell',267376,4379597,0.98999999999999999111);
INSERT INTO tracks VALUES(3387,'Disappearing Act',270,2,23,'Chris Cornell',273320,4476203,0.98999999999999999111);
INSERT INTO tracks VALUES(3388,'You Know My Name',270,2,23,'Chris Cornell',240255,3940651,0.98999999999999999111);
INSERT INTO tracks VALUES(3389,'Revelations',271,2,23,NULL,252376,4111051,0.98999999999999999111);
INSERT INTO tracks VALUES(3390,'One and the Same',271,2,23,NULL,217732,3559040,0.98999999999999999111);
INSERT INTO tracks VALUES(3391,'Sound of a Gun',271,2,23,NULL,260154,4234990,0.98999999999999999111);
INSERT INTO tracks VALUES(3392,'Until We Fall',271,2,23,NULL,230758,3766605,0.98999999999999999111);
INSERT INTO tracks VALUES(3393,'Original Fire',271,2,23,NULL,218916,3577821,0.98999999999999999111);
INSERT INTO tracks VALUES(3394,'Broken City',271,2,23,NULL,228366,3728955,0.98999999999999999111);
INSERT INTO tracks VALUES(3395,'Somedays',271,2,23,NULL,213831,3497176,0.98999999999999999111);
INSERT INTO tracks VALUES(3396,'Shape of Things to Come',271,2,23,NULL,274597,4465399,0.98999999999999999111);
INSERT INTO tracks VALUES(3397,'Jewel of the Summertime',271,2,23,NULL,233242,3806103,0.98999999999999999111);
INSERT INTO tracks VALUES(3398,'Wide Awake',271,2,23,NULL,266308,4333050,0.98999999999999999111);
INSERT INTO tracks VALUES(3399,'Nothing Left to Say But Goodbye',271,2,23,NULL,213041,3484335,0.98999999999999999111);
INSERT INTO tracks VALUES(3400,'Moth',271,2,23,NULL,298049,4838884,0.98999999999999999111);
INSERT INTO tracks VALUES(3401,'Show Me How to Live (Live at the Quart Festival)',271,2,23,NULL,301974,4901540,0.98999999999999999111);
INSERT INTO tracks VALUES(3402,'Band Members Discuss Tracks from "Revelations"',271,3,23,NULL,294294,61118891,0.98999999999999999111);
INSERT INTO tracks VALUES(3403,'Intoitus: Adorate Deum',272,2,24,'Anonymous',245317,4123531,0.98999999999999999111);
INSERT INTO tracks VALUES(3404,'Miserere mei, Deus',273,2,24,'Gregorio Allegri',501503,8285941,0.98999999999999999111);
INSERT INTO tracks VALUES(3405,'Canon and Gigue in D Major: I. Canon',274,2,24,'Johann Pachelbel',271788,4438393,0.98999999999999999111);
INSERT INTO tracks VALUES(3406,'Concerto No. 1 in E Major, RV 269 "Spring": I. Allegro',275,2,24,'Antonio Vivaldi',199086,3347810,0.98999999999999999111);
INSERT INTO tracks VALUES(3407,'Concerto for 2 Violins in D Minor, BWV 1043: I. Vivace',276,2,24,'Johann Sebastian Bach',193722,3192890,0.98999999999999999111);
INSERT INTO tracks VALUES(3408,'Aria Mit 30 Veränderungen, BWV 988 "Goldberg Variations": Aria',277,2,24,'Johann Sebastian Bach',120463,2081895,0.98999999999999999111);
INSERT INTO tracks VALUES(3409,'Suite for Solo Cello No. 1 in G Major, BWV 1007: I. Prélude',278,2,24,'Johann Sebastian Bach',143288,2315495,0.98999999999999999111);
INSERT INTO tracks VALUES(3410,'The Messiah: Behold, I Tell You a Mystery... The Trumpet Shall Sound',279,2,24,'George Frideric Handel',582029,9553140,0.98999999999999999111);
INSERT INTO tracks VALUES(3411,'Solomon HWV 67: The Arrival of the Queen of Sheba',280,2,24,'George Frideric Handel',197135,3247914,0.98999999999999999111);
INSERT INTO tracks VALUES(3412,'"Eine Kleine Nachtmusik" Serenade In G, K. 525: I. Allegro',281,2,24,'Wolfgang Amadeus Mozart',348971,5760129,0.98999999999999999111);
INSERT INTO tracks VALUES(3413,'Concerto for Clarinet in A Major, K. 622: II. Adagio',282,2,24,'Wolfgang Amadeus Mozart',394482,6474980,0.98999999999999999111);
INSERT INTO tracks VALUES(3414,'Symphony No. 104 in D Major "London": IV. Finale: Spiritoso',283,4,24,'Franz Joseph Haydn',306687,10085867,0.98999999999999999111);
INSERT INTO tracks VALUES(3415,'Symphony No.5 in C Minor: I. Allegro con brio',284,2,24,'Ludwig van Beethoven',392462,6419730,0.98999999999999999111);
INSERT INTO tracks VALUES(3416,'Ave Maria',285,2,24,'Franz Schubert',338243,5605648,0.98999999999999999111);
INSERT INTO tracks VALUES(3417,'Nabucco: Chorus, "Va, Pensiero, Sull''ali Dorate"',286,2,24,'Giuseppe Verdi',274504,4498583,0.98999999999999999111);
INSERT INTO tracks VALUES(3418,'Die Walküre: The Ride of the Valkyries',287,2,24,'Richard Wagner',189008,3114209,0.98999999999999999111);
INSERT INTO tracks VALUES(3419,'Requiem, Op.48: 4. Pie Jesu',288,2,24,'Gabriel Fauré',258924,4314850,0.98999999999999999111);
INSERT INTO tracks VALUES(3420,'The Nutcracker, Op. 71a, Act II: Scene 14: Pas de deux: Dance of the Prince & the Sugar-Plum Fairy',289,2,24,'Peter Ilyich Tchaikovsky',304226,5184289,0.98999999999999999111);
INSERT INTO tracks VALUES(3421,'Nimrod (Adagio) from Variations On an Original Theme, Op. 36 "Enigma"',290,2,24,'Edward Elgar',250031,4124707,0.98999999999999999111);
INSERT INTO tracks VALUES(3422,'Madama Butterfly: Un Bel Dì Vedremo',291,2,24,'Giacomo Puccini',277639,4588197,0.98999999999999999111);
INSERT INTO tracks VALUES(3423,'Jupiter, the Bringer of Jollity',292,2,24,'Gustav Holst',522099,8547876,0.98999999999999999111);
INSERT INTO tracks VALUES(3424,'Turandot, Act III, Nessun dorma!',293,2,24,'Giacomo Puccini',176911,2920890,0.98999999999999999111);
INSERT INTO tracks VALUES(3425,'Adagio for Strings from the String Quartet, Op. 11',294,2,24,'Samuel Barber',596519,9585597,0.98999999999999999111);
INSERT INTO tracks VALUES(3426,'Carmina Burana: O Fortuna',295,2,24,'Carl Orff',156710,2630293,0.98999999999999999111);
INSERT INTO tracks VALUES(3427,'Fanfare for the Common Man',296,2,24,'Aaron Copland',198064,3211245,0.98999999999999999111);
INSERT INTO tracks VALUES(3428,'Branch Closing',251,3,22,NULL,1814855,360331351,1.9899999999999999911);
INSERT INTO tracks VALUES(3429,'The Return',251,3,22,NULL,1705080,343877320,1.9899999999999999911);
INSERT INTO tracks VALUES(3430,'Toccata and Fugue in D Minor, BWV 565: I. Toccata',297,2,24,'Johann Sebastian Bach',153901,2649938,0.98999999999999999111);
INSERT INTO tracks VALUES(3431,'Symphony No.1 in D Major, Op.25 "Classical", Allegro Con Brio',298,2,24,'Sergei Prokofiev',254001,4195542,0.98999999999999999111);
INSERT INTO tracks VALUES(3432,'Scheherazade, Op. 35: I. The Sea and Sindbad''s Ship',299,2,24,'Nikolai Rimsky-Korsakov',545203,8916313,0.98999999999999999111);
INSERT INTO tracks VALUES(3433,'Concerto No.2 in F Major, BWV1047, I. Allegro',300,2,24,'Johann Sebastian Bach',307244,5064553,0.98999999999999999111);
INSERT INTO tracks VALUES(3434,'Concerto for Piano No. 2 in F Minor, Op. 21: II. Larghetto',301,2,24,'Frédéric Chopin',560342,9160082,0.98999999999999999111);
INSERT INTO tracks VALUES(3435,'Cavalleria Rusticana \ Act \ Intermezzo Sinfonico',302,2,24,'Pietro Mascagni',243436,4001276,0.98999999999999999111);
INSERT INTO tracks VALUES(3436,'Karelia Suite, Op.11: 2. Ballade (Tempo Di Menuetto)',303,2,24,'Jean Sibelius',406000,5908455,0.98999999999999999111);
INSERT INTO tracks VALUES(3437,'Piano Sonata No. 14 in C Sharp Minor, Op. 27, No. 2, "Moonlight": I. Adagio sostenuto',304,2,24,'Ludwig van Beethoven',391000,6318740,0.98999999999999999111);
INSERT INTO tracks VALUES(3438,'Fantasia On Greensleeves',280,2,24,'Ralph Vaughan Williams',268066,4513190,0.98999999999999999111);
INSERT INTO tracks VALUES(3439,'Das Lied Von Der Erde, Von Der Jugend',305,2,24,'Gustav Mahler',223583,3700206,0.98999999999999999111);
INSERT INTO tracks VALUES(3440,'Concerto for Cello and Orchestra in E minor, Op. 85: I. Adagio - Moderato',306,2,24,'Edward Elgar',483133,7865479,0.98999999999999999111);
INSERT INTO tracks VALUES(3441,'Two Fanfares for Orchestra: II. Short Ride in a Fast Machine',307,2,24,'John Adams',254930,4310896,0.98999999999999999111);
INSERT INTO tracks VALUES(3442,'Wellington''s Victory or the Battle Symphony, Op.91: 2. Symphony of Triumph',308,2,24,'Ludwig van Beethoven',412000,6965201,0.98999999999999999111);
INSERT INTO tracks VALUES(3443,'Missa Papae Marcelli: Kyrie',309,2,24,'Giovanni Pierluigi da Palestrina',240666,4244149,0.98999999999999999111);
INSERT INTO tracks VALUES(3444,'Romeo et Juliette: No. 11 - Danse des Chevaliers',310,2,24,NULL,275015,4519239,0.98999999999999999111);
INSERT INTO tracks VALUES(3445,'On the Beautiful Blue Danube',311,2,24,'Johann Strauss II',526696,8610225,0.98999999999999999111);
INSERT INTO tracks VALUES(3446,'Symphonie Fantastique, Op. 14: V. Songe d''une nuit du sabbat',312,2,24,'Hector Berlioz',561967,9173344,0.98999999999999999111);
INSERT INTO tracks VALUES(3447,'Carmen: Overture',313,2,24,'Georges Bizet',132932,2189002,0.98999999999999999111);
INSERT INTO tracks VALUES(3448,'Lamentations of Jeremiah, First Set \ Incipit Lamentatio',314,2,24,'Thomas Tallis',69194,1208080,0.98999999999999999111);
INSERT INTO tracks VALUES(3449,'Music for the Royal Fireworks, HWV351 (1749): La Réjouissance',315,2,24,'George Frideric Handel',120000,2193734,0.98999999999999999111);
INSERT INTO tracks VALUES(3450,'Peer Gynt Suite No.1, Op.46: 1. Morning Mood',316,2,24,'Edvard Grieg',253422,4298769,0.98999999999999999111);
INSERT INTO tracks VALUES(3451,'Die Zauberflöte, K.620: "Der Hölle Rache Kocht in Meinem Herze"',317,2,25,'Wolfgang Amadeus Mozart',174813,2861468,0.98999999999999999111);
INSERT INTO tracks VALUES(3452,'SCRIABIN: Prelude in B Major, Op. 11, No. 11',318,4,24,NULL,101293,3819535,0.98999999999999999111);
INSERT INTO tracks VALUES(3453,'Pavan, Lachrimae Antiquae',319,2,24,'John Dowland',253281,4211495,0.98999999999999999111);
INSERT INTO tracks VALUES(3454,'Symphony No. 41 in C Major, K. 551, "Jupiter": IV. Molto allegro',320,2,24,'Wolfgang Amadeus Mozart',362933,6173269,0.98999999999999999111);
INSERT INTO tracks VALUES(3455,'Rehab',321,2,14,NULL,213240,3416878,0.98999999999999999111);
INSERT INTO tracks VALUES(3456,'You Know I''m No Good',321,2,14,NULL,256946,4133694,0.98999999999999999111);
INSERT INTO tracks VALUES(3457,'Me & Mr. Jones',321,2,14,NULL,151706,2449438,0.98999999999999999111);
INSERT INTO tracks VALUES(3458,'Just Friends',321,2,14,NULL,191933,3098906,0.98999999999999999111);
INSERT INTO tracks VALUES(3459,'Back to Black',321,2,14,'Mark Ronson',240320,3852953,0.98999999999999999111);
INSERT INTO tracks VALUES(3460,'Love Is a Losing Game',321,2,14,NULL,154386,2509409,0.98999999999999999111);
INSERT INTO tracks VALUES(3461,'Tears Dry On Their Own',321,2,14,'Nickolas Ashford & Valerie Simpson',185293,2996598,0.98999999999999999111);
INSERT INTO tracks VALUES(3462,'Wake Up Alone',321,2,14,'Paul O''duffy',221413,3576773,0.98999999999999999111);
INSERT INTO tracks VALUES(3463,'Some Unholy War',321,2,14,NULL,141520,2304465,0.98999999999999999111);
INSERT INTO tracks VALUES(3464,'He Can Only Hold Her',321,2,14,'Richard Poindexter & Robert Poindexter',166680,2666531,0.98999999999999999111);
INSERT INTO tracks VALUES(3465,'You Know I''m No Good (feat. Ghostface Killah)',321,2,14,NULL,202320,3260658,0.98999999999999999111);
INSERT INTO tracks VALUES(3466,'Rehab (Hot Chip Remix)',321,2,14,NULL,418293,6670600,0.98999999999999999111);
INSERT INTO tracks VALUES(3467,'Intro / Stronger Than Me',322,2,9,NULL,234200,3832165,0.98999999999999999111);
INSERT INTO tracks VALUES(3468,'You Sent Me Flying / Cherry',322,2,9,NULL,409906,6657517,0.98999999999999999111);
INSERT INTO tracks VALUES(3469,'F**k Me Pumps',322,2,9,'Salaam Remi',200253,3324343,0.98999999999999999111);
INSERT INTO tracks VALUES(3470,'I Heard Love Is Blind',322,2,9,NULL,129666,2190831,0.98999999999999999111);
INSERT INTO tracks VALUES(3471,'(There Is) No Greater Love (Teo Licks)',322,2,9,'Isham Jones & Marty Symes',167933,2773507,0.98999999999999999111);
INSERT INTO tracks VALUES(3472,'In My Bed',322,2,9,'Salaam Remi',315960,5211774,0.98999999999999999111);
INSERT INTO tracks VALUES(3473,'Take the Box',322,2,9,'Luke Smith',199160,3281526,0.98999999999999999111);
INSERT INTO tracks VALUES(3474,'October Song',322,2,9,'Matt Rowe & Stefan Skarbek',204846,3358125,0.98999999999999999111);
INSERT INTO tracks VALUES(3475,'What Is It About Men',322,2,9,'Delroy "Chris" Cooper, Donovan Jackson, Earl Chinna Smith, Felix Howard, Gordon Williams, Luke Smith, Paul Watson & Wilburn Squiddley Cole',209573,3426106,0.98999999999999999111);
INSERT INTO tracks VALUES(3476,'Help Yourself',322,2,9,'Freddy James, Jimmy hogarth & Larry Stock',300884,5029266,0.98999999999999999111);
INSERT INTO tracks VALUES(3477,'Amy Amy Amy (Outro)',322,2,9,'Astor Campbell, Delroy "Chris" Cooper, Donovan Jackson, Dorothy Fields, Earl Chinna Smith, Felix Howard, Gordon Williams, James Moody, Jimmy McHugh, Matt Rowe, Salaam Remi & Stefan Skarbek',663426,10564704,0.98999999999999999111);
INSERT INTO tracks VALUES(3478,'Slowness',323,2,23,NULL,215386,3644793,0.98999999999999999111);
INSERT INTO tracks VALUES(3479,'Prometheus Overture, Op. 43',324,4,24,'Ludwig van Beethoven',339567,10887931,0.98999999999999999111);
INSERT INTO tracks VALUES(3480,'Sonata for Solo Violin: IV: Presto',325,4,24,'Béla Bartók',299350,9785346,0.98999999999999999111);
INSERT INTO tracks VALUES(3481,'A Midsummer Night''s Dream, Op.61 Incidental Music: No.7 Notturno',326,2,24,NULL,387826,6497867,0.98999999999999999111);
INSERT INTO tracks VALUES(3482,'Suite No. 3 in D, BWV 1068: III. Gavotte I & II',327,2,24,'Johann Sebastian Bach',225933,3847164,0.98999999999999999111);
INSERT INTO tracks VALUES(3483,'Concert pour 4 Parties de V**les, H. 545: I. Prelude',328,2,24,'Marc-Antoine Charpentier',110266,1973559,0.98999999999999999111);
INSERT INTO tracks VALUES(3484,'Adios nonino',329,2,24,'Astor Piazzolla',289388,4781384,0.98999999999999999111);
INSERT INTO tracks VALUES(3485,'Symphony No. 3 Op. 36 for Orchestra and Soprano "Symfonia Piesni Zalosnych" \ Lento E Largo - Tranquillissimo',330,2,24,'Henryk Górecki',567494,9273123,0.98999999999999999111);
INSERT INTO tracks VALUES(3486,'Act IV, Symphony',331,2,24,'Henry Purcell',364296,5987695,0.98999999999999999111);
INSERT INTO tracks VALUES(3487,'3 Gymnopédies: No.1 - Lent Et Grave, No.3 - Lent Et Douloureux',332,2,24,'Erik Satie',385506,6458501,0.98999999999999999111);
INSERT INTO tracks VALUES(3488,'Music for the Funeral of Queen Mary: VI. "Thou Knowest, Lord, the Secrets of Our Hearts"',333,2,24,'Henry Purcell',142081,2365930,0.98999999999999999111);
INSERT INTO tracks VALUES(3489,'Symphony No. 2: III. Allegro vivace',334,2,24,'Kurt Weill',376510,6129146,0.98999999999999999111);
INSERT INTO tracks VALUES(3490,'Partita in E Major, BWV 1006A: I. Prelude',335,2,24,'Johann Sebastian Bach',285673,4744929,0.98999999999999999111);
INSERT INTO tracks VALUES(3491,'Le Sacre Du Printemps: I.iv. Spring Rounds',336,2,24,'Igor Stravinsky',234746,4072205,0.98999999999999999111);
INSERT INTO tracks VALUES(3492,'Sing Joyfully',314,2,24,'William Byrd',133768,2256484,0.98999999999999999111);
INSERT INTO tracks VALUES(3493,'Metopes, Op. 29: Calypso',337,2,24,'Karol Szymanowski',333669,5548755,0.98999999999999999111);
INSERT INTO tracks VALUES(3494,'Symphony No. 2, Op. 16 -  "The Four Temperaments": II. Allegro Comodo e Flemmatico',338,2,24,'Carl Nielsen',286998,4834785,0.98999999999999999111);
INSERT INTO tracks VALUES(3495,'24 Caprices, Op. 1, No. 24, for Solo Violin, in A Minor',339,2,24,'Niccolò Paganini',265541,4371533,0.98999999999999999111);
INSERT INTO tracks VALUES(3496,'Étude 1, In C Major - Preludio (Presto) - Liszt',340,4,24,NULL,51780,2229617,0.98999999999999999111);
INSERT INTO tracks VALUES(3497,'Erlkonig, D.328',341,2,24,NULL,261849,4307907,0.98999999999999999111);
INSERT INTO tracks VALUES(3498,'Concerto for Violin, Strings and Continuo in G Major, Op. 3, No. 9: I. Allegro',342,4,24,'Pietro Antonio Locatelli',493573,16454937,0.98999999999999999111);
INSERT INTO tracks VALUES(3499,'Pini Di Roma (Pinien Von Rom) \ I Pini Della Via Appia',343,2,24,NULL,286741,4718950,0.98999999999999999111);
INSERT INTO tracks VALUES(3500,'String Quartet No. 12 in C Minor, D. 703 "Quartettsatz": II. Andante - Allegro assai',344,2,24,'Franz Schubert',139200,2283131,0.98999999999999999111);
INSERT INTO tracks VALUES(3501,'L''orfeo, Act 3, Sinfonia (Orchestra)',345,2,24,'Claudio Monteverdi',66639,1189062,0.98999999999999999111);
INSERT INTO tracks VALUES(3502,'Quintet for Horn, Violin, 2 Violas, and Cello in E Flat Major, K. 407/386c: III. Allegro',346,2,24,'Wolfgang Amadeus Mozart',221331,3665114,0.98999999999999999111);
INSERT INTO tracks VALUES(3503,'Koyaanisqatsi',347,2,10,'Philip Glass',206005,3305164,0.98999999999999999111);
ANALYZE sqlite_master;
INSERT INTO sqlite_stat1 VALUES('tracks','IFK_TrackMediaTypeId','3503 701');
INSERT INTO sqlite_stat1 VALUES('tracks','IFK_TrackGenreId','3503 141');
INSERT INTO sqlite_stat1 VALUES('tracks','IFK_TrackAlbumId','3503 11');
INSERT INTO sqlite_stat1 VALUES('playlist_track','IFK_PlaylistTrackTrackId','8715 3');
INSERT INTO sqlite_stat1 VALUES('playlist_track','sqlite_autoindex_playlist_track_1','8715 623 1');
INSERT INTO sqlite_stat1 VALUES('albums','IFK_AlbumArtistId','347 2');
INSERT INTO sqlite_stat1 VALUES('artists',NULL,'275');
INSERT INTO sqlite_stat1 VALUES('customers','IFK_CustomerSupportRepId','59 20');
INSERT INTO sqlite_stat1 VALUES('playlists',NULL,'18');
INSERT INTO sqlite_stat1 VALUES('employees','IFK_EmployeeReportsTo','8 2');
INSERT INTO sqlite_stat1 VALUES('genres',NULL,'25');
INSERT INTO sqlite_stat1 VALUES('invoices','IFK_InvoiceCustomerId','412 7');
INSERT INTO sqlite_stat1 VALUES('media_types',NULL,'5');
INSERT INTO sqlite_stat1 VALUES('invoice_items','IFK_InvoiceLineTrackId','2240 2');
INSERT INTO sqlite_stat1 VALUES('invoice_items','IFK_InvoiceLineInvoiceId','2240 6');
DELETE FROM sqlite_sequence;
INSERT INTO sqlite_sequence VALUES('genres',25);
INSERT INTO sqlite_sequence VALUES('media_types',5);
INSERT INTO sqlite_sequence VALUES('artists',275);
INSERT INTO sqlite_sequence VALUES('albums',347);
INSERT INTO sqlite_sequence VALUES('tracks',3503);
INSERT INTO sqlite_sequence VALUES('employees',8);
INSERT INTO sqlite_sequence VALUES('customers',59);
INSERT INTO sqlite_sequence VALUES('invoices',412);
INSERT INTO sqlite_sequence VALUES('invoice_items',2240);
INSERT INTO sqlite_sequence VALUES('playlists',18);
CREATE INDEX [IFK_AlbumArtistId] ON "albums" ([ArtistId]);
CREATE INDEX [IFK_CustomerSupportRepId] ON "customers" ([SupportRepId]);
CREATE INDEX [IFK_EmployeeReportsTo] ON "employees" ([ReportsTo]);
CREATE INDEX [IFK_InvoiceCustomerId] ON "invoices" ([CustomerId]);
CREATE INDEX [IFK_InvoiceLineInvoiceId] ON "invoice_items" ([InvoiceId]);
CREATE INDEX [IFK_InvoiceLineTrackId] ON "invoice_items" ([TrackId]);
CREATE INDEX [IFK_PlaylistTrackTrackId] ON "playlist_track" ([TrackId]);
CREATE INDEX [IFK_TrackAlbumId] ON "tracks" ([AlbumId]);
CREATE INDEX [IFK_TrackGenreId] ON "tracks" ([GenreId]);
CREATE INDEX [IFK_TrackMediaTypeId] ON "tracks" ([MediaTypeId]);
COMMIT;
